//Author: Neeraj Agrawal <nirajuec@cs.stanford.edu>

/**
 \file neerajCifarAccuracy.cc
*/

/*Example usage
./bin/neerajCifarAccuracy -e /afs/cs/group/brain/scratch7/paramViz/cifarScripts/neeraj-batch/trial_001/network/encNet_0001400000.dat -s /afs/cs/group/brain/scratch7/paramViz/cifarScripts/neeraj-batch/trial_001/network/supNet_0001400000.dat -i cifar -o cifar
*/


#include <cstdio>
#include <iostream>
#include "tclap/CmdLine.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/stable.h"
#include "sdn/registry.h"

//std::string networkFile;
std::string encNetworkFile;
std::string supNetworkFile;
std::string trainDatasetFile;
std::string testDatasetFile;


bool verbose;
bool debug;


boost::shared_ptr<sdn::ComposableSupervisedNetwork> supervisedNetwork;
boost::shared_ptr<sdn::ComposableAutoEncoderNetwork> encoderNetwork;

boost::shared_ptr<sdn::Dataset> trainDataset;
boost::shared_ptr<sdn::Dataset> testDataset;


void parseCommandLine(int argc, char *argv[]);

int main(int argc, char *argv[])
{


  //Get options from command line
  parseCommandLine(argc, argv);

  //Turn on debugging or verbose information if requested
  if (debug)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_DEBUG);
  else if (verbose)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_VERBOSE);


  //Load encoder  network
  std::ifstream fin;
  fin.open(encNetworkFile.c_str(), std::ios::binary);
  sdn::serial::BinInArchive ar(fin);
  boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
  fin.close();

  encoderNetwork = boost::dynamic_pointer_cast<sdn::ComposableAutoEncoderNetwork> (net);
  if (!encoderNetwork) {
    SDN_ERROR("Could not load compsable auto encoder Network  with name " << encNetworkFile);
    SDN_FATAL("Quitting");
  }


  std::cout<<"Loaded Encoder\n";

  //Load supervised  network
  std::ifstream fin2;
  fin2.open(supNetworkFile.c_str(), std::ios::binary);
  sdn::serial::BinInArchive ar2(fin2);
  boost::shared_ptr<sdn::AbstractNetwork> net2 = sdn::AbstractNetworkRegistry().load(ar2);
  fin2.close();

  supervisedNetwork = boost::dynamic_pointer_cast<sdn::ComposableSupervisedNetwork> (net2);

  if (!supervisedNetwork) {
      SDN_ERROR("Could not load composable supervised Network with name " << supNetworkFile);
      SDN_FATAL("Quitting");
  }


  std::cout<<"Loaded SupervisedNetwork\n";

  //Connect the supervised and autoencoder networks
  supervisedNetwork->useAutoEncoderNetwork(encoderNetwork);



  //Load training dataset
  trainDataset =  sdn::DatasetRegistry().make(trainDatasetFile);
    if (!trainDataset) {
      SDN_ERROR("Could not load dataset with name " << trainDatasetFile);
      SDN_FATAL("Quitting");
    }

    std::cout<<"Loaded training dataset\n";

    int inputHeight = 32;
    int inputWidth = 32;
    trainDataset->requestExampleSize(inputHeight, inputWidth);



    sdn::NormalizeInputBlock normInput;

    sdn::Matrix inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
    sdn::Matrix targetOutput = sdn::Matrix::Zero(trainDataset->numLabels(), 1);
    sdn::Matrix output = sdn::Matrix::Zero(trainDataset->numLabels(), 1);


    /////////////////////////////////////////////////////////////


    //Works fine : output = 1
    std::cout<<"No. of Sup Network blocks:"<<supervisedNetwork->blocks.size()<<"\n";
    

    int label = trainDataset->getRandomExampleAsColumn(inputExample);

    
    // Setup target output
    targetOutput.fill(0.0);
    targetOutput(label, 0) = 1.0;
    
    // Setup input
    normInput.normalizePatch(inputExample);


    std::cout<<"Obtained a training example. Doing forward prop in sup network\n";

    //Works fine: output: input size 1024 1
    //output size : 10 1
    std::cout<<"Input Example size: "<<inputExample.rows()<<"  "<<inputExample.cols()<<"\n\n";
    std::cout<<"Desired Output size: "<<output.rows()<<"  "<<output.cols()<<"\n\n";




    // Transform the input using auto encoder

    //Technically, we should not execute this line.
    //However, executing it is not helping either
    //    encoderNetwork->initTransientParameters();


    //Ouput: 2
    std::cout<<"Blocks: "<<encoderNetwork->blocks.size()<<"\n";

    //Output: 2
    std::cout<<"Intermediate Rep. : "<<encoderNetwork->intermediateRepresentations.size()<<"\n";


    // !!!!!!!!!!!!!!!!! PROBLEM !!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    //This gives output  : 0, 0
    std::cout<<"Intermediate Rep[0]. Rows : "<<encoderNetwork->intermediateRepresentations[0].rows()<<"\n";
    std::cout<<"Intermediate Rep[0]. Cols : "<<encoderNetwork->intermediateRepresentations[0].cols()<<"\n";



    //PROBLEM IN THIS LINE (which is called within encoderNetwork->encode(..) function
    //Gives output error::
    //-*- (stable/blocks/sharedBlocks.cc, 110) values.size() == trackingSize

    encoderNetwork->blocks[0]->forwardProp(inputExample, encoderNetwork->intermediateRepresentations[0]);



    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



    std::cout<<"Block 0 forward prop done\n\n";

    encoderNetwork->encode(inputExample);

    std::cout<<"Passed through encoder\n";

    //   supervisedNetwork->blocks[0]->forwardProp(encodingNetwork->encode(input),
    //                         intermediateRepresentations[0]);




    supervisedNetwork->forwardProp(inputExample, output);

    std::cout<<"Label:"<<label<<"\n"<<output<<"\n";








    ////////////////////////////////////////////////////////////

















    /*
  //Load test dataset
    testDataset =  sdn::DatasetRegistry().make(testDatasetFile);
    if (!testDataset) {
      SDN_ERROR("Could not load dataset with name " << testDatasetFile);
      //  ERROR_OUT_LIST_REGISTRY_NAMES(sdn::Dataset);
      SDN_FATAL("Quitting");
    }

    std::cout<<"Loaded test dataset\n";
    */



    /*
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);


    // set the output vector size
    targetOutput = sdn::Matrix::Zero(dataset->numLabels(), 1);
    */





  /*  
  if (net) {
    
    SDN_MSG( "Loaded a " << net->className() ); 
    
    // Ideally, we will have a functor style code here that accepts
    // any network
    
    // But to just test/demonstrate how things work, we will just
    // assume this is a composable network and look for a 
    // sdn::SparseAffineTanhBlock
    
    boost::shared_ptr<sdn::ComposableAutoEncoderNetwork> cnet = 
      boost::dynamic_pointer_cast<sdn::ComposableAutoEncoderNetwork> (net);
    boost::shared_ptr<sdn::StandardAutoEncoderNetwork> anet = 
      boost::dynamic_pointer_cast<sdn::StandardAutoEncoderNetwork> (net);
    
    sdn::Matrix weights;
      
    SDN_ASSERT_MSG(cnet || anet, "This demo visualization app only works with ComposableAutoEncoderNetworks or StandardAutoEncoderNetwork");
    

    if (cnet) {

      int foundBlock = -1;
      for(int i=0;i<cnet->blocks.size();i++) {
        SDN_MSG ( "Found a " << cnet->blocks[i]->className() << " block!" );
        if (foundBlock < 0 && cnet->blocks[i]->className() == "sdn::SparseAffineTanhBlock")
          foundBlock = i;
      }
      
      SDN_MSG( "Visualizing the " << cnet->blocks[foundBlock]->className() << " at position " << foundBlock );
      
      boost::shared_ptr<sdn::SparseAffineTanhBlock> sath = 
        boost::dynamic_pointer_cast<sdn::SparseAffineTanhBlock> (cnet->blocks[foundBlock]);

      SDN_ASSERT(sath);
      
      weights = sath->transferBlock.affineB.weights;
      
    } else if (anet) {
      
      SDN_MSG( "Visualizing the encoding weights");
      weights = anet->affineEncoder.weights;
      
    }
    
    int patchHeight = sqrt(weights.cols());      // assume square input images
    int displayGridWidth = sqrt(weights.rows()); // try to fit all nicely
    
    sdn::VizWeightMatrixBlock tmpViz;
    tmpViz.vizMatrix(weights, 
                     patchHeight, 
                     displayGridWidth, 
                     "",                         // don't save
                     true);                      // true to showWindow

    //getchar();

  }
  */  
  
  return 0;
}

void parseCommandLine(int argc, char *argv[]) {
  try {
    TCLAP::CmdLine parser("Stanford Deep Network Library training program", ' ', SDN_MAKE_STRING(SDN_VERSION) );

    TCLAP::ValueArg<std::string> encoderNetworkArg("e", "encNetwork", "Enc Network Serialized (binary) file", true, "", "string", parser);
    TCLAP::ValueArg<std::string> supNetworkArg("s", "supNetwork", "Sup Network Serialized (binary) file", true, "", "string", parser);

    TCLAP::ValueArg<std::string> t1("i", "trainDataset", "Train Dataset file", true, "", "string", parser);
    TCLAP::ValueArg<std::string> t2("o", "testDataset", "Test Dataset", true, "", "string", parser);


    
    TCLAP::SwitchArg verboseSwitch("v","verbose","print more information",parser,false);
    TCLAP::SwitchArg debugSwitch("d","debug","print debugging information (also implies -v)",parser,false);
    parser.parse( argc, argv );

    encNetworkFile = encoderNetworkArg.getValue();
    supNetworkFile = supNetworkArg.getValue();
    trainDatasetFile = t1.getValue();
    testDatasetFile = t2.getValue();

    verbose = verboseSwitch.getValue();
    debug = debugSwitch.getValue();


  } catch (TCLAP::ArgException &e) { 
    SDN_ERROR("error: " << e.error() << " for arg " << e.argId()); 
    SDN_FATAL("parseCommandLine failed.");
  }


}



