#include "nnUtil.h"
#include "stackedEncoder.h"
#include <highgui.h>
#include "util.h"
#include "fastEncoder.h"
#include "matrixUtil.h"
#include <fstream>
#include <iomanip>

//Performance todo--
// 1. Test whether caching averagingWeight, testing averagingWeight
//    to see if tanh needs to be recomputed, etc. result in
//    speedup or not (this might be a layer-wise check, since
//    currently our constructors will never make averaging
//    necessary while reconstruncting layer nEncoderLayers-1)
// 2. Find out how to disable Eigen bounds checking
// 3. Use eigen plugin for things like elementwise evaluation of tanh,
//    other places that iterate over whole matrix



USING_PART_OF_NAMESPACE_EIGEN // import most common Eigen types 
using namespace std;
using namespace util;

//------------------------------------------------------------------------------------------


FastEncoder::FastEncoder(int inputWidth, int outputSize,		\
			       double learningRate,  double biasLearningRate, 	\
			       double weightDecay, double desiredSparsityMin,\
			       double desiredSparsityMax, \
			 double sparsityPenaltyMultiplier, double initMag)
  : _nEncoderLayers(1), _nTotLayers(3),	_inputWidth(_nEncoderLayers+1),	\
    _patchStep(_nEncoderLayers+1), _numStep(_nEncoderLayers+1), \
    _outputSize(_nEncoderLayers+1), \
    _learningRate(learningRate), _biasLearningRate(biasLearningRate),	\
    _weightDecay(2), _desiredSparsityMin(2), \
    _desiredSparsityMax(2), \
    _sparsityPenaltyMultiplier(2), \
    _learnAll(false), _errorEstimate(-1),
    _L1weightsDisplay("Layer 1 Weights", inputWidth, inputWidth, 1, 2*outputSize)
{ 
  _deepBackprop = true;

  //Validate inputs
  assert(desiredSparsityMin <= desiredSparsityMax); 

  //Set size specifications

  //Visible layer
  _inputWidth[0] = 1;
  _patchStep[0] = 1;
  _numStep[0] = inputWidth;
  _outputSize[0] = 1;

  //Encoder layer
  _inputWidth[1] = inputWidth;
  _patchStep[1] = inputWidth;
  _numStep[1] = 1;
  _outputSize[1] = outputSize;

  //Assemble all the weights (making no attempt to be efficient;
  //I assume we won't be constructing these in a tight loop)
  
  //Weights out of visible layer
  _weightsOut.push_back(MatrixXd(inputWidth*inputWidth+1, outputSize));
  
  //Weights out of hidden layer
  _weightsOut.push_back(MatrixXd( outputSize+1, inputWidth * inputWidth));

  //Initialize weights to small random values
  for (unsigned i = 0; i < _weightsOut.size(); i++)
    for (unsigned k = 0; k < _weightsOut[i].cols(); k++)
      {
	for (unsigned j = 0; j < _weightsOut[i].rows()-1; j++)
	  _weightsOut[i](j,k) = randomReal(-initMag, initMag);
      }

  //Initialize encoding biases such that if weights were 0 we would be on
  //the upper bound of the desired firing rate
  _weightsOut[0].row(_weightsOut[0].rows()-1).setConstant(0.5*sigmoidInverse(max(_desiredSparsityMax[1], 0.005)));
  
  //Initialize decoding biases to 0
  _weightsOut[1].row(_weightsOut[1].rows()-1).setConstant(0);

  //Set up learning parameters-- this needs to happen before setting up biases
  _weightDecay[1] = weightDecay;
  _desiredSparsityMin[1] = desiredSparsityMin;
  _desiredSparsityMax[1] = desiredSparsityMax;			
  _sparsityPenaltyMultiplier[1] = sparsityPenaltyMultiplier;


  //Set up activations and deltas

  //visible layer
  //input not used
  _input.push_back(MatrixXd::Zero(1,1));
  //deltas not used
  _delta.push_back(MatrixXd::Zero(1,1));
  _deltaFormatted.push_back(MatrixXd::Zero(1,1));
  _deltaPreave.push_back(MatrixXd::Zero(1,1));
  //_rawOutput is used
  _rawOutput.push_back(MatrixXd::Zero(1,inputWidth*inputWidth));
  //_formattedOutput is used
  _formattedOutput.push_back(MatrixXd::Zero(inputWidth*inputWidth+1,1));
  _formattedOutput.back().row(_formattedOutput.back().rows()-1).setConstant(1); //bias constant

  //hidden layer
  _input.push_back(MatrixXd::Zero(outputSize, 1));
  _rawOutput.push_back(MatrixXd::Zero(outputSize+1, 1));
  _rawOutput.back().row(outputSize).setConstant(1); //bias term
  //_formattedOutput is not used, no need to allocate
  _delta.push_back(MatrixXd::Zero(outputSize+1, 1));
  //_deltaPreave unused
  _deltaPreave.push_back(MatrixXd(1,1));
 
  //reconstruction layer
  _input.push_back(MatrixXd::Zero(inputWidth*inputWidth, 1));
  _rawOutput.push_back(MatrixXd::Zero(2,inputWidth*inputWidth));//I put in a row for bias terms for consistency, but don't bother setting it to 1 since it will never be used
  //_formattedOutput is not used, no need to allocate
  //delta is used
  _delta.push_back(MatrixXd::Zero(2, inputWidth*inputWidth));
  //deltaMulti is used
  _deltaPreave.push_back(_input.back());

  //cout << "In constructor: "<<endl;
  //assert(_input.size() == _deltaPreave.size());

  //for (unsigned i = 0; i < _input.size(); i++)
  //cout << _input[i].rows() << " " << _input[i]. cols() << " " << _deltaPreave[i].rows() << " " << _deltaPreave[i].cols() << endl;


  assert(_deltaPreave.back().cols() == _input.back().cols());

  //Set up the sparsity estimates
  //All encoder layers are used (if _learnAll is on, otherwise only last is used)
  for (int i = 0; i <= _nEncoderLayers; i++)
    _sparsityEstimate.push_back(_desiredSparsityMax[i] * MatrixXd::Ones(_rawOutput[i].rows(),_rawOutput[i].cols()));
  
  for (int i = 0; i <= _nEncoderLayers; i++)
    _sparsityPenaltyEstimate.push_back(-1.0);
  
  
  _L1weightsDisplay.setRange(-1,1); 
}

FastEncoder::FastEncoder(int outputSize, \
			       int patchStep, int numStep, \
			       FastEncoder & smaller, bool learnAll,
			       double weightDecay, double desiredSparsityMin, double
			 desiredSparsityMax, double sparsityPenaltyMultiplier, double initMag)
  : _nEncoderLayers(smaller._nEncoderLayers+1), \
    _nTotLayers(smaller._nTotLayers+2), _inputWidth(_nEncoderLayers+1), \
    _patchStep(_nEncoderLayers+1), _numStep(_nEncoderLayers+1), \
    _outputSize(_nEncoderLayers+1), 
    _weightsOut(smaller._weightsOut),	\
    _learningRate(smaller._learningRate), \
    _biasLearningRate(smaller._learningRate), \
    _weightDecay(smaller._weightDecay), \
    _desiredSparsityMin(smaller._desiredSparsityMin),	      \
    _desiredSparsityMax(smaller._desiredSparsityMax), \
_sparsityPenaltyMultiplier(smaller._sparsityPenaltyMultiplier), \
    _learnAll(learnAll), _errorEstimate(0),
    _L1weightsDisplay("Layer 1 Weights", smaller._inputWidth[1], smaller._inputWidth[1], 1, smaller._outputSize[1]*2)
{ 
  _deepBackprop = true;

  //Validate inputs
  assert(desiredSparsityMin <= desiredSparsityMax); 

  //Set size specifications

  //Copy over old values
  for (int i = 0; i < _nEncoderLayers; i++)
    {
      _inputWidth[i] = smaller._inputWidth[i];
      _patchStep[i] = smaller._patchStep[i];
      _numStep[i] = smaller._numStep[i];
      _outputSize[i] = smaller._outputSize[i];
    }
  
  //Values for innermost code layer-- there is only one instantiation
  //of it, and it is connected to the entirety of the next layer down
  //(Note: this is the ONLY place in the code that assumes the innermost
  //layer is not convolutional. By writing a new constructor, we could
  //train the smallest network as convolutional from the start)
  _inputWidth[_nEncoderLayers] = numStep;
  _outputSize[_nEncoderLayers] = outputSize;
  _patchStep[_nEncoderLayers] = _inputWidth[_nEncoderLayers];
  _numStep[_nEncoderLayers] = 1;

  //_inputWidth and _outputWidth don't change for the intermediate layers,
  //but we do need to update their _numStep and _patchStep

  assert(_nEncoderLayers >= 2);
  
  //Second-deepest encoder layer
  _patchStep[_nEncoderLayers-1] = patchStep;
  _numStep[_nEncoderLayers-1] = numStep;

  
  //All other encoder layers
  for (int i = _nEncoderLayers-2; i >= 0; i--)
    {
      //we want ns[i] = (ns[i+1]-1)*ps[i+1]*iw[i+1]
      _numStep[i] = ((_numStep[i+1]-1)*_patchStep[i+1]+_inputWidth[i+1]);
      //cout << "ns "<<i<<" "<<_numStep[i]<<endl;
    }

  //Visible layer
  _numStep[0] = (_numStep[1]-1)*_patchStep[1]+_inputWidth[1];

  for (int i = 0; i < _nEncoderLayers; i++)
    assert( _numStep[i] == ( _numStep[i+1] -1)*_patchStep[i+1] + _inputWidth[i+1]);

  //Assemble all the weights (making no attempt to be efficient;
  //I assume we won't be constructing these in a tight loop)

  //Note: v.insert(v.begin()+i, elem) results in v[i] == elem

  //Add weights into new hidden layer
  int iw = numStep;
  assert(iw > 0);

  _weightsOut.insert(_weightsOut.begin()+_nEncoderLayers-1, \
		     MatrixXd(iw*iw*_outputSize[_nEncoderLayers-1]+1,outputSize));



  //Add weights out of new hidden layer
  _weightsOut.insert(_weightsOut.begin()+_nEncoderLayers, \
		     MatrixXd(outputSize+1, _inputWidth[_nEncoderLayers]*_inputWidth[_nEncoderLayers]*_outputSize[_nEncoderLayers-1] ));


  //Initialize weights to small random values
  for (unsigned i = _nEncoderLayers-1; i <= _nEncoderLayers; i++)
    for (unsigned k = 0; k < _weightsOut[i].cols(); k++)
      for (unsigned j = 0; j < _weightsOut[i].rows()-1; j++)
	{
	  _weightsOut[i](j,k) = randomReal(-initMag, initMag);
	}

  //Initialize encoding biases such that if weights were 0 we would be on
  //the upper bound of the desired firing rate
  _weightsOut[_nEncoderLayers-1].row(_weightsOut[_nEncoderLayers-1].rows()-1).setConstant( 0.5*sigmoidInverse(max(_desiredSparsityMax[1], 0.005)));

  //Initialize decoding biases to 0
  _weightsOut[_nEncoderLayers].row(_weightsOut[_nEncoderLayers].rows()-1).setConstant(0);

  //Set up learning parameters-- this needs to happen before setting up biases
  _weightDecay.push_back(weightDecay);
  _desiredSparsityMin.push_back(desiredSparsityMin);
  _desiredSparsityMax.push_back(desiredSparsityMax);			
  _sparsityPenaltyMultiplier.push_back(sparsityPenaltyMultiplier);



  //Set up activations and deltas
  allocateTemporaries();

  _L1weightsDisplay.setRange(-1,1); 

}

void FastEncoder::allocateTemporaries()
{
  _input.clear();
  _delta.clear();
  _deltaPreave.clear();
  _rawOutput.clear();
  _formattedOutput.clear();
  _sparsityEstimate.clear();
  _sparsityPenaltyEstimate.clear();

//visible layer
  //input is unused
  _input.push_back(MatrixXd::Zero(1,1));
  //deltas unused
  _delta.push_back(MatrixXd::Zero(1,1));
  _deltaPreave.push_back(MatrixXd::Zero(1,1));
  _deltaFormatted.push_back(MatrixXd::Zero(1,1));
  //outputs are used, but will be handled later

  //outputs for visible layer and all but last encoder layer
  for (int i = 0; i <= _nEncoderLayers-1; i++)
    {
      _rawOutput.push_back(MatrixXd::Zero(_outputSize[i],_numStep[i]*_numStep[i]));
      _formattedOutput.push_back(MatrixXd::Zero(_inputWidth[i+1]*_inputWidth[i+1]*_outputSize[i]+1, _numStep[i+1]*_numStep[i+1]));
      _formattedOutput.back().row(_formattedOutput.back().rows()-1).setConstant(1);
    }

  //output for final encoder layer
  _rawOutput.push_back(MatrixXd::Zero(_outputSize[_nEncoderLayers]+1, _numStep[_nEncoderLayers]*_numStep[_nEncoderLayers]));
  _rawOutput.back().row(_outputSize[_nEncoderLayers]).setConstant(1); //bias term
  //formattedOutput is not used

  //raw output for decoder layers
  for (int i = _nEncoderLayers-1; i >= 0; i--)
    {
      _rawOutput.push_back(MatrixXd(_rawOutput[i].rows()+1, _rawOutput[i].cols()));
      _rawOutput.back().row(_rawOutput.back().rows()-1).setConstant(1); //bias constant
      //_formattedOutput is not used
    }
  assert(_rawOutput.size() == _nTotLayers);
    

  //inputs for encoder layers
  for (int i = 1; i <= _nEncoderLayers; i++)
    {
      _input.push_back(MatrixXd::Zero(_outputSize[i],_numStep[i]*_numStep[i]));
      _delta.push_back(_rawOutput[i]);
      if (i < _nEncoderLayers)
	_deltaFormatted.push_back(_formattedOutput[i]);
      _deltaPreave.push_back(MatrixXd(1,1)); //unused
    }

  int encoderLayer = _nEncoderLayers;
  //activationsIn and deltas for decoder layers
  for (int i = _nEncoderLayers+1; i < _nTotLayers; i++)
    {
      _input.push_back(MatrixXd::Zero(_inputWidth[encoderLayer]*_inputWidth[encoderLayer]*_outputSize[encoderLayer-1], _numStep[encoderLayer]*_numStep[encoderLayer]));


      _delta.push_back(_rawOutput[i]);
      _deltaPreave.push_back(_input.back());

      encoderLayer--;
    } 


  //cout << "In constructor: "<<endl;
  //assert(_input.size() == _deltaPreave.size());

  //for (unsigned i = 0; i < _input.size(); i++)
  //cout << _input[i].rows() << " " << _input[i]. cols() << " " << _deltaPreave[i].rows() << " " << _deltaPreave[i].cols() << endl;


  //Set up the sparsity estimates

  //Vis layer is unused
  _sparsityEstimate.push_back(MatrixXd::Zero(1,1));
  //All encoder layers are used (if _learnAll is on, otherwise only last is used)
  for (int i = 1; i <= _nEncoderLayers; i++)
    _sparsityEstimate.push_back(_desiredSparsityMax[i] * MatrixXd::Ones(_rawOutput[i].rows(),_rawOutput[i].cols()));

  for (int i = 0; i <= _nEncoderLayers; i++)
    _sparsityPenaltyEstimate.push_back(-1.0);

  _errorEstimate = -1.0;
}

void FastEncoder::forwardProp(const Eigen::MatrixXd * patch, double * errorEstimate, std::vector<vector< double> > * sparsityStats,
				 std::vector<double> * sparsityPenaltyEstimate, int startLayer, int endLayer)
{

  if (endLayer == -1)
    endLayer = _nTotLayers -1;

  assert(startLayer < _nTotLayers - 1);

  //Check size of input
  //cout << _numStep[0] << endl;
  if (patch)
  {
	assert(startLayer == 0);
  	assert(patch->rows() == _numStep[0]);
  	assert(patch->cols() == _numStep[0]);
 
	//Input provides activation output for first layer
  	//image to be vectorized
  	matrixToMatrixRow(*patch, _rawOutput[0]);
  }

  //Do forward prop through encoding layers
  for (int i = startLayer; i <= min(_nEncoderLayers-1,endLayer-1); i++)
    {

      //	cout << "Making formatted output for layer "<<i<<endl;

      //make formattedOutput[i]
      int col_idx = 0; //index of the column of _formattedOutput we're populating.
      //(each column corresponds to a different hidden unit group's input)
      int row_stride = _outputSize[i];
      int ns = _numStep[i+1];
      int im_width = _numStep[i];
      int im_stride = _patchStep[i+1];
      int im_depth = _outputSize[i];
      int iw = _inputWidth[i+1];
      int formatted_stride = im_depth * iw;
      //matrix to vector uses column major format
      //so when we map the hidden unit groups of layer i+1 into column 
      //indices of _formattedOutput[i], we should step over the groups
      //in column major order 

      int first_im_col = 0;
      for (int k = 0; k < ns; k++)
	{
	  int first_im_row = 0;
	  for (int j = 0; j < ns; j++)
	    {
	      int row_idx = 0;//index into _formattedOutput

	      //each column of _rawOutput is like a different pixel of
	      //the input. each row of _rawOutput is like a different
	      //input channel.

	      //when we linearized the pixels into _rawOutput column
	      //indices, we did so in a column major way

	      int im_ind = first_im_col * im_width + first_im_row;

	      for (int l = 0; l < iw; l++)
		{
		  int im_col = first_im_col + l;
		  
		  matrixToMatrixSubCol(_rawOutput[i].block(0, im_ind, im_depth, iw), _formattedOutput[i], row_idx, col_idx);

		  im_ind += im_width;//advance to same position in next column
		  row_idx += formatted_stride;
		}


	      col_idx += 1;
	      first_im_row += im_stride;
	    }
	  first_im_col += im_stride;
	}
      
      
      std::ofstream testFile("testPatch", ios::out | ios::binary);
      if (!testFile.is_open()) {
	std::cerr << "FastEncoder::saveNN: Tried to open " << "testOutput" << "\n";
	error("FastEncoder::saveNN: Error opening file.");
      }
      //saveMatrix(testPatch,_formattedOutput[i]);
      testFile << patch;
      testFile.flush();

      //todo-- disable this check for added speed
      for (int j = 0; j < _formattedOutput[i].cols(); j++)
	{
	  assert( _formattedOutput[i](_formattedOutput[i].rows()-1, j) == 1);
	}

      /* a quick sanity check, which I think only applies in the 1 hidden layer case
      for (int j = 0; j < _numStep[i]*_numStep[i]; j++)
	assert(_rawOutput[i](0,j) == _formattedOutput[i](j,0));
      */

      //cout << "fe formatted output "<<i<<endl<<_formattedOutput[i]<<endl;
      
      //cout << "fe weightsOut "<<i<<endl<<_weightsOut[i]<<endl;

      assert(i < _weightsOut[i].size());
      assert(_weightsOut[i].rows() == _formattedOutput[i].rows());
      //cout << _weightsOut[i].transpose() << endl;
      //cout << _formattedOutput[i] << endl;

      _input[i+1] = _weightsOut[i].transpose() * _formattedOutput[i];

      //cout << "True size of input "<<i+1<<": "<<_input[i+1].rows()<<" "<<_input[i+1].cols()<<endl;



      //cout << "fe input "<<i+1<<endl;
      //cout << _input[i+1] << endl;

      //make _rawOutput[i+1];
      for (int j = 0; j < _input[i+1].rows(); j++)//note that on the last layer, this will leave the last row of _rawOutput untouched, so as not to overwrite the bias term
	for (int k = 0; k < _input[i+1].cols(); k++)
	  _rawOutput[i+1](j,k) = tanh(_input[i+1](j,k));

      //cout << "fe raw output "<<i+1<<endl << _rawOutput[i+1] << endl;

    }//end of loop over i, layers

  int encoderLayer = _nEncoderLayers;

  if (startLayer > _nEncoderLayers)
    encoderLayer = _nEncoderLayers - (startLayer-_nEncoderLayers);





  //cout << "fe _weightsOut " << _nEncoderLayers << endl << _weightsOut[_nEncoderLayers] << endl;

  //Do forward prop through decoding layers and output layer
  for (int i = max(_nEncoderLayers, startLayer); i <= min(endLayer-1,_nTotLayers-2); i++)
    { 
	//cout << "Mulitplying by _rawOutput "<<i<<endl;
        //cout << _rawOutput[i].block(0,0,_rawOutput[i].rows()-1,_rawOutput[i].cols()) << endl;	

      //todo-- disable this check for added speed
      for (int j = 0; j < _rawOutput[i].cols(); j++)
	{
	  assert( _rawOutput[i](_rawOutput[i].rows()-1, j) == 1);
	}

      _input[i+1] = _weightsOut[i].transpose() * _rawOutput[i];
      
      //cout << "True size of input "<<i+1<<": "<<_input[i+1].rows()<<" "<<_input[i+1].cols()<<endl;


      //cout << "fastEncoder input "<<i+1<<": "<<endl<<_input[i+1]<<endl;

      //make _rawOutput[i+1] -- this will require both tanh and averaging
      //NOTE-- this step here consumes a full third of forward prop's runtime!
      //apply tanh and sum up all groups' contributions to the reconstruction of
      //each pixel
      //the mapping here is inverse of the rawOutput->formatted mapping
      //i.e., _input[i+1] is the same size as _formattedOutput[encoderLayer-1]
      assert(_input[i+1].cols() == _formattedOutput[encoderLayer-1].cols());
      assert(_input[i+1].rows() == _formattedOutput[encoderLayer-1].rows()-1);

      //we can do this by repeating the code for the raw->formatted transformation,
      //except turning the matrixToMatrixSubCol operation into an operation
      //that adds tanh of the MatrixSubCol to the matrix that used to be the input

      _rawOutput[i+1].block(0,0,_rawOutput[i+1].rows()-1,_rawOutput[i+1].cols()).setConstant(0); //todo-- might add a check to see if averaging happens on this layer. if not, rather than zeroing followed by summing followed by dividing, we could just do the summing step and replace the += with = 

      int col_idx = 0;
      int row_stride = _outputSize[encoderLayer-1];
      int ns = _numStep[encoderLayer];
      int im_width = _numStep[encoderLayer-1];
      int im_stride = _patchStep[encoderLayer];
      int im_depth = _outputSize[encoderLayer-1];
      int iw = _inputWidth[encoderLayer];
      int input_stride = im_depth * iw;

      #if 0

      //Please don't delete this, just use the #if to turn it off
      //This hack lets you see what the reconstructions from each group look like before averaging

      assert(_nEncoderLayers == 2);

      if (i+1 == 4)
	{
	  cout << "Example reconstruction" << endl;
	 
	  PatchDisplay inputDisplay("Input", patch.cols(), patch.rows(), 2, 1, 1);
	  inputDisplay.addPatch(patch);

	  PatchDisplay reconstructionDisplay("Reconstructions", _inputWidth[1], _inputWidth[1], 2, _numStep[1], _numStep[1]);

	  for (int j = 0; j < _numStep[1]; j++)
	    {
	      for (int k = 0; k < _numStep[1]; k++)
		{
		  int idx = j * _numStep[1]+k;
		  
		  MatrixXd r(_inputWidth[1],_inputWidth[1]);

		  for (int l = 0; l < r.rows(); l++)
		    {
		      for (int m = 0; m < r.cols(); m++)
			{
			  if (m*r.rows()+l >= _input[i+1].rows())
			    {
			      cout << "m: "<<m<<endl;
			      cout << "r.rows(): "<<r.rows()<<endl;
			      cout << "l: "<<l<<endl;
			      cout << "m*r.rows()+l: "<<m*r.rows()+l<<endl;
			      cout << "_input[i+1].rows(): "<<_input[i+1].rows()<<endl;
			      cout << "i+1: "<<i+1<<endl;
			    }
			  if (idx >= _input[i+1].cols())
			    {
			      cout << "idx: "<<idx<<endl;
			      cout << "_input[i+1].cols(): "<<_input[i+1].cols()<<endl;
			      cout << "i+1: "<<i+1<<endl;
			    }

			  r(l,m) = tanh(_input[i+1](m*r.rows()+l, idx));
			}
		    }

		  reconstructionDisplay.addPatch(r);
		}
	    }

	  inputDisplay.enableWindow();
	  reconstructionDisplay.enableWindow();

	  cvWaitKey(0);

	}
      

#endif
	  




      int first_im_col = 0;
      for (int k = 0; k < ns; k++)
	{
	  int first_im_row = 0;
	  for (int j = 0; j < ns; j++)
	    {
	      int row_idx = 0; //index into _input
	      

	      int im_ind = first_im_col * im_width + first_im_row;

	      for (int l = 0; l < iw; l++)
		{
		  int im_col = first_im_col + l;

		  //here is where we replace teh matrixToMatrixSubCol call
		  //we want to take tanh of input_stride elements of column col_idx
		  //of input, starting at row row_idx
		  //and add these, using a column major mapping, to the
		  //the matrix block of _rawOutput starting at (0,im_ind) with
		  //height im_depth and width iw

		  int offset = 0;

		  for (int c = 0; c < iw; c++)
		    {
		      for (int r = 0; r < im_depth; r++)
			{
			  _rawOutput[i+1](r,im_ind+c) += tanh(_input[i+1](row_idx+offset,col_idx));

			  offset++;
			}
		    }

		  im_ind += im_width;
		  row_idx += input_stride;
		}
	      
	      col_idx += 1;
	      first_im_row += im_stride;
	    }
	  first_im_col += im_stride;
	}


      //we've now added up all the outputs. now got to divide by number of overlaps at each pixel
      int r = _rawOutput[i+1].rows()-1; //don't overwrite the bias term
      int c = _rawOutput[i+1].cols();
      int ps = _patchStep[encoderLayer];

      //_rawOutput[i+1].setConstant(1);

      for (int k = 0; k < c; k++) //coordinate within linearized image
	{
	  double aw = averagingWeight(k%im_width,k/im_width,ps,iw,ns);//using % and / to convert from 1D to 2D image coordinates

	  //cout << "Averaging weight: " << aw << endl;

	  for (int j = 0; j < r; j++)
	    {
	      _rawOutput[i+1](j,k) *= aw;
	    }
	}


      #if 0

      assert(_nEncoderLayers == 2);

      if (i+1 == 3)
	{
	  _rawOutput[i+1].block(0,0,_rawOutput[i+1].rows()-1,_rawOutput[i+1].cols()) = _rawOutput[1];
	  cout << "HACK: 2 layer network is actually just echoing L1 to L1 reconstruction (this lets you see what convolved L1 units code for)" << endl;
	}
      #endif

      //cout << "fe _rawOutput "<<i+1<<endl<<_rawOutput[i+1]<<endl;
      
      //if (i+1 < _weightsOut.size())
      //cout << "_weightsOut "<<i+1<<endl<<_weightsOut[i+1]<<endl;

      encoderLayer--;
    }



#if 0
  PatchDisplay p("Final reconstruction",inputWidth(), inputWidth(), 1, 1, 1);
  p.addPatch(vectorToMatrix(_rawOutput.back().row(0),inputWidth(),inputWidth()));
  p.enableWindow();
  cvWaitKey(0);
#endif

#if 0
  static int counter = 0;
  if (counter++ % 100 == 0)
    {
      int lo = _nEncoderLayers - 1;
      int lr = _nEncoderLayers + 1;
      //cout << "lo "<<_rawOutput[lo].rows()<<" "<<_rawOutput[lo].cols()<<endl;
      //cout << "lr "<<_rawOutput[lr].rows()<<" "<<_rawOutput[lr].cols()<<endl;
      PatchDisplay q("Intermediate rep and reconstruction",_rawOutput[lo].rows(),_rawOutput[lo].cols(), 3, 1, 2);
      q.addPatch(_rawOutput[lo].transpose());
      q.addPatch(_rawOutput[lr].block(0,0,_rawOutput[lr].rows()-1,_rawOutput[lr].cols()).transpose());
      q.enableWindow();
      cvWaitKey(0);
    }
#endif


  //It doesn't make sense to update the sparsity estimates if not driven by data
  //There's no way to estimate reconstruction error if there is no input
  if (startLayer != 0)
    return;

  //todo-- add option to prevent sparsity/error updates to save time, i.e. for caltech or invariance tests

  int first = _nEncoderLayers;
  int last = _nEncoderLayers;

  if (_learnAll)
    first = 1;

  if (sparsityPenaltyEstimate)
    sparsityPenaltyEstimate->resize(_nEncoderLayers+1);

  if (sparsityStats)
    sparsityStats->resize(_nEncoderLayers+1);

  for (int i = first; i <= last; i++)
    {
      _sparsityEstimate[i] = 0.999 * _sparsityEstimate[i] + 0.001 * \
	(_rawOutput[i].cwise()+1.0)/2.0;
      
      // cout << "_sparsityEstimate["<<i<<"](0,0): "<<_sparsityEstimate[i](0,0) <<endl;

      //Note-- autoencoder class doesn't actually compute a sparsity
      //penalty estimate, so I don't do so here either
      if (sparsityPenaltyEstimate)
	(*sparsityPenaltyEstimate)[i] = _sparsityPenaltyEstimate[i];

      if (sparsityStats)
	{
	  //Note-- this will report sparsity estimates for individual neurons,
	  //the gradients from these estimates are pooled when updating
	  //weights that are shared between _numStep x _numStep neurons
	  

	  int nHid = _sparsityEstimate[i].rows() * _sparsityEstimate[i].cols();

	  if (i == _nEncoderLayers)
	    nHid -= _sparsityEstimate[i].cols(); //the last row of the final encoder layer is the bias term

	  double * temp = new double[nHid];//todo-- make this a cached structure so it doesn't always need to be reallocated

	  if (i != _nEncoderLayers)
	    {
	      //This is safe because we don't care about the order of the units;
	      //we're about to sort them anyway
	      memcpy(temp, _sparsityEstimate[i].data(), nHid*sizeof(double));
	    }
	  else
	    {
	      matrixBlockToCArr(_sparsityEstimate[i], 0,0,_sparsityEstimate[i].rows()-1,_sparsityEstimate[i].cols(), temp);
	    }
	  
	  qsort_double(temp, nHid);

	  //cout << "nnHid\t\tnnSort" << endl;
	  //for (int j = 0; j < nHid; j++)
	  //cout << _sparsityEstimate[i].data()[j] << "\t\t" << temp[j] << endl;

	  (*sparsityStats)[i].resize(6);

	  (*sparsityStats)[i][0] = mean_in_arr(temp, nHid);
	  (*sparsityStats)[i][1] = temp[(int)round(nHid*0.1)]; // 10% percentile
	  (*sparsityStats)[i][2] = temp[(int)round(nHid*0.25)]; // 25% percentile
	  (*sparsityStats)[i][3] = temp[(nHid==1)?0:(int)round(nHid*0.5)]; // median]; // median
	  (*sparsityStats)[i][4] = temp[nHid-1-(int)round(nHid*0.25)]; // 75% percentile
	  (*sparsityStats)[i][5] = temp[nHid-1-(int)round(nHid*0.1)]; // 90% percentile	  					    
	  delete [] temp;
	}
    }

  //Estimate errors  //TODO-- move this to backprop?
  double thisError = 0;

  for (int j = 0; j < _rawOutput[0].cols(); j++)
      {
	//assert(!isinf(_activationOut[0](i,j)));
	//assert(!isinf(_activationOut[_nTotLayers-1](i,j)));
	thisError += sqr(_rawOutput[0](0,j) - _rawOutput[_nTotLayers-1](0,j));
      }

  //assert(!isnan(thisError));
  //assert(!isinf(thisError));


  if (_errorEstimate < 0)
    _errorEstimate = thisError;
  else
    _errorEstimate = 0.9999 * _errorEstimate + 0.0001 * thisError;

  if (errorEstimate)
    *errorEstimate = _errorEstimate;



}

void FastEncoder::backProp()
{
  //todo-- make this function work in the new setup
  
  //TODO-- if onyl learning with _deepBackprop=false, then it isn't
  //necessary to do full forwardProp. Just be careful not to break the
  //visualization code.

  int objectiveLayer = _nTotLayers - 1;
  int firstLearnableLayer = 0;
  int ref = 0;
  if (!_deepBackprop)
    {
      objectiveLayer = _nEncoderLayers+1;
      firstLearnableLayer = _nEncoderLayers-1;
      ref = _nEncoderLayers -1;
    }


  //Compute deltas for imaginary "averaging units" of output layer
  int idx = objectiveLayer;
  for (int i = 0; i < _rawOutput[idx].rows()-1; i++)
    {
      for (int j = 0; j < _rawOutput[idx].cols(); j++)
	{
	  _delta[idx](i, j) =		     \
	    (_rawOutput[idx](i,j) -		     \
	     _rawOutput[ref](i, j));
	  //cout << "output delta: "<<_delta[idx](i,j) << endl;
	  //cout << "output: "<<_activationOut[idx](i,j) << endl;
	  //cout << "input: "<<_activationOut[0](i,j) << endl;
	}
    }


  //cout << "fe _delta "<<idx << endl << _delta[idx] << endl;

	
  int encodingLayer = 1;

  if (!_deepBackprop)
    encodingLayer = _nEncoderLayers;

  
  //Compute deltas for layers whose output is averaged
  //This is basically just copying the code that does the summing from forward prop

  //(Note-- by output I mean output of the decoding layer, and input
  //I mean input to the decoding layer from a higher code layer)
  //cout << "Enter i loop"<<endl;
  for (int i = objectiveLayer - 1 ; i >= _nEncoderLayers; i--)
    {
      int e = encoderLayer(i);

      //Make _deltaPreave for layer i+1
      int col_idx = 0;
      int row_stride = _outputSize[e-1];
      int ns = _numStep[e];
      int im_width = _numStep[e-1];
      int im_stride = _patchStep[e];
      int im_depth = _outputSize[e-1];
      int iw = _inputWidth[e];
      int input_stride = im_depth * iw;
      int ps = _patchStep[e];

      int first_im_col = 0;
      for (int k = 0; k < ns; k++)
	{
	  int first_im_row = 0;
	  for (int j = 0; j < ns; j++)
	    {
	      int row_idx = 0; //index into _input
	      
	      int im_ind = first_im_col * im_width + first_im_row;
	      
	      for (int l = 0; l < iw; l++)
		{
		  int im_col = first_im_col + 1;

		  int offset = 0;

		  for (int c = 0; c < iw; c++)
		    {
		      int i_p_c = im_ind + c;
		      double aw = averagingWeight(i_p_c % im_width, i_p_c / im_width, ps,iw,ns);

		      for (int r = 0; r < im_depth; r++)
			{			  
			  _deltaPreave[i+1](row_idx+offset, col_idx) = aw * (1.0-sqr(tanh(_input[i+1](row_idx+offset,col_idx)))) \
			    * _delta[i+1](r, i_p_c);		       
			  
			  offset++;
			}
		    }
		
		  im_ind += im_width;
		  row_idx += input_stride;
		}

	      col_idx += 1;
	      first_im_row += im_stride;
	    }
	  first_im_col += im_stride;
	}

      //cout << "fe _deltaPreave "<<i+1<<endl<<_deltaPreave[i+1] << endl;


      //Make deltas for the averaged units on this layer
      _delta[i] = _weightsOut[i] * _deltaPreave[i+1];

      if (i == _nEncoderLayers)
	{
	  _delta[i] = _delta[i].cwise() * ((-1.0*(_rawOutput[i].cwise().square())).cwise()+1);
	  //todo now-- rather than * _rawOutput, this needs to be * (1.0-sqr(_rawOutput))
	}

      //cout << "fe _delta "<<i<<endl<<_delta[i]<<endl;
    }
 
  int lastLayer  = _nEncoderLayers-1;
  
  if (lastLayer < 1)
    lastLayer = 1;

  if (_learnAll)
    lastLayer = 1;

  if (firstLearnableLayer > lastLayer)
    lastLayer = firstLearnableLayer;

  //Iterate over all encoding layers whose weights we are learning
  for (int i = _nEncoderLayers-1; i >= lastLayer; i--)
    {
      if (i < _nEncoderLayers -1)
	_deltaFormatted[i] = _weightsOut[i] * _delta[i+1];
      else //delta[_nEncoderLayers] has an extra useless row because _rawOutput has the bias term there
	_deltaFormatted[i] = _weightsOut[i] * _delta[i+1].block(0,0,_delta[i+1].rows()-1,_delta[i+1].cols());


      //convert from deltaFormatted to delta
      int col_idx = 0; //index of the column of _formattedOutput we're reading back out
      int row_stride = _outputSize[i];
      int ns = _numStep[i+1];
      int im_width = _numStep[i];
      int im_stride = _patchStep[i+1];
      int im_depth = _outputSize[i];
      int iw = _inputWidth[i+1];
      int formatted_stride = im_depth * iw;

      int first_im_col = 0;
      for (int k = 0; k < ns; k++)
	{
	  int first_im_row = 0;
	  for (int j = 0; j < ns; j++)
	    {
	      int row_idx = 0;

	      int im_ind = first_im_col * im_width + first_im_row;
	      
	      for (int l = 0; l < iw; l++)
		{
		  int im_col = first_im_col + 1;
		  
		  matrixSubColToMatrixBlock(_deltaFormatted[i], row_idx, col_idx, _delta[i], 0, im_ind, im_depth, iw);

		  im_ind += im_width;
		  row_idx += formatted_stride;
		}

	      col_idx += 1;
	      first_im_row += im_stride;
	    }
	  first_im_col += im_stride;
	}
      
      //now that we have _delta de-formmatted, we can associate it with _rawOutput
      _delta[i] = _delta[i].cwise() * ((-1.0*(_rawOutput[i].cwise().square())).cwise()+1.0);


      //     cout << "fe _delta "<<i<<endl<<_delta[i]<<endl;
    }
}

void FastEncoder::updateWeights()
{
  int first = _nEncoderLayers-1;
  int last = _nEncoderLayers;

  if (_learnAll && _deepBackprop)
    {
      first = 0;
      last = _nTotLayers-2;
    }

  for (int i = first; i <= last; i++)
    {
      int encoderLayer;

      if (i < _nEncoderLayers)
	encoderLayer = i+1;
      else
	encoderLayer = 2 * _nEncoderLayers-i;

      double weightDecay = _weightDecay[encoderLayer];
      int ns = _numStep[encoderLayer];
      int ps = _patchStep[encoderLayer];
      int iw = _inputWidth[encoderLayer];
      int biasRow = _weightsOut[i].rows()-1;

      //update biases
      if (i < _nEncoderLayers)
	{
	  //cout << "nn before: "<<endl;
	  //cout << _biasOut[i] << endl;
	  
	  
	  int idx = 0;
	  
	  int numRows = _delta[i+1].rows();
	  if (i+1 == _nEncoderLayers)
	    numRows--;

	  //	  cout << "fe _biasLearningRate: "<<_biasLearningRate<<endl;

	  for (int l = 0; l < ns; l++)
	    for (int m = 0; m < ns; m++)
	      {
		_weightsOut[i].row(biasRow).transpose() -= _biasLearningRate * _delta[i+1].block(0,idx,numRows,1);
		idx++;
	      }
	  
	  //cout << "nn after: "<<endl;
	  //cout << _biasOut[i] << endl;
	}
      else
	{
	  int idx = 0;
	  for (int l = 0; l < ns; l++)
	    {
	      for (int m = 0; m < ns; m++)
		{
		  _weightsOut[i].row(biasRow).transpose() -= _biasLearningRate * _deltaPreave[i+1].col(idx);
		  idx++;
		}
	    }
	}
      
      double smx = _desiredSparsityMax[i+1];
      double smn = _desiredSparsityMin[i+1];
      double ss = _sparsityPenaltyMultiplier[i+1]*_learningRate;
      
      //weight decay must come before all other updates to weights because it depends on the value of the weights
      _weightsOut[i].block(0,0,_weightsOut[i].rows()-1,_weightsOut[i].cols()) *= (1.0 - weightDecay * _learningRate);
      
      //backprop gradient and sparsity penalty
      if (i < _nEncoderLayers)
	{
	  //for each c
	  //for each col
	  //_weightsOut[i](:,c) -= _learningRate * _delta[i+1](corresponding to c, col) * _formattedOutput[i](:,col)
	  
	  //this simplifies to
	  //for each c
	  //_weightsOut[i](:,c) -= _learningRate *  _formattedOutput[i] * _delta[i+1](corresponding to c, :).transpose();
	  
	  //I was tired when I wrote this. I think "corresponding to c" is just c. 
	  
	  
	  int numRows = _delta[i+1].rows();
	  if (i+1 == _nEncoderLayers)
	    numRows--;
	  
	  for (int c = 0; c < _weightsOut[i].cols(); c++)
	    {
	      //backprop gradient update (see calculations above)
	      assert(_formattedOutput[i].cols() == _delta[i+1].cols());
	      _weightsOut[i].block(0,c,biasRow,1) -= _learningRate * _formattedOutput[i].block(0,0,biasRow,_formattedOutput[i].cols()) * (_delta[i+1].row(c).transpose());  
	      
	      
	      //sparsity update (only applies to encoder units)
	      for (int r = 0; r < _rawOutput[i+1].cols(); r++)
		{
		  double se = _sparsityEstimate[i+1](c,r);
		  
		  if (se > smx)	 
		    _weightsOut[i](biasRow,c) -= (se-smx) * ss;
		  else if (se < smn)
		    _weightsOut[i](biasRow,c) += (smn-se) * ss;
		}
	    }
	}
      else
	{ //decoder layers
	  //for each c
	  //for each col
	  //_weightsOut[i](:,c) -= _learningRate * _deltaPreave[i+1](c,col) * _rawOutput[i](:,col)
	  
	  //if (i == 3)
	    //{
	      //cout << "fe _deltaPreave "<<i+1<<endl<<_deltaPreave[i+1] << endl;
	      //cout << "fe layer 3 weights" << endl << _weightsOut[i]<<endl;
	      //cout << "fe layer 3 updates" << endl;
	      //}
	  

	  for (int c = 0; c < _weightsOut[i].cols(); c++)
	    {
	      _weightsOut[i].block(0,c,biasRow,1) -= _learningRate * _rawOutput[i].block(0,0,biasRow,_rawOutput[i].cols()) * _deltaPreave[i+1].row(c).transpose();
	      //if (i == 3)
		//cout << - _learningRate * _rawOutput[i].block(0,0,biasRow,_rawOutput[i].cols()) * _deltaPreave[i+1].row(c).transpose() << endl ;
	    }
	}
    }
}

void FastEncoder::showLayer1Weights(bool allowRescalingColorMap, bool allowScalingUp)
{
  _L1weightsDisplay.freezeWindow();
  updateLayer1Weights(allowRescalingColorMap, allowScalingUp);
  _L1weightsDisplay.unfreezeWindow();
  _L1weightsDisplay.enableWindow();
}

void FastEncoder::updateLayer1Weights(bool allowRescalingColorMap, bool allowScalingUp)
{ 
  //this means update the display, not update the weights themselves. should probably rename.

  _L1weightsDisplay.clearDisplay(); 
  MatrixXd temp(_inputWidth[1], _inputWidth[1]);


  for (int i = 0; i < _outputSize[1]; i++)
    {
      matrixSubColToMatrix(_weightsOut[0],0,i,temp);
      
      _L1weightsDisplay.addPatch(temp,allowRescalingColorMap,allowScalingUp);
     
      vectorToMatrix(_weightsOut.back().row(i), temp);

      _L1weightsDisplay.addPatch(temp,allowRescalingColorMap,allowScalingUp);
    } 
}

void FastEncoder::saveLayer1WeightsImage(const char *filename, bool allowRescalingColorMap, bool allowScalingUp)
{
  updateLayer1Weights(allowRescalingColorMap, allowScalingUp);
  _L1weightsDisplay.saveImage(filename);
}

void FastEncoder::saveNN(const char *filename)
{
  cout << "Saving" << endl;

  std::ofstream outputFile(filename, ios::out | ios::binary);
  if (!outputFile.is_open()) {
    std::cerr << "FastEncoder::saveNN: Tried to open " << filename << "\n";
    error("FastEncoder::saveNN: Error opening file.");
  }

  //Write a header so we can verify that the file is a
  //Fast AutoEnCoder
  static char header[] = {'f','a','e','c'};
  outputFile.write((char *)header, 4);

  //Write a version number so we can change the file format later
  //and still load old files
  int version = 0;
  outputFile.write((char *)&version, sizeof(int));

  //Save all the fields of the class
  outputFile.write((char*)&_nEncoderLayers, sizeof(int));
  outputFile.write((char*)&_nTotLayers, sizeof(int));
  
  int s;

  s = _inputWidth.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_inputWidth[i], sizeof(int));

  s = _patchStep.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_patchStep[i], sizeof(int));
  
  s = _numStep.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_numStep[i], sizeof(int));

  s = _outputSize.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_outputSize[i], sizeof(int));


  s = _weightsOut.size();
  outputFile.write((char *)&s, sizeof(int));
  //cout << "Writing s: "<<s<<endl;
  for (int i = 0; i < s; i++)
    {
      saveMatrix(outputFile, _weightsOut[i]);
    }

  //just save the sizes of the temp datastructures
#define SAVE_MATRIX_VECTOR(v) s = v.size();\
  outputFile.write((char *)&s, sizeof(int));\
  for (int i = 0; i < s; i++)\
    {\
      int r = v[i].rows();\
      int c = v[i].cols();\
      outputFile.write((char *)&r, sizeof(int));\
      outputFile.write((char *)&c, sizeof(int));\
    }

  SAVE_MATRIX_VECTOR(_formattedOutput);
  SAVE_MATRIX_VECTOR(_input);
  SAVE_MATRIX_VECTOR(_rawOutput);
  SAVE_MATRIX_VECTOR(_delta);
  SAVE_MATRIX_VECTOR(_deltaPreave);
  SAVE_MATRIX_VECTOR(_deltaFormatted);



  //Now we're back to real stuff that needs to be saved
  outputFile.write((char*)&_learningRate, sizeof(double));
  outputFile.write((char*)&_biasLearningRate, sizeof(double));

  s = _weightDecay.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)& (_weightDecay[0]), sizeof(double)*s);
  s = _desiredSparsityMin.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)&(_desiredSparsityMin[0]), sizeof(double)*s);
  s = _desiredSparsityMax.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)&(_desiredSparsityMax[0]), sizeof(double)*s);
  s = _sparsityPenaltyMultiplier.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)& (_sparsityPenaltyMultiplier[0]), sizeof(double)*s);

  outputFile.write((char *)&_learnAll,sizeof(bool));

  outputFile.write((char *)&_errorEstimate, sizeof(double));

  s = _sparsityEstimate.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      saveMatrix(outputFile, _sparsityEstimate[i]);
    }

  //Only need to save the size of sparsity penalty estimate
  s = _sparsityPenaltyEstimate.size();
  outputFile.write((char *)&s, sizeof(int));


  //No need to save _L1weightsDisplay, it can be regenerated

  if (!outputFile.good())
    error("Error saving to file.");

  outputFile.close();

  cout << "Done saving " << filename  << endl;
 
}

void FastEncoder::loadNN(const char *filename)
{
  std::ifstream inputFile(filename, ios::in | ios::binary);
  if (!inputFile.is_open()) {
    std::cerr << "FastEncoder::loadNN Tried to open " << filename << "\n";
    error("FastEncoder::loadNN: Error opening file."); 
    return;
  }

  //Read header
  static char header[5];
  header[4] = '\0';
  inputFile.read((char*)header, 4);
  
  if (strcmp(header, "faec") != 0)
    {
      inputFile.close();

      if (strcmp(header,"saec") != 0)
	{
	  cerr << "FastEncoder::loadNN: file \""<<filename<<"\" does not have the right header" << endl;
	  return;
	}
     

      
      //Load StackedEncoder, then convert
      StackedEncoder sa(filename);

      _nEncoderLayers = sa._nEncoderLayers;
      _nTotLayers = sa._nTotLayers;
      _inputWidth = sa._inputWidth;
      _patchStep = sa._patchStep;
      _numStep = sa._numStep;
      _outputSize = sa._outputWidth;

      _inputWidth[0] = 1;
      _patchStep[0] = 1;
      _outputSize[0] = 1;
      _numStep[0] = sa.inputWidth();

      for (int i = 2; i <= _nEncoderLayers; i++)
	{
	  _inputWidth[i] /= _outputSize[i-1];
	  _patchStep[i] /= _outputSize[i-1];
	}

      for (int i = 1; i <= _nEncoderLayers; i++)
	{
	  _outputSize[i] *= _outputSize[i];
	}

      for (int i = 0; i < _nEncoderLayers; i++)
	{
	  //cout << "i: "<<i << endl;
	  //cout << "_numStep "<<i<<":"<<_numStep[i] << endl;
	  //cout << "_inputWidth "<<i+1<<":"<<_inputWidth[i+1]<<endl;
	  assert( _numStep[i] == ( _numStep[i+1] -1)*_patchStep[i+1] + _inputWidth[i+1]);
	}
   
      _learningRate = sa._learningRate;
      _weightDecay = sa._weightDecay;
      _desiredSparsityMin = sa._desiredSparsityMin;
      _desiredSparsityMax = sa._desiredSparsityMax;
      _sparsityPenaltyMultiplier = sa._sparsityPenaltyMultiplier;
      _learnAll = sa._learnAll;
      _L1weightsDisplay = sa._L1weightsDisplay;

      cerr << "Warning: loading StackedEncoder as FastEncoder. This discards saved error estimate, sparsity estimates." << endl;
      allocateTemporaries();

      _weightsOut.clear();

      for (int i = 0; i < _nEncoderLayers; i++)
	{
	  _weightsOut.push_back(MatrixXd(_inputWidth[i+1]*_inputWidth[i+1]*_outputSize[i]+1, _outputSize[i+1]));
	}

      int j = _nEncoderLayers;
      for (int i = _nEncoderLayers; i < _nTotLayers - 1; i++)
	{
	  _weightsOut.push_back(MatrixXd(_outputSize[j]+1,_inputWidth[j]*_inputWidth[j]*_outputSize[j-1]));
	  j--;
	}				

      copyWeights(sa);
      return;
    }

  
  //Native file loading

  int version;
  inputFile.read((char *)&version,sizeof(int));

  assert(version==0);//if/when there are multiple versions, we can put
  //a switch here

  inputFile.read((char *)&_nEncoderLayers,sizeof(int));
  inputFile.read((char *)&_nTotLayers,sizeof(int));
  
  int s;

  inputFile.read((char *)&s, sizeof(int));
  _inputWidth.resize(s);
  inputFile.read((char *)&(_inputWidth[0]), sizeof(int)*s);

  inputFile.read((char *)&s, sizeof(int));
  _patchStep.resize(s);
  inputFile.read((char *)&(_patchStep[0]), sizeof(int)*s);

  //int buf[100];
  //inputFile.read((char *) buf, sizeof(int)*100);
  //assert(false);

  inputFile.read((char *)&s, sizeof(int));
  _numStep.resize(s);
  inputFile.read((char *)&(_numStep[0]), sizeof(int)*s);
  
  inputFile.read((char *)&s, sizeof(int));
  _outputSize.resize(s);
  inputFile.read((char *)&(_outputSize[0]), sizeof(int)*s);

  inputFile.read((char *)&s, sizeof(int));


  //cout << "Resizing to size "<<s<<endl;
  _weightsOut.resize(s, MatrixXd::Zero(1,1));
  //cout << "Success" << endl;
  for (int i = 0; i < s; i++)
    {
      loadMatrix(inputFile, _weightsOut[i], true);
    }

  //Temp datastructures just need to be allocated with the right size
  
#define LOAD_MATRIX_VECTOR( v) inputFile.read((char *)&s, sizeof(int)); \
  v.resize(s, MatrixXd::Zero(1,1)); \
  for (int i = 0; i < s; i++) \
    { \
      int r, c; \
      inputFile.read((char *) &r, sizeof(int)); \
      inputFile.read((char *) &c, sizeof(int)); \
      v[i] = MatrixXd::Ones(r,c); \
    }

  LOAD_MATRIX_VECTOR(_formattedOutput);
  LOAD_MATRIX_VECTOR(_input);
  LOAD_MATRIX_VECTOR(_rawOutput);
  LOAD_MATRIX_VECTOR(_delta);
  LOAD_MATRIX_VECTOR(_deltaPreave);
  LOAD_MATRIX_VECTOR(_deltaFormatted);





  inputFile.read((char*)&_learningRate, sizeof(double));
  inputFile.read((char*)&_biasLearningRate, sizeof(double));

  inputFile.read((char *)&s, sizeof(int));
  _weightDecay.resize(s);
  inputFile.read((char*)&(_weightDecay[0]), s*sizeof(double));

  inputFile.read((char *)&s, sizeof(int));
  _desiredSparsityMin.resize(s);
  inputFile.read((char*)&(_desiredSparsityMin[0]), s*sizeof(double));

  inputFile.read((char *)&s, sizeof(int));
  _desiredSparsityMax.resize(s);
  inputFile.read((char*)&(_desiredSparsityMax[0]), s*sizeof(double));

  inputFile.read((char *)&s, sizeof(int));
  _sparsityPenaltyMultiplier.resize(s);
  inputFile.read((char*)&(_sparsityPenaltyMultiplier[0]), s*sizeof(double));

  inputFile.read((char*)&_learnAll, sizeof(bool));

  inputFile.read((char *)&_errorEstimate, sizeof(double));

  cout << "Error estimate: "<<_errorEstimate<<endl;

  inputFile.read((char *)&s, sizeof(int));
  _sparsityEstimate.resize(s, MatrixXd::Zero(1,1));
  for (int i = 0; i < s; i++)
    {
      loadMatrix(inputFile, _sparsityEstimate[i], true);
    }

  inputFile.read((char *)&s, sizeof(int));
  _sparsityPenaltyEstimate.resize(s);
  

  cout << "Read fast autoencoder." << endl;
  cout << "Hidden units in deepest layer: " << _numStep.back()*_numStep.back()*_outputSize.back() << "\nInput width: " << inputWidth()  
       << "\nLearning rate: " << _learningRate << "\n";

  cout << "Numstep: ";
  for (int i = 0; i < _patchStep.size(); i++)
    cout << _numStep[i] << "\t";
  cout << endl;

  cout << "Output sizes: ";
  for (int i = 0; i < _outputSize.size(); i++)
    cout << _outputSize[i] << "\t";
  cout << endl;
       
  if (!inputFile.good())
    error("StackedEncoder::StackedEncoder: Error reading from file.");

  _L1weightsDisplay.setNewDisplayParam(_inputWidth[1], _inputWidth[1], 1, _outputSize[1]*2);
  _L1weightsDisplay.setRange(-1,1); 

  inputFile.close();

  return;

}

void FastEncoder::showBiases(int layer)
{
  assert(false);

  //todo-- make this work in the new setup

  /*
  PatchDisplay d("Biases",_biasOut[layer].cols(),_biasOut[layer].rows(),4,2,2);
  d.setRange(-1,1);
  d.addPatch(_biasOut[layer]);

  MatrixXd activations = _biasOut[layer];
  for (int i = 0; i < activations.rows(); i++)
    for (int j = 0; j < activations.cols(); j++)
      activations(i,j) = tanh(activations(i,j));

  d.addPatch(activations);


  MatrixXd inputs = _biasOut[layer];

  for (int i = 0; i < inputs.rows(); i++)
    {
      for (int j = 0; j < inputs.cols(); j++)
	{
	  inputs(i,j) -= _weightsOut[layer][i][j].sum();
	}
    }

  d.addPatch(inputs);

  MatrixXd outputs = inputs;

  for (int i = 0; i < inputs.rows(); i++)
    {
      for (int j = 0; j < inputs.cols(); j++)
	{
	  outputs(i,j) = tanh(outputs(i,j));
	}
    }


  d.addPatch(outputs);

  d.enableWindow();
  cout << "Biases "<<layer<<endl;




  //4 panes
  //upper left: biases
  //upper right: tanh(biases)
  //lower left: activationIn when all inputs are -1 (this is more analagous to what bias usually represents in a sigmoidal network)
  //lower right: tanh of the same

  cvWaitKey(0);
  */
}

void FastEncoder::copyWeights(StackedEncoder &sa)
{
  //Check that networks are same size
  assert(sa._nEncoderLayers == _nEncoderLayers);

  for (unsigned i = 1; i <= sa._nEncoderLayers; i++)
    {
      //      cout << "ow: "<<sa._outputWidth[i] << endl;
      //cout << "os: "<<_outputSize[i] << endl;
      assert(sa._outputWidth[i] * sa._outputWidth[i] == _outputSize[i]);
      assert(sa._inputWidth[i]/((int) sqrt(_outputSize[i-1])) == _inputWidth[i]);
      assert(sa._numStep[i] == _numStep[i]);
    }

  //check that stacked encoders receptive fields align with output groups, so it can be converted to fastEncoder
  for (unsigned i = 2; i <= sa._nEncoderLayers; i++)
    {
      assert(sa._inputWidth[i] >= sa._outputWidth[i-1]);
      assert(sa._inputWidth[i] % sa._outputWidth[i-1] == 0);
    }

  for (unsigned i = 0; i < sa._weightsOut.size(); i++)
    {
      int weightsIdx = 0;

      if (i >= _nEncoderLayers)
	{
	  int encoderLayer = 2 * _nEncoderLayers - i;
	  
	  int numBlocks = _inputWidth[encoderLayer];
	  int blockWidth = (int) sqrt(_outputSize[encoderLayer-1]);
	  int blockRow = 0;
	  int blockCol = 0;
	  int inblockRow = 0;
	  int inblockCol = 0;
	  
	  for (weightsIdx = 0; weightsIdx < _weightsOut[i].cols(); weightsIdx++)
	    {

	      int row = 0;

	      int j = blockRow*blockWidth + inblockRow;
	      int k = blockCol*blockWidth + inblockCol;


	      assert(i < sa._weightsOut.size());
	      assert(j < sa._weightsOut[i].size());
	      assert(k < sa._weightsOut[i][j].size());

	      for (int m = 0; m < sa._weightsOut[i][j][k].cols(); m++)
		{
		  for (int l = 0; l < sa._weightsOut[i][j][k].rows(); l++)
		    {		      
		      _weightsOut[i](row, weightsIdx) = sa._weightsOut[i][j][k](l,m);//confirmed: l and m indices are confirmed, just got to change the order we visit each j and k
		      row++;
		    }
		}
	      _weightsOut[i](row,weightsIdx) = sa._biasOut[i](j,k);

		 
	      inblockRow++;
	      
	      if (inblockRow == blockWidth)
		{
		  inblockRow = 0;
		  
		  inblockCol++;
		  
		  if (inblockCol == blockWidth)
		    {
		      inblockCol = 0;
		      
		      blockRow ++;

		      if (blockRow == numBlocks)
			{
			  blockRow = 0;
			  
			  blockCol ++;
			  
			  if (blockCol == numBlocks)
			    assert(weightsIdx == _weightsOut[i].cols()-1);
			}
		    }
		}
	    }
	}
      else
	{
	  for (unsigned k = 0; k < sa._weightsOut[i][0].size(); k++)
	    {
	      for (unsigned j = 0; j < sa._weightsOut[i].size(); j++)
		{		  		  		
		  int blockRow = 0;
		  int blockCol = 0;
		  int ingroupRow = 0;
		  int ingroupCol = 0;
		  int blockWidth = (int) sqrt(_outputSize[i]);
		  
		  for (int row = 0; row < _weightsOut[i].rows()-1; row++)
		    {
		      assert(i < sa._weightsOut.size() );
		      assert(j < sa._weightsOut[i].size());
		      assert(k < sa._weightsOut[i].size());

		      //cout << "_weightsOut["<<i<<"]("<<row<<","<<weightsIdx<<")= sa._weightsOut["<<i<<"]["<<j<<"]["<<k<<"]("<<blockRow*blockWidth+ingroupRow<<","<<blockCol*blockWidth+ingroupCol<<")="<< sa._weightsOut[i][j][k](blockRow*blockWidth+ingroupRow, blockCol*blockWidth+ingroupCol)<<endl;

		      _weightsOut[i](row,weightsIdx) = sa._weightsOut[i][j][k](blockRow*blockWidth+ingroupRow, blockCol*blockWidth+ingroupCol);
		      
		      ingroupRow++;
		      
		      if (ingroupRow == blockWidth)
			{
			  ingroupRow = 0;
			  ingroupCol++;
			  if (ingroupCol == blockWidth)
			    {
			      ingroupCol = 0;
			      blockRow++;
			      
			      if (blockRow == _inputWidth[i+1])
				{
				  blockRow = 0;
				  blockCol++;
				  
				  if (blockCol == _inputWidth[i+1])
				    assert(row == _weightsOut[i].rows()-2);
				}
			    }
			}
		    }
		  weightsIdx++;
		}
	      
	    }
	
	  //yes, this block really is only for i < _nEncoderLayers
	  int biasIdx = 0;
	  for (unsigned k = 0; k < sa._biasOut[i].cols(); k++)
	    {
	      for (unsigned j = 0; j < sa._biasOut[i].rows(); j++)
		{
		  _weightsOut[i](_weightsOut[i].rows()-1, biasIdx) = sa._biasOut[i](j,k);
		  
		  biasIdx++;
		}
	    }
	}
    }
  
}

void FastEncoder::checkWeights(StackedEncoder &sa)
{
  double tol = 1e-10;

  //Check that networks are same size
  assert(sa._nEncoderLayers == _nEncoderLayers);

  for (unsigned i = 1; i <= sa._nEncoderLayers; i++)
    {
      //      cout << "ow: "<<sa._outputWidth[i] << endl;
      //cout << "os: "<<_outputSize[i] << endl;
      assert(sa._outputWidth[i] * sa._outputWidth[i] == _outputSize[i]);
      assert(sa._inputWidth[i]/((int) sqrt(_outputSize[i-1])) == _inputWidth[i]);
      assert(sa._numStep[i] == _numStep[i]);
    }

  //check that stacked encoders receptive fields align with output groups, so it can be converted to fastEncoder
  for (unsigned i = 2; i <= sa._nEncoderLayers; i++)
    {
      assert(sa._inputWidth[i] >= sa._outputWidth[i-1]);
      assert(sa._inputWidth[i] % sa._outputWidth[i-1] == 0);
    }

  for (unsigned i = 0; i < sa._weightsOut.size(); i++)
    {
      int weightsIdx = 0;

      if (i >= _nEncoderLayers)
	{
	  int encoderLayer = 2 * _nEncoderLayers - i;
	  
	  int numBlocks = _inputWidth[encoderLayer];
	  int blockWidth = (int) sqrt(_outputSize[encoderLayer-1]);
	  int blockRow = 0;
	  int blockCol = 0;
	  int inblockRow = 0;
	  int inblockCol = 0;
	  
	  for (weightsIdx = 0; weightsIdx < _weightsOut[i].cols(); weightsIdx++)
	    {

	      int row = 0;

	      int j = blockRow*blockWidth + inblockRow;
	      int k = blockCol*blockWidth + inblockCol;


	      assert(i < sa._weightsOut.size());
	      assert(j < sa._weightsOut[i].size());
	      assert(k < sa._weightsOut[i][j].size());

	      for (int m = 0; m < sa._weightsOut[i][j][k].cols(); m++)
		{
		  for (int l = 0; l < sa._weightsOut[i][j][k].rows(); l++)
		    {		      
		      if (abs(_weightsOut[i](row, weightsIdx) - sa._weightsOut[i][j][k](l,m)) > tol)
			cout << "differ at ("<<i<<" "<<j<<" "<<k<<" "<<m<<" "<<l<<")"<<endl;
		      row++;
		    }
		}
	      if (abs(_weightsOut[i](row,weightsIdx) - sa._biasOut[i](j,k) > tol))
		cout << "biases differ at ("<<i<<" "<<j<<" "<<k<<")"<<endl;

		 
	      inblockRow++;
	      
	      if (inblockRow == blockWidth)
		{
		  inblockRow = 0;
		  
		  inblockCol++;
		  
		  if (inblockCol == blockWidth)
		    {
		      inblockCol = 0;
		      
		      blockRow ++;

		      if (blockRow == numBlocks)
			{
			  blockRow = 0;
			  
			  blockCol ++;
			  
			  if (blockCol == numBlocks)
			    assert(weightsIdx == _weightsOut[i].cols()-1);
			}
		    }
		}
	    }
	}
      else
	{
	  for (unsigned k = 0; k < sa._weightsOut[i][0].size(); k++)
	    {
	      for (unsigned j = 0; j < sa._weightsOut[i].size(); j++)
		{		  		  		
		  int blockRow = 0;
		  int blockCol = 0;
		  int ingroupRow = 0;
		  int ingroupCol = 0;
		  int blockWidth = (int) sqrt(_outputSize[i]);
		  
		  for (int row = 0; row < _weightsOut[i].rows()-1; row++)
		    {
		      assert(i < sa._weightsOut.size() );
		      assert(j < sa._weightsOut[i].size());
		      assert(k < sa._weightsOut[i].size());

		      //cout << "_weightsOut["<<i<<"]("<<row<<","<<weightsIdx<<")= sa._weightsOut["<<i<<"]["<<j<<"]["<<k<<"]("<<blockRow*blockWidth+ingroupRow<<","<<blockCol*blockWidth+ingroupCol<<")="<< sa._weightsOut[i][j][k](blockRow*blockWidth+ingroupRow, blockCol*blockWidth+ingroupCol)<<endl;

		      if (abs(_weightsOut[i](row,weightsIdx) - sa._weightsOut[i][j][k](blockRow*blockWidth+ingroupRow, blockCol*blockWidth+ingroupCol)) > tol)
			cout << "differ at ("<<i<<" "<<j<<" "<<k<<" "<<blockRow*blockWidth+ingroupRow<<" "<<blockCol*blockWidth+ingroupCol<<")"<<endl;
		      
		      ingroupRow++;
		      
		      if (ingroupRow == blockWidth)
			{
			  ingroupRow = 0;
			  ingroupCol++;
			  if (ingroupCol == blockWidth)
			    {
			      ingroupCol = 0;
			      blockRow++;
			      
			      if (blockRow == _inputWidth[i+1])
				{
				  blockRow = 0;
				  blockCol++;
				  
				  if (blockCol == _inputWidth[i+1])
				    assert(row == _weightsOut[i].rows()-2);
				}
			    }
			}
		    }
		  weightsIdx++;
		}
	      
	    }
	
	  //yes, this block really is only for i < _nEncoderLayers
	  int biasIdx = 0;
	  for (unsigned k = 0; k < sa._biasOut[i].cols(); k++)
	    {
	      for (unsigned j = 0; j < sa._biasOut[i].rows(); j++)
		{
		  if (abs(_weightsOut[i](_weightsOut[i].rows()-1, biasIdx) - sa._biasOut[i](j,k)) > tol)
		    cout << "biases differ at ("<<i<<" "<<j<<" "<<k<<")"<<endl;
		  
		  biasIdx++;
		}
	    }
	}
    }
  
}

void FastEncoder::weightsHack()
{

  //this does NOT work with convolution, that is much tricker


  cerr << "Warning! using weights image hack that assumes no convolution!!!!!!" << endl;

  vector<vector<MatrixXd> > weights;
  vector<vector<MatrixXd> > decodeWeights;

  weights.push_back(vector<MatrixXd> ());
  decodeWeights.push_back(vector<MatrixXd> ());

  MatrixXd temp(_inputWidth[1], _inputWidth[1]);


  cout << "Processing layer 1"<< endl;

  for (unsigned i = 0; i < _outputSize[1]; i++)
    {
      matrixSubColToMatrix(_weightsOut[0],0,i,temp);
      weights[0].push_back(temp);

      vectorToMatrix(_weightsOut.back().row(i), temp);
      decodeWeights[0].push_back(temp);
    }

  int decodeLayer = _nTotLayers - 3;
  for (unsigned i = 2; i <= _nEncoderLayers; i++)
    {
      cout << "Processing layer " << i << endl;
      weights.push_back(vector<MatrixXd> ());
      decodeWeights.push_back(vector<MatrixXd> ());
      
      for (unsigned j = 0; j < _outputSize[i]; j++)
	{
	  temp.setConstant(0);



	  for (unsigned k = 0; k < _outputSize[i-1]; k++)
	    {
	      assert(i-1 > 0);
	      assert(i-1 < _weightsOut.size());
	      assert(i-1 < weights.size());
	      assert(k < weights[i-2].size());
	      assert(temp.cols() == weights[i-2][k].cols());
	      assert(temp.rows() == weights[i-2][k].rows());
	  
	      temp += _weightsOut[i-1](k,j) * weights[i-2][k];
	    }

	  weights[i-1].push_back(temp);


	  temp.setConstant(0);

	  for (unsigned k = 0; k < _outputSize[i-1]; k++)
	    {
	      assert(decodeLayer < _weightsOut.size());
	      assert(j < _weightsOut[decodeLayer].rows());
	      assert(k < _weightsOut[decodeLayer].cols());
	      assert(decodeWeights[i-2][k].rows() == temp.rows());
	      assert(decodeWeights[i-2][k].cols() == temp.cols());
	      temp += _weightsOut[decodeLayer](j,k) * decodeWeights[i-2][k];
	    }

	  decodeWeights[i-1].push_back(temp);
	}

      decodeLayer--;
    }


  for (unsigned i = 0; i < weights.size(); i++)
    {
      stringstream name;
      name << "Layer "<<(i+1)<<" Weights";
      PatchDisplay d(name.str().c_str(), _inputWidth[1],_inputWidth[1], 1, 2 * weights[i].size());

      for (unsigned j = 0; j < weights[i].size(); j++)
	{
	  d.addPatch(weights[i][j],1,1);
	  d.addPatch(decodeWeights[i][j],1,1);
	}

      d.enableWindow();

      cvWaitKey(0);
    }

  

}



    void FastEncoder::setCodeVec(int i, const Eigen::VectorXd & code)
    {
	//cout << "Setting code for layer "<<i<<endl;



        if (i < _nEncoderLayers)
                vectorToMatrix(code, _rawOutput[i]);
        else
                vectorToMatrixBlock(code, _rawOutput[i], 0, 0, _rawOutput[i].rows()-1,_rawOutput[i].cols());

	//cout << _rawOutput[i].block(0,0,_rawOutput[i].rows()-1, _rawOutput[i].cols()) << endl;
	//exit(-1);

    
	//int count = 0;
	//for (int j = 0; j < _rawOutput[i].rows()-1; j++)
	//	for (int k = 0; k < _rawOutput[i].cols(); k++)
	//	{
	//		if (_rawOutput[i](j,k) != -1.0)
	//		{
	//			count++;
	//			assert(_rawOutput[i](j,k) == 1.0);
	//		}
	//	}

	//assert(count == 1);		
   }

