
#pragma once
#include <vector>

#include <Eigen/Core>
#include <Eigen/Array>
#include "stackedEncoder.h"
#include "matrixUtil.h"

class FastEncoder
{
  public:

  //Constructs an autoencoder from file
  FastEncoder(const char * filepath) : _L1weightsDisplay(std::string("Layer 1 Weights"),1,1,1,1,1)
    {
      loadNN(filepath);
    }

  //Constructs an autoencoder with a single hidden layer
  FastEncoder(int inputWidth, int outputSize,			\
		 double learningRate, double biasLearningRate,		\
		   double weightDecay, double desiredSparsityMin,	\
		   double desiredSparsityMax,				\
	      double sparsityPenaltyMultiplier,
	      double initMag = 0.05);

  //Given an autoencoder with n encoder layers, constructs an autoencoder with
  //n+1 encoder layers. outputSize is the number of hidden units in the final 
  //hidden layer. patchStep and numStep are assigned to the
  //next layer down.

  FastEncoder(int outputSize, int patchStep, int numStep,
		 FastEncoder & smaller, bool learnAll, double weightDecay,
		 double desiredSparsityMin, double desiredSparsityMax,
	      double sparsityPenaltyMultiplier,
	      double initMag = 0.05);
  
  void setLearnAll(bool learnAll) { _learnAll = learnAll; }

  void disableDeepbackprop() { _deepBackprop = false; } //todo-- store this option in file

  int inputWidth() { return _numStep[0]; }


  void weightsHack();


  void forwardProp(const Eigen::MatrixXd * patch, double * errorEstimate = NULL,
		   std::vector<std::vector< double> > * sparsityStats = NULL,
		   std::vector<double> * sparsityPenaltyEstimate = NULL, int startLayer = 0, int endLayer = -1);//sparsityPenaltyEstimate is not actually calcualted right now
 


    void backProp();
    void updateWeights();


    Eigen::VectorXd  getCodeVec(int i) const
      {
	assert( i > 0 && i <= _nEncoderLayers);
	if (i < _nEncoderLayers)
		return matrixToVector(_rawOutput[i]);
	else
		return matrixToVector(_rawOutput[i].block(0,0,_rawOutput[i].rows()-1,_rawOutput[i].cols())); 

     }

    void getCodeVec(int i, Eigen::VectorXd & output) const
    {
      assert( i > 0 && i <= _nEncoderLayers);
	if (i < _nEncoderLayers)
      		matrixToVector(_rawOutput[i],output);
	else
		matrixToVector(_rawOutput[i].block(0,0,_rawOutput[i].rows()-1,_rawOutput[i].cols()), output);
    }

    void setCodeVec(int i, const Eigen::VectorXd & code);

    const Eigen::MatrixXd outputValues() const
      { 	
	return vectorToMatrix(_rawOutput[_nTotLayers-1].row(0), _numStep[0], _numStep[0]);
      }


    void debugSetAllWeights(double c);
    const Eigen::MatrixXd & debugDelMe2() { return _delta[1]; }

    void showLayer1Weights(bool allowRescalingColorMap , bool allowScalingUp );
    void saveLayer1WeightsImage(const char * filepath, bool allowRescalingColorMap, bool allowScalingUp);
    void showBiases(int layer);
  

    void saveNN(const char *filename); 
    void loadNN(const char *filename);

    int numEncoderLayers() const { return _nEncoderLayers; }




    void debugSetConstantWeights(double d)
    {
      for (unsigned i = 0; i < _weightsOut.size(); i++)
	_weightsOut[i].block(0,0,_weightsOut[i].rows()-1,_weightsOut[i].cols()).setConstant(d);
    }

    void copyWeights(StackedEncoder &sa);
    void checkWeights(StackedEncoder &sa);
    const std::vector<Eigen::MatrixXd> & getActivationOut() const 
      {
	return _rawOutput;
      }

 private:
    void updateLayer1Weights(bool allowRescalingColorMap = true, bool allowScalingUp = true);
    void allocateTemporaries();
    int encoderLayer(int L) const
    {
      assert(L >= _nEncoderLayers);
      assert(L < _nTotLayers);
      return 2*_nEncoderLayers-L;
    }


  private:

  int _nEncoderLayers;                      // If have I -> L1 -> L2 -> L1 -> O, then this is 2.  
  // If have I -> L1 -> L2 -> L3 -> ... 0, then this is 3
  int _nTotLayers;                          // In the two examples above, would be respectively 5 and 7.
  // _nTotlayers =  2*_nEncoderLayers+1. 

  
  // Parameters specifying the dimensions of each layer 
  std::vector<int> _inputWidth, _patchStep, _numStep, _outputSize; // vectors of size _nEncoderLayers+1.
    // Meanings of these parameters: 
    // _inputWidth[i]: Layer Li's activations is a function of an _inputWidth[i]x_inputWidth[i] patch of the lower level. 
    // _patchStep[i]:  Adjacent Li groups take their input from receptive fields that are shifted by _patchStep[i] in Li-1
    // _numStep[i]: The groups of layer Li form a numStep[i]xnumStep[i] array.
    // _outputSize[i]: Each group of layer Li is of this size. Each of these outputSize[i] units
    //                 connect to the same set of Li-1 units.  So, the total dimension (number of real numbers) of Li's 
    //                 output is numStep[i]^2 * outputSize[i]


    //For element 0 of the four arrays above, we must always have 
    //    _inputWidth = _patchStep = _outputSize= 1,
    //    and _numStep = imageSize

    // Weights
  std::vector<Eigen::MatrixXd > _weightsOut;  //column j of _weightsOut[i] is the weights and bias term for hidden unit j within its group in Li+1.
  //if these are decoding weights, note that the units in layer i+1 are grouped by the group identities of the encoder layer corresponding to layer i
  //the last unit in the input is always 1, this is the bias term.
  //_weightsOut[i] is thus of size _inputWidth[i+1]^2*_outputSize[i] + 1 x  _outputSize[i+1]   (where "x" is "by", not "times")
  //or if i is a decoding layer or last encoder layer, then let j be the 
  //index of the encoding layer corresponding to i
  //then _weightsOut[i] is of size _outputSize[j] + 1 x _inputWidth[j]^2*_outputSize[j-1]                

  std::vector<Eigen::MatrixXd > _formattedOutput; //column j of _formattedOutput[i] is the input, terminated with bias term 1, to hidden unit group j in Li+1.
  //_formattedOutput[i] is thus of size _inputWidth[i+1]^2*_outputSize[i] +1  x _numStep[i+1]^2
  //or if i is the last encoder layer/ a decoding layer, then let j be the 
  //index of the encoding layer corresponding to i
  //not used for last encoding layer or for any decoding layers
  std::vector<Eigen::MatrixXd > _input; //_input[i] is values before tanh/averaging of Li. _input[0] is unused.
  //during forward prop _input[i] = _weightsOut[i-1].transpose()*_formattedOutput[i-1];
  //so _input[i] is of size _outputSize[i] x _numStep[i]^2
  //or if i is a decoding layer, then let j be the index of the encoding layer corresponding to i.
  //then _input[i] is of size _inputWidth[j+1]^2 * _outputSize[j] x _numStep[j+1]*_numStep[j+1]     (i.e., each instantiation of the weights in code layer j+1 tries to reconstruct a different patch of code layer j)
  std::vector<Eigen::MatrixXd> _rawOutput; //_rawOutput[i] is made by applying tanh and averaging to _input[i].
  //it is of size _outputSize[i] x _numStep[i]^2
  //or if i is last encoder layer or a decoder layer, let j be the corresponding encoder layer
  //then _rawOutput[j] is of size _outputSize[j]+1 x _numStep[j]^2
  //in this case (last encoder, or any decoder) the last row is the bias term

  
  //_delta = same size as _rawOutput (except not used for layer 0), _deltaPreave = same size as _input (except not used for layers <= _nEncoderLayers)
  std::vector<Eigen::MatrixXd> _delta; //only _delta[nTotLayers-1] down to _delta[_nEncoderLayers-1] is used if we are just learning the central set of weights
  std::vector<Eigen::MatrixXd> _deltaPreave;
  //(For layers > _nEncoderLayers, _delta is actually the deltas for imaginary
  //averaging units, while _deltaPreave is the deltas  for the sigmoidal
  //units)
  std::vector<Eigen::MatrixXd> _deltaFormatted;

  // learning parameters
  double _learningRate, _biasLearningRate;
  //vectors of size _nEncoderLayers+1, with element 0 unused
  std::vector<double> _weightDecay, _desiredSparsityMin, _desiredSparsityMax,
_sparsityPenaltyMultiplier; 
  
 bool _learnAll; //if true, adjusts all weights using backprop rather
  //than just the central layer.
 bool _deepBackprop; //true = default behavior where the objective is reconstruction of the input. false = objective is reconstruction of 2nd deepest hidden layer.



 double _errorEstimate;

 //same size as _rawOutput
 std::vector<Eigen::MatrixXd> _sparsityEstimate; //element 0 is unused. element i is for encoder layer i
 std::vector<double> _sparsityPenaltyEstimate; //element 0 is unused, and if _learnAll is false, the only used element is _nEncoderLayers

 PatchDisplay _L1weightsDisplay;
};
