/// \author Ian Goodfellow
#include <sdn/base.h>
#include <sdn/old.h>
#include <sdn/dataset.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include <Eigen/Array>
using namespace boost;
using namespace sdn;
using namespace std;

/// Uses the hidden units in the deepest layer of the network
// specified by argv[1] to as features for a logistic regression
// model
// trains on the first NUM_TRAIN_EXAMPLES
// mnist digits (scaled down to 14x14) 
// tests on the next NUM_TEST_EXAMPLES
// reports test set misclassification rate


void getExamples(shared_ptr<Network> n, shared_ptr<StorageNode> h,
		 Matrix & X, Vector & y, vector<unsigned> & yv, shared_ptr<Dataset> d,
		 unsigned numExamples)
{
  X.resize(numExamples,h->getActivities().rows());
  y.resize(numExamples);
  yv.resize(numExamples);

  Matrix input = Matrix::Zero(n->getInputRows(),n->getInputCols());

  for (unsigned i = 0; i < numExamples; i++)
    {
      yv[i] = d->getRandomExample(input);
      y(i) = (yv[i] == 8);
      n->processExample(input);
      X.row(i) = h->getActivities().transpose();
    }

  cout << "A piece of x: " << X.block(0,0,4,4) << endl;
}

void writeExample(const Matrix & X, int i, string dir)
{
	stringstream name;
	name << dir << "/example_"<<i<<".txt";
	ofstream ofs(name.str().c_str());
	SDN_ASSERT(! ofs.fail());
	for (int j = 0; j < X.cols(); j++)
	{
		ofs << X(i,j) << endl;
	}
	SDN_ASSERT(! ofs.fail());
        ofs.close();	
}

void makeCache(const Matrix & X, Vector & y, string dirbase)
{
	mkdir((dirbase+"_pos").c_str());
	mkdir((dirbase+"_neg").c_str());
	for (int i = 0; i < X.rows(); i++)
	{
		if (y(i) > 0.5)
{
			writeExample(X,i,dirbase+"_pos");
}
		else
{
			writeExample(X,i,dirbase+"_neg");
}	
	}
}

int main(int argc, char ** argv)
{
  ifstream ifs(argv[1], ios::binary);
  serial::BinInArchive ar(ifs);
  shared_ptr<Network> n = NetworkRegistry().load(ar);
  SDN_ASSERT(n);
  shared_ptr<StackedNetwork> sn = dynamic_pointer_cast<StackedNetwork>(n);
  SDN_ASSERT(sn);

  shared_ptr<Dataset> d = DatasetRegistry().make("mnist_train");
  SDN_ASSERT(d);
  shared_ptr<BasicImageCollection> bic = dynamic_pointer_cast<BasicImageCollection>(d);
  SDN_ASSERT(bic);
  bic->pyramidDown();

  //Logistic l;
  //l.setPosClassWeight(1.0);

  SimpleClassifier c;

  shared_ptr<StorageNode> h = sn->getStorageNode(sn->numLayers() /2 );
  
  SDN_ASSERT(h);

  #define NUM_TRAIN_EXAMPLES 500
  #define NUM_TEST_EXAMPLES 500

  Matrix Xtr, Xte;
  Vector Ytr, Yte;
  Vector Yhat;
  vector<unsigned> Ytrv, Ytev;


  getExamples(n,h,Xtr,Ytr,Ytrv, d,NUM_TRAIN_EXAMPLES);
  getExamples(n,h,Xte,Yte,Ytev, d,NUM_TRAIN_EXAMPLES);

  makeCache(Xtr,Ytr,"train");
  makeCache(Xte,Yte,"test");


  for (unsigned i = 0; i < Xtr.cols(); i++)
    {
      double m = matrixMean(Xtr.col(i));
      Xtr.col(i).cwise() -= m;
      Xte.col(i).cwise() -= m;

      double v = sqrt(matrixVariance(Xtr.col(i)));
      if (v > 0)
	{
	  Xtr.col(i) /= v;
	  Xte.col(i) /= v;
	}
      else
	{
	  SDN_ERROR("Column "<<i<<" is constant");
	}
    }


  cout << "A piece of X after whitening: " << Xtr.block(0,0,4,4) << endl;


  Matrix cov(Matrix::Zero(Xtr.cols(),Xtr.cols()));
  cout << "Calculating cov" << endl;
  matrixCovariance(Xtr, cov);
  cout << "Done" << endl;
  cout << "cov: "<< cov.block(0,0,4,4) << endl;


  c.train(Xtr, Ytrv);
  cout << c.error(Xte, Ytev) << endl;


  //l.train(Xtr, Ytr);
  //l.getProbabilitiesFromMatrices(Xte,Yhat);

  //cout << l.mse(Yte,Yhat) << endl;
  

  return 0;
}
