/// \author Ian Goodfellow
/// \file sdnview.cc
/// \brief Loads a network file and visualizes its weights
#include "sdn/old.h"
#include "sdn/base.h"
#include <string>
using namespace std;
using namespace sdn;
using namespace boost;

void usage()
{
  SDN_MSG("Usage: ./sdnview [options] <filepath> " << endl
	  << "Loads network in <filepath> and shows visualizations of its weights" << endl
	  << "options: "<<endl
	  << "-c <config.xml>\t\tUse config.xml to configure the application and visualizer"<<endl
	  << "-o <output dir>\t\tCreate output directory and save visualization images to it" << endl
	  << "-d\t\tPrint SDN_DEBUG messages"<<endl
	  << "-v\t\tPrint SDN_VERBOSE messages");
}


class SDNViewParser : public ConfigParser
  {
  public:
    SDNViewParser() 
    {
      firstNode = true;
      p.addInt("mag",1);
      weightVisualizations.push_back(VisualizationRegistry().make("sdn::LinearVisualization") );
    }
    virtual ~SDNViewParser() {}

    ///Tells its parent ConfigListener which topics to listen to
    virtual void topics(std::vector<std::string> & topics)
    {
      topics.push_back("sdnview");
    }
    
    virtual void usage(std::ostream & os) const
    {
      os << "<sdnview " << endl 
	 << "\tmag=\"<int>\" Magnify displayed patches by this factor" << endl
      << "<UseVisualizations> </UseVisualizations>   Runs the enclosed weight visualizations. Default is to run LinearVisualization. " << endl
      <<" />" << endl;
    }
    
    ///Parse and take action on a message recieved via the listener
    virtual void readConfig(XMLNode & node)
    {
      const char * rootname = node.getName();
      SDN_ASSERT(rootname);
      string rootnamestr(rootname);
      SDN_ASSERT(rootnamestr=="sdnview");

      p.parse(node);

      //The first time we see a node requesting visualizations, override the default
      if (firstNode && node.nChildNode() != 0)
	{
	  firstNode = false;
	  weightVisualizations.clear();
	}

      for (int i = 0; i < node.nChildNode(); i++)
	{
	  XMLNode n = node.getChildNode(i);
	  const char * name = n.getName();
	  SDN_ASSERT(name);
	  string namestr(name);
	  SDN_ASSERT(namestr=="UseVisualizations");
	  for (int j = 0; j < n.nChildNode(); j++)
	    {
	      XMLNode  nc = n.getChildNode(j);
	      weightVisualizations.push_back(VisualizationRegistry().load(nc));
	    }
	}
    }

    AttributeParser p;
    vector<shared_ptr<Visualization> > weightVisualizations;
    bool firstNode;
  };


int main(int argc, char ** argv)
{
  bool debug = false;
  bool verbose = false;
  string filepath;
  bool saveImages = false;
  string outputDir;

  for (int i = 1; i < argc; i++)
    {
      string arg = string(argv[i]);
      if (arg == "-v")
	{
	  verbose = true;
	}
      else if (arg == "-d")
	{
	  debug = true;
	}
      else if (arg == "-c")
	{
	  i++;
	  SDN_ASSERT(i < argc);
	  ConfigManager::get().configure(argv[i]);
	}
      else if (arg == "-o")
	{
	  saveImages = true;
	  i++;
	  SDN_ASSERT(i < argc);
	  outputDir = argv[i];
	}
      else
	{
	  if (filepath != "")
	    {
	      usage();
	      exit(-1);
	    }
	  filepath = arg;
	}
    }

  if (filepath == "")
    {
      usage();
      exit(-1);
    }

  if (verbose)
    Logger::setLogLevel(LOGLEVEL_VERBOSE);
  if (debug)
    Logger::setLogLevel(LOGLEVEL_DEBUG);

  shared_ptr<Network> p = Network::load(filepath.c_str());

  
  shared_ptr<SDNViewParser> parser( new SDNViewParser());
  ConfigListener listener(parser);


  int sizeMult = parser->p.getInt("mag").value;


  if (saveImages)
    {
      mkdir(outputDir.c_str());
    }

  for (unsigned v = 0; v < parser->weightVisualizations.size(); v++)
    {
      shared_ptr<Visualization> wv = parser->weightVisualizations[v];

      vector<VisImageSet> imageSets = wv->makeVisualizations(p);
      
      for (unsigned i = 0; i < imageSets.size(); i++)
	{
	  int n = imageSets[i].images.size();
	  int g = imageSets[i].groupSize;
	  
	  int cols = int(sqrt(double(n)));
	  while (cols % g != 0)
	    {
	      cols++;
	    }
	  int rows = n/cols;
	  if (rows * cols < n)
	    {
	      rows++;
	    }

	  if (imageSets[i].showUnnormalized)
	  {
	    PatchDisplay v(imageSets[i].title+" (no normalization)", imageSets[i].images[0].rows(),
			 imageSets[i].images[0].cols(), sizeMult,
			 cols, rows);
	    
	    for (unsigned j = 0; j < n; j++)
	      {
		v.addPatch(imageSets[i].images[j], 0,0);
	      }
	    
	    if (saveImages)
	      {
		v.saveImage((outputDir+"/"+ v.getFileSafeName()+".jpg").c_str());
	      }
	    else
	      {
		v.enableWindow();
		v.wait(0);
	      }
	  }
	  
	  if (imageSets[i].showNormalized)
	  {
	    PatchDisplay v(imageSets[i].title+" (normalized)", imageSets[i].images[0].rows(),
			 imageSets[i].images[0].cols(), sizeMult,
			 cols, rows);
	    
	    for (unsigned j = 0; j < n; j++)
	      {
		v.addPatch(imageSets[i].images[j], 1,1);
	      }
	    
	    if (saveImages)
	      {
		v.saveImage((outputDir+"/"+ v.getFileSafeName()+".jpg").c_str());
	      }
	    else
	      {
		v.enableWindow();
		v.wait(0);
	      }
	  }
	}
    }

  return 0;
}
