//Author: Ian Goodfellow <ia3n@cs.stanford.edu>

/**
 \file train.cc
 \brief Executable for training networks
*/

#include <cv.h>
#include <fstream>
#include <highgui.h>
#include <iostream>
#include <sstream>
#include "tclap/CmdLine.h"
#include "train.h"
#include "sdn/base.h"

using namespace std;
using namespace TCLAP;
using namespace boost;
using namespace sdn;

int main(int argc, char *argv[])
{
  //Get options from command line
  CommandLineHelper cmd;
  parseCommandLine(argc, argv, cmd);

  //Load additional config data
  if (cmd.config != "") {
    ConfigManager::get().configure(cmd.config.c_str(), false);
  }
  
  //Turn on debugging or verbose information if requested
  if (cmd.debug)
    sdn::Logger::setLogLevel(LOGLEVEL_DEBUG);
  else if (cmd.verbose)
    sdn::Logger::setLogLevel(LOGLEVEL_VERBOSE);

  //Print out values of important command line settings
  SDN_VERBOSE("dataset: "<<cmd.dataset);


  //Get dataset
  SDN_VERBOSE("Loading data...");
  shared_ptr<Dataset> data =  DatasetRegistry().make(cmd.dataset);




  if (!data)
    {
      SDN_ERROR("Could not load dataset with name "<<cmd.dataset);
      string info = "Available datasets include: ";
      vector<string> datasets;
      DatasetRegistry().getInstanceNames(datasets);
      for (unsigned i = 0; i < datasets.size(); i++)
	info += "\n\t" + datasets[i];
      SDN_MSG(info);
    }


  ///\todo-- configure the dynamic range to the actual correct value, do it in a more elegant way
  data->setDynamicRange(-0.6,0.6);

  //Get network
  
  shared_ptr<Network> network = NetworkRegistry().make(cmd.type);

  ///\todo-- support other options such as loading from a file,
  // growing from a smaller network


  if (!network)
    {
      SDN_FATAL("Could not create network of type "<<cmd.type);
    }

 

  
  ///\todo-- optionally set up gui feedback to view
  // patches/reconstructions during training
  ///\todo-- optionally set up gui feedback to view weights visualizations
  // during training


  ///\todo-- convert log file to a Record or Journal class
  /*
  char logFilename[100];
  sprintf(logFilename, "output/log-%s.txt", experimentName.c_str()); 
  fstream logFile(logFilename, ios::out); 
  assert(logFile.is_open()); 


  logFile << "<head> " << "Iteration" << " " 
	      << "Sparsity_penalty_pluse_error" << " " 
	      <<  "Error" << " " 
	      << "Sparsity_penalty" << " " 
	      << "Sparsity_stats_0" << " " << "Sparsity_stats_1" << " " << "Sparsity_stats_2" << " "
	      << "Sparsity_stats_3" << " " << "Sparsity_stats_4" << " " << "Sparsity_stats_5" << " "
	  << "Time ";

  logFile << "<command> ";

  for (int i = 0; i < argc; i++)
      logFile << argv[i] << " ";

  logFile << "</command> ";

    logFile <<"</head>"
	      << std::endl; 

    logFile.close(); // we've got to close the log file except while writing to it, or it won't be visible to other computers
  */


  network->registerDataset(data);

  //Set up the runtime manager
  TrainRuntimeManager mgr(network, data);

 



  //mgr.registerCallback(shared_ptr<TrainCallback>(
  //new SaveFileCallback("quoctestfile")));

  /// \todo allow user to turn this off/set how frequently it should run
  //mgr.registerCallback(shared_ptr<TrainCallback>(new PatchVizCallback()));

  /// \todo allow user to set how frequently this should run
  mgr.registerCallback(shared_ptr<TrainCallback>(new RecordCallback()));

  

  //Run
  mgr.run();
 
  return 0;
}

void parseCommandLine(int argc, char *argv[], CommandLineHelper & cmd)
		      /*                      string &experimentName, Scalar *learningRate, Scalar *biasLearningRate, Scalar *weightDecay,
		      Scalar *desiredSparsityMin, Scalar *desiredSparsityMax, 
                      Scalar *sparsityPenaltyMult, int *nhidden, int *patchSize, int *imageReduce, 
		      bool *showWindows, int *seed1, int *seed2, int * saveIterations, string &  subNetwork, int * patchStep, int * numStep, bool * learnAll, string & dataset, int * displayIterations, Scalar & initMag, int & disableDeepbackprop, string & init, int & change) */
{
  try {
    CmdLine parser("Stanford Deep Network Library training program", ' ', SDN_MAKE_STRING(SDN_VERSION) );

    //ValueArg<std::string> nameArg("n", "name", "Experiment name", true, "INVALID", "string", parser);

    string dataDesc = "Dataset to use: ";
    vector<string> datasets;
    DatasetRegistry().getInstanceNames(datasets);
    if (datasets.size())
      {
	for (unsigned i = 0; i < datasets.size(); i++)
	  {
	    dataDesc += datasets[i];
	    if (i < datasets.size() - 1)
	      dataDesc += " | ";
	  }
      }
    else
      {
	dataDesc += "NO DATASETS REGISTERED.\n";
	dataDesc += "Edit " + configPath() + "/main.xml to register datasets.";
      }


    ValueArg<string> datasetArg("s", "data[s]et", dataDesc, false, "mnist_train", "string", parser);
    ValueArg<string> typeArg("n", "type", "Type of network", false, "sdn::StackedNetwork", "string", parser);
    ValueArg<string> configArg("c", "config", "Configuration file", false, "", "string", parser);
    //ValueArg<std::string> subNetworkArg("", "sub", "Subnetwork file", false, "", "string", parser);
    //ValueArg<std::string> initArg("", "init", "Initialize from file, not randomly", false, "", "string", parser);
    //ValueArg<Scalar> learningRateArg("a", "alpha", "Learning rate", false, 0.0003, "Scalar", parser);
    //ValueArg<Scalar> biasLearningRateArg("b", "beta", "Bias learning rate", false, -1, "Scalar", parser);
    //ValueArg<Scalar> weightDecayArg("w", "weightdecay", "Weight decay", false, 0.001, "Scalar", parser);
    //ValueArg<Scalar> desiredSparsityArg("s", "sparsity", "Desired sparsity (upperbound)", false, 0.02, "Scalar", parser);
    //ValueArg<Scalar> sparsityGapArg("z", "spzone", "Sparsity deadzone", false, 0.01, "Scalar", parser);
    //ValueArg<Scalar> sparsityPenaltyMultArg("p", "sparsitypenalty", "Sparsity penalty", false, 10, "Scalar", parser);
    //ValueArg<Scalar> initMagArg("", "initmag", "Weights are initialized uniformly in [-initMag,initMag]", false, 0.05, "Scalar", parser);
    //ValueArg<int> nhiddenArg("", "nhidden", "Number of hidden units", false, 100, "integer", parser);
    //ValueArg<int> patchSizeArg("", "patchsize", "Patch size", false, 10, "integer", parser);
    //ValueArg<int> imageReduceArg("", "imagereduce", "Image reduction factor", false, 1, "integer", parser);
    //ValueArg<int> changeArg("", "change", "On each iteration, change training image with probability 1/change", false, 1, "integer", parser);

    //ValueArg<int> seedArg("", "seed", "Random number seed", false, 12593, "integer", parser);

    SwitchArg verboseSwitch("v","verbose","print more information",parser,false);
    SwitchArg debugSwitch("d","debug","print debugging information (also implies -v)",parser,false);
    //SwitchArg showWindowsSwitch("g","graphics","Enable graphical output", parser, false);
    //SwitchArg disableDeepbackpropSwitch("","shallow","Use shallow objective (reconstruction of second deepest layer, rather than input)",parser,false);
    //SwitchArg learnAllSwitch("","learnall","Learn weights for all layers, rather than just central layer",parser,false); 
    //ValueArg<int> saveIterationsArg("","saveiter","Save weights/image every saveiter iterations",false,1000000,"integer",parser);
    //ValueArg<int>  displayIterationsArg("","displayiter","Display weights/a new sample reconstruction every saveiter iterations",false,10000,"integer",parser);
    //ValueArg<int> patchStepArg("","patchstep","Amount to step receptive field of subnetwork's top hidden layer",false,1,"integer",parser);
    //ValueArg<int> numStepArg("","numstep","Number of times to instantiate subnetwork's top hidden layer in each direction",false,1,"integer",parser);

    parser.parse( argc, argv );

    //experimentName = nameArg.getValue();
    cmd.dataset = datasetArg.getValue();
    cmd.type = typeArg.getValue();
    cmd.config = configArg.getValue();
    //subNetwork = subNetworkArg.getValue();
    //init = initArg.getValue();
    //*learningRate = learningRateArg.getValue();
    //*biasLearningRate = biasLearningRateArg.getValue();
    //if (*biasLearningRate < 0)
    //*biasLearningRate = *learningRate; 
    //*weightDecay = weightDecayArg.getValue(); 

     // --sparsity specifies sparsity max. 
     // --sparsitygap specifies the difference between desiredSparsityMin and desiredSparsityMax
    //*desiredSparsityMax = desiredSparsityArg.getValue();
    //*desiredSparsityMin = *desiredSparsityMax - sparsityGapArg.getValue(); 

    //*sparsityPenaltyMult = sparsityPenaltyMultArg.getValue(); 
    //initMag = initMagArg.getValue();
    //*nhidden = nhiddenArg.getValue(); 
    //*patchSize = patchSizeArg.getValue(); 
    //*imageReduce = imageReduceArg.getValue(); 
    //change = changeArg.getValue();
    cmd.verbose = verboseSwitch.getValue();
    cmd.debug = debugSwitch.getValue();

    //*showWindows = showWindowsSwitch.getValue();
    //disableDeepbackprop = disableDeepbackpropSwitch.getValue();
    //*learnAll = learnAllSwitch.getValue();
    //*saveIterations = saveIterationsArg.getValue();
    //*displayIterations = displayIterationsArg.getValue();
    //*patchStep = patchStepArg.getValue();
    //*numStep = numStepArg.getValue();

    //*seed1 = seedArg.getValue();
    //*seed2 = 6948;


  } catch (ArgException &e) { 
    SDN_ERROR("error: " << e.error() << " for arg " << e.argId()); 
    SDN_FATAL("parseCommandLine failed.");
  }

  //assert(*desiredSparsityMin >= 0);

  /*
  cout << "Experiment name: " << experimentName << "\n"
       << "Learning rate: " << *learningRate << "\n"
       << "Bias learning rate: " << *biasLearningRate << "\n"
       << "Weight decay: " << *weightDecay << "\n"
       << "Desired sparsity: Min=" << *desiredSparsityMin << " Max=" << *desiredSparsityMax << " \n"
       << "Sparsity penalty: " << *sparsityPenaltyMult << "\n"
       << "Hidden units: " << *nhidden << "\n"
       << "Patch size: " << *patchSize << "x" << *patchSize << "\n"
       << "Image reduce: " << *imageReduce << "\n"
       << "Show windows: " << *showWindows << "\n"
       << "Disable deep backprop: " << disableDeepbackprop << "\n"
       << "Seeds: " << *seed1 << ", " << *seed2 << std::endl; 
  */
}



