//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>

/**
 \file pcaw-dataset.cc
 \brief Executable for computing the PCA whitening matrix
*/

#include <cstdio>
#include <iostream>
#include <Eigen/Array>
#include <Eigen/QR>
#include "tclap/CmdLine.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/stable.h"
#include "sdn/registry.h"

std::string datasetName, dataMeanOutputFile, whiteningMatOutputFile;
bool verbose;
bool debug;
int width,height,nexamples,tDim;

void parseCommandLine(int argc, char *argv[]);

/// Helper Structure for Sorting EigenValues
struct sortEigenStruct {
  int   index;
  float value;
  sortEigenStruct(int i, float f) {
    index = i; value = f;
  }
};

bool operator<(const sortEigenStruct& a, const sortEigenStruct& b) {
  return a.value > b.value;   // sort descending
}


int main(int argc, char *argv[])
{

  //Get options from command line
  parseCommandLine(argc, argv);

  //Turn on debugging or verbose information if requested
  if (debug)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_DEBUG);
  else if (verbose)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_VERBOSE);


  //Load Dataset
  boost::shared_ptr<sdn::Dataset> dataset =  sdn::DatasetRegistry().make(datasetName);
  if (!dataset) {
    SDN_ERROR("Could not load dataset with name " << datasetName);
    // ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
    SDN_FATAL("Quitting");
  }
  dataset->requestExampleSize(height, width);
  
  //Open dataMeanOutputFile if any
  std::ofstream fdMean;
  if (dataMeanOutputFile != "") {
    fdMean.open(dataMeanOutputFile.c_str());
    SDN_ASSERT(fdMean.good());
  }  
  
  //Open whiteningMatOutputFile if any
  std::ofstream fWhite;
  if (whiteningMatOutputFile != "") {
    fWhite.open(whiteningMatOutputFile.c_str());
    SDN_ASSERT(fWhite.good());
  }  
  
  //Convert Dataset to BasicImageCollection
  boost::shared_ptr<sdn::BasicImageCollection> bic =
    boost::dynamic_pointer_cast<sdn::BasicImageCollection> (dataset);

  SDN_ASSERT(bic);
  
  nexamples = bic->getImgRef().size();

  //Run N examples
  sdn::Matrix example = sdn::Matrix::Zero(height*width, 1);
  
  sdn::NormalizeInputBlock normInput;
  
  sdn::Matrix data(height*width, nexamples);
  int errors =0;
  for(int i=0;i<nexamples;i++) {
    
    sdn::matrixSubColToMatrix(bic->getImgRef()[i],0,0,example); 

    // Add example to data
    data.col(i) = example;

  }

  std::cout << "Data Read Complete\n";
  
  // Subtract Mean
  sdn::Matrix mean = data.rowwise().sum() / data.cols();
  data -= mean * sdn::Matrix::Ones(1, data.cols());

  // Run PCA
  sdn::Matrix &m = data;
  
  Eigen::EigenSolver<sdn::Matrix> eig(m*m.transpose() / (m.cols()));
  sdn::Matrix eigenVec = eig.pseudoEigenvectors();
  sdn::Matrix eigenVal = eig.pseudoEigenvalueMatrix();
  
  // Sort the eigenValues
  std::vector<sortEigenStruct> eigenValArray;
  for (int i = 0 ; i < eigenVal.rows() ; i++) {
    eigenValArray.push_back(sortEigenStruct(i, eigenVal(i,i)));
  }
  std::sort(eigenValArray.begin(), eigenValArray.end());
  
  // V is going to be our projection matrix
  // Note that each column of eigenVec is a eigenvector
  sdn::Matrix reduceV = sdn::Matrix::Zero(eigenVec.rows(), tDim);
  
  double eiSum=0;
  
  // Pick the biggest X eigen stuff
  for (int i=0;i<tDim;i++) {
    reduceV.col(i) = eigenVec.col(eigenValArray[i].index) / sqrt(eigenValArray[i].value);
    eiSum += eigenValArray[i].value;
  }

  std::cout << "Eigenvalue % " << eiSum / eigenVal.sum() << "\n";
  fflush(stdout);
  
  // Transpose the Matrix
  sdn::Matrix V = reduceV.transpose();
  
  // --------------------------------
  /*
  // Compute the Inverse Matrix 
  sdn::Matrix invV = V.transpose();
  for(int i=0;i<invV.cols();i++) {
    invV.col(i) /= invV.col(i).cwise().square().sum();
  }
  
  sdn::Matrix dataNew = (V * data);
  dataNew = invV * dataNew;
  
  // Check Output Difference (for a random example)
  
  for(int i=0;i<dataNew.rows();i++)
    std::cout << data(i,3) << " " << dataNew(i,3) << "\n";
  */
  
  sdn::serial::BinOutArchive aroutW(fWhite);
  sdn::serializeMatrix(aroutW, V);
  
  sdn::serial::BinOutArchive aroutM(fdMean);
  sdn::serializeMatrix(aroutM, mean);
  
  fdMean.close();
  fWhite.close();
    
  
  return 0;
}

void parseCommandLine(int argc, char *argv[]) {
  try {
    TCLAP::CmdLine parser("Stanford Deep Network Library training program", ' ', SDN_MAKE_STRING(SDN_VERSION) );

    TCLAP::ValueArg<std::string> dataMeanOutputFileArg("m", "meanfile", "DataMeanOutputFile file", true, "", "string", parser);
    TCLAP::ValueArg<std::string> whiteningMatOutputFileArg("i", "whitefile", "WhiteningMatOutputFile file", true, "", "string", parser);
    
    TCLAP::ValueArg<std::string> datasetArg("s", "dataset", "Dataset to load", true, "", "string", parser);
    

    TCLAP::ValueArg<int> widthArg("w", "width", "Input Width", false, 28, "int", parser);
    TCLAP::ValueArg<int> heightArg("t", "height", "Input Height", false, 28, "int", parser);
 
    TCLAP::ValueArg<int> tDimArg("a", "targetdim", "Target Dimensions", true, 10, "int", parser);
    
    TCLAP::SwitchArg verboseSwitch("v","verbose","print more information",parser,false);
    TCLAP::SwitchArg debugSwitch("d","debug","print debugging information (also implies -v)",parser,false);
    
    parser.parse( argc, argv );

    whiteningMatOutputFile = whiteningMatOutputFileArg.getValue();
    dataMeanOutputFile = dataMeanOutputFileArg.getValue();
    verbose = verboseSwitch.getValue();
    debug = debugSwitch.getValue();
    datasetName = datasetArg.getValue();

    width = widthArg.getValue();
    height = heightArg.getValue();
    
    tDim = tDimArg.getValue();

  } catch (TCLAP::ArgException &e) { 
    SDN_ERROR("error: " << e.error() << " for arg " << e.argId()); 
    SDN_FATAL("parseCommandLine failed.");
  }


}



