//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>

/**
 \file run-autotunenet.cc
 \brief Executable for running auto tuned nets
*/

#include <cstdio>
#include <iostream>
#include "tclap/CmdLine.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/stable.h"
#include "sdn/registry.h"
#include "../lib/projects/jngiam/networks/autoTuneNetwork.h"

std::string networkFile, datasetName, libsvmFile, reconFile;
bool verbose;
bool debug;
int width,height,nexamples;

void parseCommandLine(int argc, char *argv[]);

int main(int argc, char *argv[])
{

  //Get options from command line
  parseCommandLine(argc, argv);

  //Turn on debugging or verbose information if requested
  if (debug)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_DEBUG);
  else if (verbose)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_VERBOSE);


  //Load network
  std::ifstream fin;
  fin.open(networkFile.c_str(), std::ios::binary);
  sdn::serial::BinInArchive ar(fin);
  boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
  fin.close();

  SDN_ASSERT(net);
  
  //Load as auto tune net
  boost::shared_ptr<sdn::AutoTuneNetwork> atnet = 
    boost::dynamic_pointer_cast<sdn::AutoTuneNetwork> (net);
  
  //Load Dataset
  boost::shared_ptr<sdn::Dataset> dataset =  sdn::DatasetRegistry().make(datasetName);
  if (!dataset) {
    SDN_ERROR("Could not load dataset with name " << datasetName);
    // ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
    SDN_FATAL("Quitting");
  }
  dataset->requestExampleSize(height, width);
  
  //Open libsvmFile if any
  std::ofstream fout;
  if (libsvmFile != "") {
    fout.open(libsvmFile.c_str());
    SDN_ASSERT(fout.good());
  }  
  
  //Open libsvmFile if any
  std::ofstream foutr;
  if (reconFile != "") {
    foutr.open(reconFile.c_str());
    SDN_ASSERT(foutr.good());
  }  
  
  //Convert Dataset to BasicImageCollection
  boost::shared_ptr<sdn::BasicImageCollection> bic =
    boost::dynamic_pointer_cast<sdn::BasicImageCollection> (dataset);

  SDN_ASSERT(bic);
  
  nexamples = bic->getImgRef().size();

  //Run N examples
  sdn::Matrix example = sdn::Matrix::Zero(height*width, 1);
  sdn::NormalizeInputBlock normInput;
  
  int errors =0;
  for(int i=0;i<nexamples;i++) {
    
    sdn::matrixSubColToMatrix(bic->getImgRef()[i],0,0,example); 
    int label = bic->getLabelRef()[i];

    normInput.normalizePatch(example);
    atnet->encode(example);
    atnet->forwardPropClassifyHidden();
    
    int pred = 0;
    for (int j=0;j<atnet->getPrediction().size();j++)
      if (atnet->getPrediction()[j] > atnet->getPrediction()[pred])
        pred = j;
    
    if (pred != label)
      errors++;

    if (fout.good()) {
      fout << label << " ";
      for (int j=0;j<atnet->getHiddenRepresentation().size() ;j++)
        fout << (j+1) << ":" << atnet->getHiddenRepresentation()[j] << " ";
      fout << "\n";
    }
    
    if (foutr.good()) {
      foutr << atnet->decode(atnet->getHiddenRepresentation()).transpose() << "\n";
    }
    
  }

  std::cout << double(errors)/(nexamples) << "\n";
  
  fout.close();
  foutr.close();
    
  
  
  

  
  return 0;
}

void parseCommandLine(int argc, char *argv[]) {
  try {
    TCLAP::CmdLine parser("Stanford Deep Network Library training program", ' ', SDN_MAKE_STRING(SDN_VERSION) );

    TCLAP::ValueArg<std::string> networkArg("n", "network", "Network Serialized (binary) file", true, "", "string", parser);
    TCLAP::ValueArg<std::string> datasetArg("s", "dataset", "Dataset to load", true, "", "string", parser);
    TCLAP::ValueArg<std::string> reconArg("r", "reconstruction", "Save Reconstructions", false, "", "string", parser);
    

    TCLAP::ValueArg<int> widthArg("w", "width", "Input Width", false, 28, "int", parser);
    TCLAP::ValueArg<int> heightArg("t", "height", "Input Height", false, 28, "int", parser);
    
    TCLAP::ValueArg<std::string> libsvmArg("m", "libsvm", "Generate LIBSVM Compatible File", false, "", "string", parser);
    
    TCLAP::SwitchArg verboseSwitch("v","verbose","print more information",parser,false);
    TCLAP::SwitchArg debugSwitch("d","debug","print debugging information (also implies -v)",parser,false);
    parser.parse( argc, argv );

    networkFile = networkArg.getValue();
    verbose = verboseSwitch.getValue();
    debug = debugSwitch.getValue();
    datasetName = datasetArg.getValue();
    libsvmFile = libsvmArg.getValue();
    reconFile = reconArg.getValue();

    width = widthArg.getValue();
    height = heightArg.getValue();

  } catch (TCLAP::ArgException &e) { 
    SDN_ERROR("error: " << e.error() << " for arg " << e.argId()); 
    SDN_FATAL("parseCommandLine failed.");
  }


}



