//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>

/**
 \file viz-network.cc
 \brief Executable for visualizing some networks networks
*/

#include <cstdio>
#include <iostream>
#include "tclap/CmdLine.h"
#include "sdn/base.h"
#include "sdn/stable.h"
#include "sdn/registry.h"

std::string networkFile;
bool verbose;
bool debug;

void parseCommandLine(int argc, char *argv[]);

int main(int argc, char *argv[])
{

  //Get options from command line
  parseCommandLine(argc, argv);

  //Turn on debugging or verbose information if requested
  if (debug)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_DEBUG);
  else if (verbose)
    sdn::Logger::setLogLevel(sdn::LOGLEVEL_VERBOSE);


  //Load network
  std::ifstream fin;
  fin.open(networkFile.c_str(), std::ios::binary);
  sdn::serial::BinInArchive ar(fin);
  boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
  fin.close();
  
  if (net) {
    
    SDN_MSG( "Loaded a " << net->className() ); 
    
    // Ideally, we will have a functor style code here that accepts
    // any network
    
    // But to just test/demonstrate how things work, we will just
    // assume this is a composable network and look for a 
    // sdn::SparseAffineTanhBlock
    
    boost::shared_ptr<sdn::ComposableAutoEncoderNetwork> cnet = 
      boost::dynamic_pointer_cast<sdn::ComposableAutoEncoderNetwork> (net);
    boost::shared_ptr<sdn::StandardAutoEncoderNetwork> anet = 
      boost::dynamic_pointer_cast<sdn::StandardAutoEncoderNetwork> (net);
    
    sdn::Matrix weights;
      
    SDN_ASSERT_MSG(cnet || anet, "This demo visualization app only works with ComposableAutoEncoderNetworks or StandardAutoEncoderNetwork");
    
    if (cnet) {

      int foundBlock = -1;
      for(int i=0;i<cnet->blocks.size();i++) {
        SDN_MSG ( "Found a " << cnet->blocks[i]->className() << " block!" );
        if (foundBlock < 0 && cnet->blocks[i]->className() == "sdn::SparseAffineTanhBlock")
          foundBlock = i;
      }
      
      SDN_MSG( "Visualizing the " << cnet->blocks[foundBlock]->className() << " at position " << foundBlock );
      
      boost::shared_ptr<sdn::SparseAffineTanhBlock> sath = 
        boost::dynamic_pointer_cast<sdn::SparseAffineTanhBlock> (cnet->blocks[foundBlock]);

      SDN_ASSERT(sath);
      
      weights = sath->transferBlock.affineB.weights;
      
    } else if (anet) {
      
      SDN_MSG( "Visualizing the encoding weights");
      weights = anet->affineEncoder.weights;
      
    }
    
    
    int patchHeight = sqrt(weights.cols());      // assume square input images
    int displayGridWidth = sqrt(weights.rows()); // try to fit all nicely
    
    sdn::VizWeightMatrixBlock tmpViz;
    tmpViz.vizMatrix(weights, 
                     patchHeight, 
                     displayGridWidth, 
                     "",                         // don't save
                     true);                      // true to showWindow

    //getchar();

  }
  
  
  return 0;
}

void parseCommandLine(int argc, char *argv[]) {
  try {
    TCLAP::CmdLine parser("Stanford Deep Network Library training program", ' ', SDN_MAKE_STRING(SDN_VERSION) );

    TCLAP::ValueArg<std::string> networkArg("n", "network", "Network Serialized (binary) file", true, "", "string", parser);
    
    TCLAP::SwitchArg verboseSwitch("v","verbose","print more information",parser,false);
    TCLAP::SwitchArg debugSwitch("d","debug","print debugging information (also implies -v)",parser,false);
    parser.parse( argc, argv );

    networkFile = networkArg.getValue();
    verbose = verboseSwitch.getValue();
    debug = debugSwitch.getValue();


  } catch (TCLAP::ArgException &e) { 
    SDN_ERROR("error: " << e.error() << " for arg " << e.argId()); 
    SDN_FATAL("parseCommandLine failed.");
  }


}



