//Author: Ian Goodfellow
//File: Archive
//Description: Class for performing serialization/deserialization

#pragma once
#include <stack>
#include <string>
#include <ostream>
#include <map>
#include "xmlUtils.h"
#include "logger.h"
#include "parser.h"
#include "matrix.h"

namespace sdn
{
  namespace serial
  {

    /// Format descriptions, made available so it is easy to tell other classes which archive class to use-- i.e. network.save("file.dat",XML)
    enum Format
      {
	BIN,
	XML
      };



    typedef unsigned long long Vaddr;
    




    /** Abstract base class for serialization. Subclasses determine
	whether the archive is used for reading or for writing, and
	the format (currently only XML or binary are supported).
        Designed to support the ability to read archives in out-of-date
        formats but not to write them (being able to write them would
	require that the classes to be serialized implement a lot more
	control logic). Supports reading/writing of objects whose
        derived type is not known at compile time. Writing out multiple
        pointers to the same object, then reading them back in will
        result in multiple pointers to the same object.
    Caveats: Anything that you serialize as an "attribute" should not
    be pointed to by anything other than a single pointer in the class
    that owns it. Class members that are part of the actual memory
    layout of the class (not something elsewhere on the heap that the
    class just points to) should not be pointed to by other things.



    Examples:

    class A
    {
      int * a;
    };

    A a1, a2;
    a1.a = new int;
    a2.a = a1.a;

    Here, a is just an int, so you may be tempted to serialize it as
    an attribute. However, only class objects are tracked by the
    archive to prevent duplicates (to avoid the overhead of
    tracking every single primitive) so serializing a as an attribute
    would mean that a1 and a2 end up pointing to two separate ints
    if you serialized and deserialized them.

    
    clas C
    {
     ...;
    };

    class B
    {
      C c;
    };

    B b;
    C * c = & b.c;

    Suppose you now serialize c, then b. When you deserialize them,
    deserialzing c will allocate a C on the heap. Then you will go
    to deserialize b. b can't be deserialized because its member c
    already has a fixed location on the heap, but the space for b
    can't be malloced to make it align with it.

    If you don't do these two things, you should be ok.

    Edit: one more, but you wouldn't do it anyway:

    don't use references in the class body:

    class A
    {
      int & a;
    };

    they're essentially const pointers that can't be const-casted into
    being non-const so the serialization code is unable to read in
    their value from the archive.

*/
    class Archive
    {
    public:
      Archive();
      virtual ~Archive();
      
      /** Announces that a base class is about to be serialized.
	  Should be followed by the class members and a call to 
	  endBase() */
      int beginBase(void * addr, int version, std::string id);
      
      /** Announces the end of the serialization of a base class */
      void endBase();
      
      /** Announces that the super-class of the current object is
	  about to be serialized. This call must follow every call
	  to beginDerived() and must be closed by a call to endSuper()
	  \return The version of the super class that the archive expects
	  to be serialized. When reading, this will be the version of
	  the superclass stored in the archive. When writing, it will
	  be the latest version. */
      void beginSuper();
      
      void endSuper();
      
      /** Announces that the serialization of a class that is not a base
	  class is beginning. 
	    Should be followed by super class serialization, then class 
	    members and a call to endDerived() */
      int beginDerived(void * addr, int version, std::string  id);
      
      /** Announces the end of the serialization of a derived class */
      void endDerived();
      
      /** Serializes an attribute of the class currently being serialized.
       (attribute names are only used for XML) */
      void attribute(std::string name, unsigned & value);

      void attribute(std::string name, double & value);

      void attribute(std::string name, bool & value);

      void attribute(std::string name, int & value);

      void attribute(std::string name, std::string & value);

      void attribute(std::string name, float & value);

      ///Uses an array of Scalars as the "custom content" of the class
      //(this just means in XML, it will go in the node body, and that
      //it is mutually exclusive with other custom content calls)
      void scalarArray(Scalar * data, unsigned numFloats);


      /// Returns true if the archive is an InArchive
      virtual bool in() = 0;

    protected:


      virtual int beginBase_(void * addr, int version, std::string id) = 0;
      virtual void endBase_() = 0;
      virtual void beginSuper_() = 0;
      virtual void endSuper_() = 0;
      virtual int beginDerived_(void * addr, int version, std::string id)= 0;
      virtual void endDerived_() = 0;
      virtual void attribute_(std::string name, unsigned & value) = 0;
      virtual void attribute_(std::string name, double & value) = 0;
      virtual void attribute_(std::string name, bool & value) = 0;

      // virtual void attribute_(std::string name, int & value) = 0;    -- cast to unsigned and use that
      virtual void attribute_(std::string name, std::string & value) = 0;
      virtual void attribute_(std::string name, float & value) = 0;

      virtual void scalarArray_(Scalar * data, unsigned numFloats) = 0;

      Parser p_;
    };

  }




}

#define ARCHIVE_STRING "sdnArchive"

