/// \author Ian Goodfellow

#include "attributeParser.h"
#include "logger.h"
#include <cstring>
#include "xmlParser/xmlParser.h"
#include <sstream>

using namespace sdn;
using namespace std;

#define GET_ATTRIB(lower, upper) \
  Attribute<lower> AttributeParser::get ##upper(const std::string & name) \
{ \
  map<string, Attribute<lower> >::iterator i =  SDN_ATTRIB_MAP_NAME(lower).find(name); \
  if (i == SDN_ATTRIB_MAP_NAME(lower).end())				\
    { \
      SDN_FATAL("Parser does not contain an entry for "<<name); \
    } \
  return i->second; \
}

GET_ATTRIB(bool, Bool);
GET_ATTRIB(int, Int);
GET_ATTRIB(double, Double);
GET_ATTRIB(attrib_str, String);

#define ADD_ATTRIB_D( lower, upper )  \
  void AttributeParser::add ## upper (const std::string & name, lower defaultVal) \
  {									\
    map<string, Attribute< lower > > & m = SDN_ATTRIB_MAP_NAME( lower); \
    if (alreadyContains_(name)) \
      { \
	SDN_FATAL("AttributeParser already contains attribute called \""<<name<<"\""); \
      } \
    m.insert(make_pair(name, Attribute< lower > ( name, true, defaultVal , false ))); \
	     }

ADD_ATTRIB_D( bool, Bool);
ADD_ATTRIB_D( int, Int);
ADD_ATTRIB_D( double, Double);
ADD_ATTRIB_D( attrib_str, String);

//defaultV is included here because defaultValue is a const field and must be
//initialized even if it is ignored
#define ADD_ATTRIB( lower, upper , defaultV)					\
  void AttributeParser::add ## upper (const std::string & name) \
  {									\
    map<string, Attribute< lower > > & m = SDN_ATTRIB_MAP_NAME( lower); \
    if (alreadyContains_(name)) \
      { \
	SDN_FATAL("AttributeParser already contains attribute called \""<<name<<"\""); \
      } \
    m.insert(make_pair(name, Attribute< lower > ( name, false , defaultV, true)));	\
	     }

ADD_ATTRIB( bool, Bool, false);
ADD_ATTRIB( int, Int, -1);
ADD_ATTRIB( double, Double, -1.0);
ADD_ATTRIB( attrib_str, String, "UNINITIALIZED STRING");

#define IGNORE_ATTRIB( lower, upper, defaultV )					\
  void AttributeParser::ignore ## upper (const std::string & name) \
  {									\
    map<string, Attribute< lower > > & m = SDN_ATTRIB_MAP_NAME( lower); \
    if (alreadyContains_(name)) \
      { \
	SDN_FATAL("AttributeParser already contains attribute called \""<<name<<"\""); \
      } \
    m.insert(make_pair(name, Attribute< lower > ( name, false , defaultV, false)));	\
	     }

IGNORE_ATTRIB( bool, Bool, false);
IGNORE_ATTRIB( int, Int, -1);
IGNORE_ATTRIB( double, Double, -1.0);
IGNORE_ATTRIB( attrib_str, String, "UNINITIALIZED STRING");

bool AttributeParser::alreadyContains_(const std::string & name) const
{
#define SDN_CHECK_MAP(T) if (SDN_ATTRIB_MAP_NAME(T).find(name) != SDN_ATTRIB_MAP_NAME(T).end()) return true;

  SDN_CHECK_MAP(bool);
  SDN_CHECK_MAP(int);
  SDN_CHECK_MAP(double);
  SDN_CHECK_MAP(attrib_str);

  return false;
}

void AttributeParser::parse(XMLNode & node, bool allowUndeclaredAttributes)
{
  //Set all attributes as not having been newly configured
#define SDN_RESET_NEW(T) for (map< string, Attribute<T> >::iterator i =\
				SDN_ATTRIB_MAP_NAME(T).begin(); i != \
				SDN_ATTRIB_MAP_NAME(T).end(); ++i) \
    {							      \
      i->second.newlyConfigured = false; \
    }

  SDN_RESET_NEW(bool);
  SDN_RESET_NEW(int);
  SDN_RESET_NEW(double);
  SDN_RESET_NEW(attrib_str);

  //Parse each declared attribute
  int parsedAttributes = 0;

#define SDN_PARSE_MAP(T, Upper) for (map< string, Attribute<T> >::iterator i = \
				SDN_ATTRIB_MAP_NAME(T).begin(); i != \
				SDN_ATTRIB_MAP_NAME(T).end(); ++i) \
    {							      \
      const char * v = node.getAttribute(i->second.name.c_str()); \
      if (v) \
	{ \
	  parse ## Upper ## _(v, i->second);		\
	  i->second.configured = true; \
	  i->second.newlyConfigured = true; \
	  ++parsedAttributes; \
	} else if (i->second.isRequired) { \
    SDN_FATAL("Node with name " << (node.getName() ? node.getName() : "unknown") << " requires attribute " << i->second.name << " to be specified."); \
  } \
    }
  
  SDN_PARSE_MAP(bool, Bool);
  SDN_PARSE_MAP(int, Int);
  SDN_PARSE_MAP(double, Double);
  SDN_PARSE_MAP(attrib_str, String);
  
  //Check that node doesn't use any attributes we don't know about
  if ( (!allowUndeclaredAttributes) && node.nAttribute() != parsedAttributes)
    {
      stringstream s;
      s << "Node ";
      if (node.getName())
	{
	  s << " with name "<<node.getName();
	}

      s << " contains the following unrecognized attributes: ";

      for (int i = 0; i < node.nAttribute(); i++)
	{
	  if (!alreadyContains_(node.getAttribute(i).lpszName))
	    {
	      s << node.getAttribute(i).lpszName << " ";
	    }
	}

      SDN_FATAL(s.str());
    }
}

void AttributeParser::parseBool_(const char * str, Attribute<bool> & a)
{
  if (!strcmp(str,"1") || ! strcmp(str,"true"))
    {
      a.value = true;
    }
  else if (!strcmp(str,"0") || ! strcmp(str,"false"))
    {
      a.value = false;
    }
  else
    {
      SDN_FATAL("Could not parse "<<str<<" as a bool");
    }
}

void AttributeParser::parseString_(const char * str, Attribute<attrib_str> & a)
{
  a.value = str;
}



#define SDN_GENERAL_PARSE(T, Upper)						\
  void AttributeParser::parse##Upper##_(const char * str, Attribute< T> & a) \
{ \
  string s(str);				\
  stringstream ss(s);				\
						\
  if ((ss >> a.value).fail())			\
    { \
      SDN_FATAL("Could not parse "<<str<<" as a "<<#T); \
    } \
		     \
  while (!ss.fail()) \
    { \
      char ch; \
      ss >> ch; \
      if (! ss.fail() && ch != ' ' && ch != '\t') \
	{ \
	  SDN_FATAL("In addition to the expected "<<#T<<" "<<a.value<<", "<<str<<" also contains "<<ch); \
	}\
    }\
}

SDN_GENERAL_PARSE(int, Int);
SDN_GENERAL_PARSE(double, Double);



void AttributeParser::reset()
{
#define SDN_RESET(T) for (map< string, Attribute<T> >::iterator i =  \
			    SDN_ATTRIB_MAP_NAME(T).begin(); i !=     \
			    SDN_ATTRIB_MAP_NAME(T).end(); ++i)	     \
    {								     \
      i->second.configured = false;			     \
      i->second.newlyConfigured = false;		     \
      if (i->second.hasDefault) \
	i->second.value = i->second.defaultValue; \
    }

  SDN_RESET(int);
  SDN_RESET(bool);
  SDN_RESET(double);
  SDN_RESET(attrib_str);
}
