/// \author Ian Goodfellow

#pragma once

#include <string>
#include <map>


class XMLNode;

namespace sdn
{

template <typename T>
class Attribute
{
  public:
  
  ///If true, this attribute has been configured at some point after the most recent time reset() was called on the enclosing AttributeParser
  bool configured;

  ///If true, this attribute was configured by the last call to parseAttribute
  bool newlyConfigured;

  ///The value of the attribute. If configured is false, this is a junk value
  T value;

  ///If true, has a defined default value.
  const bool hasDefault;

  ///If true, has is required - this overrides default
  const bool isRequired;

  ///The default value for this field. If hasDefault is true, this is a junk value
  const T defaultValue;

  ///The name of this attribute
  const std::string name;
  
  ///Easy value extraction
  operator T() { return value; }


  Attribute<T>(const std::string & n, bool hasD, const T & defaultV, bool isReq) : configured(false), newlyConfigured(false), value(defaultV), hasDefault(hasD), isRequired(isReq), defaultValue(defaultV), name(n)
										 { }
};

 typedef std::string attrib_str;

/**
A class for safely reading attributes from an XMLNode

*/

class AttributeParser
{

 public:

  /**
     Reads the attributes in node. For each of the attributes in the
     parser, their "configured" field is set to true if it is present
     in the node. 

     \param allowUndeclaredAttribues If false (default) will throw an error if an unrecognized attribute is present in node
  */
  void parse(XMLNode & node, bool allowUndeclaredAttributes = false);

  ///Sets all attributes to unconfigured and to their default value
  void reset();

  /// Adds an unconfigured attribute of the specified type, with the specified default value
  void addBool(const std::string & name, bool defaultVal);
  void addInt(const std::string & name, int defaultVal);
  void addDouble(const std::string & name, double defaultVal);
  void addString(const std::string & name, std::string defaultVal);


  void addBool(const std::string & name);
  void addInt(const std::string & name);
  void addDouble(const std::string & name);
  void addString(const std::string & name);

  void ignoreBool(const std::string & name);
  void ignoreInt(const std::string & name);
  void ignoreDouble(const std::string & name);
  void ignoreString(const std::string & name);

  Attribute<bool> getBool(const std::string & name);
  Attribute<int> getInt(const std::string & name);
  Attribute<double> getDouble(const std::string & name);
  Attribute<std::string> getString(const std::string & name);
 
  
 protected:

  bool alreadyContains_(const std::string & name) const;
  void parseBool_(const char * str, Attribute<bool> & a);
  void parseInt_(const char * str, Attribute<int> & a);
  void parseDouble_(const char * str, Attribute<double> & a);
  void parseString_(const char * str, Attribute<std::string> & a);

 protected:


#define SDN_ATTRIB_MAP_NAME( __T__ ) __T__ ## Map_
#define SDN_DECLARE_ATTRIB_MAP( T ) std::map<std::string, Attribute< T > > SDN_ATTRIB_MAP_NAME( T );

 SDN_DECLARE_ATTRIB_MAP(bool);
 SDN_DECLARE_ATTRIB_MAP(int);
 SDN_DECLARE_ATTRIB_MAP(double);


 SDN_DECLARE_ATTRIB_MAP(attrib_str);


};

}
