/// \author Ian Goodfellow

#include "binOutArchive.h"
#include "binOutClassBuilder.h"
#include "pointer.h"
#include <sstream>
using namespace std;
using namespace boost;
using namespace sdn;
using namespace serial;

/*
  Class format:
  4 bytes: has been serialized

  if serialized before:
  8 bytes: vaddr of prev instance

  if not serialized before:
  8 bytes: vaddr
  4 bytes: version
  4 bytes: length of base type string
  <base type string>
  4 bytes: length of derived type string (0 if is not a derived type)
  if derived:
    <derived type name>
    4 bytes: class version of super class
    4 bytes: length of super class's derived type name (0 if it is not a derived type)
      etc.
    super class body
  class body



*/


BinOutClassBuilder::BinOutClassBuilder()
{
}

shared_ptr<BinOutClassBuilder> BinOutClassBuilder::makeDerived(BinOutArchive * archive, void * addr, int version, std::string id, bool knownToBeSerialized, BinOutClassBuilder * subclass)
{
  shared_ptr<BinOutClassBuilder> rval = ptr(new BinOutClassBuilder());
  rval->done_ = false;
  rval->archive_ = archive;//must precede calls to writeVaddr_
  rval->version_ = version;


  if (knownToBeSerialized)
    {
      rval->alreadySerialized_ = true;
    }
  else if (subclass)
    {
      rval->alreadySerialized_ = false; //if you have a subclass, and you are not knownToBeSerialized, it is impossible to find out you are serialized

    }
  else
    {
      if (archive->alreadySerialized_(addr))
	{
	  rval->alreadySerialized_ = true;
	  rval->writeBool_(rval->alreadySerialized_);
	  /// \todo "todo-- check that rval is allowed to have been serialized already (not a member of a contiguous memory region such as an array or class body)");
	  Vaddr v = archive->getVaddr_(addr);
	  rval->writeVaddr_(v);	  
	}
      else
	{
	  rval->alreadySerialized_ = false;
	  rval->writeBool_(rval->alreadySerialized_);
	  Vaddr v = archive->registerObject_(addr);
	  rval->writeVaddr_(v);
	  rval->writeInt_(version);
	}
    }

  rval->id_ = id;
  rval->subclass_ = subclass;
  rval->expectingSuper_ = true;
  rval->parsingSuper_ = false;
  rval->parsedSuper_ = false;
  rval->subclass_ = NULL;
  rval->isDerived_ = true;
  rval->gotSuper_ = false;

  return rval;
}

shared_ptr<BinOutClassBuilder> BinOutClassBuilder::makeBase(BinOutArchive * archive, void * addr, int version, std::string id, bool knownToBeSerialized, BinOutClassBuilder * subclass)
{
  shared_ptr<BinOutClassBuilder> rval = ptr(new BinOutClassBuilder());
  rval->done_ = false;
  rval->archive_ = archive;//must precede calls to writeVaddr_
  rval->isDerived_ = false;//must precede calls to setBaseType or this class won't know it's the base
  rval->version_ = version;
  rval->subclass_ = subclass;

  //This is probably also done by the caller catching the return value,
  //for explicit-ness, but this assignment here is necessary so that they
  //are already linked up before this one calls setBaseType
  if (subclass)
    subclass->curChild_ = rval;


  if (knownToBeSerialized)
    {
      rval->alreadySerialized_ = true;
    }
  else if (subclass)
    {
      rval->alreadySerialized_ = false; //if you have a subclass, and you are not knownToBeSerialized, it is impossible for you to be serialized (your address is the same as your subclass, which has already been checked)
      subclass->setBaseType(id);
    }
  else
    {
      if (archive->alreadySerialized_(addr))
	{
	  rval->alreadySerialized_ = true;
	  rval->writeBool_(rval->alreadySerialized_);
	  SDN_WARN("todo-- check that rval is allowed to have been serialized already (not a member of a contiguous memory region such as an array or class body)");
	  Vaddr v = archive->getVaddr_(addr);
	  rval->writeVaddr_(v);	  
	}
      else
	{
	  rval->alreadySerialized_ = false;
	  rval->writeBool_(rval->alreadySerialized_);
	  Vaddr v = archive->registerObject_(addr);
	  rval->writeVaddr_(v);
	  rval->writeInt_(version);

	  int len = id.size();
	  
	  SDN_ASSERT(sizeof(len) == 4);
	  archive->out_->write((char *) & len, sizeof(len));
	  
	  archive->out_->write((char *) id.c_str(), len);

	  len = 0;
	  rval->writeInt_(len);
	}
    }
  
  
  rval->id_ = id;
  rval->subclass_ = subclass;
  rval->parsingSuper_ = false;
  rval->expectingSuper_ = false;
  rval->parsedSuper_ = false;
  rval->gotSuper_ = false;

  return rval;
}

void BinOutClassBuilder::setBaseType(std::string id)
{
  SDN_ASSERT(!done_);

  if (alreadySerialized_)
    return;

  if (subclass_)
    {
    }
  else
    {
      int len = id.size(); //id, not id_. id is base type and it goes first
      SDN_ASSERT(sizeof(len) == 4);
      archive_->out_->write( (char *) & len, sizeof(len));
      archive_->out_->write( id.c_str(), len);
      
      len = id_.size();
      SDN_ASSERT(sizeof(len) == 4);      
      archive_->out_->write( (char *)  & len, sizeof(len));
      archive_->out_->write( id_.c_str(), len);
      //SDN_DEBUG(this << " setting base type");    
      curChild_->writeHeader();
      //SDN_DEBUG("done writing headers");
    }
}

void BinOutClassBuilder::writeHeader()
{
  //SDN_DEBUG("writeHeader running");
  SDN_ASSERT(subclass_ != NULL);
  if (isDerived_)
    {
      SDN_ASSERT(sizeof(version_) == 4);
      archive_->out_->write( (char *) & version_, sizeof(version_));
      int len = id_.size();
      SDN_ASSERT(sizeof(len) == 4);      
      archive_->out_->write( (char *)  & len, sizeof(len));
      archive_->out_->write( id_.c_str(), len);
    }
  else
    {
      //SDN_DEBUG("Writing out version "<<version_);
      writeInt_(version_);
      int len = 0;
      writeInt_(len);
    }
  if (curChild_ != NULL)
    curChild_->writeHeader();
}

void BinOutClassBuilder::beginDerived(void * addr, int version, std::string id)
{
  SDN_ASSERT(!done_);
  if (curChild_ != NULL)
    curChild_->beginDerived(addr, version, id);
  else
    {
      //SDN_DEBUG(this << " gained derived class child");

      BinOutClassBuilder * subclass = NULL;
      if (expectingSuper_)
	{
	  subclass = this;
	  expectingSuper_ = false;
	  SDN_ASSERT(! parsingSuper_);
	  parsingSuper_ = true;
	}

      curChild_ = makeDerived(archive_, addr, version, id, alreadySerialized_, subclass);     
    }
}

void BinOutClassBuilder::beginBase(void * addr, int version, std::string id)
{
  SDN_ASSERT(!done_);
  if (curChild_ != NULL)
    curChild_->beginBase(addr, version, id);
  else
    {
      //SDN_DEBUG(this << " gained base class child");

      BinOutClassBuilder * subclass = NULL;

      if (expectingSuper_)
	{
	  subclass = this;
	  expectingSuper_ = false;
	  SDN_ASSERT(! parsingSuper_);
	  parsingSuper_ = true;
	}

      curChild_ = makeBase(archive_, addr, version, id, alreadySerialized_, subclass);
      SDN_ASSERT(curChild_);
      
    } 
}

bool BinOutClassBuilder::endBase()
{
  //SDN_DEBUG("endBase");
  SDN_ASSERT(!done_);
  if (curChild_ != NULL)
    {
      if (curChild_->endBase())
	{
	  if (parsingSuper_)
	    {
	      parsingSuper_ = false;
	      SDN_ASSERT( ! parsedSuper_);
	      parsedSuper_ = true;
	    }
	  else
	    {
	      //SDN_DEBUG(this << " removed its child");
	      curChild_ = ptr( (BinOutClassBuilder * ) NULL);
	    }
	}
    }
  else
    {
      SDN_ASSERT(!isDerived_);
      done_ = true;
      //SDN_DEBUG("base class done");
    }
  return done_;
}

bool BinOutClassBuilder::endDerived()
{
  SDN_ASSERT(!archive_->out_->fail());
  SDN_ASSERT(!done_);
  if (curChild_ != NULL)
    {
      if (curChild_->endDerived())
	{
	  if (parsingSuper_)
	    {
	      parsingSuper_ = false;
	      SDN_ASSERT( ! parsedSuper_);
	      parsedSuper_ = true;
	    }
	  else
	    {
	      //SDN_DEBUG(this << " removed its child");
	      curChild_ = ptr( (BinOutClassBuilder * ) NULL);
	    }
	}
    }
  else
    {
      SDN_ASSERT(isDerived_);
      SDN_ASSERT(gotSuper_);
      done_ = true;
    }
  return done_;
}


void BinOutClassBuilder::endSuper()
{
  SDN_ASSERT(curChild_ != NULL)
  if (curChild_->done_)
    {
      SDN_ASSERT(isDerived_);
      //SDN_DEBUG(this << " removed its child");
      curChild_ = ptr( (BinOutClassBuilder *) NULL);
      parsedSuper_ = false;
      gotSuper_ = true;
    }
  else
    {
      curChild_->endSuper();
    }
}
  

void BinOutClassBuilder::writeVaddr_(Vaddr v)
{
  SDN_ASSERT(sizeof(v) == 8);
  archive_->out_->write( (char *) & v, sizeof(v));
}

void BinOutClassBuilder::writeBool_(bool b)
{
  unsigned char temp;

  if (b)
    temp = 1;
  else
    temp = 0;

  SDN_ASSERT(sizeof(temp) == 1);
  archive_->out_->write( (char *) & temp, sizeof(temp));
}

void BinOutClassBuilder::writeInt_(int i)
{
  SDN_ASSERT(sizeof(i) == 4);
  archive_->out_->write( (char *) & i, sizeof(i));
}

void BinOutClassBuilder::attribute(unsigned & value)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	curChild_->attribute(value);
      else
	writeInt_(value);
    }
}

void BinOutClassBuilder::attribute(double & value)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	curChild_->attribute(value);
      else
	archive_->write_(value);
    }
}

void BinOutClassBuilder::attribute(bool & value)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	curChild_->attribute(value);
      else
	archive_->write_(value);
    }
}


void BinOutClassBuilder::attribute(float & value)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	curChild_->attribute(value);
      else
	archive_->write_(value);
    }
}


void BinOutClassBuilder::attribute(std::string & value)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	curChild_->attribute(value);
      else
	archive_->write_(value);
    }
}

void BinOutClassBuilder::ScalarArray(Scalar * data, unsigned numFloats)
{
  if (!alreadySerialized_)
    {
      if (curChild_)
	{
	curChild_->ScalarArray(data, numFloats);
	}
      else
	{
	  if (SDN_DOUBLE_PRECISION)
	  {
		SDN_ASSERT(sizeof(Scalar) == 8);
	  }
	  else
	  {
	  	SDN_ASSERT(sizeof(Scalar) == 4);
	  }
	  SDN_ASSERT(!archive_->out_->fail());
	  //for (unsigned i = 0; i < numFloats; i++)
	  //cout << data[i] << endl;
	   writeInt_(numFloats);
	SDN_DEBUG("Wrote Scalar array beginning with "<<data[0]);
	  archive_->out_->write( (char *)  data, numFloats * sizeof(Scalar));
	  SDN_ASSERT_MSG(!archive_->out_->fail(),"output stream died after trying to write "<<numFloats<<" Scalars");
	}
    }
}
