//Author: Ian Goodfellow <ia3n@cs.stanford.edu>
/** \file command.h
     \brief Command line parsing
*/

#pragma once
#include "boost/shared_ptr.hpp"
#include <vector>
#include <string>

namespace sdn
{
	///  Command line parsing
	namespace cmd
	{

		class Parser;

		/// Represents results of parsing for a single command line argument
		class ParserResult
		{
			public:
				friend class Parser;

				/// after parse() has been called, this will be true if the associated argument was used.
				/** (this lets you disambiguate between a variable being set to its default value and 
				bool wasUsed;

			private:

				ParserResult() : wasUsed(false)

		};

		/// Declares an argument that the caller may optionally use. The return value will let you later find out if it was used.
		/** 
			Type T must be int, double, or std::string.
			\param target The variable whose value is to be update if the argument is invoked. Caller is responsible for ensuring it remains in scope until after parse() is called.
			\param longName The version of the argument to be used with "--". e.g., passing "images" enables the "--images" argument. Use "" to disable the long version of the argument.
			\param shortName The version of the argument to be used with "-", e.g. passing "i" enables the "-i" argument. Use "" or leave blank to disable the long version of the argument.
			\todo Make a multi-arg version
			\todo After implementing multi-arg version, use it to make a standardized way of broadcasting config information from the command line (-s/--set <topicName> <attribname> <value> )
			\todo Make a multi-invocation version
			\todo After implementing multi-invocation version, use it to make a standardized way of loading config files (-c config1.xml -c config2.xml -c config3.xml)
		
		*/
		template <typename T>
		boost::shared_ptr<ParserResult> declareArgument<T> ( T & target, const std::string & longName, std::string shortName = "" );

		/// Declares an argument that the caller must specify
		template <typename T>
		void requireArgument<T> (T & target, const std::string & longName, std::string shortName = "");

		/// Declares a switch (argument that sets a bool by its presence or absence)
		void declareSwitch(blah)


		/// Parses the command line
		void parse(int argc, char ** argv);



		// pay no attention to the man behind the curtain
		// clients only need to invoke this class via the functions defined above
		class Parser
		{
			private:
				class ParserEntry
				{
					public:
						

				};

			public:
				//singleton class accessor method
				static Parser & get();

				template <typename T>
				boost::shared_ptr<ParserResult> declareArgument<T> ( T & target, const std::string & longName, const std::string & shortName)
				{
					boost::shared_ptr<ParserResult> rval(new ParserResult());
					SDN_ASSERT(rval);

					addArgument_(target, longname, shortname, rval);
					
					return rval;
				}

				template <typename T>
				void requireArgument<T> ( T & target, const std::string & longName, const std::string & shortName)
				{
					boost::shared_ptr<ParserResult> nullPtr;
					SDN_ASSERT(rval);

					addArgument_(target, longname, shortname, rval);
					
					return rval;
				}
				



				void parse(int argc, char ** argv);

			private:
				//singleton class, hide constructor
				Parser();	


				//if result is NULL, then this is a required argument
				void addArgument_<T> (T & target, const std::string & longName, const std::string & shortname, boost::shared_ptr<ParserResult> result)
				{
					checkNames_(longName, shortName);
				
					/// \todo finish implementing	
				}

				//Verifies that names are valid/not in use
				void checkNames_(const std::string & longName, const std::string & shortName);


			private:
				//if true, parse has been called already. parse may only be called once
				bool parseCalled_;
		};

		template <typename T>
		boost::shared_ptr<ParserResult> sdn::cmd::declareArgument<T> ( T & target, const std::string & longName, std::string shortName)
		{
			return Parser::get().declareArgument(target,longName,shortName);
		}

		template <typename T>
		void sdn::cmd::requireArgument<T> ( T & target, const std::string & longName, std::string shortName)
		{
			Parser::get().requireArgument(target,longName,shortName);
		}


	}
}


