//Author: Ian Goodfellow <ia3n@cs.stanford.edu>
//File: ConfigManager.h
//Based on svlConfigManager.h by Stephen Gould, from the STAIR Vision Library
//Ian's modifications are to allow each listener to subscribe to multiple topics,
//each topic to have multiple subscribers + to use sdn style 
/*** DESCRIPTION:
**   Library-wide onfiguration code. Handles configuration of static members in the
**   SDN libraries and applications from XML config files. Also allows projects 
**   to register their own configuration code. 
**   
**   All identifiers are case-sensitive. Implementers should inherit from
**   ConfigParser and override the usage() and readConfig() member 
**   functions, then create a ConfigListener initialized with an
**   instance of their ConfigParser class.
**
*****************************************************************************/

#pragma once

#include <map>
#include <boost/smart_ptr.hpp>
#include <vector>

class XMLNode;

// note: ConfigManager uses raw pointers rather than smart pointers because
// shared_ptr would cause trouble with both statically and
// dynamically allocated objects.
// When static objects get destructed, they'd unsubscribe themselves,
// and when
// the ConfigManager removed the shared_ptr to them, they'd get 
// destructed again). Since objects are only removed from the 
// Dynamically allocated objects would end up living forever since they
// are only unsubscribed when their destructor is called.

namespace sdn
{
  // configuration interface ----------------------------------------------------

  /// Interprets messages received by a ConfigListener
  /** To set up your own code to receive configuration from the
      ConfigManager, subclass ConfigParser and create a ConfigListener
      that is initialized with an instance of your parser.
  */
  class ConfigParser
  {
	public:
		ConfigParser();
		virtual ~ConfigParser();

///Tells its parent ConfigListener which topics to listen to
		virtual void topics(std::vector<std::string> & topics) = 0;
///Displays usage instructions. ///TODO-- standardize the usage format / decide how we display it to the user
		virtual void usage(std::ostream & os) const = 0;

///Parse and take action on a message recieved via the listener
		virtual void readConfig(XMLNode & node)=0;
  };

  /// Receives messages from the ConfigManager
/** As a client, all you need to do is declare an instance of this class
   that is initialized with your parser.
   This class is mainly needed because virtual methods can't be called
   safely from a constructor of any class that is not the most derived.
   Having ConfigListener wrap ConfigParser allows ConfigListener to 
  call virtual methods of ConfigParser upon construction.
  These calls are indirect in nature-- ConfigListener calls ConfigManager
  which then calls ConfigListener/Parser.
 */

  class ConfigListener
  {
  private:
    std::vector<std::string> _topics;
    boost::shared_ptr<ConfigParser> _parser;    

  public:
/// Pass your parser in here
    ConfigListener(boost::shared_ptr<ConfigParser> p);
    virtual ~ConfigListener();
    
    inline const std::vector<std::string>& topics() const { return _topics; }
    

    void readConfig(const char *filename);
    void readConfig(XMLNode& node);
  };

  class ConfigTopic
  {
  public:
    std::vector<sdn::ConfigListener *> subscribers;
//messages that were sent earlier. these will be resent to new subscribers 
   std::vector<XMLNode> backIssues;
  };
  
  // library configuration registry ---------------------------------------------
  
/// Broadcasts configuration information to registered ConfigListener objects
/** 
At construction, ConfigManager checks ${CONFIG_PATH}/main.xml for
configuration information. This should contain a single <sdnConfig>
</sdnCongfig> node. Each of this node's children will be sent to
registered listeners (the name of the node determines which topic
it will be sent to, i.e. <colorTopic theColor="red"> will be sent
to listeners subscribed to the topic "red").
Prior to being broadcast, the XML will be preprocessed somewhat.
${SDN_PATH} and ${CONFIG_PATH} will be resolved to the appropriate
value. 
It is also possible to make hierarchical xml files by using
<include file="filepath"/> tags. (Note: The include tag is implemented
using a ConfigListener that calls the ConfigManager again, not
with hard-coded preprocessing)

In addition to loading ${CONFIG_PATH}/main.xml on launch, the 
ConfigManager can also be called explicitly, for example to
load experiment-specific config files passed in on the command line.
*/
  class ConfigManager {
   
  private:

    friend class ConfigListener;

    typedef std::map<std::string, sdn::ConfigTopic * > ConfigRegistry;
    ConfigRegistry _registry;
    
  public:
    ~ConfigManager();
    
///Provides a reference to the ConfigManager
    static ConfigManager& get();
    
///Broadcast configuration information from a file
    void configure(const char *filename, bool noErrors=false);//if noErrors=true, will fail silently when file doesn't exist

///Broadcast configuration information from an XML node
    void configure(XMLNode& root);
    
///Prints out all registered topics
    void showRegistry() const;
   
//returns true if destructor has not been called yet. this prevents
//listeners from trying to unregister themselves after the ConfigManager
//has already died
    static bool alive; 
  protected:
    ConfigManager(); // singleton class so hide constructor
    
    void _subscribe(ConfigListener * m);//this is called whenever and only when a ConfigListener constructor runs
    void _unsubscribe(ConfigListener * m);//this is called whenever and only when a ConfigListener destructor runs

    ConfigRegistry::iterator _createTopic(const std::string & name);
  };
  


///todo-- a known issues with the design: saving backissues could waste memory.



/// ConfigParser that implements <include> for the ConfigManager

  class IncludeParser : public ConfigParser
  {
  public:
    IncludeParser();
    virtual ~IncludeParser();

    void topics(std::vector<std::string> & topics);
    void usage(std::ostream & os) const;
    
    void readConfig(XMLNode & node);
  };

}
