/// \author Ian Goodfellow
/// \brief An archive that deserializes objects

#pragma once

#include "archive.h"
#include <map>
//#include <string>
#include <vector>
#include <boost/shared_ptr.hpp>

namespace sdn
{
  namespace serial
  {

    class PointerDeleter
    {
    };

    template <typename T>
    class DerivedDeleter : public PointerDeleter
    {
	public:
		DerivedDeleter(boost::shared_ptr<T> * p) { p_ = p; }
		virtual ~DerivedDeleter()
		{
			delete p_;
		}
	protected:
		boost::shared_ptr<T> * p_;
    };


    //The strings are the names of the base types of the objects, so that type safety
    //can be somewhat enforced
    //The void *s can be casted to shared pointers of the base type.
    //stored as void *s
    //because there's not a good way of making them polymorphic
    typedef std::map<Vaddr, std::pair<std::string, void *> > VaddrToAddr;

    class InArchive : public Archive
    {
    public:
      InArchive();
      
      //should only be called by Registry or the implementations of functions
      //in pointer.h. Lets the registry see what's
      //coming in advance so that it can allocate the object that will
      //be able to serialize the next contents of the archive
      //If the next object has been serialized already, the address of
      //a shared pointer to it is written into prevPtr. Otherwise, prevPtr
      //is set to NULL and the derived class id and version of the class
      //are written into the arguments
      void peek(void * & prevPtr,
			std::string & id, int & version);
      

      //should only be called by the registry. after the registry allocates
      //an object from the archive, it should put a shared pointer on the
      //heap (yes, the shared pointer itself should be on the heap) and
      //give a void * to that shared pointer back to the archive so the
      //archive can keep on handing out shared pointers to that object
      void reportPointer(void * p, PointerDeleter * d);


      bool in()
      {
	return true;
      }

	virtual ~InArchive();

    protected:
      virtual Vaddr peek_(void * & prevPtr,
	  std::string & id, int & version) = 0;

      std::string vaddrToBaseName_(Vaddr v);
      //returns the address of a shared pointer to the object with virtual address v
      void * vaddrToAddrOfPtr_(Vaddr v);
      

    protected:
      VaddrToAddr vaddrToAddr_;
      bool expectingReportPointer_;
      Vaddr expectedVaddr_;
      std::string peekedDerivedName_;
      int peekedVersion_;
      bool hasPeeked_; //subclasses are responsible for turning this off
      std::vector<PointerDeleter *> deleters_;
    };
  }
}
