//Based on svlLogger.h by Stephen Gould and Ian Goodfellow from the STAIR Vision Library
//For SDN, this has been modified to use shorter names for the macros
/** \file logger.h
  \brief Message and error logging. This class is not thread-safe in the interest
  of not having to flush the log on every message.
*/
#pragma once

//#include <iostream>
#include <fstream>
#include <sstream>


#define SDN_LOG_MESSAGE(L, M) \
  if ((L) > sdn::Logger::getLogLevel()) { } \
  else { std::stringstream __s; if (L == sdn::LOGLEVEL_FATAL) { __s << "(" << __FILE__ << ", " << __LINE__ << ") "; } \
    __s << M; sdn::Logger::log(L, __s.str()); }

#define SDN_ASSERT(C) \
  if (!(C)) { SDN_FATAL( #C); }

#define SDN_ASSERT_MSG(C, M) \
  if (!(C)) { SDN_FATAL( #C << ": " << M); }

/// Display message M and kill the program
#define SDN_FATAL( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_FATAL, M )
/// Display message M as an error
#define SDN_ERROR( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_ERROR, M )
/// Display message M as a warning
#define SDN_WARN( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_WARNING, M )
/// Display message M with no decoration
#define SDN_MSG( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_MSG, M )
/// Display message M only if in verbose mode
#define SDN_VERBOSE( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_VERBOSE, M )
/// Display message M only if in debug mode
#define SDN_DEBUG( M ) SDN_LOG_MESSAGE( sdn::LOGLEVEL_DEBUG, M )

namespace sdn
{

  enum LogLevel { //note that there is a sorted order to these, see implementation of SDN_LOG
    LOGLEVEL_FATAL = 0, //fatal error message
    LOGLEVEL_ERROR,     //non-fatal error message
    LOGLEVEL_WARNING,   //warning message
    LOGLEVEL_MSG,    //standard message for displaying output
    LOGLEVEL_VERBOSE,   //message to be shown only in verbose mode
    LOGLEVEL_DEBUG      //message to be shown only in debugging mode
  };

  /// Class that implements the output macros
  class Logger
  {

/// \todo make it possible to compile out SDN_DEBUG, SDN_VERBOSE statements (and then print a warning if these log levels are requested at runtime)
/// Logs output.
/** In general, the output file and log level will be set by the command line parser, so the only
reason a client needs to worry about this class is if they want to change the log level away from
what it was set to on the command line. Pretty much all access to this class class should be through
the SDN_FATAL, SDN_ERROR, SDN_WARN, SDN_MSG, SDN_VERBOSE, and SDN_DEBUG logging macros.
*/  


  public:
    static void (*showErrorCallback)(const char *message);
    static void (*showWarningCallback)(const char *message);
    static void (*showMessageCallback)(const char *message);
    
 private:
    static std::ofstream _output;
    static LogLevel _logLevel;
    
  public:
    Logger();
    ~Logger();
    
    static void initialize(const char *filename,
        bool bOverwrite = false);
    static void initialize(const char *filename,
			   bool bOverwrite, LogLevel level);
    static inline void setLogLevel(LogLevel level) {
      _logLevel = level;
    }
    static inline LogLevel getLogLevel() {
      return _logLevel;
    }
    
    static void log(LogLevel level, const std::string& msg);
  };



  
}
