//Author: Ian Goodfellow <ia3n@cs.stanford.edu>
//File: matrixUtils.cc

#include <iostream>
using namespace std;
#include "matrixUtils.h"
#include "logger.h"
using namespace sdn;
//USING_PART_OF_NAMESPACE_EIGEN // import most common Eigen types 
#include <Eigen/Array>
#include <boost/make_shared.hpp>
#include "archive.h"
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"
#include "matrix.h"

using namespace boost;

Matrix sdn::tanhM(const Matrix& m) {
#if SDN_USE_CUDA
  return m.cwise().tanh();
#else
  Matrix res(m.rows(), m.cols());
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      res(i,j) = tanh(m(i,j));
  return res;
#endif
}

Matrix sdn::coshM(const Matrix& m) {
#if SDN_USE_CUDA
  return m.cwise().cosh();
#else
  Matrix res(m.rows(), m.cols());
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      res(i,j) = cosh(m(i,j));
  return res;
#endif
}

void sdn::loadMatrixFromImage(const char * filepath, Matrix & m, bool resizeOkay)
{
  IplImage * im = cvLoadImage(filepath, CV_LOAD_IMAGE_GRAYSCALE);

  SDN_ASSERT_MSG(im, "Could not load image \""<<filepath<<"\"");
  
  if (m.rows() != im->height || m.cols() != im->width)
    {
      if (resizeOkay)
	{
	  m = Matrix::Zero(im->height, im->width);
	}
      else
	{
	  SDN_FATAL("Error while loading \""<<filepath<<"\" as matrix. Expected width "<<m.cols()<<" and height "<<m.rows()<<" but got "<<im->width<<"x"<<im->height);
	}
    }

  for (int i = 0; i < m.rows(); i++)
    for (int j = 0; j < m.cols(); j++)
      m(i,j) = cvGetReal2D(im,i,j);


  cvReleaseImage(&im);
}


static IplImage * matrixToIplImage(const Matrix & m)
{
IplImage * im = cvCreateImage(cvSize(m.cols(),m.rows()), IPL_DEPTH_8U, 3);

	for (int i = 0; i < m.rows(); i++)
		for (int j = 0; j < m.cols(); j++)
		{
			Scalar val = m(i,j) * 255;
			cvSet2D(im,i,j,cvScalar(val,val,val));
		}

	return im;
}

void sdn::saveMatrixAsImage(const char * filepath, const Matrix & m)
{
  IplImage * im = matrixToIplImage(m);

  cvSaveImage(filepath, im);
  cvReleaseImage(&im);	
  

}


void sdn::showMatrix(const Matrix & m, const char * windowName)
{
  IplImage * im = matrixToIplImage(m);

  cvNamedWindow(windowName, CV_WINDOW_AUTOSIZE);

  cvShowImage(windowName, im);

  cvWaitKey(0);

  cvReleaseImage(&im);
}


void sdn::saveMatrix(std::ofstream &outputFile, const Matrix &m)
{
  assert(outputFile.is_open());
  
  int header[3];
  header[0] = 0x5852544D;    // Magic number.  Writes "MTRX" into file.
  header[1] = m.rows();
  header[2] = m.cols();
  
  outputFile.write((char*)header, sizeof(int)*3);
  if (!outputFile.good())
  {
    SDN_FATAL("saveMatrix: SDN_FATAL saving to file.");
  }

  Scalar *data = new Scalar[m.rows()*m.cols()];
  assert(data != NULL);
  
  for (int i=0; i < m.rows(); i++)
    for (int j=0; j < m.cols(); j++)
      data[i*m.cols()+j] = m(i,j);
  
  outputFile.write((char*)data, sizeof(Scalar)*m.rows()*m.cols());
  if (!outputFile.good())
  {
  SDN_FATAL("saveMatrix: SDN_FATAL saving to file.");
  }  

  delete[] data;
  
  return;
}

void sdn::loadMatrix(std::ifstream &inputFile, Matrix &m, bool resizeOkay)
{
  assert(inputFile.is_open());
  assert(!inputFile.eof());
  
  int header[3];
  
  inputFile.read((char*)header, sizeof(int)*3);
  if (!inputFile.good())
  {
   SDN_FATAL("loadMatrix: SDN_FATAL loading from file.");
  }
 
  if (header[0] != 0x5852544D)
  {
   SDN_FATAL("loadMatrix: Didn't get expected magic number.");
  }
  //std::cout << header[1] << "x" << header[2] << " matrix" << std::endl;
  
  if (resizeOkay)
  {
   m = Matrix(header[1], header[2]);
  }
  else if (m.rows() != header[1] || m.cols() != header[2])
  {
   SDN_FATAL("loadMatrix: Input matrix m inconsistent with size in file.");
  } 
 
  Scalar *data = new Scalar[m.rows()*m.cols()];
  assert(data != NULL);
  
  inputFile.read((char*)data, sizeof(Scalar)*m.rows()*m.cols());
  
  if (inputFile.bad())
  {
   SDN_FATAL("loadMatrix: SDN_FATAL loading from file.");
  } 
 
  for (int i=0; i < m.rows(); i++)
    for (int j=0; j < m.cols(); j++)
      m(i,j) = data[i*m.cols()+j];
  
  return;
}

Vector sdn::matrixToVector(const Matrix & m)
{
  //  cout << "Step within row: " << (long( &(m(0,1))) - (m.data())) << endl;

  //Vector::block(i,0,n,1) returns elements i:i+n-1 of the vector

  Vector rval(m.rows()*m.cols());

  /* 
  int c = m.cols();
  int idx = 0;
  for (int i = 0; i < m.rows(); i++)
    {
      rval.block(idx,0,c,1).transpose() = m.row(i);
      idx += c;
    }
  */

  //copying it by columns seems to be faster, guess default is column major
 
  int r = m.rows();
  int idx = 0;
  for (int i = 0; i < m.cols(); i++)
    {
      rval.block(idx,0,r,1) = m.col(i);
      idx += r;
    }

  //note-- even when we get the matrix format wrong, block copying is faster
  //than using two for loops
 

  //todo-- figure out how vector blocks work for sure

  return rval;
}

unsigned sdn::vectorCoord(unsigned i, unsigned j, unsigned r, unsigned c)
{
  SDN_ASSERT_MSG(i < r, "row "<<i<<" is not valid, matrix only has "<<r << " rows");
  SDN_ASSERT_MSG(j < c, "col "<<j<<" is not valid, matrix only has "<<c << " cols");
  return j * r + i;
}


void sdn::matrixToVector(const Matrix & m, Vector & v)
{
  //  cout << "Step within row: " << (long( &(m(0,1))) - (m.data())) << endl;

  //Vector::block(i,0,n,1) returns elements i:i+n-1 of the vector

  /* 
  int c = m.cols();
  int idx = 0;
  for (int i = 0; i < m.rows(); i++)
    {
      rval.block(idx,0,c,1).transpose() = m.row(i);
      idx += c;
    }
  */
 
  int r = m.rows();
  int c = m.cols();

  assert(v.size() == r*c);

  int idx = 0;
  for (int i = 0; i < c; i++)
    {
      v.block(idx,0,r,1) = m.col(i);
      idx += r;
    }
 
}

void sdn::matrixToMatrixRow(const Matrix & m, Matrix & output, int row )
{
  int r = m.rows();
  int c = m.cols();

  assert(output.cols() == r*c);

  int idx = 0;

  for (int i = 0; i < c; i++)
    {
      output.block(row,idx,1,r).transpose() = m.col(i);
      idx += r;
    }
}

void sdn::matrixToMatrixCol(const Matrix & m, Matrix & output, int col)
{
  int r = m.rows();
  int c = m.cols();
  /*cout << "in matrixtomatrixcol " << endl;
    cout << output.rows() << endl;
    cout << r << " " << c << endl;*/
  SDN_ASSERT_MSG(output.rows() == r*c, "output.cols() = "<<output.cols()<<"; r="<<r<<", c="<<c<<",so r*c="<<r*c);

  int idx = 0;

  for (int i = 0; i < c; i++)
    {
      output.block(idx,col,r,1) = m.col(i);
      idx += r;
    }
}

void sdn::matrixToMatrixSubCol(const Matrix & a, Matrix & b, int r, int c)
{
  int a_r = a.rows();
  int a_c = a.cols();

  for (int i = 0; i < a_c; i++)
    {
      b.block(r,c,a_r,1) = a.col(i);
      r += a_r;
    }
}

void sdn::matrixSubColToMatrix(const Matrix & b, int r, int c, Matrix & a)
{
  int a_r = a.rows();
  int a_c = a.cols();

  for (int i = 0; i < a_c; i++)
    {
      a.col(i) = b.block(r,c,a_r,1);
      r += a_r;
    }
}

void sdn::matrixSubColToMatrixBlock(const Matrix & b, int r, int c, Matrix & a, int startRow, int startCol, int rows, int cols)
{
  //assert(r+rows*cols <= b.rows());
  //assert(startRow+rows<= a.rows());
  //assert(startCol+cols <= a.cols());

  for (int i = 0; i < cols; i++)
    {
      a.block(startRow, startCol+i, rows, 1) = b.block(r,c,rows,1);
      r += rows;
    }
}

void sdn::vectorToMatrixBlock(const Vector & b, Matrix & a, int startRow, int startCol, int rows, int cols)
{
	int r = 0;
	for (int i = 0; i < cols; i++)
	{
		a.block(startRow, startCol+i, rows, 1) = b.block(r, 0 , rows, 1);
		r += rows;
	}
}

void sdn::vectorToMatrix(const Vector & v, Matrix & m)
{
  int r = m.rows();
  int c = m.cols();
  assert(v.size() == r * c);

  int idx = 0;  

  for (int i = 0; i < c; i++)
    {
      m.col(i) = v.block(idx,0,r,1);
      idx += r;
    }
}



Matrix sdn::vectorToMatrix(const Vector & v, int r, int c)
{
  Matrix m(r,c);

  int idx = 0;  

  for (int i = 0; i < c; i++)
    {
      m.col(i) = v.block(idx,0,r,1);
      idx += r;
    }

  return m;
}

Matrix sdn::vectorToMatrixHack(const Vector & v, int r, int c)
{
  Matrix m(r,c);

  assert(v.size() == r*c);

  memcpy(m.data(),v.data(),r*c*sizeof(Scalar));

  return m;
}


void sdn::vectorToMatrixHack(const Vector & v, Matrix & m)
{
  int r = m.rows();
  int c = m.cols();
  assert(v.size() == r * c);



  memcpy(m.data(), v.data(),r*c*sizeof(Scalar));
}


void sdn::matrixBlockToCArr(const Matrix & m, int startRow, int startCol, int rowSpan, int colSpan, Scalar * output)
{
  assert(isColumnMajor(m));

  int idx = 0;
  for (int c = 0; c < colSpan; c++)
    {
      memcpy(output+idx, m.data()+(startCol+c)*m.rows()*sizeof(Scalar) + startRow*sizeof(Scalar), rowSpan*sizeof(Scalar));
      idx += rowSpan * sizeof(Scalar);
    }
}

void sdn::matrixMean(const Matrix& m, 
		Matrix& output, 
		bool rowwise) {
	if (rowwise) {
		output = m.rowwise().sum()/m.cols();
	} else {
		output = m.colwise().sum()/m.rows();
	}
}


void sdn::matrixCovariance(const Matrix& m, Matrix& cov) {
  Matrix mean;
  SDN_MSG("A");
  matrixMean(m, mean);
  SDN_MSG("B");
  SDN_MSG(mean.rows()<<" "<<mean.cols());

  cov = m.transpose()*m/m.rows() - mean.transpose()*mean;
  SDN_MSG("C");
  if(m.rows() > 1) {
    cov *= (Scalar)m.rows()/(Scalar)(m.rows()-1);
  }
  SDN_MSG("D");
}

void sdn::serializeMatrixSize(serial::Archive & ar, Matrix & m)
{
  SDN_DEBUG("serializeMatrixSize");

  int version = 64;

  version = ar.beginBase( & m, version, "sdn::SizedMatrix");

    switch (version)
      {
	case 64:
	
	  {
	    unsigned r = (unsigned) m.rows();
	    unsigned c = (unsigned) m.cols();
	    
	    ar.attribute("rows",r);
	    ar.attribute("cols",c);

	    SDN_DEBUG("rows: "<<r<<" cols: "<<c);
	    
	    if (ar.in())
	      {
		m.resize(r,c);
	      }
	    
	    ar.endBase();
	  }

	break;

	default:
	SDN_FATAL("Invalid class version requested");
      }

}

void sdn::serializeMatrix(serial::Archive & ar, Matrix & m)
{
  SDN_DEBUG("running serializeMatrix");

  int version = 65;

  version = ar.beginBase( & m, version, "sdn::Matrix");

  switch (version)
    {
    case 65:
      {
	
	unsigned r = (unsigned) m.rows();
	unsigned c = (unsigned) m.cols();
	
	bool format = isColumnMajor(m);
	
	ar.attribute("rows",r);
	ar.attribute("cols",c);
	ar.attribute("columnMajor",format);
	
	  if (ar.in())
	    {
	      SDN_ASSERT_MSG(isColumnMajor(m) == format, "You are trying to deserialize an Eigen matrix whose compile-time row major vs column major flag differs from the matrix that was written into the archive. This is weird and I didn't expect it to happen, but you can fix it by altering serializeMatrix to re-arrange the data after it gets read back in");	    
	      m.resize(r,c);
	    }
	  
	  unsigned size = r*c;
#if SDN_USE_CUDA
	  size = m.dataSize();
#endif

	  ar.scalarArray(m.data(),size);
	  
	  ar.endBase();
      }
      break;
      
    default:
      SDN_FATAL("Invalid matrix version requested");
    }
  
}

double sdn::matrixMean(const Matrix & m)
{
  return m.sum() / Scalar(m.rows() * m.cols());
}

Matrix sdn::pyramidDown(const Matrix & m)
{
  SDN_ASSERT(m.rows() % 2 == 0);
  SDN_ASSERT(m.cols() % 2 == 0);
  Matrix rval = Matrix::Zero(m.rows()/2, m.cols()/2);

  for (int i = 0; i < rval.rows(); i++)
    {
      for (int j = 0; j < rval.cols(); j++)
	{
	  rval(i,j) = matrixMean(m.block(2*i,2*j,2,2));
	}
    }

  return rval;
}

double sdn::matrixVariance(const Matrix & m)
 {
   if (m.rows() * m.cols() == 1)
     {
       return 0;
     }
 
   return (m.cwise()-matrixMean(m)).cwise().square().sum()/double(m.rows()*m.cols()-1);
 }

unsigned sdn::maxIndex(Vector & v)
{
  Scalar maxVal = v(0);
  unsigned maxInd = 0;

  for (int i = 0; i < v.size(); i++)
    {
      if (v[i] > maxVal)
	{
	  maxVal = v[i];
	  maxInd = i;
	}
    }

  return maxInd;
}


void sdn::scrambleRowwise(Matrix & m, boost::shared_ptr<RandomIntGenerator<int> > g)
{
  if (!g)
    {
      g = make_shared<RandomIntGenerator<int> >();
    }

  for (int i = 0; i < m.rows(); i++)
    for (int j = 0; j < m.cols(); j++)
      {
	int swap = g->randUnif(0,m.cols()-1);
	Scalar temp = m(i,swap);
	m(i,swap) = m(i,j);
	m(i,j) = temp;
      }
}
