//Author: Ian Goodfellow <ia3n@cs.stanford.edu>
//(some functions by Andrew Ng)
//File: matrixUtils.h
//Description: Convenience functions for using Eigen matrices

#pragma once

#include "matrix.h"
#include "random.h"

namespace sdn
{

  //Forward declaration to reduce compile time-- this class is defined in archive.h
  namespace serial
  {
    class Archive;
  }


  Matrix tanhM(const Matrix& m);
  Matrix coshM(const Matrix& m);

///Saves a matrix to file
//todo-- switch to use ostream rather than ofstream
void saveMatrix(std::ofstream & outputFile, const Matrix & m);

/// Loads a matrix from file
/** If resizeOkay is true, then will resize m to fit the size of the matrix we're loading. Otherwise, m must already be the correct size.  */
void loadMatrix(std::ifstream &inputFile, Matrix &m, bool resizeOkay=0);


/// loads a matrix from a grayscale image
/** \param resizeOkay If true, then will resize m to fit the size of the matrix
    we're loading. Otherwise, m must already be the correct size. */
 void loadMatrixFromImage(const char * filepath, Matrix & m, bool resizeOkay = false);

/// saves a matrix as a grayscale image
void saveMatrixAsImage(const char * filepath, const Matrix & m);

///Reformats a vector as a matrix of the desired size. Use this version when initializing a matrix's constructor
Matrix vectorToMatrix(const Vector & v, int r, int c);

///Reformats a vector as matrix; matrix argument should have the correct size
void vectorToMatrix(const Vector & v, Matrix & m);

///A fast version of vectorToMatrix; I'm uncertain if it is safe
void vectorToMatrixHack(const Vector & v, Matrix & m);
///A fast version of vectorToMatrix; I'm uncertain if it is safe. This is the fastest way to get the matrix if constructing it from scratch
Matrix vectorToMatrixHack(const Vector & v, int r, int c);

///Reformats a vector as a matrix. Fastest way to construct a vector from a matrix
Vector matrixToVector(const Matrix & m);
///Reformats a vector as a matrix
void matrixToVector(const Matrix & m, Vector & v);
///Suppose $M \in \mathcal{R}^{r x c}$, and you have called matrixToVector(M,v). Then vectorCoord(i,j,r,c) gives the coordinate of $M_{i,j}$ in $v$  (c is only used for bounds-checking)
 unsigned vectorCoord(unsigned i, unsigned j, unsigned r, unsigned c);


//benchmarking shows that for small matrices, the memcpy hack could be worthwhile
//but for large matrices it can actually be slower. I guess we won't worry about
//trying to use it.

//also, benchmarking shows that passing a vector/matrix as a return value is
//a bad idea, unless it is passed to a constructor. in that case, there is
//a small benefit.

/// reformats a matrix as a row of a different matrix
void matrixToMatrixRow(const Matrix & m, Matrix & output, int row = 0);

/// reformats a matrix as a column of a different matrix
 void matrixToMatrixCol(const Matrix & m, Matrix & output, int col = 0);

///writes a column-major vectorization of a into column c of b, starting at row r
/**(no need for this to reach all the way to the end of the column in b) */
void matrixToMatrixSubCol(const Matrix & a, Matrix & b, int r, int c);

///inverse of matrixToMatrixSubCol
void matrixSubColToMatrix(const Matrix & b, int r, int c, Matrix & a);

///like matrixSubColToMatrix, but uses a.block(startRow,startCol,rows,cols) as the output (for some reason you can't just pass a block as a non-const matrix reference)
void matrixSubColToMatrixBlock(const Matrix & b, int r, int c, Matrix & a, int startRow, int startCol, int rows, int cols);
/// like matrixSubColToMatrixBlock, but input is a vector
void vectorToMatrixBlock(const Vector & b, Matrix & a, int startRow, int startCol, int rows, int cols);

///Like matrix to vector, but you can specify a sub-matrix, and output is a C array. Checks that the input matrix is column major
void matrixBlockToCArr(const Matrix & m, int startRow, int startCol, int rowSpan, int colSpan, Scalar * output);


//Shrinks an image by a factor of 2 by averaging each 2x2 block into one pixel
 Matrix pyramidDown(const Matrix & m);

///Checks if a matrix is column major
template<int Options>
bool isColumnMajor(const Eigen::Matrix<Scalar, Eigen::Dynamic, Eigen::Dynamic, Options>& m)
{
  return ! (Options&Eigen::RowMajor); // Eigen::ColMajor is 0, i.e., it turns off the RowMajor flag, so we can't check by &ing with it
}

///Returns the mean of each row or column of the matrix (defaults to col)
/**
	Identical to mean() in matlab.
*/
void matrixMean(const Matrix& m, Matrix& output, bool rowwise = false);

/// Returns the mean, with each coefficient in the matrix as a sample of the same random var
double matrixMean(const Matrix & m);

///Returns the variance, with each coefficient in the matrix as a sample of the same random var
 double matrixVariance(const Matrix & m);

///Returns the empirical covariance of the matrix m.
/**
	Each row should be an observation, and each column a variable.
	Identical to the covariance call in matlab. Normalizes by N-1 for N > 1
	where N is the number of observations.
*/
void matrixCovariance(const Matrix& m, Matrix& cov);


/// Randomly rearranges the elements within each row of m
 void scrambleRowwise(Matrix & m, boost::shared_ptr<RandomIntGenerator<int> > g = boost::shared_ptr<RandomIntGenerator<int> >());


///Serializes a matrix in such a way that only its size is stored in/restored from the archive
 void serializeMatrixSize(serial::Archive & ar, Matrix & m);
 
 ///Serializes the matrix, both in terms of size and in terms of content
 void serializeMatrix(serial::Archive & ar, Matrix & m);


 void showMatrix(const Matrix & m, const char * windowName);

 unsigned maxIndex(Vector & v);

class MatrixSerializer
{
	public:
		void operator()(serial::Archive & ar, Matrix & m)
		{
			serializeMatrix(ar,m);
		}

		std::string getClassName()
		{
		  return "sdn::Matrix";
		}
};


}

