/// \author Ian Goodfellow


#include "parser.h"
#include "logger.h"
#include <string>
using namespace std;
using namespace sdn;
using namespace serial;

string toString(Symbol t)
{
#define SDN_SERIAL_SYMBOL_STR(x) case x: return #x;

  switch (t)
    {
      SDN_SERIAL_SYMBOL_STR(FLUSH);
      SDN_SERIAL_SYMBOL_STR(BEGIN_BASE);
      SDN_SERIAL_SYMBOL_STR(END_BASE);
      SDN_SERIAL_SYMBOL_STR(BEGIN_DERIVED);
      SDN_SERIAL_SYMBOL_STR(END_DERIVED);
      SDN_SERIAL_SYMBOL_STR(BEGIN_SUPER);
      SDN_SERIAL_SYMBOL_STR(END_SUPER);
      SDN_SERIAL_SYMBOL_STR(ATTRIBUTE);
      SDN_SERIAL_SYMBOL_STR(BEGIN_PTR);
      SDN_SERIAL_SYMBOL_STR(END_PTR);
      SDN_SERIAL_SYMBOL_STR(CUSTOM_CONTENT);
      SDN_SERIAL_SYMBOL_STR(S);
      SDN_SERIAL_SYMBOL_STR(CLASS_INSTANCES);
      SDN_SERIAL_SYMBOL_STR(FLUSHES);
      SDN_SERIAL_SYMBOL_STR(CLASS_INSTANCE);
      SDN_SERIAL_SYMBOL_STR(BASE_CLASS_INSTANCE);
      SDN_SERIAL_SYMBOL_STR(DERIVED_CLASS_INSTANCE);
      SDN_SERIAL_SYMBOL_STR(SUPER_CLASS_INSTANCE);
      SDN_SERIAL_SYMBOL_STR(INSTANCE_BODY);
      SDN_SERIAL_SYMBOL_STR(ATTRIBUTE_NT);
      SDN_SERIAL_SYMBOL_STR(CHILD);
      SDN_SERIAL_SYMBOL_STR(PTR);
      SDN_SERIAL_SYMBOL_STR(CONTENT);
      SDN_SERIAL_SYMBOL_STR(ELEMENTS);
      SDN_SERIAL_SYMBOL_STR(ELEMENT);
    }
  SDN_FATAL("Uncrecognized symbol");
  return "";
}


Parser::Parser()
{
  stack_.push(S);
}


void Parser::debug(Symbol t)
{
#if 0
  string msg = "Processed "+toString(t)+", stack is now ";
  stack<Symbol> other(stack_);
  while (!other.empty())
    {
      msg += toString(other.top())+" ";
      other.pop();
    }
  SDN_DEBUG(msg);
#endif
}

void Parser::processTerminal(Symbol t)
{
  /// verifies that the incoming terminal is valid using an LL(1) parse table (if you don't know what this means, look it up before you edit anything)
  /// Note-- if you think there are mistakes in here, or want to make lots of changes to the language, it probably woudln't be too hard to auto-generate the parse table from the language description
  switch (stack_.top())
    {
    case S:
      switch (t)
	{
	case FLUSH:
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //predict S->CLASS_INSTANCES
	  stack_.pop();
	  stack_.push(CLASS_INSTANCES);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command. This command can't be the first command to an archive");
	}
      break;


    case CLASS_INSTANCES:
      switch (t)
	{
	case FLUSH:
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //predict CLASS_INSTANCES->FLUSHES CLASS_INSTANCE CLASS_INSTANCES
	  stack_.pop();
	  stack_.push(CLASS_INSTANCES);
	  stack_.push(CLASS_INSTANCE);
	  stack_.push(FLUSHES);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing class instances list. Told to "<<toString(t)<<", expected FLUSH, BEGIN_BASE, or BEGIN_DERIVED");
	}
      break;


    case FLUSHES:
      switch (t)
	{
	case FLUSH:
	  //Match first token of FLUSHES -> FLUSH FLUSHES
	  //do nothing, because we would actually pop FLUSHES and then push it back
	  debug(t);
	  break;
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //Match FLUSHES -> e, keep looking for match to t
	  stack_.pop();
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command at a time when all object definitions are complete.");
	}
      break;

    case CLASS_INSTANCE:
      switch (t)
	{
	case BEGIN_BASE:
	  //Predict CLASS_INSTANCE->BASE_CLASS_INSTANCE
	  stack_.pop();
	  stack_.push(BASE_CLASS_INSTANCE);
	  debug(t);
	  processTerminal(t);
	  break;
	case BEGIN_DERIVED:
	  //Preict CLASS_INSTANCE->DERIVED_CLASS_INSTANCE
	  stack_.pop();
	  stack_.push(DERIVED_CLASS_INSTANCE);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing a class instance. Got "<<toString(t)<<", expected BEGIN_BASE or BEGIN_DERIVED");
	}
      break;

    case BASE_CLASS_INSTANCE:
      switch (t)
	{
	case BEGIN_BASE:
	  //match first token of BASE_CLASS_INSTANCE -> BEGIN_BASE INSTANCE_BODY END_BASE
	  stack_.pop();
	  stack_.push(END_BASE);
	  stack_.push(INSTANCE_BODY);
	  debug(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing a base class instance");
	}
      break;

    case DERIVED_CLASS_INSTANCE:
      switch (t)
	{
	case BEGIN_DERIVED:
	  //Predict DERIVED_CLASS_INSTANCE -> SUPER_CLASS_INSTANCE BEGIN_DERIVED INSTANCE_BODY END_DERIVED
	  stack_.pop();
	  stack_.push(END_DERIVED);
	  stack_.push(INSTANCE_BODY);
	  stack_.push(SUPER_CLASS_INSTANCE);
	  stack_.push(BEGIN_DERIVED);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing a derived class instance. Got "<<toString(t)<<", expected BEGIN_DERIVED");
	}
      break;

    case SUPER_CLASS_INSTANCE:
      switch(t)
	{
	case BEGIN_SUPER:
	  //Match first token of SUPER_CLASS_INSTANCE->BEGIN_SUPER CLASS_INSTANCE END_SUPER
	  stack_.pop();
	  stack_.push(END_SUPER);
	  stack_.push(CLASS_INSTANCE);
	  debug(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command "<<toString(t)<<" while serializing a superclass");
	}
      break;

    case INSTANCE_BODY:
      switch (t)
	{
	case ATTRIBUTE:
	case BEGIN_PTR:
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	case CUSTOM_CONTENT:
	case END_BASE:
	case END_DERIVED:
	  //Predict INSTANCE_BODY -> ELEMENTS CONTENT
	  stack_.pop();
	  stack_.push(CONTENT);
	  stack_.push(ELEMENTS);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing class instance body");
	}
      break;





    case ATTRIBUTE_NT:
      switch (t)
	{
	case ATTRIBUTE:
	  //Match ATTRIBUTE_NT -> ATTRIBUTE
	  stack_.pop();
	  debug(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing attribute");
	}
      break;

    case CHILD:
      switch (t)
	{
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //Predict CHILD->CLASS_INSTANCE
	  stack_.pop();
	  stack_.push(CLASS_INSTANCE);
	  debug(t);
	  processTerminal(t);
	  break;
	case BEGIN_PTR:
	  //Predict CHILD->PTR
	  stack_.pop();
	  stack_.push(PTR);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing class member");
	}
      break;

    case PTR:
      switch(t)
	{
	case BEGIN_PTR:
	  //Match first token of PTR -> BEGIN_PTR CLASS_INSTANCE END_PTR
	  stack_.pop();
	  stack_.push(END_PTR);
	  stack_.push(CLASS_INSTANCE);
	  debug(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing a pointer");
	}
      break;

    case CONTENT:
      switch (t)
	{
	case CUSTOM_CONTENT:
	  //Match CONTENT -> CUSTOM_CONTENT
	  stack_.pop();
	  debug(t);
	  break;
	case END_BASE:
	case END_DERIVED:
	  //Match CONTENT -> e, keep looking for a match for t
	  stack_.pop();
	  debug(t);
	  processTerminal(t);
	  break;

	default:
	  SDN_FATAL("Invalid serialization command while serializing custom content");

	}
      break;


    case ELEMENTS:
      switch(t)
	{
	case ATTRIBUTE:
	case BEGIN_PTR:
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //Predict ELEMENTS -> ELEMENT ELEMENTS
	  stack_.pop();
	  stack_.push(ELEMENTS);
	  stack_.push(ELEMENT);
	  debug(t);
	  processTerminal(t);
	  break;
	case CUSTOM_CONTENT:
	case END_BASE:
	case END_DERIVED:
	  //Match ELEMENTS -> e, keep looking for T
	  stack_.pop();
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command "<<toString(t)<<" while serializing class members");
	}

      break;

    case ELEMENT:
      switch(t)
	{
	case ATTRIBUTE:
	  //Match ELEMENT -> ATTRIBUTE
	  stack_.pop();
	  debug(t);
	  break;
	case BEGIN_PTR:
	case BEGIN_BASE:
	case BEGIN_DERIVED:
	  //Predict ELEMENT -> CHILD
	  stack_.pop();
	  stack_.push(CHILD);
	  debug(t);
	  processTerminal(t);
	  break;
	default:
	  SDN_FATAL("Invalid serialization command while serializing class member");
	}
      break;

    default:
      //We've enumerated all of the non-terminals, so we should only be left with terminals that can only match themselves
      if (stack_.top() == t)
	{
	  stack_.pop();
	  debug(t);
	}
      else
	SDN_FATAL("Expected "<<toString(stack_.top())<<", got "<<toString(t));
    }

  
}
