/// \author Ian Goodfellow
/// \file pointer.h
/// \brief Utilities for working with boost/shared_ptr

#pragma once

#include <boost/smart_ptr.hpp>
#include "registry.h"
#include "inArchive.h"

namespace sdn
{
  template <typename T>
  class Registry;


  /** 
      Note-- we should probably quit using this function and switch to using boost/make_shared.hpp  -Ian

      Returns a shared pointer to an object. You should only
      call this on the output of operator new.
      (It is just for convenience so you don't have to write out
      shared_ptr and the template argument everywhere) */
  template <typename T>   boost::shared_ptr<T> ptr(T * p);

  /** Returns a shared_ptr that will not do reference counting.
      Use this when taking the address of a statically allocated
      object.
      If you create such a pointer, you must be certain that
      it / any pointers copied from it will die before the object
      does. */
  template <typename T> boost::shared_ptr<T> owned_ptr(T * p);
 
  /// Serializes a polymorphic shared pointer
  template <typename T>
    void serialize_shared_ptr(serial::Archive & ar, Registry<T> & r, boost::shared_ptr<T> & p);
 
  /// Serializes a fixed-type shared pointer
  template <typename T>
    void serialize_shared_ptr(serial::Archive & ar, boost::shared_ptr<T> & p);


  /// Serializes a fixed-type shared pointer to a type that doesn't have a serialize method
  template <typename T, typename Serializer>
    void serialize_shared_ptr(serial::Archive & ar, boost::shared_ptr<T> & p, Serializer s);



  /// Serializer object (suitable for use with STL vector serialization) for polymorphic shared pointers
  template<typename T>
    class PolymorphicSharedPointerSerializer;

  /// Serializer object (suitable for use with STL vector serialization) for fixed-type shared pointers
  template<typename T>
    class SharedPointerSerializer;



  //Implementations
  
  
  template <typename T> boost::shared_ptr<T> ptr(T * p)
    {
      return boost::shared_ptr<T>(p);
    }
  
  
  /*A function object that does nothing, used to implement 
    owned_ptr */
  struct NoOp
  {
    void operator()(void const *) const
    {
    }
  };
  
  
  template <typename T> boost::shared_ptr<T> owned_ptr(T * p)
    {
      return boost::shared_ptr<T>(p, NoOp());
    }
  
  template <typename T>
    void serialize_shared_ptr(serial::Archive & ar, Registry<T> & r, boost::shared_ptr<T> & p)
    {
      if (ar.in())
	{
	  p = r.load(ar);
	}
      else
	{
	  SDN_ASSERT_MSG(p, "todo-- support null pointers");
	  p->serialize(ar);
	}
    }
  
  template <typename T>
    void serialize_shared_ptr(serial::Archive & ar, boost::shared_ptr<T> & p)
    {
      if (ar.in())
	{
	  serial::InArchive * iar = dynamic_cast<serial::InArchive *>(&ar);
	  
	  //Pointer to a shared pointer for an object that has already been deserialized
	  void * prevPtr = NULL;
	  std::string derivedId;
	  int version;
	  
	  iar->peek(prevPtr,derivedId, version);
	  if (prevPtr)
	    {
	      p =  * (boost::shared_ptr<T> *) prevPtr;
	    }
	  else
	    {
	      boost::shared_ptr<T> * temp = new boost::shared_ptr<T>(NoArgFactory<T>()());
	      SDN_ASSERT(temp);

	      if (*temp)
		{
		  iar->reportPointer(temp, new sdn::serial::DerivedDeleter<T>(temp));
		  (*temp)->serialize(ar);
		  p = *temp;
		}
	      else
		{
		  delete temp;
		  p = ptr( (T*) NULL);
		}
	    }
	}
      else
	{
	  SDN_ASSERT_MSG(p, "todo-- support null pointers");
	  p->serialize(ar);
	}
    }

  template <typename T, typename Serializer>
    void serialize_shared_ptr(serial::Archive & ar, boost::shared_ptr<T> & p, Serializer s)
  {
    if (ar.in())
      {
	serial::InArchive * iar = dynamic_cast<serial::InArchive *>(&ar);
	
	//Pointer to a shared pointer for an object that has already been deserialized
	void * prevPtr = NULL;
	std::string derivedId;
	int version;
	
	iar->peek(prevPtr,derivedId, version);


	if (prevPtr)
	  {
	    p = * (boost::shared_ptr<T> *) prevPtr;
	    SDN_DEBUG("Using pre-existing shared pointer");
	  }
	else
	  {
	    boost::shared_ptr<T> * temp = new boost::shared_ptr<T>(NoArgFactory<T>()());
	    SDN_ASSERT(temp);
	    
	    if (*temp)
	      {
		iar->reportPointer(temp, new sdn::serial::DerivedDeleter<T>(temp));
		SDN_DEBUG("De-serializing weights now");
		s(ar, **temp);
		p = *temp;
	      }
	    else
	      {
		delete temp;
		p = ptr( (T*) NULL);
	      }
	  }
      }
    else
      {
	SDN_ASSERT_MSG(p, "TODO-- support NULL pointers");
	s(ar,*p);
      }
  }
  

template<typename T>
    class PolymorphicSharedPointerSerializer
    {
    public:

      PolymorphicSharedPointerSerializer(Registry<T> & registry) : registry_(registry)
	{
	}

      void operator()(serial::Archive & ar, boost::shared_ptr<T> & p)
      {
	serialize_shared_ptr(ar, registry_, p);
      }

    private:

      Registry<T> & registry_;

    };


template<typename T>
    class SharedPointerSerializer
    {
    public:

      SharedPointerSerializer() 
	{
	}

      void operator()(serial::Archive & ar, boost::shared_ptr<T> & p)
      {
	serialize_shared_ptr(ar, p);
      }

  
    };
}


