/// \author Ian Goodfellow
/// \file random.h
/// \brief Random number generation classes


#pragma once

#include <boost/random.hpp>
#include <boost/smart_ptr.hpp>
#include "logger.h"
#include "stlUtils.h"


namespace sdn
{

  typedef boost::mt19937 generator_type;
  

  /* Base class from which other random number generators are derived.
      Does nothing on its own-- just holds the Mersenne Twister class
      that the others use to get their random primitives */
  class RNG
  {
  public:
    RNG();
    virtual ~RNG();

    virtual void setSeed(unsigned int seed);
  protected:


    generator_type generator_;
  };


  /** Generates random Scalaring point numbers.
      The point of putting this in a class is that its seed can be
      controlled independent of other random number generators in
      use elsewhere. This makes experiments easier to control-- e.g.
      if data is randomly loaded using one random number generator,
      and network is trained using sampling from another random
      number generator, then it's possible to change one component
      without altering all of the random numbers in the other. */
template <typename T=double>
  class RandomRealGenerator : public RNG
{
  public:
  RandomRealGenerator() {}
  virtual ~RandomRealGenerator() {}
  
  //returns a random real drawn from the unit interval
  T unit()
  {
    boost::uniform_real<> range(0.0,1.0);
    boost::variate_generator<generator_type&, boost::uniform_real<> > callable(generator_, range);
    return callable();
  }

  //returns a random real drawn from a uniform distribution over the specified interval
  T randUnif(T minVal, T maxVal)
  {
    boost::uniform_real<> range(minVal, maxVal);
    boost::variate_generator<generator_type&, boost::uniform_real<> > callable(generator_, range);
    return callable();
  }
};




  ///Generates random integers
  template <typename T=int>
    class RandomIntGenerator : public RNG
    {
      public:
      RandomIntGenerator() {}
      virtual ~RandomIntGenerator() {}
      
      //returns a number $\in \{ lower, \dots, upper \} $
      T randUnif(T lower, T upper)
      {
	SDN_ASSERT_MSG(lower<=upper,"Asked for random integer in interval ["<<lower<<","<<upper<<"]");
	boost::uniform_int<> range(lower,upper);
	boost::variate_generator<generator_type&,boost::uniform_int<> > callable(generator_, range);
	return callable();
      }


      /** returns a number $\in \{ 0, \dots, \text{dist.size()-1} \}$ where
	  $P(i) = dist[i]$ or if normalize is true, $P(i) \propto dist[i] $ */
      template<typename R>
      T rand(const std::vector<R>  & dist, bool normalize = false)
      {
	RandomRealGenerator<double> g;
	double v = g.unit();
	double sum = 1.0;
	if (normalize)
	  sum = double(sdn::sum(dist));

	v *= sum;

	for (unsigned i = 0; i < dist.size(); i++)
	  {
	    v -= double(dist[i]);
	    if (v <= 0)
	      return i;
	  }
	SDN_FATAL("Distribution sums to less than one");
	return -1;//included to suppress compiler warnings
      }
    };
}
