/******************************************************************************
** FILENAME:    registry.h
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              
** DESCRIPTION:
**  Allows construction of objects in cases where the type is not known
**  until runtime, either by id or from a file.
**
**  Based on svlFactory.h, by the same author, from the STAIR Vision Library
**
*****************************************************************************/

#pragma once

//#include <map>
#include <string>
#include <vector>
#include "logger.h"
#include "strUtils.h"
#include "xmlUtils.h"
#include "inArchive.h"
//#include <cstring>

/// Stanford Deep Network
namespace sdn
{


	//Functions that examine a file and return "" if the type is not
	//recognized as a member of the family or return a string identifying
	//the type of object from the family that is in the file
  typedef std::string (*FileChecker)(const char *);


/// Class for creating objects via some Registry
/**
  If you need a custom means of creating objects, subclass this class.
  Implementations that make a class based on its no-arg constructor and
  that make a class based on an instance description stored in a node
  already exist.
*/
  template <typename C>
  class Factory
  {
	public:
		virtual ~Factory() {}
		virtual boost::shared_ptr<C> operator()() = 0; 
  };

/// Creates instances of objects based on descriptions stored in XML nodes
template <typename C>
  class InstanceFactory; //forward declaration

  
/// Creates classes/instances based on string identifiers or file contents
/** This makes it unnecessary to know the type at compile-time. */
  template <typename T>
    class Registry
    {
    public:
 
     /// Constructs a registry.
/**
 \param familyName The name of the base class the registry should construct
*/ 
      Registry(const std::string & familyName);

/// Constructs an object of the class type or instnace name specified by id
      boost::shared_ptr<T> make(const std::string & id);

      /// Constructs an object from a file. Type is inferred at runtime.
      boost::shared_ptr<T> load(const char * filename);

      /// Constructs an object from an archive. Type is inferred at runtime
      boost::shared_ptr<T> load(serial::Archive & ar);


/// Constructs an object from an XML node. Type is inferred at runtime.
      boost::shared_ptr<T> load(XMLNode & node);

/// Registers a new function used to infer the type of an object stored in a file
      void registerFileChecker(FileChecker checker);

/// Registers a new factory to run when make(id) is called
/** id must be unique within each factory object. */
      void registerFactory(const std::string & id, Factory<T> * creator);
/// Registers an XMLNode to be used to create an object when make(id) is called
/** if id is left blank, the id is read from the "name" field of the XMLNode */
      void registerInstance(XMLNode & node, std::string id = "");
      ///fetches a list of the names of all registered instances
      void getInstanceNames(std::vector<std::string> & rval);

//used by the autoregistration macros
      class autoreg
      {
      public:
	autoreg(const char * id,  Factory<T> * factory , sdn::Registry<T> & base)
	  {
	    base.registerFactory( std::string(id), factory);
	  }


	
	autoreg(sdn::FileChecker checker, Registry<T> & base)
	  {
	    base.registerFileChecker(checker);
	  }

	
      private:
	autoreg() {} //disallow the no-arg constructor
      };
      
      
    protected:
      
      std::string defaultFileChecker(XMLNode & root);
      
      std::string _familyName;
      std::map<std::string, Factory<T> * > _registry;
      std::vector<FileChecker> _fileCheckers;
      std::vector<boost::shared_ptr<InstanceFactory<T> > > _instanceFactories; //stored here so that the shared_ptr will destroy them when the registry goes down / keep them alive until then
    };
  


template <typename C>
    class InstanceFactory : public Factory<C>
  {
  public:
    InstanceFactory(const XMLNode & node, Registry<C> & registry) : _node(node), _registry(registry) {}
     
	std::string getID() const
	{
		const char * name = _node.getAttribute("name");
		SDN_ASSERT(name);
		return std::string(name);
	}
	 
      boost::shared_ptr<C> operator()() { 
		return _registry.load(_node);
          }
     
     virtual ~InstanceFactory() {} 

  protected:
    XMLNode _node;
    Registry<C> & _registry;
  };



  //Registry implementation


  template <typename T>
    Registry<T>::Registry(const std::string & familyName) : _familyName(familyName)
    {
    }
    
    template <typename T>
      boost::shared_ptr<T> Registry<T>::make(const std::string & type)
      {
	typename std::map<std::string, Factory<T> * >::iterator result = _registry.find(type);
	if (result == _registry.end())
	  {
	    SDN_WARN( "sdn::Registry<"<<_familyName<<"> asked to make object of unknown id \""<<type<<"\"");
	    return boost::shared_ptr<T>((T *) NULL);
	  }
	return (*result->second)();
      }
    
    template <typename T>
      boost::shared_ptr<T> Registry<T>::load(const char * filename)
      {
	std::ifstream ifs;
	ifs.open(filename);
	if (ifs.fail())
     	  {
	    SDN_WARN( "sdn::Registry"<<_familyName<<">::load: File \""<<filename<<"\" does not exist.");
	    return boost::shared_ptr<T>((T *) NULL);
    	  }
	ifs.close();
	
	XMLNode root = openXML(filename);
	
	std::string xmlResult = defaultFileChecker(root);
	
	if (xmlResult != "")
    	  {
	    boost::shared_ptr<T> rval = make(xmlResult);
	    if (rval)
	      {
		if (rval->load(root))
		  {
		    return rval;
		  }
		else
		  {
		    delete rval;
		    return NULL;
		  }
	      }
	    
	    SDN_WARN( "sdn::Registry<"<<_familyName<<">::load: Identified file \""<<filename<<"\" as being of type \""<<xmlResult<<"\", but could not allocate an object of that type.");
	    return NULL;
    	  }
	
  	for (unsigned i = 0; i < _fileCheckers.size(); i++)
	  {
	    std::string result = _fileCheckers[i](filename);
	    if (result != "")
	      {
	  	boost::shared_ptr<T> rval = make(result);
	  	if (rval)
		  {
		    if (rval->load(filename))
		      {
			return rval;
		      }
		    else
		      {
			delete rval;
			return NULL;
		      }
	    	}
	  	else
		  {
		    SDN_WARN( "sdn::Registry<"<<_familyName<<">::load: Identified file \""<<filename<<"\" as being of type \""<<result<<"\", but could not allocate an object of that type.");
		    return NULL;
		  }
	      }
	  }
	
	SDN_WARN( "sdn::Registry<"<<_familyName<<">::load: No registered classifier file checking function could identify the contents of file \""<<filename<<"\"");
	
	return NULL;
      }
    
    template <typename T>
      boost::shared_ptr<T> sdn::Registry<T>::load(XMLNode & node)
      {
	std::string xmlResult = defaultFileChecker(node);
	
	if (xmlResult != "")
	  {
	    boost::shared_ptr<T> rval = make(xmlResult);
	    if (rval)
	      {
		rval->load(node); //todo-- catch exception and return NULL if load fails
		return rval;
	      }
	  }
	else
	  {
	    std::string optname;
	    if (node.getName())
	      {
		optname=std::string(" (\"")+node.getName()+"\")";
	      }
	    SDN_WARN("sdn::Registry<"<<_familyName<<">::load: Could not identify class"<<optname<<" contained in XMLNode");
	  }
	return boost::shared_ptr<T>( (T * ) NULL);
      }
    
    template <typename T>
    boost::shared_ptr<T> Registry<T>::load(serial::Archive & ar)
      {
	serial::InArchive * iar = dynamic_cast<serial::InArchive *>(&ar);

	if (!iar)
	  {
	    SDN_FATAL("sdn::Registry<"<<_familyName<<">::load was called on an archive that does not support reading");
	  }

	//Pointer to a shared pointer for an object that has already been deserialized
	void * prevPtr = NULL;
	std::string derivedId;
	int version;

	iar->peek(prevPtr,derivedId, version);
	if (prevPtr)
	  {
	    return * (boost::shared_ptr<T> *) prevPtr;
	  }

	boost::shared_ptr<T> * rval = new boost::shared_ptr<T>(make(derivedId));
	SDN_ASSERT(rval);
	if (*rval)
	  {	    
	    iar->reportPointer(rval, new serial::DerivedDeleter<T>(rval));
	    (*rval)->serialize(ar);
	    return *rval;
	  }

	delete rval;
	
	SDN_WARN("sdn::Registry<"<<_familyName<<">::load: Could not allocate class of type "<<derivedId<<" as requested by archive.");
	return boost::shared_ptr<T>( (T * ) NULL);
      }




    template <typename T>
      void Registry<T>::registerFileChecker(FileChecker checker)
      {
	
	for (unsigned i = 0; i < _fileCheckers.size(); i++)
	  if (_fileCheckers[i] == checker)
	    {
	      SDN_WARN( "sdn::Registry<"<<_familyName<<">::registerTypeFileChecker: prevented attempt to register the same checker twice");
	      return;
	    }
	
	_fileCheckers.push_back(checker);
      }
    
    template <typename T>
      void Registry<T>::registerFactory(const std::string & id, Factory<T> * creator)
      {
	if (_registry.find(id) != _registry.end())
	  {
	    SDN_WARN( "sdn::Registry<"<<_familyName<<"::registerType: id \""<<id<<"\" registered twice.");
	  }
	
	_registry.insert(std::make_pair(std::string(id), creator));
      }

template <typename T>
  void Registry<T>::registerInstance(XMLNode & node, std::string id)
  {
    InstanceFactory<T> * temp = new InstanceFactory<T>(node, *this);
    SDN_ASSERT(temp);
    _instanceFactories.push_back(boost::shared_ptr<InstanceFactory<T> >(temp));
    if (id == "")
	id = temp->getID();
    registerFactory(id, dynamic_cast<Factory<T> *>(temp));
  }

 template <typename T>
   void Registry<T>::getInstanceNames(std::vector<std::string> & rval)
   {
     rval.resize(_instanceFactories.size());
     for (unsigned i = 0; i < _instanceFactories.size(); i++)
       rval[i] = _instanceFactories[i]->getID();
   }

    
    template <typename T>
      std::string Registry<T>::defaultFileChecker(XMLNode & root)
      {
	//This function should not display error messages because it could just
	//be that it is not the right file checker being called. The caller
	//should already have checked that the file exists, so we don't need
	//to report that either (though none of this code should crash if
	//the file does not exist)
	
	if (root.isEmpty() || ! root.getName())
	  {
	    return "";
	  }
	
	if (!strcmp(root.getName(),_familyName.c_str()))
	  {
	    return std::string(root.getAttribute("id"));
	  }
	
	return "";
      }
  



    template <typename C,typename D=C>
      class NoArgFactory : public Factory<C>
    {
    public:
      boost::shared_ptr<C> operator()()
      {
	return boost::shared_ptr<C>(new D());
      }

    virtual ~NoArgFactory() {}

    };
    
    
    //Macro for registering a class to be automatically created
    //by name using the no-arg constructor

//For instance, to autogregister StandardLayer to be
//loaded by LayerRegistry(), use 
// SDN_AUTOREGISTER( Layer , StandardLayer )
#define SDN_AUTOREGISTER( ___baseType___, ___derivedType___ )	      \
    namespace sdn { \
      static sdn::NoArgFactory<sdn::___baseType___, sdn::___derivedType___> noArgAutoreg ## ___baseType___ ## ___derivedType___; \
  sdn::Registry<sdn::___baseType___>::autoreg Autoreg ## ___baseType___ ## ___derivedType___ \
      ( (std::string("sdn::") + SDN_MAKE_STRING(___derivedType___)).c_str() ,\
	& sdn:: noArgAutoreg ## ___baseType___ ## ___derivedType___ ,	\
	___baseType___ ## Registry()); \
    }
    
    
//Similar macros for registering file checkers
#define SDN_AUTOREGISTER_FILECHECKER_H( fileChecker, objectType)	\
    extern svlRegistry<objectType>::autoreg sdnAutoreg ## fileChecker ## objectType;
    
#define SDN_AUTOREGISTER_FILECHECKER_CPP( fileChecker, objectType) \
    svlRegistry<objectType>::autoreg sdnAutoreg ## fileChecker ## objectType \
      ( & fileChecker, objectType ## Registry());
    
    
}


