/// \author Ian Goodfellow
/// \brief Utilties for working with the STL (string functionality is in strUtils.h)

#pragma once

#include "archive.h"
#include <vector>
#include <string>
#include <algorithm>

namespace sdn
{
  template<typename T, typename Serializer>
    void serializeSTLVector(serial::Archive & ar, std::vector<T> & v, Serializer s)
  {
    int version = 22;
    version = ar.beginBase(&v, version, std::string("std::vector<>"));
    
    
    switch (version)
    {
    case 22:
      {
	unsigned sz = v.size();
	ar.attribute("size",sz);
	
	if (ar.in())
	  v.resize(sz);
	
	for (unsigned i = 0; i < sz; i++)
	  s(ar,v[i]);
	
	ar.endBase();
      }
      break;
    default:
      SDN_FATAL("Invalid std::vector version requested");     
    }
   
  } 

  //Returns the maximum element of an STL vector
  template <typename T>
    T maxElem(const std::vector<T> & v)
    {
      SDN_ASSERT(v.size());
      T maxVal = v[0];

      for (unsigned i = 1; i < v.size(); i++)
	{
	  if (v[i] > maxVal)
	    maxVal = v[i];
	}
      return maxVal;
    }

  //Returns the index of the maximum element of an STL vector
  template <typename T>
    T maxIndex(const std::vector<T> & v)
    {
      SDN_ASSERT(v.size());
      T maxVal = v[0];
      unsigned maxInd = 0;

      for (unsigned i = 1; i < v.size(); i++)
	{
	  if (v[i] > maxVal)
	    {
	      maxVal = v[i];
	      maxInd = i;
	    }
	}

      return maxInd;
    }
  
  ///Returns the sum of an STL vector
  template <typename T>
    T sum(const std::vector<T> & v)
    {
      T rval = 0;
      for (unsigned i = 0; i < v.size(); i++)
	rval += v[i];
      return rval;
    }

  ///Returns the mean of an STL vector
  template <typename T>
    T mean(const std::vector<T> & v)
    {
      return sum(v)/T(v.size());
    }


  //Performs comparison on any object that uses operator< (useful for 
  template <typename T>
  class Comparator
    {
    public:
      bool operator()(const T & a, const T & b) { return a < b; }
    };

  ///Adds an element to a vector in sorted position (assuming vector is already sorted)
  template <typename T>
    void addSorted(std::vector<T> & v, const T & elem)
    {
      v.insert(std::lower_bound(v.begin(),v.end(), elem),elem);
    }


}
