#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <algorithm>
#include <cuda_runtime.h>
#include <cublas.h>
//#include "../base/logger.h"
#include "cudaMatrixUtils.h"
#include "cudaMemHandler.h"
#include "Eigen/Core"

#include "cudaMatrix.h"
#include "cudaVector.h"
// #include <culapackdevice.h>

// Decouple from SDN
#ifndef SDN_ASSERT
#define SDN_ASSERT(x) assert(x)
#endif

#ifndef SDN_FATAL
#define SDN_FATAL(x) {std::cerr << x << "\n"; assert(0);}
#endif

#ifndef SDN_ASSERT_MSG
#define SDN_ASSERT_MSG(C, M) \
  if (!(C)) { SDN_FATAL( #C << ": " << M); }
#endif

bool randInit = false;

inline bool checkSuccess(cublasStatus status) {
  return status == CUBLAS_STATUS_SUCCESS;
}

bool sdn::isColumnMajor(const CudaMatrix& matrix) { return true; }

int sdn::pad(int d) {
  if(d <= 1) return 1;
  if(d <= 64)
    return ((d - 1) | (CUDA_MATRIX_PAD_SMALL - 1)) + 1;
  return ((d - 1) | (CUDA_MATRIX_PAD - 1)) + 1;
}

sdn::CudaMatrix sdn::CudaMatrix::Zero(int m, int n) {
  return Constant(m, n, 0.0);
}

sdn::CudaMatrix sdn::CudaMatrix::Ones(int m, int n) {
  return Constant(m, n, 1.0);
}

sdn::CudaMatrix sdn::CudaMatrix::Eye(int m, int n, float scalar) {
  SDN_ASSERT(m > 0 && n > 0);

  // implemented on CPU for now

  int pm = pad(m), pn = pad(n);
  float* matrix = (float*)malloc(pm*pn*sizeof(float));
  if(matrix == NULL) {
    SDN_FATAL("CudaMatrix::Eye malloc failed!\n");
  }

  for(int i = 0; i < pn; i++)
    for(int j = 0; j < pm; j++)
      if (i < n && j < m && i == j)
          matrix[i*pm + j] = scalar;
        else
          matrix[i*pm + j] = 0.0f;

  return CudaMatrix(m, n, matrix, true, false);
}

sdn::CudaMatrix sdn::CudaMatrix::Constant(int m, int n, float scalar) {
  SDN_ASSERT(m > 0 && n > 0);

  int pm = pad(m), pn = pad(n);
  float* matrix = CudaMemHandler::alloc(pm*pn*sizeof(float));
  cudaMatrixSet(matrix, scalar, pm, pn, m, n);
  return CudaMatrix(m, n, matrix, false, false);
}

sdn::CudaMatrix sdn::CudaMatrix::Random(int m, int n) {
  SDN_ASSERT(m > 0 && n > 0);

  if(!randInit) {
    srand(time(NULL));
    randInit = true;
  }

  int pm = pad(m), pn = pad(n);
  float* matrix = (float*)malloc(pm*pn*sizeof(float));
  if(matrix == NULL) {
    SDN_FATAL("CudaMatrix::Random malloc failed!\n");
  }
  for(int i = 0; i < pn; i++)
    for(int j = 0; j < pm; j++)
      matrix[i*pm + j] = (i < n && j < m) ?
	(2.0*rand() - RAND_MAX) / RAND_MAX : 0.0;
  return CudaMatrix(m, n, matrix, true, false);
}

sdn::CudaMatrix sdn::CudaMatrix::Random(int m) {
  return Random(m, 1);
}

sdn::CudaMatrix::CudaMatrix() {
  m_ = 0; n_ = 0;
  pm_ = 1; pn_ = 1;
  matrixCPU_ = matrixGPU_ = NULL;
  dirtyCPU_ = dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;
}

sdn::CudaMatrix::CudaMatrix(int n)
  : m_(n), n_(n) {
  SDN_ASSERT(n > 0);

  pm_ = pn_ = pad(n);
  matrixCPU_ = matrixGPU_ = NULL;
  dirtyCPU_ = dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;
}

sdn::CudaMatrix::CudaMatrix(int m, int n)
  : m_(m), n_(n) {
  SDN_ASSERT(m > 0 && n > 0);

  pm_ = pad(m); pn_ = pad(n);
  matrixCPU_ = matrixGPU_ = NULL;
  dirtyCPU_ = dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;
}

sdn::CudaMatrix::CudaMatrix(const CudaMatrix& other) {
  m_ = n_ = pm_ = pn_ = 0;
  matrixCPU_ = matrixGPU_ = NULL;
  dirtyCPU_ = dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;

  *this = other;
}

sdn::CudaMatrix::CudaMatrix(const Eigen::MatrixXf& other) {
  m_ = other.rows();
  n_ = other.cols();
  pm_ = pad(m_);
  pn_ = pad(n_);

  matrixGPU_ = NULL;
  dirtyCPU_ = false;
  dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;

  matrixCPU_ = (float*)malloc(pm_ * pn_ * sizeof(float));
  if (matrixCPU_ == NULL) {
    SDN_FATAL("CudaMatrix::CudaMatrix malloc failed!\n");
  }

  for (int i = 0; i < pn_; i++)
    for (int j = 0; j < pm_; j++)
      matrixCPU_[i*pm_ + j] = (i < n_ && j < m_) ? other(j, i) : 0.0f;
}

sdn::CudaMatrix::CudaMatrix(const CudaMatrixBlock& block) {
  m_ = n_ = pm_ = pn_ = 0;
  matrixCPU_ = matrixGPU_ = NULL;
  dirtyCPU_ = dirtyGPU_ = true;
  transpose_ = false;

  randSeeds_ = NULL;
  nSeeds_ = 0;

  *this = block;
}

sdn::CudaMatrix::~CudaMatrix() {
  if(matrixCPU_ != NULL)
    free(matrixCPU_);
  if(matrixGPU_ != NULL)
    CudaMemHandler::free(matrixGPU_);
  if (randSeeds_ != NULL)
    cudaFree(randSeeds_);
}

Eigen::MatrixXf sdn::CudaMatrix::ToEigenMatrix() {
  Eigen::MatrixXf ret(rows(), cols());
  dirtyCPU_ = true; // force the update in the first call to ()
  for (int i = 0; i < rows(); i++)
    for (int j = 0; j < cols(); j++)
      ret(i, j) = (*this)(i, j);
  return ret;
}

float& sdn::CudaMatrix::operator()(int i, int j) {
  SDN_ASSERT(i >= 0 && i < rows() && j >= 0 && j < cols());

  updateMatrixCPU();
  dirtyGPU_ = true;
  if(transpose_) return matrixCPU_[i*pm_+j];
  return matrixCPU_[j*pm_+i];
}

const float sdn::CudaMatrix::operator()(int i, int j) const {
  SDN_ASSERT(i >= 0 && i < rows() && j >= 0 && j < cols());

  updateMatrixCPU();
  if(transpose_) return matrixCPU_[i*pm_+j];
  return matrixCPU_[j*pm_+i];
}

float& sdn::CudaMatrix::operator[](int i) {
  return (*this)(i%rows(), (int)(i/rows()));
}

const float sdn::CudaMatrix::operator[](int i) const {
  return (*this)(i%rows(), (int)(i/rows()));
}

sdn::CudaMatrix::CudaMatrixBlock
sdn::CudaMatrix::block(int start_row, int start_col,
		       int block_rows, int block_cols) {
  SDN_ASSERT(start_row >= 0 && start_row < rows() &&
	     start_col >= 0 && start_col < cols() &&
	     block_rows > 0 && start_row + block_rows <= rows() &&
	     block_cols > 0 && start_col + block_cols <= cols());

  return CudaMatrixBlock(*this, start_row, start_col, block_rows, block_cols,
			 false);
}

const sdn::CudaMatrix::CudaMatrixBlock
sdn::CudaMatrix::block(int start_row, int start_col,
		       int block_rows, int block_cols) const {
  SDN_ASSERT(start_row >= 0 && start_row < rows() &&
	     start_col >= 0 && start_col < cols() &&
	     block_rows > 0 && start_row + block_rows <= rows() &&
	     block_cols > 0 && start_col + block_cols <= cols());

  return CudaMatrixBlock(*this, start_row, start_col, block_rows, block_cols,
			 false);
}

sdn::CudaMatrix::CudaMatrixBlock sdn::CudaMatrix::col(int index) {
  return block(0, index, rows(), 1);
}

const sdn::CudaMatrix::CudaMatrixBlock sdn::CudaMatrix::col(int index) const {
  return block(0, index, rows(), 1);
}

sdn::CudaMatrix::CudaMatrixBlock sdn::CudaMatrix::row(int index) {
  return block(index, 0, 1, cols());
}

const sdn::CudaMatrix::CudaMatrixBlock sdn::CudaMatrix::row(int index) const {
  return block(index, 0, 1, cols());
}

sdn::CudaMatrix::CudaMatrixCwise sdn::CudaMatrix::cwise() {
  return CudaMatrixCwise(*this);
}

const sdn::CudaMatrix::CudaMatrixCwise sdn::CudaMatrix::cwise() const {
  return CudaMatrixCwise(*this);
}

sdn::CudaMatrix::CudaMatrixRedux sdn::CudaMatrix::rowwise() const {
  return CudaMatrixRedux(*this, false);
}

sdn::CudaMatrix::CudaMatrixRedux sdn::CudaMatrix::colwise() const {
  return CudaMatrixRedux(*this, true);
}

sdn::CudaMatrix& sdn::CudaMatrix::operator=(const CudaMatrix& rhs) {
  if(this == &rhs)
    return *this;

  m_ = rhs.m_;
  n_ = rhs.n_;

  if(pm_*pn_ != rhs.pm_*rhs.pn_) {
    // if dimensions are wrong, free CPU and GPU matrices
    if(matrixCPU_ != NULL) {
      free(matrixCPU_);
      matrixCPU_ = NULL;
    }
    if(matrixGPU_ != NULL) {
      CudaMemHandler::free(matrixGPU_);
      matrixGPU_ = NULL;
    }
    if (randSeeds_ != NULL) {
      cudaFree(randSeeds_);
      randSeeds_ = NULL;
    }
  }

  pm_ = rhs.pm_;
  pn_ = rhs.pn_;
  transpose_ = rhs.transpose_;

  // copy on GPU
  rhs.updateMatrixGPU();
  if(matrixGPU_ == NULL)
    allocateMatrixGPU();
  cudaMatrixCopy(matrixGPU_, rhs.matrixGPU_, pm_*pn_);
  dirtyCPU_ = true;
  dirtyGPU_ = false;
  
  return *this;
}

bool sdn::CudaMatrix::operator==(const CudaMatrix& rhs) const {
  return CudaMatrixBlock(*this) == rhs;
}

bool sdn::CudaMatrix::operator!=(const CudaMatrix& rhs) const {
  return CudaMatrixBlock(*this) != rhs;
}

sdn::CudaMatrix& sdn::CudaMatrix::operator=(const CudaMatrixBlock& block) {
  // synchronize GPU matrix
  const CudaMatrix& m = block.matrix_;
  m.updateMatrixGPU();

  if(m.transpose_)
    m.forceTranspose();

  float* rhsMatrixGPU = m.matrixGPU_;
  int rpm = m.pm_, rpn = m.pn_;
  bool saved = false;
  if(&m == this) {
    saved = true;
    matrixGPU_ = NULL;
    dirtyGPU_ = true;
  }
  
  m_ = block.block_rows_;
  n_ = block.block_cols_;

  if(pm_*pn_ != pad(m_)*pad(n_)) {
    // if dimensions are wrong, free CPU and GPU matrices
    if(matrixCPU_ != NULL) {
      free(matrixCPU_);
      matrixCPU_ = NULL;
    }
    if(matrixGPU_ != NULL) {
      CudaMemHandler::free(matrixGPU_);
      matrixGPU_ = NULL;
    }
  }

  pm_ = pad(m_);
  pn_ = pad(n_);
  transpose_ = block.transpose_;

  if(matrixGPU_ == NULL)
    allocateMatrixGPU();
  else
    cudaMatrixSetZero(matrixGPU_, pm_*pn_);

  cudaMatrixBlockCopy(matrixGPU_, rhsMatrixGPU, pm_, pn_, 0, 0, rpm, rpn,
		      block.start_row_, block.start_col_, m_, n_, false);
  dirtyCPU_ = true;
  dirtyGPU_ = false;

  if(saved) {
    CudaMemHandler::free(rhsMatrixGPU);
  }

  return *this;
}

sdn::CudaMatrix sdn::CudaMatrix::operator*(const CudaMatrix& rhs) const {
  SDN_ASSERT(cols() == rhs.rows());

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // multiply on GPU
  int tpm = pad(rows()), tpn = pad(rhs.cols());
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cublasSgemm((transpose_?'t':'n'), (rhs.transpose_?'t':'n'),
	      prows(), rhs.pcols(), pcols(),
	      1.0, matrixGPU_, pm_, rhs.matrixGPU_, rhs.pm_,
	      0.0, tmatrix, tpm);
  return CudaMatrix(rows(), rhs.cols(), tmatrix, false, false);
}

sdn::CudaMatrix sdn::CudaMatrix::operator+(const CudaMatrix& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());
  // return cwise() + rhs.cwise();

  // forceTranspose when one is transposed and the other is not
  if(transpose_ && !rhs.transpose_)
    forceTranspose();
  else if(!transpose_ && rhs.transpose_)
    rhs.forceTranspose();

  SDN_ASSERT(m_ == rhs.m_ && n_ == rhs.n_);

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // add on GPU
  float* tmatrix = CudaMemHandler::alloc(pm_*pn_*sizeof(float));
  cudaMatrixCopy(tmatrix, matrixGPU_, pm_*pn_);
  cublasSaxpy(pm_*pn_, 1.0, rhs.matrixGPU_, 1, tmatrix, 1);
  return CudaMatrix(m_, n_, tmatrix, false, transpose_);
}

sdn::CudaMatrix sdn::CudaMatrix::operator-(const CudaMatrix& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // forceTranspose when one is transposed and the other is not
  if(transpose_ && !rhs.transpose_)
    forceTranspose();
  else if(!transpose_ && rhs.transpose_)
    rhs.forceTranspose();

  SDN_ASSERT(m_ == rhs.m_ && n_ == rhs.n_);
  
  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // subtract on GPU
  float* tmatrix = CudaMemHandler::alloc(pm_*pn_*sizeof(float));
  cudaMatrixCopy(tmatrix, matrixGPU_, pm_*pn_);
  cublasSaxpy(pm_*pn_, -1.0, rhs.matrixGPU_, 1, tmatrix, 1);
  return CudaMatrix(m_, n_, tmatrix, false, transpose_);
}

sdn::CudaMatrix& sdn::CudaMatrix::operator+=(const CudaMatrix& rhs) {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // forceTranspose when one is transposed and the other is not
  if(transpose_ && !rhs.transpose_)
    forceTranspose();
  else if(!transpose_ && rhs.transpose_)
    rhs.forceTranspose();

  SDN_ASSERT(m_ == rhs.m_ && n_ == rhs.n_);

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // add on GPU, overwriting this
  cublasSaxpy(pm_*pn_, 1.0, rhs.matrixGPU_, 1, matrixGPU_, 1);
  dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix& sdn::CudaMatrix::operator-=(const CudaMatrix& rhs) {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // forceTranspose when one is transposed and the other is not
  if(transpose_ && !rhs.transpose_)
    forceTranspose();
  else if(!transpose_ && rhs.transpose_)
    rhs.forceTranspose();

  SDN_ASSERT(m_ == rhs.m_ && n_ == rhs.n_);

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // subtract on GPU, overwriting this
  cublasSaxpy(pm_*pn_, -1.0, rhs.matrixGPU_, 1, matrixGPU_, 1);
  dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix sdn::CudaMatrix::operator*(float scalar) const {
  // synchronize GPU matrix
  updateMatrixGPU();
  
  // multiply on GPU
  int tpm = pad(m_), tpn = pad(n_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixCopy(tmatrix, matrixGPU_, tpm*tpn);
  cublasSscal(tpm*tpn, scalar, tmatrix, 1);
  return CudaMatrix(m_, n_, tmatrix, false, transpose_);
}

sdn::CudaMatrix sdn::CudaMatrix::operator/(float scalar) const {
  SDN_ASSERT(scalar != 0);

  return operator*(1 / scalar);
}

sdn::CudaMatrix& sdn::CudaMatrix::operator*=(float scalar) {
  // synchronize GPU matrix
  updateMatrixGPU();

  // multiply on GPU, overwriting matrixGPU_
  cublasSscal(pm_*pn_, scalar, matrixGPU_, 1);
  dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix& sdn::CudaMatrix::operator/=(float scalar) {
  // synchronize GPU matrix
  updateMatrixGPU();

  // multiply on GPU, overwriting matrixGPU_
  cublasSscal(pm_*pn_, 1 / scalar, matrixGPU_, 1);
  dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix sdn::CudaMatrix::transpose() const {
  // synchronize GPU matrix
  updateMatrixGPU();

  // copy on GPU
  float* matrixGPU = CudaMemHandler::alloc(pm_*pn_*sizeof(float));
  cudaMatrixCopy(matrixGPU, matrixGPU_, pm_*pn_);
  return CudaMatrix(m_, n_, matrixGPU, false, !transpose_);
}

sdn::CudaMatrix sdn::CudaMatrix::operator-() const {
  return 0.0 - cwise();
}


sdn::CudaMatrix sdn::CudaMatrix::asDiagonal() const {
  return CudaMatrixBlock(*this).asDiagonal();
}

void sdn::CudaMatrix::transposeThis() const {
  transpose_ = !transpose_;
}

float sdn::CudaMatrix::sum() const {
  return CudaMatrixBlock(*this).sum();
}

float sdn::CudaMatrix::minCoeff() const {
  return CudaMatrixBlock(*this).minCoeff();
}

float sdn::CudaMatrix::maxCoeff() const {
  return CudaMatrixBlock(*this).maxCoeff();
}

float sdn::CudaMatrix::norm() const {
  return sqrt(squaredNorm());
}

float sdn::CudaMatrix::squaredNorm() const {
  return CudaMatrixBlock(*this).squaredNorm();
}

float sdn::CudaMatrix::trace() const {
  SDN_ASSERT(rows() == cols());
  updateMatrixGPU();
  float result;
  cudaMatrixTrace(&result, matrixGPU_, pm_, pn_, rows());
  return result;
}

void sdn::CudaMatrix::resize(int m, int n) {
  m_ = m; n_ = n;
  if(pm_ != pad(m_) || pn_ != pad(n_)) {
    pm_ = pad(m_);
    pn_ = pad(n_);
    if(matrixCPU_ != NULL) {
      free(matrixCPU_);
      matrixCPU_ = NULL;
    }
    if(matrixGPU_ != NULL) {
      CudaMemHandler::free(matrixGPU_);
      matrixGPU_ = NULL;
    }
    allocateMatrixCPU();
    dirtyCPU_ = false;
    dirtyGPU_ = true;
  }
}

sdn::CudaMatrix& sdn::CudaMatrix::setConstant(float scalar) {
  updateMatrixGPU();
  cudaMatrixSet(matrixGPU_, scalar, pm_, pn_, m_, n_);
  dirtyCPU_ = true;
  return *this;
}

void sdn::CudaMatrix::fill(float scalar) {
  setConstant(scalar);
}

void sdn::CudaMatrix::initRandomSeeds() {
  if (!cudaMatrixInitRandom(&randSeeds_, nSeeds_, pm_, pn_)) {
    SDN_FATAL("CudaMatrix::initRandomSeeds() cudaMatrixInitRandom failed");
  }
}

void sdn::CudaMatrix::fillRandomUniform() {
  // either seeds never allocated, or dimensions have changed
  
  if (randSeeds_ == NULL || cudaMatrixNumberRandomSeeds(pm_, pn_) != nSeeds_)
    initRandomSeeds();

  if (matrixGPU_ == NULL)
    allocateMatrixGPU();

  if (!cudaMatrixFillRandomUniform(matrixGPU_, randSeeds_, pm_, pn_, m_, n_)) {
    SDN_FATAL("CudaMatrix::fillRandomUniform cudaMatrixFillRandomUniform failed");
  }

  dirtyCPU_ = true;
  dirtyGPU_ = false;
}

void sdn::CudaMatrix::fillRandomGaussian() {
  // either seeds never allocated, or dimensions have changed
  
  if (randSeeds_ == NULL || cudaMatrixNumberRandomSeeds(pm_, pn_) != nSeeds_)
    initRandomSeeds();

  if (matrixGPU_ == NULL)
    allocateMatrixGPU();

  if (!cudaMatrixFillRandomGaussian(matrixGPU_, randSeeds_, pm_, pn_, m_, n_)) {
    SDN_FATAL("CudaMatrix::fillRandomGaussian cudaMatrixFillRandomGaussian failed");
  }

  dirtyCPU_ = true;
  dirtyGPU_ = false;
}

sdn::CudaMatrix::CommaInitializer sdn::CudaMatrix::operator<<(float scalar) {
  return CommaInitializer(*this, scalar);
}

int sdn::CudaMatrix::rows() const {
  return transpose_ ? n_ : m_;
}

int sdn::CudaMatrix::cols() const {
  return transpose_ ? m_ : n_;
}

int sdn::CudaMatrix::size() const {
  return rows() * cols();
}

float* sdn::CudaMatrix::data() {
  if(transpose_)
    forceTranspose();
  updateMatrixCPU();
  dirtyGPU_ = true;
  return matrixCPU_;
}

const float* sdn::CudaMatrix::data() const {
  if(transpose_)
    forceTranspose();
  updateMatrixCPU();
  return matrixCPU_;
}

int sdn::CudaMatrix::dataSize() const {
  return pm_ * pn_;
}

sdn::CudaMatrix::CudaMatrix(int m, int n, float* matrix, bool onCPU,
			    bool transpose)
  : m_(m), n_(n) {
  pm_ = pad(m); pn_ = pad(n);
  if(onCPU) {
    matrixCPU_ = matrix;
    matrixGPU_ = NULL;
    dirtyCPU_ = false;
    dirtyGPU_ = true;
  }
  else {
    matrixCPU_ = NULL;
    matrixGPU_ = matrix;
    dirtyCPU_ = true;
    dirtyGPU_ = false;
  }

  randSeeds_ = NULL;
  nSeeds_ = 0;

  transpose_ = transpose;
}

int sdn::CudaMatrix::prows() const {
  return (transpose_ ? pn_ : pm_);
}

int sdn::CudaMatrix::pcols() const {
  return (transpose_ ? pm_ : pn_);
}

void sdn::CudaMatrix::allocateMatrixCPU() const {
  SDN_ASSERT(matrixCPU_ == NULL);

  matrixCPU_ = (float*)malloc(pm_*pn_*sizeof(float));
  if(matrixCPU_ == NULL) {
    SDN_FATAL("CudaMatrix::allocateMatrixCPU malloc failed!");
  }
  memset(matrixCPU_, 0, pm_*pn_*sizeof(float));
  dirtyCPU_ = true;
}

void sdn::CudaMatrix::allocateMatrixGPU() const {
  SDN_ASSERT(matrixGPU_ == NULL);

  matrixGPU_ = CudaMemHandler::alloc(pm_*pn_*sizeof(float));
  cudaMatrixSetZero(matrixGPU_, pm_*pn_);
  dirtyGPU_ = true;
}

bool sdn::CudaMatrix::updateMatrixCPU() const {
  if(!dirtyCPU_) return false;
  
  if(matrixCPU_ == NULL)
    allocateMatrixCPU();

  // synchronize with GPU matrix (if it exists)
  if(matrixGPU_ != NULL) {
    if(!checkSuccess(cublasGetMatrix(pm_, pn_, sizeof(float), matrixGPU_, pm_,
				     matrixCPU_, pm_))) {
      SDN_FATAL("CudaMatrix::updateMatrixCPU cublasGetMatrix failed!");
    }
  }
  dirtyCPU_ = false;
  return true;
}

bool sdn::CudaMatrix::updateMatrixGPU() const {
  if(!dirtyGPU_) return false;

  if(matrixGPU_ == NULL)
    allocateMatrixGPU();

  // synchronize with CPU matrix (if it exists)
  if(matrixCPU_ != NULL) {
    if(!checkSuccess(cublasSetMatrix(pm_, pn_, sizeof(float), matrixCPU_, pm_,
				     matrixGPU_, pm_))) {
      SDN_FATAL("CudaMatrix::updateMatrixGPU cublasSetMatrix failed!");
    }
  }
  dirtyGPU_ = false;
  return true;
}

bool sdn::CudaMatrix::forceTranspose() const {
  SDN_ASSERT(transpose_);

  if(pm_ != 1 && pn_ != 1) {
    // save old matrix and allocate new matrix
    updateMatrixGPU();
    float* oldMatrixGPU_ = matrixGPU_;
    matrixGPU_ = NULL;
    allocateMatrixGPU();

    // transpose on GPU
    cudaMatrixTranspose(matrixGPU_, oldMatrixGPU_, pm_, pn_);
    
    CudaMemHandler::free(oldMatrixGPU_);
    dirtyCPU_ = true;
    dirtyGPU_ = false;
  }

  std::swap(m_, n_);
  std::swap(pm_, pn_);
  transpose_ = false;
  return true;
}

bool sdn::CudaMatrix::forceTransposeCPU() const {
  if(!transpose_)
    return true;

  if(pm_ != 1 && pn_ != 1) {
    // save old matrix and allocate new matrix
    updateMatrixCPU();
    float* oldMatrixCPU_ = matrixCPU_;
    matrixCPU_ = NULL;
  
    std::swap(m_, n_);
    std::swap(pm_, pn_);
    transpose_ = false;
  
    allocateMatrixCPU();
    
    for (int i = 0; i < m_; i++) {
      for (int j = 0; j < n_; j++)
        matrixCPU_[i+j*pm_] = oldMatrixCPU_[j+i*pn_];
    }
    
    free(oldMatrixCPU_);
    dirtyCPU_ = false;
    dirtyGPU_ = true;
  }
  
  return true;
}

sdn::CudaMatrix::CudaMatrixBlock::CudaMatrixBlock(const CudaMatrix& matrix)
  : matrix_(matrix) {
  transpose_ = false;
  start_row_ = 0; start_col_ = 0;
  block_rows_ = matrix.rows(); block_cols_ = matrix.cols();
}

sdn::CudaMatrix::
CudaMatrixBlock::CudaMatrixBlock(const CudaMatrix& matrix, int start_row,
				 int start_col, int block_rows, int block_cols,
				 bool transpose)
  : matrix_(matrix), transpose_(transpose), start_row_(start_row),
    start_col_(start_col), block_rows_(block_rows), block_cols_(block_cols) { }

const float sdn::CudaMatrix::CudaMatrixBlock::operator()(int i, int j) const {
  SDN_ASSERT(i >= 0 && i < rows() && j >= 0 && j < cols());
  
  if(transpose_) return matrix_(start_col_ + j, start_row_ + i);
  return matrix_(start_row_ + i, start_col_ + j);
}

sdn::CudaMatrix::CudaMatrixCwise
sdn::CudaMatrix::CudaMatrixBlock::cwise() {
  return CudaMatrixCwise(*this);
}

const sdn::CudaMatrix::CudaMatrixCwise
sdn::CudaMatrix::CudaMatrixBlock::cwise() const {
  return CudaMatrixCwise(*this);
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator=(const CudaMatrix& matrix) {
  return (*this = CudaMatrixBlock(matrix));
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator=(const CudaMatrixBlock& rhs) {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  bool trans = m.transpose_;
  cudaMatrixBlockCopy(m.matrixGPU_, rhsM.matrixGPU_, m.pm_, m.pn_,
		      trans ? start_col_ : start_row_,
		      trans ? start_row_ : start_col_,
		      rhsM.pm_, rhsM.pn_,
		      trans ? rhsB.start_col_ : rhsB.start_row_,
		      trans ? rhsB.start_row_ : rhsB.start_col_,
		      trans ? block_cols_ : block_rows_,
		      trans ? block_rows_ : block_cols_,
		      rhsB.transpose_);
  matrix_.dirtyCPU_ = true;
  matrix_.dirtyGPU_ = false;

  return *this;
}

bool
sdn::CudaMatrix::CudaMatrixBlock::operator==(const CudaMatrixBlock& rhs) const {
  // Jaehyun: Use a kernel to parallelize?

  float tol = 1e-5;
  if(rows() != rhs.rows() || cols() != rhs.cols()) return false;
  for(int i = 0; i < rows(); i++)
    for(int j = 0; j < cols(); j++) {
      float a = operator()(i,j), b = rhs(i,j);
      if(fabs(a) < tol && fabs(b) < tol) continue;
      if(fabs(a-b) / fmax(fabs(a), fabs(b)) > tol) return false;
    }
  return true;
}

bool
sdn::CudaMatrix::CudaMatrixBlock::operator!=(const CudaMatrixBlock& rhs) const {
  return !operator==(rhs);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator+(const CudaMatrixBlock& rhs) const {
  return cwise() + rhs.cwise();
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator-(const CudaMatrixBlock& rhs) const {
  return cwise() - rhs.cwise();
}

float sdn::CudaMatrix::CudaMatrixBlock::dot(const CudaMatrixBlock& rhs) const {
  return (cwise() * rhs.cwise()).sum();
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator+=(const CudaMatrixBlock& rhs) {
  cwise() += rhs;
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator-=(const CudaMatrixBlock& rhs) {
  cwise() -= rhs;
  return *this;
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator+(float scalar) const {
  return cwise() + scalar;
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator-(float scalar) const {
  return cwise() - scalar;
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator*(float scalar) const {
  return cwise() * scalar;
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator/(float scalar) const {
  return cwise() / scalar;
}

namespace sdn {
  CudaMatrix operator*(float scalar, const CudaMatrix::CudaMatrixBlock& rhs) {
    return rhs.cwise() * scalar;
  }
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator+=(float scalar) {
  cwise() += scalar;
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator-=(float scalar) {
  cwise() -= scalar;
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator*=(float scalar) {
  cwise() *= scalar;
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::operator/=(float scalar) {
  cwise() /= scalar;
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock
sdn::CudaMatrix::CudaMatrixBlock::transpose() const {
  return CudaMatrixBlock(matrix_, start_row_, start_col_, block_rows_,
			 block_cols_, !transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::asDiagonal() const {
  SDN_ASSERT(rows() == 1 || cols() == 1);

  // synchronize GPU matrix
  matrix_.updateMatrixGPU();

  // create diagonal matrix on GPU
  int m = (block_rows_ > block_cols_) ? block_rows_ : block_cols_;
  int pm = pad(m);
  float* matrix = CudaMemHandler::alloc(pm*pm*sizeof(float));
  cudaMatrixSetZero(matrix, pm*pm);
  bool trans = matrix_.transpose_;
  cudaMatrixAsDiagonal(matrix, matrix_.matrixGPU_, pm,
		       matrix_.pm_, matrix_.pn_,
		       trans ? start_col_ : start_row_,
		       trans ? start_row_ : start_col_,
		       trans ? block_cols_ : block_rows_,
		       trans ? block_rows_ : block_cols_);

  return CudaMatrix(m, m, matrix, false, false);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixBlock::operator-() const {
  return -CudaMatrix(*this);
}

float sdn::CudaMatrix::CudaMatrixBlock::sum() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // compute sum on GPU
  bool trans = m.transpose_;
  float result;
  cudaMatrixSum(&result, m.matrixGPU_, m.pm_, m.pn_,
		trans ? start_col_ : start_row_,
		trans ? start_row_ : start_col_,
		trans ? block_cols_ : block_rows_,
		trans ? block_rows_ : block_cols_);
  return result;
}

float sdn::CudaMatrix::CudaMatrixBlock::minCoeff() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // compute min on GPU
  bool trans = m.transpose_;
  float result;
  cudaMatrixMin(&result, m.matrixGPU_, m.pm_, m.pn_,
		trans ? start_col_ : start_row_,
		trans ? start_row_ : start_col_,
		trans ? block_cols_ : block_rows_,
		trans ? block_rows_ : block_cols_);
  return result;
}

float sdn::CudaMatrix::CudaMatrixBlock::maxCoeff() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // compute max on GPU
  bool trans = m.transpose_;
  float result;
  cudaMatrixMax(&result, m.matrixGPU_, m.pm_, m.pn_,
		trans ? start_col_ : start_row_,
		trans ? start_row_ : start_col_,
		trans ? block_cols_ : block_rows_,
		trans ? block_rows_ : block_cols_);
  return result;
}

float sdn::CudaMatrix::CudaMatrixBlock::norm() const {
  return sqrt(squaredNorm());
}

float sdn::CudaMatrix::CudaMatrixBlock::squaredNorm() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // compute max on GPU
  bool trans = m.transpose_;
  float result;
  cudaMatrixSquaredNorm(&result, m.matrixGPU_, m.pm_, m.pn_,
			trans ? start_col_ : start_row_,
			trans ? start_row_ : start_col_,
			trans ? block_cols_ : block_rows_,
			trans ? block_rows_ : block_cols_);
  return result;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::setConstant(float scalar) {
  cwise().setConstant(scalar);
  return *this;
}

sdn::CudaMatrix::CudaMatrixBlock&
sdn::CudaMatrix::CudaMatrixBlock::setZero() {
  cwise().setConstant(0.0);
  return *this;
}

void sdn::CudaMatrix::CudaMatrixBlock::swap(CudaMatrixBlock other) {
  SDN_ASSERT(rows() == other.rows() && cols() == other.cols());
  CudaMatrix temp = *this;
  *this = other;
  other = temp;
}

int sdn::CudaMatrix::CudaMatrixBlock::rows() const {
  return transpose_ ? block_cols_ : block_rows_;
}

int sdn::CudaMatrix::CudaMatrixBlock::cols() const {
  return transpose_ ? block_rows_ : block_cols_;
}

sdn::CudaMatrix::CudaMatrixBlock
sdn::CudaMatrix::CudaMatrixBlock::start(int numRows) const {
  SDN_ASSERT(numRows <= rows());
  return matrix_.block(0, start_col_, numRows, cols());
}

sdn::CudaMatrix::CudaMatrixBlock
sdn::CudaMatrix::CudaMatrixBlock::end(int numRows) const {
  SDN_ASSERT(numRows <= rows());
  return matrix_.block(start_row_ + rows() - numRows, start_col_, numRows, cols());
}

namespace sdn {
  std::ostream& operator<<(std::ostream& out,
			   const CudaMatrix::CudaMatrixBlock& rhs) {
    if(rhs.transpose_) {
      for(int i = 0; i < rhs.block_cols_; i++) {
	for(int j = 0; j < rhs.block_rows_; j++) {
	  out << rhs.matrix_(rhs.start_row_ + j, rhs.start_col_ + i);
	  if(j < rhs.block_rows_ - 1) out << " ";
	}
	if(i < rhs.block_cols_ - 1) out << std::endl;
      }
    }
    else {
      for(int i = 0; i < rhs.block_rows_; i++) {
	for(int j = 0; j < rhs.block_cols_; j++) {
	  out << rhs.matrix_(rhs.start_row_ + i, rhs.start_col_ + j);
	  if(j < rhs.block_cols_ - 1) out << " ";
	}
	if(i < rhs.block_rows_ - 1) out << std::endl;
      }
    }
    return out;
  }
}

sdn::CudaMatrix::CudaMatrixCwise::CudaMatrixCwise(const CudaMatrixBlock& block)
  : CudaMatrixBlock(block) { }

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator+(float scalar) const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // add on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, scalar,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_ADDITION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator-(float scalar) const {
  return operator+(-scalar);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator*(float scalar) const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // multiply on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, scalar,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_MULTIPLICATION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator/(float scalar) const {
  SDN_ASSERT(scalar != 0);

  return operator*(1.0 / scalar);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::pow(float scalar) const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // exponentiate on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, scalar,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
		  false, CWISE_POWER);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

// needs to be declared inside namespace
namespace sdn {
  CudaMatrix operator+(float scalar, const CudaMatrix::CudaMatrixCwise& rhs) {
    return rhs + scalar;
  }

  CudaMatrix operator-(float scalar, const CudaMatrix::CudaMatrixCwise& rhs) {
    // synchronize GPU matrix
    const CudaMatrix& m = rhs.matrix_;
    m.updateMatrixGPU();
    
    // subtract on GPU
    bool trans = m.transpose_;
    int tpm = pad(trans ? rhs.block_cols_ : rhs.block_rows_);
    int tpn = pad(trans ? rhs.block_rows_ : rhs.block_cols_);
    float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
    cudaMatrixSetZero(tmatrix, tpm*tpn);
    cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, scalar,
			    tpm, tpn, m.pm_, m.pn_,
			    trans ? rhs.start_col_ : rhs.start_row_,
			    trans ? rhs.start_row_ : rhs.start_col_,
			    trans ? rhs.block_cols_ : rhs.block_rows_,
			    trans ? rhs.block_rows_ : rhs.block_cols_,
			    false, CWISE_SUBTRACTION);
    return CudaMatrix(trans ? rhs.block_cols_ : rhs.block_rows_,
		      trans ? rhs.block_rows_ : rhs.block_cols_,
		      tmatrix, false, trans ^ rhs.transpose_);
  }
  
  CudaMatrix operator/(float scalar, const CudaMatrix::CudaMatrixCwise& rhs) {
    // synchronize GPU matrix
    const CudaMatrix& m = rhs.matrix_;
    m.updateMatrixGPU();
    
    // multiply on GPU
    bool trans = m.transpose_;
    int tpm = pad(trans ? rhs.block_cols_ : rhs.block_rows_);
    int tpn = pad(trans ? rhs.block_rows_ : rhs.block_cols_);
    float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
    cudaMatrixSetZero(tmatrix, tpm*tpn);
    cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, scalar,
			    tpm, tpn, m.pm_, m.pn_,
			    trans ? rhs.start_col_ : rhs.start_row_,
			    trans ? rhs.start_row_ : rhs.start_col_,
			    trans ? rhs.block_cols_ : rhs.block_rows_,
			    trans ? rhs.block_rows_ : rhs.block_cols_,
			    false, CWISE_DIVISION);
    return CudaMatrix(trans ? rhs.block_cols_ : rhs.block_rows_,
		      trans ? rhs.block_rows_ : rhs.block_cols_,
		      tmatrix, false, trans ^ rhs.transpose_);
  }
}


sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator+=(float scalar) {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  bool trans = m.transpose_;
  cudaMatrixCWiseScalarOp(m.matrixGPU_, m.matrixGPU_, scalar,
			  m.pm_, m.pn_, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  true, CWISE_ADDITION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator-=(float scalar) {
  return operator+=(-scalar);
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator*=(float scalar) {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  bool trans = m.transpose_;
  cudaMatrixCWiseScalarOp(m.matrixGPU_, m.matrixGPU_, scalar,
			  m.pm_, m.pn_, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  true, CWISE_MULTIPLICATION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator/=(float scalar) {
  SDN_ASSERT(scalar != 0);

  return operator*=(1 / scalar);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator+(const CudaMatrixBlock& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // add on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseOp(tmatrix, m.matrixGPU_, rhsM.matrixGPU_, tpm, tpn,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    false, rhsB.transpose_, CWISE_ADDITION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator-(const CudaMatrixBlock& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // subtract on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseOp(tmatrix, m.matrixGPU_, rhsM.matrixGPU_, tpm, tpn,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    false, rhsB.transpose_, CWISE_SUBTRACTION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator*(const CudaMatrixBlock& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // multiply on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseOp(tmatrix, m.matrixGPU_, rhsM.matrixGPU_, tpm, tpn,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    false, rhsB.transpose_, CWISE_MULTIPLICATION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::operator/(const CudaMatrixBlock& rhs) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // divide on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseOp(tmatrix, m.matrixGPU_, rhsM.matrixGPU_, tpm, tpn,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    false, rhsB.transpose_, CWISE_DIVISION);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator+=(const CudaMatrixBlock& rhs) {
  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // add on GPU
  bool trans = m.transpose_;
  cudaMatrixCWiseOp(m.matrixGPU_, m.matrixGPU_, rhsM.matrixGPU_, m.pm_, m.pn_,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    true, rhsB.transpose_, CWISE_ADDITION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator-=(const CudaMatrixBlock& rhs) {
  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // subtract on GPU
  bool trans = m.transpose_;
  cudaMatrixCWiseOp(m.matrixGPU_, m.matrixGPU_, rhsM.matrixGPU_, m.pm_, m.pn_,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    true, rhsB.transpose_, CWISE_SUBTRACTION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator*=(const CudaMatrixBlock& rhs) {
  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // multiply on GPU
  bool trans = m.transpose_;
  cudaMatrixCWiseOp(m.matrixGPU_, m.matrixGPU_, rhsM.matrixGPU_, m.pm_, m.pn_,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    true, rhsB.transpose_, CWISE_MULTIPLICATION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::operator/=(const CudaMatrixBlock& rhs) {
  // synchronize GPU matrices
  const CudaMatrix& m = matrix_;
  const CudaMatrixBlock& rhsB = (transpose_ ? rhs.transpose() : rhs);
  const CudaMatrix& rhsM = rhsB.matrix_;

  m.updateMatrixGPU();
  rhsM.updateMatrixGPU();

  if(m.transpose_ && !rhsM.transpose_)
    m.forceTranspose();
  else if(!m.transpose_ && rhsM.transpose_)
    rhsM.forceTranspose();

  // divide on GPU
  bool trans = m.transpose_;
  cudaMatrixCWiseOp(m.matrixGPU_, m.matrixGPU_, rhsM.matrixGPU_, m.pm_, m.pn_,
		    m.pm_, m.pn_,
		    trans ? start_col_ : start_row_,
		    trans ? start_row_ : start_col_,
		    rhsM.pm_, rhsM.pn_,
		    trans ? rhs.start_col_ : rhs.start_row_,
		    trans ? rhs.start_row_ : rhs.start_col_,
		    trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    true, rhsB.transpose_, CWISE_DIVISION);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::square() const {
  return operator*(*this);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::exp() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // exponentiate on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_EXP);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::log() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // log on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
                          tpm, tpn, m.pm_, m.pn_,
                          trans ? start_col_ : start_row_,
                          trans ? start_row_ : start_col_,
                          trans ? block_cols_ : block_rows_,
                          trans ? block_rows_ : block_cols_,
                          false, CWISE_LOG);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
                    trans ? block_rows_ : block_cols_,
                    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::inverse() const {
  return 1 / (*this);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::abs() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // abs on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_ABS);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::tanh() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // tanh on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_TANH);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::cosh() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // cosh on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
                          tpm, tpn, m.pm_, m.pn_,
                          trans ? start_col_ : start_row_,
                          trans ? start_row_ : start_col_,
                          trans ? block_cols_ : block_rows_,
                          trans ? block_rows_ : block_cols_,
                          false, CWISE_COSH);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
                    trans ? block_rows_ : block_cols_,
                    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::sqrt() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // sqrt on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_SQRT);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::sin() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // sqrt on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_SIN);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix
sdn::CudaMatrix::CudaMatrixCwise::cos() const {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  // sqrt on GPU
  bool trans = m.transpose_;
  int tpm = pad(trans ? block_cols_ : block_rows_);
  int tpn = pad(trans ? block_rows_ : block_cols_);
  float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
  cudaMatrixSetZero(tmatrix, tpm*tpn);
  cudaMatrixCWiseScalarOp(tmatrix, m.matrixGPU_, 0.0,
			  tpm, tpn, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  false, CWISE_COS);
  return CudaMatrix(trans ? block_cols_ : block_rows_,
		    trans ? block_rows_ : block_cols_,
		    tmatrix, false, trans ^ transpose_);
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::setConstant(float scalar) {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  bool trans = m.transpose_;
  cudaMatrixCWiseScalarOp(m.matrixGPU_, m.matrixGPU_, scalar,
			  m.pm_, m.pn_, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  true, CWISE_CONSTANT);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::CudaMatrixCwise&
sdn::CudaMatrix::CudaMatrixCwise::setMax(float scalar) {
  // synchronize GPU matrix
  const CudaMatrix& m = matrix_;
  m.updateMatrixGPU();

  bool trans = m.transpose_;
  cudaMatrixCWiseScalarOp(m.matrixGPU_, m.matrixGPU_, scalar,
			  m.pm_, m.pn_, m.pm_, m.pn_,
			  trans ? start_col_ : start_row_,
			  trans ? start_row_ : start_col_,
			  trans ? block_cols_ : block_rows_,
			  trans ? block_rows_ : block_cols_,
			  true, CWISE_SETMAX);
  m.dirtyCPU_ = true;

  return *this;
}

sdn::CudaMatrix::
CudaMatrixRedux::CudaMatrixRedux(const CudaMatrix& matrix, bool colwise)
  : matrix_(matrix), colwise_(colwise) { }


sdn::CudaMatrix sdn::CudaMatrix::CudaMatrixRedux::sum() const {
  // synchronize GPU matrix
  matrix_.updateMatrixGPU();

  // compute sum on GPU
  int m = (colwise_ ? matrix_.cols() : matrix_.rows());
  int pm = pad(m);
  float* matrix = CudaMemHandler::alloc(pm*sizeof(float));
  cudaMatrixPartialSum(matrix, matrix_.matrixGPU_, matrix_.pm_, matrix_.pn_,
		       colwise_, matrix_.transpose_);

  return CudaMatrix(colwise_ ? 1 : m, colwise_ ? m : 1, matrix, false, false);
}

sdn::CudaMatrix sdn::CudaMatrix::CudaMatrixRedux::squaredNorm() const {
  // synchronize GPU matrix
  matrix_.updateMatrixGPU();

  // compute sum on GPU
  int m = (colwise_ ? matrix_.cols() : matrix_.rows());
  int pm = pad(m);
  float* matrix = CudaMemHandler::alloc(pm*sizeof(float));
  cudaMatrixPartialNormSquared(matrix, matrix_.matrixGPU_, matrix_.pm_, matrix_.pn_,
		       colwise_, matrix_.transpose_);

  return CudaMatrix(colwise_ ? 1 : m, colwise_ ? m : 1, matrix, false, false);
}

sdn::CudaMatrix sdn::CudaMatrix::CudaMatrixRedux::norm() const {
  return squaredNorm().cwise().sqrt();
}

sdn::CudaMatrix sdn::CudaMatrix::CudaMatrixRedux::maxCoeff() const {
  if(colwise_) {
    CudaMatrix m(1, matrix_.cols());
    for(int i = 0; i < m.cols(); i++)
      m(0,i) = matrix_.col(i).maxCoeff();
    return m;
  } else {
    CudaMatrix m(matrix_.rows(), 1);
    for(int i = 0; i < m.rows(); i++)
      m(i,0) = matrix_.row(i).maxCoeff();
    return m;
  }
}

sdn::CudaMatrix sdn::CudaMatrix::CudaMatrixRedux::minCoeff() const {
    if(colwise_) {
    CudaMatrix m(1, matrix_.cols());
    for(int i = 0; i < m.cols(); i++)
      m(0,i) = matrix_.col(i).minCoeff();
    return m;
  } else {
    CudaMatrix m(matrix_.rows(), 1);
    for(int i = 0; i < m.rows(); i++)
      m(i,0) = matrix_.row(i).minCoeff();
    return m;
  }
}

sdn::CudaMatrix::CommaInitializer::CommaInitializer(CudaMatrix& matrix,
						    float scalar)
  : matrix_(matrix), row_(0), col_(1) {
  matrix_(0, 0) = scalar;
}

sdn::CudaMatrix::CommaInitializer&
sdn::CudaMatrix::CommaInitializer::operator,(float scalar) {
  if(col_ == matrix_.cols()) {
    row_++;
    col_ = 0;
    SDN_ASSERT_MSG(row_ < matrix_.rows(),
		   "Too many values passed to operator<<");
  }
  SDN_ASSERT_MSG(col_ < matrix_.cols(),
		 "Too many values passed to operator<<");

  matrix_(row_, col_++) = scalar;
  return *this;
}

void sdn::CudaMatrix::syncWithGPU() {
  cudaThreadSynchronize();
}

/*
static bool culaInitialized = false;
void sdn::CudaMatrix::svd(CudaMatrix& u, CudaMatrix& vt, CudaMatrix& s) const {
  if(!culaInitialized) {
    SDN_ASSERT(culaInitialize() == culaNoError);
    culaInitialized = true;
  }

  CudaMatrix copy(*this); // need copy so we don't destroy matrix
  copy.updateMatrixGPU();

  int min_mn = std::min(copy.rows(), copy.cols());
  u = Zero(copy.rows(), min_mn);
  vt = Zero(min_mn, copy.cols());
  CudaVector sigma = CudaVector::Zero(min_mn);
  SDN_ASSERT(culaDeviceSgesvd('S', 'S', copy.rows(), copy.cols(),
			      copy.matrixGPU_, copy.pm_,
			      sigma.matrixGPU_, u.matrixGPU_, u.pm_,
			      vt.matrixGPU_, vt.pm_) == culaNoError);
  s = sigma.asDiagonal();
}
*/
