#pragma once
#include <iostream>
#include "Eigen/Core"

// Cublas documentation:
// http://developer.download.nvidia.com/compute/cuda/2_1/toolkit/docs/CUBLAS_Library_2.1.pdf

#define CUDA_MATRIX_PAD 64 // should always be power of 2
#define CUDA_MATRIX_PAD_SMALL 16 // should always be a power of 2 and div by 16

namespace sdn {

  class SparseCudaMatrix;//forward declare for friend functions

  class CudaMatrix {
  protected:
    class CudaMatrixBlock;
    class CudaMatrixCwise;
    class CudaMatrixRedux;
    class CommaInitializer;

  public:
    // for timing use
    static void syncWithGPU();

    // CudaMatrix static allocators (start in CPU memory)
    static CudaMatrix Zero(int m, int n);
    static CudaMatrix Ones(int m, int n);
    static CudaMatrix Eye(int m, int n, float scalar);
    static CudaMatrix Constant(int m, int n, float scalar);
    static CudaMatrix Random(int m, int n);
    static CudaMatrix Random(int m);

    // CudaMatrix constructors/destructors
    CudaMatrix();
    CudaMatrix(int n);
    CudaMatrix(int m, int n);
    CudaMatrix(const CudaMatrix& other);
    CudaMatrix(const Eigen::MatrixXf& other);
    CudaMatrix(const CudaMatrixBlock& block);
    ~CudaMatrix();
    
    // CudaMatrix exporter
    Eigen::MatrixXf ToEigenMatrix();

    // CudaMatrix element getting/setting
    float& operator()(int i, int j);
    const float operator()(int i, int j) const;
    float& operator[](int i);
    const float operator[](int i) const;

    // CudaMatrix block getting/setting
    CudaMatrixBlock block(int start_row, int start_col,
			  int block_rows, int block_cols);
    const CudaMatrixBlock block(int start_row, int start_col,
				int block_rows, int block_cols) const;

    // CudaMatrix column getting/setting
    CudaMatrixBlock col(int index);
    const CudaMatrixBlock col(int index) const;

    // CudaMatrix row getting/setting
    CudaMatrixBlock row(int index);
    const CudaMatrixBlock row(int index) const;

    // CudaMatrix cwise operations
    CudaMatrixCwise cwise();
    const CudaMatrixCwise cwise() const;

    // CudaMatrix rowwise/colwise reductions
    CudaMatrixRedux rowwise() const;
    CudaMatrixRedux colwise() const;

    // CudaMatrix binary operations
    CudaMatrix& operator=(const CudaMatrix& rhs);
    bool operator==(const CudaMatrix& rhs) const;
    bool operator!=(const CudaMatrix& rhs) const;
    CudaMatrix& operator=(const CudaMatrixBlock& block);
    CudaMatrix operator*(const CudaMatrix& rhs) const;
    CudaMatrix operator+(const CudaMatrix& rhs) const;
    CudaMatrix operator-(const CudaMatrix& rhs) const;

    // CudaMatrix binary operations with assignment
    CudaMatrix& operator+=(const CudaMatrix& rhs);
    CudaMatrix& operator-=(const CudaMatrix& rhs);

    // CudaMatrix scalar operations
    CudaMatrix operator*(float scalar) const;
    CudaMatrix operator/(float scalar) const;

    // CudaMatrix scalar operations with assignment
    CudaMatrix& operator*=(float scalar);
    CudaMatrix& operator/=(float scalar);

    // CudaMatrix unary operations
    CudaMatrix transpose() const;
    CudaMatrix operator-() const;
    CudaMatrix asDiagonal() const;
    
    // CudaMatrix custom transpose operations
    void transposeThis() const;
    bool forceTransposeCPU() const;

    // CudaMatrix reduction operations
    float sum() const;
    float minCoeff() const;
    float maxCoeff() const;
    float norm() const;
    float squaredNorm() const;
    float trace() const;

    // CudaMatrix resizing
    void resize(int m, int n);

    // CudaMatrix setting
    CudaMatrix& setConstant(float scalar);
    void fill(float scalar);

    // CudaMatrix random
    void initRandomSeeds();
    void fillRandomUniform();
    void fillRandomGaussian();

    // CudaMatrix comma initialization
    CommaInitializer operator<<(float scalar);

    // CudaMatrix getters
    int rows() const;
    int cols() const;
    int size() const;

    // CudaMatrix unsafe pointer access
    float* data();
    const float* data() const;
    int dataSize() const;

    // CudaMatrixBlock scalar operations
    friend CudaMatrix operator*(float scalar, const CudaMatrixBlock& rhs);

    // CudaMatrixBlock output
    friend std::ostream& operator<<(std::ostream& out,
				    const CudaMatrixBlock& rhs);

    friend SparseCudaMatrix denseTimesDenseCWiseTimesSparse(const CudaMatrix& lhs,
							    const CudaMatrix& rhs,
							    const SparseCudaMatrix& cw);
    friend SparseCudaMatrix sparseTimesDenseToSparse(const SparseCudaMatrix& lhs,
						     const CudaMatrix& rhs);
    friend CudaMatrix sparseTimesDenseToDense(const SparseCudaMatrix& lhs,
					      const CudaMatrix& rhs);

    friend CudaMatrix weightsTimesData(const SparseCudaMatrix& W, const CudaMatrix &data);

    friend class SparseCudaMatrix;

    // CudaMatrixCwise scalar operations
    friend CudaMatrix operator+(float scalar, const CudaMatrixCwise& rhs);
    friend CudaMatrix operator-(float scalar, const CudaMatrixCwise& rhs);
    friend CudaMatrix operator/(float scalar, const CudaMatrixCwise& rhs);

    // CudaMatrix other operations
    // void svd(CudaMatrix& u, CudaMatrix& vt, CudaMatrix& s) const;

    // moved out to enable timing
    bool updateMatrixCPU() const;
    bool updateMatrixGPU() const;
    
  protected:
    CudaMatrix(int m, int n, float* matrix, bool onCPU, bool transpose);
    
    // CudaMatrix protected getters
    int prows() const;
    int pcols() const;

    // methods to manage lazy operations; note they are const even though
    // they modify fields in the class
    void allocateMatrixCPU() const;
    void allocateMatrixGPU() const;
    bool forceTranspose() const;

    // mutable to get around const; make sure const methods keep the class
    // logically constant even if they modify these fields
    mutable int m_, n_;
    mutable int pm_, pn_;
    mutable float* matrixCPU_;
    mutable float* matrixGPU_;
    mutable bool dirtyCPU_;
    mutable bool dirtyGPU_;
    mutable bool transpose_;

    // for random matrix
    unsigned* randSeeds_;
    int nSeeds_;

    // protected nested class to handle block getting/setting; should never be
    // stored or manipulated by users
    class CudaMatrixBlock {
    public:
      // CudaMatrixBlock constructors
      CudaMatrixBlock(const CudaMatrix& matrix);
      CudaMatrixBlock(const CudaMatrix& matrix, int start_row, int start_col,
		      int block_rows, int block_cols, bool transpose);
      
      // CudaMatrixBlock element getting
      const float operator()(int i, int j) const;

      // CudaMatrixBlock cwise operations
      CudaMatrixCwise cwise();
      const CudaMatrixCwise cwise() const;

      // CudaMatrixBlock binary operations
      CudaMatrixBlock& operator=(const CudaMatrix& matrix);
      CudaMatrixBlock& operator=(const CudaMatrixBlock& rhs);
      bool operator==(const CudaMatrixBlock& rhs) const;
      bool operator!=(const CudaMatrixBlock& rhs) const;
      CudaMatrix operator+(const CudaMatrixBlock& rhs) const;
      CudaMatrix operator-(const CudaMatrixBlock& rhs) const;
      float dot(const CudaMatrixBlock& rhs) const;

      // CudaMatrixBlock binary operations with assignment
      CudaMatrixBlock& operator+=(const CudaMatrixBlock& rhs);
      CudaMatrixBlock& operator-=(const CudaMatrixBlock& rhs);

      // CudaMatrixBlock scalar operations
      CudaMatrix operator+(float scalar) const;
      CudaMatrix operator-(float scalar) const;
      CudaMatrix operator*(float scalar) const;
      CudaMatrix operator/(float scalar) const;

      friend CudaMatrix operator*(float scalar, const CudaMatrixBlock& rhs);

      // CudaMatrixBlock scalar operations with assignment
      CudaMatrixBlock& operator+=(float scalar);
      CudaMatrixBlock& operator-=(float scalar);
      CudaMatrixBlock& operator*=(float scalar);
      CudaMatrixBlock& operator/=(float scalar);

      // CudaMatrixBlock unary operations
      CudaMatrixBlock transpose() const;
      CudaMatrix asDiagonal() const;
      CudaMatrix operator-() const;

      // CudaMatrixBlock reduction operations
      float sum() const;
      float minCoeff() const;
      float maxCoeff() const;
      float norm() const;
      float squaredNorm() const;
      
      // CudaMatrixBlock setting
      CudaMatrixBlock& setConstant(float scalar);
      CudaMatrixBlock& setZero();
      void swap(CudaMatrixBlock other);

      // CudaMatrixBlock getters
      int rows() const;
      int cols() const;
      CudaMatrixBlock start(int numRows) const;
      CudaMatrixBlock end(int numRows) const;

      // CudaMatrixBlock output
      friend std::ostream& operator<<(std::ostream& out,
				      const CudaMatrixBlock& rhs);

      friend class CudaMatrix;

    protected:
      const CudaMatrix& matrix_;
      bool transpose_;
      int start_row_, start_col_;
      int block_rows_, block_cols_;
    };

    // protected class to handle cwise operations; should never be stored or
    // manipulated by users
    class CudaMatrixCwise : public CudaMatrixBlock {
    public:
      // CudaMatrixCwise constructors
      CudaMatrixCwise(const CudaMatrixBlock& block);

      // CudaMatrixCwise binary operations
      CudaMatrix operator+(float scalar) const;
      CudaMatrix operator-(float scalar) const;
      CudaMatrix operator*(float scalar) const;
      CudaMatrix operator/(float scalar) const;
      CudaMatrix pow(float scalar) const;
      
      friend CudaMatrix operator+(float scalar, const CudaMatrixCwise& rhs);
      friend CudaMatrix operator-(float scalar, const CudaMatrixCwise& rhs);
      friend CudaMatrix operator/(float scalar, const CudaMatrixCwise& rhs);
      
      // CudaMatrixCwise scalar operations with assignment
      CudaMatrixCwise& operator+=(float scalar);
      CudaMatrixCwise& operator-=(float scalar);
      CudaMatrixCwise& operator*=(float scalar);
      CudaMatrixCwise& operator/=(float scalar);

      // CudaMatrixCwise binary cwise operations
      CudaMatrix operator+(const CudaMatrixBlock& rhs) const;
      CudaMatrix operator-(const CudaMatrixBlock& rhs) const;
      CudaMatrix operator*(const CudaMatrixBlock& rhs) const;
      CudaMatrix operator/(const CudaMatrixBlock& rhs) const;

      // CudaMatrixCwise binary cwise operations with assignment
      CudaMatrixCwise& operator+=(const CudaMatrixBlock& rhs);
      CudaMatrixCwise& operator-=(const CudaMatrixBlock& rhs);
      CudaMatrixCwise& operator*=(const CudaMatrixBlock& rhs);
      CudaMatrixCwise& operator/=(const CudaMatrixBlock& rhs);

      // CudaMatrixCwise unary operations
      CudaMatrix square() const;
      CudaMatrix exp() const;
      CudaMatrix log() const;
      CudaMatrix inverse() const;
      CudaMatrix abs() const;
      CudaMatrix tanh() const;
      CudaMatrix cosh() const;
      CudaMatrix sqrt() const;
      CudaMatrix cos() const;
      CudaMatrix sin() const;

      // CudaMatrixCwise setting
      CudaMatrixCwise& setConstant(float scalar);
      CudaMatrixCwise& setMax(float scalar);
    };

    // protected nested class to handle rowwise/colwise reductions; should never
    // be stored or manipulated by users
    class CudaMatrixRedux {
    public:
      // CudaMatrixRedux constructors
      CudaMatrixRedux(const CudaMatrix& matrix, bool colwise);

      // CudaMatrixRedux reduction operations
      CudaMatrix sum() const;
      CudaMatrix squaredNorm() const;
      CudaMatrix norm() const;
      CudaMatrix maxCoeff() const;
      CudaMatrix minCoeff() const;

    protected:
      const CudaMatrix& matrix_;
      bool colwise_;
    };

    // protected nested class to handle comma initialization; should never be
    // stored or manipulated by users
    class CommaInitializer {
    public:
      CommaInitializer(CudaMatrix& matrix, float scalar);

      CommaInitializer& operator,(float scalar);

    protected:
      CudaMatrix& matrix_;
      int row_, col_;
    };
  };

  // CudaMatrix utility functions
  bool isColumnMajor(const CudaMatrix& matrix);
  int pad(int d);
  
}
