#include <math.h>
#include <stdio.h>
#include <cuda.h>
#include <cutil_inline.h>
#include <cublas.h>
#include <float.h>

#include "cudaMatrixUtils.h"

#define BLOCK_DIM                       (16)
#define CUBLAS_SAXPY_CTAS_MIN           (1)
#define CUBLAS_SAXPY_CTAS_MAX           (256)
#define CUBLAS_SAXPY_THREAD_MIN         (256)
#define CUBLAS_SAXPY_THREAD_MAX         (512)

struct cWiseParams {
  float* so;
  float* sl;
  float* sr;
  int n;
  int opm, opn;
  int lpm, lpn;
  int lsrow, lscol;
  int rpm, rpn;
  int rsrow, rscol;
  int nrow, ncol;
  bool inPlace;
  bool rtrans;
  CUDA_MATRIX_OPTYPE type;
};

struct cWiseScalarParams {
  float* so;
  float* si;
  float sa;
  int n;
  int opm, opn;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  bool inPlace;
  bool rtrans;
  CUDA_MATRIX_OPTYPE type;
};

struct reduxParams {
  float* so;
  float* si;
  int pm, pn;
  bool colwise;
  bool trans;
};

// For a given vector size, cublasVectorSplay() determines what CTA grid 
// size to use, and how many threads per CTA.
void cublasVectorSplay (int n, int tMin, int tMax, int gridW, int *nbrCtas, 
                        int *elemsPerCta, int *threadsPerCta) {
  if (n < tMin) {
    *nbrCtas = 1;
    *elemsPerCta = n;
    *threadsPerCta = tMin;
  } else if (n < (gridW * tMin)) {
    *nbrCtas = ((n + tMin - 1) / tMin);
    *threadsPerCta = tMin;
    *elemsPerCta = *threadsPerCta;
  } else if (n < (gridW * tMax)) {
    int grp;
    *nbrCtas = gridW;
    grp = ((n + tMin - 1) / tMin);
    *threadsPerCta = (((grp + gridW -1) / gridW) * tMin);
    *elemsPerCta = *threadsPerCta;
  } else {
    int grp;
    *nbrCtas = gridW;
    *threadsPerCta = tMax;
    grp = ((n + tMin - 1) / tMin);
    grp = ((grp + gridW - 1) / gridW);
    *elemsPerCta = grp * tMin;
  }
}

// Gets a block of a matrix and an index into that block and converts it to
// an index with respect to the whole matrix.
__device__ inline int getIndex(int pm, int pn, int srow, int scol,
			       int nrow, int ncol, int ind, bool trans) {
  if(trans) return (scol+ind%nrow)*pm + (srow+ind/nrow);
  return (scol+ind/nrow)*pm + (srow+ind%nrow);
}

// This kernel is optimized to ensure all global reads and writes are coalesced,
// and to avoid bank conflicts in shared memory.
__global__ void transpose(float *odata, float *idata, int width, int height) {
  __shared__ float block[BLOCK_DIM][BLOCK_DIM+1];
	
  // read the matrix tile into shared memory
  unsigned int xIndex = blockIdx.x * BLOCK_DIM + threadIdx.x;
  unsigned int yIndex = blockIdx.y * BLOCK_DIM + threadIdx.y;
  if((xIndex < width) && (yIndex < height)) {
    unsigned int index_in = yIndex * width + xIndex;
    block[threadIdx.y][threadIdx.x] = idata[index_in];
  }

  __syncthreads();

  // write the transposed matrix tile to global memory
  xIndex = blockIdx.y * BLOCK_DIM + threadIdx.x;
  yIndex = blockIdx.x * BLOCK_DIM + threadIdx.y;
  if((xIndex < height) && (yIndex < width)) {
    unsigned int index_out = yIndex * height + xIndex;
    odata[index_out] = block[threadIdx.x][threadIdx.y];
  }
}

void cudaMatrixTranspose(float* matrixOut, float* matrixIn, int pm, int pn) {
  dim3 grid(pm / BLOCK_DIM, pn / BLOCK_DIM, 1);
  dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
  transpose<<<grid,threads>>>(matrixOut, matrixIn, pm, pn);
}

__global__ void cudaMatrixBlockCopyMain(struct cWiseParams parms) {
  int i, tid, totalThreads, ctaStart;
  float *sl;
  const float *sr;
  int lpm, lpn;
  int lsrow, lscol;
  int rpm, rpn;
  int rsrow, rscol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  sl = parms.sl;
  sr = parms.sr;
  lpm = parms.lpm;
  lpn = parms.lpn;
  lsrow = parms.lsrow;
  lscol = parms.lscol;
  rpm = parms.rpm;
  rpn = parms.rpn;
  rsrow = parms.rsrow;
  rscol = parms.rscol;
  nrow = parms.nrow;
  ncol = parms.ncol;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
    int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, parms.rtrans);
    sl[ind1] = sr[ind2];
  }
}

int cudaMatrixBlockCopy(float* matrixLhs, float* matrixRhs, int lpm, int lpn,
			int lsrow, int lscol, int rpm, int rpn, int rsrow,
			int rscol, int nrow, int ncol, bool rtrans) {
  struct cWiseParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n  = nrow*ncol;
  params.sl = matrixLhs;
  params.sr = matrixRhs;
  params.lpm = lpm;
  params.lpn = lpn;
  params.lsrow = lsrow;
  params.lscol = lscol;
  params.rpm = rpm;
  params.rpn = rpn;
  params.rsrow = rsrow;
  params.rscol = rscol;
  params.nrow = nrow;
  params.ncol = ncol;
  params.rtrans = rtrans;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixBlockCopyMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cWiseOpMain(struct cWiseParams parms) {
  int i, tid, totalThreads, ctaStart;
  const float *sl;
  const float *sr;
  float *so;
  int opm, opn;
  int lpm, lpn;
  int lsrow, lscol;
  int rpm, rpn;
  int rsrow, rscol;
  int nrow, ncol;
  bool rtrans;
  CUDA_MATRIX_OPTYPE type;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  sl = parms.sl;
  sr = parms.sr;
  so = parms.so;
  opm = parms.opm;
  opn = parms.opn;
  lpm = parms.lpm;
  lpn = parms.lpn;
  lsrow = parms.lsrow;
  lscol = parms.lscol;
  rpm = parms.rpm;
  rpn = parms.rpn;
  rsrow = parms.rsrow;
  rscol = parms.rscol;
  nrow = parms.nrow;
  ncol = parms.ncol;
  type = parms.type;
  rtrans = parms.rtrans;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;

  if(parms.inPlace) {
    switch(type) {
    case CWISE_ADDITION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	so[ind1] = sl[ind1] + sr[ind2];
      }
      break;
    case CWISE_SUBTRACTION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	so[ind1] = sl[ind1] - sr[ind2];
      }
      break;
    case CWISE_MULTIPLICATION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	so[ind1] = sl[ind1] * sr[ind2];
      }
      break;
    case CWISE_DIVISION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	so[ind1] = sl[ind1] / sr[ind2];
      }
      break;
    }
  }
  else {
    switch(type) {
    case CWISE_ADDITION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	int ind = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind] = sl[ind1] + sr[ind2];
      }
      break;
    case CWISE_SUBTRACTION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	int ind = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind] = sl[ind1] - sr[ind2];
      }
      break;
    case CWISE_MULTIPLICATION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	int ind = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind] = sl[ind1] * sr[ind2];
      }
      break;
    case CWISE_DIVISION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind1 = getIndex(lpm, lpn, lsrow, lscol, nrow, ncol, i, false);
	int ind2 = getIndex(rpm, rpn, rsrow, rscol, nrow, ncol, i, rtrans);
	int ind = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind] = sl[ind1] / sr[ind2];
      }
      break;
    }
  }
}

int cudaMatrixCWiseOp(float* matrixOut, float* matrixLhs, float* matrixRhs,
		      int opm, int opn, int lpm, int lpn, int lsrow, int lscol,
		      int rpm, int rpn, int rsrow, int rscol, int nrow,
		      int ncol, bool inPlace, bool rtrans,
		      CUDA_MATRIX_OPTYPE type) {
  struct cWiseParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n  = nrow*ncol;
  params.so = matrixOut;
  params.sl = matrixLhs;
  params.sr = matrixRhs;
  params.opm = opm;
  params.opn = opn;
  params.lpm = lpm;
  params.lpn = lpn;
  params.lsrow = lsrow;
  params.lscol = lscol;
  params.rpm = rpm;
  params.rpn = rpn;
  params.rsrow = rsrow;
  params.rscol = rscol;
  params.nrow = nrow;
  params.ncol = ncol;
  params.inPlace = inPlace;
  params.type = type;
  params.rtrans = rtrans;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cWiseOpMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cWiseScalarOpMain(struct cWiseScalarParams parms) {
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float sa;
  float *so;
  int opm, opn;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  CUDA_MATRIX_OPTYPE type;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  sa = parms.sa;
  so = parms.so;
  opm = parms.opm;
  opn = parms.opn;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;
  type = parms.type;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;

  if(parms.inPlace) {
    switch(type) {
    case CWISE_ADDITION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = sa + si[ind2];
      }
      break;
    case CWISE_SUBTRACTION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = sa - si[ind2];
      }
      break;
    case CWISE_MULTIPLICATION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = sa * si[ind2];
      }
      break;
    case CWISE_DIVISION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = sa / si[ind2];
      }
      break;
    case CWISE_POWER:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = __powf(si[ind2], sa);
      }
      break;
    case CWISE_EXP:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = __expf(si[ind2]);
      }
      break;
    case CWISE_LOG:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
        int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
        so[ind2] = __logf(si[ind2]);
      }
      break;
    case CWISE_ABS:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = fabsf(si[ind2]);
      }
      break;
    case CWISE_TANH:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = tanhf(si[ind2]);
      }
      break;
    case CWISE_COSH:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
        int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
        so[ind2] = coshf(si[ind2]);
      }
      break;
    case CWISE_CONSTANT:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	so[ind2] = sa;
      }
			break;
		case CWISE_SETMAX:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
		if (so[ind2] < sa)
				so[ind2] = sa;
			}
      break;
    }
  }
  else {
    switch(type) {
    case CWISE_ADDITION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sa + si[ind2];
      }
      break;
    case CWISE_SUBTRACTION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sa - si[ind2];
      }
      break;
    case CWISE_MULTIPLICATION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sa * si[ind2];
      }
      break;
    case CWISE_DIVISION:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sa / si[ind2];
      }
      break;
    case CWISE_POWER:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = __powf(si[ind2], sa);
      }
      break;;
    case CWISE_EXP:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = __expf(si[ind2]);
      }
      break;
    case CWISE_LOG:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
        int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
        int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
        so[ind1] = __logf(si[ind2]);
      }
      break;
    case CWISE_ABS:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = fabsf(si[ind2]);
      }
      break;
    case CWISE_TANH:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = tanhf(si[ind2]);
      }
      break;
    case CWISE_COSH:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
        int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
        int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
        so[ind1] = coshf(si[ind2]);
      }
      break;
    case CWISE_SQRT:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sqrtf(si[ind2]);
      }
      break;
    case CWISE_CONSTANT:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sa;
      }
      break;
		case CWISE_SIN:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = sinf(si[ind2]);
      }
      break;
    case CWISE_COS:
      for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
	int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
	int ind1 = getIndex(opm, opn, 0, 0, nrow, ncol, i, false);
	so[ind1] = cosf(si[ind2]);
      }
      break;
    }
  }
}

int cudaMatrixCWiseScalarOp(float* matrixOut, float* matrixIn, float scalar,
			    int opm, int opn, int pm, int pn, int srow,
			    int scol, int nrow, int ncol, bool inPlace,
			    CUDA_MATRIX_OPTYPE type) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.so = matrixOut;
  params.si = matrixIn;
  params.sa = scalar;
  params.opm = opm;
  params.opn = opn;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;
  params.inPlace = inPlace;
  params.type = type;

  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cWiseScalarOpMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixSumMain(struct cWiseScalarParams parms) {
  extern __shared__ float sdata[];
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float *so;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  so = parms.so;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;

  sdata[tid] = 0;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  float t, y, c = 0;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
    y = si[ind] - c;
    t = sdata[tid] + y;
    c = (t - sdata[tid]) - y;
    sdata[tid] = t;
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = 0;
    for (i = 0; i < blockDim.x; i++)
      so[blockIdx.x] += sdata[i];
  }
}

int cudaMatrixSum(float* result, float* matrixIn, int pm, int pn,
		  int srow, int scol, int nrow, int ncol) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  int sMemSize = threadsPerCta * sizeof(float);
  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, nbrCtas * sizeof(float)));
  params.so = d_block;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixSumMain<<<nbrCtas,threadsPerCta,sMemSize>>>(params);

  float* h_block = (float*)malloc(nbrCtas * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, nbrCtas * sizeof(float),
				 cudaMemcpyDeviceToHost));

  *result = 0;
  for (int i = 0; i < nbrCtas; i++)
    *result += h_block[i];

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixTraceMain(float *result, float *matrixIn, int pm,
	   int pn, int n) {
  *result = 0;
  float t, y, c = 0;
  int i = 0;
  for (; i < n; i++) {
    int ind = i * (pm + 1);
    y = matrixIn[ind] - c;
    t = *result + y;
    c = (t - *result) - y;
    *result = t;
  }
}

int cudaMatrixTrace(float* result, float* matrixIn, int pm, int pn,
		  int n) {
  float *res;
  cutilSafeCallNoSync(cudaMalloc((void**)&res, sizeof(float)));
  cudaError_t cudaStat = cudaGetLastError(); /* clear error status */

  cudaMatrixTraceMain<<<1,1>>>(res, matrixIn, pm, pn, n);

  cutilSafeCallNoSync(cudaMemcpy(result, res, sizeof(float),
				 cudaMemcpyDeviceToHost));
  cutilSafeCallNoSync(cudaFree(res));
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixMinMain(struct cWiseScalarParams parms) {
  extern __shared__ float sdata[];
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float *so;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  so = parms.so;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;

  sdata[tid] = FLT_MAX;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
    sdata[tid] = fmin(sdata[tid], si[ind]);
  }
  
  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = FLT_MAX;
    for (i = 0; i < blockDim.x; i++)
      so[blockIdx.x] = fmin(so[blockIdx.x], sdata[i]);
  }
}

int cudaMatrixMin(float* result, float* matrixIn, int pm, int pn,
		  int srow, int scol, int nrow, int ncol) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  int sMemSize = threadsPerCta * sizeof(float);
  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, nbrCtas * sizeof(float)));
  params.so = d_block;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixMinMain<<<nbrCtas,threadsPerCta,sMemSize>>>(params);

  float* h_block = (float*)malloc(nbrCtas * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, nbrCtas * sizeof(float),
				 cudaMemcpyDeviceToHost));

  *result = FLT_MAX;
  for (int i = 0; i < nbrCtas; i++)
    *result = fmin(*result, h_block[i]);

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixMaxMain(struct cWiseScalarParams parms) {
  extern __shared__ float sdata[];
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float *so;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  so = parms.so;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;

  sdata[tid] = FLT_MIN;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
    sdata[tid] = fmax(sdata[tid], si[ind]);
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = FLT_MIN;
    for (i = 0; i < blockDim.x; i++)
      so[blockIdx.x] = fmax(so[blockIdx.x], sdata[i]);
  }
}

int cudaMatrixMax(float* result, float* matrixIn, int pm, int pn,
		  int srow, int scol, int nrow, int ncol) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  int sMemSize = threadsPerCta * sizeof(float);
  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, nbrCtas * sizeof(float)));
  params.so = d_block;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixMaxMain<<<nbrCtas,threadsPerCta,sMemSize>>>(params);

  float* h_block = (float*)malloc(nbrCtas * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, nbrCtas * sizeof(float),
				 cudaMemcpyDeviceToHost));

  *result = FLT_MIN;
  for (int i = 0; i < nbrCtas; i++)
    *result = fmax(*result, h_block[i]);

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixSquaredNormMain(struct cWiseScalarParams parms) {
  extern __shared__ float sdata[];
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float *so;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  so = parms.so;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;

  sdata[tid] = 0;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  float t, y, c = 0;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
    y = si[ind] * si[ind] - c;
    t = sdata[tid] + y;
    c = (t - sdata[tid]) - y;
    sdata[tid] = t;
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = 0;
    for (i = 0; i < blockDim.x; i++)
      so[blockIdx.x] += sdata[i];
  }
}

int cudaMatrixSquaredNorm(float* result, float* matrixIn, int pm, int pn,
			  int srow, int scol, int nrow, int ncol) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;
  
  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  int sMemSize = threadsPerCta * sizeof(float);
  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, nbrCtas * sizeof(float)));
  params.so = d_block;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixSquaredNormMain<<<nbrCtas,threadsPerCta,sMemSize>>>(params);

  float* h_block = (float*)malloc(nbrCtas * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, nbrCtas * sizeof(float),
				 cudaMemcpyDeviceToHost));

  *result = 0;
  for (int i = 0; i < nbrCtas; i++)
    *result += h_block[i];

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

#define NUM_SUMS 64

__global__ void cudaMatrixPartialSumMain(struct reduxParams parms) {
  int i, j, tid, totalThreads, ctaStart;
  float *so;
  float *si;
  int pm, pn;
  bool colwise;
  bool trans;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  so = parms.so;
  si = parms.si;
  pm = parms.pm;
  pn = parms.pn;
  colwise = parms.colwise;
  trans = parms.trans;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  float y, t, c;
  if (colwise ^ trans) {
    for (i = ctaStart + tid; i < pn; i += totalThreads) {
      so[i] = 0; c = 0;
      for (j = 0; j < pm; j++) {
	y = si[i*pm+j] - c;
	t = so[i] + y;
	c = (t - so[i]) - y;
	so[i] = t;
      }
    }
  }
  else {
    for (i = ctaStart + tid; i < pm; i += totalThreads) {
      so[i] = 0; c = 0;
      for (j = 0; j < pn; j++) {
	y = si[j*pm+i] - c;
	t = so[i] + y;
	c = (t - so[i]) - y;
	so[i] = t;
      }
    }
  }
}

int cudaMatrixPartialSum(float* result, float* matrixIn, int pm, int pn,
			 bool colwise, bool trans) {
  struct reduxParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.so = result;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.colwise = colwise;
  params.trans = trans;
  
  cublasVectorSplay (((colwise ^ trans) ? pn : pm), CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixPartialSumMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixPartialNormSquaredMain(struct reduxParams parms) {
  int i, j, tid, totalThreads, ctaStart;
  float *so;
  float *si;
  int pm, pn;
  bool colwise;
  bool trans;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  so = parms.so;
  si = parms.si;
  pm = parms.pm;
  pn = parms.pn;
  colwise = parms.colwise;
  trans = parms.trans;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  float y, t, c;
  if (colwise ^ trans) {
    for (i = ctaStart + tid; i < pn; i += totalThreads) {
      so[i] = 0; c = 0;
      for (j = 0; j < pm; j++) {
	y = si[i*pm+j]*si[i*pm+j] - c;
	t = so[i] + y;
	c = (t - so[i]) - y;
	so[i] = t;
      }
    }
  }
  else {
    for (i = ctaStart + tid; i < pm; i += totalThreads) {
      so[i] = 0; c = 0;
      for (j = 0; j < pn; j++) {
	y = si[j*pm+i]*si[j*pm+i] - c;
	t = so[i] + y;
	c = (t - so[i]) - y;
	so[i] = t;
      }
    }
  }
}

int cudaMatrixPartialNormSquared(float* result, float* matrixIn, int pm, int pn,
			 bool colwise, bool trans) {
  struct reduxParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  memset (&params, 0, sizeof(params));
  params.so = result;
  params.si = matrixIn;
  params.pm = pm;
  params.pn = pn;
  params.colwise = colwise;
  params.trans = trans;
  
  cublasVectorSplay (((colwise ^ trans) ? pn : pm), CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);

  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixPartialNormSquaredMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixAsDiagonalMain(struct cWiseScalarParams parms) {
  int i, tid, totalThreads, ctaStart;
  const float *si;
  float *so;
  int opm;
  int pm, pn;
  int srow, scol;
  int nrow, ncol;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  tid = threadIdx.x;
  si = parms.si;
  so = parms.so;
  opm = parms.opm;
  pm = parms.pm;
  pn = parms.pn;
  srow = parms.srow;
  scol = parms.scol;
  nrow = parms.nrow;
  ncol = parms.ncol;
  
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < parms.n; i += totalThreads) {
    int ind2 = getIndex(pm, pn, srow, scol, nrow, ncol, i, false);
    so[i*opm + i] = si[ind2];
  }
}

int cudaMatrixAsDiagonal(float* matrixOut, float* matrixIn, int opm, int pm,
			 int pn, int srow, int scol, int nrow, int ncol) {
  struct cWiseScalarParams params;
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;

  memset (&params, 0, sizeof(params));
  params.n = nrow*ncol;
  params.so = matrixOut;
  params.si = matrixIn;
  params.opm = opm;
  params.pm = pm;
  params.pn = pn;
  params.srow = srow;
  params.scol = scol;
  params.nrow = nrow;
  params.ncol = ncol;

  cublasVectorSplay (nrow*ncol, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); /* clear error status */
  cudaMatrixAsDiagonalMain<<<nbrCtas,threadsPerCta>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixSetZeroMain(float* si, int n) {
  int i, tid, totalThreads, ctaStart;
  tid = threadIdx.x;
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < n; i += totalThreads)
    si[i] = 0;
}

int cudaMatrixSetZero(float* matrixIn, int n) {
  if(n <= 1024) {
    cudaMemset(matrixIn, 0, n*sizeof(float));
    return 1;
  }

  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  cudaStat = cudaGetLastError(); // clear error status
  
  cublasVectorSplay (n, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); // clear error status
  cudaMatrixSetZeroMain<<<nbrCtas,threadsPerCta>>>(matrixIn, n);
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixSetMain(float* si, float sa, int pm, int pn,
				  int m, int n) {
  unsigned int xIndex = blockIdx.x*BLOCK_DIM + threadIdx.x;
  unsigned int yIndex = blockIdx.y*BLOCK_DIM + threadIdx.y;
  if(xIndex < m && yIndex < n)
    si[yIndex*pm+xIndex] = sa;
  else if(xIndex < pm && yIndex < pn)
    si[yIndex*pm+xIndex] = 0;
}

int cudaMatrixSet(float* matrixIn, float scalar, int pm, int pn, int m, int n) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  
  dim3 grid((pm+BLOCK_DIM-1) / BLOCK_DIM, (pn+BLOCK_DIM-1) / BLOCK_DIM, 1);
  dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
  cudaMatrixSetMain<<<grid,threads>>>(matrixIn, scalar, pm, pn, m, n);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixCopyMain(float* so, float* si, int n) {
  int i, tid, totalThreads, ctaStart;
  tid = threadIdx.x;
  totalThreads = gridDim.x*blockDim.x;
  ctaStart = blockDim.x*blockIdx.x;
  for (i = ctaStart + tid; i < n; i += totalThreads)
    so[i] = si[i];
}

int cudaMatrixCopy(float* matrixOut, float* matrixIn, int n) {
  cudaError_t cudaStat;
  int nbrCtas;
  int elemsPerCta;
  int threadsPerCta;
  
  cudaStat = cudaGetLastError(); // clear error status
  
  cublasVectorSplay (n, CUBLAS_SAXPY_THREAD_MIN,
		     CUBLAS_SAXPY_THREAD_MAX, CUBLAS_SAXPY_CTAS_MAX,
		     &nbrCtas, &elemsPerCta, &threadsPerCta);
  
  cudaStat = cudaGetLastError(); // clear error status
  cudaMatrixCopyMain<<<nbrCtas,threadsPerCta>>>(matrixOut, matrixIn, n);
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

int cudaMatrixNumberRandomSeeds(int pm, int pn) {
  return ((pm+BLOCK_DIM-1) / BLOCK_DIM) * ((pn+BLOCK_DIM-1) / BLOCK_DIM) * BLOCK_DIM * BLOCK_DIM;
}

int cudaMatrixInitRandom(unsigned **randSeeds, int &nSeeds, int pm, int pn) {
  if (*randSeeds != NULL) {
    cudaFree(*randSeeds);
    *randSeeds = NULL;
  }

  nSeeds = cudaMatrixNumberRandomSeeds(pm, pn);

  cudaMalloc((void**)randSeeds, nSeeds * 4 * sizeof(unsigned));
  if (*randSeeds == NULL)
    return 0;
  
  unsigned *temp = (unsigned *) malloc(nSeeds * 4 * sizeof(unsigned));
  
  for (int i = 0; i < nSeeds * 4; i++)
    temp[i] = rand();

  cudaMemcpy(*randSeeds, temp, nSeeds * 4 * sizeof(unsigned), cudaMemcpyHostToDevice);  
  free(temp);

  return 1;
}

__device__ unsigned TausStep(unsigned &z, int S1, int S2, int S3, unsigned M)  
{  
  unsigned b=(((z << S1) ^ z) >> S2);  
  return z = (((z & M) << S3) ^ b);  
}  

__device__ unsigned LCGStep(unsigned &z, unsigned A, unsigned C) {
  return z = (A * z + C);
}

__device__ float hybridTaus(unsigned &z1, unsigned &z2, unsigned &z3, unsigned &z4) {
  // Combined period is lcm(p1,p2,p3,p4)~ 2^121  

  return 2.3283064365387e-10f * (              // Periods  
      TausStep(z1, 13, 19, 12, 4294967294UL) ^  // p1=2^31-1  
      TausStep(z2, 2, 25, 4, 4294967288UL) ^    // p2=2^30-1  
      TausStep(z3, 3, 11, 17, 4294967280UL) ^   // p3=2^28-1  
      LCGStep(z4, 1664525, 1013904223UL)        // p4=2^32  
      );  
}

__global__ void cudaMatrixFillRandomUniformMain(float* si, unsigned *seeds , int pm, int pn,
				  int m, int n) {
  unsigned int xIndex = blockIdx.x*BLOCK_DIM + threadIdx.x;
  unsigned int yIndex = blockIdx.y*BLOCK_DIM + threadIdx.y;
  unsigned int nThreads = pm * pn;
  unsigned int idx = yIndex * pm + xIndex;

  if(xIndex < m && yIndex < n)
    si[idx] = hybridTaus(
        seeds[idx],
        seeds[idx + nThreads],
        seeds[idx + nThreads * 2],
        seeds[idx + nThreads * 3]);
  else if(xIndex < pm && yIndex < pn)
    si[idx] = 0.0f;
}

int cudaMatrixFillRandomUniform(float *matrixIn, unsigned *randSeeds, int pm, int pn, int m, int n) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  
  dim3 grid((pm+BLOCK_DIM-1) / BLOCK_DIM, (pn+BLOCK_DIM-1) / BLOCK_DIM, 1);
  dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
  cudaMatrixFillRandomUniformMain<<<grid,threads>>>(matrixIn, randSeeds, pm, pn, m, n);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__device__ float boxMuller(unsigned &z1, unsigned &z2, unsigned &z3, unsigned &z4) {
  float u0 = hybridTaus(z1, z2, z3, z4);
  float u1 = hybridTaus(z1, z2, z3, z4);
  float r = sqrt(-2.0f * log(u0 < 1e-6 ? 1e-6 : u0));
  float theta = 2.0f * 3.14159f * u1;
  
  return r * sin(theta);
}

__global__ void cudaMatrixFillRandomGaussianMain(float* si, unsigned *seeds , int pm, int pn,
				  int m, int n) {
  unsigned int xIndex = blockIdx.x*BLOCK_DIM + threadIdx.x;
  unsigned int yIndex = blockIdx.y*BLOCK_DIM + threadIdx.y;
  unsigned int nThreads = pm * pn;
  unsigned int idx = yIndex * pm + xIndex;

  if(xIndex < m && yIndex < n)
    si[idx] = boxMuller(
        seeds[idx],
        seeds[idx + nThreads],
        seeds[idx + nThreads * 2],
        seeds[idx + nThreads * 3]);
  else if(xIndex < pm && yIndex < pn)
    si[idx] = 0.0f;
}

int cudaMatrixFillRandomGaussian(float *matrixIn, unsigned *randSeeds, int pm, int pn, int m, int n) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  
  dim3 grid((pm+BLOCK_DIM-1) / BLOCK_DIM, (pn+BLOCK_DIM-1) / BLOCK_DIM, 1);
  dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
  cudaMatrixFillRandomGaussianMain<<<grid,threads>>>(matrixIn, randSeeds, pm, pn, m, n);

  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}
