#include <cuda.h>
#include <cuda_runtime.h>
//#include "../base/logger.h"

#include "cudaMemHandler.h"

#define USE_OLD_MEM_HANDLER 0

// Decouple from SDN
#include <iostream>
#include <assert.h>

#ifndef SDN_ASSERT
#define SDN_ASSERT(x) assert(x)
#endif

#ifndef SDN_FATAL
#define SDN_FATAL(x) std::cerr << x << "\n"; assert(0);
#endif

std::map<int, std::list<float*> > sdn::CudaMemHandler::sizeToPtr;
std::map<float*, int> sdn::CudaMemHandler::ptrToSize;
std::map<float*, int> sdn::CudaMemHandler::freeMap;

float* sdn::CudaMemHandler::alloc(int size) {
  if(size == 0) {
    SDN_FATAL("CudaMemHandler::alloc is called with size = 0");
  }

  size = ((size - 1) | 7) + 1;
  if(USE_OLD_MEM_HANDLER || size >= ALLOC_PAGE_SIZE - 8)
    return allocLarge(size);

  for(std::map<float*, int>::reverse_iterator rit = freeMap.rbegin();
      rit != freeMap.rend(); ++rit) {
    if(rit->second >= size) {
      float* retPtr = (float*)((char*)(rit->first) + rit->second - size);
      if(rit->second == size)
        freeMap.erase(--(rit.base()));
      else
        rit->second -= size;

      SDN_ASSERT(ptrToSize.find(retPtr) == ptrToSize.end());
      ptrToSize[retPtr] = size;
      return retPtr;
    }
  }

  float* ptr;
  if(cudaMalloc((void**)&ptr, ALLOC_PAGE_SIZE)) {
    clear();
    if(cudaMalloc((void**)&ptr, ALLOC_PAGE_SIZE)) {
      SDN_FATAL("CudaMemHandler::alloc cudaMalloc failed!");
    }
  }

  SDN_ASSERT(freeMap.find(ptr) == freeMap.end());
  freeMap[ptr] = (ALLOC_PAGE_SIZE - 8) - size;

  float *retPtr = (float*)((char*)(ptr) + (ALLOC_PAGE_SIZE - 8) - size);
  SDN_ASSERT(ptrToSize.find(retPtr) == ptrToSize.end());
  ptrToSize[retPtr] = size;
  return retPtr;
}

void sdn::CudaMemHandler::free(float* ptr) {
  SDN_ASSERT(ptrToSize.find(ptr) != ptrToSize.end());

  int size = ptrToSize[ptr];
  if(USE_OLD_MEM_HANDLER || size >= ALLOC_PAGE_SIZE - 8) {
    sizeToPtr[size].push_back(ptr);
    ptrToSize.erase(ptr);
  }
  else {
    insertFreeChunk(ptr, size);
    ptrToSize.erase(ptr);
  }
}

float* sdn::CudaMemHandler::allocLarge(int size) {
  std::map<int, std::list<float*> >::iterator it = sizeToPtr.find(size);
  if(it == sizeToPtr.end() || it->second.size() == 0) {
    float* ptr;
    if(cudaMalloc((void**)&ptr, size)) {
      clear();
      if(cudaMalloc((void**)&ptr, size)) {
        SDN_FATAL("CudaMemHandler::allocLarge cudaMalloc failed!");
      }
    }
    ptrToSize[ptr] = size;
    return ptr;
  }
  
  float* ptr = it->second.front();
  it->second.pop_front();
  ptrToSize[ptr] = size;
  return ptr;
}

void sdn::CudaMemHandler::insertFreeChunk(float* ptr, int size) {
  SDN_ASSERT(freeMap.find(ptr) == freeMap.end());

  std::map<float*, int>::iterator it, nextIt;
  nextIt = freeMap.insert(std::make_pair(ptr, size)).first;
  it = nextIt++;
  if(nextIt != freeMap.end() && 
     (float*)((char*)(it->first)+it->second) == nextIt->first) {
    it->second += nextIt->second;
    freeMap.erase(nextIt);
    SDN_ASSERT(it->second <= ALLOC_PAGE_SIZE);
  }
  
  if(it != freeMap.begin()) {
    nextIt = it--;
    if((float*)((char*)(it->first)+it->second) == nextIt->first) {
      it->second += nextIt->second;
      freeMap.erase(nextIt);
      SDN_ASSERT(it->second <= ALLOC_PAGE_SIZE);
    }
  }
}

void sdn::CudaMemHandler::clear() {
  for(std::map<int, std::list<float*> >::iterator it = sizeToPtr.begin();
      it != sizeToPtr.end(); ++it) {
    for(std::list<float*>::iterator lit = it->second.begin();
	lit != it->second.end(); ++lit) {
      cudaFree(*lit);
      ptrToSize.erase(*lit);
    }
  }
  sizeToPtr.clear();

  for(std::map<float*, int>::iterator it = freeMap.begin();
      it != freeMap.end();) {
    if(it->second == ALLOC_PAGE_SIZE) {
      cudaFree(it->first);
      freeMap.erase(it++);
    }
    else
      ++it;
  }
}
