#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <algorithm>
#include <cuda_runtime.h>
#include <cublas.h>
// #include "../base/logger.h"
#include "cudaMatrixUtils.h"
#include "sparseCudaMatrixUtils.h"
#include "cudaMemHandler.h"
#include "Eigen/Core"

#include "sparseCudaMatrix.h"
#include "cudaVector.h"
// #include <culapackdevice.h>

#define VERIFY 1

// Decouple from SDN
#ifndef SDN_ASSERT
#define SDN_ASSERT(x) assert(x)
#endif

#ifndef SDN_FATAL
#define SDN_FATAL(x) std::cerr << x << "\n"; assert(0);
#endif 

namespace sdn {
  void verifyCanCwise(const sdn::SparseCudaMatrix& lhs, const sdn::SparseCudaMatrix& rhs) {
    SDN_ASSERT(lhs.m_ == rhs.m_ && lhs.n_ == rhs.n_ &&
	       lhs.rowEntries_ == rhs.rowEntries_ &&
	       lhs.colEntries_ == rhs.colEntries_ &&
	       lhs.numEntries_ == rhs.numEntries_);
    lhs.updateMatrixGPU();
    rhs.updateMatrixGPU();
    int notEqual = 0;
    sparseCudaMatrixVerifyIndexEntries(&notEqual, lhs.rowIndexGPU_, rhs.rowIndexGPU_,
				       lhs.rowEntries_, toBlockSize(lhs.rowEntries_+1));
    SDN_ASSERT(notEqual == 0);
    
    sparseCudaMatrixVerifyEntries(&notEqual, lhs.rowMatrixGPU_, rhs.rowMatrixGPU_,
				  lhs.numEntries_, lhs.pNumEntries_);
    SDN_ASSERT(notEqual == 0);
  }
}
  
inline bool checkSuccess(cublasStatus status) {
  return status == CUBLAS_STATUS_SUCCESS;
}

int sdn::toBlockSize(int d) {
  return ((d - 1) | (BLOCK_DIM - 1)) + 1;
}

sdn::SparseCudaMatrix::SparseCudaMatrix() {
  initializeEmpty();
}

void sdn::SparseCudaMatrix::initializeEmpty() {
  m_ = 0;
  n_ = 0;
  rowEntries_ = 0;
  colEntries_ = 0;
  numEntries_ = 0;
  pNumEntries_ = 0;

  rowIndexGPU_ = NULL;
  colIndexGPU_ = NULL;
  rowMatrixGPU_ = NULL;
  colMatrixGPU_ = NULL;
  
  rowIndexCPU_ = NULL;
  colIndexCPU_ = NULL;
  rowMatrixCPU_ = NULL;
  colMatrixCPU_ = NULL;

  dirtyCPU_ = true;
  dirtyGPU_ = true;
}

void sdn::SparseCudaMatrix::initialize(int m, int n, int numEntries, 
				       std::vector<int> rows, std::vector<int> cols,
				       std::vector<std::vector<entry> > rowEntries,
				       std::vector<std::vector<entry> > colEntries) {
  m_ = m;
  n_ = n;
  rowEntries_ = rows.size();
  colEntries_ = cols.size();
  numEntries_ = numEntries;
  pNumEntries_ = toBlockSize(numEntries);

  rowIndexGPU_ = NULL;
  colIndexGPU_ = NULL;
  rowMatrixGPU_ = NULL;
  colMatrixGPU_ = NULL;

  rowIndexCPU_ = (indexEntry*)malloc(toBlockSize(rowEntries_+1) * sizeof(indexEntry));
  colIndexCPU_ = (indexEntry*)malloc(toBlockSize(colEntries_+1) * sizeof(indexEntry));
  rowMatrixCPU_ = (entry*)malloc(pNumEntries_ * sizeof(entry));
  colMatrixCPU_ = (entry*)malloc(pNumEntries_ * sizeof(entry));

  int curNum = 0;
  for(unsigned int i = 0; i < rows.size(); i++) {
    rowIndexCPU_[i].index = rows[i];
    rowIndexCPU_[i].value = curNum;
    std::vector<entry> thisRow = rowEntries[i];
    for(unsigned int j = 0; j < thisRow.size(); j++) {
      rowMatrixCPU_[curNum + j] = thisRow[j];
    }
    curNum += thisRow.size();
  }
  rowIndexCPU_[rowEntries_].value = curNum;

  curNum = 0;
  for(unsigned int i = 0; i < cols.size(); i++) {
    colIndexCPU_[i].index = cols[i];
    colIndexCPU_[i].value = curNum;
    std::vector<entry> thisCol = colEntries[i];
    for(unsigned int j = 0; j < thisCol.size(); j++) {
      colMatrixCPU_[curNum + j] = thisCol[j];
    }
    curNum += thisCol.size();
  }
  colIndexCPU_[colEntries_].value = curNum;

  dirtyCPU_ = false;
  dirtyGPU_ = true;
}

sdn::SparseCudaMatrix::SparseCudaMatrix(int m, int n, int numEntries,
					std::vector<int> rows, std::vector<int> cols,
					std::vector<std::vector<entry> > rowEntries,
					std::vector<std::vector<entry> > colEntries) {
  initialize(m, n, numEntries, rows, cols, rowEntries, colEntries);
}

bool rowSortFn(sparseEntry a, sparseEntry b) {
  return (a.row != b.row) ? (a.row < b.row) : (a.col < b.col);
}

bool colSortFn(sparseEntry a, sparseEntry b) {
  return (a.col != b.col) ? (a.col < b.col) : (a.row < b.row);
}

sdn::SparseCudaMatrix::SparseCudaMatrix(int m, int n, std::vector<sparseEntry> entries) {
  std::vector<int> rowIndex;
  std::vector<int> colIndex;
  std::vector<std::vector<entry> > rows;
  std::vector<std::vector<entry> > cols;
  int numEntries = entries.size();

  sort(entries.begin(), entries.end(), rowSortFn);
  int row = -1;
  int index = -1;
  for(unsigned int i = 0; i < entries.size(); i++) {
    sparseEntry next = entries[i];
    if(next.row != row) {
      rowIndex.push_back(next.row);
      row = next.row;
      std::vector<entry> nextRow;
      index++;
      rows.push_back(nextRow);
    }
    entry e;
    e.index = next.col;
    e.value = next.val;
    rows[index].push_back(e);
  }

  sort(entries.begin(), entries.end(), colSortFn);
  int col = -1;
  index = -1;
  for(unsigned int i = 0; i < entries.size(); i++) {
    sparseEntry next = entries[i];
    if(next.col != col) {
      colIndex.push_back(next.col);
      col = next.col;
      std::vector<entry> nextCol;
      index++;
      cols.push_back(nextCol);
    }
    entry e;
    e.index = next.row;
    e.value = next.val;
    cols[index].push_back(e);
  }

  initialize(m, n, numEntries, rowIndex, colIndex, rows, cols);
}

sdn::SparseCudaMatrix::SparseCudaMatrix(const SparseCudaMatrix& other) {
  initializeEmpty();
  *this = other;
}

sdn::SparseCudaMatrix::SparseCudaMatrix(const CudaMatrix& other) {
  int m = other.rows(), n = other.cols(), numEntries = 0;
  std::vector<int> rows, cols;
  std::vector<std::vector<entry> > rowEntries, colEntries;
  for (int i = 0; i < m; i++) {
    std::vector<entry> thisRow;
    bool rowEmpty = true;
    for (int j = 0; j < n; j++) {
      if (other(i, j) != 0.0) {
        numEntries++;
        rowEmpty = false;
        entry newEntry;
        newEntry.index = j;
        newEntry.value = other(i, j);
        thisRow.push_back(newEntry);
      }
    }
    if (!rowEmpty) {
      rows.push_back(i);
      rowEntries.push_back(thisRow);
    }
  }
  for (int j = 0; j < n; j++) {
    std::vector<entry> thisCol;
    bool colEmpty = true;
    for (int i = 0; i < m; i++) {
      if (other(i, j) != 0.0) {
        colEmpty = false;
        entry newEntry;
        newEntry.index = i;
        newEntry.value = other(i, j);
        thisCol.push_back(newEntry);
      }
    }
    if (!colEmpty) {
      cols.push_back(j);
      colEntries.push_back(thisCol);
    }
  }
  initialize(m, n, numEntries, rows, cols, rowEntries, colEntries);
}

sdn::SparseCudaMatrix::~SparseCudaMatrix() {
  if(rowMatrixCPU_ != NULL) {
    free(rowMatrixCPU_);
    free(colMatrixCPU_);
    free(rowIndexCPU_);
    free(colIndexCPU_);
  }
  if(rowMatrixGPU_ != NULL) {
    CudaMemHandler::free((float*)rowMatrixGPU_);
    CudaMemHandler::free((float*)colMatrixGPU_);
    CudaMemHandler::free((float*)rowIndexGPU_);
    CudaMemHandler::free((float*)colIndexGPU_);
  }
}

void sdn::SparseCudaMatrix::set(int i, int j, float f) {
  SDN_ASSERT(i >= 0 && i < rows() && j >= 0 && j < cols());

  updateMatrixCPU();
  dirtyGPU_ = true;
  int min = 0, max = rowEntries_ - 1;
  int cur = (min + max) / 2;
  while (min <= max) {
    if(rowIndexCPU_[cur].index > i)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  if (rowIndexCPU_[cur].index != i)
    SDN_FATAL("Cannot access unstored elements of SparseCudaMatrix!");

  min = rowIndexCPU_[cur].value;
  max = rowIndexCPU_[cur+1].value - 1;
  cur = (min + max) / 2;
  while (min <= max) {
    if(rowMatrixCPU_[cur].index > j)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  if (rowMatrixCPU_[cur].index != j)
    SDN_FATAL("Cannot access unstored elements of SparseCudaMatrix!");
  rowMatrixCPU_[cur].value = f;
  
  min = 0, max = colEntries_ - 1;
  cur = (min + max) / 2;
  while (min <= max) {
    if(colIndexCPU_[cur].index > j)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  
  min = colIndexCPU_[cur].value;
  max = colIndexCPU_[cur+1].value - 1;
  cur = (min + max) / 2;
  while (min <= max) {
    if(colMatrixCPU_[cur].index > i)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  colMatrixCPU_[cur].value = f;
}

const float sdn::SparseCudaMatrix::get(int i, int j) const {
  SDN_ASSERT(i >= 0 && i < rows() && j >= 0 && j < cols());

  updateMatrixCPU();
  int min = 0, max = rowEntries_ - 1;
  int cur = (min + max) / 2;
  while (min <= max) {
    if(rowIndexCPU_[cur].index > i)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  if (rowIndexCPU_[cur].index != i)
    return 0.0;

  min = rowIndexCPU_[cur].value;
  max = rowIndexCPU_[cur+1].value - 1;
  cur = (min + max) / 2;
  while (min <= max) {
    if(rowMatrixCPU_[cur].index > j)
      max = cur - 1;
    else
      min = cur + 1;
    cur = (max + min) / 2;
  }
  if (rowMatrixCPU_[cur].index != j)
    return 0.0;
  return rowMatrixCPU_[cur].value;
}

void sdn::SparseCudaMatrix::set(int i, float f) {
  set(i%rows(), (int)(i/rows()), f);
}

const float sdn::SparseCudaMatrix::get(int i) const {
  return get(i%rows(), (int)(i/rows()));
}

const float sdn::SparseCudaMatrix::operator()(int i, int j) const {
  return get(i, j);
}

const float sdn::SparseCudaMatrix::operator[](int i) const {
  return get(i%rows(), (int)(i/rows()));
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise sdn::SparseCudaMatrix::cwise() {
  return SparseCudaMatrixCwise(*this);
}

const sdn::SparseCudaMatrix::SparseCudaMatrixCwise sdn::SparseCudaMatrix::cwise() const {
  return SparseCudaMatrixCwise(*this);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator=(const SparseCudaMatrix& rhs) {
  if(this == &rhs)
    return *this;

  if(toBlockSize(rowEntries_+1) != toBlockSize(rhs.rowEntries_+1) ||
     toBlockSize(colEntries_+1) != toBlockSize(rhs.colEntries_+1) ||
     pNumEntries_ != rhs.pNumEntries_) {
    // if dimensions are wrong, free CPU and GPU matrices
    if(rowIndexCPU_ != NULL) {
      free(rowIndexCPU_);
      free(colIndexCPU_);
      free(rowMatrixCPU_);
      free(colMatrixCPU_);
      rowIndexCPU_ = NULL;
      colIndexCPU_ = NULL;
      rowMatrixCPU_ = NULL;
      colMatrixCPU_ = NULL;
    }
    if(rowIndexGPU_ != NULL) {
      CudaMemHandler::free((float*)rowIndexGPU_);
      CudaMemHandler::free((float*)colIndexGPU_);
      CudaMemHandler::free((float*)rowMatrixGPU_);
      CudaMemHandler::free((float*)colMatrixGPU_);
      rowIndexGPU_ = NULL;
      colIndexGPU_ = NULL;
      rowMatrixGPU_ = NULL;
      colMatrixGPU_ = NULL;
    }
  }

  m_ = rhs.m_;
  n_ = rhs.n_;
  rowEntries_ = rhs.rowEntries_;
  colEntries_ = rhs.colEntries_;
  numEntries_ = rhs.numEntries_;
  pNumEntries_ = rhs.pNumEntries_;
  // copy on GPU
  rhs.updateMatrixGPU();
  if(rowMatrixGPU_ == NULL)
    allocateMatrixGPU();
  sparseCudaMatrixEntryCopy(rowMatrixGPU_, rhs.rowMatrixGPU_, pNumEntries_);
  sparseCudaMatrixEntryCopy(colMatrixGPU_, rhs.colMatrixGPU_, pNumEntries_);
  sparseCudaMatrixIndexCopy(rowIndexGPU_, rhs.rowIndexGPU_, toBlockSize(rowEntries_+1));
  sparseCudaMatrixIndexCopy(colIndexGPU_, rhs.colIndexGPU_, toBlockSize(colEntries_+1));
  dirtyCPU_ = true;
  dirtyGPU_ = false;
 
  return *this;
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator=(const CudaMatrix& rhs) {
  int m = rhs.rows(), n = rhs.cols(), numEntries = 0;
  std::vector<int> rows, cols;
  std::vector<std::vector<entry> > rowEntries, colEntries;
  for (int i = 0; i < m; i++) {
    std::vector<entry> thisRow;
    bool rowEmpty = true;
    for (int j = 0; j < n; j++) {
      if (rhs(i, j) != 0.0) {
        numEntries++;
        rowEmpty = false;
        entry newEntry;
        newEntry.index = j;
        newEntry.value = rhs(i, j);
        thisRow.push_back(newEntry);
      }
    }
    if (!rowEmpty) {
      rows.push_back(i);
      rowEntries.push_back(thisRow);
    }
  }
  for (int j = 0; j < n; j++) {
    std::vector<entry> thisCol;
    bool colEmpty = true;
    for (int i = 0; i < m; i++) {
      if (rhs(i, j) != 0) {
        colEmpty = false;
        entry newEntry;
        newEntry.index = i;
        newEntry.value = rhs(i, j);
        thisCol.push_back(newEntry);
      }
    }
    if (!colEmpty) {
      cols.push_back(j);
      colEntries.push_back(thisCol);
    }
  }
  *this = SparseCudaMatrix(m, n, numEntries, rows, cols,
    rowEntries, colEntries);
  return *this;
}

bool sdn::SparseCudaMatrix::operator==(const SparseCudaMatrix& rhs) const {
  if(m_ != rhs.m_ || n_ != rhs.n_ || rowEntries_ != rhs.rowEntries_ ||
     colEntries_ != rhs.colEntries_ || numEntries_ != rhs.numEntries_)
    return false;
  updateMatrixGPU();
  rhs.updateMatrixGPU();
  int notEqual;
  sparseCudaMatrixVerifyIndexEntries(&notEqual, rowIndexGPU_, rhs.rowIndexGPU_,
				     rowEntries_, toBlockSize(rowEntries_+1));
  if(notEqual == 1)
    return false;

  sparseCudaMatrixTestEquality(&notEqual, rowMatrixGPU_, rhs.rowMatrixGPU_,
			       numEntries_, pNumEntries_);
  return notEqual == 0;
}

bool sdn::SparseCudaMatrix::operator!=(const SparseCudaMatrix& rhs) const {
  return !(*this == rhs);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator*(const SparseCudaMatrix& rhs) const {
  SDN_ASSERT(cols() == rhs.rows());

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  // multiply on GPU
  int sr = rowEntries_;
  int sc = rhs.colEntries_;
  int numEntries = sr * sc;
  int pNumEntries = toBlockSize(numEntries);
  indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sr+1)
							       * sizeof(indexEntry));
  indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sc+1)
							       * sizeof(indexEntry));
  entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
  entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
  sparseCudaMatrixMultiply(rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU,
			   rowIndexGPU_, rowMatrixGPU_, rhs.colIndexGPU_,
			   rhs.colMatrixGPU_, sr, sc);
  return SparseCudaMatrix(rows(), rhs.cols(), sr, sc, numEntries, pNumEntries,
			  rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU, false);
}

sdn::SparseCudaMatrix&
sdn::SparseCudaMatrix::cWiseOpEq(const SparseCudaMatrix& rhs, CUDA_MATRIX_OPTYPE op) {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());

  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();
  if(VERIFY)
    verifyCanCwise(*this, rhs);

  sparseCudaMatrixCWiseOp(rowMatrixGPU_, rowMatrixGPU_, rhs.rowMatrixGPU_,
			  numEntries_, pNumEntries_, op);
  sparseCudaMatrixCWiseOp(colMatrixGPU_, colMatrixGPU_, rhs.colMatrixGPU_,
			  numEntries_, pNumEntries_, op);
  return *this;
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::cWiseOp(const SparseCudaMatrix& rhs, CUDA_MATRIX_OPTYPE op) const {
  SDN_ASSERT(rows() == rhs.rows() && cols() == rhs.cols());
  // synchronize GPU matrices
  updateMatrixGPU();
  rhs.updateMatrixGPU();

  if(VERIFY)
    verifyCanCwise(*this, rhs);

  indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(rowEntries_+1)
							       * sizeof(indexEntry));
  indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(colEntries_+1)
							       * sizeof(indexEntry));
  entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));

  sparseCudaMatrixIndexCopy(rowIndexGPU, rowIndexGPU_, toBlockSize(rowEntries_+1));
  sparseCudaMatrixIndexCopy(colIndexGPU, colIndexGPU_, toBlockSize(colEntries_+1));
  sparseCudaMatrixCWiseOp(rowMatrixGPU, rowMatrixGPU_, rhs.rowMatrixGPU_,
			  numEntries_, pNumEntries_, op);
  sparseCudaMatrixCWiseOp(colMatrixGPU, colMatrixGPU_, rhs.colMatrixGPU_,
			  numEntries_, pNumEntries_, op);
  return SparseCudaMatrix(rows(), cols(), rowEntries_, colEntries_, numEntries_,
			  pNumEntries_, rowIndexGPU, rowMatrixGPU, colIndexGPU,
			  colMatrixGPU, false);
}

sdn::SparseCudaMatrix&
sdn::SparseCudaMatrix::cWiseScalarOpEq(float rhs, CUDA_MATRIX_OPTYPE op) {
  // synchronize GPU matrices
  updateMatrixGPU();
  sparseCudaMatrixCWiseScalarOp(rowMatrixGPU_, rowMatrixGPU_, rhs,
				numEntries_, pNumEntries_, op);
  sparseCudaMatrixCWiseScalarOp(colMatrixGPU_, colMatrixGPU_, rhs,
				numEntries_, pNumEntries_, op);
  return *this;
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::cWiseScalarOp(float rhs, CUDA_MATRIX_OPTYPE op) const {
  // synchronize GPU matrices
  updateMatrixGPU();

  indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(rowEntries_+1)
							       * sizeof(indexEntry));
  indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(colEntries_+1)
							       * sizeof(indexEntry));
  entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));

  sparseCudaMatrixIndexCopy(rowIndexGPU, rowIndexGPU_, toBlockSize(rowEntries_+1));
  sparseCudaMatrixIndexCopy(colIndexGPU, colIndexGPU_, toBlockSize(colEntries_+1));
  sparseCudaMatrixCWiseScalarOp(rowMatrixGPU, rowMatrixGPU_, rhs,
				numEntries_, pNumEntries_, op);
  sparseCudaMatrixCWiseScalarOp(colMatrixGPU, colMatrixGPU_, rhs,
				numEntries_, pNumEntries_, op);
  return SparseCudaMatrix(rows(), cols(), rowEntries_, colEntries_, numEntries_,
			  pNumEntries_, rowIndexGPU, rowMatrixGPU, colIndexGPU,
			  colMatrixGPU, false);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator+(const SparseCudaMatrix& rhs) const {
  return cWiseOp(rhs, CWISE_ADDITION);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator-(const SparseCudaMatrix& rhs) const {
  return cWiseOp(rhs, CWISE_SUBTRACTION);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator+=(const SparseCudaMatrix& rhs) {
  return cWiseOpEq(rhs, CWISE_ADDITION);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator-=(const SparseCudaMatrix& rhs) {
  return cWiseOpEq(rhs, CWISE_SUBTRACTION);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator*(float scalar) const {
  return cWiseScalarOp(scalar, CWISE_MULTIPLICATION);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator/(float scalar) const {
  SDN_ASSERT(scalar != 0);

  return operator*(1 / scalar);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator*=(float scalar) {
  return cWiseScalarOpEq(scalar, CWISE_MULTIPLICATION);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::operator/=(float scalar) {
  SDN_ASSERT(scalar != 0);

  return operator*=(1 / scalar);
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::transpose() const {
  // synchronize GPU matrix
  updateMatrixGPU();
  indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(rowEntries_+1)
						   * sizeof(indexEntry));
  indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(colEntries_+1)
						    * sizeof(indexEntry));
  entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  sparseCudaMatrixEntryCopy(colMatrixGPU, rowMatrixGPU_, pNumEntries_);
  sparseCudaMatrixEntryCopy(rowMatrixGPU, colMatrixGPU_, pNumEntries_);
  sparseCudaMatrixIndexCopy(colIndexGPU, rowIndexGPU_, toBlockSize(rowEntries_+1));
  sparseCudaMatrixIndexCopy(rowIndexGPU, colIndexGPU_, toBlockSize(colEntries_+1));
  return SparseCudaMatrix(cols(), rows(), colEntries_, rowEntries_,
			  numEntries_, pNumEntries_, rowIndexGPU,
			  rowMatrixGPU, colIndexGPU, colMatrixGPU, false);
}

sdn::SparseCudaMatrix& sdn::SparseCudaMatrix::transposeEquals() {
  indexEntry* temp = rowIndexCPU_;
  rowIndexCPU_ = colIndexCPU_;
  colIndexCPU_ = temp;
  temp = rowIndexGPU_;
  rowIndexGPU_ = colIndexGPU_;
  colIndexGPU_ = temp;

  entry* temp2 = rowMatrixCPU_;
  rowMatrixCPU_ = colMatrixCPU_;
  colMatrixCPU_ = temp2;
  temp2 = rowMatrixGPU_;
  rowMatrixGPU_ = colMatrixGPU_;
  colMatrixGPU_ = temp2;

  return *this;
}

sdn::SparseCudaMatrix sdn::SparseCudaMatrix::operator-() const {
  return 0.0 - cwise();
}

float sdn::SparseCudaMatrix::sum() const {
  updateMatrixGPU();
  float result;
  sparseCudaMatrixSum(&result, rowMatrixGPU_, numEntries_, pNumEntries_);
  return result;
}

float sdn::SparseCudaMatrix::minCoeff() const {
  updateMatrixGPU();
  float result;
  sparseCudaMatrixMin(&result, rowMatrixGPU_, numEntries_, pNumEntries_);
  return fmin(result, 0.0);
}

float sdn::SparseCudaMatrix::maxCoeff() const {
  updateMatrixGPU();
  float result;
  sparseCudaMatrixMax(&result, rowMatrixGPU_, numEntries_, pNumEntries_);
  return fmax(result, 0.0);
}

float sdn::SparseCudaMatrix::norm() const {
  return sqrt(squaredNorm());
}

float sdn::SparseCudaMatrix::squaredNorm() const {
  updateMatrixGPU();
  float result;
  sparseCudaMatrixSquaredNorm(&result, rowMatrixGPU_, numEntries_, pNumEntries_);
  return result;
}

sdn::SparseCudaMatrix::SparseCudaMatrixRedux sdn::SparseCudaMatrix::rowwise() const {
  return SparseCudaMatrixRedux(*this, false);
}

void sdn::SparseCudaMatrix::transposeThis() {
  int itmp;
  indexEntry* ieptmp;
  entry* eptmp;
  itmp = m_; m_ = n_; n_ = itmp;
  itmp = rowEntries_; rowEntries_ = colEntries_; colEntries_= itmp;
  ieptmp = rowIndexCPU_; rowIndexCPU_ = colIndexCPU_; colIndexCPU_ = ieptmp;
  eptmp = rowMatrixCPU_; rowMatrixCPU_ = colMatrixCPU_; colMatrixCPU_ = eptmp;
  ieptmp = rowIndexGPU_; rowIndexGPU_ = colIndexGPU_; colIndexGPU_ = ieptmp;
  eptmp = rowMatrixGPU_; rowMatrixGPU_ = colMatrixGPU_; colMatrixGPU_ = eptmp;
}

int sdn::SparseCudaMatrix::rows() const {
  return m_;
}

int sdn::SparseCudaMatrix::cols() const {
  return n_;
}

int sdn::SparseCudaMatrix::size() const {
  return m_*n_;
}

sdn::SparseCudaMatrix::SparseCudaMatrix(int m, int n, int sr, int sc,
					int numEntries, int pNumEntries,
					indexEntry* ri, entry* rm,
					indexEntry* ci, entry* cm, bool onCPU)
  : m_(m), n_(n), rowEntries_(sr), colEntries_(sc), numEntries_(numEntries), pNumEntries_(pNumEntries) {
  if(onCPU) {
    rowIndexCPU_ = ri;
    colIndexCPU_ = ci;
    rowMatrixCPU_ = rm;
    colMatrixCPU_ = cm;
    rowIndexGPU_ = NULL;
    colIndexGPU_ = NULL;
    rowMatrixGPU_ = NULL;
    colMatrixGPU_ = NULL;
    dirtyCPU_ = false;
    dirtyGPU_ = true;
  }
  else {
    rowIndexCPU_ = NULL;
    colIndexCPU_ = NULL;
    rowMatrixCPU_ = NULL;
    colMatrixCPU_ = NULL;
    rowIndexGPU_ = ri;
    colIndexGPU_ = ci;
    rowMatrixGPU_ = rm;
    colMatrixGPU_ = cm;
    dirtyCPU_ = true;
    dirtyGPU_ = false;
  }
}

int sdn::SparseCudaMatrix::rowEntries() const {
  return rowEntries_;
}

int sdn::SparseCudaMatrix::colEntries() const {
  return colEntries_;
}

int sdn::SparseCudaMatrix::numEntries() const {
  return numEntries_;
}

float sdn::SparseCudaMatrix::density() const {
  int totalSize = m_ * n_;
  if (totalSize == 0)
    return 1.0;
  return (float)numEntries_ / (float)totalSize;
}

entry* sdn::SparseCudaMatrix::getColMatrixEntries() {
  dirtyCPU_ = true; // force the update
  updateMatrixCPU();
  return colMatrixCPU_;
}

indexEntry* sdn::SparseCudaMatrix::getColMatrixIndex() {
  dirtyCPU_ = true; // force the update
  updateMatrixCPU();
  return colIndexCPU_;
}

entry* sdn::SparseCudaMatrix::getRowMatrixEntries() {
  dirtyCPU_ = true;
  updateMatrixCPU();
  return rowMatrixCPU_;
}

indexEntry* sdn::SparseCudaMatrix::getRowMatrixIndex() {
  dirtyCPU_ = true;
  updateMatrixCPU();
  return rowIndexCPU_;
}

void sdn::SparseCudaMatrix::checkIndexConsistency() {
  updateMatrixCPU();
  bool checkPassed = true;
  for (int i = 0; i < colEntries_-1; i++) {
    if (colIndexCPU_[i].value > colIndexCPU_[i+1].value) {
      std::cout << "Consistency Check Failed at Column " << i << " : " << colIndexCPU_[i].value  << "-" << colIndexCPU_[i+1].value << "-" << colIndexCPU_[i+2].value << "\n";
      checkPassed = false;
    }      
  }
  for (int i = 0; i < rowEntries_-1; i++) {
    if (rowIndexCPU_[i].value > rowIndexCPU_[i+1].value) {
      std::cout << "Consistency Check Failed at Row " << i << " : " << rowIndexCPU_[i].value  << "-" << rowIndexCPU_[i+1].value << "-" << rowIndexCPU_[i+2].value << "\n";
      checkPassed = false;
    }      
  }
  assert(checkPassed);
  std::cout << "Consistency Check Passed\n";
}

std::vector<sparseEntry> sdn::SparseCudaMatrix::getSparseEntries() {
  dirtyCPU_ = true;
  updateMatrixCPU();
  std::vector<sparseEntry> ret;
  int start, end;
  
  for (int i = 0; i < colEntries_; i++) {
    start = colIndexCPU_[i].value;
    if(i == colEntries_-1)
      end = numEntries_;
    else
      end = colIndexCPU_[i+1].value;
    assert(start <= end); // In case the GPU messes up, this will catch it
    for (int j = start; j < end; j++){
      ret.push_back(sparseEntry(colMatrixCPU_[j].index, colIndexCPU_[i].index, colMatrixCPU_[j].value));
    }
  }
  return ret;
}

void sdn::SparseCudaMatrix::allocateMatrixCPU() const {
  SDN_ASSERT(rowIndexCPU_ == NULL);

  rowIndexCPU_ = (indexEntry*)malloc(toBlockSize(rowEntries_+1)*sizeof(indexEntry));
  if(rowIndexCPU_ == NULL) {
    SDN_FATAL("SparseCudaMatrix::allocateMatrixCPU malloc failed!");
  }
  colIndexCPU_ = (indexEntry*)malloc(toBlockSize(colEntries_+1)*sizeof(indexEntry));
  if(colIndexCPU_ == NULL) {
    SDN_FATAL("SparseCudaMatrix::allocateMatrixCPU malloc failed!");
  }
  rowMatrixCPU_ = (entry*)malloc(pNumEntries_*sizeof(entry));
  if(rowMatrixCPU_ == NULL) {
    SDN_FATAL("SparseCudaMatrix::allocateMatrixCPU malloc failed!");
  }
  colMatrixCPU_ = (entry*)malloc(pNumEntries_*sizeof(entry));
  if(colMatrixCPU_ == NULL) {
    SDN_FATAL("SparseCudaMatrix::allocateMatrixCPU malloc failed!");
  }
  dirtyCPU_ = true;
}

void sdn::SparseCudaMatrix::allocateMatrixGPU() const {
  SDN_ASSERT(rowIndexGPU_ == NULL);

  rowIndexGPU_ = (indexEntry*)CudaMemHandler::alloc(toBlockSize(rowEntries_+1)
						   * sizeof(indexEntry));
  colIndexGPU_ = (indexEntry*)CudaMemHandler::alloc(toBlockSize(colEntries_+1)
						    * sizeof(indexEntry));
  rowMatrixGPU_ = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  colMatrixGPU_ = (entry*)CudaMemHandler::alloc(pNumEntries_ * sizeof(entry));
  dirtyGPU_ = true;
}

bool sdn::SparseCudaMatrix::updateMatrixCPU() const {
  if(!dirtyCPU_) return false;
  
  if(rowIndexCPU_ == NULL)
    allocateMatrixCPU();

  // synchronize with GPU matrix (if it exists)
  if(rowMatrixGPU_ != NULL) {
    if(!checkSuccess(cublasGetVector(toBlockSize(rowEntries_+1), sizeof(indexEntry),
				     rowIndexGPU_, 1, rowIndexCPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixCPU cublasGetVector failed!");
    }
    if(!checkSuccess(cublasGetVector(toBlockSize(colEntries_+1), sizeof(indexEntry),
				     colIndexGPU_, 1, colIndexCPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixCPU cublasGetVector failed!");
    }
    if(!checkSuccess(cublasGetVector(pNumEntries_, sizeof(entry),
				     rowMatrixGPU_, 1, rowMatrixCPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixCPU cublasGetVector failed!");
    }
    if(!checkSuccess(cublasGetVector(pNumEntries_, sizeof(entry),
				     colMatrixGPU_, 1, colMatrixCPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixCPU cublasGetVector failed!");
    }
    dirtyCPU_ = false;
    return true;
  }
  return false;
}

bool sdn::SparseCudaMatrix::updateMatrixGPU() const {
  if(!dirtyGPU_) return false;
  
  if(rowIndexGPU_ == NULL)
    allocateMatrixGPU();

  // synchronize with CPU matrix (if it exists)
  if(rowMatrixCPU_ != NULL) {
    if(!checkSuccess(cublasSetVector(toBlockSize(rowEntries_+1), sizeof(indexEntry),
				     rowIndexCPU_, 1, rowIndexGPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixGPU cublasSetVector failed!");
    }
    if(!checkSuccess(cublasSetVector(toBlockSize(colEntries_+1), sizeof(indexEntry),
				     colIndexCPU_, 1, colIndexGPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixGPU cublasSetVector failed!");
    }
    if(!checkSuccess(cublasSetVector(pNumEntries_, sizeof(entry),
				     rowMatrixCPU_, 1, rowMatrixGPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixGPU cublasSetVector failed!");
    }
    if(!checkSuccess(cublasSetVector(pNumEntries_, sizeof(entry),
				     colMatrixCPU_, 1, colMatrixGPU_, 1))) {
      SDN_FATAL("SparseCudaMatrix::updateMatrixGPU cublasSetVector failed!");
    }
    dirtyGPU_ = false;
    return true;
  }
  return false;
}

namespace sdn {
  std::ostream& operator<<(std::ostream& out, const SparseCudaMatrix& rhs) {
    rhs.updateMatrixCPU();
    for(int i = 0; i < rhs.rowEntries_; i++) {
      indexEntry actualRow = rhs.rowIndexCPU_[i];
      int row = actualRow.index;
      int start = actualRow.value;
      int end = rhs.rowIndexCPU_[i+1].value;
      for(int j = start; j < end; j++) {
	entry actualCol = rhs.rowMatrixCPU_[j];
	int col = actualCol.index;
	float val = actualCol.value;
	out << row << " " << col << " " << val << std::endl;
      }
    }
    return out;
  }
 
  SparseCudaMatrix denseTimesDenseCWiseTimesSparse(const CudaMatrix& lhs,
						   const CudaMatrix& rhs,
						   const SparseCudaMatrix& cw) {
    if(lhs.transpose_)
      lhs.forceTranspose();
    if(rhs.transpose_)
      rhs.forceTranspose();

    SDN_ASSERT(lhs.n_ == rhs.m_ &&
               lhs.m_ == cw.m_ && rhs.n_ == cw.n_);

    lhs.updateMatrixGPU();
    rhs.updateMatrixGPU();
    cw.updateMatrixGPU();

    int sr = cw.rowEntries_;
    int sc = cw.colEntries_;
    int numEntries = cw.numEntries_;
    int pNumEntries = cw.pNumEntries_;
    indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sr+1)
								 * sizeof(indexEntry));
    indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sc+1)
								 * sizeof(indexEntry));
    entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
    entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
    cudaMatrixTimesCWiseTimesSparse(rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU,
				    lhs.matrixGPU_, rhs.matrixGPU_,
				    cw.rowIndexGPU_, cw.rowMatrixGPU_, cw.colIndexGPU_, cw.colMatrixGPU_,
				    lhs.m_, lhs.n_, lhs.pm_, lhs.pn_,
				    rhs.m_, rhs.n_, rhs.pm_, rhs.pn_,
				    sr, sc);
    return SparseCudaMatrix(lhs.rows(), rhs.cols(), sr, sc, numEntries, pNumEntries,
			    rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU, false);
  }

  SparseCudaMatrix sparseTimesDenseToSparse(const SparseCudaMatrix& lhs,
					    const CudaMatrix& rhs) {
    if(rhs.transpose_)
      rhs.forceTranspose();

    SDN_ASSERT(lhs.n_ == rhs.m_);

    lhs.updateMatrixGPU();
    rhs.updateMatrixGPU();

    int sr = lhs.rowEntries_;
    int sc = rhs.n_;
    int numEntries = sr * sc;
    int pNumEntries = toBlockSize(numEntries);
    indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sr+1)
								 * sizeof(indexEntry));
    indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sc+1)
								 * sizeof(indexEntry));
    entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
    entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
    sparseTimesDenseToSparse(rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU,
			     lhs.rowIndexGPU_, lhs.rowMatrixGPU_, rhs.matrixGPU_,
			     lhs.rowEntries_, rhs.m_, rhs.n_, rhs.pm_, rhs.pn_);
    return SparseCudaMatrix(lhs.rows(), rhs.cols(), sr, sc, numEntries, pNumEntries,
			    rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU, false);
  }

  CudaMatrix sparseTimesDenseToDense(const SparseCudaMatrix& lhs, const CudaMatrix& rhs) {
    if(rhs.transpose_)
      rhs.forceTranspose();

    SDN_ASSERT(lhs.n_ == rhs.m_);
    
    lhs.updateMatrixGPU();
    rhs.updateMatrixGPU();
    
    int tpm = pad(lhs.rows());
    int tpn = pad(rhs.cols());
    float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
    cudaMatrixSetZero(tmatrix, tpm*tpn);
    sparseTimesDenseToDense(tmatrix, lhs.rowIndexGPU_, lhs.rowMatrixGPU_,
			    rhs.matrixGPU_, lhs.rowEntries_, tpm,
			    rhs.m_, rhs.n_, rhs.pm_, rhs.pn_);
    return CudaMatrix(lhs.rows(), rhs.cols(), tmatrix, false, false);
  }

  SparseCudaMatrix online_orthogonalization(const SparseCudaMatrix& W, 
					    const int maxNonZeroPerRow, 
					    int windowSize, int step, 
					    int windowsPerRow, 
					    int windowsPerCol, 
					    int outputChannels, 
					    int inputChannels) {
    W.updateMatrixGPU();

    int sr = W.rowEntries_;
    int sc = W.colEntries_;
    int numEntries = W.numEntries_;
    int pNumEntries = W.pNumEntries_;
    indexEntry* rowIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sr+1)
								 * sizeof(indexEntry));
    indexEntry* colIndexGPU = (indexEntry*)CudaMemHandler::alloc(toBlockSize(sc+1)
								 * sizeof(indexEntry));
    entry* rowMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));
    entry* colMatrixGPU = (entry*)CudaMemHandler::alloc(pNumEntries * sizeof(entry));

    sparseCudaMatrixIndexCopy(rowIndexGPU, W.rowIndexGPU_, toBlockSize(W.rowEntries_+1));
    sparseCudaMatrixIndexCopy(colIndexGPU, W.colIndexGPU_, toBlockSize(W.colEntries_+1));

    if (!onlineOrthogonalization(rowIndexGPU, colIndexGPU, W.rowMatrixGPU_, W.colMatrixGPU_, rowMatrixGPU, colMatrixGPU, sr, sc, maxNonZeroPerRow, windowSize, step, windowsPerRow, windowsPerCol, outputChannels, inputChannels))
    {
      SDN_FATAL("SparseCudaMatrix::online_orthogonalization failed!");
    }

    return SparseCudaMatrix(W.rows(), W.cols(), sr, sc, numEntries, pNumEntries, rowIndexGPU, rowMatrixGPU, colIndexGPU, colMatrixGPU, false);
  }

  CudaMatrix weightsTimesData(const SparseCudaMatrix& W, const CudaMatrix &data) {
    if(data.transpose_)
      data.forceTranspose();

    SDN_ASSERT(W.n_ == data.m_);

    int sr = W.rowEntries_;
    int nonZeroPerRow = W.numEntries_ / sr;

    SDN_ASSERT(W.numEntries_ == nonZeroPerRow * sr);
    
    W.updateMatrixGPU();
    data.updateMatrixGPU();
    
    int tpm = pad(W.rows());
    int tpn = pad(data.cols());
    float* tmatrix = CudaMemHandler::alloc(tpm*tpn*sizeof(float));
    cudaMatrixSetZero(tmatrix, tpm*tpn);

    weightsTimesData(tmatrix, W.rowMatrixGPU_, data.matrixGPU_, tpm, W.rowEntries_, nonZeroPerRow, data.m_, data.n_, data.pm_);

    return CudaMatrix(W.rows(), data.cols(), tmatrix, false, false);
  }
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise::SparseCudaMatrixCwise(const SparseCudaMatrix& matrix)
  : matrix_(matrix) { }

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator+(float scalar) const {
  return matrix_.cWiseScalarOp(scalar, CWISE_ADDITION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator-(float scalar) const {
  return operator+(-scalar);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator*(float scalar) const {
  return matrix_.cWiseScalarOp(scalar, CWISE_MULTIPLICATION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator/(float scalar) const {
  SDN_ASSERT(scalar != 0);

  return operator*(1.0 / scalar);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator+(const SparseCudaMatrix& rhs) const {
  return matrix_.cWiseOp(rhs, CWISE_ADDITION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator-(const SparseCudaMatrix& rhs) const {
  return matrix_.cWiseOp(rhs, CWISE_SUBTRACTION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator*(const SparseCudaMatrix& rhs) const {
  return matrix_.cWiseOp(rhs, CWISE_MULTIPLICATION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator/(const SparseCudaMatrix& rhs) const {
  return matrix_.cWiseOp(rhs, CWISE_DIVISION);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::pow(float scalar) const {
  return matrix_.cWiseScalarOp(scalar, CWISE_POWER);
}

// SparseCudamatrix redux operations

sdn::SparseCudaMatrix::SparseCudaMatrixRedux::SparseCudaMatrixRedux(const SparseCudaMatrix& matrix, bool colwise)
  : matrix_(matrix), colwise_(colwise) {}

sdn::CudaMatrix sdn::SparseCudaMatrix::SparseCudaMatrixRedux::sum() const {
  matrix_.updateMatrixGPU();

  if (!colwise_) {
    int m = matrix_.rows();
    int pm = pad(m);
    float* matrix = CudaMemHandler::alloc(pm*sizeof(float));

    if (!sparseCudaMatrixPartialSum(matrix, pm, matrix_.rowEntries_, matrix_.rowIndexGPU_, matrix_.rowMatrixGPU_))
    {
      SDN_FATAL("SparseCudaMatrix::rowwise()::sum() failed!\n");
    }

    return CudaMatrix(m, 1, matrix, false, false);
  }
  else
  {
    SDN_FATAL("colwise sum() NOT IMPLEMENTED\n");
    return CudaMatrix(1, 1, NULL, false, false);
  }

}

// needs to be declared inside namespace
namespace sdn {
  SparseCudaMatrix operator+(float scalar,
			     const SparseCudaMatrix::SparseCudaMatrixCwise& rhs) {
    return rhs + scalar;
  }

  SparseCudaMatrix operator-(float scalar, 
			     const SparseCudaMatrix::SparseCudaMatrixCwise& rhs) {
    return rhs.matrix_.cWiseScalarOp(scalar, CWISE_SUBTRACTION);
  }
  
  SparseCudaMatrix operator*(float scalar, 
			     const SparseCudaMatrix::SparseCudaMatrixCwise& rhs) {
    return rhs * scalar;
  }
  
  SparseCudaMatrix operator/(float scalar,
			     const SparseCudaMatrix::SparseCudaMatrixCwise& rhs) {
    return rhs.matrix_.cWiseScalarOp(scalar, CWISE_DIVISION);
  }
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator+=(float scalar) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  sparseCudaMatrixCWiseScalarOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_, scalar,
				matrix_.numEntries_, matrix_.pNumEntries_,
				CWISE_ADDITION);
  sparseCudaMatrixCWiseScalarOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_, scalar,
				matrix_.numEntries_, matrix_.pNumEntries_,
				CWISE_ADDITION);
  return *this;
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator-=(float scalar) {
  return operator+=(-scalar);
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator*=(float scalar) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  sparseCudaMatrixCWiseScalarOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_, scalar,
				matrix_.numEntries_, matrix_.pNumEntries_,
				CWISE_MULTIPLICATION);
  sparseCudaMatrixCWiseScalarOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_, scalar,
				matrix_.numEntries_, matrix_.pNumEntries_,
				CWISE_MULTIPLICATION);
  return *this;
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator/=(float scalar) {
  SDN_ASSERT(scalar != 0);

  return operator*=(1 / scalar);
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator+=(const SparseCudaMatrix& rhs) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  rhs.updateMatrixGPU();

  if(VERIFY)
    verifyCanCwise(matrix_, rhs);

  sparseCudaMatrixCWiseOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_,
			  rhs.rowMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_ADDITION);
  sparseCudaMatrixCWiseOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_,
			  rhs.colMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_ADDITION);
  return *this;
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator-=(const SparseCudaMatrix& rhs) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  rhs.updateMatrixGPU();

  if(VERIFY)
    verifyCanCwise(matrix_, rhs);

  sparseCudaMatrixCWiseOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_,
			  rhs.rowMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_SUBTRACTION);
  sparseCudaMatrixCWiseOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_,
			  rhs.colMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_SUBTRACTION);
  return *this;
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator*=(const SparseCudaMatrix& rhs) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  rhs.updateMatrixGPU();

  if(VERIFY)
    verifyCanCwise(matrix_, rhs);

  sparseCudaMatrixCWiseOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_,
			  rhs.rowMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_MULTIPLICATION);
  sparseCudaMatrixCWiseOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_,
			  rhs.colMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_MULTIPLICATION);
  return *this;
}

sdn::SparseCudaMatrix::SparseCudaMatrixCwise&
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::operator/=(const SparseCudaMatrix& rhs) {
  // synchronize GPU matrices
  matrix_.updateMatrixGPU();
  rhs.updateMatrixGPU();

  if(VERIFY)
    verifyCanCwise(matrix_, rhs);

  sparseCudaMatrixCWiseOp(matrix_.rowMatrixGPU_, matrix_.rowMatrixGPU_,
			  rhs.rowMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_DIVISION);
  sparseCudaMatrixCWiseOp(matrix_.colMatrixGPU_, matrix_.colMatrixGPU_,
			  rhs.colMatrixGPU_, matrix_.numEntries_,
			  matrix_.pNumEntries_, CWISE_DIVISION);
  return *this;
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::square() const {
  return operator*(matrix_);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::exp() const {
  return matrix_.cWiseScalarOp(1, CWISE_EXP);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::log() const {
  return matrix_.cWiseScalarOp(1, CWISE_LOG);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::inverse() const {
  return 1 / (*this);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::abs() const {
  return matrix_.cWiseScalarOp(1, CWISE_ABS);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::tanh() const {
  return matrix_.cWiseScalarOp(1, CWISE_TANH);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::cosh() const {
  return matrix_.cWiseScalarOp(1, CWISE_COSH);
}

sdn::SparseCudaMatrix
sdn::SparseCudaMatrix::SparseCudaMatrixCwise::sqrt() const {
  return matrix_.cWiseScalarOp(1, CWISE_SQRT);
}

void sdn::SparseCudaMatrix::validate() {
  dirtyCPU_ = true;
  updateMatrixCPU();
  int i, j, k, l;
//  for(i=0;i<100;i++) printf("%d ", rowMatrixCPU_[i].index);
//  printf("\n");
//  return;
  for(i=0;i<rowEntries_;i++)
  {
    for(j=rowIndexCPU_[i].value;j<rowIndexCPU_[i+1].value;j++)
    {
      for(k=0;k<colEntries_;k++)
      {
        if(colIndexCPU_[k].index!=rowMatrixCPU_[j].index) continue;
        for(l=colIndexCPU_[k].value;l<colIndexCPU_[k+1].value;l++)
        {
          if(colMatrixCPU_[l].index!=rowIndexCPU_[i].index) continue;
          SDN_ASSERT(rowMatrixCPU_[j].value == colMatrixCPU_[l].value);
        }
      }
    }
  }
  for(k=0;k<colEntries_;k++)
  {
    for(l=colIndexCPU_[k].value;l<colIndexCPU_[k+1].value;l++)
    {
      for(i=0;i<rowEntries_;i++)
      {
        if(rowIndexCPU_[i].index!=colMatrixCPU_[l].index) continue;
        for(j=rowIndexCPU_[i].value;j<rowIndexCPU_[i+1].value;j++)
        {
          if(rowMatrixCPU_[j].index!=colIndexCPU_[k].index) continue;
          SDN_ASSERT(rowMatrixCPU_[j].value == colMatrixCPU_[l].value);
        }
      }
    }
  }
  SDN_ASSERT(rowIndexCPU_[rowEntries_].value == numEntries_);
  SDN_ASSERT(colIndexCPU_[colEntries_].value == numEntries_);
  printf("Passed:\n");
/*  int cnt=0;
  for(i=0;i<rowEntries_;i++)
    for(j=rowIndexCPU_[i].value;j<rowIndexCPU_[i+1].value;j++) {
      printf("(%d %d) ", rowIndexCPU_[i].index, rowMatrixCPU_[j].index);
      cnt++;
      if(cnt%16==0) printf("\n");
    }
  printf("\n");*/
}
