#pragma once
#include <iostream>
#include <vector>
#include "sparseCudaMatrixUtils.h"
#include "cudaMatrix.h"

// Cublas documentation:
// http://developer.download.nvidia.com/compute/cuda/2_1/toolkit/docs/CUBLAS_Library_2.1.pdf

struct sparseEntry {
  int row;
  int col;
  float val;
  sparseEntry(int _row=0, int _col=0, float _val=0.0) {
    row = _row; col = _col; val = _val;
  }
};

namespace sdn {

  class SparseCudaMatrix {
  protected:
    class SparseCudaMatrixCwise;
    class SparseCudaMatrixRedux;

  public:
    // SparseCudaMatrix constructors/destructors
    SparseCudaMatrix(); // Do not use this constructor unless you are sure what you're doing.
    SparseCudaMatrix(int m, int n, int numEntries, 
		     std::vector<int> rows, std::vector<int> cols,
		     std::vector<std::vector<entry> > rowEntries,
		     std::vector<std::vector<entry> > colEntries);
    SparseCudaMatrix(int m, int n, std::vector<sparseEntry> entries);
    SparseCudaMatrix(const SparseCudaMatrix& other);
		SparseCudaMatrix(const CudaMatrix& other);
    ~SparseCudaMatrix();

    // SparseCudaMatrix element getting/setting
    void set(int i, int j, float f);
    const float get(int i, int j) const;
    void set(int i, float f);
    const float get(int i) const;
    const float operator()(int i, int j) const;
    const float operator[](int i) const;

    // SparseCudaMatrix cwise operations
    SparseCudaMatrixCwise cwise();
    const SparseCudaMatrixCwise cwise() const;

    // SparseCudaMatrix binary operations
    SparseCudaMatrix& operator=(const SparseCudaMatrix& rhs);
		SparseCudaMatrix& operator=(const CudaMatrix& rhs);
    bool operator==(const SparseCudaMatrix& rhs) const;
    bool operator!=(const SparseCudaMatrix& rhs) const;
    SparseCudaMatrix operator*(const SparseCudaMatrix& rhs) const;
    SparseCudaMatrix operator+(const SparseCudaMatrix& rhs) const;
    SparseCudaMatrix operator-(const SparseCudaMatrix& rhs) const;

    // SparseCudaMatrix binary operations with assignment
    SparseCudaMatrix& operator+=(const SparseCudaMatrix& rhs);
    SparseCudaMatrix& operator-=(const SparseCudaMatrix& rhs);

    // SparseCudaMatrix scalar operations
    SparseCudaMatrix operator*(float scalar) const;
    SparseCudaMatrix operator/(float scalar) const;

    // SparseCudaMatrix scalar operations with assignment
    SparseCudaMatrix& operator*=(float scalar);
    SparseCudaMatrix& operator/=(float scalar);

    // SparseCudaMatrix unary operations
    SparseCudaMatrix& transposeEquals();
    SparseCudaMatrix transpose() const;
    SparseCudaMatrix operator-() const;

    // SparseCudaMatrix reduction operations
    float sum() const;
    float minCoeff() const;
    float maxCoeff() const;
    float norm() const;
    float squaredNorm() const;

    // SparseCudaMatrix redux
    SparseCudaMatrixRedux rowwise() const;

    // SparseCudaMatrix resizing
    void resize(int m, int n);
    
    // Constant-time transpose
    void transposeThis();

    // SparseCudaMatrix getters
    int rows() const;
    int cols() const;
    int size() const;
    int rowEntries() const;
    int colEntries() const;
    int numEntries() const;
    float density() const;
    
    // SparseCudaMatrix direct data access
    entry* getColMatrixEntries();
    indexEntry* getColMatrixIndex();
    
    entry* getRowMatrixEntries();
    indexEntry* getRowMatrixIndex();

    std::vector<sparseEntry> getSparseEntries();
    void checkIndexConsistency();


    // SparseCudaMatrixBlock output
    friend std::ostream& operator<<(std::ostream& out,
				    const SparseCudaMatrix& rhs);

    friend void verifyCanCwise(const SparseCudaMatrix& lhs, const SparseCudaMatrix& rhs);

    friend SparseCudaMatrix denseTimesDenseCWiseTimesSparse(const CudaMatrix& lhs,
							    const CudaMatrix& rhs,
							    const SparseCudaMatrix& cw);
    friend SparseCudaMatrix sparseTimesDenseToSparse(const SparseCudaMatrix& lhs,
						     const CudaMatrix& rhs);

    friend CudaMatrix sparseTimesDenseToDense(const SparseCudaMatrix& lhs,
					      const CudaMatrix& rhs);

    // localrf methods
    // method for localrf orthogonalization. probably want relocate this somewhere else in future

    friend SparseCudaMatrix online_orthogonalization(const SparseCudaMatrix& W, 
						     const int maxNonZeroPerRow, 
						     const int windowSize, 
						     const int step, 
						     const int windowsPerRow, 
						     const int windowsPerCol, 
						     const int outputChannels, 
						     const int inputChannels);
    friend CudaMatrix weightsTimesData(const SparseCudaMatrix& W, const CudaMatrix &data);

    // SparseCudaMatrixCwise scalar operations
    friend SparseCudaMatrix operator+(float scalar, const SparseCudaMatrixCwise& rhs);
    friend SparseCudaMatrix operator-(float scalar, const SparseCudaMatrixCwise& rhs);
    friend SparseCudaMatrix operator*(float scalar, const SparseCudaMatrixCwise& rhs);
    friend SparseCudaMatrix operator/(float scalar, const SparseCudaMatrixCwise& rhs);
    
    void validate(); // temporary function for debugging
    
  protected:
  
    void initializeEmpty();
  
    void initialize(int m, int n, int numEntries, 
		    std::vector<int> rows, std::vector<int> cols,
		    std::vector<std::vector<entry> > rowEntries,
		    std::vector<std::vector<entry> > colEntries);
    
    SparseCudaMatrix(int m, int n, int sr, int sc, int numEntires, int pNumEntries,
		     indexEntry* ri, entry* rm, indexEntry* ci, entry* cm, bool onCPU);
    SparseCudaMatrix& cWiseOpEq(const SparseCudaMatrix& rhs, CUDA_MATRIX_OPTYPE op);
    SparseCudaMatrix cWiseOp(const SparseCudaMatrix& rhs, CUDA_MATRIX_OPTYPE op) const;
    SparseCudaMatrix& cWiseScalarOpEq(float rhs, CUDA_MATRIX_OPTYPE op);
    SparseCudaMatrix cWiseScalarOp(float rhs, CUDA_MATRIX_OPTYPE op) const;

    // methods to manage lazy operations; note they are const even though
    // they modify fields in the class
    void allocateMatrixCPU() const;
    void allocateMatrixGPU() const;
    bool updateMatrixCPU() const;
    bool updateMatrixGPU() const;

    // mutable to get around const; make sure const methods keep the class
    // logically constant even if they modify these fields
    mutable int m_, n_;
    mutable int rowEntries_, colEntries_, numEntries_, pNumEntries_;
    mutable indexEntry* rowIndexCPU_; // indicates where the first element of each row is
    mutable indexEntry* colIndexCPU_; // indicates where the first element of each column is
    mutable entry* rowMatrixCPU_; // row-major ordered entries on CPU
    mutable entry* colMatrixCPU_; // column-major ordered entries on CPU
    mutable indexEntry* rowIndexGPU_;
    mutable indexEntry* colIndexGPU_;
    mutable entry* rowMatrixGPU_; // row-major ordered entries on GPU
    mutable entry* colMatrixGPU_; // column-major ordered entries on GPU
    mutable bool dirtyCPU_;
    mutable bool dirtyGPU_;

    // protected class to handle cwise operations; should never be stored or
    // manipulated by users
    class SparseCudaMatrixCwise {
    public:
      // SparseCudaMatrixCwise constructors
      SparseCudaMatrixCwise(const SparseCudaMatrix& matrix);

      // SparseCudaMatrixCwise binary operations
      SparseCudaMatrix operator+(float scalar) const;
      SparseCudaMatrix operator-(float scalar) const;
      SparseCudaMatrix operator*(float scalar) const;
      SparseCudaMatrix operator/(float scalar) const;
      SparseCudaMatrix pow(float scalar) const;
      
      friend SparseCudaMatrix operator+(float scalar, const SparseCudaMatrixCwise& rhs);
      friend SparseCudaMatrix operator-(float scalar, const SparseCudaMatrixCwise& rhs);
      friend SparseCudaMatrix operator/(float scalar, const SparseCudaMatrixCwise& rhs);
      
      // SparseCudaMatrixCwise scalar operations with assignment
      SparseCudaMatrixCwise& operator+=(float scalar);
      SparseCudaMatrixCwise& operator-=(float scalar);
      SparseCudaMatrixCwise& operator*=(float scalar);
      SparseCudaMatrixCwise& operator/=(float scalar);

      // SparseCudaMatrixCwise binary cwise operations
      SparseCudaMatrix operator+(const SparseCudaMatrix& rhs) const;
      SparseCudaMatrix operator-(const SparseCudaMatrix& rhs) const;
      SparseCudaMatrix operator*(const SparseCudaMatrix& rhs) const;
      SparseCudaMatrix operator/(const SparseCudaMatrix& rhs) const;

      // SparseCudaMatrixCwise binary cwise operations with assignment
      SparseCudaMatrixCwise& operator+=(const SparseCudaMatrix& rhs);
      SparseCudaMatrixCwise& operator-=(const SparseCudaMatrix& rhs);
      SparseCudaMatrixCwise& operator*=(const SparseCudaMatrix& rhs);
      SparseCudaMatrixCwise& operator/=(const SparseCudaMatrix& rhs);

      // SparseCudaMatrixCwise unary operations
      //Note these are only applied to the nonzero entries!
      SparseCudaMatrix square() const;
      SparseCudaMatrix exp() const;
      SparseCudaMatrix log() const;
      SparseCudaMatrix inverse() const;
      SparseCudaMatrix abs() const;
      SparseCudaMatrix tanh() const;
      SparseCudaMatrix cosh() const;
      SparseCudaMatrix sqrt() const;

    protected:
      const SparseCudaMatrix& matrix_;
    };

    class SparseCudaMatrixRedux {
      public:

        //constructors
        SparseCudaMatrixRedux(const SparseCudaMatrix& matrix, bool colwise);

        //reduction operations
        
        CudaMatrix sum() const;

      protected:

        const SparseCudaMatrix& matrix_;
        bool colwise_;
    };
  };

  // SparseCudaMatrix utility functions
  int toBlockSize(int d);
}

