#include <math.h>
#include <stdio.h>
#include <cuda.h>
#include <cutil_inline.h>
#include <cublas.h>
#include <float.h>

#include "sparseCudaMatrixUtils.h"

#ifndef __MAX___
#define MAX(a,b) (a) > (b) ? a : b 
#endif

#ifndef __MIN__
#define MIN(a,b) (a) < (b) ? a : b
#endif

struct sparseCWiseParams {
  entry* so;
  entry* sl;
  entry* sr;
  int n;
  CUDA_MATRIX_OPTYPE type;
};

struct sparseCWiseScalarParams {
  entry* so;
  entry* si;
  float sa;
  int n;
  CUDA_MATRIX_OPTYPE type;
};

struct sparseReduxParams {
  float* so;
  entry* si;
  indexEntry* sIndex;
  int n;
  int per;
};

__global__ void sparseCWiseOpMain(struct sparseCWiseParams parms) {
  int i;
  const entry *sl;
  const entry *sr;
  entry *so;
  CUDA_MATRIX_OPTYPE type;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  sl = parms.sl;
  sr = parms.sr;
  so = parms.so;
  type = parms.type;

  i = blockIdx.x * blockDim.x + threadIdx.x;
  if(i < parms.n) {
    so[i].index = sl[i].index;
    switch(type) {
    case CWISE_ADDITION:
      so[i].value = sl[i].value + sr[i].value;
      break;
    case CWISE_SUBTRACTION:
      so[i].value = sl[i].value - sr[i].value;
      break;
    case CWISE_MULTIPLICATION:
      so[i].value = sl[i].value * sr[i].value;
      break;
    case CWISE_DIVISION:
      so[i].value = sl[i].value / sr[i].value;
      break;
    }
  }
}

int sparseCudaMatrixCWiseOp(entry* matrixOut, entry* matrixLhs, entry* matrixRhs,
                            int numEntries, int pNumEntries, CUDA_MATRIX_OPTYPE type) {
  struct sparseCWiseParams params;
  cudaError_t cudaStat;

  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.so = matrixOut;
  params.sl = matrixLhs;
  params.sr = matrixRhs;
  params.type = type;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCWiseOpMain<<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCWiseScalarOpMain(struct sparseCWiseScalarParams parms) {
  int i;
  const entry *si;
  float sa;
  entry *so;
  CUDA_MATRIX_OPTYPE type;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  si = parms.si;
  sa = parms.sa;
  so = parms.so;
  type = parms.type;
  
  i = blockIdx.x * blockDim.x + threadIdx.x;
  if(i < parms.n) {
    so[i].index = si[i].index;
    switch(type) {
    case CWISE_ADDITION:
      so[i].value = sa + si[i].value;
      break;
    case CWISE_SUBTRACTION:
      so[i].value = sa - si[i].value;
      break;
    case CWISE_MULTIPLICATION:
      so[i].value = sa * si[i].value;
      break;
    case CWISE_DIVISION:
      so[i].value = sa / si[i].value;
      break;
    case CWISE_POWER:
      so[i].value = __powf(si[i].value, sa);
      break;
    case CWISE_EXP:
      so[i].value = __expf(si[i].value);
      break;
    case CWISE_LOG:
      so[i].value = __logf(si[i].value);
      break;
    case CWISE_ABS:
      so[i].value = fabsf(si[i].value);
      break;
    case CWISE_TANH:
      so[i].value = tanhf(si[i].value);
      break;
    case CWISE_COSH:
      so[i].value = coshf(si[i].value);
      break;
    case CWISE_CONSTANT:
      so[i].value = sa;
      break;
    }
  }
}

int sparseCudaMatrixCWiseScalarOp(entry* matrixOut, entry* matrixIn, float scalar,
                                  int numEntries, int pNumEntries,
                                  CUDA_MATRIX_OPTYPE type) {
  struct sparseCWiseScalarParams params;
  cudaError_t cudaStat;
  
  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.so = matrixOut;
  params.si = matrixIn;
  params.sa = scalar;
  params.type = type;

  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCWiseScalarOpMain<<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(params);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixSumMain(struct sparseReduxParams parms) {
  __shared__ float sdata[BLOCK_DIM];
  int i, per, n, tid = threadIdx.x;
  const entry *si;
  float *so;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  si = parms.si;
  so = parms.so;
  n = parms.n;
  per = parms.per;

  sdata[tid] = 0;
  
  int start = (blockIdx.x * blockDim.x + tid) * per;
  int end = start + per;
  float t, y, c = 0;
  for (i = start; i < n && i < end; i++) {
    y = si[i].value - c;
    t = sdata[tid] + y;
    c = (t - sdata[tid]) - y;
    sdata[tid] = t;
  }

  __syncthreads();

  if(tid % HALF_DIM == 0) {
    c = 0;
    for (i = tid + 1; i < tid + HALF_DIM; i++) {
      y = sdata[i] - c;
      t = sdata[tid] + y;
      c = (t - sdata[tid]) - y;
      sdata[tid] = t;
    }
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = sdata[0];
    for (i = HALF_DIM; i < BLOCK_DIM; i += HALF_DIM)
      so[blockIdx.x] += sdata[i];
  }
}

int sparseCudaMatrixSum(float* result, entry* matrixIn, int numEntries, int pNumEntries) {
  struct sparseReduxParams params;
  cudaError_t cudaStat;
  
  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.si = matrixIn;
  int per = (int)sqrt(pNumEntries / BLOCK_DIM + 0.1);
  int numBlocks;
  if(pNumEntries % per == 0)
    numBlocks = pNumEntries / per;
  else
    numBlocks = pNumEntries / per + 1;

  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, numBlocks * sizeof(float)));
  params.so = d_block;
  params.per = per;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCudaMatrixSumMain<<<numBlocks,BLOCK_DIM>>>(params);

  float* h_block = (float*)malloc(numBlocks * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, numBlocks * sizeof(float),
                                 cudaMemcpyDeviceToHost));

  *result = 0;
  for (int i = 0; i < numBlocks; i++)
    *result += h_block[i];

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixMinMain(struct sparseReduxParams parms) {
  __shared__ float sdata[BLOCK_DIM];
  int i, per, n, tid = threadIdx.x;
  const entry *si;
  float *so;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  si = parms.si;
  so = parms.so;
  n = parms.n;
  per = parms.per;

  sdata[tid] = FLT_MAX;
  
  int start = (blockIdx.x * blockDim.x + tid) * per;
  int end = start + per;
  for (i = start; i < n && i < end; i++) {
    sdata[tid] = fmin(sdata[tid], si[i].value);
  }

  __syncthreads();

  if(tid % HALF_DIM == 0) {
    for (i = tid + 1; i < tid + HALF_DIM; i++) {
      sdata[tid] = fmin(sdata[tid], sdata[i]);
    }
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = sdata[0];
    for (i = HALF_DIM; i < BLOCK_DIM; i += HALF_DIM)
      so[blockIdx.x] = fmin(sdata[i], so[blockIdx.x]);
  }
}

int sparseCudaMatrixMin(float* result, entry* matrixIn, int numEntries, int pNumEntries) {
  struct sparseReduxParams params;
  cudaError_t cudaStat;
  
  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.si = matrixIn;
  int per = (int)sqrt(pNumEntries / BLOCK_DIM + 0.1);
  int numBlocks;
  if(pNumEntries % per == 0)
    numBlocks = pNumEntries / per;
  else
    numBlocks = pNumEntries / per + 1;

  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, numBlocks * sizeof(float)));
  params.so = d_block;
  params.per = per;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCudaMatrixMinMain<<<numBlocks,BLOCK_DIM>>>(params);

  float* h_block = (float*)malloc(numBlocks * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, numBlocks * sizeof(float),
                                 cudaMemcpyDeviceToHost));

  *result = h_block[0];
  for (int i = 1; i < numBlocks; i++)
    *result = fmin(h_block[i], *result);

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixMaxMain(struct sparseReduxParams parms) {
  __shared__ float sdata[BLOCK_DIM];
  int i, per, n, tid = threadIdx.x;
  const entry *si;
  float *so;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  si = parms.si;
  so = parms.so;
  n = parms.n;
  per = parms.per;

  sdata[tid] = -FLT_MAX;
  
  int start = (blockIdx.x * blockDim.x + tid) * per;
  int end = start + per;
  for (i = start; i < n && i < end; i++) {
    sdata[tid] = fmax(sdata[tid], si[i].value);
  }

  __syncthreads();

  if(tid % HALF_DIM == 0) {
    for (i = tid + 1; i < tid + HALF_DIM; i++) {
      sdata[tid] = fmax(sdata[tid], sdata[i]);
    }
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = sdata[0];
    for (i = HALF_DIM; i < BLOCK_DIM; i += HALF_DIM)
      so[blockIdx.x] = fmax(sdata[i], so[blockIdx.x]);
  }
}

int sparseCudaMatrixMax(float* result, entry* matrixIn, int numEntries, int pNumEntries) {
  struct sparseReduxParams params;
  cudaError_t cudaStat;
  
  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.si = matrixIn;
  int per = (int)sqrt(pNumEntries / BLOCK_DIM + 0.1);
  int numBlocks;
  if(pNumEntries % per == 0)
    numBlocks = pNumEntries / per;
  else
    numBlocks = pNumEntries / per + 1;

  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, numBlocks * sizeof(float)));
  params.so = d_block;
  params.per = per;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCudaMatrixMaxMain<<<numBlocks,BLOCK_DIM>>>(params);

  float* h_block = (float*)malloc(numBlocks * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, numBlocks * sizeof(float),
                                 cudaMemcpyDeviceToHost));

  *result = h_block[0];
  for (int i = 1; i < numBlocks; i++)
    *result = fmax(h_block[i], *result);

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixSquaredNormMain(struct sparseReduxParams parms) {
  __shared__ float sdata[BLOCK_DIM];
  int i, per, n, tid = threadIdx.x;
  const entry *si;
  float *so;
  
  /* NOTE: wrapper must ensure that parms.n > 0  */
  si = parms.si;
  so = parms.so;
  n = parms.n;
  per = parms.per;

  sdata[tid] = 0;
  
  int start = (blockIdx.x * blockDim.x + tid) * per;
  int end = start + per;
  float t, y, c = 0;
  for (i = start; i < n && i < end; i++) {
    y = si[i].value * si[i].value - c;
    t = sdata[tid] + y;
    c = (t - sdata[tid]) - y;
    sdata[tid] = t;
  }

  __syncthreads();

  if(tid % HALF_DIM == 0) {
    c = 0;
    for (i = tid + 1; i < tid + HALF_DIM; i++) {
      y = sdata[i] - c;
      t = sdata[tid] + y;
      c = (t - sdata[tid]) - y;
      sdata[tid] = t;
    }
  }

  __syncthreads();

  // write result for this block to global mem
  if (tid == 0) {
    so[blockIdx.x] = sdata[0];
    for (i = HALF_DIM; i < BLOCK_DIM; i += HALF_DIM)
      so[blockIdx.x] += sdata[i];
  }
}

int sparseCudaMatrixSquaredNorm(float* result, entry* matrixIn,
                                int numEntries, int pNumEntries) {
  struct sparseReduxParams params;
  cudaError_t cudaStat;
  
  memset (&params, 0, sizeof(params));
  params.n = numEntries;
  params.si = matrixIn;
  int per = (int)sqrt(pNumEntries / BLOCK_DIM + 0.1);
  int numBlocks;
  if(pNumEntries % per == 0)
    numBlocks = pNumEntries / per;
  else
    numBlocks = pNumEntries / per + 1;

  float* d_block = NULL;
  cutilSafeCallNoSync(cudaMalloc((void**)&d_block, numBlocks * sizeof(float)));
  params.so = d_block;
  params.per = per;
  
  cudaStat = cudaGetLastError(); /* clear error status */
  sparseCudaMatrixSquaredNormMain<<<numBlocks,BLOCK_DIM>>>(params);

  float* h_block = (float*)malloc(numBlocks * sizeof(float));
  cutilSafeCallNoSync(cudaMemcpy(h_block, d_block, numBlocks * sizeof(float),
                                 cudaMemcpyDeviceToHost));

  *result = 0;
  for (int i = 0; i < numBlocks; i++)
    *result += h_block[i];

  cutilSafeCallNoSync(cudaFree(d_block));
  free(h_block);
  
  cudaStat = cudaGetLastError(); /* check for launch error */
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixEntryCopyMain(entry* so, entry* si, int n) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  if (i < n) so[i] = si[i];
}

int sparseCudaMatrixEntryCopy(entry* matrixOut, entry* matrixIn, int pNumEntries) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixEntryCopyMain
    <<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(matrixOut, matrixIn, pNumEntries);
    
  cudaThreadSynchronize();
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixIndexCopyMain(indexEntry* so, indexEntry* si, int n) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  if (i < n) so[i] = si[i];
}

int sparseCudaMatrixIndexCopy(indexEntry* matrixOut, indexEntry* matrixIn,
                              int pNumEntries) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixIndexCopyMain
    <<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(matrixOut, matrixIn, pNumEntries);

  cudaThreadSynchronize();
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixTestEqualityMain(int* result, entry* matrixLhs,
                                                 entry* matrixRhs, int numEntries) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  if(i < numEntries && (matrixLhs[i].index != matrixRhs[i].index ||
                        matrixLhs[i].value != matrixRhs[i].value))
    *result = 1;
}

int sparseCudaMatrixTestEquality(int* result, entry* matrixLhs, entry* matrixRhs,
                                 int numEntries, int pNumEntries) {
  cudaError_t cudaStat;

  int* error = NULL;
  
  cutilSafeCallNoSync(cudaMalloc((void**)&error, sizeof(int)));
  cutilSafeCallNoSync(cudaMemset(error, 0, sizeof(int)));
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixTestEqualityMain
    <<<(pNumEntries-BLOCK_DIM+1)/BLOCK_DIM,BLOCK_DIM>>>(error, matrixLhs, matrixRhs, numEntries);

  cutilSafeCallNoSync(cudaMemcpy(result, error, sizeof(int), cudaMemcpyDeviceToHost));

  cutilSafeCallNoSync(cudaFree(error));
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixVerifyEntriesMain(int* result, entry* matrixLhs,
                                                  entry* matrixRhs, int numEntries) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  if(i < numEntries && matrixLhs[i].index != matrixRhs[i].index)
    *result = 1;
}

int sparseCudaMatrixVerifyEntries(int* result, entry* matrixLhs, entry* matrixRhs,
                                  int numEntries, int pNumEntries) {
  cudaError_t cudaStat;

  int* error = NULL;
  
  cutilSafeCallNoSync(cudaMalloc((void**)&error, sizeof(int)));
  cutilSafeCallNoSync(cudaMemset(error, 0, sizeof(int)));
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixVerifyEntriesMain
    <<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(error, matrixLhs, matrixRhs, numEntries);
  cudaThreadSynchronize();

  cutilSafeCallNoSync(cudaMemcpy(result, error, sizeof(int), cudaMemcpyDeviceToHost));

  cutilSafeCallNoSync(cudaFree(error));
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixVerifyIndexEntriesMain(int* result, indexEntry* matrixLhs,
                                                       indexEntry* matrixRhs,
                                                       int numEntries) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  if(i < numEntries && (matrixLhs[i].index != matrixRhs[i].index ||
                       matrixLhs[i].value != matrixRhs[i].value))
    *result = 1;
}

int sparseCudaMatrixVerifyIndexEntries(int* result, indexEntry* matrixLhs,
                                       indexEntry* matrixRhs, int numEntries,
                                       int pNumEntries) {
  cudaError_t cudaStat;
  
  int* error = NULL;
  
  cutilSafeCallNoSync(cudaMalloc((void**)&error, sizeof(int)));
  cutilSafeCallNoSync(cudaMemset(error, 0, sizeof(int)));
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixVerifyIndexEntriesMain
    <<<(pNumEntries+BLOCK_DIM-1)/BLOCK_DIM,BLOCK_DIM>>>(error, matrixLhs, matrixRhs, numEntries);
  cudaThreadSynchronize();

  cutilSafeCallNoSync(cudaMemcpy(result, error, sizeof(int), cudaMemcpyDeviceToHost));

  cutilSafeCallNoSync(cudaFree(error));
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixMultiplyMain(indexEntry* rowIndexOut, entry* rowOut,
                                             indexEntry* colIndexOut, entry* colOut,
                                             indexEntry* lhsIndexIn, entry* lhsIn,
                                             indexEntry* rhsIndexIn, entry* rhsIn,
                                             int lhsRows, int rhsCols) {
  int row = blockIdx.x;
  indexEntry actualRow = lhsIndexIn[row];
  rowIndexOut[row].index = actualRow.index;
  rowIndexOut[row].value = row * rhsCols; //dotting each row with each col
  rowIndexOut[row+1].value = (row+1) * rhsCols; //don't forget extra entry at end
  int rowEnd = lhsIndexIn[row+1].value;
  
  int per = (rhsCols + blockDim.x - 1) / blockDim.x; //DIV_ROUND_UP
  int startCol = threadIdx.x * per;
  int endCol = startCol + per;
  if (endCol > rhsCols)
    endCol = rhsCols;

  int col;
  for(col = startCol; col < endCol; col++) {
    indexEntry actualCol = rhsIndexIn[col];
    colIndexOut[col].index = actualCol.index;
    colIndexOut[col].value = col * lhsRows;
    colIndexOut[col+1].value = (col+1) * lhsRows;

    //perform merge
    int rowIndex = actualRow.value;
    int colIndex = actualCol.value;
    int colEnd = rhsIndexIn[col+1].value;
    float sum = 0;
    while(rowIndex < rowEnd && colIndex < colEnd) {
      entry rowEntry = lhsIn[rowIndex];
      entry colEntry = rhsIn[colIndex];
      if(rowEntry.index < colEntry.index)
        ++rowIndex;
      else if(rowEntry.index > colEntry.index)
        ++colIndex;
      else {
        sum += rowEntry.value * colEntry.value;
        ++rowIndex;
        ++colIndex;
      }
    }

    //store
    rowOut[row * rhsCols + col].index = actualCol.index;
    rowOut[row * rhsCols + col].value = sum;
    colOut[col * lhsRows + row].index = actualRow.index;
    colOut[col * lhsRows + row].value = sum;
  }
  colIndexOut[col].value = col * lhsRows; //don't forget extra entry at end
}
  
int sparseCudaMatrixMultiply(indexEntry* rowIndexOut, entry* rowOut,
                             indexEntry* colIndexOut, entry* colOut,
                             indexEntry* lhsIndexIn, entry* lhsIn,
                             indexEntry* rhsIndexIn, entry* rhsIn,
                             int lhsRows, int rhsCols) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseCudaMatrixMultiplyMain
    <<<lhsRows,BLOCK_DIM>>>(rowIndexOut, rowOut, colIndexOut, colOut,
                            lhsIndexIn, lhsIn, rhsIndexIn, rhsIn,
                            lhsRows, rhsCols);

  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseTimesDenseToDenseMain(float* matrixOut, indexEntry* rowIndexIn,
                                            entry* rowIn, float* rhsIn, int lhsRows,
                                            int lhsPRows, int m, int n, int pm, int pn) { 
  indexEntry rowIndex = rowIndexIn[blockIdx.x];
  __shared__ entry sdata[BLOCK_DIM];

  int rowEnd = rowIndexIn[blockIdx.x + 1].value;

  for (int lhsColOffset = rowIndex.value; lhsColOffset < rowEnd; lhsColOffset += BLOCK_DIM) {
    int col = lhsColOffset + threadIdx.x;

    // load row segment into sdata

    if (col < rowEnd)
      sdata[threadIdx.x] = rowIn[col];

    __syncthreads();

    for (int rhsColOffset = 0; rhsColOffset < n; rhsColOffset += BLOCK_DIM) {
      float sum = 0.0f;
      int rhsCol = threadIdx.x + rhsColOffset;

      if (rhsCol < n) {
        for (int k = 0; k < BLOCK_DIM && lhsColOffset + k < rowEnd ; k++)
          sum += sdata[k].value * rhsIn[rhsCol * pm + sdata[k].index];

        matrixOut[rhsCol * lhsPRows + rowIndex.index] += sum;
      }
    }

    __syncthreads();
  }
}

int sparseTimesDenseToDense(float* matrixOut, indexEntry* rowIndexIn, entry* rowIn,
                            float* rhsIn, int lhsRows, int lhsPRows, int m, int n, int pm, int pn) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseTimesDenseToDenseMain
    <<<lhsRows,BLOCK_DIM>>>(matrixOut, rowIndexIn, rowIn, rhsIn,
                            lhsRows, lhsPRows, m, n, pm, pn);
  
  cudaThreadSynchronize();
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void sparseTimesDenseToSparseMain(indexEntry* rowIndexOut, entry* rowOut,
                                             indexEntry* colIndexOut, entry* colOut,
                                             indexEntry* rowIndexIn, entry* rowIn,
                                             float* rhsIn, int lhsRows, int m, int n,
                                             int pm, int pn) { 
  indexEntry rowIndex = rowIndexIn[blockIdx.x];
  int row = rowIndex.index;
  int start = rowIndex.value;
  int rowLoc = n * blockIdx.x;
  rowIndexOut[blockIdx.x].index = row;
  rowIndexOut[blockIdx.x].value = rowLoc;
  rowIndexOut[blockIdx.x+1].value = rowLoc + n;
  int end = rowIndexIn[blockIdx.x+1].value;
  int per = (n - 1) / blockDim.x + 1;//DIV_ROUND_UP
  int startCol = threadIdx.x * per;
  int endCol = startCol + per;
  int col;
  for(col = startCol; col < endCol && col < n; col++) {
    int colVal = lhsRows * col;
    colIndexOut[col].index = col;
    colIndexOut[col].value = colVal;
    colIndexOut[col+1].value = colVal + lhsRows;
    float sum = 0;
    int cur;
    int colLoc = col * pn;
    for(cur = start; cur < end; cur++) {
      entry thisRow = rowIn[cur];
      sum += thisRow.value * rhsIn[colLoc + thisRow.index];
    }
    rowOut[rowLoc + col].index = col;
    rowOut[rowLoc + col].value = sum;
    colOut[colVal + blockIdx.x].index = row;
    colOut[colVal + blockIdx.x].value = sum;
  }
}

int sparseTimesDenseToSparse(indexEntry* rowIndexOut, entry* rowOut,
                             indexEntry* colIndexOut, entry* colOut,
                             indexEntry* rowIndexIn, entry* rowIn,
                             float* rhsIn, int lhsRows, int m, int n, int pm, int pn) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  sparseTimesDenseToSparseMain
    <<<lhsRows,BLOCK_DIM>>>(rowIndexOut, rowOut, colIndexOut, colOut,
                            rowIndexIn, rowIn, rhsIn, lhsRows, m, n, pm, pn);
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__global__ void cudaMatrixTimesCTSRowMain(indexEntry* indexOut, entry* matOut,
                                          float* lhsIn, float* rhsIn,
                                          indexEntry* indexIn, entry* matIn,
                                          int lm, int ln, int lpm, int lpn,
                                          int rm, int rn, int rpm, int rpn) {
  __shared__ float sdata[BLOCK_DIM];
  extern __shared__ float rowVals[];
  
  indexEntry actualRow = indexIn[blockIdx.x];
  indexOut[blockIdx.x] = indexIn[blockIdx.x];
  indexOut[blockIdx.x+1] = indexIn[blockIdx.x+1];
  int row = actualRow.index;
  int tid = threadIdx.x;
  int i;
  for(i = tid; i < ln; i += blockDim.x)
    rowVals[i] = lhsIn[i * lpm + row];

  __syncthreads();

  int start = actualRow.value;
  int end = indexIn[blockIdx.x+1].value;
  for(i = start; i < end; i++) {
    entry actualCol = matIn[i];
    int col = actualCol.index;
    int colLoc = col * rpm;
    int j;
    float t, y, c = 0;
    sdata[tid] = 0;
    for (j = threadIdx.x; j < rm; j += blockDim.x) {
      y = rowVals[j] * rhsIn[colLoc + j] - c;
      t = sdata[tid] + y;
      c = (t - sdata[tid]) - y;
      sdata[tid] = t;
    }
    
    __syncthreads();
    
    if(tid % HALF_DIM == 0) {
      c = 0;
      for (j = tid + 1; j < tid + HALF_DIM; j++) {
        y = sdata[j] - c;
        t = sdata[tid] + y;
        c = (t - sdata[tid]) - y;
        sdata[tid] = t;
      }
    }
    
    __syncthreads();
    
    // write result for this block to global mem
    if (tid == 0) {
      matOut[i].index = actualCol.index;
      matOut[i].value = sdata[0];
      for (j = HALF_DIM; j < BLOCK_DIM; j += HALF_DIM)
        matOut[i].value += sdata[j];
      matOut[i].value *= actualCol.value;
    }

    __syncthreads();
  }
}

__global__ void cudaMatrixTimesCTSColMain(indexEntry* indexOut, entry* matOut,
                                          float* lhsIn, float* rhsIn,
                                          indexEntry* indexIn, entry* matIn,
                                          int lm, int ln, int lpm, int lpn,
                                          int rm, int rn, int rpm, int rpn) {
  __shared__ float sdata[BLOCK_DIM];
  extern __shared__ float colVals[];
  
  indexEntry actualCol = indexIn[blockIdx.x];
  indexOut[blockIdx.x] = indexIn[blockIdx.x];
  indexOut[blockIdx.x+1] = indexIn[blockIdx.x+1];
  int col = actualCol.index;
  int lColLoc = col * lpm;
  int tid = threadIdx.x;
  int i;
  for(i = tid; i < lm; i += blockDim.x)
    colVals[i] = lhsIn[lColLoc + i];

  __syncthreads();

  int start = actualCol.value;
  int end = indexIn[blockIdx.x+1].value;
  for(i = start; i < end; i++) {
    entry actualRow = matIn[i];
    int row = actualRow.index;
    int j;
    float t, y, c = 0;
    sdata[tid] = 0;
    for (j = threadIdx.x; j < rn; j += blockDim.x) {
      y = colVals[j] * rhsIn[j * rpm + row] - c;
      t = sdata[tid] + y;
      c = (t - sdata[tid]) - y;
      sdata[tid] = t;
    }
    
    __syncthreads();
    
    if(tid % HALF_DIM == 0) {
      c = 0;
      for (j = tid + 1; j < tid + HALF_DIM; j++) {
        y = sdata[j] - c;
        t = sdata[tid] + y;
        c = (t - sdata[tid]) - y;
        sdata[tid] = t;
      }
    }
    
    __syncthreads();
    
    // write result for this block to global mem
    if (tid == 0) {
      matOut[i].index = actualRow.index;
      matOut[i].value = sdata[0];
      for (j = HALF_DIM; j < BLOCK_DIM; j += HALF_DIM)
        matOut[i].value += sdata[j];
      matOut[i].value *= actualRow.value;
    }

    __syncthreads();
  }
}

//the mother of all methods: 20 parameters
int cudaMatrixTimesCWiseTimesSparse(indexEntry* rowIndexOut, entry* rowOut,
                                    indexEntry* colIndexOut, entry* colOut,
                                    float* lhsIn, float* rhsIn,
                                    indexEntry* rowIndexIn, entry* rowIn,
                                    indexEntry* colIndexIn, entry* colIn,
                                    int lm, int ln, int lpm, int lpn,
                                    int rm, int rn, int rpm, int rpn,
                                    int sr, int sc) {
  cudaError_t cudaStat;
  
  cudaStat = cudaGetLastError(); // clear error status
  cudaMatrixTimesCTSRowMain
    <<<sr,BLOCK_DIM,ln*sizeof(float)>>>(rowIndexOut, rowOut, lhsIn, rhsIn,
                                        rowIndexIn, rowIn, lm, ln, lpm, lpn,
                                        rm, rn, rpm, rpn);
  cudaThreadSynchronize();
  cudaThreadSynchronize();
  cudaStat = cudaGetLastError(); // check for launch error
  if (cudaStat != cudaSuccess)
    return 0;

  cudaMatrixTimesCTSColMain
    <<<sc,BLOCK_DIM,rm*sizeof(float)>>>(colIndexOut, colOut, rhsIn, lhsIn,
                                        colIndexIn, colIn, rm, rn, rpm, rpn,
                                        lm, ln, lpm, lpn);
  cudaThreadSynchronize();
  cudaThreadSynchronize();
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}

__device__ unsigned int nextPowerOfTwo(unsigned int v) {
  v--;
  v |= v >> 1;
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;
  return v + 1;
}

__global__ void onlineOrthogonalizationMain(indexEntry *rowIndex, indexEntry *colIndex, 
					    entry *rowMatrixIn, entry *colMatrixIn, entry *rowMatrixOut, 
					    entry *colMatrixOut, int sr, int sc, int windowSize, 
					    int step, int windowsPerRow, int windowsPerCol, 
					    int outputChannels, int inputChannels) {
  extern __shared__ entry sdata[];
  entry *srow = sdata;
  entry *krow = &sdata[blockDim.x];
  float *dsums = (float *) &sdata[2*blockDim.x];

  indexEntry row = rowIndex[blockIdx.x];

  // load in row i
  srow[threadIdx.x] = rowMatrixIn[row.value + threadIdx.x];
  int col = srow[threadIdx.x].index;

  __syncthreads();

  // locate colOffsetStart
  int colOffsetStart, colOffsetEnd, colOffsetOut;

  for (int i = 0; i < sc; i++)
    if (colIndex[i].index == col)
    {
      colOffsetStart = colIndex[i].value;
      colOffsetEnd = colIndex[i+1].value;
    }

  // locate offset into colmatrix
  entry kcol = colMatrixIn[colOffsetStart];

  float sum = 0.0f;
  int nonZeroPerInput = windowSize*windowSize;
  int windowsPerImg = windowsPerRow*windowsPerCol;

  int in = blockIdx.x / (outputChannels * windowsPerImg);
  int idx = blockIdx.x  % windowsPerImg;
  int wr = idx / windowsPerRow;
  int wc = idx % windowsPerRow;

  int lb_r = wr - windowSize/step + 1;
  int lb_c = wc - windowSize/step + 1;
  int ub_r = wr + windowSize/step;
  int ub_c = wc + windowSize/step;

  if(windowSize % step) {
    lb_r -= 1;
    lb_c -= 1;
    ub_r += 1;
    ub_c += 1;  
  }

  int start_r =  max(0, lb_r);
  int start_c =  max(0, lb_c);
  int end_r = min(windowsPerRow, ub_r);
  int end_c = min(windowsPerCol, ub_c);

  int k;

  for(int out = 0; out < outputChannels; out++) {
    for (int r = start_r; r < end_r; r++) {
      for (int c = start_c; c < end_c; c++) {

        k = in*outputChannels*windowsPerImg + out*windowsPerImg + r * windowsPerRow + c;

        // load in row k
        indexEntry rowk_index = rowIndex[k];

        krow[threadIdx.x] = rowMatrixIn[rowk_index.value + threadIdx.x];
        __syncthreads();


        // compute dot product
        int krow_index;

        dsums[threadIdx.x] = 0.0f;

        for (krow_index = 0; krow_index < blockDim.x; krow_index++) {
          entry row = krow[krow_index];

          if (row.index == srow[threadIdx.x].index)
            dsums[threadIdx.x] = row.value * srow[threadIdx.x].value;
          else if (row.index > srow[threadIdx.x].index)
            break;
        }

        __syncthreads();

        // do reduce!

        int input;
        int offInput;
        for (unsigned int offset = nextPowerOfTwo(nonZeroPerInput) / 2; offset > 0; offset >>= 1) {
          input = threadIdx.x / nonZeroPerInput;
          offInput = (threadIdx.x + offset) / nonZeroPerInput; 
          if (threadIdx.x - input*nonZeroPerInput < offset && threadIdx.x + offset < blockDim.x && input == offInput)
            dsums[threadIdx.x] += dsums[threadIdx.x + offset];

          __syncthreads();
        }


        // locate Wkj

        float wkj = 0.0f;

        while(colOffsetStart + 1 < colOffsetEnd && kcol.index < rowk_index.index) {
          colOffsetStart++;
          kcol = colMatrixIn[colOffsetStart];
        }

        if (kcol.index == rowk_index.index)
          wkj = colMatrixIn[colOffsetStart].value;

        if (k == blockIdx.x)
          colOffsetOut = colOffsetStart;

        // add to sum
        int idx2 = threadIdx.x / nonZeroPerInput;
        int blockToGet = idx2 * nonZeroPerInput;
        sum += wkj * dsums[blockToGet];

        __syncthreads();
      }
    }
  }

  // write out rowMatrix && colMatrix
  entry temp;

  temp.index = srow[threadIdx.x].index;
  temp.value = sum;
  rowMatrixOut[row.value + threadIdx.x] = temp;

  temp.index = row.index;
  colMatrixOut[colOffsetOut] = temp;
}


int onlineOrthogonalization(indexEntry *rowIndex, indexEntry *colIndex, entry *rowMatrixIn, 
			    entry *colMatrixIn, entry *rowMatrixOut, entry *colMatrixOut, 
			    int sr, int sc, int maxNonZeroPerRow, int windowSize, int step, 
			    int windowsPerRow, int windowsPerCol, int outputChannels, 
			    int inputChannels) {

  cudaError_t cudaStat;

  cudaThreadSynchronize();
  cudaStat = cudaGetLastError();

  onlineOrthogonalizationMain
    <<<sr, maxNonZeroPerRow, maxNonZeroPerRow * 2 * sizeof(entry) + maxNonZeroPerRow * sizeof(float)>>>
    (rowIndex, colIndex, rowMatrixIn, colMatrixIn, rowMatrixOut, colMatrixOut, sr, sc, windowSize, 
     step, windowsPerRow, windowsPerCol, outputChannels, inputChannels);

  cudaThreadSynchronize();
  cudaStat = cudaGetLastError();
  printf("%s\n", cudaGetErrorString(cudaStat));

  return (cudaStat == cudaSuccess);
}

__global__ void sparseCudaMatrixPartialSumMain(float *matrixOut, int dim, indexEntry *index, entry *matrixIn) {
  __shared__ float sdata[BLOCK_DIM];
  __shared__ float res;

  indexEntry start = index[blockIdx.x];
  indexEntry end = index[blockIdx.x + 1];

  unsigned int cols = end.value - start.value;

  for (unsigned int offset = 0; offset < cols; offset += BLOCK_DIM) {
    unsigned int idx = threadIdx.x + offset;

    if (idx < cols)
      sdata[threadIdx.x] = matrixIn[start.value + idx].value;
    else
      sdata[threadIdx.x] = 0.0f;

    __syncthreads();

    // do reduction

    for (unsigned int o = BLOCK_DIM / 2; o > 0; o >>= 1)
    {
      if (threadIdx.x < o)
        sdata[threadIdx.x] += sdata[threadIdx.x + o];
      
      __syncthreads();
    }

    if (!threadIdx.x) {
      if (offset == 0)
        res = sdata[0];
      else
        res += sdata[0];
    }

    __syncthreads();
  }

  //write out result to matrix

  if (!threadIdx.x)
    matrixOut[start.index] = res;
}

int sparseCudaMatrixPartialSum(float *matrixOut, int pdim, int dim, indexEntry *index, entry *matrixIn) {
  cudaError_t cudaStat;

  cudaMemset(matrixOut, 0, pdim * sizeof(float));

  sparseCudaMatrixPartialSumMain<<<dim, BLOCK_DIM>>>(matrixOut, dim, index, matrixIn);

  //cudaThreadSynchronize();
  cudaStat = cudaGetLastError();

  return (cudaStat == cudaSuccess);
}

__global__ void weightsTimesDataMain(float* matrixOut, entry* rowIn, float* rhsIn, int outPm, int nonZeroPerRow, int m, int n, int pm) { 
  __shared__ entry lrow[HALF_DIM];
  __shared__ float rdata[HALF_DIM][HALF_DIM+1];
  __shared__ float res[HALF_DIM];

  int col = HALF_DIM * blockIdx.x + threadIdx.y;
  int col2 = HALF_DIM * blockIdx.x + threadIdx.x;

  if (threadIdx.y == 0)
    res[threadIdx.x] = 0.0f;

  for (int colOffset = 0; colOffset < nonZeroPerRow; colOffset += HALF_DIM) {
    // load partial row into shared memory
    int rowIndex = colOffset + threadIdx.x;

    if (threadIdx.y == 0 && rowIndex < nonZeroPerRow)
      lrow[threadIdx.x] = rowIn[blockIdx.y * nonZeroPerRow + rowIndex];

    __syncthreads();

    // load tile of values from dense matrix
    
    if (col < n && rowIndex < nonZeroPerRow)
      rdata[threadIdx.x][threadIdx.y] = rhsIn[col * pm + lrow[threadIdx.x].index]; 

    __syncthreads();

    // compute dot product

    if (threadIdx.y == 0 && col2 < n)
      for (int i = 0; i < HALF_DIM && colOffset + i < nonZeroPerRow; i++)
        res[threadIdx.x] += lrow[i].value * rdata[i][threadIdx.x];

    __syncthreads();
  }

  if (threadIdx.y == 0 && col2 < n)
    matrixOut[outPm * col2 + blockIdx.y] = res[threadIdx.x];
}

int weightsTimesData(float* matrixOut, entry* rowIn, float* rhsIn, int outPm, int lhsRows, int nonZeroPerRow, int m, int n, int pm) {
  cudaError_t cudaStat;

  cudaStat = cudaGetLastError(); // clear error status

  dim3 gridSize((n + HALF_DIM - 1) / HALF_DIM, lhsRows), blockSize(HALF_DIM, HALF_DIM);
  weightsTimesDataMain<<<gridSize, blockSize>>>(matrixOut, rowIn, rhsIn, outPm, nonZeroPerRow, m, n, pm);
  
  cudaStat = cudaGetLastError(); // check for launch error
  return (cudaStat == cudaSuccess);
}
