///Author: Ian Goodfellow
/// \file dataset.h
/// \brief base class for loading datasets
// New dataset formats should be supported by subclassing
// this class
// All data is assumed to be two-dimensional, though there's
// nothing to stop you from using only one dimension, e.g.
// for audio


#pragma once

#include "../base/matrix.h"
#include <vector>
class XMLNode;


namespace sdn
{
  template <typename T>
    class Registry;

  /// This is the main dataset class
  /** If you need to represent a new KIND of image data (e.g., color video,
     video augmented with information about types of transformation)
     then you should subclass this.     
     If you just need to post-process images differently or load a different
    file format, then you should subclass an existing subclass of 
     ImageCollection. */

  class Dataset
  {
    
  public:

    /// Constructs a Dataset with a specific dynamic range   
    Dataset(Scalar minVal = -1.0, Scalar maxVal = 1.0); //todo-- set these parameters up to work with ConfigManager framework

    /// Empty destructor
    virtual ~Dataset();
    
    /// Opens a file as XML and passes the root to the derived class
    virtual void load(const char * filepath);   

    /// Pure virtual method for loading a dataset from XML
    virtual void load(XMLNode & root) = 0;

    /// Changes to using the given dynamic range
    virtual void setDynamicRange(Scalar minVal, Scalar maxVal);
    
    /// Specify the size of examples that should be returned later
    virtual void requestExampleSize(int rows, int cols);

    /// Test if a specified size is supported
    virtual bool supportsExampleSize(unsigned rows, unsigned cols)=0;


    /// Find out what size of example the dataset does support
    virtual unsigned getPreferredRows()=0;
    virtual unsigned getPreferredCols()=0;
    
    /// Returns true if the dataset contains class labels
    virtual bool containsLabels();

    /// Returns the number of class labels
    /**
       The labels returned by other functions will be in the range
       [0,numLabels()].

       Default implementation returns false.
    */
    virtual unsigned int numLabels();

    /// Randomly chooses an example and returns it as a column
    /**
       The default implementation is to call getRandomExample
       and convert it to a column, but you may override this
       to use more efficient methods. 

       Default implementation returns zero

       \returns The associated class label 
                (valid only if containsLabels() is true)
    */
    virtual unsigned int getRandomExampleAsColumn(Matrix & x);

    /// Randomly chooses an example and returns it as a matrix
    /**
       \returns The associated class label 
               (valid only if containsLabels() is true)
    */
    virtual unsigned int getRandomExample(Matrix &x) = 0;



    // ******* Methods for invariance testing  *****************

    ///Returns true if the getRandomIndexedExample, getTrajectoryFromIndex, and trajectoryOverlap methods are supported
    virtual bool supportsTrajectoryIndexing();

    /// Like getRandomExample, but also writes out an index so this example can be found later
    virtual unsigned int getRandomIndexedExample(Matrix & x, unsigned & index);
    
    /// Returns a sequence of related patches. Relatedness will mean different things in different contexts-- it could be patches that appear at the same location in the frame, but with different time coordinates, or it could be gratings that all have the same orientation and frequency but different phase, etc. The central patch will be the patch that was returned by getRandomIndexedExample with index index
    virtual std::vector<Matrix> getTrajectoryFromIndex(unsigned index);

    virtual unsigned getExampleFromIndex(Matrix & x, unsigned index);
    
    /// Returns true if the the trajectories associated with two indices overlap
    virtual bool trajectoryOverlap(unsigned index1, unsigned index2);

    

    // ******* End of methods for invariance testing **********

  protected:
    
    Scalar _minVal;
    Scalar _maxVal;        

    int _requestedRows;
    int _requestedCols;
  };

/// Access to the ImageCollectionRegistry
  sdn::Registry<Dataset> & DatasetRegistry();

}
