/// \author Ian Goodfellow

#include "grating.h"
#include "math.h"
using namespace sdn;


static const double grating_pi = 3.14159265358979323864;

Grating::Grating(unsigned rows, unsigned cols, double minValue, double maxValue) : image_(Matrix::Zero(rows,cols))
{
  dirty_ = true;
  amplitude_ = (maxValue-minValue)/2.0;
  baseline_ = (maxValue+minValue)/2.0;
  
  frequency_ = 1;
  theta_ = 0;
  phaseShift_ = 0;

  minX_  = - grating_pi;
  width_ = 2.0 * grating_pi;
  maxY_ =  grating_pi;
  height_ = 2.0 * grating_pi;
  
  pixelWidth_ = width_ / double(cols);
  pixelHeight_ = height_ / double(rows);
  
  setSamplingRate(8,8);
}

void Grating::setSamplingRate(unsigned xSamplesPerPixel, unsigned ySamplesPerPixel)
{
  xSamplesPerPixel_ = xSamplesPerPixel;
  ySamplesPerPixel_ = ySamplesPerPixel;
  
  sampleWidth_ = pixelWidth_ / double(xSamplesPerPixel_);
  sampleHeight_ = pixelHeight_ / double(ySamplesPerPixel_);
}

Matrix Grating::getImage()
{
  if (dirty_)
    render_();
  return image_;
}

void Grating::setFrequency(double f)
{
  frequency_ = f;
  dirty_ = true;
}

void Grating::setOrientation(double t)
{
  theta_ = t;
  dirty_ = true;
}

void Grating::setPhaseShift(double p)
{
  phaseShift_ = p;
  dirty_ = true;
}

double Grating::getFrequency() const
{
  return frequency_;
}

double Grating::getOrientation() const
{
  return theta_;
}

double Grating::getPhaseShift() const
{
  return phaseShift_;
}

void Grating::render_()
{
  for (unsigned i = 0; i < image_.rows(); i++)
    {
      for (unsigned j = 0; j < image_.cols(); j++)
	{
	  drawPixel_(i,j);
	}
    }

  dirty_ = false;
}

void Grating::drawPixel_(unsigned i, unsigned j)
{
  double accum = 0;

  double y = maxY_ - double(i) * pixelHeight_ + 0.5 * sampleHeight_;
  double startX = minX_ + double(j) * pixelWidth_ + 0.5 * sampleWidth_;

  for (unsigned k = 0; k < ySamplesPerPixel_; k++)
    {
      double x = startX;
      for (unsigned l = 0; l < xSamplesPerPixel_; l ++)
	{
	  accum += eval_(x,y);

	  x+= sampleWidth_;
	}
      y += sampleHeight_;
    }

  accum /= double(xSamplesPerPixel_ * ySamplesPerPixel_);

  image_(i,j) = accum;
}

double Grating::eval_(double x, double y) const
{
  return baseline_ + amplitude_ * sin( frequency_ * ( x * cos(theta_) + y*sin(theta_) )- phaseShift_  );
}

unsigned Grating::rows()
{
  return image_.rows();
}

unsigned Grating::cols()
{
  return image_.cols();
}
