/// \author Ian Goodfellow

#include "gratingDataset.h"
#include "../base/constants.h"
#include "../base/logger.h"
#include "../base/attributeParser.h"

using namespace sdn;
using namespace std;

/// \todo-- set the dynamic range of the grating appropriately, rather than hardcoding it to [-0.8,0.8]
GratingDataset::GratingDataset() : grating_(14,14,-0.8,0.8)
{
  _minVal = -0.8;
  _maxVal = 0.8;
  phaseTrajectory_ = true;
  trajSize_ = 11;
  SDN_ASSERT_MSG(trajSize_ %2 == 1, "trajectories must be of odd length");
  trajWingSize_ = trajSize_ / 2;
  SDN_ASSERT(trajWingSize_ * 2 + 1 == trajSize_);

  fMin_ = 1.0;
  fMax_ = 4.0;

  
  phaseMin_ = 0;
  phaseMax_ = 2.0 * pi;

  thetaMin_ = 0;
  thetaMax_ = 2.0 * pi;

  phaseStep_ = 0.025;
  thetaStep_ = 0.025;

  phaseTol_ = 0.02;
  thetaTol_ = 0.02;
  fTol_ = 0.3;

}

GratingDataset::~GratingDataset()
{

}

void GratingDataset::load(XMLNode & root)
{
  SDN_ASSERT(root.getName());
  SDN_ASSERT(string(root.getName()) == "sdn::Dataset");
  SDN_ASSERT(root.getAttribute("id"));
  SDN_ASSERT(string(root.getAttribute("id")) == "sdn::GratingDataset");

  AttributeParser p;
  p.addBool("phaseTrajectory");
  p.addString("id");
  p.parse(root);
  phaseTrajectory_ = p.getBool("phaseTrajectory").value;
}

void GratingDataset::setDynamicRange(Scalar minVal, Scalar maxVal)
{
  SDN_FATAL("Not implemented yet");
}

void GratingDataset::requestExampleSize(int rows, int cols)
{
  Dataset::requestExampleSize(rows,cols);

  grating_ = Grating(_requestedRows, _requestedCols, _minVal, _maxVal);
}

bool GratingDataset::supportsExampleSize(unsigned rows, unsigned cols)
{
  return (rows * cols > 0);
}

unsigned GratingDataset::getPreferredRows()
{
  if (_requestedRows  > 0)
    {
      return _requestedRows;
    }

  return 14;
}

unsigned GratingDataset::getPreferredCols()
{
  if (_requestedCols > 0)
    {
      return _requestedCols;
    }

  return 14;
}

unsigned GratingDataset::getRandomExample(Matrix & x)
{
  double p, f, t;
  getRandomExample(x,p,f,t);
  return 0;
}

bool GratingDataset::supportsTrajectoryIndexing()
{
  return true;
}

void GratingDataset::getRandomExample(Matrix & x, double & phase, double & freq, double & theta)
{
  SDN_ASSERT_MSG( _requestedRows > 0 && _requestedCols > 0 , "Calling getRandomIndexedExample without first calling requestExampleSize");
  SDN_ASSERT(x.rows() == _requestedRows);
  SDN_ASSERT(x.cols() == _requestedCols);

  phase = rrg_.randUnif(phaseMin_,phaseMax_);
  freq = rrg_.randUnif(fMin_,fMax_);
  theta = rrg_.randUnif(thetaMin_, thetaMax_);

  grating_.setPhaseShift(phase);
  grating_.setFrequency(freq);
  grating_.setOrientation(theta);

  x = grating_.getImage();
    
}

unsigned GratingDataset::getRandomIndexedExample(Matrix &x, unsigned & index)
{
  IndexedExample e;
  getRandomExample(x, e.phase, e.f, e.theta);
  index = indexedExamples_.size();
  indexedExamples_.push_back(e);
  return 0;
}


vector<Matrix> GratingDataset::getTrajectoryFromIndex(unsigned index)
{
  IndexedExample e = indexedExamples_[index];

  if (phaseTrajectory_)
    {
      e.phase -= double(trajWingSize_)*phaseStep_;
    }
  else
    {
      e.theta -= double(trajWingSize_)*thetaStep_;
    }

  vector<Matrix> rval;

  for (unsigned i = 0; i < trajSize_; i++)
    {
        grating_.setFrequency(e.f);
	grating_.setPhaseShift(e.phase);
	grating_.setOrientation(e.theta);
      
	rval.push_back(grating_.getImage());

	if (phaseTrajectory_)
	  {
	    e.phase +=phaseStep_;
	  }
	else
	  {
	    e.theta += thetaStep_;
	  }
    }
  
  return rval;
}

unsigned GratingDataset::getExampleFromIndex(Matrix & x, unsigned index)
{
  IndexedExample e = indexedExamples_[index];

  grating_.setFrequency(e.f);
  grating_.setPhaseShift(e.phase);
  grating_.setOrientation(e.theta);
  
  x = grating_.getImage();

  return 0;
}

static double standardize(double p)
{
  double rval = fmod(p,2.0*pi);
  if (rval < 0)
    {
      rval += 2.0 * pi;
    }
  return rval;
}

static double cyclicDistance(double p1, double p2)
{
  p1 = standardize(p1);
  p2 = standardize(p2);

  double v1 = abs(p1-p2);
  double v2 = abs(p1 - 2.0 * pi - p2);
  double v3 = abs(p1-(p2-2.0*pi));

  return min(min(v1,v2),v3);
}

bool GratingDataset::trajectoryOverlap(unsigned index1, unsigned index2)
{
  IndexedExample e1 = indexedExamples_[index1];
  IndexedExample e2 = indexedExamples_[index2];

  if (abs(e1.f - e2.f) > fTol_)
    {
      return false;
    }

  if (phaseTrajectory_)
    {
      if (abs(e1.theta-e2.theta) > thetaTol_)
	{
	  return false;
	}
      
      if (cyclicDistance(e1.phase,e2.phase) > double(trajWingSize_)*phaseStep_)
	{
	  return false;
	}
    }
  else
    {
      if (abs(e1.phase-e2.phase) > phaseTol_)
	{
	  return false;
	}

      if (cyclicDistance(e1.theta, e2.theta) > double(trajWingSize_)*thetaStep_)
	{
	  return false;
	}
    }
  return true;
}
