/// \author Ian Goodfellow
/// \file gratingDataset.h
/// \brief A Dataset that synthetically generates examples using a Grating

#pragma once

#include "dataset.h"
#include "grating.h"
#include "../base/random.h"

namespace sdn
{

 
/// A Dataset that synthetically generates examples using a Grating
  class GratingDataset : public Dataset
  {
    
  public:

    /// Constructs a Dataset with a specific dynamic range   
    GratingDataset();

    /// Empty destructor
    virtual ~GratingDataset();
    
    /// Pure virtual method for loading a dataset from XML
    void load(XMLNode & root);

    /// Changes to using the given dynamic range
    void setDynamicRange(Scalar minVal, Scalar maxVal);
    
    /// Specify the size of examples that should be returned later
    void requestExampleSize(int rows, int cols);

    /// Test if a specified size is supported
    bool supportsExampleSize(unsigned rows, unsigned cols);


    /// Find out what size of example the dataset does support
    unsigned getPreferredRows();
    unsigned getPreferredCols();
    
    /// Randomly chooses an example and returns it as a matrix
    /**
       \returns The associated class label 
               (valid only if containsLabels() is true)
    */
    unsigned int getRandomExample(Matrix &x);



    // ******* Methods for invariance testing  *****************

    ///Returns true if the getRandomIndexedExample, getTrajectoryFromIndex, and trajectoryOverlap methods are supported
    bool supportsTrajectoryIndexing();

    /// Like getRandomExample, but also writes out an index so this example can be found later
    unsigned int getRandomIndexedExample(Matrix & x, unsigned & index);
    
    /// Returns a sequence of related patches. Relatedness will mean different things in different contexts-- it could be patches that appear at the same location in the frame, but with different time coordinates, or it could be gratings that all have the same orientation and frequency but different phase, etc. The central patch will be the patch that was returned by getRandomIndexedExample with index index
    std::vector<Matrix> getTrajectoryFromIndex(unsigned index);

    unsigned getExampleFromIndex(Matrix & x, unsigned index);
    
    /// Returns true if the the trajectories associated with two indices overlap
    bool trajectoryOverlap(unsigned index1, unsigned index2);

    

    // ******* End of methods for invariance testing **********

  protected:

    void getRandomExample(Matrix & x, double & phase, double & freq, double & theta);

  protected:
    
    Grating grating_;
    bool phaseTrajectory_;
    unsigned trajSize_;
    unsigned trajWingSize_;
    double fMin_;
    double fMax_;
    double phaseMin_;
    double phaseMax_;
    double thetaMin_;
    double thetaMax_;
    double phaseStep_;
    double thetaStep_;


    struct IndexedExample
    {
      double f;
      double theta;
      double phase;
    };

    std::vector<IndexedExample> indexedExamples_;

    /** two orientation-varying trajectories are considered to overlap if their 
	orientation ranges have a non-empty intersection and their phases are
	within phaseTol_ of each other
    */
    double phaseTol_;
    double thetaTol_;

    //frequence can distinguish either phase or theta trajectories
    double fTol_;

    RandomRealGenerator<double> rrg_; 
  };


}
