// Author: Quoc V. Le
// \file gratingTestDataset.cc
// \brief an grating dataset generator
#include "sdn/base.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include <vector>
#include "gratingTestDataset.h"
#include "grating.h"

USING_PART_OF_NAMESPACE_EIGEN // import most common Eigen types 

using namespace std; 
using namespace sdn;




void GratingTestDataset::generateExhaustivePatchesOnce(){
  /* generating detailed patches right at the beginning */
  vector<Scalar> orientations;
  Scalar minOrientation = MIN_ORIENTATION_EXHAUSTIVE;
  Scalar maxOrientation = MAX_ORIENTATION_EXHAUSTIVE;
  Scalar numOrientation = NUM_ORIENTATION_EXHAUSTIVE;
  Scalar deltaOrientation = (maxOrientation - minOrientation) / numOrientation;
  for (unsigned int i=0;i<=numOrientation;i++){
    orientations.push_back(minOrientation + i*deltaOrientation);
  }

  vector<Scalar> frequencies;
  Scalar minFrequency = MIN_FREQUENCY_EXHAUSTIVE;
  Scalar maxFrequency = MAX_FREQUENCY_EXHAUSTIVE;
  Scalar numFrequency = NUM_FREQUENCY_EXHAUSTIVE;
  Scalar deltaFrequency = (maxFrequency - minFrequency) / numFrequency;
  for (unsigned int i=0;i<=numFrequency;i++){
    frequencies.push_back(minFrequency + i*deltaFrequency);
  }

  vector<Scalar> phases;
  Scalar minPhase = MIN_PHASE_EXHAUSTIVE;
  Scalar maxPhase = MAX_PHASE_EXHAUSTIVE;
  Scalar numPhase = NUM_PHASE_EXHAUSTIVE;
  Scalar deltaPhase = (maxPhase - minPhase) / numPhase;
  for (unsigned int i=0;i<=numPhase;i++){
    phases.push_back(minPhase + i*deltaPhase);
  }
  

  for (unsigned int j=0;j<frequencies.size();j++){
    for (unsigned int i=0;i<phases.size(); i++){
      for (unsigned int k=0;k<orientations.size();k++){
	gratingGenerator->setPhaseShift(phases[i]);
	gratingGenerator->setOrientation(orientations[k]);
	gratingGenerator->setFrequency(frequencies[j]);
	
	Matrix m = gratingGenerator->getImage();
	normalizePatch(m);
	exhaustivePatches_.push_back(m);
      }
    }
  }  
}

void GratingTestDataset::generateInexhaustivePatchesOnce(){
  /* generating detailed patches right at the beginning */
  vector<Scalar> orientations;
  Scalar minOrientation = MIN_ORIENTATION_INEXHAUSTIVE;
  Scalar maxOrientation = MAX_ORIENTATION_INEXHAUSTIVE;
  Scalar numOrientation = NUM_ORIENTATION_INEXHAUSTIVE;
  Scalar deltaOrientation = (maxOrientation - minOrientation) / numOrientation;
  for (unsigned int i=0;i<=numOrientation;i++){
    orientations.push_back(minOrientation + i*deltaOrientation);
  }

  vector<Scalar> frequencies;
  Scalar minFrequency = MIN_FREQUENCY_INEXHAUSTIVE;
  Scalar maxFrequency = MAX_FREQUENCY_INEXHAUSTIVE;
  Scalar numFrequency = NUM_FREQUENCY_INEXHAUSTIVE;
  Scalar deltaFrequency = (maxFrequency - minFrequency) / numFrequency;
  for (unsigned int i=0;i<=numFrequency;i++){
    frequencies.push_back(minFrequency + i*deltaFrequency);
  }

  vector<Scalar> phases;
  Scalar minPhase = MIN_PHASE_INEXHAUSTIVE;
  Scalar maxPhase = MAX_PHASE_INEXHAUSTIVE;
  Scalar numPhase = NUM_PHASE_INEXHAUSTIVE;
  Scalar deltaPhase = (maxPhase - minPhase) / numPhase;
  for (unsigned int i=0;i<=numPhase;i++){
    phases.push_back(minPhase + i*deltaPhase);
  }
  

  for (unsigned int j=0;j<frequencies.size();j++){
    for (unsigned int i=0;i<phases.size(); i++){
      for (unsigned int k=0;k<orientations.size();k++){
	gratingGenerator->setPhaseShift(phases[i]);
	gratingGenerator->setOrientation(orientations[k]);
	gratingGenerator->setFrequency(frequencies[j]);
	
	Matrix m = gratingGenerator->getImage();
	normalizePatch(m);
	inexhaustivePatches_.push_back(m);
	inexhaustiveFrequencies_.push_back(frequencies[j]);
	inexhaustivePhases_.push_back(phases[i]);
	inexhaustiveOrientations_.push_back(orientations[k]);
      }
    }
  }

}

GratingTestDataset::GratingTestDataset(int patchWidth, int patchHeight)

{
  int seed = 10;
  g_.setSeed(seed);

  gratingGenerator = new Grating(patchWidth, patchHeight, -PATCH_RANGE, PATCH_RANGE);
  
  generateExhaustivePatchesOnce();

  generateInexhaustivePatchesOnce();
  
  categories_.push_back(TRANSLATION);
  types_.push_back("translation");

  categories_.push_back(ROTATION);
  types_.push_back("rotation");
  
  categories_.push_back(FREQUENCY);
  types_.push_back("frequency");
}



void GratingTestDataset::getExhaustivePatches(vector<Matrix> &allPatches)
{
  
  allPatches = exhaustivePatches_;
}


void GratingTestDataset::getLocalPatches(int type, 
					 int index, 
					 vector<Matrix> &allPatches)
{
  allPatches.clear();
  Scalar phase = inexhaustivePhases_[index];
  Scalar orientation = inexhaustiveOrientations_[index];
  Scalar frequency = inexhaustiveFrequencies_[index];
  
  SDN_ASSERT(type == TRANSLATION || type == ROTATION || type == FREQUENCY);
  /* translation: vary the phases */
  if (type == TRANSLATION){
    vector<Scalar> phases;
    Scalar minPhase = MIN_PHASE_EXHAUSTIVE;
    Scalar maxPhase = MAX_PHASE_EXHAUSTIVE;
    Scalar numPhase = NUM_PHASE_EXHAUSTIVE;
    Scalar deltaPhase = (maxPhase - minPhase) / numPhase;
    for (unsigned int i=0;i<=numPhase;i++){
    //for (int i=-2;i<=2;i++){
      phases.push_back(phase + i*deltaPhase);
    }
  

    for (unsigned int i=0;i<phases.size(); i++){
      gratingGenerator->setPhaseShift(phases[i]);
      gratingGenerator->setOrientation(orientation);
      gratingGenerator->setFrequency(frequency);
	
      Matrix m = gratingGenerator->getImage();
      normalizePatch(m);
      allPatches.push_back(m);
    }
  }
  
  /* orientation: vary the orientations */
  if (type == ROTATION){
    vector<Scalar> orientations;
    Scalar minOrientation = MIN_ORIENTATION_EXHAUSTIVE;
    Scalar maxOrientation = MAX_ORIENTATION_EXHAUSTIVE;
    Scalar numOrientation = NUM_ORIENTATION_EXHAUSTIVE;
    Scalar deltaOrientation = (maxOrientation - minOrientation) / numOrientation;
    for (unsigned int i=0;i<=numOrientation;i++){
      //for (unsigned int i=-2;i<=2;i++){
      orientations.push_back(orientation + i*deltaOrientation);
    }
  

    for (unsigned int i=0;i<orientations.size(); i++){
      gratingGenerator->setPhaseShift(phase);
      gratingGenerator->setOrientation(orientations[i]);
      gratingGenerator->setFrequency(frequency);
      
      Matrix m = gratingGenerator->getImage();
      normalizePatch(m);
      allPatches.push_back(m);
    }
  }

/* frequences: vary the frequencies */
  if (type == FREQUENCY){
    vector<Scalar> frequencies;
    Scalar minFrequency = MIN_FREQUENCY_EXHAUSTIVE;
    Scalar maxFrequency = MAX_FREQUENCY_EXHAUSTIVE;
    Scalar numFrequency = NUM_FREQUENCY_EXHAUSTIVE;
    Scalar deltaFrequency = (maxFrequency - minFrequency) / numFrequency;
    for (unsigned int i=0;i<=numFrequency;i++){
      frequencies.push_back(frequency + i*deltaFrequency);
    }
    
    for (unsigned int i=0;i<frequencies.size(); i++){
      gratingGenerator->setPhaseShift(phase);
      gratingGenerator->setOrientation(orientation);
      gratingGenerator->setFrequency(frequencies[i]);
      
      Matrix m = gratingGenerator->getImage();
      normalizePatch(m);
      allPatches.push_back(m);
    }
  }
}

void GratingTestDataset::getInexhaustivePatches(int inv_type, vector<Matrix> &allPatches)
{
  allPatches = inexhaustivePatches_;
  /*phasesOut = inexhaustivePhases_;
    orientationsOut = inexhaustiveOrientations_;
    frequenciesOut = inexhaustiveFrequencies_;*/
}


void GratingTestDataset::getPatch(int type, 
				  Scalar phase, 
				  Scalar orientation,
				  Scalar frequency,
				  Matrix &patch)
{
  SDN_ASSERT(type == TRANSLATION || type == ROTATION || type == FREQUENCY);
  
  gratingGenerator->setPhaseShift(phase);
  gratingGenerator->setOrientation(orientation);
  gratingGenerator->setFrequency(frequency);
  
  Matrix m = gratingGenerator->getImage();
  normalizePatch(m);
  patch = m;
}

void GratingTestDataset::getRandomPatch(Matrix &patch)
{
  int i = g_.randUnif(0, exhaustivePatches_.size()-1); 
  //cout << "i = " << i << endl;
  patch = exhaustivePatches_[i];
  //normalizePatch(patch);
}



