// Author: Quoc V. Le
// \file gratingTestDataset.h
// \brief an grating dataset generator
#pragma once
#include <vector>
#include <stdlib.h>
#include "grating.h"
#include "../base/random.h"
#include "invarianceTestDataset.h"

#define PATCH_RANGE 0.8
#define TRANSLATION 0
#define ROTATION 1
#define FREQUENCY 2
#define NUM_GRATING_COLLECTION 3
#define NUM_GRATING_CATEGORY 3

#define MIN_ORIENTATION_EXHAUSTIVE 0
#define MAX_ORIENTATION_EXHAUSTIVE 3.14159265358979323846264338
#define NUM_ORIENTATION_EXHAUSTIVE 40

#define MIN_FREQUENCY_EXHAUSTIVE 2
#define MAX_FREQUENCY_EXHAUSTIVE 8
#define NUM_FREQUENCY_EXHAUSTIVE 4

#define MIN_PHASE_EXHAUSTIVE 0
#define MAX_PHASE_EXHAUSTIVE 2*3.14159265358979323846264338
#define NUM_PHASE_EXHAUSTIVE 20

#define MIN_ORIENTATION_INEXHAUSTIVE 0
#define MAX_ORIENTATION_INEXHAUSTIVE 3.14159265358979323846264338
#define NUM_ORIENTATION_INEXHAUSTIVE 20

#define MIN_FREQUENCY_INEXHAUSTIVE 2
#define MAX_FREQUENCY_INEXHAUSTIVE 4
#define NUM_FREQUENCY_INEXHAUSTIVE 2

#define MIN_PHASE_INEXHAUSTIVE 0
#define MAX_PHASE_INEXHAUSTIVE 2*3.14159265358979323846264338
#define NUM_PHASE_INEXHAUSTIVE 20



namespace sdn
{
  
  /* This is the Grating test dataset 
     This dataset can be used as follows
     step 1: getInexhaustivePatches
     step 2: run through all these patches to find activated patches
     step 3: use inexhaustiveOrientations_, inexhaustivePhases_, inexhaustiveFrequencies_, to keep track of orientations, phases, and frequencies
     step 4: getLocallyVaryingPatches
     step 5: run through these patches to see if the hidden unit achieves invariance
     step 6: getExhaustivePatches
     step 7: run through these patches to see how active is the hidden unit in global scope
  */
  class GratingTestDataset: public InvarianceTestDataset
  {
  protected: 
    Grating *gratingGenerator;
    
    /// Random generator
    RandomIntGenerator<> g_;
    
    std::vector<int> categories_;
    std::vector<std::string> types_;
    
    std::vector<Matrix> exhaustivePatches_; // contains more detailed ranges of variations
    std::vector<Matrix> inexhaustivePatches_; // contains less detailed ranges of variations
    std::vector<Scalar> inexhaustiveOrientations_; // use this to keep track of orientations
    std::vector<Scalar> inexhaustivePhases_; // use this to keep track of phases
    std::vector<Scalar> inexhaustiveFrequencies_; // use this to keep track of frequencies
    // generate exhaustive patches once at the beginning
    void generateExhaustivePatchesOnce();
    // generate inexhaustive patches once at the beginning
    void generateInexhaustivePatchesOnce();
  public: 
    // constructor
    GratingTestDataset(int patchWidth, int patchHeight);

    // get a random patch
    void getRandomPatch(Matrix &patch);
    
      

    /// \todo change type to be an enum
    void getPatch(int type, 
		  Scalar phase, 
		  Scalar orientation,
		  Scalar frequency,
		  Matrix &patch);
    

    // get exhaustive, detailed patches (a huge number of patches)
    void getExhaustivePatches(std::vector<Matrix> &allPatches);
    
    // generate locally varying patches by varying one parameter (either phase, orientation, frequency)
    // if type = TRANSLATION: vary only phases
    // if type = ROTATION: vary only orientations
    // if type = FREQUENCY: vary only frequencies
    /// \todo change type to be an enum
    void getLocalPatches(int type, 
			 int index,
			 std::vector<Matrix> &allPatches);
    
    // get inexhaustive patches (a small number of patches)
    void getInexhaustivePatches(int type, std::vector<Matrix> &allPatches);

    /// Returns number of video sequences in the collection
    int getNumCollection(){ return NUM_GRATING_COLLECTION; }
    
    int getCategory(int i){
      return categories_[i];
    }
    int getNumCategory(){
      return NUM_GRATING_CATEGORY;
    }
    std::string getType(int i){
      return types_[i];
    }
    virtual ~GratingTestDataset(){}
  };
}


