// Author: Quoc V. Le
// \file naturalTestDataset.cc
// \brief an natural dataset reader
#include "sdn/base.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include "highgui.h"
#include "cv.h"
#include "cxcore.h"
#include "naturalTestDataset.h"
#include <stdlib.h>



using namespace std; 
using namespace sdn;



NaturalTestDataset::NaturalTestDataset(int patchWidth, int patchHeight)

{

	patchWidth_ = patchWidth;
	patchHeight_ = patchHeight;

	double scaleFactor = SCALE_FACTOR;

	int seed = 10;
	g_.setSeed(seed);
	_startFrame = START_FRAME;
	_endFrame = END_FRAME;

	types_.push_back("illum_pan_tripod_flowers");       categories_.push_back(ILLUMINATION);
	types_.push_back("illum_tripod_leaves");            categories_.push_back(ILLUMINATION);

	types_.push_back("still_handheld_flowers");         categories_.push_back(STILL);
	types_.push_back("still_tripod_flowers");           categories_.push_back(STILL);
	types_.push_back("still_tripod_leaves");            categories_.push_back(STILL);
	types_.push_back("still_tripod_palm");              categories_.push_back(STILL);
	types_.push_back("still_tripod_trunk");             categories_.push_back(STILL);

	types_.push_back("pan_tripod_flowers");             categories_.push_back(PAN);
	types_.push_back("pan_tripod_leaves");              categories_.push_back(PAN);

	types_.push_back("slow_tilt_tripod_flowers");       categories_.push_back(TILT);
	types_.push_back("tilt_tripod_flowers");            categories_.push_back(TILT);
	types_.push_back("tilt_tripod_palm");               categories_.push_back(TILT);
	types_.push_back("tilt_tripod_trunk");              categories_.push_back(TILT);

	types_.push_back("zoom_tripod_flowers");            categories_.push_back(ZOOM);
	types_.push_back("zoom_tripod_palm");               categories_.push_back(ZOOM);
	types_.push_back("zoom_tripod_tree");               categories_.push_back(ZOOM);

	types_.push_back("translate_handheld_leaves");      categories_.push_back(TRANSLATE);
	types_.push_back("translate_tripod_trunk");         categories_.push_back(TRANSLATE);
	types_.push_back("translate_handheld_flowers");     categories_.push_back(TRANSLATE);
	types_.push_back("wind_motion_tripod_flowers");     categories_.push_back(TRANSLATE);

	types_.push_back("rotate_handheld_flowers");        categories_.push_back(ROTATE);
	types_.push_back("rotate_handheld_palm");           categories_.push_back(ROTATE);
	types_.push_back("rotate_off_axis_tripod_flowers"); categories_.push_back(ROTATE);
	types_.push_back("rotate_off_axis_tripod_leaves");  categories_.push_back(ROTATE);
	types_.push_back("rotate_off_axis_tripod_palm");    categories_.push_back(ROTATE);

	types_.push_back("orbit_handheld_trunk");           categories_.push_back(ORBIT);
	types_.push_back("orbit_turntable_mouse");          categories_.push_back(ORBIT);
	types_.push_back("orbit_turntable_mug1");           categories_.push_back(ORBIT);
	types_.push_back("orbit_turntable_mug2");           categories_.push_back(ORBIT);
	types_.push_back("orbit_turntable_checker");        categories_.push_back(ORBIT);
	types_.push_back("orbit_turntable_leaves");         categories_.push_back(ORBIT);


	//types_.push_back("squirrel");

	char filename[200];
	for (unsigned int t=0;t<types_.size();t++){

		vector<IplImage* > this_type_images;
		for (int i=_startFrame;i<_endFrame;i++){

			sprintf(filename, "/afs/cs.stanford.edu/group/brain/scratch6/zhenghao/AndrewSaxe2/%s/whitened%08d.png", types_[t].c_str(), i); 
			IplImage* src = cvLoadImage(filename, 1);

			int new_width = (int)(src->width*scaleFactor); //320
			int new_height = (int)(src->height*scaleFactor); //180

			IplImage* dst = cvCreateImage(cvSize(new_width,new_height), src->depth, src->nChannels);

			cvResize(src, dst);
			cvReleaseImage(&src);
			this_type_images.push_back(dst);
		}
		all_images.push_back(this_type_images);
	}

	width_ = all_images[0][0]->width;
	height_ = all_images[0][0]->height;
	numImagePerCollection_ = all_images[0].size();
	numCollection_ = all_images.size();
	testStartFrame_ = 0;
}


Scalar NaturalTestDataset::getPixelValue(IplImage* frame, int x, int y){
	return (double)CV_IMAGE_ELEM(frame, uchar, x, y) / 255.0 * 2 * PATCH_RANGE - PATCH_RANGE; // TANH
}

void NaturalTestDataset::getExhaustivePatches(vector<Matrix> &allPatches)
{
	if (_randomSampledPatches.empty()){

		allPatches.clear();
		for (int i=0;i<NUM_RANDOM_PATCHES;i++){
			Matrix patch(patchWidth_, patchHeight_);
			randomPatchTesting(patch, true);
			Matrix patchCV(patchWidth_*patchHeight_,1);
			matrixToMatrixCol(patch,patchCV);
			allPatches.push_back(patchCV);
		}
		_randomSampledPatches = allPatches;  
	}
	else{
		allPatches = _randomSampledPatches;
	}
}

int NaturalTestDataset::getMiddleIndex(){
	return numImagePerCollection_/2;
}



void NaturalTestDataset::getPatch(int type, int index, Matrix &patch)
{
	int i = _rowIndices[index];
	int j = _colIndices[index];
	int offset = _offsets[index];

	SDN_ASSERT(type >= 0 && type <= numCollection_ - 1);
	IplImage *frame = all_images[type][getMiddleIndex()+offset]; 
	patch = Matrix::Zero(patchWidth_, patchHeight_);
	for (int idxI = 0; idxI < patchWidth_; idxI++){
		for (int idxJ = 0; idxJ < patchHeight_; idxJ++) {
			patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
		}
	}
	normalizePatch(patch);

}

void NaturalTestDataset::getInexhaustivePatches(int type, 
		vector<Matrix> &allPatches)
{
	_rowIndices.clear();
	_colIndices.clear();
	_offsets.clear();
	SDN_ASSERT(type >= 0 && type <= numCollection_ - 1);
	for (int offset=MIDDLE_FRAME_MIN_OFFSET;offset<=MIDDLE_FRAME_MAX_OFFSET;offset+=MIDDLE_FRAME_STEP_OFFSET){
		IplImage *frame = all_images[type][getMiddleIndex()+offset]; 
		for (int i=RECTANGLE_X;i<RECTANGLE_X+RECTANGLE_WIDTH-patchWidth_;i+=STEP){
			for (int j=RECTANGLE_Y;j<RECTANGLE_Y+RECTANGLE_HEIGHT-patchHeight_;j+=STEP){
				Matrix patch(patchWidth_, patchHeight_);
				for (int idxI = 0; idxI < patchWidth_; idxI++){
					for (int idxJ = 0; idxJ < patchHeight_; idxJ++) {
						patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
					}
				}
				normalizePatch(patch);
				Matrix patchCV(patchWidth_*patchHeight_,1);
				matrixToMatrixCol(patch,patchCV);
				allPatches.push_back(patchCV);
				_offsets.push_back(offset);
				_rowIndices.push_back(i);
				_colIndices.push_back(j);
			}
		}
	}
}

void NaturalTestDataset::getLocalPatches(int type, int index, vector<Matrix> &allPatches)
{
	int i = _rowIndices[index];
	int j = _colIndices[index];
	int offset = _offsets[index];
	allPatches.clear();
	SDN_ASSERT(type >= 0 && type <= numCollection_ - 1);
	for (int f=-TRAJECTORY_RANGE;f<=TRAJECTORY_RANGE;f++){
		//cout << "f = " << f;
		IplImage *frame = all_images[type][getMiddleIndex()+offset+f]; 
		Matrix patch(patchWidth_, patchHeight_);
		for (int idxI = 0; idxI < patchWidth_; idxI++)
			for (int idxJ = 0; idxJ < patchHeight_; idxJ++) {
				patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
			}
		normalizePatch(patch);
		Matrix patchCV(patchWidth_*patchHeight_,1);
		matrixToMatrixCol(patch,patchCV);
		allPatches.push_back(patchCV);
	}
	//cout << endl;
}

void NaturalTestDataset::getLocalTrajectory(int type, int index, vector<Matrix> &allPatches)
{
	int i = _rowIndices[index];
	int j = _colIndices[index];
	int offset = _offsets[index];
	allPatches.clear();
	SDN_ASSERT(type >= 0 && type <= numCollection_ - 1);
	for (int f=_startFrame;f<_endFrame;f++){
		//cout << "f = " << f;
		IplImage *frame = all_images[type][f]; 
		Matrix patch(patchWidth_, patchHeight_);
		for (int idxI = 0; idxI < patchWidth_; idxI++)
			for (int idxJ = 0; idxJ < patchHeight_; idxJ++) {
				patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
			}
		normalizePatch(patch);
		Matrix patchCV(patchWidth_*patchHeight_,1);
		matrixToMatrixCol(patch,patchCV);
		allPatches.push_back(patchCV);
	}
	//cout << endl;
}


void NaturalTestDataset::getRandomPatch(Matrix &patch)
{
	int i =  g_.randUnif(0, _randomSampledPatches.size()-1); 
	patch = _randomSampledPatches[i];
}


void NaturalTestDataset::randomPatchTesting(Matrix &patch, bool normalizePatchValues)
{
	SDN_ASSERT(patch.cols() == patchWidth_ && patch.rows() == patchHeight_);

	int typeIndex = g_.randUnif(0, numCollection_-1); 
	int frameIndex = g_.randUnif(0,numImagePerCollection_-1); 
	IplImage *frame = all_images[typeIndex][frameIndex]; 
	int i = g_.randUnif(0, width_-patchWidth_);  
	int j = g_.randUnif(0, height_-patchHeight_);
	for (int idxI = 0; idxI < patchWidth_; idxI++)
		for (int idxJ = 0; idxJ < patchHeight_; idxJ++)
			//patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
			patch(idxJ,idxI) = getPixelValue(frame, j+idxJ, i+idxI);
	if (normalizePatchValues)
		normalizePatch(patch); 
	return; 
}

