// Author: Quoc V. Le
// \file naturalTestDataset.cc
// \brief an Natural dataset reader
#pragma once


#include <vector>
#include "../base/random.h"
#include "cv.h"
#include "invarianceTestDataset.h"

#define NUM_RANDOM_PATCHES 50000
#define PATCH_RANGE 0.8
#define STEP 5
#define RECTANGLE_X 120
#define RECTANGLE_Y 60
#define RECTANGLE_WIDTH 80
#define RECTANGLE_HEIGHT 70 
#define START_FRAME 0
#define END_FRAME 100
#define MIDDLE_FRAME_MIN_OFFSET -20
#define MIDDLE_FRAME_MAX_OFFSET 20
#define MIDDLE_FRAME_STEP_OFFSET 5
#define TRAJECTORY_RANGE 5
#define SCALE_FACTOR 0.5

#define ILLUMINATION 0
#define STILL 1
#define PAN 2
#define TILT 3
#define ZOOM 4
#define TRANSLATE 5
#define ROTATE 6
#define ORBIT 7

#define NUM_NATURAL_CATEGORY 8

namespace sdn
{

  /// todo-- this should be renamed something like VideoCollection and made a subclass of Dataset. 
  
  // This is the Natural test dataset 
  class NaturalTestDataset: public InvarianceTestDataset{
    
  protected: 
    std::vector<int> _rowIndices;
    std::vector<int> _colIndices;
    std::vector<int> _offsets;
    
    // sampled random patches
    std::vector<Matrix> _randomSampledPatches;
    
    // names of test 
    std::vector<std::string> types_;
    std::vector<int> categories_;
      
    /// Width of each image
    int width_;
    
    /// Height of each image
    int height_;
    /// Width of each patch
    int patchWidth_;
    
    /// Height of each patch
    int patchHeight_;
    
    /// Index of the first image that we should load
    int _startFrame;
    
    /// Index of the last image that we should load
    int _endFrame;
    
    /// Storage of all images in the dataset
    std::vector< std::vector< IplImage* > > all_images;
    
    /// Number of images per collection 
    int numImagePerCollection_;
    
    /// Number of test collections (video sequences) that we use
    int numCollection_;

    /// Starting frame to the test data
    int testStartFrame_;

    /// Random generator

    RandomIntGenerator<> g_;
    
    /// Returns the index of the middle frame
    int getMiddleIndex();
    
  public: 
    
    NaturalTestDataset(int patchWidth, int patchHeight);
    
    /// Returns the width of images in the collection
    int imageWidth() { return width_; }
    
    /// Returns the height of images in the collection
    int imageHeight() { return height_; } 
    
    /// Returns number of video sequences in the collection
    int getNumCollection(){ return numCollection_; }
    
    
    void randomPatchTesting(Matrix &patch, bool normalizePatchValues);
      
    /// Returns a random patch
    void getRandomPatch(Matrix &patch);
    
    /// Returns patch at location i, j in the middle frame
    void getPatch(int type, int index, Matrix &patch);
    

    /// Get a pixel value from frame indexed by x and y
    Scalar getPixelValue(IplImage* frame, int x, int y);
    
    std::string getType(int i){
      return types_[i];
    }
    int getCategory(int i){
      return categories_[i];
    }
    int getNumCategory(){
      return NUM_NATURAL_CATEGORY;
    }


    
    /// Returns random patches. The number of patches is equal to number.
    void getExhaustivePatches(std::vector<Matrix> &allPatches);
    

    /// Returns all patches in a trajectory for a given type. Here, i,
    /// j are the row and column index of the place in an image where
    /// we want to extract a trajectory and type is the type of video
    /// sequence
    void getLocalPatches(int type, int index,
			 std::vector<Matrix> &allPatches);
    
    void getLocalTrajectory(int type, int index,
			 std::vector<Matrix> &allPatches);
    

    /// Returns all patches from the middle frame, we can use this to
    /// compute the optimal stimulus. rowIndices and columnIndices
    /// store the locations wehere each patch in allPatches are
    /// located.
    void getInexhaustivePatches(int type, std::vector<Matrix> &allPatches);
    

    virtual ~NaturalTestDataset(){}
  };
}


