/// \author Ian Goodfellow
/// \file pairedVideoCollection.cc
/// \brief Implemention of PairedVideoCollection


#include "sdn/base.h"
#include "pairedVideoCollection.h"
#include <string>
#include <Eigen/Array>
#include <Eigen/Core>
using namespace std;
using namespace sdn;

PairedVideoCollection::PairedVideoCollection()
{
  minRows_ = 0;
  minCols_ = 0;
}

PairedVideoCollection::~PairedVideoCollection()
{

}

void PairedVideoCollection::load(XMLNode & root)
{
  SDN_ASSERT(root.getName());
  SDN_ASSERT(string(root.getName()) == "sdn::Dataset");
  SDN_ASSERT(root.getAttribute("id"));
  SDN_ASSERT(string(root.getAttribute("id")) == "sdn::PairedVideoCollection");
  SDN_ASSERT(root.getAttribute("path"));
  string path = root.getAttribute("path");
  
  /** Currently the only protocol that exists is called "a". Make sure we are
     using that protocol. Later, we may use other protocols for loading the
     data.
     
     Protocol a: Assumes that the data is stored in this format:
     a directory pointed to by "path". Inside this directory are an unknown
     number of subdirectories.
     Each of these subdirectories contains a sequence of images. When sorted
     alpahbetically, they will be in the correct order (i.e., any frame
     numbers have the same number of digits and are padded with leading
     zeroes).
     All of the images within a subdirectory should be the same size,
     otherwise it won't make sense to have corresponding pairs.

     Protocol "a" initializes the PairedVideoCollection instance's random
     number generated with "seed". It chooses "numImgPair" image pairs
     from all subdirectories, uniformly at random. Each pair of images
     is separated by "frameOffset" 

     Note the difference between a "protocol" and a "format". A format
     specifies how the data is stored. A protocol specifies how the data
     will be loaded, and in the case of protocol "a" only a random subset
     of the data is loaded.
  */


  SDN_ASSERT(root.getAttribute("protocol"));
  SDN_ASSERT(string(root.getAttribute("protocol")) == "a");

  vector<string> subdirs = listSubdirs(path);
  vector<vector<string> > frames(subdirs.size());
  vector<unsigned> movieSizes(subdirs.size());

  for (unsigned i = 0; i < subdirs.size(); i++)
    frames[i] = listFiles(subdirs[i]);

  for (unsigned i = 0; i < subdirs.size(); i++)
    movieSizes[i] = frames[i].size();

  SDN_ASSERT(root.getAttribute("seed"));
  unsigned int seed = atoi(root.getAttribute("seed"));

  g_.setSeed(seed);
  
  SDN_ASSERT(root.getAttribute("frameOffset"));
  unsigned int frameOffset = atoi(root.getAttribute("frameOffset"));

  SDN_ASSERT(root.getAttribute("numImgPair"));
  unsigned int numImgPair = atoi(root.getAttribute("numImgPair"));

  _imagePairs.resize(numImgPair);

  Scalar minVal = 0;
  Scalar maxVal = 0;

  for (unsigned i = 0; i < numImgPair; i++)
    {
      unsigned dirIdx = g_.rand(movieSizes, true);
      unsigned frame2Idx = g_.randUnif(frameOffset,frames[dirIdx].size()-1);
      unsigned frame1Idx = frame2Idx - frameOffset;

      loadMatrixFromImage(frames[dirIdx][frame1Idx].c_str(),
			  _imagePairs[i].first, true);
      loadMatrixFromImage(frames[dirIdx][frame2Idx].c_str(),
			  _imagePairs[i].second, true);
      SDN_ASSERT(_imagePairs[i].first.rows() == _imagePairs[i].second.rows());
      SDN_ASSERT(_imagePairs[i].first.cols() == _imagePairs[i].second.cols());

      Scalar curMin = min(_imagePairs[i].first.minCoeff(),
			 _imagePairs[i].second.minCoeff());
      Scalar curMax = max(_imagePairs[i].first.maxCoeff(),
			 _imagePairs[i].second.maxCoeff());

      if (i == 0 || curMin < minVal)
	{
	  minVal = curMin;
	}

      if (i == 0 || curMax > maxVal)
	{
	  maxVal = curMax;
	}

      unsigned curRows = _imagePairs[i].first.rows();
      unsigned curCols = _imagePairs[i].first.cols();

      if (i == 0 || curRows < minRows_)
	minRows_ = curRows;

      if (i == 0 || curCols < minCols_)
	minCols_ = curCols;
    }

  _fixDynamicRange(minVal,maxVal);
}


void PairedVideoCollection::setDynamicRange(Scalar minVal, Scalar maxVal)
{
  Scalar oldMinVal = _minVal;
  Scalar oldMaxVal = _maxVal;

  Dataset::setDynamicRange(minVal, maxVal);

  _fixDynamicRange(oldMinVal, oldMaxVal);

}

void PairedVideoCollection::_fixDynamicRange(Scalar curMinVal, Scalar curMaxVal)
{
  if (curMinVal == _minVal && curMaxVal == _maxVal)
    return;

  SDN_ASSERT(curMinVal != curMaxVal);

  
  Scalar scale = (_maxVal - _minVal) / (curMaxVal - curMinVal);

  for (unsigned i = 0; i < _imagePairs.size(); i++)
  {
/// \todo : make more efficient
	_imagePairs[i].first.cwise() -= curMinVal;
	_imagePairs[i].first *= scale;
        _imagePairs[i].first.cwise() += _minVal;
	_imagePairs[i].second.cwise() -= curMinVal;
	_imagePairs[i].second *= scale;
        _imagePairs[i].second.cwise() += _minVal;
  }
}


unsigned int PairedVideoCollection::getRandomExample(Matrix & x)
{
  SDN_ASSERT_MSG(_requestedRows > 0, "PairedVideoCollection::getRandomExample called without first calling requestExampleSize");
  SDN_ASSERT_MSG(x.rows() == _requestedRows, "Expected request for "<<_requestedRows<<" rows, got request for "<<x.rows());
  SDN_ASSERT(x.cols() == _requestedCols);


  unsigned int idx = g_.randUnif(0, _imagePairs.size()-1);
  Matrix * p = NULL;
  if (g_.randUnif(0,1))
    {
      p = & _imagePairs.at(idx).first;
    }
  else
    {
      p = & _imagePairs.at(idx).second;
    }

  x = p->block( g_.randUnif(0, p->rows()- _requestedRows),
		g_.randUnif(0, p->cols()- _requestedCols),
		_requestedRows,
		_requestedCols);

  return 0;
}
