/// \author Ian Goodfellow
/// \file videoCollection.cc
/// \brief Implemention of VideoCollection


#include "sdn/base.h"
#include "videoCollection.h"
#include <string>
#include <Eigen/Array>
#include <Eigen/Core>
using namespace std;
using namespace sdn;


//Set to true to turn on sanity checking mode where the images are displayed after being cropped and loaded
const static bool showImages = false;


VideoCollection::VideoCollection()
{
  minRows_ = 0;
  minCols_ = 0;


  ///todo-- trajectory size is hardcoded, make it configurable
  trajSize_ = 11;
  SDN_ASSERT(trajSize_ %2 == 1);
  trajWingSize_ = trajSize_ / 2;
  SDN_ASSERT(trajSize_ == 2 * trajWingSize_ + 1);
}

VideoCollection::~VideoCollection()
{

}

void VideoCollection::load(XMLNode & root)
{
  SDN_VERBOSE("Loading VideoCollection");

  indexedExamples_.clear();

  SDN_ASSERT(root.getName());
  SDN_ASSERT(string(root.getName()) == "sdn::Dataset");
  SDN_ASSERT(root.getAttribute("id"));
  SDN_ASSERT(string(root.getAttribute("id")) == "sdn::VideoCollection");

  AttributeParser p;
  //if these options are set, uses the central crop_x x crop_y pixels of the video. regards a value <= 0 as "don't crop"
  p.addInt("crop_x",-1);
  p.addInt("crop_y",-1);
  p.addString("id");

  p.parse(root);

  if ((p.getInt("crop_x").value > 0) != (p.getInt("crop_y").configured))
    {
      SDN_FATAL("So far, VideoCollection only supports cropping both x and y or neither x nor y. You asked to crop one but not the other");
    }

  bool crop = (p.getInt("crop_x").value > 0);
  unsigned crop_x = p.getInt("crop_x").value;
  unsigned crop_y = p.getInt("crop_y").value;

  vector<string> videoDirs;

  for (int i = 0; i < root.nChildNode(); i++)
    {
      XMLNode node = root.getChildNode(i);
      SDN_ASSERT(node.getName());
      SDN_ASSERT(string(node.getName()) == "videodir");
      SDN_ASSERT(node.getAttribute("path"));
      videoDirs.push_back(node.getAttribute("path"));
    }

  videos_.resize(videoDirs.size());

  vector<vector<string> > frames(videoDirs.size());

  for (unsigned i = 0; i < videoDirs.size(); i++)
    frames[i] = listFiles(videoDirs[i]);

  for (unsigned i = 0; i < videoDirs.size(); i++)
    {
      videos_[i].resize( frames[i].size());
    }

  Scalar minVal = 0;
  Scalar maxVal = 0;

  for (unsigned dirIdx = 0; dirIdx < videos_.size(); dirIdx++)
    {
      int thisVidRows = 0;
      int thisVidCols = 0;

      for (unsigned frameIdx = 0; frameIdx < videos_[dirIdx].size(); frameIdx++)
	{
	  loadMatrixFromImage(frames[dirIdx][frameIdx].c_str(),
			  videos_[dirIdx][frameIdx], true);

	  if (frameIdx == 0)
	    {
	      thisVidRows = videos_[dirIdx][frameIdx].rows();
	      thisVidCols = videos_[dirIdx][frameIdx].cols();
	    }
	  else
	    {
	      SDN_ASSERT(videos_[dirIdx][frameIdx].rows() == thisVidRows);
	      SDN_ASSERT(videos_[dirIdx][frameIdx].cols() == thisVidCols);
	    }

	  if (crop)
	    {
	      Matrix temp = videos_[dirIdx][frameIdx];
	      videos_[dirIdx][frameIdx] = temp.block(temp.rows()/2-crop_y/2, temp.cols()/2-crop_x/2, crop_y, crop_x);
	    }

	  Scalar curMin = videos_[dirIdx][frameIdx].minCoeff();
	  Scalar curMax = videos_[dirIdx][frameIdx].maxCoeff();

	  if (dirIdx + frameIdx == 0 || curMin < minVal)
	    {
	      minVal = curMin;
	    }

	  if (dirIdx + frameIdx == 0 || curMax > maxVal)
	    {
	      maxVal = curMax;
	    }

	  unsigned curRows = videos_[dirIdx][frameIdx].rows();
	  unsigned curCols = videos_[dirIdx][frameIdx].cols();

	  if (dirIdx + frameIdx == 0 || curRows < minRows_)
	    minRows_ = curRows;

	  if (dirIdx + frameIdx == 0 || curCols < minCols_)
	    minCols_ = curCols;
	}
    }
  
  _fixDynamicRange(minVal,maxVal);

  if (showImages)
    {
      for (unsigned i = 0; i < videos_.size(); i++)
	{
	  for (unsigned j = 0; j < videos_[i].size(); j++)
	    {
	      showMatrix(videos_[i][j],"VideoCollection entry");
	    }
	}

    }

  SDN_VERBOSE("Done loading VideoCollection");
}


void VideoCollection::setDynamicRange(Scalar minVal, Scalar maxVal)
{
  Scalar oldMinVal = _minVal;
  Scalar oldMaxVal = _maxVal;

  Dataset::setDynamicRange(minVal, maxVal);

  _fixDynamicRange(oldMinVal, oldMaxVal);

}

void VideoCollection::_fixDynamicRange(Scalar curMinVal, Scalar curMaxVal)
{
  if (curMinVal == _minVal && curMaxVal == _maxVal)
    return;

  SDN_ASSERT(curMinVal != curMaxVal);

  
  Scalar scale = (_maxVal - _minVal) / (curMaxVal - curMinVal);

  for (unsigned i = 0; i < videos_.size(); i++)
  {
    for (unsigned j = 0; j < videos_[i].size(); j++)
      {
/// \todo : make more efficient
	videos_[i][j].cwise() -= curMinVal;
	videos_[i][j] *= scale;
        videos_[i][j].cwise() += _minVal;

      }
  }
}

unsigned int VideoCollection::getRandomExample(Matrix & x)
{
  SDN_ASSERT_MSG(_requestedRows > 0, "VideoCollection::getRandomExample called without first calling requestExampleSize");
  SDN_ASSERT_MSG(x.rows() == _requestedRows, "Expected request for "<<_requestedRows<<" rows, got request for "<<x.rows());
  SDN_ASSERT(x.cols() == _requestedCols);

  Matrix * p = NULL;
  unsigned int i = g_.randUnif(0, videos_.size()-1);
  unsigned j = g_.randUnif(0, videos_[i].size()-1);

  p = & videos_[i][j];

  x = p->block( g_.randUnif(0, p->rows()- _requestedRows),
		g_.randUnif(0, p->cols()- _requestedCols),
		_requestedRows,
		_requestedCols);

  return 0;
}

bool VideoCollection::supportsTrajectoryIndexing()
{
  return true;
}

unsigned int VideoCollection::getRandomIndexedExample(Matrix & x, unsigned & index)
{
   SDN_ASSERT_MSG(_requestedRows > 0, "VideoCollection::getRandomExample called without first calling requestExampleSize");
  SDN_ASSERT_MSG(x.rows() == _requestedRows, "Expected request for "<<_requestedRows<<" rows, got request for "<<x.rows());
  SDN_ASSERT(x.cols() == _requestedCols);

  Matrix * p = NULL;
  unsigned int video = g_.randUnif(0, videos_.size()-1);
  //Don't pick frames that are too early or late in the video or we won't be able to construct a trajectory
  unsigned frame = g_.randUnif(trajWingSize_, videos_[video].size()-1-trajWingSize_);

  p = & videos_[video][frame];

  unsigned i = g_.randUnif(0,p->rows() - _requestedRows);
  unsigned j = g_.randUnif(0,p->cols() - _requestedCols);

  x = p->block( i,
		j,
		_requestedRows,
		_requestedCols);

  index = indexedExamples_.size();

  IndexedExample r;
  r.video = video;
  r.frame = frame;
  r.i = i;
  r.j = j;

  indexedExamples_.push_back(r);

  return 0; 
}


vector<Matrix> VideoCollection::getTrajectoryFromIndex(unsigned index)
{
  vector<Matrix> rval;

  IndexedExample r = indexedExamples_[index];

  unsigned firstFrame = r.frame-trajWingSize_;

  for (unsigned i = 0; i < trajSize_; i++)
    {
      rval.push_back( videos_[r.video][firstFrame+i].block(r.i,r.j,_requestedRows,_requestedCols));
    }

  return rval;
}

bool VideoCollection::trajectoryOverlap(unsigned index1, unsigned index2)
{
  IndexedExample r1 = indexedExamples_[index1];
  IndexedExample r2 = indexedExamples_[index2];

  int f1 = r1.frame;
  int f2 = r2.frame;

  if (std::abs(f1-f2) >= (int) trajWingSize_)
    {
      return false;
    }

  int i1 = r1.i;
  int i2 = r2.i;

  if (std::abs(i1-i2) >= _requestedRows)
    {
      return false;
    }

  int j1 = r1.j;
  int j2 = r2.j;

  if (std::abs(j1-j2) >= _requestedCols)
    {
      return false;
    }

  return true;
}

unsigned VideoCollection::getExampleFromIndex(Matrix & x, unsigned index)
{
  x = videos_[indexedExamples_[index].video][indexedExamples_[index].frame].block(
										indexedExamples_[index].i,indexedExamples_[index].j,_requestedRows,_requestedCols);
  return 0;
}
