/// \author Ian Goodfellow <ia3n@cs.stanford.edu>
/// \file videoCollection.h
/// \brief A dataset class for holding pairs of frames sampled from videos

#pragma once


#include "dataset.h"
#include <map>
#include <vector>
#include "../base/random.h"

namespace sdn
{

  /// A Dataset capapble of storing videos as subsampled pairs
  class VideoCollection : public Dataset
  {
  public:
    
    VideoCollection();
    virtual ~VideoCollection();
    
    void load (XMLNode & root);
    void setDynamicRange(Scalar minVal, Scalar maxVal);
    unsigned int getRandomExample(Matrix & x);

    bool supportsExampleSize(unsigned rows, unsigned cols)
    {
      return rows <= minRows_ && cols <= minCols_;
    }

    unsigned getPreferredRows()
    {
      return std::min(14u,minRows_);
    }

    unsigned getPreferredCols()
    {
      return std::min(14u,minCols_);
    }


  ///Returns true if the getRandomIndexedExample, getTrajectoryFromIndex, and trajectoryOverlap methods are supported
    virtual bool supportsTrajectoryIndexing();


    /// Like getRandomExample, but also writes out an index so this example can be found later
    virtual unsigned int getRandomIndexedExample(Matrix & x, unsigned & index);

    unsigned getExampleFromIndex(Matrix & x, unsigned index);
    
    /// Returns a sequence of related patches. Relatedness will mean different things in different contexts-- it could be patches that appear at the same location in the frame, but with different time coordinates, or it could be gratings that all have the same orientation and frequency but different phase, etc. The central patch will be the patch that was returned by getRandomIndexedExample with index index
    virtual std::vector<Matrix> getTrajectoryFromIndex(unsigned index);
    
    /// Returns true if the the trajectories associated with two indices overlap
    virtual bool trajectoryOverlap(unsigned index1, unsigned index2);
    



  protected:

    ///  videos_[i][j] is frame j of video i
    std::vector<std::vector<Matrix> > videos_;

    RandomIntGenerator<> g_;

    ///if values in _images are actually in the range [curMinVal,curMaxVal], this scale/translates everything to range [_minVal, _maxVal]
    void _fixDynamicRange(Scalar curMinVal, Scalar curMaxVal);

    unsigned minRows_;
    unsigned minCols_;


    /// number of elements in a trajectory. must be odd
    unsigned trajSize_;
    ///number of elements in the trajectory occuring to the left of center (equivalently, to the right of center)
    unsigned trajWingSize_;


    struct IndexedExample
    {
      unsigned video;
      unsigned frame;
      unsigned i;
      unsigned j;
    };

    std::vector<IndexedExample> indexedExamples_;

  };

}

