// Copyright 2009, All Rights Reserved.
// Author: Andrew Ng (ang@cs.stanford.edu), Quoc V. Le (quocle@cs.stanford.edu)
/** \file visualizer.h
    \brief Abstract class for visualization.
*/
    
#pragma once
#include <string>
#include "../base/matrix.h"

/// Abstract class that declares an interface for visualization.
/**
 * Abstract class that declares an interface for visualization.
 */
namespace sdn
{

/// Displays collections of fixed-sized image patches in a grid
class AbstractPatchDisplay
{
 public:
  /// Add a patch matrix to the visualizer
  /* 
     Add a patch (in eigen matrix) to the visualizer.
     \param patch the patch to be added to the visualizer.
     \param allow_rescaling_colormap ?
     \param allow_scaling_up ?
   */
  virtual void addPatch(const Matrix &patch, 
			bool allow_rescaling_colormap = 0, 
			bool allow_scaling_up = 0) = 0;
  /* 
     Save the current display to an image file.
     \param filename the filename to be saved as.
  */
  virtual void saveImage(const char *filename) = 0; 
  
  /* 
     Overloaded function to save an image from a string 
     instead of a char array.
     \param filename the filename to be saved as (ends with png, jpg etc).
  */
  virtual void saveImage(const std::string filename){
    saveImage(filename.c_str());
  }
  
  /* 
     Turns on display in window. 
  */
  virtual void enableWindow(void) = 0;       
  
  /* 
     Turns off display in window. 
  */
  virtual void disableWindow() = 0;
  
  /* 
     Clears the current display.
  */
  virtual void clearDisplay(void) = 0;

  /* 
     Waits some time (in milliseconds) before adding a new patch 
     \param time time in milliseconds to be delayed.
  */
  virtual void wait(int time)=0;
  
  /* 
     Shows the window and stop there.
     Disappears as soon as any key is pressed.
  */
  virtual void showAndStop()=0; 


  virtual unsigned getPatchWidth()=0;

  virtual unsigned getPatchHeight()=0;

  virtual std::string getName()=0;

  virtual ~AbstractPatchDisplay() {} 
};
}

