// Copyright 2009, All Rights Reserved.
// Author: Quoc Le
// Based on PatchDisplay by Andrew Ng and Ian Goodfellow
/** \file opencvPatchDisplay.cc
    \brief OpenCV's utilities for loading/saving/etc. image data.
*/
#include <cv.h>
#include <Eigen/Core>
#include <Eigen/Array>
#include <fstream>
#include <highgui.h>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include "patchDisplay.h"
#include "sdn/base.h"
#include <assert.h>

#define BACKGROUND_COLOR 192 // grayish color

using namespace std; 
using namespace sdn;
USING_PART_OF_NAMESPACE_EIGEN // import most common Eigen types 

PatchDisplay::PatchDisplay(string display_name, 
				   int patch_width, 
				   int patch_height, 
				   int size_mult, 
				   int cols, int rows)
  : patch_width_(patch_width),
    patch_height_(patch_height), 
    size_mult_(size_mult), cols_(cols), 
    rows_(rows),
    window_created_(0) 
{
  update_window_ = true;
  int image_width = MARGIN*(cols+1) + size_mult*patch_width*cols;
  int image_height = MARGIN*(rows+1) + size_mult*patch_height*rows;

  low_lim_ = 0.0; high_lim_ = 1.0; 
  window_name_ =  display_name; 
  display_ = cvCreateImage(cvSize(image_width, image_height), IPL_DEPTH_8U, 3);
  next_col_ = next_row_ = 0; 
  cvSet(display_, cvScalar(BACKGROUND_COLOR,BACKGROUND_COLOR,BACKGROUND_COLOR));

}

PatchDisplay::PatchDisplay(string display_name, 
				   int patch_width, 
				   int patch_height, 
				   int size_mult, 
				   int numPatches)
  : patch_width_(patch_width), 
    patch_height_(patch_height), 
    size_mult_(size_mult), 
    window_created_(0) 
{
  update_window_ = true;


  #define NUMPATCHES_TO_RC rows_ = (int) sqrt((double) numPatches); \
  cols_ = numPatches / rows_; \
  if (rows_ * cols_ != numPatches) \
    { \
      if (rows_ * cols_ < numPatches) \
	{ \
	  if (cols_ %2 != 0) \
	    cols_ ++; \
	  else \
	    rows_ ++; \
	  assert(rows_*cols_ >= numPatches); \
	} \
    } 

  NUMPATCHES_TO_RC;

  int image_width = MARGIN*(cols_+1) + size_mult*patch_width*cols_;
  int image_height = MARGIN*(rows_+1) + size_mult*patch_height*rows_;

  low_lim_ = 0.0; high_lim_ = 1.0; 
  window_name_ = display_name; 
  display_ = cvCreateImage(cvSize(image_width, image_height), IPL_DEPTH_8U, 3);
  // cout << "Display is " << image_width << "x" << image_height << std::endl; 

  next_col_ = next_row_ = 0; 

  cvSet(display_, cvScalar(BACKGROUND_COLOR,BACKGROUND_COLOR,BACKGROUND_COLOR));

}

void PatchDisplay::setNewDisplayParam(int patch_width, 
					  int patch_height, 
					  int size_mult, 
					  int numPatches)
{
  NUMPATCHES_TO_RC;

  cvReleaseImage(&display_);

  patch_width_ = patch_width;
  patch_height_ = patch_height;
  size_mult_ = size_mult;


  int image_width = MARGIN*(cols_+1) + size_mult*patch_width*cols_;
  int image_height = MARGIN*(rows_+1) + size_mult*patch_height*rows_;
  display_ = cvCreateImage(cvSize(image_width, image_height), IPL_DEPTH_8U, 3);

  next_col_ = next_row_ = 0; 

  cvSet(display_, cvScalar(BACKGROUND_COLOR,BACKGROUND_COLOR,BACKGROUND_COLOR));


  return;
}

void PatchDisplay::setNewDisplayParam(int patch_width, 
					  int patch_height, 
					  int size_mult, 
					  int cols, 
					  int rows)
{
  cvReleaseImage(&display_);

  patch_width_ = patch_width;
  patch_height_ = patch_height;
  size_mult_ = size_mult;
  cols_ = cols;
  rows_ = rows;

  int image_width = MARGIN*(cols+1) + size_mult*patch_width*cols;
  int image_height = MARGIN*(rows+1) + size_mult*patch_height*rows;
  display_ = cvCreateImage(cvSize(image_width, image_height), IPL_DEPTH_8U, 3);

  next_col_ = next_row_ = 0; 

  cvSet(display_, cvScalar(BACKGROUND_COLOR,BACKGROUND_COLOR,BACKGROUND_COLOR));


  return;
}

void PatchDisplay::enableWindow(void)
{
  SDN_DEBUG("enableWindow executing");
  if (!window_created_) {
    SDN_DEBUG("making window "<<window_name_);
    window_created_ = 1;
    cvNamedWindow(window_name_.c_str(), CV_WINDOW_AUTOSIZE);
    cvShowImage(window_name_.c_str(), display_);
  }
}

void PatchDisplay::disableWindow()
{
  if (window_created_)
    {
      window_created_ = 0;
      cvDestroyWindow(window_name_.c_str());
    }
}

PatchDisplay::~PatchDisplay()
{
  cvReleaseImage(&display_);
  if (window_created_) 
    cvDestroyWindow(window_name_.c_str());

}

void PatchDisplay::addPatch(const Matrix &patch, 
				bool allow_rescaling_colormap, 
				bool allow_scaling_up)
{
  if (next_row_ == rows_) {
    assert(next_col_ == 0);
    // clear the display
    clearDisplay(); 
  }



  //cout << "Adding "<<patch.cols()<<"x"<<patch.rows()<<" patch to "<<this<<endl;

  assert(patch.rows() == patch_height_ && patch.cols() == patch_width_);

  int upperLeftI = (MARGIN + patch_height_ * size_mult_) * next_row_ + MARGIN; 
  int upperLeftJ = (MARGIN + patch_width_ * size_mult_) * next_col_ + MARGIN; 

  double patchMin = patch.minCoeff(), patchMax = patch.maxCoeff();

  double thisHighLim, thisLowLim; 
  if (allow_rescaling_colormap) {
    thisHighLim = max(high_lim_, patchMax);
    thisLowLim = min(low_lim_, patchMin);
  } else {
    thisHighLim = high_lim_;
    thisLowLim = low_lim_;
  }

  if (allow_scaling_up)
    {
      if (patchMax > 0 && patchMax < high_lim_)
	thisHighLim = patchMax;
      
      if (patchMin < 0 && patchMin > low_lim_)
	thisLowLim = patchMin;
    }      
  
  //make sure grey point remains consistent
  if (thisHighLim > -thisLowLim)
    thisLowLim = - thisHighLim;
  if (-thisLowLim > thisHighLim)
    thisHighLim = - thisLowLim;

  //unsigned long minAddr = ~ long(0);
  //unsigned long maxAddr = 0;

  for (int i=0; i < patch_height_*size_mult_; i++)
    for (int j=0; j < patch_width_*size_mult_; j++) {
      int thisPixelValue = (int)round((patch(i/size_mult_,j/size_mult_)-thisLowLim)/(thisHighLim-thisLowLim)*255.0);
      thisPixelValue = max(thisPixelValue, 0);
      thisPixelValue = min(thisPixelValue, 255);


      CvScalar s = cvScalar(thisPixelValue,thisPixelValue,thisPixelValue);
      cvSet2D(display_,upperLeftI+i, upperLeftJ+j, s);

      /*  IG-- I was getting bugs when using non-square patches with this code. Haven't had time to fix,
	  instead I wrote the above cvSet2D code which is probably a lot slower

#if CHECK_IMAGE_ACCESS
      assert(display_->nChannels == 3);
      assert(display_->depth == IPL_DEPTH_8U);
      assert(upperLeftI+i < display_->height);
      assert(upperLeftJ+j < display_->width);

      unsigned long firstAddr = (unsigned long) &CV_IMAGE_ELEM(display_, uchar, upperLeftI + i, (upperLeftJ + j)*3 + 0);
      unsigned long lastAddr = (unsigned long) & CV_IMAGE_ELEM(display_, uchar, upperLeftI + i, (upperLeftJ + j)*3 + 2);
      unsigned long baseAddr = (unsigned long) display_->imageData;

      assert(firstAddr > baseAddr);
      assert(lastAddr < baseAddr+display_->width*display_->height*3);

      //cout << this << " range: "<<baseAddr<<" "<<baseAddr+display_->width*display_->height*3 << endl;


      //if (firstAddr < minAddr)
	//minAddr = firstAddr;

      //if (lastAddr > maxAddr)
	//maxAddr = lastAddr;
      
#endif

           CV_IMAGE_ELEM(display_, uchar, upperLeftI + i, (upperLeftJ + j)*3 + 0)
      = CV_IMAGE_ELEM(display_, uchar, upperLeftI + i, (upperLeftJ + j)*3 + 1)
      = CV_IMAGE_ELEM(display_, uchar, upperLeftI + i, (upperLeftJ + j)*3 + 2)
       = thisPixelValue; 
      */
    }

  //cout << this << " wrote to "<<minAddr <<" through "<<maxAddr << endl;

  if (window_created_ && update_window_)
    cvShowImage(window_name_.c_str(), display_);

  next_col_++;
  if (next_col_ == cols_) {
    next_col_=0;
    next_row_++;
  }
}

void PatchDisplay::unfreezeWindow()
{
  update_window_ = true;
  if (window_created_)
    cvShowImage(window_name_.c_str(), display_);


}


/// \todo-- on failure this seg faults inside opencv, refer to stair vision for how to trap the opencv error
void PatchDisplay::saveImage(const char *filename)
{
  SDN_DEBUG("Saving "<<filename<<endl);
  bool success = cvSaveImage(filename, display_);
  if (!success)
    {
      SDN_WARN("PatchDisplay::SaveImage: Error saving image file");
    }
}

void PatchDisplay::copy(const PatchDisplay & other)
{
  display_ = cvCloneImage(other.display_);

  patch_width_ = other.patch_width_;
  patch_height_ = other.patch_height_;
  size_mult_ = other.size_mult_;
  cols_ = other.cols_;
  rows_ = other.rows_;
  next_col_ = other.next_row_;
  window_created_ = other.window_created_;
  update_window_ = other.update_window_;
  window_name_ = other.window_name_;
  low_lim_ = other.low_lim_;
  high_lim_ = other.high_lim_;
}

PatchDisplay::PatchDisplay(const PatchDisplay & other)
{
  copy(other);
}

const PatchDisplay & PatchDisplay::operator=(const PatchDisplay & other)
{
  if (&other != this)
    copy(other);
  return *this;
}

void PatchDisplay::clearDisplay(void){
  cvSet(display_, cvScalar(BACKGROUND_COLOR,BACKGROUND_COLOR,BACKGROUND_COLOR)); next_col_ = next_row_ = 0;
}

void PatchDisplay::freezeWindow(void){
  update_window_ = false;
}

void PatchDisplay::setLimits(double low_lim, double high_lim){
  low_lim_ = low_lim; 
  high_lim_ = high_lim;
}

void PatchDisplay::wait(int time){
  cvWaitKey(time);
}

void PatchDisplay::showAndStop(){
  cvWaitKey(0);
}

