// Copyright 2009, All Rights Reserved.
// Author: Quoc Le
// Based on PatchDisplay by Andrew Ng and Ian Goodfellow
/** \file opencvVisualizer.h
    \brief OpenCV's utilities for loading/saving/etc. image data.
*/
    
#pragma once

#include <cv.h>
#include <string>
#include <vector>
#include "abstractPatchDisplay.h"
#include "../base/fileUtils.h"


/// OpenCV's Visualization
/**
 * OpenCV's Visualization, an implementation of the PatchDisplay class.
 */

///Stanford Deep Network
namespace sdn
{

/// OpenCV implementation of AbstractPatchDisplay

class PatchDisplay: public AbstractPatchDisplay
{

 public:
  /// Constructs an  visualizer
  /** 
     PatchDisplay constructor 
     \param display_name the title in the window display.
     \param patch_width the width of the patches to be added.
     \param patch_height the height of the patches to be added.
     \param size_mult ?
     \param cols number of columns in the visualizer.
     \param rows number of rows in the visualizer
  */
  PatchDisplay(std::string display_name, int patch_width, 
		   int patch_height, int size_mult, int cols, int rows);
  /// Constructs a PatchDisplay
  /** 
     PatchDisplay constructor 
     \param display_name the title in the window display.
     \param patch_width the width of the patches to be added.
     \param patch_height the height of the patches to be added.
     \param size_mult ?
     \param num_patches number of patches to be added. 
     This function will sort out to find the number of rows and columns
     in the visualizer.
  */
  PatchDisplay(std::string display_name, int patch_width, 
	     int patch_height, int size_mult, int num_patches);
  /// Constructs a PatchDisplay
  /** 
     PatchDisplay constructor from other visualizer. 
     \param other other visualizer.
  */
  PatchDisplay(const PatchDisplay & other);
  /// Destructor of this class.
  virtual ~PatchDisplay();
  
  /// Add a patch matrix to the visualizer
  /** 
     Add a patch (in eigen matrix) to the visualizer.
     \param patch the patch to be added to the visualizer.
     \param allow_rescaling_colormap ?
     \param allow_scaling_up ?
   */
  virtual void addPatch(const Matrix &patch, 
			bool allow_rescaling_colormap = 0, 
			bool allow_scaling_up = 0); //doing this while window is enabled is a very slow operation
  
  /// Change display parameters
  /**
    Change the display parameters of the current visualizer.
    \param patch_width the width of the patches to be added.
    \param patch_height the height of the patches to be added.
    \param size_mult ?
    \param cols number of columns in the visualizer.
    \param rows number of rows in the visualizer
   */  
  void setNewDisplayParam(int patch_width, int patch_height, int size_mult, int cols, int rows);
  /** 
     Change the display parameters of the current visualizer.
     \param display_name the title in the window display.
     \param patch_width the width of the patches to be added.
     \param patch_height the height of the patches to be added.
     \param size_mult ?
     \param num_patches number of patches to be added. 
  */
  void setNewDisplayParam(int patch_width, int patch_height, int size_mult, int num_patches);
  

  /** 
     Save the current display to an image file.
     \param filename the filename to be saved as (ends with png, jpg etc).
  */
  virtual void saveImage(const char *filename); 
  
  void setLimits(double low_lim, double high_lim);
  /** 
     Turns on display in window. 
  */
  virtual void enableWindow(void);       
  
  /** 
     Turns off display in window. 
  */
  virtual void disableWindow();          
  
  /** 
     Clears the current display.
  */
  virtual void clearDisplay(void);

  /** 
     Freezes the current display.
  */
  void freezeWindow();

  /** 
     Unfreezes the current display.
  */
  void unfreezeWindow();
  
  /** 
     waits some time (in milliseconds)
     \param time time in milliseconds to be delayed.
  */
  virtual void wait(int time);
  
  /** 
     Shows the window and stop there.
     Disappears as soon as any key is pressed.
  */
  virtual void showAndStop();

  /** 
     Compares the current display with another display.
  */
  const PatchDisplay & operator=(const PatchDisplay & other);
  

  unsigned getPatchWidth()
  {
    return patch_width_;
  }

  unsigned getPatchHeight()
  {
    return patch_height_;
  }

  std::string getName() { return window_name_; }

  std::string getFileSafeName() { return makeValidFilename(window_name_); }

 private:
  
  IplImage *display_;
  
  int patch_width_, patch_height_, size_mult_, cols_, rows_; 
  
  int next_col_, next_row_; // index of next rectangle to display image in 
  
  int window_created_;      // whether window has been created yet 
  
  int update_window_;       // whether the window should be updated on each 
                            // addPatch. controlled with freezeWindow / unfreezeWindow. 
                            // (use freezeWindow before adding a large batch of patches 
                            // to make it run faster)
  
  static const int MARGIN = 10;  // number of pixels of margin between displays 
  
  std::string window_name_; 
  double low_lim_, high_lim_; 
  


 private:
  /*
    Copy the other display to this display.
  */
  void copy(const PatchDisplay & other);
};

}
