/// \author Ian Goodfellow

#pragma once
#include <vector>

namespace sdn
{

  /// Records the activation of a single hidden unit and the index of the example on which it occurred
  struct ActivationInstance
  {
    Scalar value;
    unsigned index;

    bool operator<(const ActivationInstance & other) const { return value < other.value; }
  };

  /// A record of how a hidden unit has fired over several examples. Sorted in order of increasing activity
  typedef std::vector<ActivationInstance> UnitActivationRecord;

  /// A record of how a population of hidden units have fired, over several examples
  class ActivationRecord
  {
  protected:
    typedef std::vector<UnitActivationRecord> ActivationMatrixRow;
    typedef std::vector<ActivationMatrixRow> ActivationMatrix;

  protected:

    void add(ActivationInstance r, unsigned i, unsigned j)
    {
      addSorted(matrix_[i][j],r);
    }

  protected:
    unsigned rows_;
    unsigned cols_;
    ActivationMatrix matrix_;

  public:

    /**
       \param rows Number of rows in the activity matrices for the population of hidden units being tracked
       \param cols Number of cols in the activity matrices for the population of hidden units being tracked
       \param expectNumActivations: reserves space for this many activations, as an optimization
    */
    ActivationRecord(unsigned rows, unsigned cols, unsigned expectNumActivations = 0)
    {
      rows_ = rows;
      cols_ = cols;
      UnitActivationRecord elem;
      elem.reserve(expectNumActivations);
      ActivationMatrixRow row(cols, elem);
      matrix_.resize(rows,row);
    }

    /// Makes an ActivationInstance and puts it, in sorted order, into the appropriate UnitActivationRecord for each hidden unit
    void add(const Matrix & values, unsigned index)
    {
      SDN_ASSERT((unsigned) values.rows() == rows_);
      SDN_ASSERT((unsigned) values.cols() == cols_);

      for (unsigned i = 0; i < rows_; i++)
	for (unsigned j = 0; j < cols_; j++)
	  {
	    ActivationInstance r;
	    r.value = values(i,j);
	    r.index = index;
	    
	    add(r,i,j);
	  }
    }

    unsigned numEntries()
    {
      if (rows_ * cols_ == 0)
	return 0;

      return matrix_[0][0].size();
    }
 
    UnitActivationRecord & getUnitRecord(unsigned i, unsigned j) { return matrix_[i][j]; }


  };

}

