#include "bestInputVisualization.h"
#include "storageNode.h"
#include <limits>
#include "stackedNetwork.h"
#include "../base/registry.h"
using namespace sdn;
using namespace std;
using namespace boost;

namespace sdn
{

  class BestInputUnitVisualizationMaker
  {
    
  public:
    
    BestInputUnitVisualizationMaker(int numWinners)
    {
      winners_.resize(numWinners);
      scores_.resize(numWinners,std::numeric_limits<double>::min());
    }
    
    void processExample(const Matrix & m, double value)
    {
      unsigned i = 0;
      for (i = 0; i < winners_.size(); i++)
	{
	  if (scores_[i] == numeric_limits<double>::min())
	    {
		continue;
	    }
	  if (scores_[i] > value)
	    {
	      break;
	    }
	}
      if (i > 0)
	{
	  scores_.insert(scores_.begin()+i,value);
	  winners_.insert(winners_.begin()+i,m);
	  scores_.erase(scores_.begin());
	  winners_.erase(winners_.begin());
	}
    }
    
    void appendImages(vector<Matrix> & output)
    {
      for (unsigned i = 0; i < winners_.size(); i++)
	{
	  output.push_back(winners_[i]);
	}
    }
    
  protected:
    vector<Matrix> winners_;
    vector<double> scores_;
    
  };
  
  class BestInputLayerVisualizationMaker
  {
  public:
    
    BestInputLayerVisualizationMaker(shared_ptr<StorageNode> n, int numWinners) : node_(n), numWinners_(numWinners)
    {
      BestInputUnitVisualizationMaker temp(numWinners);
      
      unitVisualizationMakers_.resize(n->getActivities().rows(), vector<BestInputUnitVisualizationMaker>(n->getActivities().cols(),temp));
    }
    
    void processExample(const Matrix & m)
    {
      for (unsigned i = 0; i<  unitVisualizationMakers_.size(); i++)
	{
	  for (unsigned j = 0; j < unitVisualizationMakers_[i].size(); j++)
	    {
	      unitVisualizationMakers_[i][j].processExample(m, node_->getActivities()(i,j));
	    }
	}
    }
    
    
    VisImageSet getVisImageSet()
    {
      SDN_DEBUG("Enter getVisImageSet()");

      VisImageSet rval;

      if (numWinners_ == 1)
	{
	  rval.groupSize = 1;
	}
      else
	{
	rval.groupSize = numWinners_+1;
	}
      
      for (unsigned j = 0; j < unitVisualizationMakers_[0].size(); j++)
	{
	  for (unsigned i = 0; i < unitVisualizationMakers_.size(); i++)
	      {unitVisualizationMakers_[i][j].appendImages(rval.images);
	      if (numWinners_ != 1)
		{
		  SDN_ASSERT(rval.images.size());
		  SDN_ASSERT(rval.images.back().rows());
		  SDN_ASSERT(rval.images.back().cols());
		rval.images.push_back(rval.images.back());
		rval.images.back().setConstant(0);
		}
	    }
	}
      

      return rval;
    }
    
    
    
  protected:
    
    shared_ptr<StorageNode> node_;
    int numWinners_;
    vector<vector<BestInputUnitVisualizationMaker> > unitVisualizationMakers_;
    
  };
  
}

BestInputVisualization::BestInputVisualization()
{
  p_.addInt("numExamples",10000);//number of examples to present while taking the max
  p_.addInt("numWinners",4);//number of top patches to show for each position
  p_.addString("dataset","ng_video");
  p_.addString("id");
}

void BestInputVisualization::load(XMLNode & node)
{
  p_.parse(node);
}

std::vector<VisImageSet> BestInputVisualization::makeVisualizations(boost::shared_ptr<Network> network)
{
  shared_ptr<StackedNetwork> p = dynamic_pointer_cast<StackedNetwork>(network);

  SDN_ASSERT(p);

  

  shared_ptr<Dataset> d = DatasetRegistry().make(p_.getString("dataset").value);

  network->registerDataset(d);

  vector<BestInputLayerVisualizationMaker> layerVisualizationMakers;

  int numWinners = p_.getInt("numWinners").value;


 


  for (unsigned i = 0; i < p->numLayers()/2; i++)
    layerVisualizationMakers.push_back(BestInputLayerVisualizationMaker(p->getStorageNode(i+1), numWinners));
  



  Matrix input = Matrix::Zero(network->getInputRows(),network->getInputCols());

  int numExamples = p_.getInt("numExamples").value;

  for (int i = 0; i < numExamples; i++)
    {

      d->getRandomExample(input);


      network->processExample(input);


      for (unsigned j = 0; j < layerVisualizationMakers.size(); j++)
	{
	  layerVisualizationMakers[j].processExample(input);
	}

    }

  vector<VisImageSet> rval;



  for (unsigned i = 0; i < layerVisualizationMakers.size(); i++)
    {
      rval.push_back(layerVisualizationMakers[i].getVisImageSet());
      stringstream s;
	s << "Layer "<<i+1<<": Top "<<numWinners<<" patches";
	rval.back().title = s.str();
    }



  return rval;
}
