// Copyright 2009, All Rights Reserved.
// Author: Andrew Saxe (asaxe@stanford.edu)
                                                                                
/** \file convolutionLayer.cc
    \brief  Defines the interface of the convolution layer.
*/

#include "costFunction.h"                                                                                
#include "convolutionLayer.h"
#include "sdn/base.h"
#include "storageNode.h"


#include <assert.h>

#include <Eigen/Array>

using namespace sdn;


Convolver::Convolver(int patch_width, 
		     int patch_step, 
		     int output_width) {

  //Can't have parts of the input that do not propagate forward right now
  //since then there is no way to reconstruct them
  assert(patch_width >= patch_step);

  patch_width_ = patch_width;
  patch_step_ = patch_step;
  output_width_ = output_width;
  input_width_ = patch_step*(output_width-1) + patch_width;

  //Compute and cache the averaging coefficients to be used in the 
  //deconvolution process
 
  Matrix averaging_matrix_ = Matrix::Zero(input_width_, input_width_);
  
  //int in_r;
  //int in_c; 
  for(int out_c = 0,in_c = 0; out_c  < output_width_; 
      ++out_c, in_c += patch_step_) {
    for( int out_r = 0,in_r = 0; out_r < output_width_; 
	 ++out_r, in_r += patch_step_) {
      
      averaging_matrix_.block(in_r, 
			      in_c, 
			      patch_width_, 
			      patch_width_).cwise() += 1;
    }
  }
  //std::cout << averaging_matrix_ << "\n";
  averaging_coefficients_ = Vector::Zero(input_width_*input_width_);
  for(int c = 0; c < averaging_matrix_.cols(); ++c) {
    averaging_coefficients_.segment(c*input_width_,input_width_) =
      averaging_matrix_.col(c);
  }
}

//Takes a matrix that represents a number of 2D arrays of inputs. In particular,
/*

 */
const Matrix Convolver::convolve(const Matrix& input) {

  SDN_ASSERT_MSG(input_width_*input_width_ == input.cols(),
		 "Input should be of size " << input_width_*input_width_ << 
		 " but it was of size " << input.cols());

  //Todo(asaxe): add a check to make sure the input is big enough to do the
  //convolution.
  
  Matrix output = Matrix::Zero(input.rows()*patch_width_*patch_width_,
				       output_width_*output_width_);

  //These first two loops iterate over the location of the patch in the input.
  //Here in_r and in_c are the row and column coordinates of the upper left 
  //corner of the patch in the input that will be extracted, and out_r and out_c
  //are the indices in the output array at which the extracted patch will be 
  //placed. (Note that these 2D arrays are never actually formed, but are
  //represented in column major order in the columns of input_mat and 
  //output_mat respectively.)
  //int in_r;
  //int in_c; 
  for(int out_c = 0,in_c = 0; out_c  < output_width_; 
      ++out_c, in_c += patch_step_) {
    for( int out_r = 0,in_r = 0; out_r < output_width_; 
	 ++out_r, in_r += patch_step_) {

      //These next two loops iterate over each element of the patch with upper 
      //left corner at (in_r,in_c). Patch_r and _c are offsets from the upper 
      //left corner of the patch. 
      int itr = 0;
      for(int patch_c = 0; patch_c < patch_width_; ++patch_c) {
	for(int patch_r = 0; patch_r < patch_width_; ++patch_r) {
	  //TODO(asaxe): The below could be sped up by using pointer arithmetic 
	  //instead of the OutputIndex and InputIndex calls. 
	  
	  output.block(itr*input.rows(),outputIndex(out_r,out_c),
		    input.rows(), 1) = 
	    input.col(inputIndex(in_r + patch_r, in_c + patch_c));
	 
	  itr++;
	}
      }
      
      
    }
  }

  return output;
  
}

const Matrix Convolver::deconvolve(const Matrix& output) {
  //Output must be a multiple of the patch_width_ 
  assert(output.rows() % (patch_width_*patch_width_) == 0);

  Matrix input = Matrix::Zero(output.rows()/patch_width_/patch_width_, 
				      input_width_*input_width_);

  //These first two loops iterate over the location of the patch in the input.
  //Here in_r and in_c are the row and column coordinates of the upper left 
  //corner of the patch in the input that will be reconstructed, 
  //and out_r and out_c are the indices in the output array from which the
  //values will be extracted. (Note that these 2D arrays are never actually 
  //formed, but are represented in column major order in the columns of 
  //input_mat and output_mat respectively.)
  //int in_r;
  //int in_c; 
  for(int out_c = 0,in_c = 0; out_c  < output_width_; 
      ++out_c, in_c += patch_step_) {
    for( int out_r = 0,in_r = 0; out_r < output_width_; 
	 ++out_r, in_r += patch_step_) {

      //These next two loops iterate over each element of the patch with upper 
      //left corner at (in_r,in_c). Patch_r and _c are offsets from the upper 
      //left corner of the patch. 
      int itr = 0;
      for(int patch_c = 0; patch_c < patch_width_; ++patch_c) {
	for(int patch_r = 0; patch_r < patch_width_; ++patch_r) {
	  //TODO(asaxe): The below could be sped up by using pointer arithmetic 
	  //instead of the OutputIndex and InputIndex calls. 
	  
	  input.col(inputIndex(in_r + patch_r, in_c + patch_c)) += 
	    output.block(itr*input.rows(),outputIndex(out_r,out_c),
		      input.rows(), 1);
	 
	  itr++;
	}
      }
      
      
    }
  }

  //Finally, apply the appropriate averaging constants.
  for(int c = 0; c < input.cols(); c++) {
    input.col(c) /= averaging_coefficients_(c);
  }

  return input;

}

int Convolver::inputIndex(int r, int c) {
  return c*input_width_ + r;
}

int Convolver::outputIndex(int r, int c) {
  return c*output_width_ + r;
}

int Convolver::patch_width() {
  return patch_width_;
}
int Convolver::patch_step() {
  return patch_step_;
}
int Convolver::output_width() {
  return output_width_;
}

ConvolutionLayer::ConvolutionLayer()  {
}

ConvolutionLayer::ConvolutionLayer(int input_dimension,
				   int patch_width,
				   int patch_step,
				   int output_width) 
                                     {
 

  expectDimensions(input_dimension, input_dimension * patch_width * patch_width);
 
  SDN_ASSERT(patch_width > 0);
  SDN_ASSERT(patch_step >= 0);
  SDN_ASSERT(output_width >= 0);

  init(patch_width, patch_step, output_width);
}

ConvolutionLayer::ConvolutionLayer(int patch_width, 
				   int patch_step, 
				   int output_width, 
				   boost::shared_ptr<StorageNode> input, 
				   boost::shared_ptr<StorageNode> output) {

  SDN_ASSERT(input != NULL);
  SDN_ASSERT(output != NULL);
  SDN_ASSERT(input->size()*patch_width*patch_width == output->size());

  init(patch_width, patch_step, output_width);
  
  link(input, output);

  checkInvariants();
}

ConvolutionLayer::~ConvolutionLayer() {
}

void ConvolutionLayer::forwardProp() {
  checkInvariants();
  output_->setActivities(convolver_->convolve(input_->getActivities()));
}

void ConvolutionLayer::backProp() {
  checkInvariants();
  input_->setDerivatives(convolver_->deconvolve(output_->getDerivatives()));
}

void ConvolutionLayer::presentTarget(const Matrix& target) {
  assert(output_->getActivities().cols() == target.cols());
  // TODO(asaxe): Prevent ppl from calling this before calling ForwardProp.

  output_->setDerivatives(
			  cost_fn_->jacobian(
			  output_->getActivities(), target).transpose());
  
}

Scalar ConvolutionLayer::calculateError(const Matrix& target) {
  checkInvariants();
  SDN_ASSERT_MSG(output_->getActivities().cols() == target.cols(),
		 "Output cols: " << output_->getActivities().cols() <<
		 " target cols: " << target.cols());
  assert(output_->getActivities().rows() == target.rows());
  Matrix ret = cost_fn_->output(output_->getActivities(), target);
  return ret(0,0);
}

int ConvolutionLayer::patch_width() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->patch_width();
}
int ConvolutionLayer::patch_step() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->patch_step();
}
int ConvolutionLayer::output_width() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->output_width();
}

void ConvolutionLayer::init(int patch_width,
			    int patch_step,
			    int output_width) {
  convolver_ = boost::shared_ptr<Convolver>(new Convolver(patch_width, patch_step, output_width));
  cost_fn_ = boost::shared_ptr<CostFunction>(new SumOfSquaresCostFunction());
}

void ConvolutionLayer::checkInvariants() {
  SDN_ASSERT(input_ != NULL);
  SDN_ASSERT(output_ != NULL);
  SDN_ASSERT(inputDimension()*patch_width()*patch_width() == outputDimension());

  SDN_ASSERT(convolver_ != NULL);
  SDN_ASSERT(cost_fn_ != NULL);
  
}
  
DeconvolutionLayer::DeconvolutionLayer()  {
}

DeconvolutionLayer::DeconvolutionLayer(int output_dimension,
				       int patch_width,
				       int patch_step,
				       int output_width) 
                                      {

  expectDimensions(output_dimension * patch_width * patch_width, output_dimension);
  
  SDN_ASSERT(patch_width > 0);
  SDN_ASSERT(patch_step >= 0);
  SDN_ASSERT(output_width >= 0);

  init(patch_width, patch_step, output_width);
}

DeconvolutionLayer::DeconvolutionLayer(int patch_width, 
				       int patch_step, 
				       int output_width, 
				       boost::shared_ptr<StorageNode> input, 
				       boost::shared_ptr<StorageNode> output) {

  SDN_ASSERT(input != NULL);
  SDN_ASSERT(output != NULL);
  SDN_ASSERT(input->size() == output->size()*patch_width*patch_width);
  
  init(patch_width, patch_step, output_width);
  
  link(input, output);
}


DeconvolutionLayer::~DeconvolutionLayer() {

}

void DeconvolutionLayer::forwardProp() {
  checkInvariants();
  output_->setActivities(convolver_->deconvolve(input_->getActivities()));
}

void DeconvolutionLayer::backProp() {
  checkInvariants();
  input_->setDerivatives(convolver_->convolve(output_->getDerivatives()));
}

void DeconvolutionLayer::presentTarget(const Matrix& target) {
  assert(output_->getActivities().cols() == target.cols());
  // TODO(asaxe): Prevent ppl from calling this before calling ForwardProp.

  output_->setDerivatives(
			  cost_fn_->jacobian(
			  output_->getActivities(), target).transpose());
  
}

Scalar DeconvolutionLayer::calculateError(const Matrix& target) {
  checkInvariants();
  SDN_ASSERT_MSG(output_->getActivities().cols() == target.cols(),
		 "Output cols: " << output_->getActivities().cols() <<
		 " target cols: " << target.cols());
  assert(output_->getActivities().rows() == target.rows());
  Matrix ret = cost_fn_->output(output_->getActivities(), target);
  return ret(0,0);
}

int DeconvolutionLayer::patch_width() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->patch_width();
}
int DeconvolutionLayer::patch_step() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->patch_step();
}
int DeconvolutionLayer::output_width() {
  SDN_ASSERT(convolver_ != NULL);
  return convolver_->output_width();
}

void DeconvolutionLayer::init(int patch_width,
			      int patch_step,
			      int output_width) {
  convolver_ = boost::shared_ptr<Convolver>(new Convolver(patch_width, patch_step, output_width));
  cost_fn_ = boost::shared_ptr<SumOfSquaresCostFunction>(new SumOfSquaresCostFunction());
}

void DeconvolutionLayer::checkInvariants() {
  SDN_ASSERT(input_ != NULL);
  SDN_ASSERT(output_ != NULL);
  SDN_ASSERT_MSG(inputDimension() == outputDimension()*patch_width()*patch_width(),
		 "Left hand side: " << inputDimension()*
		 patch_width()*patch_width()
		 << " Right hand side: " << outputDimension()
		 << " Input dimension: " << inputDimension() 
		 << " Patch width: " << patch_width() 
		 << " output dimension: " << outputDimension());

  SDN_ASSERT(convolver_ != NULL);
  SDN_ASSERT(cost_fn_ != NULL);
  
}
