// Copyright 2009, All Rights Reserved.
// Author: Andrew Saxe (asaxe@stanford.edu)

/** \file convolution-layer.h
    \brief  Declares the interface of a layer that performs convolution 
    by pasting.
*/

#pragma once

#include "layer.h"
#include <Eigen/Core>

namespace sdn 
{

class CostFunction;
class StorageNode;

class Convolver {
 public:
  Convolver(int patch_width, int patch_step, int output_width);
  
  const Matrix convolve(const Matrix& input);

  const Matrix deconvolve(const Matrix& output);

  ///Returns the index into the input array corresponding to (r,c)
  int inputIndex(int r, int c);
  int outputIndex(int r, int c);

  int patch_width();
  int patch_step();
  int output_width();
  
 private:
  int patch_width_;
  int patch_step_;
  int input_width_;
  int output_width_;
  boost::shared_ptr<StorageNode> input_;
  boost::shared_ptr<StorageNode> output_;
  Vector averaging_coefficients_;
};


///Implements convolution by pasting, and can backpropagate gradients
 class ConvolutionLayer : public Layer {
 public:
  ///Creates an empty convolution layer
  ConvolutionLayer();

  ///Creates a convolution with the specified parameters
  /**
     These parameters deserve a good figure, but for now, they are
     identical to the parameters Ian used in FastEncoder so he will be able
     to answer questions.

     The convolution accepts a 2D input array, and takes patches from this
     input to form a 2D output array. These 2D arrays are not represented
     as matrices, but rather are represented in column major order in the 
     columns of the input and output matrices. Rows indicate features. The
     parameters defining the convolution are related in certain ways:

     input_dimension*patch_width*patch_width = output_dimension

     num_input_cols = patch_width + patch_step*(output_width-1)

     num_output_cols = output_width*output_width

     \param input_dimension The number of features (rows) in the input. (The
     number of rows in the output will be 
     input_dimension*patch_width*patch_width)

     \param patch_width The width of the patch in the input to be extracted.
     In particular, a patch_width by patch_width square of the input will be
     taken to form each column of the output.

     \param patch_step The number of input elements to skip between adjacent
     patches in the input.

     \param output_width The width of the 2D output array to be formed. The
     output array is represented in column major order in the columns of the 
     output.
    
  */
  ConvolutionLayer(int input_dimension,
		   int patch_width,
		   int patch_step,
		   int output_width);
		   
  ///Creates a convolution with the specified parameters
  /**
     These parameters deserve a good figure, but for now, they are
     identical to the parameters Ian used in FastEncoder so he will be able
     to answer questions.

     The convolution accepts a 2D input array, and takes patches from this
     input to form a 2D output array. These 2D arrays are not represented
     as matrices, but rather are represented in column major order in the 
     columns of the input and output matrices. Rows indicate features. The
     parameters defining the convolution must be related in certain ways:

     input_dimension*patch_width*patch_width = output_dimension

     num_input_cols = patch_width + patch_step*(output_width-1)

     num_output_cols = output_width*output_width

     \param input_dimension The number of features (rows) in the input. (The
     number of rows in the output will be 
     input_dimension*patch_width*patch_width)

     \param patch_width The width of the patch in the input to be extracted.
     In particular, a patch_width by patch_width square of the input will be
     taken to form each column of the output.

     \param patch_step The number of input elements to skip between adjacent
     patches in the input.

     \param output_width The width of the 2D output array to be formed. The
     output array is represented in column major order in the columns of the 
     output.
    
     \param input The input storage layer to use. Its size is used as the 
     input dimension.
     
     \param output The output storage layer to use. Its size must be 
     input->size()*patch_width*patch_width.
  */
  ConvolutionLayer(int patch_width, 
		   int patch_step, 
		   int output_width, 
		   boost::shared_ptr<StorageNode> input, 
		   boost::shared_ptr<StorageNode> output);

  ~ConvolutionLayer();

  void forwardProp();
  void backProp();

  void presentTarget(const Matrix& target);
  Scalar calculateError(const Matrix& target);

  ///Returns the index into the input array corresponding to (r,c)
  int inputIndex(int r, int c);
  ///Returns the index into the output array corresponding to (r,c)
  int outputIndex(int r, int c);
  
  int patch_width();
  int patch_step();
  int output_width();
 protected:
  void init(int patch_width, int patch_step, int output_width);
  void checkInvariants();
  
  boost::shared_ptr<Convolver> convolver_;
  boost::shared_ptr<CostFunction> cost_fn_;

  Vector averaging_coefficients_;
};

 class DeconvolutionLayer : public Layer {
 public:
  DeconvolutionLayer();

  DeconvolutionLayer(int input_dimension,
		     int patch_width,
		     int patch_step,
		     int output_width);

  DeconvolutionLayer(int patch_width, 
		     int patch_step,
		     int output_width, 
		     boost::shared_ptr<StorageNode> input, 
		     boost::shared_ptr<StorageNode> output);

  ~DeconvolutionLayer();
  
  void forwardProp();
  void backProp();

  void presentTarget(const Matrix& target);
  Scalar calculateError(const Matrix& target);

  ///Returns the index into the input array corresponding to (r,c)
  int inputIndex(int r, int c);
  ///Returns the index into the output array corresponding to (r,c)
  int outputIndex(int r, int c);

  int patch_width();
  int patch_step();
  int output_width();
 protected:
  void init(int patch_width, int patch_step, int output_width);
  void checkInvariants();
  
  boost::shared_ptr<Convolver> convolver_;
  boost::shared_ptr<CostFunction> cost_fn_;

  Vector averaging_coefficients_;
};

}

//SDN_AUTOREGISTER_H(Layer, ConvolutionLayer);
//SDN_AUTOREGISTER_H(Layer, DeconvolutionLayer);



