/// \author Ian Goodfellow
#include "firingRateTest.h"
#include "../base/attributeParser.h"
#include "../base/registry.h"
#include "storageNode.h"
#include "../gui/patchDisplay.h"
#include "activationRecord.h"
#include "trajectoryManager.h"

//Set to 1 to enable debug mode that displays all the trajectories
#define SHOW_TRAJECTORIES 1
//Set to 1 to calculate global firing rate explicitly and check that my bookkeeping is correct
#define CHECK_GLOBAL 0

using namespace std;
using namespace sdn;
using namespace boost;

/*

The detailed documentation:

The test is run on the hidden units given by network->getTestingNode(). For StackedNetwork
this is the output of the deepest encoding layer.

The dataset argument is ignored.

The dataset that will be used is given by the XMLNode used to load the test.

The test proceeds as follows:

usedPatches = 0
while usedPatches <  numGlobalPatches or some hidden unit does not have numLocalTrajectories non-overlapping trajectories
    Run network on random patch
    Store output
    Update thresholds so that each hidden unit fires at the minimum rate possible that is >= targetFiringRate
    A local trajectory for a hidden unit is a trajectory for which the hidden unit fires above threshold on the central patch of the trajectory
    Two trajectories are non-overlapping if they do not contain the same patches

For each hidden unit
   for each trajectory
      compute firing rate over the trajectory
   average the firing rates over each trajectory
   report local firing rate / global firing rate

*/


std::vector<double> FiringRateTest::getScoresForSign_(shared_ptr<Network> network, double sign)
{
  shared_ptr<StorageNode> temp = network->getTestingNode();
  if (!temp)
    {
      SDN_FATAL("FiringRateTest passed network that doesn't specify which node to run the test on");
    }



  StorageNode & node = * temp;
  int numUsed = 0;
  bool trajectoriesFound = false;

  Matrix inputPatch(Matrix::Zero(network->getInputRows(),network->getInputCols()));

  int rows = node.getActivities().rows();
  int cols = node.getActivities().cols();

  ActivationRecord record(rows, cols, numGlobalPatches_);

  TrajectoryManager trajectoryManager(rows,cols, numLocalTrajectories_, testData_);

  unsigned numAboveThreshold = 0; //above means >=, threshold is record[numEntries-numAboveThreshold]

  double maxIterations = 5000;

  bool debugTrajRan = false;

  while ( (numUsed < numGlobalPatches_ || !trajectoriesFound) && numUsed < maxIterations)
    {
      if (numUsed % 100 == 0)
	{
	  SDN_VERBOSE("FiringRateTest iteration "<<numUsed);
	}


      unsigned index;
      testData_->getRandomIndexedExample(inputPatch, index);
      
      network->processExampleToHighestLayer(inputPatch);
      
      record.add(sign * node.getActivities(), index);
      
      numUsed++;

      if (numUsed == numGlobalPatches_)
	{
	  SDN_MSG("Acquired the necessary "<<numGlobalPatches_<<" global patches");
	}

      if (numUsed >= numGlobalPatches_ || numUsed == maxIterations)
	{	  
	  numAboveThreshold = unsigned(targetFiringRate_ * double(record.numEntries()));
	  
	if (numAboveThreshold == 0)
	{
		numAboveThreshold = 1;
	}

	SDN_ASSERT(numAboveThreshold > 0);

	debugTrajRan = true;
	  trajectoriesFound = trajectoryManager.findTrajectories(record, numAboveThreshold, numUsed == maxIterations);
	}
    }

	SDN_ASSERT(debugTrajRan);

    SDN_ASSERT(numAboveThreshold > 0);

  SDN_VERBOSE("Done finding "<<numLocalTrajectories_<<" trajectories");

  vector<double> indivScores;

  for (int i = 0; i < rows; i++)
    for (int j = 0; j < cols; j++)
      {
	SDN_VERBOSE("Evaluating trajectories for unit ("<<i<<","<<j<<")");

	double threshold = record.getUnitRecord(i,j).at(record.numEntries()-numAboveThreshold).value;
	vector<unsigned> trajectories = trajectoryManager.getTrajectories(i,j);
	vector<double> localRates;

#if SHOW_TRAJECTORIES
	vector<Matrix> temp = testData_->getTrajectoryFromIndex(trajectories[0]);
	PatchDisplay viz("FiringRateTest trajectories",network->getInputCols(),network->getInputRows(),1,temp.size(),numLocalTrajectories_);
#endif


	SDN_ASSERT(trajectories.size());

	for (unsigned k = 0; k < trajectories.size(); k++)
	  {
	    vector<Matrix> trajectory = testData_->getTrajectoryFromIndex(trajectories[k]);

	    unsigned count = 0;

	    for (unsigned l = 0; l < trajectory.size(); l++)
	      {
#if SHOW_TRAJECTORIES
		viz.addPatch(trajectory[l],0,0);
#endif
		network->processExampleToHighestLayer(trajectory[l]);
		if (sign * node.getActivities()(i,j) >= threshold)
		  count++;
	      }

		SDN_ASSERT(trajectory.size());
	    
	    localRates.push_back(double(count)/double(trajectory.size()));

		SDN_ASSERT(!isnan(localRates.back()));
		SDN_ASSERT(!isnan(mean(localRates)));
	    //SDN_DEBUG("\t\tLocal firing rate: "<<localRates.back());
	  }

#if SHOW_TRAJECTORIES
	viz.enableWindow();
	SDN_MSG("Waiting for you to press a key...");
	viz.wait(0);
#endif

	
	/// Calculate the true global firing rate of this unit, accounting for ties
	ActivationInstance dummy;
	dummy.value = threshold;	
	UnitActivationRecord & firingRecord = record.getUnitRecord(i,j);
	UnitActivationRecord::iterator iter = lower_bound(firingRecord.begin(),firingRecord.end(),dummy);
	unsigned dist = distance(firingRecord.begin(), iter);

	SDN_ASSERT_MSG(dist != firingRecord.size(), " couldn't insert threshold "<<threshold<<" into list ending in "<<firingRecord.back().value);

	SDN_ASSERT(firingRecord[dist].value >= threshold);
	if (dist > 0)
	  {
	    SDN_ASSERT(firingRecord[dist-1].value <= threshold);
	  }
	       
	double globalFiringRate = double(firingRecord.size()-dist)/double(firingRecord.size());

	SDN_ASSERT_MSG(globalFiringRate > 0, "globalFiringRate="<<globalFiringRate);	



#if CHECK_GLOBAL

	SDN_DEBUG("Checking global firing rate");

	unsigned count = 0;

	for (unsigned e = 0; e < firingRecord.size(); e++)
	  {
	    Matrix patch(Matrix::Zero(network->getInputRows(),network->getInputCols()));
	    testData_->getExampleFromIndex(patch,firingRecord[e].index);
	    
	    network->processExampleToHighestLayer(patch);

	    if (node.getActivities()(i,j) >= threshold)
	      {
		count++;
	      }
	  }

	double explicitGlobal = double(count)/double(firingRecord.size());

	SDN_ASSERT_MSG(globalFiringRate == explicitGlobal, "global firing rate failed: "<<globalFiringRate<<" vs. "<<explicitGlobal);

	SDN_MSG("Global check passed");
#endif



	//SDN_DEBUG("\tglobal firing rate: "<<globalFiringRate);
	//SDN_DEBUG("\ttarget firing rate: "<<targetFiringRate_);

	SDN_ASSERT(localRates.size())
	SDN_ASSERT(!isnan(mean(localRates)));
	SDN_ASSERT(globalFiringRate > 0);
	SDN_ASSERT(!isnan(globalFiringRate));

	indivScores.push_back( mean(localRates) / globalFiringRate);

	SDN_DEBUG("Score so far: "<<mean(indivScores));

	SDN_ASSERT(!isnan(indivScores.back()));

	//SDN_DEBUG("\tindiv score: " <<indivScores.back());
      }      

return indivScores;
}


std::vector<double> FiringRateTest::run(boost::shared_ptr<Network> network, boost::shared_ptr<Dataset> dataset)
{
  testData_->requestExampleSize(network->getInputRows(),network->getInputCols());

  //Ignore the dataset argument, we don't care what the network is trained on
  dataset.reset();


  vector<double> indivScoresPos = getScoresForSign_(network, 1);
  //vector<double> indivScoresNeg = getScoresForSign_(network, -1);

  vector<double> indivScores;

  //for (unsigned i = 0; i < indivScoresPos.size(); i++)
    //indivScores.push_back(max(indivScoresPos[i],indivScoresNeg[i]));



  vector<double> rval;
  
  rval.push_back(mean(indivScoresPos));


  SDN_VERBOSE("FiringRateTest: final score "<<rval.back());
  return rval;
}

void FiringRateTest::load(XMLNode & node)
{
  SDN_ASSERT(node.getName());
  SDN_ASSERT(string(node.getName()) == "sdn::Test");
  SDN_ASSERT(node.getAttribute("id"));
  SDN_ASSERT(string(node.getAttribute("id")) == "sdn::FiringRateTest");


  AttributeParser p;
  p.addInt("numGlobalPatches",1000);
  p.addInt("numLocalTrajectories",10);
  p.addDouble("targetFiringRate",0.01);
  p.addString("id");

  p.parse(node);

  numGlobalPatches_ = p.getInt("numGlobalPatches").value;
  numLocalTrajectories_ = p.getInt("numLocalTrajectories").value;
  targetFiringRate_ = p.getDouble("targetFiringRate").value;

  if (node.nChildNode() != 1)
    {
      SDN_FATAL("FiringRateTest requires a single XML node specifying the dataset to run on");
    }

  XMLNode dataNode = node.getChildNode(0);
  testData_ = DatasetRegistry().load(dataNode);

  if (!testData_)
    {
      SDN_FATAL("FiringRateTest could not load the requested dataset");
    }

  SDN_ASSERT(testData_->supportsTrajectoryIndexing());
}

