/// \author Ian Goodfellow
#include "implicitPositiveTest.h"
#include "../base/attributeParser.h"
#include "../base/registry.h"
#include "storageNode.h"
#include "../gui/patchDisplay.h"
#include "activationRecord.h"
#include "trajectoryManager.h"
#include <iostream>
#include "classificationPRCurve.h"

//Set to 1 to enable debug mode that displays all the trajectories
#define SHOW_TRAJECTORIES 0
//Set to 1 to enable debug mode that shows the PR curves
#define SHOW_PR_CURVES 0

using namespace std;
using namespace sdn;
using namespace boost;

/*

The detailed documentation:

The test is run on the hidden units given by network->getTestingNode(). For StackedNetwork
this is the output of the deepest encoding layer.

The dataset argument is ignored.

The dataset that will be used is given by the XMLNode used to load the test.

The test proceeds as follows:

usedPatches = 0
while usedPatches <  numGlobalPatches or some hidden unit does not have numLocalTrajectories non-overlapping trajectories
    Run network on random patch
    Store output
    Update thresholds so that each hidden unit fires at the minimum rate possible that is >= targetFiringRate
    A local trajectory for a hidden unit is a trajectory for which the hidden unit fires above threshold on the central patch of the trajectory
    Two trajectories are non-overlapping if they do not contain the same patches

For each hidden unit
   Treat all patches in local trajectories as "positive examples"
   Treat all patches in global sample whose trajectories do not overlap with the positive trajectories as "negative examples"
   The hidden unit's score include max F_1 score and area under PR curve

*/

std::vector<double> ImplicitPositiveTest::run(boost::shared_ptr<Network> network, boost::shared_ptr<Dataset> dataset)
{
  testData_->requestExampleSize(network->getInputRows(),network->getInputCols());

  //Ignore the dataset argument, we don't care what the network is trained on
  dataset.reset();

  shared_ptr<StorageNode> temp = network->getTestingNode();
  if (!temp)
    {
      SDN_FATAL("ImplicitPositiveTest passed network that doesn't specify which node to run the test on");
    }

  StorageNode & node = * temp;


  int numUsed = 0;
  bool trajectoriesFound = false;

  Matrix inputPatch(Matrix::Zero(network->getInputRows(),network->getInputCols()));

  int rows = node.getActivities().rows();
  int cols = node.getActivities().cols();

  ActivationRecord record(rows, cols, numGlobalPatches_);

  TrajectoryManager trajectoryManager(rows,cols, numLocalTrajectories_, testData_);

  unsigned numAboveThreshold = 0; //above means >=, threshold is record[numEntries-numAboveThreshold]

  while (numUsed < numGlobalPatches_ || !trajectoriesFound)
    {
      if (numUsed % 100 == 0)
	{
	  SDN_VERBOSE("ImplicitPositiveTest iteration "<<numUsed);
	}


      unsigned index;
      testData_->getRandomIndexedExample(inputPatch, index);
      
      network->processExampleToHighestLayer(inputPatch);
      
      record.add(node.getActivities(), index);
      
      numUsed++;

      if (numUsed == numGlobalPatches_)
	{
	  SDN_MSG("Acquired the necessary "<<numGlobalPatches_<<" global patches");
	}

      if (numUsed >= numGlobalPatches_)
	{	  
	  numAboveThreshold = unsigned(targetFiringRate_ * double(record.numEntries()));
	  
	  trajectoriesFound = trajectoryManager.findTrajectories(record, numAboveThreshold, false);
	}
    }

  SDN_VERBOSE("Done finding "<<numLocalTrajectories_<<" trajectories");

  vector<double> indivFScores;
  vector<double> indivPRAreas;

  for (int i = 0; i < rows; i++)
    for (int j = 0; j < cols; j++)
      {
	SDN_VERBOSE("Evaluating trajectories for unit ("<<i<<","<<j<<")");

	vector<unsigned> trajectories = trajectoryManager.getTrajectories(i,j);
	vector<double> posScores;

#if SHOW_TRAJECTORIES
	vector<Matrix> temp = testData_->getTrajectoryFromIndex(trajectories[0]);
	PatchDisplay viz("ImplicitPositiveTest trajectories",network->getInputCols(),network->getInputRows(),1,temp.size(),numLocalTrajectories_);
#endif

	for (unsigned k = 0; k < trajectories.size(); k++)
	  {
	    vector<Matrix> trajectory = testData_->getTrajectoryFromIndex(trajectories[k]);

	    for (unsigned l = 0; l < trajectory.size(); l++)
	      {
#if SHOW_TRAJECTORIES
		viz.addPatch(trajectory[l],0,0);
#endif
		network->processExampleToHighestLayer(trajectory[l]);
		posScores.push_back(node.getActivities()(i,j));
	      }	

	    //SDN_DEBUG("\t\tLocal firing rate: "<<localRates.back());
	  }

#if SHOW_TRAJECTORIES
	viz.enableWindow();
	SDN_MSG("Waiting for you to press a key...");
	viz.wait(0);
#endif

	
	//Now we go through and find all the "negative examples"-- things we ran the network on before whose trajectories don't overlap with positive trajectories
	UnitActivationRecord r = record.getUnitRecord(i,j);

	vector<double> negScores;

	for (unsigned k = 0; k < r.size(); k++)
	  {
	    bool include = true;
	    for (unsigned l = 0; l < trajectories.size(); l++)
	      {
		if (testData_->trajectoryOverlap(r[k].index,trajectories[l]))
		  {
		    include = false;
		    break;
		  }		    
	      }
	    if (include)
	      {
		negScores.push_back(r[k].value);
	      }
	  }


	ClassificationPRCurve pr(posScores,negScores);


	indivPRAreas.push_back(pr.area());
	indivFScores.push_back(pr.maxF());

	SDN_DEBUG("pr area: "<<indivPRAreas.back());
	SDN_DEBUG("f score: "<<indivFScores.back());

	#if SHOW_PR_CURVES
	pr.show();
	char ch;
	cin >> ch;
	#endif
	
	//SDN_DEBUG("\tindiv score: " <<indivScores.back());
      }      

  vector<double> rval;
  
  rval.push_back(mean(indivFScores));
  rval.push_back(mean(indivPRAreas));

  SDN_VERBOSE("ImplicitPositiveTest: final F score "<<rval[0]);
  SDN_VERBOSE("ImplicitPositiveTest: final area under PR curve "<<rval[1]);
  return rval;
}

void ImplicitPositiveTest::load(XMLNode & node)
{
  SDN_ASSERT(node.getName());
  SDN_ASSERT(string(node.getName()) == "sdn::Test");
  SDN_ASSERT(node.getAttribute("id"));
  SDN_ASSERT(string(node.getAttribute("id")) == "sdn::ImplicitPositiveTest");


  AttributeParser p;
  p.addInt("numGlobalPatches",1000);
  p.addInt("numLocalTrajectories",10);
  p.addDouble("targetFiringRate",0.01);
  p.addString("id");

  p.parse(node);

  numGlobalPatches_ = p.getInt("numGlobalPatches").value;
  numLocalTrajectories_ = p.getInt("numLocalTrajectories").value;
  targetFiringRate_ = p.getDouble("targetFiringRate").value;

  if (node.nChildNode() != 1)
    {
      SDN_FATAL("ImplicitPositiveTest requires a single XML node specifying the dataset to run on");
    }

  XMLNode dataNode = node.getChildNode(0);
  testData_ = DatasetRegistry().load(dataNode);

  if (!testData_)
    {
      SDN_FATAL("ImplicitPositiveTest could not load the requested dataset");
    }

  SDN_ASSERT(testData_->supportsTrajectoryIndexing());
}

