#include <Eigen/Core>
#include "sdn/dataset.h"
#include "network.h"
#include <vector>
#include <iostream>  // I/O 
#include "invarianceTest.h"
#include "templateMatching.h"
#include "maxpoolingTemplateMatching.h"
#include "boost/shared_ptr.hpp"

using namespace std;
using namespace boost;
USING_PART_OF_NAMESPACE_EIGEN


namespace sdn{
  

  void computeThresholdAndGlobalCount(std::vector<Scalar> activations,
				      const Scalar quantile,
				      Scalar &threshold,
				      Scalar &globalCount)
  {
    SDN_ASSERT(quantile <= 1 && quantile >= 0);
    vector<int> indices(activations.size());
    iota(indices.begin(), indices.end(), 0);
    indirect_less_than<Scalar> igt(&(activations[0]));
    sort(indices.begin(), indices.end(),igt);
    
    int index_quantile = (int) (indices.size()*quantile);
    threshold = activations[indices[index_quantile]];
    globalCount = 0;//indices.size() - index_quantile;
    for (int i=0;i<activations.size();i++){
      if (activations[i] >= threshold)
	globalCount = globalCount + 1;
    }
  }
  

  // compute average firing rate of top scoring neurons in a network with sign compensation
  void computeThresholdScore(const Matrix localPositive, 
			     const Matrix middlePositive,
			     const Matrix globalPositive,
			     const Matrix localNegative, 
			     const Matrix middleNegative,
			     const Matrix globalNegative,
			     Scalar neuronFraction,
			     Scalar &networkScore)
  {
    SDN_ASSERT(localPositive.rows() == middlePositive.rows() && 
	       localPositive.cols() == middlePositive.cols());
    SDN_ASSERT(localNegative.rows() == middleNegative.rows() && 
	       localNegative.cols() == middleNegative.cols());
    SDN_ASSERT(localNegative.rows() == localPositive.rows() && 
	       localNegative.cols() == localPositive.cols());
    vector<Scalar> max_scores;
    for (int i=0;i<localPositive.rows();i++){
      for (int j=0;j<localPositive.cols();j++){
	Scalar positive_score;
	if (middlePositive(i,j) != 0){
	  positive_score = localPositive(i,j) / middlePositive(i,j) / globalPositive(i,j);
	}
	else{
	  positive_score = 0;
	}
	Scalar negative_score;
	if (middleNegative(i,j) != 0){
	  negative_score = localNegative(i,j) / middleNegative(i,j) / globalNegative(i,j);
	}
	else{
	  negative_score = 0;
	}
	Scalar score = MAX(positive_score, negative_score);
	max_scores.push_back(score);
      }
    }
    vector<int> indices(max_scores.size());
    iota(indices.begin(), indices.end(), 0);
    indirect_greater_than<Scalar> igt(&(max_scores[0]));
    sort(indices.begin(), indices.end(),igt);
    int k = (int) (((float)neuronFraction * (float) localPositive.rows() * (float) localPositive.cols()) +0.5);
    Scalar mean_score = 0;
    for (int i=0; i<k;i++){
      mean_score = mean_score + max_scores[indices[i]];
    }
    networkScore = mean_score / k;
    cout << "networkScore for this test : " << networkScore << endl;
  }

  
  // this function returns the threshold that activates quantile percentage of cells
  // and use this for counting the number of cells that activate globally
  void computeGlobalStatistics(shared_ptr<Network> network,
			       InvarianceTestDataset &itd, 
			       int sign,
			       Scalar quantile,
			       Matrix &thresholds,
			       Matrix &globalCounts)
  {
    SDN_ASSERT(sign==-1 || sign==1);
    vector<Matrix> exhaustivePatches;
    itd.getExhaustivePatches(exhaustivePatches);
    vector<Matrix> responses;
    
    vector<Scalar> activations;
    // run through all random patches
    for (unsigned int k=0;k<exhaustivePatches.size();k++){
      // collect the response for a random patch
      network->processExampleToHighestLayer(exhaustivePatches[k]);
      Matrix highestLayer = network->getHighestLayerActivities();
      responses.push_back(highestLayer);
    }		
    thresholds = Matrix::Zero(responses[0].rows(),responses[0].cols());
    globalCounts = Matrix::Zero(responses[0].rows(),responses[0].cols());
    for (int i=0;i<responses[0].rows();i++){
      for (int j=0;j<responses[0].cols();j++){
	vector<Scalar> activations;
	for (unsigned int k=0;k<responses.size();k++){
	  activations.push_back(sign*responses[k](i,j));
	}
	Scalar threshold;
	Scalar globalCount;
	computeThresholdAndGlobalCount(activations, quantile, threshold, globalCount);
	thresholds(i,j) = threshold;
	globalCounts(i,j) = globalCount;
      }
    }
    globalCounts = globalCounts / responses.size();
  }
  
  // compute the firing rate for local transformed images
  void computeLocalStatistics(shared_ptr<Network> network,
			      int inv_type, 
			      int sign,
			      InvarianceTestDataset &itd, 
			      Matrix thresholds, 
			      Matrix &localCounts,
			      Matrix &middleCounts)
  {
    SDN_ASSERT(sign==NEGATIVE || sign==POSITIVE);
    vector<Matrix> inexhaustivePatches;
    vector<Scalar> inexhaustivePhases;
    vector<Scalar> inexhaustiveOrientations;
    vector<Scalar> inexhaustiveFrequencies;
    
    itd.getInexhaustivePatches(inv_type, 
			       inexhaustivePatches);  
    network->processExampleToHighestLayer(inexhaustivePatches[0]);
    Matrix highestLayer = network->getHighestLayerActivities();
    int nHiddenRows = highestLayer.rows();
    int nHiddenCols = highestLayer.cols();
    
    localCounts = Matrix::Zero(nHiddenRows, nHiddenCols);
    middleCounts = Matrix::Zero(nHiddenRows, nHiddenCols);
    // run through each hidden unit
    for (unsigned int i=0;i<inexhaustivePatches.size();i++){
      // check hidden unit (r,c) response for a particular patch index by i
      network->processExampleToHighestLayer(inexhaustivePatches[i]);
      highestLayer = network->getHighestLayerActivities();
      for (int r=0;r<nHiddenRows;r++){
	for (int c=0;c<nHiddenCols;c++){
	  Scalar val = sign*highestLayer(r,c);
	  // run through each patch in the middle frame
	  // if the response is higher than a threshold, we can use this as a stimulus
	  if (val >= thresholds(r,c)){
	    middleCounts(r,c) = middleCounts(r,c) + 1;
	    // get locally varying data
	    vector<Matrix> locallyVaryingPatches;
	    itd.getLocalPatches(inv_type, i, locallyVaryingPatches);
	    
	    // run through network trajectory and check the response
	    for (unsigned int k=0;k<locallyVaryingPatches.size();k++){
	      network->processExampleToHighestLayer(locallyVaryingPatches[k]);
	      Matrix highestLayer_temp = network->getHighestLayerActivities();
	      Scalar trajVal = sign*highestLayer_temp(r,c);
	      if (trajVal >= thresholds(r,c)){
		localCounts(r,c) = localCounts(r,c) + 1.0 / (locallyVaryingPatches.size()-1);
	      }
	    }
	    localCounts(r,c) = localCounts(r,c) - 1.0 / (locallyVaryingPatches.size()-1);
	  }
	}
      }
    }
  }


  // perform firing rate test, local over global firing rates
  void firingRateTest(shared_ptr<Network> network,
		      InvarianceTestDataset &itd,
		      Scalar quantile,
		      Scalar neuronFraction,
		      Scalar &networkScoreOut,
		      Scalar &templateScoreOut,
		      Scalar &maxpoolingScoreOut,
		      vector<Scalar> &scorePerCategory)
  {
    Scalar networkScore = 0;
    Scalar tmScore = 0;
    Scalar mtmScore = 0;
    Matrix networkThresholdsPositive;
    Matrix globalPositive;
    Matrix networkThresholdsNegative;
    Matrix globalNegative;
    int counter = 0;
    int numCategory = itd.getNumCategory();
    vector<int> counterPerCategory;
    for (int i=0;i<numCategory;i++){
      scorePerCategory.push_back(0);
      counterPerCategory.push_back(0);
    }
    
    computeGlobalStatistics(network, itd, POSITIVE, quantile, 
			    networkThresholdsPositive, 
			    globalPositive);
    computeGlobalStatistics(network, itd, NEGATIVE, quantile, 
			    networkThresholdsNegative, 
			    globalNegative);
    vector<Scalar> scores;

    for (int inv_type = 0;inv_type<itd.getNumCollection();inv_type++){
      cout << "invariance type = " << itd.getType(inv_type) << endl;
      // compute local firing rate with sign = positive
      Matrix localPositive;
      Matrix middlePositive;
      computeLocalStatistics(network,
			     inv_type, POSITIVE, itd,  
			     networkThresholdsPositive,  
			     localPositive, 
			     middlePositive);
      // compute local firing rate with sign = negative
      Matrix localNegative;
      Matrix middleNegative;
      computeLocalStatistics(network, 
			     inv_type, NEGATIVE, itd,  
			     networkThresholdsNegative, 
			     localNegative, 
			     middleNegative);
      // use both local firing rates to compute a firing rate score
      Scalar score = 0;
      computeThresholdScore(localPositive, 
			    middlePositive,
			    globalPositive,
			    localNegative, 
			    middleNegative,
			    globalNegative,
			    neuronFraction,
			    score);
      int category = itd.getCategory(inv_type);
      scorePerCategory[category] = scorePerCategory[category] + score;
      counterPerCategory[category] = counterPerCategory[category] + 1;
      

      //computeThresholdScore(networkLocalCounts, 
      //middleCounts,
      //networkGlobalCounts,
      //score);
      networkScore = networkScore + score;
      scores.push_back(score);
      counter = counter + 1;
    }
    for (unsigned int i=0;i<scorePerCategory.size();i++){
      scorePerCategory[i] = scorePerCategory[i] / counterPerCategory[i];
    }
    cout << "*all scores ";
    for (unsigned int i=0;i<scores.size();i++){
      cout << scores[i] << " ";
    }
    cout << endl;
    
    networkScoreOut = networkScore /counter;
    templateScoreOut = tmScore /counter;
    maxpoolingScoreOut = mtmScore /counter;
  }

}
