#include "invariantNetwork.h"
#include "sdn/base.h"
#include "sdn/gui.h"
#include <cmath>
#include <sstream>
#include <algorithm>
#include <string>
#include "Eigen/Core"
#include "Eigen/SVD"
#include <time.h>
#include <fstream>

using namespace sdn;

#define SLOWROT_EXAMPLES 100000
#define CIFAR_CLASSES  10
#define CIFAR_EXAMPLES 150000 // 50000
#define CIFAR_TEST_EXAMPLES 30000 // 10000
#define CIFAR_IPE 1 // 3
#define NORB_CLASSES 5
#define NORB_EXAMPLES 24300
#define NORB_TEST_EXAMPLES 24300
#define NORB_IPE 18
#define MAX_GPU_INPUT_SIZE (3 << 29)

InvariantNetwork::InvariantNetwork(int neighborhoodSize, int W_rows) {
  Wr = W_rows;
  ns = neighborhoodSize;
  C = .01;
  int dim = (int)sqrt(Wr+0.01);
  xdim = ydim = dim;
  update_w = update_v = false;
}

Matrix InvariantNetwork::neighborhoodTimes(Matrix m) {
  Matrix ret = m;
  bool computeNRs = false;
  if(neighborhoodRows.rows() != m.rows()) {
    neighborhoodRows = Matrix::Ones(m.rows(), 1);
  }
  for(int dr = -ns; dr <= ns; dr++)
    for(int dc = -ns; dc <= ns; dc++) {
      if(dr * dr + dc * dc > ns * ns)
        continue;
      if(dr == 0 && dc == 0)
        continue;
      int dif = ydim * dr + dc;
      if(dif > 0) {
        ret.block(0, 0, ret.rows() - dif, ret.cols()) +=
          m.block(dif, 0, ret.rows() - dif, ret.cols());
        if(computeNRs)
          neighborhoodRows.block(0, 0, ret.rows() - dif, 1).cwise() += 1;
      } else {
        ret.block(-dif, 0, ret.rows() + dif, ret.cols()) +=
          m.block(0, 0, ret.rows() + dif, ret.cols());
        if(computeNRs)
          neighborhoodRows.block(-dif, 0, ret.rows() + dif, 1).cwise() += 1;
      }
    }
  if(computeNRs)
    neighborhoodRows = neighborhoodRows.cwise().sqrt().cwise().inverse();
  for(int i = 0; i < neighborhoodRows.rows(); i++)
    ret.row(i) *= neighborhoodRows(i,0);
  return ret;
}


void InvariantNetwork::initRandWeights() {
  W.resize(Wr, N);
  //int dim = (int)sqrt(Wr+0.01);
  //V = neighborhoodMatrix(dim, dim, ns);

  for(int i = 0; i < W.rows(); i++)
    for(int j = 0; j < W.cols(); j++)
      W(i,j) = (1.0/sqrt(N))*(2.0*rand()-RAND_MAX)/RAND_MAX;

  update_w = true;
  //update_v = true;
  performProjections();
}

void InvariantNetwork::readInput(std::ifstream& data, DATASET D)
{
  SDN_DEBUG("Reading data file...");
  N = -1;
  std::string line;
  int index = 0;
  while(getline(data, line)) {
    std::stringstream ss;
    ss << line;

    Scalar val;
    std::vector<Scalar> cur;
    while(ss >> val)
      cur.push_back(val);

    if(N == -1) {
      N = (int)cur.size();
      switch(D){
      case SLOWROT:
        inputs.resize(N, SLOWROT_EXAMPLES);
        break;
      case CIFAR:
        inputs.resize(N, CIFAR_EXAMPLES*CIFAR_IPE);
        break;
      case CIFAR_LABELS:
        multilabels = Matrix::Zero(CIFAR_CLASSES-1, CIFAR_EXAMPLES);
        labels.clear();
        break;
      case CIFAR_TEST:
        inputs.resize(N, CIFAR_TEST_EXAMPLES*CIFAR_IPE);
        break;
      case CIFAR_TEST_LABELS:
        labels.clear();
        break;
      case NORB:
        inputs.resize(N, NORB_EXAMPLES*NORB_IPE);
        break;
      case NORB_LABELS:
        multilabels = Matrix::Zero(NORB_CLASSES-1, NORB_EXAMPLES);
        labels.clear();
        break;
      case NORB_TEST:
        inputs.resize(N, NORB_TEST_EXAMPLES*NORB_IPE);
        break;
      case NORB_TEST_LABELS:
        labels.clear();
        break;
      }
    }
    else if((int)cur.size() != N) {
      SDN_FATAL("Inputs are not all the same size!");
    }

    // temporaries for processing cifar
    int ch;
    int nExamples = 0, col, exNum = 0;
    
    switch(D){
    case SLOWROT:
    case NORB:
    case NORB_TEST:
      for(int i = 0; i < N; i++)
        inputs(i,index) = cur[i];
      break;
    case CIFAR:
      nExamples = CIFAR_EXAMPLES; 
    case CIFAR_TEST:
      if(nExamples == 0)
        nExamples = CIFAR_TEST_EXAMPLES;
      ch = index / nExamples;
      exNum = index % nExamples;
      col = exNum*CIFAR_IPE+ch;
      for(int i = 0; i < N; i++)
        inputs(i,col) = cur[i];
      break;
    case CIFAR_LABELS:
      if(index < CIFAR_EXAMPLES) {
        if (cur[0] < CIFAR_CLASSES-1)
          multilabels((int)cur[0],index) = 1;
        labels.push_back(cur[0]);
      }
      break;
    case NORB_LABELS:
      if (cur[0] < NORB_CLASSES-1)
        multilabels((int)cur[0],index) = 1;
      labels.push_back(cur[0]);
      break;
    case CIFAR_TEST_LABELS:
      if(index < CIFAR_TEST_EXAMPLES) {
        labels.push_back(cur[0]);
      }
      break;
    case NORB_TEST_LABELS:
      labels.push_back(cur[0]);
      break;
    default:
      break;
    }
    index++;
  }
  SDN_DEBUG("Done reading data; input size: " << N);
}

void InvariantNetwork::setInputsOnCPU(bool onCPU) {
  inputs_on_cpu = onCPU;
}

void InvariantNetwork::initWeight(std::ifstream& data, WEIGHT_MATRIX M)
{
  SDN_DEBUG("Reading weight matrix file...");
  N = -1;
  std::string line;
  std::vector<std::vector<Scalar> > buffer; 

  while(getline(data, line)) {
    std::stringstream ss;
    ss << line;

    Scalar val;
    std::vector<Scalar> cur;
    while(ss >> val)
      cur.push_back(val);

    if(N == -1) {
      N = (int)cur.size();
    }
    
    else if((int)cur.size() != N) {
      SDN_FATAL("Inputs are not all the same size!");
    }
    
    buffer.push_back(cur);
  }

  switch(M){
  case W_MAT:
    W.resize(buffer.size(), buffer[0].size());
    Wr = W.rows();
    for(int i = 0; i < (int)buffer.size(); i++)
      for(int j = 0; j < (int)buffer[i].size(); j++)
        W(i,j) = buffer[i][j];
    break;
  case V_MAT:
    V.resize(buffer.size(), buffer[0].size());
    T = Matrix::Zero(CIFAR_CLASSES, V.rows());
    for(int i = 0; i < (int)buffer.size(); i++)
      for(int j = 0; j < (int)buffer[i].size(); j++)
        V(i,j) = buffer[i][j];
    break;
  case T_MAT:
    T.resize(buffer.size(), buffer[0].size());
    for(int i = 0; i < (int)buffer.size(); i++)
      for(int j = 0; j < (int)buffer[i].size(); j++)
        T(i,j) = buffer[i][j];
    break;
  default:
    break;
  }
  
  SDN_DEBUG("Done initializing weight matrix");
}

bool doIsNan = false;
static bool matrixIsNan(Matrix m) {
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      if(isnan(m(i,j)))
        return true;
  return false;
}

void InvariantNetwork::feedForward() {
  // feed forward to layer 2
  z2 = W * a1;

  switch(act_f) {
  case SQUARE_SQRT_ACT:
    // square activation (layer 2)
    a2 = z2.cwise().square();
    break;
  case LOGCOSH_LINEAR_ACT:
    // log cosh activation (layer 2)
    a2 = coshM(z2).cwise().log();
    break;
  }

  // feed forward to layer 3
  z3 = ns == 0 ? a2 : neighborhoodTimes(a2); // V * a2;
  
  switch(act_f) {
  case SQUARE_SQRT_ACT:
    // sqrt activation (layer 3)
    a3 = (z3.cwise() + C).cwise().sqrt();
    break;
  case LOGCOSH_LINEAR_ACT:
    // linear activation (layer 3)
    a3 = z3;
    break;
  }

  if(doIsNan) {
    SDN_DEBUG("Is z2 Nan: " << matrixIsNan(z2));
    SDN_DEBUG("Is z3 Nan: " << matrixIsNan(z3));
    SDN_DEBUG("Min z2: " << z2.minCoeff());
    SDN_DEBUG("Max z2: " << z2.maxCoeff());
    SDN_DEBUG("Min z3: " << z3.minCoeff());
    SDN_DEBUG("Max z3: " << z3.maxCoeff());
    SDN_DEBUG("Min W: " << W.minCoeff());
    SDN_DEBUG("Max W: " << W.maxCoeff());
    SDN_DEBUG("Is a3 Nan: " << matrixIsNan(a3));
  }
}

void InvariantNetwork::concatOutputs(int inputsPerExample) {
  if(inputsPerExample > 1) {
    a3t.resize(a3.rows() * inputsPerExample, a3.cols() / inputsPerExample);
    for(int i = 0; i < a3t.cols(); i++) {
      for(int j = 0; j < inputsPerExample; j++) {
        a3t.block(a3.rows() * j, i, a3.rows(), 1) =
          a3.col(inputsPerExample * i + j);
      }
    }
  }
  else {
    a3t = a3;
  }
}

void InvariantNetwork::feedForwardSoftmax() {
  Matrix z4temp = T * a3t;

  Matrix z4Maxes = z4temp.colwise().maxCoeff();    
  Matrix temp = z4temp;
  for(int i = 0; i < temp.rows(); i++)
    temp.row(i) -= z4Maxes;
  Matrix logsumexp = z4Maxes + temp.cwise().exp().colwise().sum().cwise().log();
  
  for(int i = 0; i < z4temp.rows(); i++)
    z4temp.row(i) -= logsumexp;
  z4 = z4temp.cwise().exp().block(0, 0, z4temp.rows()-1, z4temp.cols());
}

/*
void InvariantNetwork::feedForwardInput() {
  //inputs = 

}
*/

Scalar InvariantNetwork::backProp(Scalar oldError) {
  Matrix grad_W, grad_V;

  {
    // compute layer 3 derivatives
    Matrix da3; computeDerivatives(da3);
    
    Matrix delta3;
    switch(act_f) {
    case SQUARE_SQRT_ACT:
      // sqrt activation derivative (layer 3)
      delta3 = da3.cwise() * (0.5 * a3.cwise().inverse());
      break;
    case LOGCOSH_LINEAR_ACT:
      // linear activation derivative (layer 3)
      delta3 = da3;
      break;
    }

    // compute layer 2 derivatives
    Matrix da2 = ns == 0 ? delta3 : neighborhoodTimes(delta3);//V.transpose() * delta3;
   
    Matrix delta2;
    switch(act_f) {
    case SQUARE_SQRT_ACT:
      // square activation derivative (layer 2)
      delta2 = da2.cwise() * (2 * z2);
      break;
    case LOGCOSH_LINEAR_ACT:
      // log cosh activation derivative (layer 2)
      delta2 = da2.cwise() * tanhM(z2);
      break;
    }

    // compute gradients
    //grad_V = delta3 * a2.transpose();
    grad_W = delta2 * a1.transpose();

    // regularization
    //grad_V -= 2 * lambdaV * (Matrix::Ones(V.rows(), 1) * V.colwise().sum() - V);
    grad_W -= 4 * lambdaW * (W * W.transpose() * W - (W.rowwise().squaredNorm() * Matrix::Ones(1, W.cols())).cwise() * W);
  }

  oldError = computeCost();
  //Scalar oldError = computeWholeBatchCost(oldIndex,  batches, batchSize);

  // computing numerical derivatives
  /*
  Scalar eps = 1e-5;
  SDN_DEBUG("Computing numerical derivatives");
  
  for(int i = 0; i < W.rows(); i++) {
    for(int j = 0; j < W.cols(); j++) {
      W(i,j) += eps;
      
      feedForward();
      Scalar newError = computeCost();
      Scalar numDeriv = (newError - oldError) / eps;
      Scalar diff = std::abs(numDeriv - grad_W(i,j));
      if(diff / std::max(std::abs(numDeriv), std::abs(grad_W(i,j))) > 0.01) {
        SDN_DEBUG("Numerical derivative does not match: " << numDeriv << " "
                  << grad_W(i,j));
      }
      else {
        SDN_DEBUG("Numerical derivative ok: " << numDeriv << " " << grad_W(i,j));
      }

      W(i,j) -= eps;
    }
  }
  
  for(int i = 0; i < V.rows(); i++) {
    for(int j = 0; j < V.cols(); j++) {
      V(i,j) += eps;
      
      feedForward();
      Scalar newError = computeCost();
      Scalar numDeriv = (newError - oldError) / eps;
      Scalar diff = std::abs(numDeriv - grad_V(i,j));
      if(diff / std::max(std::abs(numDeriv), std::abs(grad_V(i,j))) > 0.01) {
        SDN_DEBUG("Numerical derivative does not match: " << numDeriv << " "
                  << grad_V(i,j));
      }
      else {
        SDN_DEBUG("Numerical derivative ok: " << numDeriv << " " << grad_V(i,j));
      }

      V(i,j) -= eps;
    }
  }
  */

  Scalar alpha_star = alpha;
  Scalar newError;
  Matrix oldW = W, oldV = V;
  do {
    if(update_w) {
      W = oldW + alpha_star * grad_W;
    }
    if(update_v) {
      V = oldV + alpha_star * grad_V;
      if(hasNegativeV()) {
        newError = oldError - 1;
        alpha_star /= 2;
        continue;
      }
    }

    performProjections();
    feedForward();
    newError = computeCost();
    //newError = computeWholeBatchCost(oldIndex,  batches, batchSize);

    //SDN_DEBUG("Done computing objective for alpha = " << alpha_star);
    
    alpha_star /= 2;
  } while (newError <= oldError && alpha_star > 1e-10);
  SDN_DEBUG("Learning rate used: " << 2 * alpha_star);
  //alpha = std::min((Scalar)alpha_init, 4 * alpha_star);
  alpha = 2 * alpha_star;

  return newError;
}

void InvariantNetwork::getA1(int batchSize, int batchIndex) {
  if(inputs_on_cpu) {
    a1.resize(inputs.rows(), batchSize);
    int start = batchSize * batchIndex;
    for(int i = 0; i < batchSize; i++)
      for(int j = 0; j < a1.rows(); j++)
        a1(j,i) = inputs(j,i+start);
  }
  else {
    a1 = inputs.block(0, batchSize*batchIndex, inputs.rows(), batchSize);
  }
}

Scalar InvariantNetwork::computeWholeBatchCost(int oldIndex, int batches, int batchSize) {
  Scalar totalCost = 0.0;
  for(int batchIndex = 0; batchIndex < batches; batchIndex++) {
    getA1(batchSize, batchIndex);
    feedForward();
    totalCost += computeCost();
  }
  
  // Set a1 back to proper mini-batch
  getA1(batchSize, oldIndex);
  feedForward();

  return totalCost;
}

Scalar InvariantNetwork::computeCost() {
  Scalar reg = (W * W.transpose()).squaredNorm()
    - W.rowwise().squaredNorm().cwise().square().sum();
  SDN_DEBUG("regularization val for W: " << reg);

  Scalar cost = 0;
  switch(cost_f) {
  case CORRELATION_COST:
    /*
      Correlation cost
      f = sum(x1x2+x2x3+x3x4+...xn-1*xn);
      df/dx = (x2, x1+x3, x2+x4, x3+x5, ..., xn-2+xn, xn-1)
    */
    cost += (a3.block(0, 0, a3.rows(), a3.cols() - 1).cwise() *
             a3.block(0, 1, a3.rows(), a3.cols() - 1)).sum() / a3.cols();
    break;
  case SPARSE_COST:
    /*
      Sparse cost
      f = -sum(x1+x2+...+xn)
      df/dx = (-1, -1, ..., -1)
    */
    cost -= a3.sum() / a3.cols();
    break;
  case LOGCOSH_DIFF_COST:
    /*
      Logcosh difference cost
      f = sum(logcosh(x0-x1), logcosh(x1-x2), ..., logcosh(xn-1-xn))
      df/dx = (tanh(x0-x1), tanh(x1-x2)-tanh(x0-x1), ...)
    */

    cost -= coshM(a3.block(0, 0, a3.rows(), a3.cols() - 1)
                  - a3.block(0, 1, a3.rows(), a3.cols() - 1))
      .cwise().log().sum() / a3.cols();
  }

  //cost -= lambdaV * (V.colwise().sum().cwise().square().sum()
  //                     - V.cwise().square().sum());
  cost -= lambdaW * reg;

  return cost;
}

Scalar InvariantNetwork::computeLL(int batchIndex, int batchSize) {
  Matrix test = (z4.cwise() * multilabels.block(0, batchIndex * batchSize, 
                                                multilabels.rows(), 
                                                batchSize)).colwise().sum();

  Matrix temp = multilabels.block(0, batchIndex*batchSize,
				  multilabels.rows(), batchSize);
  Matrix labelssum = temp.colwise().sum();

  for (int i = 0; i < batchSize; i++) {
    if (labelssum(0,i) < 0.5) {
      test(0,i) = std::max((Scalar)1e-6, 1 - z4.col(i).sum());
    }
    else {
      test(0,i) = std::max((Scalar)1e-6, test(0,i));
    }
  }
  return test.cwise().log().sum();
}

void InvariantNetwork::computeDerivatives(Matrix& deriv) {
  switch(cost_f) {
  case CORRELATION_COST:
    deriv = Matrix::Zero(a3.rows(), a3.cols());
    deriv.block(0, 0, a3.rows(), a3.cols() - 1) +=
      a3.block(0, 1, a3.rows(), a3.cols() - 1);
    deriv.block(0, 1, a3.rows(), a3.cols() - 1) +=
      a3.block(0, 0, a3.rows(), a3.cols() - 1);
    deriv /= a3.cols();
    break;
  case SPARSE_COST:
    deriv = -Matrix::Ones(a3.rows(), a3.cols()) / a3.cols();
    break;
  case LOGCOSH_DIFF_COST:
    deriv = Matrix::Zero(a3.rows(), a3.cols());
    Matrix tanhdiff = tanhM(a3.block(0, 0, a3.rows(), a3.cols() - 1) -
                            a3.block(0, 1, a3.rows(), a3.cols() - 1));
    deriv.block(0, 0, a3.rows(), a3.cols() - 1) -= tanhdiff;
    deriv.block(0, 1, a3.rows(), a3.cols() - 1) += tanhdiff;
    deriv /= a3.cols();
    break;
  }
}

void InvariantNetwork::train(Scalar learnRate, int iterations, int batchSize, 
                             Scalar tolerance, bool train_w, bool train_v, 
                             bool train_t, Scalar reg_v, Scalar reg_w, 
                             COST_FUNCTION cost, ACTIVATION_FUNCTION act) {
  SDN_DEBUG("Training...");
  
  alpha_init = alpha = learnRate;
  update_w = train_w;
  update_v = train_v;
  update_t = train_t;
  lambdaV = reg_v;
  lambdaW = reg_w;
  cost_f = cost;
  act_f = act;

  batchSize = std::min(batchSize, inputs.cols());
  batchSize = std::max(batchSize, 2);
  if(inputs.cols() % batchSize != 0) {
    SDN_WARN("Number of inputs is not divisible by batch size; some data will not be used");
  }

  int i = 0, batchIndex;
  int batches = inputs.cols() / batchSize;
  std::vector<Scalar> batchError(batches), batchChange(batches, 0),
    trainingError(batches), trainingChange(batches, 0),
    logLikelihood(batches), likelihoodChange(batches, 0);

  // compute initial objective values
  for(batchIndex = 0; batchIndex < batches; batchIndex++) {
    getA1(batchSize, batchIndex);
    feedForward();
    if(update_w || update_v)
      batchError[batchIndex] = computeCost();
    if(update_t) {
      trainingError[batchIndex] = 99999999.9;
      logLikelihood[batchIndex] = -99999999.9;
    }
  }
  batchIndex = 0;
  //doIsNan = true;
  
  do {
    clock_t iter = clock();
    SDN_DEBUG("Iteration " << i << ", batch " << batchIndex);
    SDN_DEBUG("Training" << (update_w ? " W" : "") << (update_v ? " V" : "")
              << (update_t ? " T" : ""));
    

    if(batches > 1) {
      getA1(batchSize, batchIndex);
      feedForward();
    }
    
    // Train the softmax layer T
    /*
    if(update_t) {
      feedForwardSoftmax();

      Matrix test = (z4.cwise() * labels.block(0, batchIndex * batchSize, labels.rows(), batchSize)).colwise().sum();

      Matrix labelssum = labels.block(0,batchIndex*batchSize, labels.rows(), batchSize);
      labelssum = labelssum.colwise().sum();

      for (int i = 0; i < batchSize; i++) {
        if (labelssum(0,i) < 0.5) {
          test(0,i) = std::max((Scalar)1e-6, 1 - z4.col(i).sum());
        }
        else {
          test(0,i) = std::max((Scalar)1e-6, test(0,i));
        }
      }
      Scalar LL = test.cwise().log().sum();
      likelihoodChange[batchIndex] = LL - logLikelihood[batchIndex];
      logLikelihood[batchIndex] = LL;

      SDN_DEBUG("Log Likelihood: " << LL << ", change: " <<  likelihoodChange[batchIndex]);

      z4 = -z4 + labels.block(0, batchIndex * batchSize, labels.rows(), batchSize);
      Matrix gradT = (z4 * a3.transpose()) / batchSize;
      T.block(0, 0, T.rows()-1, T.cols()) += alpha * gradT;
      
      if(batchIndex == (batches - 1)) {
        Scalar totalChange = 0.0;
        for(int i = 0; i < batches; i++) {
          totalChange += likelihoodChange[i];
        }
        
        if(totalChange < tolerance) {
          SDN_DEBUG("Converged!");
          break;
        }
      }
    }
    */
    
    // Train the network weights W or V
    if(update_w || update_v) {
      Scalar oldError = batchError[batchIndex];
      batchError[batchIndex] = backProp(oldError);
      batchChange[batchIndex] = batchError[batchIndex] - oldError;
      
      SDN_DEBUG("Objective: " << batchError[batchIndex] << ", change: " <<  
                batchChange[batchIndex]);
      
      if(batchIndex == (batches - 1)) {
        Scalar avgChange = 0.0;
        for(int i = 0; i < batches; i++) {
          avgChange += batchChange[i];
        }
	avgChange /= batches;
        
        if(avgChange < tolerance) {
          SDN_DEBUG("Converged!");
          break;
        }
      }
    }
    
    batchIndex++;
    if(batchIndex == batches) {
      batchIndex = 0;
    }
    
    double secs = ((double)(clock() - iter)) / CLOCKS_PER_SEC;
    SDN_DEBUG("Iteration time: " << secs);
    // saveWeights("temp");
  } while(i++ < iterations);
}

void InvariantNetwork::trainSoftmax(Scalar learnRate, int iterations,
                                    int batchSize, Scalar tolerance,
                                    int inputsPerExample, int classes,
                                    ACTIVATION_FUNCTION act, 
                                    REGULARIZATION reg, Scalar lambda) {
  alpha = learnRate;
  act_f = act;

  int exBatchSize = batchSize;
  T = Matrix::Zero(classes, W.rows() * inputsPerExample);
  batchSize *= inputsPerExample;

  int batches = inputs.cols() / batchSize;
  std::vector<Scalar> batchError(batches), batchChange(batches, 0);

  // std::ofstream hist("hist.mat");

  // compute initial objective values
  int batchIndex;
  for(batchIndex = 0; batchIndex < batches; batchIndex++) {
    getA1(batchSize, batchIndex);
    feedForward();
    concatOutputs(inputsPerExample);

    /* if(batchIndex == 0) {
      for(int i = 0; i < 100; i++)
	for(int j = 0; j < a3t.rows(); j++)
	  hist << a3t(j,i) << (j == a3t.rows()-1 ? "\n" : " ");
      hist.close();
      std::cout << "done" << std::endl;
      } */

    feedForwardSoftmax();
    batchError[batchIndex] = computeLL(batchIndex, exBatchSize) / exBatchSize;
  }
  batchIndex = 0;
  
  int iter = 0;
  do {
    clock_t start = clock();
    SDN_DEBUG("Iteration " << iter << ", batch " << batchIndex);
    SDN_DEBUG("Training T");
    
    if(batches > 1) {
      getA1(batchSize, batchIndex);
      feedForward();
      concatOutputs(inputsPerExample);
    }
    
    // train the softmax layer T
    feedForwardSoftmax();
    Matrix gradT = -z4 + multilabels.block(0, batchIndex * exBatchSize,
					   multilabels.rows(), exBatchSize);
    gradT = (gradT * a3t.transpose()) / ((Scalar)exBatchSize);
    
    switch(reg) {
    case L1:
      // subgradient
      for(int i = 0; i < gradT.rows(); i++) {
        for(int j = 0; j < gradT.cols(); j++) {
	  gradT(i,j) -= lambda * (T(i,j) > 0 ? 1: (T(i,j) < 0 ? -1 : 0));
        }
      }
      break;
    case L2:
      gradT -= lambda*T.block(0,0,T.rows()-1,T.cols());
      break;
    case NONE:
      break;
    }

    /*
    // numerical derivatives check with regularization
    Scalar oldLL = computeLL(batchIndex, exBatchSize)/(Scalar)(exBatchSize);
    switch(reg) {
    case L1:
      oldLL -= lambda*T.cwise().abs().sum();
      break;
    case L2:
      oldLL -= lambda*0.5*T.squaredNorm();
      break;
    case NONE:
      break;
    }
    
    Scalar eps = 1e-5;
    for (int i = 0; i < gradT.rows(); i++) {
      for (int j = 0; j < gradT.cols(); j++) {
	T(i,j) += eps;
	feedForwardSoftmax();
	Scalar newLL = computeLL(batchIndex, exBatchSize)/(Scalar)(exBatchSize);
	switch(reg) {
	case L1:
	  newLL -= lambda*T.cwise().abs().sum();
	  break;
	case L2:
	  newLL -= lambda*0.5*T.squaredNorm();
	  break;
	case NONE:
	  break;
	}
	
	SDN_DEBUG("OLD LL: " << oldLL << " NEW LL: " << newLL );

	Scalar numDeriv = (newLL - oldLL)/eps;
	Scalar diff = std::abs(numDeriv - gradT(i,j));

	if(diff / std::max(std::abs(numDeriv), std::abs(gradT(i,j))) > 0.01) {
	  SDN_DEBUG("Numerical derivative does not match: " << numDeriv << " "
		    << gradT(i,j));
	}
	else {
	  SDN_DEBUG("Numerical derivative ok: " << numDeriv << " " << gradT(i,j));
	}
	T(i,j) -= eps;
      }
    }
    // end of numerical derivatives check
    */

    // actual update
    T.block(0, 0, T.rows()-1, T.cols()) += alpha * gradT;

    // compute new log likelihoods
    feedForwardSoftmax();
    Scalar LL = computeLL(batchIndex, exBatchSize) / exBatchSize;

    Scalar regv = 0;
    switch(reg) {
    case L1:
      regv = lambda*T.cwise().abs().sum();
      break;
    case L2:
      regv = lambda*0.5*T.squaredNorm();
      break;
    case NONE:
      break;
    }

    batchChange[batchIndex] = LL - regv - batchError[batchIndex];
    batchError[batchIndex] = LL - regv;

    SDN_DEBUG("Log Likelihood: " << LL << ", total cost: " << LL - regv <<
	      ", change: " << batchChange[batchIndex]);
    SDN_DEBUG("Regularization value: " << regv);

    if(batchIndex == batches - 1) {
      Scalar avgChange = 0.0;
      for(int i = 0; i < batches; i++) {
        avgChange += batchChange[i];
      }
      avgChange /= batches;
      
      if(avgChange < tolerance) {
        SDN_DEBUG("Converged!");
        break;
      }
    }
    
    batchIndex++;
    if(batchIndex == batches) {
      batchIndex = 0;
    }
    
    double secs = ((double)(clock() - start)) / CLOCKS_PER_SEC;
    SDN_DEBUG("Iteration time: " << secs);
    // saveWeights("temp");
  } while(iter++ < iterations);
}


Matrix InvariantNetwork::getW() {
  return W;
}

Matrix InvariantNetwork::getV() {
  return V;
}

Matrix InvariantNetwork::getT() {
  return T;
}

Matrix InvariantNetwork::neighborhoodMatrix(int xdim, int ydim, int size) {
  Matrix ret = Matrix::Zero(xdim*ydim, xdim*ydim);
  for(int r = 0; r < xdim; r++)
    for(int c = 0; c < ydim; c++) {
      int cellIndex = ydim * r + c;
      for(int dr = -size; dr <= size; dr++)
        for(int dc = -size; dc <= size; dc++) {
          if(dr * dr + dc * dc > size * size)
            continue;
          int nr = r + dr, nc = c + dc;
          if(nr < 0 || nc < 0 || nr >= xdim || nc >= ydim)
            continue;
          int newIndex = ydim * nr + nc;
          ret(cellIndex,newIndex) = 1.0;
        }
    }
  return ret;
}

void InvariantNetwork::performProjections() {
  //clock_t proj = clock();
  if(update_w) {
    Vector rowNorm = W.rowwise().norm();
    Matrix norm = 1.0 / sqrt(W.cols()) * Matrix::Ones(1, W.cols());
    for (int j = 0; j < W.rows(); j++) {
      if(rowNorm(j) > 0) {
        W.row(j) /= rowNorm(j);
      } else {
        W.row(j) = norm;
      }
    }
  }

  if(update_v) {
    for(int i = 0; i < V.rows(); i++)
      for(int j = 0; j < V.cols(); j++)
        V(i,j) = std::max(V(i,j), (Scalar)0.0);
    
    Vector rowNorm = V.rowwise().norm();
    Matrix norm = 1.0 / sqrt(V.cols()) * Matrix::Ones(1, V.cols());
    for (int j = 0; j < V.rows(); j++) {
      if(rowNorm(j) > 0) {
        V.row(j) /= rowNorm(j);
      } else {
        V.row(j) = norm;
      }
    }
  }
  //double secs = ((double)(clock() - proj)) / CLOCKS_PER_SEC;
  //SDN_DEBUG("Projection time: " << secs);
}

bool InvariantNetwork::hasNegativeV() {
  if(!update_v) return false;
  for(int i = 0; i < V.rows(); i++)
    for(int j = 0; j < V.cols(); j++)
      if(V(i,j) < 0)
        return true;
  return false;
}

void InvariantNetwork::saveWeights(std::string filePrefix) {
  std::string filePath = "/home/jeffrey/libsdn-trunk/tests/bin/invarianceResults/";

  switch(cost_f) {
  case CORRELATION_COST:
    filePrefix += "_tc";
    break;
  case SPARSE_COST:
    filePrefix += "_sc";
    break;
  case LOGCOSH_DIFF_COST:
    filePrefix += "_ldc";
    break;
  }

  switch(act_f) {
  case SQUARE_SQRT_ACT:
    filePrefix += "_square";
    break;
  case LOGCOSH_LINEAR_ACT:
    filePrefix += "_logcosh";
    break;
  }

#if SDN_USE_CUDA
  filePrefix += "_cuda";
#endif

  std::string filenameW = filePath + "weights1" + filePrefix + ".mat";
  //std::string filenameV = filePath + "weights2" + filePrefix + ".mat";
  
  std::ofstream fileW(filenameW.data());  
  //std::ofstream fileV(filenameV.data());

  fileW << W << std::endl;
  //fileV << V << std::endl;
}

void InvariantNetwork::saveT(std::string filePrefix) {
  std::string filePath = "/home/jeffrey/libsdn-trunk/tests/bin/invarianceResults/";

  switch(cost_f) {
  case CORRELATION_COST:
    filePrefix += "_tc";
    break;
  case SPARSE_COST:
    filePrefix += "_sc";
    break;
  case LOGCOSH_DIFF_COST:
    filePrefix += "_ldc";
    break;
  }

  switch(act_f) {
  case SQUARE_SQRT_ACT:
    filePrefix += "_square";
    break;
  case LOGCOSH_LINEAR_ACT:
    filePrefix += "_logcosh";
    break;
  }

#if SDN_USE_CUDA
  filePrefix += "_cuda";
#endif

  std::string filenameT = filePath + "weights3" + filePrefix + ".mat";
  std::ofstream fileT(filenameT.data());
  fileT << T << std::endl;
}


void InvariantNetwork::saveFeatures(std::string filePrefix, int batchSize, 
                                    int inputsPerExample,
                                    ACTIVATION_FUNCTION act_f, FORMAT f, 
                                    bool test)
{
  std::string filePath = "/home/jeffrey/libsdn-trunk/tests/bin/invarianceResults/";

  switch(cost_f) {
  case CORRELATION_COST:
    filePrefix += "_tc";
    break;
  case SPARSE_COST:
    filePrefix += "_sc";
    break;
  case LOGCOSH_DIFF_COST:
    filePrefix += "_ldc";
    break;
  }
  
  switch(act_f) {
  case SQUARE_SQRT_ACT:
    filePrefix += "_square";
    break;
  case LOGCOSH_LINEAR_ACT:
    filePrefix += "_logcosh";
    break;
  }
  
#if SDN_USE_CUDA
  filePrefix += "_cuda";
#endif

  std::string filename = filePath + "features" + filePrefix;
  if (f == NORMAL) {
    if (test)
      filename += "_test";
    else
      filename += "_train";

    std::ofstream file(filename.data());
    
    int exBatchSize = batchSize; 
    batchSize *= inputsPerExample; 

    for(int batchIndex=0; batchIndex*batchSize < inputs.cols(); batchIndex++) {
      getA1(batchSize, batchIndex);
      feedForward();
      concatOutputs(inputsPerExample);

      for (int ex = 0; ex < a3t.cols(); ex++) {
         for (int i = 0; i < a3t.rows(); i++) {
           file << a3t(i,ex);
           if (i < a3t.rows()-1)
             file << " ";
         }
         file << std::endl;
      }
    }
  }

  if (f == LIBSVM) {
    filename += "_libsvm";
    if (test)
      filename += "_test";
    else
      filename += "_train";
    
    std::ofstream file(filename.data());
    
    int exBatchSize = batchSize; 
    batchSize *= inputsPerExample; 

    for(int batchIndex=0; batchIndex*batchSize < inputs.cols(); batchIndex++) {
      getA1(batchSize, batchIndex);
      feedForward();
      concatOutputs(inputsPerExample);

      for (int ex = 0; ex < a3t.cols(); ex++) {
        int label = labels[ex+batchIndex*exBatchSize];
         file << label << " ";
         for (int i = 0; i < a3t.rows(); i++) {
           file << i+1 << ":" << a3t(i,ex);
           if (i < a3t.rows()-1)
             file << " ";
         }
         file << std::endl;
      }
     }
  }
} 


/*
void InvariantNetwork::mergeRGBFeatures(std::string filePrefix)
{
  // Merges and saves the rgb features for training T
  readInput(testData, CIFAR_TEST);
  readInput(testLabels, CIFAR_TEST_LABELS);

  Matrix out (3*Wr, 10000);
  
  batchSize = 2500;

  for (int ch = 0; ch < 3; ch++) {
    for (int i = 0; i < 10000/batchSize; i++) {
      out.block(ch*V.rows(), i*batchSize, V.rows(), batchSize) = a3.block( V.rows(), ch*10000 + i*batchSize, V.rows(), batchSize);
    }
  }

  std::string filePath = "/home/jeffrey/libsdn-trunk/tests/bin/invarianceResults/";
  
  switch(cost_f) {
  case CORRELATION_COST:
    filePrefix += "_tc";
    break;
  case SPARSE_COST:
    filePrefix += "_sc";
    break;
  case LOGCOSH_DIFF_COST:
    filePrefix += "_ldc";
    break;
  }

  switch(act_f) {
  case SQUARE_SQRT_ACT:
    filePrefix += "_square";
    break;
  case LOGCOSH_LINEAR_ACT:
    filePrefix += "_logcosh";
    break;
  }
  
#if SDN_USE_CUDA
  filePrefix += "_cuda";
#endif
  
  std::string filename = filePath + "test_features" + filePrefix + ".mat";
  
  std::ofstream file(filename.data());
  
  file << out << std::endl;
  
} 
*/

void InvariantNetwork::testAccuracy(int batchSize, int inputsPerExample,
                                    ACTIVATION_FUNCTION act) {
  act_f = act;

  int exBatchSize = batchSize;
  int examples = inputs.cols() / inputsPerExample;
  batchSize *= inputsPerExample;

  // Straight test accuracy
  int correct = 0, incorrect = 0;
  std::vector<int> classCorrect(T.rows(), 0), classTotal(T.rows(), 0);
  std::vector<int> channelCorrect(3, 0), channelTotal(3, 0);
  for(int batchIndex = 0; batchIndex * batchSize < inputs.cols(); batchIndex++) {
    getA1(batchSize, batchIndex);
    feedForward();
    concatOutputs(inputsPerExample);
    feedForwardSoftmax();
    
    Matrix z4Sums = z4.colwise().sum();

    // compute test accuracy
    for (int i = 0; i < z4.cols(); i++) {
      Scalar mostLikely = 1 - z4Sums(0,i);
      int classSelected = T.rows()-1;
      for (int j = 0; j < T.rows()-1; j++) {
        classSelected = (z4(j,i) > mostLikely) ? j : classSelected;
        mostLikely = std::max(z4(j,i), mostLikely);
      }
      
      int label = labels[i+batchIndex*exBatchSize];
      if (std::abs(classSelected - label) < 0.1) {
        correct++;
	classCorrect[label]++;
	channelCorrect[(i+batchIndex*exBatchSize)/(inputs.cols()/3)]++;
      }
      else
        incorrect++;
      classTotal[label]++;
      channelTotal[(i+batchIndex*exBatchSize)/(inputs.cols()/3)]++;
    }
  }

  // Test accuracy with majority voting
  /*
  int trueIndex = 0;
  Matrix votes = Matrix::Zero(CIFAR_TEST_EXAMPLES/3, CIFAR_CLASSES);
  int correctVote = 0;
  int incorrectVote = 0;
  for(int batchIndex = 0; batchIndex*batchSize < CIFAR_TEST_EXAMPLES; batchIndex++) {
    trueIndex = batchIndex*batchSize;
    getA1(batchSize, batchIndex);
    feedForward();
    feedForwardSoftmax();
    
    Matrix z4Sums = z4.colwise().sum();
    
    // get predictions
    for (int i = 0; i < z4.cols(); i++) {
      Scalar mostLikely = 1 - z4Sums(0,i);
      int classSelected = CIFAR_CLASSES-1;
      for (int j = 0; j < CIFAR_CLASSES-1; j++) {
        classSelected = (z4(j,i) > mostLikely) ? j : classSelected;
        mostLikely = std::max(z4(j,i), mostLikely);
      }
      
      votes(trueIndex%10000,classSelected) = votes(trueIndex%10000,classSelected) + 1;

      trueIndex++;
    }
  }

  unsigned int iseed = (unsigned int)time(NULL);
  srand (iseed);

  for(int i = 0; i < CIFAR_TEST_EXAMPLES/3; i++) {
    int maxVoted = 0;
    int classSelected = 0;
    
    for (int j = 0; j < CIFAR_CLASSES; j++) {
      classSelected = (votes(i,j) > maxVoted) ? j : classSelected;
      maxVoted = std::max((int)votes(i,j) , maxVoted);
    }
    
    if (maxVoted == 1) {
      int possible[3];
      int c = 0;
      for (int j = 0; j < CIFAR_CLASSES; j++) {
        if (votes(i,j) > 0)
          possible[c++] = j;
      }
      
      int chosen = (int)(2.0*rand()/(RAND_MAX+1.0));
      classSelected = possible[chosen];
    }

    if (std::abs(classSelected - labels(0,i)) < 0.1)
      correctVote++;
    else
      incorrectVote++;
  }
  
  Scalar accuracy = ((Scalar) correct) / CIFAR_TEST_EXAMPLES;
  Scalar votingAccuracy = 3.0 * ((Scalar) correctVote) / CIFAR_TEST_EXAMPLES;
  
  int tally[4];
  memset(tally, 0, sizeof(int)*4);
  
  votes = votes.rowwise().maxCoeff();
  
  for (int i = 0; i < CIFAR_TEST_EXAMPLES/3; i++) {
    tally[(int)votes(i,0)]++; 
  }
  */

  Scalar accuracy = ((Scalar) correct) / examples;
  SDN_DEBUG("Accuracy: " << accuracy);
  for(int i = 0; i < (int)classTotal.size(); i++) {
    Scalar classAcc = ((Scalar) classCorrect[i]) / classTotal[i];
    SDN_DEBUG("Class " << i << " accuracy: " << classAcc);
  }

  for(int i = 0; i < (int)channelTotal.size(); i++) {
    Scalar channelAcc = ((Scalar) channelCorrect[i]) / channelTotal[i];
    SDN_DEBUG("Channel " << i << " accuracy: " << channelAcc);
  }

  /*
  SDN_DEBUG("Accuracy on CIFAR-10 test set (with majority voting): " << votingAccuracy);
  SDN_DEBUG("Percentage of unanimous votes: " << 3.0*((double)tally[3])/CIFAR_TEST_EXAMPLES);
  SDN_DEBUG("Percentage of 2-1 votes: " << 3.0*((double)tally[2])/CIFAR_TEST_EXAMPLES);
  SDN_DEBUG("Percentage of evenly split votes: " << 3.0*((double)tally[1])/CIFAR_TEST_EXAMPLES);
  */
}
