
// Copyright 2009, All Rights Reserved.
/// \author Andrew Saxe (asaxe@stanford.edu)                                                                                
/// \author Ian Goodfellow (ia3n@cs.stanford.edu)
/** \file layer.cc
    \brief  Defines the interface of the abstract layer.
*/
                                                                                
#include "layer.h"
#include "standardLayer.h"
#include "convolutionLayer.h"
#include "storageNode.h"
#include "sdn/base.h"

#include <Eigen/Array>

using namespace sdn;

Layer::Layer(boost::shared_ptr<StorageNode> input, boost::shared_ptr<StorageNode> output) : input_(input), output_(output) {
  resizePermitted = false;
  expectDimensions_ = false;
}

Layer::~Layer() {

}

void Layer::expectDimensions(unsigned inputSize, unsigned outputSize)
{
  if (expectDimensions_ && !resizePermitted)
    {
      SDN_ASSERT_MSG(inputSize == expectInputSize_ && outputSize == expectOutputSize_, "You may not change the expected dimensions when resizePermitted is false");
    }

  expectDimensions_ = true;
  expectInputSize_ = inputSize;
  expectOutputSize_ = outputSize;
}


void Layer::link(boost::shared_ptr<StorageNode> input, boost::shared_ptr<StorageNode> output) {
  setInputStorageNode(input);
  setOutputStorageNode(output);
}

int Layer::inputDimension() {
  if (input_ == NULL)
    {
      if (expectDimensions_)
	return expectInputSize_;
      SDN_FATAL("Can't get input dimension of runtime-dimensionless Layer");
    }

  return input_->size();
}

int Layer::outputDimension() {
  if (output_ == NULL)
    {
      if (expectDimensions_)
	return expectOutputSize_;
      return -1;
    }

  return output_->size();
}

void Layer::forwardProp(const Matrix& input 
) {
  SDN_ASSERT_MSG(input.rows() == inputDimension(), 
		 "input.rows()="<<input.rows()<<
		 " but inputDimension()="<<inputDimension());
 
  setInput(input);
  forwardProp();
}


//const Matrix& Layer::errorDerivativeWrtInput() {
//  SDN_ASSERT(input_!=NULL);
//  return input_->getDerivatives();
//}

void Layer::setOutput(const Matrix& output) {
  SDN_ASSERT(output_ != NULL);
  SDN_ASSERT(outputDimension() == output.rows());
  output_->setActivities(output);
}

void Layer::setErrorDerivativeWrtOutput(
    Matrix& error_derivative_wrt_output) {
  SDN_ASSERT(output_ != NULL);
  SDN_ASSERT(outputDimension() == error_derivative_wrt_output.rows());
  output_->setDerivatives(error_derivative_wrt_output);
}

void Layer::setInput(const Matrix& input) {
  SDN_ASSERT(input_ != NULL);
  SDN_ASSERT(inputDimension() == input.rows());
  input_->setActivities(input);
}

void Layer::setErrorDerivativeWrtInput(
    Matrix& error_derivative_wrt_input) {
  SDN_ASSERT(input_ != NULL);
  SDN_ASSERT(inputDimension() == error_derivative_wrt_input.rows());
  input_->setDerivatives(error_derivative_wrt_input);
}

const Matrix& Layer::getOutput() {
  SDN_ASSERT(output_ != NULL);
  return output_->getActivities();
}

const Matrix& Layer::getErrorDerivativeWrtOutput() {
  SDN_ASSERT(output_ != NULL);
  return output_->getDerivatives();
}

const Matrix& Layer::getInput() {
  SDN_ASSERT(input_ != NULL);
  return input_->getActivities();
}

const Matrix& Layer::getErrorDerivativeWrtInput() {
  SDN_ASSERT(input_ != NULL);
  return input_->getDerivatives();
}

void Layer::setStorageNode_(boost::shared_ptr<StorageNode> newLayer,
			     boost::shared_ptr<StorageNode> & destination,
			     unsigned expectedDimension)
{
  SDN_ASSERT(newLayer != NULL);
  
  if (!resizePermitted)
    {
      if (destination == NULL)
	{
	  if (expectDimensions_)
	    {
	      SDN_ASSERT(newLayer->size() == expectedDimension);
	    }
	}
      else
	{
	  SDN_ASSERT(newLayer->size() == destination->size());
	}
    }

  destination = newLayer;
}    

void Layer::setInputStorageNode(boost::shared_ptr<StorageNode> input) {
  setStorageNode_(input,input_,expectInputSize_);
}

void Layer::setOutputStorageNode(boost::shared_ptr<StorageNode> output) {
  setStorageNode_(output,output_,expectOutputSize_);
}

void Layer::serialize(serial::Archive & ar)
{
  SDN_DEBUG("Layer::serialize");
  int version = 3;

  version = ar.beginBase(this,version,"sdn::Layer");

  switch(version)
    {
    case 3:
      ar.attribute("resizePermitted",resizePermitted);
            
      serialize_shared_ptr(ar,  input_);
      serialize_shared_ptr(ar, output_);
      
      ar.attribute("expectDimensions_",expectDimensions_);
      ar.attribute("expectInputSize_",expectInputSize_);
      ar.attribute("expectOutputSize_",expectOutputSize_);
      
     

      ar.endBase();
      break;
    default:
      SDN_FATAL("Unrecognized Layer version");
    }

}

Registry<Layer>& sdn::LayerRegistry() {
  static Registry<Layer> registry("sdn::Layer");
  return registry;
}

SDN_AUTOREGISTER(Layer, StandardLayer);
SDN_AUTOREGISTER(Layer, ConvolutionLayer);
SDN_AUTOREGISTER(Layer, DeconvolutionLayer);
