// Copyright 2009, All Rights Reserved.
/// \author Andrew Saxe (asaxe@stanford.edu)
/// \author Ian Goodfellow (ia3n@cs.stanford.edu)

/** \file layer.h
    \brief  Declares the interface of an abstract layer.
*/

#pragma once


//if turned on, turns hiden units in encoding StandardLayers off
//with probability NOISY_HIDDEN (defined in StandardLayer.cc)
#define NOISY_HIDDEN_HACK 0


#include <boost/smart_ptr.hpp>
#include "../base/matrix.h"
#include "../base/logger.h"

class XMLNode;

namespace sdn
{
  
  //Forward declarations
  template <typename T>
  class Registry;
  class StorageNode;
  namespace serial
  {
	class Archive;
  }

  
  /// Defines a basic layer interface
  /**
     This class is the basic building block of the library. It permits
     automatic differentiation through arbitrary feed forward configurations
     of properly implemented Layer classes.
  */
  class Layer {
  public:
    ///True if the link method is permitted to change the size of the storage layers (false by default)
  bool resizePermitted;
  
  
 
  
  public:
  
  ///Creates a layer with given storage layers
  /**
     The default of NULL indicates that the storage layers will be provided
     later by the link function
  */
  Layer(boost::shared_ptr<StorageNode> input = boost::shared_ptr<StorageNode>(), boost::shared_ptr<StorageNode> output = boost::shared_ptr<StorageNode>() );
  
  virtual ~Layer(); 
  
  ///Links the layer to the specified input and output storage layers.
  /**
     If resizePermitted is false, the input and output storage layer
     sizes must match the pre-existing ones
     
     \param input The storage layer to use as input. The size of the
     storage layer determines the size of the input.
     
     \param output The storage layer to use as output. The size of the
     storage layer determines the size of the output.
     
  */
  virtual void link(boost::shared_ptr<StorageNode> input, boost::shared_ptr<StorageNode> output);
  
  ///Returns the input dimension. Negative values mean this has not been set.
  virtual int inputDimension();
  
  ///Returns the output dimension. Negative values mean this has not been set.
  virtual int outputDimension();
  
  ///Propagates from input to output
  /**
     The input storage layer must already contain the input pattern. Features
     are in the rows, examples in the columns.
  */
  virtual void forwardProp
(
) = 0;

 #if NOISY_HIDDEN_HACK
	virtual void forwardProp(bool noise)
	{
		forwardProp();
	}
#endif

  
  ///Propagates the given input matrix to the output 
  /**
     \param x The input matrix must have inputDimension() rows representing
     the features of the input. Different examples (of which there can be any
     number) are in the columns.
  */
  virtual void forwardProp(const Matrix& input);


#if NOISY_HIDDEN_HACK
	virtual void forwardProp(const Matrix & input, bool noise)
	{
		setInput(input);
		forwardProp(noise);
	}
#endif

  
  ///Propagates error derivative signal from output to input.
  /**
     The error derivative w.r.t. the output must have been
     set previously, e.g. through presentTarget.
  */
  virtual void backProp() = 0;
  
  ///Presents a target to the layer
  /**
     Calculates the error derivative w.r.t. the output based on
     the last applied input and the supplied target. You must
     have previously called forwardProp.
     
     \param target must be identical in dimension to the size
     of the input.
  */
  virtual void presentTarget(const Matrix& target) = 0;
  
  ///Calculates the error of the layer
  /**
     Calculates the error between the last applied input and
     the target provided.
     \param target must be identical in dimension to the size
     of the input.
  */
  virtual Scalar calculateError(const Matrix& target) = 0;
  
  ///Applies a gradient descent update using the supplied learning rate
  /**
     Default implementation is no op.
  */
  virtual void gradientDescentUpdate(Scalar learning_rate) {};

  ///Applies a gradient descent update using parameters stored in class
  /**
     Default implementation is no op.
  */
  virtual void gradientDescentUpdate() {};

  ///Applies a gradient descent update w.r.t. a sparsity objective
  /**
     Default implementation is no op.
  */
  virtual void gradientDescentUpdateSparsity(Scalar learning_rate) {};

  ///Applies a gradient descent update for sparsity with params in class
  /**
     Default implementation is no op.
  */
  virtual void gradientDescentUpdateSparsity() {};

  virtual void setOutput(const Matrix& output);
  
  virtual void setErrorDerivativeWrtOutput(
					   Matrix& error_derivative_wrt_output);
  
  virtual void setInput(const Matrix& input);
  
  virtual void setErrorDerivativeWrtInput(
					  Matrix& error_derivative_wrt_input);
  
  virtual const Matrix& getOutput();
  
  virtual const Matrix& getErrorDerivativeWrtOutput();
  
  virtual const Matrix& getInput();
  
  virtual const Matrix& getErrorDerivativeWrtInput();
  
  virtual void setInputStorageNode(boost::shared_ptr<StorageNode> input);
  
  virtual void setOutputStorageNode(boost::shared_ptr<StorageNode> output);

  /// Sets up layer-specific parameters specified as an xml node
  virtual void setParams(XMLNode& params) {};
  
  /** Call on a layer with no input or output to tell it to expect
      a given input/output size.
      This will only be enforced if resizePermitted is false.
      expectDimensions may be called repeatedly but changing the
      expectedDimensions when resizePermitted is false is a runtime
      error.
  */
  void expectDimensions(unsigned inputSize, unsigned outputSize);
  
  
  virtual void serialize(serial::Archive & ar);
  
  virtual bool equals(const Layer & other) {SDN_FATAL("Layer::Equals not implemented"); return false;}

  virtual void scrambleWeights() {}

  protected:
  
  boost::shared_ptr<StorageNode> input_;
  boost::shared_ptr<StorageNode> output_;
  
  bool expectDimensions_;
  unsigned expectInputSize_;
  unsigned expectOutputSize_;
  
  void setStorageNode_(boost::shared_ptr<StorageNode> newLayer,
			boost::shared_ptr<StorageNode> & destination,
			unsigned expectedDimension);
  
  
  
  };
  Registry<Layer>& LayerRegistry();
}
