/// \author Ian Goodfellow <ia3n@cs.stanford.edu>


#include "network.h"
#include "stackedNetwork.h"
#include "sdn/base.h"
#include "rbm.h"
using namespace std;
using namespace sdn;
using namespace boost;

Network::Network() : _dataset((Dataset *) NULL), _doneTraining(false)
{
}

Network::~Network()
{
  //do nothing
}

bool Network::doneTraining() const
{
	return _doneTraining;
}

void Network::endTraining() {
  _doneTraining = true;
}

void Network::registerDataset(shared_ptr<Dataset> dataset)
{
  _dataset = dataset;

  if (dataset->supportsExampleSize(getInputRows(),getInputCols()))
    dataset->requestExampleSize(getInputRows(),getInputCols());
  else
    {
      /// \todo-- add a resizePermitted variable like the one in the Layer class
      SDN_WARN("Resizing network to fit preferred size of dataset");
      setInputSize(dataset->getPreferredRows(),dataset->getPreferredCols());
    }

}

void Network::save(const char * filepath, serial::Format format)
{
  switch (format)
    {
    case serial::BIN:
      {
	ofstream ofs(filepath, ios::binary);

	serial::BinOutArchive ar(ofs);
	serialize(ar);
      }
      break;
    case serial::XML:
      SDN_FATAL("XML serialization is not supported right now due to a bug in the external xmlParser library");
      /*
      serial::XmlOutArchive ar;
      serialize(ar);
      ar.getNode().writeToFile(filepath);*/
      break;
    }
}


void Network::processExample(const Matrix & x)
{
  //cout << "processExample " << endl;
  //cout << x.rows() << endl;
  //cout << x.cols() << endl;
  SDN_ASSERT_MSG(x.rows()*x.cols() == _input.rows(), "x.rows() = " << x.rows() << ", x.cols() = " << x.cols() << "_input.rows() = " << _input.rows());
  matrixToMatrixCol(x,_input);
  processExampleAsColumn(_input);
}

bool Network::supportsReconstruction() const
{
  return false;
}

void Network::getReconstruction(Matrix & x)
{
  SDN_FATAL("getReconstruction called on a class that does not support reconstruction.");
}

Registry<Network> & sdn::NetworkRegistry()
{
  static Registry<Network> registry("sdn::Network");
  return registry;
}


void Network::setInputSize(unsigned int s)
{
  SDN_ASSERT(!isTwoDimensional());
  SDN_ASSERT(s > 0);

  if (inputSize_ == s)
    return;

  inputSize_ = s;
  _input.resize(inputSize_,1);

  setInputSize_(s);
}

void Network::setInputSize(unsigned int rows, unsigned int cols)
{
  SDN_ASSERT(cols == 1 || isTwoDimensional());
  SDN_ASSERT(cols > 0);
  SDN_ASSERT(rows > 0);


  if (inputRows_ == rows && inputCols_ == cols)
    return;

  inputRows_ = rows;
  inputCols_ = cols;
  _input.resize(inputRows_ * inputCols_,1);

  setInputSize_(rows,cols);
}

unsigned Network::getInputSize()
{
  if (isTwoDimensional())
    {
      return _input.rows();
    }
  
  return inputSize_;
}

unsigned Network::getInputRows()
{
  if (isTwoDimensional())
    return inputRows_;
  else
    return inputSize_;
}

unsigned Network::getInputCols()
{
  if (isTwoDimensional())
    return inputCols_;
  else
    return 1;
}


/*
ObjectDescription * Network::getObjectDescription_()
{
  ClassDescription<Network> * od = new ClassDescription<Network>();
  od->descendantDepth = 0;
  od->object = this;
  od->classId = "sdn::Network";
  od->baseId = "sdn::Network";
  /// \todo-- add all the real fields
  return dynamic_cast<ObjectDescription *>(od);
}
*/



void Network::serialize(serial::Archive & ar)
{
  SDN_DEBUG("Network::serialize");

  int version = 127;
  
  version = ar.beginBase(this, version, "sdn::Network");
  switch (version)
    {
    case 127:
      ar.attribute("inputSize",inputSize_);
      ar.attribute("inputRows",inputRows_);
      ar.attribute("inputCols",inputCols_);

      serializeMatrixSize(ar, _input);
      ar.attribute("doneTraining",_doneTraining);
      
      ar.endBase();

      break;
      
	default:
	  SDN_FATAL("Unrecognized class version requested");
    }
}

bool Network::equals(const Network & other)
{
	if (inputSize_ != other.inputSize_)
		return false;
	if (inputRows_ != other.inputRows_)
		return false;
	if (inputCols_ != other.inputCols_)
		return false;
	if (_input.rows() != other._input.rows())
		return false;
	if (_input.cols() != other._input.cols())
		return false;
	if (_doneTraining != other._doneTraining)
		return false;
	return equals_(other);
}


boost::shared_ptr<Network> Network::load(const char * filepath)
{
  ifstream ifs(filepath, ios::binary);
  if (ifs.fail())
    {
      SDN_FATAL("Could not open "<<filepath);
    }
  serial::BinInArchive ar(ifs);
  return NetworkRegistry().load(ar);
}

boost::shared_ptr<StorageNode> Network::getTestingNode()
{
  return boost::shared_ptr<StorageNode>();
}

void Network::scrambleWeights(bool deepestOnly)
{
  SDN_FATAL("scrambleWeights() called on a Network type that doesn't support it");
}


SDN_AUTOREGISTER(Network, StackedNetwork);
SDN_AUTOREGISTER(Network, RBM);
