//Author: Ian Goodfellow <ia3n@cs.stanford.edu>
//File: network.h
//Description: A general interface for a network

#pragma once

#include <boost/shared_ptr.hpp>
#include "../dataset/dataset.h"

///\todo get these out of this file!
#include "../dataset/naturalTestDataset.h"
#include "../dataset/gratingTestDataset.h"

namespace sdn
{

  class StorageNode;

  /// Network class

  /**
	The Network class provides a single access point for configuration
        methods, serialization and deserialization methods, evaluation
        methods, etc. It is the also the main interface between
        the actual network code and the training runtime environment.

        To implement a new type of deep network, subclass this class.
  */


  class Network
  {

  public:

    Network();
    virtual ~Network();
    

    /// trains for nIter iterations (one minibatch update = one iteration)
    /**
    if nIter = -1, trains forever
    unless an internal halting condition is met. if allowHalting = false,
    runs until nIter iterations have completed, even if internal
    halting condition is met.
    obviously, you may not set nIter = -1 and allowHalting = false
    returns number of iterations that it ran for
    implementers should set _doneTraining to true on convergence/fatal error
    */

    virtual long long train(long long nIter, bool allowHalting = true) = 0;

    /// returns true if the network thinks it can't train anymore
    /// this should happen if some convergence criterion is met or if there is
    /// some sort of fatal error
    bool doneTraining() const;

    /// Ends training at next opportunity.
    virtual void endTraining();

    virtual void registerDataset(boost::shared_ptr<Dataset> dataset);

    /// Save to file
    void save(const char * filepath, serial::Format format = serial::BIN);

    /// Load from file  \todo-- assumes binary, ought to be able to inspect file and do the right thing
    static boost::shared_ptr<Network> load(const char * filepath);


    /// Save to archive object-- note: does not save the actual content of the input, or the dataset that the network is linked to
    virtual void serialize(serial::Archive & ar);

    /// runs an example through the network
    /**
       This enables later queries for the example's reconstruction
       or classification or distribution over possible classifications

       Base class converts this to a column example and calls
       processExampleAsColumn
    */
    virtual void processExample(const Matrix & x);


    /// like processExample, but assumes the input is formatted as a column
    
    virtual void processExampleAsColumn(const Matrix & x) = 0;

    /// returns true if reconstruction is supported
    /**
       The base class returns false so you only need to implement this
       if you support reconstruction
    */
    virtual bool supportsReconstruction() const;

    /// gets the reconstruction of the last example passed to
    /// processExample
    /**
       The base class throws an assert, so if you don't support this you
       don't need to implement any checks
    */

    virtual void getReconstruction(Matrix & x);



    //later we should add  processBatchAsColumns,
    //getReconstructionAsColumn, getReconstructionAsColumns, and similar
    //methods for getting classification, etc.

    

    //todo-- add methods for converting a string ID into some sort of
    //reference to a storage layer, and then methods for accessing
    //the values of these storage layers, running the network only
    //to or only from that layer, etc. (not all of which need to be
    //supported by all classes) This will be useful for invariance
    //tests and also for visualization (turn on a unit and see what
    //it reconstructs)
    

    virtual void display()=0;


    
    /*Returns true if network uses 2D data (data that has 2D semantics,
    such as images, regardless of whether it is represented as a matrix
    or a vector under the hood.
    If false the network uses 1D data (such as audio) */
    virtual bool isTwoDimensional() = 0;


    /** Sets the size of the input for a network that works on 1-dimensional data
	Runtime error if network is not 1D. This implementation assumes column vector, you may wish to override this to get a different behavior.

	Calls setInputSize_(), override that to respond to the change in size

     */
    virtual void setInputSize(unsigned int s);

    /** Sets the size of the input for a network that works on 2-dimensional data. Runtime error if the network is not 2D, unless cols=1. This implementation
	assumes that you represent your 2D data as a column vector. You may
	wish to override this to get a different behavior.
	Most classes will need to override this in order to change some other
	fields, but will start off by calling this method for setup.
	
	Calls setInputSize_(), override that to respond to the change in size
    */
    virtual void setInputSize(unsigned int rows, unsigned int cols);


    /// Gets the size of the input for a network that works on 1-dimensional data. Runtime error if it is not 1-dimensional (unless it only has 1 column)
    virtual unsigned getInputSize();

    ///Gets the number of rows of the input. For a 1D network this is the same as getInputSize()
    virtual unsigned getInputRows();

    //Gets the number of cols of the input. For a 1D network this is always 1
    virtual unsigned getInputCols();

    /// \todo-- replace this with overall objective value so it's meaningful for all networks
    /// \todo-- logging system with different channels so you can request reconstruction error by some string identifier
    virtual Scalar getAverageError() { return 0; }

    /// Returns the StorageNode of the network that should be used for tests of hidden unit properties (such as invariance tests). Returns NULL if not supported
    virtual boost::shared_ptr<StorageNode> getTestingNode();


    /// \todo-- come up with a more general way for tests to have access to this method of StackedNetwork without needing to do casts
    virtual void forwardPropToHighestLayer(const Matrix& x) = 0;

    virtual void processExampleToHighestLayer(const Matrix & x) = 0;

    virtual const Matrix& getHighestLayerActivities()=0;
    /** For each hidden unit in the network, randomly rearranges the input weights
	@param deepestOnly If true, only scrambles the weights at the deepest layer. This is usually
	necessary to avoid saturating the output.
    */
    virtual void scrambleWeights(bool deepestOnly);


    bool equals(const Network & other);


  protected:
    unsigned inputSize_; 
    unsigned inputRows_;
    unsigned inputCols_;
    //_input can either be a vector or a matrix depending on subclass
    //but it should always have input_size_ elements if 1D 
    //or input_rows * input_cols elements if 2D
    Matrix _input;
    boost::shared_ptr<Dataset> _dataset;
    bool _doneTraining;


    /**Subclasses should override the relevant version of this in order to
       respond to resize instructions

\todo-- it's actually redundant to put the size in the argument since
the method isn't called until after the member vars storing the size
have been updated
 */
    virtual void setInputSize_(unsigned s) { setInputSize_(s,1); }
    virtual void setInputSize_(unsigned rows, unsigned cols) {}


    virtual bool equals_(const Network & other) = 0;


  };


  sdn::Registry<Network> & NetworkRegistry();

}


