/// \author Ian Goodfellow

#include "rbm.h"
#include "../base/matrixUtils.h"
#include "../base/random.h"
#include <Eigen/Array>
#include <Eigen/Core>


using namespace sdn;

void checkNan(const Matrix & m)
{
  for (int i = 0; i < m.rows(); i++)
    for (int j = 0; j < m.cols(); j++)
      SDN_ASSERT_MSG(!isnan(m(i,j)), i << " " << j);
}


RBM::RBM()
{
  numHidden_ = 50;
  batchSize_ = 1;
  numFantasy_ = 1;
  fastWeightDecay_ = 0.05;

  learningRate_ = 0.1;
  fastLearningRate_ = 0.1;

  targetMeanActivation_ = 0.5;
  sparsityPenalty_ = 0.0;

  weightDecay_ = 0.00;

  activationAveragingTimeConstant_ = 0.005;

  setInputSize(28,28);

  initialize_();
 
}


RBM::~RBM()
{

}


void RBM::initialize_()
{
  vizBiases_.resize(inputRows_ * inputCols_, 1);
  vizBiases_.setConstant(0);
  hidBiases_.resize(numHidden_,1);
  hidBiases_.setConstant(log(targetMeanActivation_)-log(1.0-targetMeanActivation_));
  meanActivation_.resize(numHidden_,1);
  meanActivation_.setConstant(targetMeanActivation_);
  weights_ = Matrix::Random(numHidden_,inputRows_*inputCols_) * 0.05;

  fastVizBiases_ = vizBiases_;
  fastVizBiases_.setConstant(0);
  fastHidBiases_ = hidBiases_;
  fastHidBiases_.setConstant(0);
  fastWeights_ = weights_;
  fastWeights_.setConstant(0);

  h_.resize(numHidden_,1);

  _input.setConstant(0);
  vFantasy_.resize(numFantasy_,_input);
  hFantasy_.resize(numFantasy_, h_);


  vizBiasGradient_ = vizBiases_;
  hidBiasGradient_ = hidBiases_;
  weightsGradient_ = weights_;
}



bool RBM::isTwoDimensional()
{
  return true;
}

void RBM::setInputSize_(unsigned rows, unsigned cols)
{
  initialize_();
}

RandomRealGenerator< > rrg;

void RBM::binarize_(Matrix & m)
{
  for (int i = 0; i < m.rows(); i ++)
    for (int j = 0; j < m.cols(); j++)
      {
	m(i,j) =  (m(i,j) > rrg.randUnif(0,1));     	  
      }
}

long long RBM::train(long long nIter, bool allowHalting)
{
  SDN_ASSERT_MSG(nIter != -1, "RBM doesn't implement a halting condition yet");


  for (unsigned i = 0; i < nIter; i++)
    {
      vizBiasGradient_.setConstant(0);
      hidBiasGradient_.setConstant(0);
      weightsGradient_.setConstant(0);

      for (unsigned j = 0; j < batchSize_; j++)
	{
	  _dataset->getRandomExampleAsColumn(_input);

	  fireUp_(_input,h_);

	  vizBiasGradient_ += _input / double(batchSize_);
	  hidBiasGradient_ += h_ /double(batchSize_);
	  weightsGradient_ += h_ * _input.transpose() / double(batchSize_) ;

	  meanActivation_ = (1.0 - activationAveragingTimeConstant_) * meanActivation_ + activationAveragingTimeConstant_ * h_;

	  hidBiasGradient_ -=  ((meanActivation_.cwise() - targetMeanActivation_)* sparsityPenalty_ / double(batchSize_));


	  //Debugging
	  /*
	  fireDown_(h_, _input);

	  Matrix recons = vectorToMatrix(_input, getInputRows(), getInputCols());

	  SDN_MSG(recons.minCoeff()<<" "<<recons.maxCoeff());
	  recons = recons.cwise() -  recons.minCoeff();
	  if (recons.maxCoeff() > 0)
	    recons /= recons.maxCoeff();
	  SDN_MSG("Waiting");
	  showMatrix(recons, "Reconstruction");
	  SDN_MSG("Running");
	  */
	}

      for (unsigned k = 0; k < numFantasy_; k++)
	{
	  fireUp_(vFantasy_[k], hFantasy_[k],true);
	  
	  vizBiasGradient_ -=  vFantasy_[k] / double(numFantasy_);
	  hidBiasGradient_ -= hFantasy_[k] / double(numFantasy_);
	  weightsGradient_ -= hFantasy_[k] * vFantasy_[k].transpose() / double(numFantasy_);
	  
	  binarize_(hFantasy_[k]);

	  fireDown_(hFantasy_[k], vFantasy_[k], true);

	  binarize_(vFantasy_[k]);
	}

      //showFantasy();

      checkNan(fastWeights_);
      checkNan(fastVizBiases_);
      checkNan(fastHidBiases_);

      weights_ *= (1.0-learningRate_ * double(batchSize_) * weightDecay_);
      fastWeights_ *= (1.0-fastWeightDecay_);
      fastVizBiases_ *= (1.0-fastWeightDecay_);
      fastHidBiases_ *= (1.0-fastWeightDecay_);

      vizBiases_ += learningRate_ * vizBiasGradient_;
      fastVizBiases_ += fastLearningRate_ * vizBiasGradient_;

      hidBiases_ += learningRate_ * hidBiasGradient_;
      fastHidBiases_ += fastLearningRate_ * hidBiasGradient_;

      weights_ += learningRate_ * weightsGradient_;
      fastWeights_ += fastLearningRate_ * weightsGradient_;
      
      checkNan(fastWeights_);
      checkNan(fastVizBiases_);
      checkNan(fastHidBiases_);


    }
  

  return nIter;

}

void RBM::showFantasy()
{

  int r = getInputRows();
  int c = getInputCols();

  int rr = (int) sqrt(numFantasy_);
  int cc = rr;

  if (rr * cc < numFantasy_)
    {
      cc++;
      SDN_ASSERT(rr * cc >= (int) numFantasy_);
    }


  Matrix means = Matrix::Zero(numFantasy_,1);

  for (unsigned i = 0; i < numFantasy_; i++)
    means(i,0) = matrixMean(vFantasy_[i]);

  double m = matrixMean(means);

  Matrix img = Matrix::Zero(rr * r , cc * c);
  
  img.setConstant(m);


  for (int i = 0; i < rr; i++)
    {
      for (int j = 0; j < cc; j++)
	{
	  int f = i *cc +j;
	  if (f < numFantasy_)
	    {
	      img.block(i*r,j*c,r,c) = vectorToMatrix(vFantasy_[f],r,c);
	    }
	}
    }

  SDN_MSG(img.minCoeff()<<" "<<img.maxCoeff());
  img = img.cwise() - img.minCoeff();
  if (img.maxCoeff() > 0)
   img /= img.maxCoeff();
  SDN_MSG("Waiting");
  showMatrix(img, "Fantasy particles");
  SDN_MSG("Running");
}


void RBM::fireUp_(const Matrix & v, Matrix & h, bool useFastWeights)
{


  checkNan(v);
  checkNan(h);
  checkNan(weights_);
  checkNan(fastWeights_);
  checkNan(hidBiases_);
  checkNan(fastHidBiases_);


  SDN_DEBUG("weights in ["<<weights_.minCoeff()<<","<<weights_.maxCoeff()<<"], v in ["<<v.minCoeff()<<","<<v.maxCoeff()<<"]"<<" hidBiases in ["<<hidBiases_.minCoeff()<<","<<hidBiases_.maxCoeff()<<"]");

  h = weights_*v + hidBiases_;

  checkNan(h);

  if (useFastWeights)
    {
      h += fastWeights_*v  + fastHidBiases_;
    }

  checkNan(h);

  for (int i = 0; i < h.rows(); i++)
    for (int j = 0; j < h.cols(); j++)
      {
	h(i,j) = 1.0/(1.0+exp(-h(i,j)));
      }
}

void RBM::fireDown_(const Matrix & h, Matrix & v, bool useFastWeights)
{
  v = weights_.transpose()*h + vizBiases_;
  
  if (useFastWeights)
    {
      v += fastWeights_.transpose()*h + fastVizBiases_;
    }

  for (int i = 0; i < v.rows(); i++)
    for (int j = 0; j < v.cols(); j++)
      {
	v(i,j) = 1.0/(1.0+exp(-v(i,j)));
      }
}


void RBM::serialize(serial::Archive& ar) {
  SDN_DEBUG("RBM::Serialize");

  int version = 256;
  
  version = ar.beginDerived(this,version,"sdn::RBM");
  
  switch (version)
    {
    case 256:
      ar.beginSuper();
      Network::serialize(ar);
      ar.endSuper();

      ar.attribute("numHidden",numHidden_);
      ar.attribute("batchSize",batchSize_);
      serializeMatrix(ar, vizBiases_);
      serializeMatrix(ar, hidBiases_);
      serializeMatrix(ar, weights_);
      serializeMatrix(ar, fastWeights_);

      serializeMatrix(ar, h_);

      SDN_WARN("Warning-- most fields of RBM are not serialized yet");

      ar.endDerived();

      break;
    default:
      SDN_FATAL("Unrecognized StackedNetwork version");
    }
}

const sdn::Matrix& RBM::getHighestLayerActivities()
{
	return h_;
}
