//Authors: Andrew Saxe (asaxe@cs.stanford.edu)
///        Ian Goodfellow (ia3n@cs.stanford.edu)
/// \file stackedNetwork.cc
/// \brief Implements a stacked network consisting of a number of Layer objects
#include "activationFunction.h"
#include "costFunction.h"
#include "convolutionLayer.h"
#include <iostream>
#include "stackedNetwork.h"
#include "storageNode.h"
#include "sdn/base.h"
#include "../gui/patchDisplay.h"
#include <fstream>
#include <time.h>
#include <boost/pointer_cast.hpp>
#include <Eigen/Array>
#include <ext/numeric>         // Quoc: for iota

#define ENABLE_TEMPLATE 1

using namespace std;
using namespace boost;
using namespace sdn;

class StackedNetworkConfig : public ConfigParser
{
public:
  // bool unsupervisedTrainingConfigured;
  //  int unsupervisedTraining;
  //bool supervisedTrainingConfigured;
  //int supervisedTraining;
  //bool errorAveragingTimeconstantConfigured;
  //int normalizeContrast;
  //bool normalizeContrastConfigured;
  //Scalar errorAveragingTimeconstant;
  //bool inputWidthConfigured;
  //int inputWidth;
  //bool inputHeightConfigured;
  //int inputHeight;

  AttributeParser p;

  

 public:
  StackedNetworkConfig(StackedNetwork& parent) 
    : //unsupervisedTrainingConfigured(false),
      //supervisedTrainingConfigured(false),
      //errorAveragingTimeconstantConfigured(false),
      //normalizeContrastConfigured(false),
      //inputWidthConfigured(false),
      //inputHeightConfigured(false),
      parent_(parent) 
  {
    p.addBool("train_unsupervised");
    p.addBool("train_supervised");
    p.addDouble("error_averaging_timeconstant");
    p.addBool("normalize_contrast");
    p.addInt("input_width"); /// \todo-- add unsigned to AttributeParser class
    p.addInt("input_height");

    p.addDouble("denoising",0.0);
    p.addInt("denoiseLayer",-1);
    p.addInt("reconstructionTarget",-1);
    p.addInt("reconstructionLayer",-1);

    p.addString("load_archive");
  }
 
  void topics(vector<string>& topics) { 
    topics.push_back("sdn::StackedNetworkConfig"); 
    topics.push_back("sdn::NetworkConfig"); 
  }
  void readConfig(XMLNode& node)
  {
    p.parse(node);




    /// \todo-- throw error if given unrecognized child. should make general framework for making such errors easy to generate

    Attribute<string> la = p.getString("load_archive");
    if (la.newlyConfigured && la.value != "") {
      std::ifstream ifs(la.value.c_str(), ios::binary);
      serial::BinInArchive ar(ifs);
      parent_.serialize(ar);
    }

    XMLNode architecture = node.getChildNode("Architecture");  
    if(!architecture.isEmpty())
      parent_.clear(); //Wipe out any architecture previously there
      
      
    Attribute<bool> ut = p.getBool("train_unsupervised");
    if (ut.newlyConfigured) {
      parent_.train_unsupervised = ut.value;
    }

    Attribute<bool> st = p.getBool("train_supervised");
    if (st.newlyConfigured) {
      parent_.train_supervised = st.value; //no need to check if active since this doesn't involve a method call
    }
    
    Attribute<double> eat = p.getDouble("error_averaging_timeconstant");
    if (eat.newlyConfigured) {
      parent_.error_averaging_timeconstant = eat.value; //no need to check if active since this doesn't involve a method call
    }

    Attribute<bool> nc = p.getBool("normalize_contrast");
    if (nc.newlyConfigured) {
      parent_.normalize_contrast = nc.value; //no need to check if active since this doesn't involve a method call
    }

    Attribute<int> iw = p.getInt("input_width");
    Attribute<int> ih = p.getInt("input_height");

    if (iw.newlyConfigured) {
      SDN_DEBUG("sdn::StackedNetworkConfig: input width of "<<iw.value);
      if (!ih.newlyConfigured) {//if both input dimensions were set, do them both at once later
	
	parent_.setInputSize(parent_.getInputRows(),iw.value);
	
      }
    }

    if (ih.newlyConfigured) {
      SDN_DEBUG("sdn::StackedNetworkConfig: input height of "<<ih.value);
      if (!iw.newlyConfigured) { //if both input dimensions were set, do them both at once later
	parent_.setInputSize(ih.value,parent_.getInputCols());
      }
    }
    
    if (ih.newlyConfigured && iw.newlyConfigured) {
      parent_.setInputSize(ih.value, iw.value);
    }

    if(!architecture.isEmpty())
	parent_.addLayers(architecture);
  }
  void usage(ostream & os) const
  {
    os << "<sdn::StackedAutoencoderConfig learningRate=\"<Scalar>\" learningRateSparsity=\"<Scalar>\" architecture=\"<string>\" where the string consists of a sequence of autoencoder items \"a(<numHidden>,<applyGradientUpdate>,<applyGradientUpdateSparsity>,<sparsityTarget>);\" and convolution items \"c(<patch_width>,<patch_step>,<output_width>);\" />" << endl;
  }
  
protected:
  StackedNetwork& parent_;
};

StackedNetwork::StackedNetwork() : active_(false) {

  

  //Set up default parameters
  setDefaults_();

 
  
  //Set up parameters specified in config file
  boost::shared_ptr<StackedNetworkConfig> 
    parser(new StackedNetworkConfig(*this));
  ConfigListener autoconfig(parser);
  
  denoising_ = parser->p.getDouble("denoising").value;
  denoiseLayer_ = parser->p.getInt("denoiseLayer").value;
  reconstructionTarget_ = parser->p.getInt("reconstructionTarget").value;
  reconstructionLayer_ = parser->p.getInt("reconstructionLayer").value;

  active_ = true;
  
}


StackedNetwork::~StackedNetwork() {

}

namespace sdn
{
#define SDN_PATCH_RANGE  0.8
  void normalizePatch(Matrix &patch)
  {
    //const double VARIANCE_REGULARIZATION = 1.0;	// parameter used to regularize our estimate of the variance 
    
    double mean = patch.sum() / (patch.rows() * patch.cols());
    //double avgsqr = (patch.cwise() * patch).sum() / (patch.rows() * patch.cols());
    //double variance = avgsqr - sqr(mean);
    
    patch.cwise() -= mean;
    //patch /= (variance + VARIANCE_REGULARIZATION);
    
    double maxAbs = max(patch.maxCoeff(), fabs(patch.minCoeff()));
    patch /= (maxAbs+ 1e-6);            // now all values between [-1,1]
    
    patch *= SDN_PATCH_RANGE;               // now all values between [-_PATCH_RANGE, _PATCH_RANGE] 
    
    // patch *= 0.4;       // now between [-0.4, 0.4]
    // patch.cwise() += 0.5;       // now between [0.1, 0.9]
  }
}

    
long long StackedNetwork::train(long long nIter, bool allowHalting) {
  if(doneTraining()) /// \todo Figure out how to make this play nice with allowHalting.
    return 0;
  
  if (denoiseLayer_ == -1)
    {
      //SDN_DEBUG("Using input for reconstruction");
      desiredOutput_ = & _input;
    }
  else
    {
      //SDN_DEBUG("Using prenoised input for reconstruction");
      desiredOutput_ = & prenoisedInput_;
    }


  if (nIter == -1) {
    SDN_WARN("sdn::Autoencoder doesn't have a termination criterion yet so it will train forever"); /// \todo add termination criterion
  }

  for (long long iter = 0; nIter == -1 || iter < nIter; iter++) {
    /// \todo  support variable batch size
    int label = _dataset->getRandomExampleAsColumn(_input); 
    

    //Set this to 1 to enable debug mode where network repeatedly trains on same example-- definitely ought to be able to reconstruct it
#define BURN_IN_TEST 0
#if BURN_IN_TEST
    int numRan = 0;
    while (numRan == 0 || (numLayers() > 2 && dynamic_cast<StandardLayer *>(layers_[1].get())->learning_rate > 0))
      {
	numRan++;
	if (numRan > 1)
	  cout << numRan << " times burned in" << endl;
#endif


          
    if(should_transpose_input) //If convolution is first layer.
      _input = _input.transpose();

    forwardProp(_input, true);   
 
    
    if(train_unsupervised) {
     presentTarget(*desiredOutput_);    
      
      //Calculate sparsity average \todo put this in storageLayers?
      /*Matrix act = storageNodes_[1]->getActivities();
	if(!sparsity_initialized_) {
	sparsity_avg_ = act.sum()/(act.rows()*act.cols());
	sparsity_initialized_ = TRUE;
	} else {
	sparsity_avg_ = sparsity_avg_*.95 + .05*act.sum()/(act.rows()*act.cols());
	}*/
      
      backProp();

      gradientDescentUpdate();
      gradientDescentUpdateSparsity();
    }

 
    if(train_supervised) {
      supervised_layer_->forwardProp();
 
      Vector target_class = Vector::Zero(_dataset->numLabels());
      target_class(label) = 1.0;
      
      supervised_layer_->presentTarget(target_class);
          supervised_layer_->backProp();
   
      for(int i = numLayers()/2-1; i >= 0; --i)
	layers_[i]->backProp();
   

      supervised_layer_->gradientDescentUpdate();
      
      for(int i = numLayers()/2-1; i>= 0; --i)
	layers_[i]->gradientDescentUpdate();
    

      //Calculate error average (This was a hack for debugging--asaxe)
      /*if (error_avg_ < 0) {
	error_avg_ = supervised_layer_->calculateError(target_class);     
      } else {
	int pred_label;
	int tmp;
	supervisedStorageNode_->getActivities().maxCoeff(&pred_label,&tmp);
	Scalar error = 1.0;
	if(pred_label == label)
	  error = 0;
	
	error_avg_ = error_avg_*(1-error_averaging_timeconstant) + error_averaging_timeconstant*supervised_layer_->calculateError(target_class);//error;
	}

	SDN_MSG(error_avg_);*/
    }
 
#if BURN_IN_TEST
      }
#endif  
  
  }
  
  return nIter;
}


shared_ptr<Layer> StackedNetwork::getLayer(unsigned d) const {
  SDN_ASSERT(d < numLayers());
  
  return layers_[d];
}

shared_ptr<StorageNode> StackedNetwork::getStorageNode(unsigned d) const {
  SDN_ASSERT(d < numLayers()+1);

  return storageNodes_[d];
}

void StackedNetwork::addLayer(shared_ptr<Layer> layer) {
  if(numLayers() > 0)
    checkInvariants();

  SDN_ASSERT(layer != NULL);
  if(numLayers() > 0)
    SDN_ASSERT(layer->inputDimension() == getLayer(numLayers()-1)->outputDimension());
 

  //Allocate new storage layers
  if(numLayers() == 0) {
    storageNodes_.push_back(shared_ptr<StorageNode>(new StorageNode(layer->inputDimension())));
  }

  storageNodes_.push_back(shared_ptr<StorageNode>(new StorageNode(layer->outputDimension())));


  //Attach layer to these storage layers
  layer->link(storageNodes_[numLayers()], storageNodes_[numLayers()+1]);

  layers_.push_back(layer);
 
  checkInvariants();
}

void StackedNetwork::addLayers(XMLNode& architecture) {
  const char* name_c = architecture.getName();
  SDN_ASSERT(name_c);
  
  string name(name_c);
  SDN_ASSERT(name == "Architecture");

  for(int i = 0; i < architecture.nChildNode(); ++i) {
    XMLNode layer = architecture.getChildNode(i);
    const char* layer_type_c = layer.getName();
    if(layer_type_c == NULL) {
      SDN_WARN("<Architecture> tag could not load its " << i << "th child.");
    } else {
      string layer_type(layer_type_c);
      
      if(layer_type == "sdn::Autoencoder") {

	int num_hidden;
	
	const char* num_hidden_str = layer.getAttribute("num_hidden");
	if(num_hidden_str) {
	  num_hidden = atoi(num_hidden_str);
	} else {
	  SDN_FATAL("Must specify number of hidden units in autoencoder layer.");
	}
	
	addAutoencoder(num_hidden, layer);

	//Hack to make first layer not have biases set to regularize sparsity
	if(numLayers() == 2) {
	  shared_ptr<StandardLayer> decoder = dynamic_pointer_cast<StandardLayer>(getLayer(numLayers()-1));
	SDN_WARN("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   hacking decoder weights not to be initialized based on sparsity target !!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	  decoder->setTargetSparsity(0.5); /// \todo This has to be changed soon, very dangerous
	  decoder->initializeWeights();
	}	  

      } else if(layer_type == "sdn::ConvolutionLayer") {
	int patch_width;
	int patch_step;
	int output_width;

	const char* patch_width_str = layer.getAttribute("patch_width");
	if(patch_width_str) {
	  patch_width = atoi(patch_width_str);
	} else {
	  SDN_FATAL("Must specify patch_width in convolution layer.");
	}

	const char* patch_step_str = layer.getAttribute("patch_step");
	if(patch_step_str) {
	  patch_step = atoi(patch_step_str);
	} else {
	  SDN_FATAL("Must specify patch_step in convolution layer.");
	}

	const char* output_width_str = layer.getAttribute("output_width");
	if(output_width_str) {
	  output_width = atoi(output_width_str);
	} else {
	  SDN_FATAL("For now, must specify output_width (will have the option of setting this based on higher layers soon)");
	}

	addConvolution(patch_width, patch_step, output_width);

      } else {
	SDN_WARN("Could not recognize " << i << "th child of <Architecture> of type " << layer_type);
      }
      
    }
  }
}

void StackedNetwork::resetLayers(XMLNode& architecture) {
  clear();
  addLayers(architecture);
}

void StackedNetwork::addSupervisedLayer() {
  shared_ptr<SumOfSquaresCostFunction> cost_fn(new SumOfSquaresCostFunction);
  shared_ptr<SoftmaxActivationFunction> act_fn(new SoftmaxActivationFunction);

  supervised_layer_.reset(new StandardLayer(act_fn, cost_fn));
  supervisedStorageNode_.reset(new StorageNode(_dataset->numLabels()));

  //Allocate input storage layer if there is only a supervised layer
  if(numLayers() == 0) {
    storageNodes_.push_back(shared_ptr<StorageNode>(new StorageNode(getInputSize())));
  }

  supervised_layer_->link(storageNodes_[numLayers()/2], supervisedStorageNode_);
}

unsigned StackedNetwork::numLayers() const {
  return layers_.size();
}

void StackedNetwork::addAutoencoder(int num_hidden) {
  if(numLayers() == 0) {

    shared_ptr<StandardLayer> encoder(
        new StandardLayer(getInputSize(), num_hidden));

    shared_ptr<StandardLayer> decoder(
        new StandardLayer(num_hidden, getInputSize()));

    addLayer(encoder);
    addLayer(decoder);
  } else {

    //Writing this function was a huge pain. there must be a better way. Would be good to refactor.
  
    unsigned index = numLayers()/2;
 
    shared_ptr<StandardLayer> encoder(
	new StandardLayer(layers_[index-1]->outputDimension(),num_hidden));

    shared_ptr<StandardLayer> decoder(
	new StandardLayer(num_hidden, layers_[index]->inputDimension()));
  
    shared_ptr<StorageNode> encoder_input(
	new StorageNode(encoder->inputDimension()));

    shared_ptr<StorageNode> encoder_output( 
	new StorageNode(encoder->outputDimension()));
 
    //Hook up storage layers
    layers_[index-1]->setOutputStorageNode(encoder_input);
    encoder->link(encoder_input, encoder_output);
    decoder->link(encoder_output, storageNodes_[index]);
  

    StandardLayer * decoderStandardLayer = dynamic_cast<StandardLayer *>(decoder.get());

    layers_.insert(layers_.begin()+index, decoder);
    layers_.insert(layers_.begin()+index, encoder);
    
    storageNodes_.insert(storageNodes_.begin()+index, encoder_output);
    storageNodes_.insert(storageNodes_.begin()+index, encoder_input);
  }    
}

void StackedNetwork::addAutoencoder(int num_hidden,
				    Scalar learning_rate,
				    Scalar learning_rate_sparsity,
				    Scalar sparsity_averaging_timeconstant,
				    Scalar target_sparsity) {
  addAutoencoder(num_hidden);
  unsigned index = numLayers()/2;

  shared_ptr<StandardLayer> encoder = dynamic_pointer_cast<StandardLayer>(layers_[index-1]);
  if(encoder) {
    encoder->learning_rate = learning_rate;
    encoder->learning_rate_sparsity =  learning_rate_sparsity;
    encoder->sparsity_averaging_timeconstant = sparsity_averaging_timeconstant;
    encoder->setTargetSparsity(target_sparsity);
  } else {
    SDN_ERROR("Bad dynamic cast: Encoder layer was not a singleLayerNetwork");
  }
  shared_ptr<StandardLayer> decoder = dynamic_pointer_cast<StandardLayer>(layers_[index]);
  if(decoder) {
    decoder->learning_rate = learning_rate;
    decoder->learning_rate_sparsity =  learning_rate_sparsity;
    decoder->sparsity_averaging_timeconstant = sparsity_averaging_timeconstant;
    if(index > 0) { ///Don't set sparsity target for reconstruction layer /todo make this an option 
      decoder->setTargetSparsity(target_sparsity); 
    }
  } else {
    SDN_ERROR("Bad dynamic cast: Decoder layer was not a singleLayerNetwork");
  }
}

void StackedNetwork::addAutoencoder(int num_hidden,
				    XMLNode& params) {
  addAutoencoder(num_hidden);
  unsigned index = numLayers()/2;

  setParams(index, params);
  setParams(index-1, params);
}


void StackedNetwork::addConvolution(int patch_width, 
				    int patch_step, 
				    int output_width) {
  if(numLayers() == 0) {
    shared_ptr<ConvolutionLayer> encoder(new ConvolutionLayer(1,
							      patch_width,
							      patch_step,
							      output_width));
    shared_ptr<DeconvolutionLayer> decoder(
					  new DeconvolutionLayer(1,
								 patch_width,
								 patch_step,
								 output_width));
    addLayer(encoder);
    addLayer(decoder);
    
    should_transpose_input = true;
  } else {
    unsigned index = numLayers()/2;

    int input_dim = layers_[index-1]->outputDimension();
 
    shared_ptr<ConvolutionLayer> cl(new ConvolutionLayer(input_dim, 
							 patch_width,
							 patch_step,
							 output_width));
    shared_ptr<DeconvolutionLayer> dl(new DeconvolutionLayer(input_dim,
							     patch_width,
							     patch_step,
							     output_width));
    shared_ptr<StorageNode> conv_input(new StorageNode(cl->inputDimension()));
    shared_ptr<StorageNode> conv_output(new StorageNode(cl->outputDimension()));
    
    //Hook up storage layers
    layers_[index-1]->setOutputStorageNode(conv_input);
    cl->link(conv_input, conv_output);
    dl->link(conv_output, storageNodes_[index]);
    
    
    layers_.insert(layers_.begin()+index, dl);
    layers_.insert(layers_.begin()+index, cl);
    
    storageNodes_.insert(storageNodes_.begin()+index, conv_output);
    storageNodes_.insert(storageNodes_.begin()+index, conv_input);
  }
}

void StackedNetwork::registerDataset(shared_ptr<Dataset> dataset) {
  Network::registerDataset(dataset);
  if(train_supervised) {
    if(_dataset->containsLabels()) {
      addSupervisedLayer();
    } else {
      SDN_FATAL("Asked to set up supervised layer on dataset without labels.");
    }
  }
}

void StackedNetwork::setParams(XMLNode& params) {
  SDN_ASSERT(params.getName());
  SDN_ASSERT(string("Params") == params.getName());

  

  for(int i = 0; i < params.nChildNode(); ++i) {
    XMLNode child = params.getChildNode(i);

    if (string(child.getName()) == "networkUpdate")
    {
	reconstructionLayer_ = atoi(child.getAttribute("reconstructionLayer"));
	reconstructionTarget_ = atoi(child.getAttribute("reconstructionTarget"));
	SDN_ASSERT(child.getAttribute("denoising"));
	denoising_ = atof(child.getAttribute("denoising"));
	denoiseLayer_ = reconstructionTarget_;
	SDN_DEBUG("Set deep denoising parameters");

    }

    if(child.getAttribute("layer")) {
      int layer = atoi(child.getAttribute("layer"));
     
	SDN_ASSERT(layer < numLayers() / 2);
 
      setParams(layer, child);
      setParams(numLayers()-1-layer, child);
    } else if (child.getAttribute("index")) {
      int index = atoi(child.getAttribute("index"));
      setParams(index, child);
    } else {
    
      if (child.getAttribute("train_unsupervised")) {
	train_unsupervised = atoi(child.getAttribute("train_unsupervised"))==1;
      }
      if (child.getAttribute("train_supervised")) {
	train_supervised = atoi(child.getAttribute("train_supervised"))==1;
      }
    }
  }
}
void StackedNetwork::setParams(unsigned index, XMLNode& params) {
  SDN_ASSERT(index >= 0);
  SDN_ASSERT(index < numLayers());
  layers_[index]->setParams(params);
}
//void StackedNetwork::clear() {
//  layers_.clear();
//  storageNodes_.clear();
//  
//  supervised_layer_.clear();
//  supervisedStorageNode_.clear();
//}
void StackedNetwork::forwardProp(const Matrix& input
, bool train
) {

  _input = input;

  if( normalize_contrast ) {
    normalizePatch(_input);
  }
 
  if(storageNodes_.size() > 0)
    storageNodes_[0]->setActivities(_input);

  if(layers_.size() == 0)
    return;
  SDN_ASSERT_MSG(input.rows() == layers_[0]->inputDimension(),
		 "Input.rows(): " << input.rows() <<
		 " Input dimension:" << layers_[0]->inputDimension(););
  checkInvariants();
  
  unsigned propDist = numLayers();

  if (train && reconstructionLayer_ != -1)
    propDist = reconstructionLayer_ ;
  
  for(unsigned i = 0; i < propDist; ++i)
    {
      if (train && i == denoiseLayer_)
	{
	  //SDN_DEBUG("Adding noise");

	  SDN_ASSERT(denoiseLayer_ == reconstructionTarget_);
	  Matrix & m = storageNodes_[i]->getActivities();	  

	  prenoisedInput_ = m;
	  
	  double noiseCount = 0;

	  for (int i = 0; i < m.rows(); i++)
	    for (int j = 0; j < m.cols(); j++)
	      if ( double(rand() % 1000)/1000.0 < denoising_)
		{
		  //SDN_DEBUG("noise!");
		  m(i,j) = 0.0;
		  noiseCount++;
		}

	  //SDN_DEBUG("denoising: "<<denoising_);
	  //SDN_DEBUG("noise ratio: "<<noiseCount/double(m.rows()*m.cols()));
	}
      else
	{
	  //SDN_DEBUG("Train: "<<train);
	  //SDN_DEBUG("i: "<<i);
	  //SDN_DEBUG("denoiseLayer_"<<denoiseLayer_);
	}

      //SDN_DEBUG("forward prop at layer "<<i);
      layers_[i]->forwardProp(
#if NOISY_HIDDEN_HACK
	i < numLayers()/2
#endif
);
    }
  
  /*
  if (propDist == 3 && dynamic_cast<StandardLayer *>(layers_[1].get())->learning_rate > 0)
    {

      PatchDisplay p("Ian's Debugger", 20, 20, 4, 5, 5);

      Matrix input = vectorToMatrix(prenoisedInput_,20,20);
      Matrix output = vectorToMatrix(storageNodes_[3]->getActivities(),20,20);
      
      cout << "error: "<<(output-input).cwise().square().sum() << endl;
      cout << "mean(input): "<<matrixMean(input)<<endl;
      cout << "mean(output): "<<matrixMean(output)<<endl;

      p.addPatch(input);
      p.addPatch(output);

      p.enableWindow();

      p.wait(0);

    }
  */


  checkInvariants();
}



void StackedNetwork::presentTarget(const Matrix& target) {

  int outputLayer = numLayers()-1;
  if (reconstructionLayer_ != -1)
    outputLayer = reconstructionLayer_ - 1;


  //SDN_DEBUG("presentTarget at layer " <<outputLayer);


  checkInvariants();
  SDN_ASSERT(layers_.size() > 0);
  SDN_ASSERT_MSG(storageNodes_[outputLayer]->numExamples() == target.cols(),"Ran on "<<storageNodes_[outputLayer]->numExamples()<<" but given target with "<<target.cols()<<" examples. outputLayer = "<<outputLayer<<", numLayers="<<numLayers());
  
  layers_[outputLayer]->presentTarget(target);

//Calculate error average
      if (error_avg_ < 0)
	error_avg_ = layers_[outputLayer]->calculateError(target);     
      else
	error_avg_ = error_avg_*(1-error_averaging_timeconstant) + error_averaging_timeconstant*layers_[outputLayer]->calculateError(target); 
}

void StackedNetwork::backProp() {
  checkInvariants();

  int outputLayer = numLayers()-1;
  if (reconstructionLayer_ != -1)
    outputLayer = reconstructionLayer_ - 1;

  int firstLayer = 0;
  SDN_ASSERT(denoiseLayer_ == reconstructionTarget_);
  if (reconstructionTarget_ != -1)
    firstLayer = reconstructionTarget_;

  for(int i = outputLayer; i >= firstLayer; --i)
    {
	if (i == 0)
		SDN_DEBUG("backprop at layer 0");
      //SDN_DEBUG("backprop at layer "<<i);
      layers_[i]->backProp();
    }

  checkInvariants();
}

void StackedNetwork::gradientDescentUpdate() {
  checkInvariants();

  if (reconstructionTarget_ != -1)
    {
      for (int i = reconstructionTarget_ ; i < reconstructionLayer_ ; i++)
	{
	  //SDN_DEBUG("Gradient descent update at layer "<<i);
	  layers_[i]->gradientDescentUpdate();
	}
    }
  else
    {
      for(unsigned i = 0; i < numLayers(); ++i) {
	layers_[i]->gradientDescentUpdate();
      }
    }
  
  checkInvariants();
}

void StackedNetwork::gradientDescentUpdateSparsity() {
  checkInvariants();

  //Sparsity update only applies to the encoding layers
  for(unsigned i = 0; i < numLayers()/2; ++i) { 
    if (reconstructionTarget_ == -1 || i >= reconstructionTarget_)
      {
	//SDN_DEBUG("Gradient descent update sparsity at layer "<<i);
	layers_[i]->gradientDescentUpdateSparsity();
      }
  }
  
  checkInvariants();
}

Scalar StackedNetwork::getAverageError() {
  return error_avg_;
}

double StackedNetwork::getAverageSparsity() {
  //SDN_ASSERT(sparsity_initialized_);

  shared_ptr<StandardLayer> sln = dynamic_pointer_cast<StandardLayer>(layers_[numLayers()/2-1]);
  return sln->getAverageSparsity();
  
}


void StackedNetwork::processExampleAsColumn(const Matrix& x) {

  if(should_transpose_input)
    forwardProp(x.transpose());
  else
    forwardProp(x);
}

bool StackedNetwork::supportsReconstruction() const
{
  return true;
}

void StackedNetwork::getReconstruction(Matrix & x)
{
  checkInvariants();

  

  if(should_transpose_input) { //Convolution is first layer
    SDN_ASSERT_MSG(x.rows()*x.cols() == 
		 storageNodes_[numLayers()]->getActivities().cols(),
		 "rows()*cols(): " << x.rows()*x.cols() <<
		 " storage layer rows: " << 
		 storageNodes_[numLayers()]->getActivities().cols());

    matrixSubColToMatrix(
	     storageNodes_[numLayers()]->getActivities().transpose(),0,0,x);
  } else {

    SDN_ASSERT_MSG(x.rows()*x.cols() == 
		   storageNodes_[numLayers()]->getActivities().rows(),
		   "rows()*cols(): " << x.rows()*x.cols() <<
		   " storage layer rows: " << 
		   storageNodes_[numLayers()]->getActivities().rows());

     matrixSubColToMatrix(
			  storageNodes_[numLayers()]->getActivities(),0,0,x);
  }
  
}

const Matrix& StackedNetwork::getActivities(unsigned index) {
  checkInvariants();

  SDN_ASSERT(index < storageNodes_.size());
  return storageNodes_[index]->getActivities();
}

const Matrix& StackedNetwork::getHighestLayerActivities() {
  checkInvariants();
  return getActivities(numLayers()/2);
}

const Matrix& StackedNetwork::getDerivatives(unsigned index) {
  checkInvariants();

  SDN_ASSERT(index < storageNodes_.size());
  return storageNodes_[index]->getDerivatives();
}

bool StackedNetwork::active() {
  return active_;
}

void StackedNetwork::display(){

  int index = 1; ///Layer of weights to show \todo Add options to show different layers individually, and deep visualizations spanning multiple layers
  shared_ptr<StandardLayer> encoder = dynamic_pointer_cast<StandardLayer>(layers_[index-1]);
  SDN_ASSERT_MSG(encoder,"Bad dynamic cast: Encoder layer was not a singleLayerNetwork in display()");
  
  shared_ptr<StandardLayer> decoder = dynamic_pointer_cast<StandardLayer>(layers_[numLayers()-index]);
  SDN_ASSERT_MSG(decoder,"Bad dynamic cast: Decoder layer was not a singleLayerNetwork");
  
  if(visualizer_ == NULL) {
    visualizer_.reset(
      new PatchDisplay("Weight matrices (encoding/decoding pairs)", 
		     this->getInputCols(), 
		     this->getInputRows(),
		     1,
		     encoder->outputDimension()*2));
    visualizer_->setLimits(-3/sqrt(encoder->outputDimension()),3/sqrt(encoder->outputDimension()));
    //visualizer_->enableWindow();
  }
  
  //visualizer_->freezeWindow();
  
  Matrix encWeights = encoder->getWeights();
  Matrix decWeights = decoder->getWeights();
  
  visualizer_->clearDisplay();
  
  for(int i = 0; i < encoder->outputDimension(); ++i) {
    
    visualizer_->addPatch(vectorToMatrix(encWeights.row(i), 
					 this->getInputRows(), 
					 this->getInputCols()),
			  false,
			  false);
    visualizer_->addPatch(vectorToMatrix(decWeights.col(i), 
					 this->getInputRows(), 
					 this->getInputCols()),
			  false,
			  false);
    
  }
  
  //visualizer_->unfreezeWindow();
  //visualizer_->wait(10); 
  
  SDN_DEBUG("Ran StackedNetwork::display");

  char name[300];
  sprintf(name, "weight_patches.png");
  visualizer_->saveImage(name);
  
}
void StackedNetwork::clear() {
  layers_.clear();
  storageNodes_.clear();
  
  supervised_layer_ = ptr((StandardLayer*)NULL);
  supervisedStorageNode_ = ptr((StorageNode*)NULL);
  
  //  visualizer_ = ptr(NULL);
  
  setDefaults_();
}

void StackedNetwork::printDebugStuff() {
  SDN_DEBUG("-----Encoding layer info:");
  shared_ptr<StandardLayer> enc = dynamic_pointer_cast<StandardLayer>(getLayer(0));
  enc->printDebugStuff();

  //SDN_DEBUG("Decoding layer info:");
  shared_ptr<StandardLayer> dec = dynamic_pointer_cast<StandardLayer>(getLayer(1));
  //dec->printDebugStuff();

}


void StackedNetwork::serialize(serial::Archive& ar) {
  SDN_DEBUG("StackedNetwork::Serialize");

    int version = 256;

  version = ar.beginDerived(this,version,"sdn::StackedNetwork");
  


  switch (version)
    {
    case 256:
      ar.beginSuper();
      Network::serialize(ar);
      ar.endSuper();
      ar.attribute("errorAveragingTimeconstant",error_averaging_timeconstant);
      ar.attribute("shouldTransposeInput",should_transpose_input);
      ar.attribute("trainUnsupervised",train_unsupervised);
      ar.attribute("trainSupervised",train_supervised);
      ar.attribute("normalizeContrast",normalize_contrast);
      
      //Serialize layers
      serializeSTLVector(ar, layers_, PolymorphicSharedPointerSerializer<Layer>(LayerRegistry()));

      //Serialize storage layers
      serializeSTLVector(ar, storageNodes_, SharedPointerSerializer<StorageNode>()); 

      //serialize_shared_ptr(ar, supervised_layer_);
      //serialize_shared_ptr(ar, supervisedStorageNode_);

      ar.attribute("errorAvg",error_avg_);

      //Ian, do I need this? -- asaxe
      //really we ought to keep the visualizer separate, but yeah you do need it or you might
      //change the size of the network without changing the patch size of the visualizer
      //serialize_shared_ptr(ar, visualizer_);

      ar.attribute("active",active_);

      ar.endDerived();

      if (ar.in())
      {
      reconstructionTarget_ = -1;
      reconstructionLayer_ = -1;
      denoiseLayer_ = -1;
      }

      break;
    default:
      SDN_FATAL("Unrecognized StackedNetwork version");
    }
}

bool StackedNetwork::equals_(const Network & o)
{
  SDN_DEBUG("Running StackedNetwork::equals_");
  const StackedNetwork* sn = dynamic_cast<const StackedNetwork*>(&o);
  if (!sn)
    return false;
  const StackedNetwork & other = *sn;
  if (error_averaging_timeconstant != other.error_averaging_timeconstant)
    return false;
  if (should_transpose_input != other.should_transpose_input)
    return false;
  if (train_unsupervised != other.train_unsupervised)
    return false;
  
  SDN_DEBUG("Checking STL Vector for layers");
  if(numLayers() != other.numLayers())
    return false;
  for( unsigned i = 0; i < numLayers(); ++i) {
    if( !getLayer(i)->equals(*other.getLayer(i))) {
      return false;
    }
  }
  SDN_DEBUG("Checking STL Vector for storage layers");
  if(storageNodes_.size() != storageNodes_.size())
    return false;
  for( unsigned i = 0; i < storageNodes_.size(); ++i) {
    if( !storageNodes_[i]->equals(*other.getStorageNode(i)))
      return false;
  }
  
  if (error_avg_ != other.error_avg_)
    return false;
  if (active_ != other.active_)
    return false;
  
  return true;
}

void StackedNetwork::checkInvariants() {
  SDN_ASSERT_MSG(layers_.size() == storageNodes_.size() - 1, "Expected " << storageNodes_.size()-1<< " layers, got "<<layers_.size());
  
  for(unsigned int i = 0; i < layers_.size(); ++i) {
    SDN_ASSERT(layers_[i]->inputDimension() != -1 && (unsigned) layers_[i]->inputDimension() == storageNodes_[i]->size());
    SDN_ASSERT(layers_[i]->outputDimension() != -1 && (unsigned) layers_[i]->outputDimension() == storageNodes_[i+1]->size());
  }
}

void StackedNetwork::setDefaults_() {
  setInputSize(28,28);

  error_avg_ = -1.0;
  error_averaging_timeconstant = .05;
  
  train_unsupervised = true;
  train_supervised = false;
  
  normalize_contrast = false;
  
  should_transpose_input = false;
}





std::vector<Matrix> StackedNetwork::getEncoderWeights()
{
  SDN_ASSERT(layers_.size() %2 == 0);
  vector<Matrix> rval(layers_.size()/2);

  for (unsigned i = 0; i < layers_.size()/2; i++)
    {
      shared_ptr<StandardLayer> p = dynamic_pointer_cast<StandardLayer>(layers_[i]);
      SDN_ASSERT_MSG(p, "getEncoderWeights assumes all layers are StandardLayers");
      rval[i] = p->getWeights();
    }

  return rval;
}

std::vector<Matrix> StackedNetwork::getDecoderWeights()
{
  SDN_ASSERT(layers_.size() %2 == 0);
  vector<Matrix> rval(layers_.size()/2);

  for (unsigned i = 0; i < rval.size(); i++)
    {
      shared_ptr<StandardLayer> p = dynamic_pointer_cast<StandardLayer>(layers_.at(layers_.size()-1-i));
      SDN_ASSERT_MSG(p, "getEncoderWeights assumes all layers are StandardLayers");
      rval.at(i) = p->getWeights().transpose();
    }

  return rval;  
}

boost::shared_ptr<StorageNode> StackedNetwork::getTestingNode()
{
  return getStorageNode(numLayers()/2);
}

void StackedNetwork::processExampleToHighestLayer(const Matrix & x)
{
  SDN_ASSERT_MSG(x.rows()*x.cols() == _input.rows(), "x.rows() = " << x.rows() << ", x.cols() = " << x.cols() << "_input.rows() = " << _input.rows());
  matrixToMatrixCol(x,_input);
  //processExampleAsColumn(_input);
  if(should_transpose_input)
    forwardPropToHighestLayer(_input.transpose());
  else
    forwardPropToHighestLayer(_input);

}


/// \todo-- it would probably be a good idea to have both this and forwardProp call the same helper function so they don't have to be maintained separately
void StackedNetwork::forwardPropToHighestLayer(const Matrix& input) {

  _input = input;

  if( normalize_contrast ) {
    normalizePatch(_input);
  }
 
  if(storageNodes_.size() > 0)
    storageNodes_[0]->setActivities(_input);

  if(layers_.size() == 0)
    return;
  SDN_ASSERT_MSG(input.rows() == layers_[0]->inputDimension(),
		 "Input.rows(): " << input.rows() <<
		 " Input dimension:" << layers_[0]->inputDimension(););
  checkInvariants();
  
    for(unsigned i = 0; i < numLayers()/2; ++i)
    layers_[i]->forwardProp();

  checkInvariants();

}

void StackedNetwork::scrambleWeights(bool deepestOnly)
{
  if (deepestOnly)
    {
      layers_[layers_.size()/2]->scrambleWeights();
      layers_[layers_.size()/2-1]->scrambleWeights();
    }
  else
    {
      for (unsigned i = 0; i < layers_.size(); i++)
	layers_[i]->scrambleWeights();
    }
}
