/// Author: Andrew Saxe (asaxe@cs.stanford.edu)
/// \file stackedNetwork.h
/// \brief Defines a stacked network consisting of a number of Layer objects
//

#pragma once

#include "network.h"
#include "layer.h"
#include "standardLayer.h"
#include "storageNode.h"
//#include "sdn/gui.h"
#include <fstream>


namespace sdn
{

	class PatchDisplay;

/// A stacked network capable of containing a number of Layer objects
/** 
    Useful as a simple way to implement a network. Layer objects can
    be constructed and then added to the network with the AddLayer function.
    
    StorageNodes can be automatically created.
*/
class StackedNetwork : public Network {
  

  public:

    ///Sets up a StackedNetwork with default parameters
    StackedNetwork();
    virtual ~StackedNetwork();
    
    virtual long long train(long long nIter, bool allowHalting);
    
    /// Returns a pointer to the layer at depth d
    /**
       Depth d=0 corresponds to the layer closest to the input, and d=numLayers()-1
       corresponds to the layer nearest the output.

       \param d The depth of the layer to return. Must lie in [0,numLayers()-1].
    */
    boost::shared_ptr<Layer> getLayer(unsigned d) const;

    /// Returns a pointer to the storage layer at depth d
    /**
       Depth d=0 corresponds to the input storage layer, depth d=numLayers()+1 corresponds
       to the output storage layer.
    */
    boost::shared_ptr<StorageNode> getStorageNode(unsigned d) const;

    /// Adds layer to the network, and sets up appropriate storage layers
    /**
       Adds layer to the network and sets up appropriate storage layers using
       the outputDimension() and inputDimension() methods of the layer. Therefore, it is
       a checked runtime error for layer to be null. The layer is added on top
       of those added previously, and hence must satisfy 
       layer->inputDimension() == getLayer(numLayers()-1)->outputDimension() if 
       numLayers() > 0. This is a checked runtime error. After being added
       the layer belongs to the StackedNetwork class (will be destroyed upon
       destruction of the StackedNetwork class).

       \param layer The layer to add. Must not be null and must have input 
       dimension equal to the output dimension of the current lowest layer.
       The layer is owned by the StackedNetwork class.
    */
    void addLayer(boost::shared_ptr<Layer> layer);
    
    ///Adds layers based on an architecture XMLNode
    void addLayers(XMLNode& architecture);

    ///Resets everything to defaults, and then loads the supplied architecture
    void resetLayers(XMLNode& architecture);

    ///Adds a softmax layer for use in classification tasks
    /**
       The layer is added to the top of the encoding chain
       (middle of the linear stack) and sized to fit the 
       number of labels in the dataset (based on the numLabels() 
       property). Should only be called after a dataset that
       has labels has been registered (and will be automatically
       called in that case if shouldSetupSupervisedLayer(true) has
       been called previously.)
    */
    void addSupervisedLayer();

    /** Returns the number of Layers in the stacked network
	(encoding, convolution, decoding, and deconvolution
	layers, but not supervised layer) */
    unsigned numLayers() const;
    
 

    ///Adds an autoencoder (two single layer networks) to the stack
    void addAutoencoder(int num_hidden);

    ///Adds an autoencoder (two single layer networks) to the stack
    /**
      \param num_hidden The size of the hidden layer to add
       
      \param learning_rate The learning rate to use
      
      \param learning_rate_sparsity The learning rate for the sparsity
      update
      
      \param sparsity_averaging_timeconstant The time constant for averaging
      the sparsity measure. Should be between 0 and 1. Smaller values weight
      history more heavily.
    */
    void addAutoencoder(int num_hidden,
			Scalar learning_rate,
			Scalar learning_rate_sparsity,
			Scalar sparsity_averaging_timeconstant,
			Scalar target_sparsity);



  std::vector<Matrix> getEncoderWeights();
  std::vector<Matrix> getDecoderWeights();


    ///Adds an autoencoder (two single layer networks) to the stack
    /**
      \param num_hidden The size of the hidden layer to add
       
      \param params An xml node specifying other attributes (will be
      passed to the setParams() function.
    */
    void addAutoencoder(int num_hidden, XMLNode& params);

    ///Adds a convolution and deconvolution layer
    /**
       The input dimension is taken from the layer below, and the output
       dimension will be patch_width*patch_width*inputDimension().

       The convolution works by representing the 2D grid of inputs and
       outputs in column-major order. Therefore, the number of columns
       supplied as input must be equal to the number of elements covered 
       by the convolution, namely patch_width+patch_step*(output_width-1).
       
       Note that it is currently not supported to specify parameters that 
       yield a noninvertible convolution (that is, some inputs don't 
       propagate to the output), i.e., patch_step must be <= patch_width. 

       \param patch_width The width of the patch in the input that will be
       taken to form one element of the output. The extracted input is a 
       patch_width by patch_width square of elements.
       
       \param patch_step The spacing between neighboring patches in the input.
       
       \param output_width The output of the layer consists of an output_width
       by output_width square, represented in column-major order in the columns
       of the output.
    */
    void addConvolution(int patch_width, int patch_step, int output_width);


    ///Sets whether to call gradientDescentUpdate on the layer
    void setApplyGradientUpdate(int layer, bool applyGradientUpdate);
    
    ///Returns true if the network calls gradientDescentUpdate on the layer
    bool getApplyGradientUpdate(int layer);
    
    ///Sets whether to call gradientDescentUpdateSparsity on the layer
    void setApplyGradientUpdateSparsity(int layer, bool applyGradientUpdate);

    ///Returns true if the network calls gradientDescentUpdateSparsity on layer
    bool getApplyGradientUpdateSparsity(int layer);

    ///Registers a dataset with the class
    /**
       Creates a supervised layer for classification if
       shouldSetupSupervisedLayer(true) has been called previously.
    */
    void registerDataset(boost::shared_ptr<Dataset> dataset);

    ///Sets parameters of layers specified as a Params xml node
    void setParams(XMLNode& params);

    ///Sets parameters of item specified as xml node
    void setParams(unsigned index, XMLNode& params);
    
    /// Propagates input through all layers_ of the StackedNetwork
    /**
       The input is accepted as a matrix with features in the rows
       and examples in the columns. Thus input.rows() must equal 
       the input dimension of the first layer. This is a checked
       runtime error.
       
       \param input The input to propagate.
       \param train: True if called while training. Currently the only effect is that this may activate denoising
    */
    virtual void forwardProp(const Matrix& x , bool train=false);


    /**
       The highest layer is defined as layer numLayers()/2

       The input is accepted as a matrix with features in the rows
       and examples in the columns. Thus input.rows() must equal 
       the input dimension of the first layer. This is a checked
       runtime error.
       
       \param input The input to propagate.
    */
    void forwardPropToHighestLayer(const Matrix& x);

    /// Applies the target pattern to the last layer in the StackedNetwork
    /**
       You must have first called forwardProp with the corresponding input.
       
       \param target  The target pattern to be applied. A target must be supplied
       for each example that was supplied in fowardProp, that is, the matrix input 
       used in forwardProp must have the same number of columns as the matrix 
       target supplied here. The target must have the same size as the output 
       dimension of the last layer.
    */
    virtual void presentTarget(const Matrix& target);

    /// Backpropagates error gradients through the StackedNetwork
    /**
       You must first have called forwardProp and presentTarget
       Computes gradients w.r.t. the cost function and sparsity,
       which are then stored in each layer.
    */
    virtual void backProp();

    /// Calls gradient descent on all layers
    /**
       You must first have called forwardProp, presentTarget, and backProp.
    */
    virtual void gradientDescentUpdate();

    /// Calls gradientDescentSparsity on all layers
    /**
       You must first have called forwardProp, presentTarget, and backProp.
    */
    virtual void gradientDescentUpdateSparsity();

    /// Returns a running average of the error rate on the training set
    Scalar getAverageError();

    /// Returns a running average of the sparsity of hidden activities in the deepest layer
    double getAverageSparsity();

    /// Save to archive object-- note: does not save the actual content of the input, or the dataset that the network is linked to
    virtual void serialize(serial::Archive & ar);

    /// Identical to forwardProp(x) \deprecated forwardProp(x) is prefered \todo phase this out in favor of forwardProp (which is already columnwise).
    void processExampleAsColumn(const Matrix& x); 

    /// Returns true by default
    bool supportsReconstruction() const;

    /// Returns the output at the last layer
    /**
       You must have previously called forwardProp.
       
       \param x The reconstruction is returned in this variable.
    */
    void getReconstruction(Matrix & x);

    /// Returns the activities at the given index
    /**
       Indices start from 0 (0 would be the input that is supplied, 1 the output of the first Layer, etc).
    */
    const Matrix& getActivities(unsigned index);

    /// Returns the activities of the 'highest' layer
    /**
       The highest layer is defined as the layer in the middle of the stack (numLayers()/2),
       which is appropriate for stacked autoencoder type architectures.
    */
    const Matrix& getHighestLayerActivities();
    

    /// Returns the derivatives at the given index
    /**
       Indices start from 0 (0 would be the input that is supplied, 1 the output of the first Layer, etc).
    */
    const Matrix& getDerivatives(unsigned index);

    bool active();
    
    void display();



    /// Right now, just returns the output of the deepest encoding layer. We could make this configurable though, so the test could be run on different layers by calling a method to change the behavior of this function prior to running the test
    boost::shared_ptr<StorageNode> getTestingNode();


 
    void scrambleWeights(bool deepestOnly);
    
    
    virtual void processExampleToHighestLayer(const Matrix & x);

    /// \todo-- maybe this should query its layers or something?
    bool isTwoDimensional()
    {
      return true;
    }

    /// Clears all storage layers and resets parameters to their default settings
    /**
       The storage layer will be left in the same state as if you had just 
       constructed it with the no args constructor and no config file. 
    */
    void clear();

    void printDebugStuff();
    
  public:
    /// Time constant used in exponentially weighted avg of error
    /**
       Should be between 0 and 1.
    */
    double error_averaging_timeconstant;

    /// True if the input column obtained during training should be transposed
    /**
       When the first layer is a convolution, this is necessary. Then each 
       component in the input is not a feature, but a position in the grid, 
       and hence represented in columns as opposed to rows.
    */
    bool should_transpose_input;

    bool train_unsupervised;
    bool train_supervised;

    /// True if the network should rescale each input patch to the range [-1,1]
    bool normalize_contrast;
    
  protected:


    bool equals_(const Network & other);

    void checkInvariants();

    void setDefaults_();

    /// Clears all 
    void clear_();
    

    std::vector<boost::shared_ptr<Layer> > layers_;
    std::vector<boost::shared_ptr<StorageNode> > storageNodes_;
    
    boost::shared_ptr<StandardLayer> supervised_layer_;
    boost::shared_ptr<StorageNode> supervisedStorageNode_;

    double error_avg_;

    boost::shared_ptr<PatchDisplay> visualizer_;
    bool active_;

    //amount to denoise by. only applied if denoiseLayer_ != -1
    double denoising_;
    //add noise to storageNodes_[denoiseLayer]. if -1, don't use denoising
    int denoiseLayer_;
    //a clean copy of the input that had noise added to it.
    Matrix prenoisedInput_;
    // a pointer to the matrix to reconstruct. currently only behaves correctly if reconstructing the input or if using denoising.
    Matrix * desiredOutput_;

    //Index of layer to learn to reconstruct during training (-1 = just assume reconstruct the original input)
    int reconstructionTarget_;
    //Layer than contains the reconstruction of reconstructionTarget_
    int reconstructionLayer_;
  };
}

//SDN_AUTOREGISTER_H(Network, StackedNetwork);


