// Copyright 2009, All Rights Reserved.
// Authors: Andrew Saxe (asaxe@stanford.edu)
//          Ian Goodfellow (ia3n@cs.stanford.edu)
//          Honglak Lee (hllee@cs.stanford.edu)                                                                                
/** \file standardLayer.cc
    \brief  Defines the interface of the SingleLayerNetwork.
*/

#define NOISY_HIDDEN 0.5
                                                                                
#include "standardLayer.h"
#include "activationFunction.h"
#include "costFunction.h"
#include "storageNode.h"
#include "sdn/base.h"
#include "normPenalty.h"
#include <iomanip>
#include <assert.h>
#include "../base/matrixUtils.h"
#include <iostream>
using namespace std;

#include <Eigen/Array>
using namespace sdn;
using namespace boost;

static double sign(double n)
{
  if (n > 0)
    return 1.0;
  else if (n < 0)
    return -1.0;
  return 0.0;
}


StandardLayer::StandardLayer(
     boost::shared_ptr<ActivationFunction> activation_fn,
     boost::shared_ptr<CostFunction> cost_fn )
  : Layer()
{
  initialize(
	     boost::shared_ptr<StorageNode>(), 
	     boost::shared_ptr<StorageNode>(), 
	     activation_fn,
	     cost_fn,
             false);
}

StandardLayer::StandardLayer(
				       int input_dimension,
				       int output_dimension,
				       boost::shared_ptr<ActivationFunction> activation_fn,
				       boost::shared_ptr<CostFunction> cost_fn ) 
  
{
  expectDimensions(input_dimension, output_dimension);

  initialize(
	     boost::shared_ptr<StorageNode>(), 
	     boost::shared_ptr<StorageNode>(),
	     activation_fn,
	     cost_fn,
             false);
}



void StandardLayer::initialize(boost::shared_ptr<StorageNode> input,
				    boost::shared_ptr<StorageNode> output,
				    shared_ptr<ActivationFunction> 
				    activation_fn,
				    shared_ptr<CostFunction> cost_fn,
                                    bool storageExists) {



  if (!weights_)
    weights_ = ptr(new Matrix());

  


  //Default learning rates
  learning_rate = 0.001;
  learning_rate_sparsity = 0.0;
  sparsity_averaging_timeconstant = 0.05;
  target_sparsity_ = 0.5;
  sparsity_deadzone_ = 0.0;
  sparsity_initialized_ = false;
  


  if (storageExists) {
    link(input, output);
  }

  if (activation_fn == NULL)
    activation_fn_ = shared_ptr<TanhActivationFunction>(new TanhActivationFunction());
  else
    activation_fn_ = activation_fn;

  if (cost_fn == NULL)
    cost_fn_ = shared_ptr<SumOfSquaresCostFunction>(new SumOfSquaresCostFunction());
  else
    cost_fn_ = cost_fn;
 
  
  if (storageExists) 
    checkInvariants();

}

StandardLayer::~StandardLayer() {
  //do nothing
}


void StandardLayer::link(boost::shared_ptr<StorageNode> input, boost::shared_ptr<StorageNode> output) {
  Layer::link(input, output);
  initializeWeights();
}

void StandardLayer::initializeWeights() {
  SDN_ASSERT(inputDimension() > 0);
  SDN_ASSERT(outputDimension() > 0);
    
  //initialize the weights in the range [-1/sqrt(k), 1/sqrt(k)] as suggested  
  //by Erhan, Manzagol, Bengio, Bengio and Vincent.

  *weights_ = Matrix::Random(outputDimension(), inputDimension())/sqrt(inputDimension());

  RandomRealGenerator<double> rrg;
  rrg.setSeed(time(NULL));

  for (int i = 0; i < weights_->rows(); i++)
	for (int j = 0; j < weights_->cols(); j++)
		(*weights_)(i,j) = rrg.randUnif(-1.0,1.0)/sqrt(inputDimension());


  //SDN_MSG("Weight range: [" << -1.0/sqrt(inputDimension()) << ", " << 1.0/sqrt(inputDimension()) << "] actual: [" << weights_->minCoeff() << ", " << weights_->maxCoeff() << "]" );
  
  if (activation_fn_->hasInverse())
    {
      //Initialize the biases to approximately set the neuron firing at the desired sparsity target
      //Only works if the data is approximately zero mean.
      biases_ = Matrix::Constant(outputDimension(), 
				 1, 
				 activation_fn_->inverse(
							 (activation_fn_->max() - activation_fn_->min())
							 *target_sparsity_ + activation_fn_->min()));
    }
  else
    {
      biases_ = Matrix::Zero(outputDimension(), 1);
    }
}

void StandardLayer::resync()
{
  initialize(input_,output_, activation_fn_, cost_fn_);
}

void StandardLayer::forwardProp(
#if NOISY_HIDDEN_HACK
	bool noise
#endif

) {
  checkInvariants();



  // Treat biases as a matrix
  int numsamples= input_->getActivities().cols();

  net_input_ = (*weights_)*input_->getActivities() + biases_*
                                                  Matrix::Ones(1, numsamples);

#if NOISY_HIDDEN_HACK
	if (noise)
	{
		for (int i = 0; i < net_input_.rows(); i++)
			for (int j = 0; j < net_input_.cols(); j++)
			{
				if ( double (rand() % 1000) / 1000.0 < NOISY_HIDDEN)
				{
					net_input_(i,j) = -1.0;
				}
			}
	}  
#endif



  output_->computeActivities(activation_fn_, net_input_);


}

void StandardLayer::forwardProp(const Matrix& input) {


   SDN_ASSERT_MSG(input.rows() == input_->size(), 
		 "input.rows()="<<input.rows()<<
		 " but input_->size()="<<input_->size());
  checkInvariants();

  setInput(input);
  forwardProp();

}

void StandardLayer::backProp() {

  checkInvariants();


  const Matrix& output_activities = output_->getActivities();
  const Matrix& error_derivative_wrt_output = output_->getDerivatives();

  assert(error_derivative_wrt_output.rows() == output_activities.rows());
  assert(error_derivative_wrt_output.cols() == output_activities.cols());
  assert(input_->getActivities().cols() == output_activities.cols());


  if (net_input_.rows() != weights_->rows() || net_input_.cols() != input_->getActivities().cols())
  {
        SDN_WARN(this<<".backProp() resizing net input (did you forget to run forwardProp?)");
        net_input_ = Matrix::Zero(weights_->rows(),input_->getActivities().cols());

  }

  //Calculate delta
  if (delta_.rows() != output_activities.rows() || delta_.cols() != output_activities.cols())
  {
	SDN_DEBUG("Resizing delta from "<<delta_.rows()<<" rows to "<<output_activities.rows());
  	SDN_DEBUG("Resizing delta from "<<delta_.cols()<<" cols to "<<output_activities.cols());
        delta_ = Matrix::Zero(output_activities.rows(), output_activities.cols());
  }

  for(int c = 0; c < output_activities.cols(); c++) {
    activation_fn_->outputDerivative(net_input_.col(c),derivative_,true);
    if (derivative_.cols() == 1)
      {  delta_.col(c) = derivative_.col(0).cwise() *
	error_derivative_wrt_output.col(c);
      }
    else
      {
	SDN_ASSERT(derivative_.rows() == derivative_.cols());
	delta_.col(c) = 
	  derivative_.transpose()
	  *error_derivative_wrt_output.col(c);
      }
  }


  //Calculate derivative w.r.t. input
  input_->setDerivatives(weights_->transpose()*delta_);

  //Calculate derivative w.r.t. weights
  error_derivative_wrt_weights_ = delta_*input_->getActivities().transpose();

  //Calculate derivative w.r.t. biases
  error_derivative_wrt_biases_ = delta_.rowwise().sum();

  //Calculate (approximate) derivative w.r.t. biases (sparsity)
  //asaxe: trying a weighted average to remove the need for minibatches
  if(!sparsity_initialized_) {
    sparsity_avg_act_ = Matrix::Constant(output_activities.rows(), 1, target_sparsity_);//(output_activities.rowwise().sum().cwise() + 1.0)/2.0;
    //SDN_DEBUG("Initial sparsity avg act : " << sparsity_avg_act_);
    sparsity_initialized_ = true;
    sparsity_avg_act_ = (output_activities.rowwise().sum().cwise()+1.0)/2.0 * sparsity_averaging_timeconstant + sparsity_avg_act_*(1.0 - sparsity_averaging_timeconstant);
    
  } else {
    //SDN_DEBUG("Updated sparsity avg act with time constant "<<sparsity_averaging_timeconstant);
    //SDN_DEBUG("Activities: "<<output_activities);
    sparsity_avg_act_ = ((output_activities.cwise()+1.0)/2.0).rowwise().sum() * sparsity_averaging_timeconstant + sparsity_avg_act_*(1.0 - sparsity_averaging_timeconstant);
  }
 
  //SDN_DEBUG("sparsity_avg_act: "<<sparsity_avg_act_);

  error_derivative_wrt_sparsity_ = sparsity_avg_act_.cwise()-target_sparsity_*output_activities.cols();//output_activities.rowwise().sum().cwise()    



  if (sparsity_deadzone_ > 0.0)
    {
      for (int i = 0; i < error_derivative_wrt_sparsity_.rows(); i++)
	for (int j = 0; j < error_derivative_wrt_sparsity_.cols(); j++)
	  if (abs(error_derivative_wrt_sparsity_(i,j)) < sparsity_deadzone_ / 2.0)
	    {
	      error_derivative_wrt_sparsity_(i,j) = 0.0;
	    }
	  else
	    {
	      error_derivative_wrt_sparsity_(i,j) -= sign(error_derivative_wrt_sparsity_(i,j)) * sparsity_deadzone_ / 2.0;
	    }
	      
    }

                                //- target_sparsity_*output_activities.cols();

  //SDN_MSG("Sparsity: " << getAverageSparsity());
  //cout << setprecision(24);
  //SDN_DEBUG(sparsity_avg_act_(0,0) - 0.5);
  //SDN_DEBUG( setprecision(24) << "edws: " << setprecision(24) <<  error_derivative_wrt_sparsity_ << " = " << setprecision(24) <<sparsity_avg_act_*1000000.0 << " - " << setprecision(24) << target_sparsity_*output_activities.cols());
}

void StandardLayer::backProp(
     const Matrix& error_derivative_wrt_output) {

  checkInvariants();
  SDN_ASSERT_MSG(error_derivative_wrt_output.rows() == 
	     output_->getActivities().rows(), error_derivative_wrt_output.rows()<<" "<<output_->getActivities().rows());
  SDN_ASSERT(error_derivative_wrt_output.cols() == 
	     output_->getActivities().cols());
   
/** \todo-- guess this means we're storing the derivatives in two places
      (or the caller is calculating it in a temporary). Opportunity for a design
      improvement? */
  output_->setDerivatives(error_derivative_wrt_output);
   
  backProp();
}

Scalar StandardLayer::calculateError( 
    const Matrix& target) {
  checkInvariants();
  SDN_ASSERT(output_->getActivities().cols() == target.cols());
  assert(output_->getActivities().rows() == target.rows());
  Matrix ret = cost_fn_->output(output_->getActivities(), target);
  return ret(0,0);
}

void StandardLayer::presentTarget(
    const Matrix& target) {
  SDN_ASSERT_MSG(output_->getActivities().rows() == target.rows(), 
		 "Output rows: " << output_->getActivities().rows() << 
		 " Target rows: " << target.rows() );
  SDN_ASSERT_MSG(output_->getActivities().cols() == target.cols(), 
		 "Output cols: " << output_->getActivities().cols() << 
		 " Target cols: " << target.cols() );
  // \todo(asaxe): Prevent ppl from calling this before calling ForwardProp.

/** \todo-- investigate whether computing the derivatives in a temporary and then
    copying is causing too much of a slowdown (doubt it because memory isn't
    fluctuating visibly) */
  output_->setDerivatives(
			  cost_fn_->jacobian(
			  output_->getActivities(), target).transpose());
  
}

/*const Matrix& StandardLayer::errorDerivativeWrtInput() {
  checkInvariants();
  return input_->getDerivatives();
  }*/

const Matrix& StandardLayer::errorDerivativeWrtWeights() {
  return error_derivative_wrt_weights_;
}

const Matrix& StandardLayer::errorDerivativeWrtBiases() {
  return error_derivative_wrt_biases_;
}

void StandardLayer::gradientDescentUpdate(Scalar learning_rate) {

 if(learning_rate == 0.0)
    return;

  for (vector<shared_ptr<UnaryObjectiveTerm> >::iterator i = 
	 weightPenalties_.begin(); i != weightPenalties_.end(); ++i)
    {

      //SDN_DEBUG("Applying a weight penalty");
      
      //SDN_DEBUG("weights_: "<<weights_->rows()<< " " << weights_->cols());
      //SDN_DEBUG("derivatives_: "<< ((ObjectiveTerm *) i->get())->matrixDerivative().rows() << " " << ((ObjectiveTerm *) i->get())->matrixDerivative().cols());
      
      /// \todo Why is this cast necessary? Do methods in derived classes shadow methods in parent classes that have the same name but different signature? 
      /// \todo Technically, all of the matrixDerivatives should be evaluated from the original value of the weights, but actually that will only hold for the first of these. is it safe to rename error_derivative_wrt_weights as objective_derivative_wrt_weights and have this for loop accumulate to it? (I see that the value gets passed out of the class and I don't know what other classes use it / what assumptions they make)
     

      *weights_ -= learning_rate*((ObjectiveTerm *) i->get())->matrixDerivative();
    }

 
  //SDN_DEBUG("gradient descent update: "<<learning_rate);

  
  *weights_ -= learning_rate*error_derivative_wrt_weights_;

  //SDN_DEBUG("error_derivative_wrt_biases_(0,0) "<<error_derivative_wrt_biases_(0,0));
  
  biases_ -= learning_rate*error_derivative_wrt_biases_;
}

void StandardLayer::gradientDescentUpdate() {

  gradientDescentUpdate(learning_rate);
}

void StandardLayer::gradientDescentUpdateSparsity(Scalar learning_rate_sparsity) {
 
  //SDN_DEBUG("sparsity update " <<learning_rate_sparsity);

  if(learning_rate_sparsity == 0)
    return;


  SDN_ASSERT(error_derivative_wrt_sparsity_.rows() == biases_.rows());
  SDN_ASSERT(error_derivative_wrt_sparsity_.cols() == biases_.cols());

  //SDN_DEBUG(" SPARSITY EQUATION:" << biases_ << " -= " << learning_rate << " * "  << learning_rate_sparsity << " * (" << sparsity_avg_act_ << " - " << target_sparsity_ << ") or " << error_derivative_wrt_sparsity_);

  biases_ -= learning_rate*learning_rate_sparsity*error_derivative_wrt_sparsity_;
  
  //SDN_DEBUG("Which equals " << biases_);
  
}
void StandardLayer::gradientDescentUpdateSparsity() {
  gradientDescentUpdateSparsity(learning_rate_sparsity);
}

Scalar StandardLayer::getAverageSparsity() {
  return sparsity_avg_act_.sum()/sparsity_avg_act_.rows();
}

/*void StandardLayer::setOutput(const Matrix& output) {
  checkInvariants();
  output_->setActivities(output);
}

void StandardLayer::setErrorDerivativeWrtOutput(
    Matrix& error_derivative_wrt_output) {
  checkInvariants();
  output_->setDerivatives(error_derivative_wrt_output);
  }*/
///sets the input to the given value
/*void StandardLayer::setInput(const Matrix& input) {
  checkInvariants();
  input_->setActivities(input);
}

void StandardLayer::setErrorDerivativeWrtInput(
    Matrix& error_derivative_wrt_input) {
  checkInvariants();
  input_->setDerivatives(error_derivative_wrt_input);
  }*/

const Scalar StandardLayer::targetSparsity() {
	return target_sparsity_;
}

void StandardLayer::setTargetSparsity(Scalar sparsity) {
  target_sparsity_ = sparsity;
}

const Matrix& StandardLayer::getWeights() {
  return *weights_;
}

void StandardLayer::setWeights(const Matrix& weights) {
  assert(weights_->rows() == weights.rows());
  assert(weights_->cols() == weights.cols());
  *weights_ = weights;
}

void StandardLayer::setBiases(const Matrix& biases) {
  assert(biases_.rows() == biases.rows());
  assert(biases_.cols() == biases.cols());
  biases_ = biases;
}

void StandardLayer::setAllBiases(Scalar b)
{
  biases_.setConstant(b);
}

void StandardLayer::checkInvariants() {
  assert(input_ != NULL);     
  assert(output_ != NULL);
  
  SDN_ASSERT_MSG((unsigned) weights_->rows() == output_->size(), "StandardLayer::checkInvariants(): output storageNode has size "<<output_->size()<<" but layer has "<<weights_->rows()<<" rows of weights");
  assert((unsigned) weights_->cols() == input_->size());
  
  assert((unsigned) biases_.rows() == output_->size());
  assert(biases_.cols() == 1);

  assert(target_sparsity_ >= activation_fn_->min());
  assert(target_sparsity_ <= activation_fn_->max());
}

void StandardLayer::addWeightPenalty(boost::shared_ptr<UnaryObjectiveTerm> p)
{
  weightPenalties_.push_back(p);
  p->registerArgument(weights_);
}

void StandardLayer::addWeightDecay(double lambda)
{
  shared_ptr<L2Penalty> temp = ptr(new L2Penalty());
  temp->lambda = lambda;
  addWeightPenalty(temp);
}

void StandardLayer::serialize(serial::Archive & ar)
{
  SDN_DEBUG("StandardLayer::serialize");
  
  int version = 37;

  version = ar.beginDerived(this,version,"sdn::StandardLayer");
  
  switch (version)
    {
    case 37:
      ar.beginSuper();
      Layer::serialize(ar);
      ar.endSuper();

      serializeMatrixSize(ar,error_derivative_wrt_weights_);
      serializeMatrixSize(ar,derivative_);
      serializeMatrixSize(ar,error_derivative_wrt_sparsity_);
      serializeMatrixSize(ar,net_input_);
      serializeMatrixSize(ar,delta_);

      serialize_shared_ptr(ar, CostFunctionRegistry(), cost_fn_);
      serialize_shared_ptr(ar, ActivationFunctionRegistry(), activation_fn_);
    
      serialize_shared_ptr(ar, weights_,  MatrixSerializer());
      serializeMatrix(ar,biases_);

      serializeSTLVector(ar, weightPenalties_, PolymorphicSharedPointerSerializer<UnaryObjectiveTerm>(UnaryObjectiveTermRegistry()));

  
      ar.endDerived();
    }

}

bool StandardLayer::equals(const Layer & o)
{
  const StandardLayer* sln = dynamic_cast<const StandardLayer*>(&o);
  if( !sln )
    return false;
  const StandardLayer& other = *sln;
  
  if (*weights_ != *other.weights_)
    {
      SDN_DEBUG("Weights don't match");

      if (weights_->rows() != other.weights_->rows())
	{
	  SDN_DEBUG("Rows don't match: "<<weights_->rows()<<" vs "<<other.weights_->rows());
	}

      if (weights_->cols() != other.weights_->cols())
	{
	  SDN_DEBUG("Cols don't match: "<<weights_->cols()<<" vs "<<other.weights_->cols());
	}

      for (int i = 0; i < weights_->rows(); i++)
	{
	  bool found = false;

	  for (int j = 0; j < weights_->cols(); j++)
	    {
	      if ((*weights_)(i,j) != (*other.weights_)(i,j))
		{
		  found = true;
		  SDN_DEBUG("Don't match at ("<<i<<","<<j<<"): "<<(*weights_)(i,j) << " vs "<<(*other.weights_)(i,j));
		  break;
		}
	    }

	  if (found)
	    break;
	}

      //SDN_DEBUG("weights_(0,0): "<<(*weights_)(0,0)<<" vs "<<(*other.weights_)(0,0));
      
      return false;
    }
  
  if (weightPenalties_.size() != other.weightPenalties_.size())
    {
      SDN_DEBUG("weight penalties don't match");
      return false;
    }
  if (biases_ != other.biases_)
    {
      SDN_DEBUG("Biases don't match");
      return false;
    }
  
  SDN_WARN("StandardLayer::equals is far from thorough");
  
  return true;
}

void StandardLayer::setParams(XMLNode& node) {
 
 SDN_DEBUG(this << ": StandardLayer::setParams");
 
  const char* learning_rate_str = node.getAttribute("learning_rate");
  if(learning_rate_str) {
    learning_rate = atof(learning_rate_str);
  }

  const char* learning_rate_sparsity_str = node.getAttribute("learning_rate_sparsity");
  if(learning_rate_sparsity_str) {
    learning_rate_sparsity = atof(learning_rate_sparsity_str);
    SDN_DEBUG(this<< ": setting learning rate sparsity" << learning_rate_sparsity);
  }

  const char* sparsity_averaging_timeconstant_str = node.getAttribute("sparsity_averaging_timeconstant");
  if(sparsity_averaging_timeconstant_str) {
    sparsity_averaging_timeconstant = atof(sparsity_averaging_timeconstant_str);
    SDN_DEBUG("Setting sparsity avg " << sparsity_averaging_timeconstant);
  }

  const char* target_sparsity_str = node.getAttribute("target_sparsity");
  if(target_sparsity_str) {
    target_sparsity_ = atof(target_sparsity_str);
    SDN_DEBUG(this << ": setting sparsity target " << target_sparsity_);
  }

  const char * sparsity_deadzone_str = node.getAttribute("sparsity_deadzone");
  if (sparsity_deadzone_str)
    {
      sparsity_deadzone_ = atof(sparsity_deadzone_str);
      SDN_DEBUG("Setting sparsity deadzone "<< sparsity_deadzone_);
    }

  
  const char* weight_decay_str = node.getAttribute("weight_decay");
  if(weight_decay_str) {
    Scalar weight_decay = atof(weight_decay_str);

  SDN_DEBUG(this << ": setting weight decay "<<weight_decay);    

    if( weightPenalties_.size() == 1 ) { /// \todo make this work for multiple penalty types (e.g., add a setWeightDecay method)
      
      shared_ptr<L2Penalty> pen = dynamic_pointer_cast<L2Penalty>(weightPenalties_[0]);
      SDN_ASSERT_MSG(pen, "Dynamic cast failed.");
      pen->lambda = weight_decay;

    } else { //Need to add weight decay term

      addWeightDecay(weight_decay);

    }
    
  }
  
}

void StandardLayer::printDebugStuff() {
  SDN_DEBUG("input:\n" << input_->getActivities());
  SDN_DEBUG("error d wrt_input:\n" << input_->getDerivatives());
  SDN_DEBUG("activations:\n" << output_->getActivities());
  SDN_DEBUG("error derivative wrt output:\n" << output_->getDerivatives());
  SDN_DEBUG("weights:\n" << weights_);
  SDN_DEBUG("error_d_wrt_w:\n" << error_derivative_wrt_weights_);
  SDN_DEBUG("derivative:\n" << derivative_);
  SDN_DEBUG("sparsity_avg_act:\n" << sparsity_avg_act_);
  SDN_DEBUG("biases:\n" << biases_);
  SDN_DEBUG("error_d_wrt_biases:\n" << error_derivative_wrt_biases_);
  SDN_DEBUG("error_d_wrt_sparsity:\n" << error_derivative_wrt_sparsity_);
  SDN_DEBUG("net_input:\n" << net_input_);
  SDN_DEBUG("delta_:\n" << delta_);
  SDN_DEBUG("sparsity_avg_: " << sparsity_avg_);
  SDN_DEBUG("target_sparsity_: " << target_sparsity_);
  SDN_DEBUG("learning_rate: " << learning_rate);
  SDN_DEBUG("learning_rate_sparsity: " << learning_rate_sparsity);
  SDN_DEBUG("sparsity_averaging_timeconstant: " << sparsity_averaging_timeconstant);
  
}

void StandardLayer::scrambleWeights()
{
  scrambleRowwise(*weights_);
}
