// Copyright 2009, All Rights Reserved.
// Authors: Andrew Saxe (asaxe@stanford.edu)
//          Ian Goodfellow (ia3n@cs.stanford.edu)
//          Honglak Lee (hllee@cs.stanford.edu)

/** \file standardLayer.h
    \brief  Declares the interface of a simple one layer network.
*/

#pragma once

#include "boost/shared_ptr.hpp"
#include "layer.h"
//#include "costFunction.h"
//#include "activationFunction.h"
#include <vector>
//#include "unaryObjectiveTerm.h"



namespace sdn
{

//Forward declarations
class ActivationFunction;
class CostFunction;
class StorageNode;
class UnaryObjectiveTerm;

/**
   Standard neural network layer consisting of an affine transform
   followed by a non-linearity.
 */

 class StandardLayer : public Layer {

 public:

  ///Constructs a single layer network with undefined dimensions
  //StandardLayer();

  /// Constructs a single layer network.
  /** 
      If no arguments are given, this
       constructor automatically sets the layer's
      activation function to the TanhActivationFunction, and the cost
      function to the SumOfSquaresCostFunction.

      The object cannot be used until the link function has been
      called to supply input and output storage layers.

     \param activation_fn The activation function to use.

     \param cost_fn The cost function to use.
   */


  StandardLayer(boost::shared_ptr<ActivationFunction> \
     activation_fn
		     = boost::shared_ptr<ActivationFunction>()
		     , 
     boost::shared_ptr<CostFunction> cost_fn
		     = boost::shared_ptr<CostFunction>()
		     );

  /// Constructs a single layer network of the specified dimension
  /**
     Creates a single layer network with the specified input and output
     dimensions, but does not connect the StandardLayer to any
     StorageNodes. This must be done subsequently via link, and then
     the supplied StorageNodes must match the dimensions given.

  */
  StandardLayer(int input_dimension, int output_dimension,
		     boost::shared_ptr<ActivationFunction>	\
		     activation_fn = boost::shared_ptr<ActivationFunction>(),
		     boost::shared_ptr<CostFunction> cost_fn = 
		     boost::shared_ptr<CostFunction>());

 

  /// Links the single layer network from one storage layer to another

  /** 
     \param input The storage layer to use as input. The size of the
     storage layer determines the size of the input.

     \param output The storage layer to use as output. The size of the
     storage layer determines the size of the output.
  */
  void link(boost::shared_ptr<StorageNode> input, boost::shared_ptr<StorageNode> output);

  /// Initializes the weight matrix randomly
  /**
     Initializes the weight matrix to values drawn uniformly in 
     \f$ \left[-\frac{1}{\sqrt(n)},\frac{1}{\sqrt(n)} \right] \f$ where 
     \f$ n \f$ is the input dimension. This initialization was suggested by
     Erhan, Manzagol, Bengio, Bengio, and Vincent.
  */
  void initializeWeights();

  ///Destroys the network (but not the input and output StorageNodes).
  ~StandardLayer();

  ///Notifies the class that its storage layers have changed in size
  /**
     Triggers a re-initialization of the weights.
  */
  void resync();


  ///Propagates activity from the input to the output.
  /**
     Computes the output activation vector for each pattern supplied
     as input.  Does _not_ sample, and hence represents probability
     for RBM interpretations. In particular, it calculates
     \f$h=f(WX)\f$, where \f$h\f$ is the output, \f$f(x)\f$ is the
     activation function, \f$W\f$ is the weight matrix, and \f$X\f$ is
     the input matrix with different input examples in each column.
     
     You must have previously set an input source. It is a checked
     runtime error for this input source to be null. It is a checked
     runtime error for this input source to have a number of rows that
     differs from InputSize().
  */
  virtual void forwardProp(
#if NOISY_HIDDEN_HACK
bool noise 
#endif
);

#if NOISY_HIDDEN_HACK
	virtual void forwardProp()
	{
		forwardProp(false);
	}
#endif

  virtual void forwardProp(const Matrix& input);

  ///Calculates the derivative of the error w.r.t. the input and weights.
  /**
     Computes the derivatives \f$\frac{\partial E}{\partial I} \f$ and
     \f$\frac{\partial E}{\partial W}\f$ where \f$E\f$ is the cost
     function, \f$I\f$ is the input, and \f$W\f$ is the weight matrix.
     These derivatives are computed and stored for later retrieval
     with the ErrorDerivativeWrtInput() and
     ErrorDerivativeWrtWeights() functions.

     An input source, output source, and error derivative
     w.r.t. output must have been populated previously, e.g. by
     calling PresentTarget(), or calling BackProp() on a higher layer.

     \return A pointer to the error derivative w.r.t. the weights.
   */
  void backProp();

 
  ///Calculates the derivative of the error w.r.t. the input and weights.
  /**
     Computes the derivatives \f$ \frac{\partial E}{\partial I} \f$
     and \f$\frac{\partial E}{\partial W}\f$ where \f$E\f$ is the cost
     function, \f$I\f$ is the input, and \f$W\f$ is the weight matrix.
     These derivatives are computed and stored for later retrieval
     with the ErrorDerivativeWrtInput() and
     ErrorDerivativeWrtWeights() functions.

     An input source and output source must have been populated
     previously.
   
     \param error_derivative Contains the derivative of the error
     function w.r.t. the output activations, that is, \f$
     \frac{\partial E}{\partial f} \f$ where \f$ f \f$ are the unit
     activities. The row indices specify the derivatives
     w.r.t. different unit activities \f$ f_i \f$, and the columns
     specify derivatives w.r.t.  different input patterns.
   */
  void backProp(const Matrix& error_derivative);

  ///Calculates the error between the last computed network output and the given target pattern using this network's cost function.
  /**
     \param target The matrix of target patterns, with a pattern in
     each column. Must have the same dimensions as the output.
   */
  Scalar calculateError(const Matrix& target);

  ///Calculates the derivative of the cost function with respect to the output activations.
  /**
     Calculates \f$\frac{\partial E}{\partial f}\f$ where \f$E\f$ is
     the cost function and \f$f\f$ is the vector of output unit
     activation. This is the first step to computing the gradient of
     the cost function w.r.t. the parameters.
   */
  void presentTarget(const Matrix& target);

  ///Returns the derivative of the error w.r.t. the input.
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). It is not yet a checked runtime error to do
     otherwise but it should be.
   */
  //const Matrix& errorDerivativeWrtInput();

  ///Returns the derivative of the error w.r.t. the weights.
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). It is not yet a checked runtime error to do
     otherwise but it should be.
   */
  const Matrix& errorDerivativeWrtWeights();

  ///Returns the derivative of the error w.r.t. the biases.
  const Matrix& errorDerivativeWrtBiases();


  /// Adds an objective term that is a function of the weights
  void addWeightPenalty(boost::shared_ptr<UnaryObjectiveTerm> p);
  
  /// Adds L2 weight decay with coefficient $\lambda$
  void addWeightDecay(double lambda);


  ///Takes a gradient descent step.
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). It applies the simple weight update \f[W\leftarrow
     W-\alpha \frac{\partial E}{\partial W}\f] where \f$W\f$ is the
     weight matrix, \f$\alpha\f$ is the learning rate, and \f$E\f$ is
     the error function.

     This will also follow the gradient of any registered weight penalties


     \param learning_rate The learning rate to use (\f$\alpha\f$ in
     the equation above.)
   */
  void gradientDescentUpdate(Scalar learning_rate);
  
  ///Takes a gradient descent step.
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). It applies the simple weight update \f[W\leftarrow
     W-\alpha \frac{\partial E}{\partial W}\f] where \f$W\f$ is the
     weight matrix, \f$\alpha\f$ is the learning rate, and \f$E\f$ is
     the error function. The learning rate used is the learning_rate variable
     stored in the class.

     This will also follow the gradient of any registered weight penalties

   */
  void gradientDescentUpdate();

  

  ///Takes a gradient descent step (with respect to sparsity penalty).
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). This will update using only the gradient of
     sparsity penalty. If you don't use sparsity regularization, then
     simply skip calling this function. 
  **/
  void gradientDescentUpdateSparsity(Scalar learning_rate);

  ///Takes a gradient descent step (with respect to sparsity penalty).
  /**
     This function can only be called after calling both ForwardProp()
     and BackProp(). This will update using only the gradient of
     sparsity penalty. If you don't use sparsity regularization, then
     simply skip calling this function. Uses the learning_rate_sparsity
     variable in the class.
  **/
  void gradientDescentUpdateSparsity();

  ///Returns a running average of the sparsity of the output layer
  Scalar getAverageSparsity();

  ///Sets the output to the given value. 
  //void setOutput(const Matrix& output);
  
  ///Sets the error derivative w.r.t. the output.
  /**
     The derivative w.r.t. the input or weights can then be calculated
     by BackProp. Usually this derivative is set via PresentTarget() or
     by a higher layer's BackProp() routine.
   */
  //void setErrorDerivativeWrtOutput(
  //				  Matrix& error_derivative_wrt_output);

  ///Sets the input to the given value.
  //void setInput(const Matrix& input);

  ///Sets the error derivative w.r.t. the input.
  //void setErrorDerivativeWrtInput(
  //				  Matrix& error_derivative_wrt_input);

  const Matrix& getWeights();

  ///Sets the weights to the given value.
  void setWeights(const Matrix& weights);

  ///Sets the biases to the given value
  void setBiases(const Matrix& biases);

  const Matrix & getBiases() const
    {
      return biases_;
    }


  ///Sets all biases to the same constant
  void setAllBiases(Scalar b);

  const Scalar targetSparsity();

  ///Sets the target sparsity to the given value. 
  /**
     The default value is 0.0, which means that the sparsity was not set up.
  */
  void setTargetSparsity(Scalar sparsity);


  ///Checks that weight matrices have appropriate size, etc
  /**
     Assumes link has been called, so it will fail before that
  */
  void checkInvariants();

  void serialize(serial::Archive & ar);

  bool equals(const Layer & other);
  /// Sets various parameters via an xml specification
  /**
     Supported attributes: learning_rate, learning_rate_sparsity, 
     sparsity_averaging_timeconstant, target_sparsity, weight_decay
   */
  void setParams(XMLNode& node);

  void printDebugStuff();


  virtual void scrambleWeights();


 public:
  Scalar learning_rate;
  Scalar learning_rate_sparsity;

  /// Time constant used in exponentially weighted avg of error
  /**
     Should be between 0 and 1.
  */
  Scalar sparsity_averaging_timeconstant;

 protected:
  
  //input and output are disregarded unless storageExists = true
  void initialize(boost::shared_ptr<StorageNode> input,
		  boost::shared_ptr<StorageNode> output,
		  boost::shared_ptr<ActivationFunction> activation_fn,
		  boost::shared_ptr<CostFunction> cost_fn,
                  bool storageExists = true);

  
  /// weights_(i,j) is synapse strength from input j to output i
  boost::shared_ptr<Matrix> weights_; 
  Matrix error_derivative_wrt_weights_;

  std::vector<boost::shared_ptr<UnaryObjectiveTerm> > weightPenalties_;


  /** 
      temporary turned into a class member so that continual
      allocation/deallocation doesn't hurt performance
  \todo-- is this why we use more memory than FastEncoder?
      doesn't seem sufficient to explain a 5x increase */
  Matrix derivative_;
  Matrix sparsity_avg_act_;

  //HL: addition of biases
  Matrix biases_; 
  Matrix error_derivative_wrt_biases_;
  Matrix error_derivative_wrt_sparsity_;

  Matrix  net_input_;
  Matrix  delta_;

  boost::shared_ptr<ActivationFunction> activation_fn_;
  boost::shared_ptr<CostFunction> cost_fn_;
  

  bool sparsity_initialized_;
  double sparsity_avg_;
   
  Scalar target_sparsity_;
  Scalar sparsity_deadzone_;
};

}

//SDN_AUTOREGISTER_H(Layer, StandardLayer);
