// Copyright 2009, All Rights Reserved.
// Authors: Andrew Saxe (asaxe@stanford.edu)
//          Ian Goodfellow (ia3n@cs.stanford.edu)

/** \file storageLayer.cc
    \brief  Defines a storage layer that holds inputs and outputs.
*/

#include "storageNode.h"
#include "../base/logger.h"
#include "../base/archive.h"
#include "../base/matrixUtils.h"
#include "activationFunction.h"

using namespace std;
using namespace sdn;
using namespace boost;
#include <assert.h>


#include <Eigen/Array>


StorageNode::StorageNode(int layer_size) {
  init(layer_size);
}

unsigned StorageNode::size() {
  return layer_size_;
}

int StorageNode::numExamples() {
  return activities_.cols();
}

void StorageNode::resize(int newSize)
{
  init(newSize);
}

#ifndef NOT_CONST
const 
#endif
Matrix& StorageNode::getActivities() {
  return activities_;
}

const Matrix& StorageNode::getDerivatives() {
  return derivatives_;
}

void StorageNode::setActivities(const Matrix & activities) {
  SDN_ASSERT_MSG(activities.rows() == size(), "Attempted to assign matrix with "<<activities.rows()<<" rows to StorageNode of size "<<size());
  activities_ = activities;
}

void StorageNode::computeActivities(shared_ptr<ActivationFunction> f,
const Matrix & x)
{
	f->output(x, activities_);
	SDN_ASSERT(activities_.rows() == size());
}

void StorageNode::setDerivatives(const Matrix & derivatives) {
  assert(derivatives.rows() == size());
  derivatives_ = derivatives;
}

void StorageNode::init(int layer_size)
{
  layer_size_ = layer_size;
  activities_ = Matrix::Zero(layer_size_,1);
  derivatives_ = Matrix::Zero(layer_size_,1);
}

/*
Registry<StorageNode> & sdn::StorageNodeRegistry()
{
  static Registry<StorageNode> registry("sdn::StorageNode");
  return registry;
}
*/


void StorageNode::serialize(serial::Archive & ar)
{
  SDN_DEBUG("-------StorageNode::serialize-----------------");
  if (ar.in())
    {
      SDN_DEBUG("(loading it)");
    }
  int version = 93;
  
  version = ar.beginBase(this,version,"sdn::StorageNode");
  
  switch(version)
    {
    case 93:
      ar.attribute("layerSize",layer_size_);
      serializeMatrixSize(ar,activities_);
      serializeMatrixSize(ar,derivatives_);
      ar.endBase();
    }
}

bool StorageNode::equals(const StorageNode & other)
{
  return layer_size_ == other.layer_size_;
}


