/// \author Ian Goodfellow

#include "translationTransformer.h"

using namespace sdn;


TranslationTransformer::TranslationTransformer()
{
	dirX_ = 1.0;
	dirY_ = 0.0;
	offset_ = 0.0;
	xSamples_ = 8;
	ySamples_ = 8;
	minOffset_ = -1.0;
	maxOffset_ = 1.0;
	rrg_.setSeed(42);
}

TranslationTransformer::~TranslationTransformer()
{
}

void TranslationTransformer::setDirection(double theta)
{
	dirX_ = cos(theta);
	dirY_ = -sin(theta);
	recalculateSize_();
}

void TranslationTransformer::setOffsetRange(double minOffset, double maxOffset)
{
	minOffset_ = minOffset;
	maxOffset_ = maxOffset;
	if (offset_ < minOffset_ || offset_ > maxOffset_)
		sampleOffset_();
	recalculateSize_();
}

void TranslationTransformer::next()
{
	sampleOffset_();
	recalculateWeights_();
}

void TranslationTransformer::setOffset(double offset)
{
  SDN_ASSERT(offset_ >= minOffset_);
  SDN_ASSERT(offset_ <= maxOffset_);
  offset_ = offset;
  recalculateWeights_();
}

void TranslationTransformer::sampleOffset_()
{
  offset_ = rrg_.randUnif(minOffset_,maxOffset_);
}

void TranslationTransformer::recalculateSize_()
{
  calculateRangeAndOrigin_(outputRows_, dirY_, yOrigin_, ySamples_, inputRows_);


  calculateRangeAndOrigin_(outputCols_, dirX_, xOrigin_, xSamples_, inputCols_);
  SDN_DEBUG("TranslationTransformer resized to "<<inputRows_<<" input rows and "<<inputCols_<<" input cols");

  resize_(inputRows_,inputCols_,outputRows_,outputCols_);
  recalculateWeights_();
}

void TranslationTransformer::calculateRangeAndOrigin_(unsigned outputSize, double stepSize, double & origin, unsigned samplingRate, unsigned & inputSize)
{
  double minVal = 0.5/double(samplingRate); //coordinate of first sample
  double maxVal = double(outputSize) - 0.5/double(samplingRate); //coordinate of last sample

  SDN_DEBUG("First sample: "<<minVal);
  SDN_DEBUG("Last sample: "<<maxVal);

  SDN_DEBUG("Step range: " << stepSize * maxOffset_ <<"," << stepSize*minOffset_);


  minVal += std::min(stepSize * maxOffset_, stepSize * minOffset_);
  maxVal += std::max(stepSize * maxOffset_, stepSize * minOffset_);

  SDN_DEBUG("minVal: "<<minVal);
  SDN_DEBUG("maxVal: "<<maxVal);

  int minCoord = floor(minVal);
  int maxCoord = floor(maxVal);
  origin = double(- minCoord);
  
  inputSize = maxCoord - minCoord + 1;
}

void TranslationTransformer::recalculateWeights_()
{
  weights_->setConstant(0.0);
  for (unsigned i = 0; i < outputRows_; i++)
    {
      for (unsigned j = 0; j < outputCols_; j++)
	{
	  recalculateWeight_(i,j);
	}
    }
}

void TranslationTransformer::recalculateWeight_(unsigned i, unsigned j)
{
  //SDN_DEBUG("Recalculating weight "<<i<<" " << j);
  double y = double(i) + 0.5 / double(ySamples_);
  for (unsigned k = 0; k < ySamples_; k++)
    {
      double x = double(j) + 0.5/double(xSamples_);

      for (unsigned l = 0; l < xSamples_; l++)
	{
	  //SDN_DEBUG("Sample x: "<<x);
	  
	  unsigned inputX = unsigned( floor( x + offset_ * dirX_ + xOrigin_));
	  unsigned inputY = unsigned( floor(y + offset_ * dirY_ + yOrigin_ ));

	  getWeight_(inputY,inputX,i,j) += 1.0/double(xSamples_ * ySamples_);
	  

	  x += 1.0/double(xSamples_); 
	}


      y += 1.0/double(ySamples_);
    }
}

void TranslationTransformer::setOutputSize(unsigned rows, unsigned cols)
{
  outputRows_ = rows;
  outputCols_ = cols;

  recalculateSize_();
}
