//Author: Daniel Chia <danchia@stanford.edu>
//File: crossValidationBlock.cc
//Description: implements cross validation on a dataset for a supervised network

#include "crossValidationBlock.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn{

	void CrossValidationBlock::loadXML(XMLNode& node) {
		AttributeParser p;

		p.addString("dataset");
		p.addInt("inputHeight");
		p.addInt("inputWidth");
		p.addBool("normalizeInput", false);

		p.parse(node);

		datasetName = p.getString("dataset");
		inputHeight = p.getInt("inputHeight");
		inputWidth = p.getInt("inputWidth");
		normalizeInput = p.getBool("normalizeInput");

		initPersistentParameters();
		initTransientParameters();
	}

	void CrossValidationBlock::initTransientParameters() {
		dataset = boost::dynamic_pointer_cast<BasicImageCollection> (DatasetRegistry().make(datasetName));

		if (!dataset) {
			SDN_ERROR("Could not load cross validation dataset with name " << datasetName);
			ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
			SDN_FATAL("Quitting");
		}

		dataset->requestExampleSize(inputHeight, inputWidth);

		inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
		prediction = sdn::Matrix::Zero(dataset->numLabels(), 1);

		numCorrect = 0;
	}

	void CrossValidationBlock::evaluateNetwork(boost::shared_ptr<AbstractSupervisedNetwork> network) {
		
		numCorrect = 0;

		Scalar max;
		unsigned int label, actualLabel;

		for (unsigned int index = 0; index < dataset->numExamples(); index++) {
			actualLabel = dataset->getExampleAsColumn(inputExample, index);

			if (normalizeInput)
				normInput.normalizePatch(inputExample);
			
			network->forwardProp(inputExample, prediction);

			max = prediction[0];
			label = 0;

			for (unsigned int j = 1; j < dataset->numLabels(); j++)
				if (prediction[j] > max)
				{
					max = prediction[j];
					label = j;
				}

			if (label == actualLabel)
				numCorrect++;
		}
	}

	double CrossValidationBlock::getError() {
		return (double)(dataset->numExamples()-numCorrect)/dataset->numExamples();
	}

	void CrossValidationBlock::initPersistentParameters() {
	}

	void CrossValidationBlock::serialize(serial::Archive &ar) {
		SDN_FATAL("CrossValidationBlock::serialize not implemented yet");
	}

	CrossValidationBlock::CrossValidationBlock() {}
	CrossValidationBlock::~CrossValidationBlock() {}

}
