//Author: Daniel Chia <danchia@stanford.edu>
//File: danchiaCrossValidationScript.cc
//Description: 

#include "danchiaCrossValidationScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"


namespace sdn {

  class DanchiaCrossValidationScriptConfigParser : public ConfigParser {

  protected:
    DanchiaCrossValidationScript & parentScript;

  public:
  
    DanchiaCrossValidationScriptConfigParser(DanchiaCrossValidationScript & parent) : parentScript(parent) {}
    ~DanchiaCrossValidationScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::DanchiaCrossValidationScript");
    }

    void readConfig(XMLNode & node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::DanchiaCrossValidationScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  DanchiaCrossValidationScript::DanchiaCrossValidationScript() {
    loadUsingConfigManager();
  }
  
  DanchiaCrossValidationScript::~DanchiaCrossValidationScript() {
  }
  
  void DanchiaCrossValidationScript::loadXML(XMLNode& node) {
    AttributeParser p;
    
    p.addString("supervisedNetworkFile");
		p.addInt("startIteration");
		p.addInt("iterationsPerStep");
    
    p.parse(node);
    
    // Set Values
    supervisedNetworkFile = p.getString("supervisedNetworkFile");
		startIteration = p.getInt("startIteration");
		iterationsPerStep = p.getInt("iterationsPerStep");
  
    // ---------------------------------------------------
    
    

    // Deep Loading
    XMLNode crossValidationBlockNode = node.getChildNode("CrossValidationBlock");
    SDN_ASSERT_MSG(!crossValidationBlockNode.isEmpty(), "You need to specify <CrossValidationBlock> !")
    crossValidationBlock.loadXML(crossValidationBlockNode); 
    
  }

  void DanchiaCrossValidationScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new DanchiaCrossValidationScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();
  }
  
  void DanchiaCrossValidationScript::initTransientParameters() {
  }
  
  void DanchiaCrossValidationScript::initPersistentParameters() {
   }
  
  void DanchiaCrossValidationScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void DanchiaCrossValidationScript::executeOneIteration() {
    
    SDN_ASSERT(manager);

		std::string iterationString = itos(startIteration + manager->currentIteration*iterationsPerStep, 10);

		std::string currentFile = supervisedNetworkFile;

		str::replaceAll(currentFile, "${ITERATION}", iterationString);

 		// Load Supervised network


		std::ifstream fin;

		fin.open(currentFile.c_str(), std::ios::binary);
		SDN_ASSERT_MSG(fin.good(), "Unable to open " << currentFile);

		sdn::serial::BinInArchive ar(fin);
		boost::shared_ptr<AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
		fin.close();

		SDN_ASSERT_MSG(net, "Unable to load network from file");

		supervisedNetwork = boost::dynamic_pointer_cast<AbstractSupervisedNetwork> (net);

		SDN_ASSERT_MSG(net, net->className() << "is not a supervised network");
		SDN_VERBOSE("Loaded " << currentFile << " as supervised network");


		// Evaluate network

    crossValidationBlock.evaluateNetwork(supervisedNetwork);
		std::cout << "[" << iterationString << "]" << " Cross validation result: " << crossValidationBlock.getError() << std::endl;
  }
  

}
