//Author: Daniel Chia
//File: danchiaScript.cc
//Description: 

#include "danchiaScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn {

  class DanchiaScriptConfigParser : public ConfigParser {

  protected:
    DanchiaScript & parentScript;

  public:
  
    DanchiaScriptConfigParser(DanchiaScript & parent) : parentScript(parent) {}
    ~DanchiaScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::DanchiaScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::DanchiaScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  DanchiaScript::DanchiaScript() {
    loadUsingConfigManager();
  }
  
  DanchiaScript::~DanchiaScript() {
  }
  
  void DanchiaScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("network");
		p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    
    p.parse(node);
    
    // Note that values are set *directly* into the network
		networkName = p.getString("network");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    
    // We need to create the network here because the XML parameters
    // for it need to be read directly to the network    
    network = boost::dynamic_pointer_cast<StandardAutoEncoderNetwork> (AbstractNetworkRegistry().make(networkName));
    if (!network) {
      SDN_ERROR("Could not load network with name " << networkName);
      SDN_ERROR("Is " << networkName << " a StandardAutoEncoderNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network->loadXML(networkNode); 

  }

  void DanchiaScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new DanchiaScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();

  }
  
  void DanchiaScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
  }
  
  void DanchiaScript::initPersistentParameters() {
    network->initTransientParameters();
    network->initPersistentParameters();
  }
  
  void DanchiaScript::serialize(serial::Archive & ar) {
    SDN_FATAL("DanchiaScript is not serializable yet");
  }

  
  void DanchiaScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // do fancy stuff here
    normInput.normalizePatch(inputExample);
    network->unsupervisedTrain(inputExample);
    
  }
  
  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> DanchiaScript::getReconstructionCostFunction() {
    return network->getCostFunction();
  }

  const sdn::Matrix& DanchiaScript::getReconstruction() {
    return network->getReconstruction();
  }
  
  const sdn::Matrix& DanchiaScript::getOriginal() {
    return inputExample;
  }
 
	// ---- Report Progress Callback ----
 
	void DanchiaScript::reportProgress(std::ostream& o) {
		o << "Iteration: " << itos(manager->currentIteration, 10) << "\n";
	}	

	// ---- Visualization Callback ----
	const sdn::Matrix& DanchiaScript::getEncoderWeights() {
		return network->affineEncoder.weights;
	}

	const int DanchiaScript::getWeightPatchHeight() {
		return inputHeight;
	}

	const sdn::Matrix& DanchiaScript::getDecoderWeights() {
		return network->affineDecoder.weights;
	}

	// ---- Save Model Callback ----
	const void DanchiaScript::saveModelInDirectory(std::string dir) {
		network->quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");

		//backup saving...
		std::ofstream fout;

		fout.open(quickString("%s%010u-ew.amat", dir.c_str(), manager->currentIteration).c_str());
		fout << network->affineEncoder.weights;
		fout.close();

    fout.open( quickString("%s%010u-eb.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineEncoder.bias;
    fout.close();

    fout.open( quickString("%s%010u-dw.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineDecoder.weights;
    fout.close();

    fout.open( quickString("%s%010u-db.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineDecoder.bias;
    fout.close();
	}
}
