//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: testAffineBlock.cc
//Description: 

#include "testAffineBlock.h"

namespace sdn {
  
  // TestAffineBlock ------------------------------------------------
  
  TestAffineBlock::TestAffineBlock() {
  }
  
  TestAffineBlock::~TestAffineBlock() {
  }

  void TestAffineBlock::setConfigurableParameters(sdn::Scalar defLearningRate,
                                              sdn::Scalar defWeightDecay,
                                              int defInputSize,
                                              int defOutputSize) {
    learningRate = defLearningRate;
    weightDecay = defWeightDecay;
    outputSize = defOutputSize;
    inputSize = defInputSize;
    
    initPersistentParameters();
    initTransientParameters();
  }

  void TestAffineBlock::loadXML(XMLNode& node) {

    AttributeParser p;

    p.addDouble("learningRate");
    p.addDouble("weightDecay", NO_WEIGHT_DECAY);
    p.addDouble("fixedWeightNorm");
    p.addInt("outputSize", SIZE_AUTOSET);
    p.addInt("inputSize", SIZE_AUTOSET);
    p.addBool("lockBias", false);
    p.addBool("shareBias", false);
    p.addDouble("initialBias", 0);
    //p.addBool("unityWeights", false);
        
    p.parse(node);

    weightDecay = p.getDouble("weightDecay");
    learningRate =p.getDouble("learningRate");
    outputSize = p.getInt("outputSize");
    inputSize = p.getInt("inputSize");
    lockBias = p.getBool("lockBias");
    shareBias = p.getBool("shareBias");
    //unityWeights = p.getBool("unityWeights");
    fixedWeightNorm = p.getDouble("fixedWeightNorm");
    
    initPersistentParameters();
    initTransientParameters();

    double initialBias;
    initialBias = p.getDouble("initialBias");
    bias = sdn::Matrix::Constant(outputSize, 1, initialBias);
    

  }
  
  void TestAffineBlock::initTransientParameters() {
    if (outputSize > 0 && inputSize > 0) {
      weightsGradient = sdn::Matrix::Zero(outputSize, inputSize);
      biasGradient = sdn::Matrix::Zero(outputSize, 1);
    }
  }
  
  void TestAffineBlock::initPersistentParameters() {
    if (outputSize > 0 && inputSize > 0) {
      weights = sdn::Matrix::Zero(outputSize, inputSize);
      bias = sdn::Matrix::Zero(outputSize, 1);

      RandomizeWeightMatrix().randomize(weights, inputSize);
    }
  }
  
  void TestAffineBlock::serialize(serial::Archive & ar) {
    
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1010;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
        
      case 1010:
        ar.attribute("fixedWeightNorm", fixedWeightNorm);
                
      case 1000:
      
        // Persistent
        serializeMatrix(ar, weights);
        serializeMatrix(ar, bias);
        
        // Configurable
        ar.attribute("learningRate", learningRate);
        ar.attribute("weightDecay", weightDecay);
        ar.attribute("outputSize", outputSize);
        ar.attribute("inputSize", inputSize);

        ar.attribute("lockBias", lockBias);
        ar.attribute("shareBias", shareBias);
        
        // backwards compatible
        if (ar.in() && version == 1000) {
          bool uw;
          ar.attribute("unityWeights", uw);
          if (uw) {
            fixedWeightNorm = 1.0;
          }
        }

        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }
  
  void TestAffineBlock::forwardProp(const sdn::Matrix& input,
                                sdn::Matrix& output) {

    if (outputSize <= 0 || inputSize <= 0)  {
      if (inputSize <= 0)
        inputSize = input.rows();
      if (outputSize <= 0)
        outputSize = output.rows();
      initPersistentParameters();
      initTransientParameters();
    }

    if (weights.cols() != input.rows()) {
      SDN_FATAL(className() << " weight matrix #columns does not match input #rows");
    }
    
    
    // Multiple examples, hence, need to use ones * bias
    if (ones.cols() != input.cols())
      ones = Matrix::Ones(1, input.cols());
    
    output = weights * input + bias * ones;
  }
 
  void TestAffineBlock::backProp(const sdn::Matrix& input,
                             const sdn::Matrix& outputDerivatives,
                             sdn::Matrix& inputDerivatives) {
    
    if (outputSize <= 0 || inputSize <= 0)  {
      inputSize = input.size();
      outputSize = outputDerivatives.size();
      initTransientParameters();
      initPersistentParameters();
    } 
    
    inputDerivatives = weights.transpose() * outputDerivatives;
    weightsGradient = outputDerivatives * input.transpose();
    biasGradient    = outputDerivatives;
    
    if (outputDerivatives.cols() > 1) {
      // Handle Multiple Examples. Normalize the Gradients
      weightsGradient *= 1.0 / outputDerivatives.cols();
      biasGradient = biasGradient.rowwise().sum() / biasGradient.cols();
    }
    
  }
  
  void TestAffineBlock::gradientUpdate() {

    // If weightDecay enabled
    if (weightDecay > 0 && (fixedWeightNorm==0)) 
      weights *= (1.0 - learningRate * weightDecay);  
      
        
    // Assume Minimizing the Objective
    weights -= learningRate * weightsGradient;    
    
    if (fixedWeightNorm > 0) {
      // Normalize weights to some fixed number
      weightNorms = (weights.cwise() * weights).rowwise().sum();
      for(int i=0;i<weights.rows();i++)
        weights.row(i) *= fixedWeightNorm / sqrt(weightNorms[i]);
    }
    
    if (!lockBias) {
      if (shareBias) {
        bias.cwise() -= learningRate * (biasGradient.sum() / biasGradient.size());
      } else {
        bias -= learningRate * biasGradient;
      }
    }
    
  }

  void TestAffineBlock::forceGradientUpdateBias(sdn::Scalar lr) {
    bias.cwise() -= lr * (biasGradient.sum() / biasGradient.size());
  }

  
}
