//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: autoTuneNetwork.cc
//Description: 

#include "autoTuneNetwork.h"

#define XML_LOAD(paramName) \
  XMLNode paramName##Node = node.getChildNode(SDN_MAKE_STRING(paramName)) ; \
  if (!paramName##Node.isEmpty()) \
    paramName.loadXML(paramName##Node); \
  else \
    SDN_FATAL("Need to specify " << SDN_MAKE_STRING(paramName) << " in config!");

const int AFFINE_ENCODE = 0;
const int SCALE_ENCODE = 1;
const int TANH_ENCODE = 2;

const int AFFINE_DECODE = 3;
const int SCALE_DECODE = 4;
const int TANH_DECODE = 5;

const int AFFINE_CLASSIFY = 6;
const int SOFTMAX_CLASSIFY = 7;

const int TOTAL_BLOCKS = 8;
    
namespace sdn {

  void AutoTuneNetwork::loadXML(XMLNode& node) {

    // Setting the End Encode Block makes it an AutoEncoder
    AttributeParser p;
    
    p.addString("reconstructionCostFunctionName", "sdn::SumOfSquaresCostFunction");
    p.addString("supervisedCostFunctionName", "sdn::SumOfSquaresCostFunction");
    p.addDouble("hyperParameterLearningRate");
    p.addDouble("hyperEncScaleLearningRate",0.0);

    p.addDouble("targetActivation", 0.0);
    p.addDouble("learningRateSparsity", 0.0);
    
    p.parse(node);
    
    reconstructionCostFunctionName = p.getString("reconstructionCostFunctionName");
    supervisedCostFunctionName = p.getString("supervisedCostFunctionName");
    hyperParameterLearningRate =p.getDouble("hyperParameterLearningRate");
    hyperEncScaleLearningRate = p.getDouble("hyperEncScaleLearningRate");

    targetActivation = p.getDouble("targetActivation");
    learningRateSparsity = p.getDouble("learningRateSparsity");
        
    // Load up internal structures
    XML_LOAD(encodeAffine);
    XML_LOAD(encodeScale);
    XML_LOAD(encodeTanh);

    XML_LOAD(decodeAffine);
    XML_LOAD(decodeScale);
    XML_LOAD(decodeTanh);
    
    XML_LOAD(classifyAffine);
    XML_LOAD(classifySoftmax);
        
    initPersistentParameters();
    initTransientParameters();
    
  }
  
  void AutoTuneNetwork::initTransientParameters() {
    
    // Initialize the CostFunction
    supervisedErrorCostFunction = CostFunctionRegistry().make(supervisedCostFunctionName);
    if (!supervisedErrorCostFunction) {
      SDN_ERROR("Could not load CostFunction with name " << supervisedCostFunctionName);
      ERROR_OUT_LIST_REGISTRY_NAMES(CostFunction);
      SDN_FATAL("Quitting");
    }
    
    reconstructionErrorCostFunction = CostFunctionRegistry().make(reconstructionCostFunctionName);
    if (!reconstructionErrorCostFunction) {
      SDN_ERROR("Could not load CostFunction with name " << reconstructionErrorCostFunction);
      ERROR_OUT_LIST_REGISTRY_NAMES(CostFunction);
      SDN_FATAL("Quitting");
    }

    
    // Setup the intermediate variables
    intermediateRepresentations.resize(TOTAL_BLOCKS);        // NOTE: Fixed Value Here
    intermediateDerivatives.resize(TOTAL_BLOCKS);            // NOTE: Fixed Value Here


    activationTracker.setConfigurableParameters(0.001,
                                                TRACKER_AUTOSET,
                                                SIZE_AUTOSET);
    reconstructionErrorTracker.setConfigurableParameters(0.001,
                                                TRACKER_AUTOSET,
                                                SIZE_AUTOSET);
    supervisedTrainingErrorTracker.setConfigurableParameters(0.001,
                                                TRACKER_AUTOSET,
                                                SIZE_AUTOSET);
    supervisedValidationErrorTracker.setConfigurableParameters(0.001,
                                                TRACKER_AUTOSET,
                                                SIZE_AUTOSET);

  }
  
  void AutoTuneNetwork::initPersistentParameters() {
    
  }
  
  void AutoTuneNetwork::serialize(serial::Archive & ar) {
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1020;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
      case 1020: 
        ar.attribute("targetActivation", targetActivation);
        ar.attribute("learningRateSparsity", learningRateSparsity); 

      case 1010:
        ar.attribute("hyperParameterLearningRate",hyperParameterLearningRate);
        ar.attribute("hyperEncScaleLearningRate",hyperEncScaleLearningRate);
        
      case 1000:
      
        // Persistent-Configurable
        encodeAffine.serialize(ar);
        encodeScale.serialize(ar);
        encodeTanh.serialize(ar);
        
        // Decoding Portion
        decodeAffine.serialize(ar);
        decodeScale.serialize(ar);
        decodeTanh.serialize(ar);
        
        // Supervised Stuff
        classifyAffine.serialize(ar);
        classifySoftmax.serialize(ar);
        
        ar.attribute("reconstructionCostFunctionName",reconstructionCostFunctionName);
        ar.attribute("supervisedCostFunctionName",supervisedCostFunctionName);
        
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }

  }

  void AutoTuneNetwork::supervisedTrain(const sdn::Matrix& example, 
                                        const sdn::Matrix& label)  {

      SDN_FATAL("Do not use " << __FUNCTION__<< " with " << className() );
  }
  
  void AutoTuneNetwork::gradientUpdate() {
      SDN_FATAL("Do not use " << __FUNCTION__<< " with " << className() );
  }

  const boost::shared_ptr<CostFunction> AutoTuneNetwork::getCostFunction() {
    return supervisedErrorCostFunction;
  }

  
  sdn::Matrix& AutoTuneNetwork::getClassificationOutputGradient() {
    return intermediateDerivatives[SOFTMAX_CLASSIFY];
  }
  
  sdn::Matrix& AutoTuneNetwork::getReconstructionOutputGradient() {
    return intermediateDerivatives[TANH_DECODE];

  }

  const sdn::Matrix& AutoTuneNetwork::getPrediction() {
    return intermediateRepresentations[SOFTMAX_CLASSIFY];
  }

  /*
    const int AFFINE_ENCODE = 0;
    const int SCALE_ENCODE = 1;
    const int TANH_ENCODE = 2;

    const int AFFINE_DECODE = 3;
    const int SCALE_DECODE = 4;
    const int TANH_DECODE = 5;

    const int AFFINE_CLASSIFY = 6;
    const int SOFTMAX_CLASSIFY = 7;

    const int TOTAL_BLOCKS = 8;
  */
  
  void AutoTuneNetwork::forwardPropClassifyHidden() {
    classifyAffine.forwardProp( intermediateRepresentations[TANH_ENCODE],
                                intermediateRepresentations[AFFINE_CLASSIFY] );
    classifySoftmax.forwardProp(intermediateRepresentations[AFFINE_CLASSIFY],
                                intermediateRepresentations[SOFTMAX_CLASSIFY] );
  }
  
  void AutoTuneNetwork::backPropClassifyHidden() {
    classifySoftmax.backProp(intermediateRepresentations[AFFINE_CLASSIFY],
                             getClassificationOutputGradient(),
                             intermediateDerivatives[AFFINE_CLASSIFY]);
    classifyAffine.backProp(intermediateRepresentations[TANH_ENCODE],
                            intermediateDerivatives[AFFINE_CLASSIFY],
                            intermediateDerivatives[TANH_ENCODE]);
  }
  
  void AutoTuneNetwork::supervisedGradientUpdate() {
    classifySoftmax.gradientUpdate();
    classifyAffine.gradientUpdate();
  }

  void AutoTuneNetwork::trainTrainingExample(const sdn::Matrix& example,
                                             const sdn::Matrix& label) {
    
    // Encode and classify
    encode(example);
    decode(getHiddenRepresentation());
    forwardPropClassifyHidden();
    
    // ---- Update Unsupervised Portion of Network ----
    intermediateDerivatives[TANH_DECODE] = reconstructionErrorCostFunction->jacobian(getReconstruction(),
                                                                                     example).transpose();
    thisError = reconstructionErrorCostFunction->output(getReconstruction(), example);
    reconstructionErrorTracker.addValues(thisError / example.cols());
    
    decodeBackProp(getHiddenRepresentation(),
                   intermediateDerivatives[TANH_DECODE],
                   intermediateDerivatives[TANH_ENCODE]);

    encodeBackProp(example,
                   intermediateDerivatives[TANH_ENCODE],
                   inputDerivatives);

    // Add to Activation Tracker
    activationTracker.addValues( encodeTanh.normalizeActivationValues( intermediateRepresentations[TANH_ENCODE]  ) );

    if (learningRateSparsity > 0) {
      sparsityError = activationTracker.getMean().cwise() - targetActivation;
      encodeAffine.bias -= encodeAffine.learningRate*learningRateSparsity*sparsityError;
    }


    // Gradient Update
    encodeGradientUpdate();
    decodeGradientUpdate();
    
    // ---- Update Supervised Portion of Network ----
    intermediateDerivatives[SOFTMAX_CLASSIFY] = supervisedErrorCostFunction->jacobian(getPrediction(), label).transpose();
    thisError = supervisedErrorCostFunction->output(getPrediction(), label);
    supervisedTrainingErrorTracker.addValues(thisError / example.cols());
    backPropClassifyHidden();

    // Gradient Update
    classifyAffine.gradientUpdate();
    classifySoftmax.gradientUpdate();
    
  }

  void AutoTuneNetwork::trainValidationExample(const sdn::Matrix& example,
                                               const sdn::Matrix& label) {
    
    // Encode and classify
    encode(example);
    forwardPropClassifyHidden();
    
    // Compute Error and Gradients
    sdn::Matrix& outputGradient = getClassificationOutputGradient();
    outputGradient = supervisedErrorCostFunction->jacobian(getPrediction(), label).transpose();
    thisError = supervisedErrorCostFunction->output(getPrediction(), label);
    supervisedValidationErrorTracker.addValues(thisError / example.cols());
    
    // Back Prop
    backPropClassifyHidden();
    encodeBackProp(example, intermediateDerivatives[TANH_ENCODE], inputDerivatives);

    // Update Hyper Parameters
    encodeAffine.forceGradientUpdateBias(hyperParameterLearningRate);
    targetActivation -= hyperParameterLearningRate * (encodeAffine.biasGradient.sum() / encodeAffine.biasGradient.size());
    
    // Update Scale Parameter for Encoder
    encodeScale.scale -= hyperEncScaleLearningRate * encodeScale.scaleGradient;
  }

  
  void AutoTuneNetwork::unsupervisedTrain(const sdn::Matrix& example)  {
    SDN_FATAL("Do not use " << __FUNCTION__<< " with " << className() );
  }

  const sdn::Matrix& AutoTuneNetwork::encode(const sdn::Matrix& input) {
    encodeAffine.forwardProp( input, 
                              intermediateRepresentations[AFFINE_ENCODE] );
    encodeScale.forwardProp ( intermediateRepresentations[AFFINE_ENCODE], 
                              intermediateRepresentations[SCALE_ENCODE] );
    encodeTanh.forwardProp  ( intermediateRepresentations[SCALE_ENCODE], 
                              intermediateRepresentations[TANH_ENCODE] );
                              
    return intermediateRepresentations[TANH_ENCODE] ;
  }
  
  const sdn::Matrix& AutoTuneNetwork::decode(const sdn::Matrix& hidden){
    decodeAffine.forwardProp( hidden, 
                              intermediateRepresentations[AFFINE_DECODE] );
    decodeScale.forwardProp ( intermediateRepresentations[AFFINE_DECODE], 
                              intermediateRepresentations[SCALE_DECODE] );
    decodeTanh.forwardProp  ( intermediateRepresentations[SCALE_DECODE], 
                              intermediateRepresentations[TANH_DECODE] );
                              
    return intermediateRepresentations[TANH_DECODE];
  }

  void AutoTuneNetwork::encodeBackProp(const sdn::Matrix& input,
                        const sdn::Matrix& hiddenErrorDerivatives,
                        sdn::Matrix& inputErrorDerivatives) {
                          
    encodeTanh.backProp(intermediateRepresentations[SCALE_ENCODE], 
                        hiddenErrorDerivatives,
                        intermediateDerivatives[SCALE_ENCODE]);
                        
    encodeScale.backProp(intermediateRepresentations[AFFINE_ENCODE], 
                         intermediateDerivatives[SCALE_ENCODE],
                         intermediateDerivatives[AFFINE_ENCODE]);

    encodeAffine.backProp(input, 
                          intermediateDerivatives[AFFINE_ENCODE],
                          inputErrorDerivatives);
                         
  }

  void AutoTuneNetwork::decodeBackProp(const sdn::Matrix& hidden,
                        const sdn::Matrix& outputErrorDerivatives,
                        sdn::Matrix& hiddenErrorDerivatives)  {
                          
    decodeTanh.backProp(intermediateRepresentations[SCALE_DECODE], 
                        outputErrorDerivatives,
                        intermediateDerivatives[SCALE_DECODE]);
                        
    decodeScale.backProp(intermediateRepresentations[AFFINE_DECODE], 
                         intermediateDerivatives[SCALE_DECODE],
                         intermediateDerivatives[AFFINE_DECODE]);

    decodeAffine.backProp(hidden, 
                          intermediateDerivatives[AFFINE_DECODE],
                          hiddenErrorDerivatives);
  }
                        
  void AutoTuneNetwork::encodeGradientUpdate() {
    encodeTanh.gradientUpdate();
    encodeScale.gradientUpdate();
    encodeAffine.gradientUpdate();
  }
  
  void AutoTuneNetwork::decodeGradientUpdate() {
    decodeTanh.gradientUpdate();
    decodeScale.gradientUpdate();
    decodeAffine.gradientUpdate();
  }

  const sdn::Matrix& AutoTuneNetwork::getReconstruction() {
    return intermediateRepresentations[TANH_DECODE];
  }
  
  const sdn::Matrix& AutoTuneNetwork::getHiddenRepresentation() {
    return intermediateRepresentations[TANH_ENCODE];
  }
    
  
}
