//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: autoTuneNetwork.h
//Description: Uses ComposableBlocks to create an AutoEncoder

#pragma once

#include "sdn/base.h"
#include "sdn/stable.h"
#include "sdn/registry.h"
#include "../blocks/testAffineBlock.h"

namespace sdn {
  
  class AutoTuneNetwork : public AbstractAutoEncoderNetwork{
  
  public:

    virtual void loadXML(XMLNode& node) ;
    
    virtual void initTransientParameters() ;
    virtual void initPersistentParameters() ;
    
    virtual void serialize(serial::Archive & ar) ;


    virtual void supervisedTrain(const sdn::Matrix& example, 
                                 const sdn::Matrix& label) ;

    virtual const boost::shared_ptr<CostFunction> getCostFunction();
    
    
    // ---- GENERAL ----
    virtual void forwardPropClassifyHidden();
    virtual void backPropClassifyHidden();
    
    virtual void trainTrainingExample(const sdn::Matrix& example,
                                      const sdn::Matrix& label);

    virtual void trainValidationExample(const sdn::Matrix& example,
                                        const sdn::Matrix& label);

    // ---- SUPERVISED STUFF ----
    virtual const sdn::Matrix& getPrediction();
    virtual void supervisedGradientUpdate();

    // ---- UNSUPERVISED STUFF ----
    
    virtual void unsupervisedTrain(const sdn::Matrix& example) ;
  
    virtual const sdn::Matrix& encode(const sdn::Matrix& input) ;
    virtual const sdn::Matrix& decode(const sdn::Matrix& hidden) ;

    virtual void encodeBackProp(const sdn::Matrix& input,
                                const sdn::Matrix& hiddenErrorDerivatives,
                                sdn::Matrix& inputErrorDerivatives) ;

    virtual void decodeBackProp(const sdn::Matrix& hidden,
                                const sdn::Matrix& outputErrorDerivatives,
                                sdn::Matrix& hiddenErrorDerivatives) ;
                                
    virtual void encodeGradientUpdate() ;
    virtual void decodeGradientUpdate() ;

    virtual const sdn::Matrix& getReconstruction() ;
    virtual const sdn::Matrix& getHiddenRepresentation() ;
    
    virtual void gradientUpdate();

    // ---- HELPERS ----
    void setupGoodInitialParameters();
    sdn::Matrix& getClassificationOutputGradient();
    sdn::Matrix& getReconstructionOutputGradient();

    // Persistent-Transient Parameters
    
    // Unsupervised Stuff
    sdn::Scalar  targetActivation;
    sdn::Scalar  learningRateSparsity; 
    
    // Encoding Portion
    TestAffineBlock encodeAffine;
    ScaleBlock      encodeScale;
    TanhBlock       encodeTanh;
    
    // Decoding Portion
    TestAffineBlock decodeAffine;
    ScaleBlock      decodeScale;
    TanhBlock       decodeTanh;
    
    // Supervised Stuff
    TestAffineBlock classifyAffine;
    SoftmaxBlock    classifySoftmax;
    
    // Configurable Parameters
    std::string reconstructionCostFunctionName, supervisedCostFunctionName;
    sdn::Scalar hyperParameterLearningRate, hyperEncScaleLearningRate;
    
    // Transient Parameters
    boost::shared_ptr<CostFunction> supervisedErrorCostFunction, reconstructionErrorCostFunction;
    ExponentiallyWeightedMeanBlock activationTracker;
    ExponentiallyWeightedMeanBlock reconstructionErrorTracker, supervisedValidationErrorTracker, supervisedTrainingErrorTracker;
    
    sdn::Matrix sparsityError;
    sdn::Matrix thisError;
    
    std::vector<sdn::Matrix> intermediateRepresentations;// Representations for the values *after* the block (so [blocks.size()-1] = reconstruction)
    std::vector<sdn::Matrix> intermediateDerivatives;    // Gradients for the values *after* the block (so [blocks.size()-1] = outputGradient)
    
    sdn::Matrix inputDerivatives; // Gradient at the inputs (how to change the inputs to decrease error)
    
    sdn::Matrix fineTuneInputDerivatives;
    

  };
  
}
