//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: JngiamAutoEncoderNetwork.cc
//Description: Uses AutoEncoderBlocks to create a Network

#include "jngiamAutoEncoderNetwork.h"
#include <Eigen/Array>

namespace sdn {
  
  void JngiamAutoEncoderNetwork::unsupervisedTrain(const sdn::Matrix& example) {
   
    if (reconstruction.size() != example.size()) {
      reconstruction = sdn::Matrix::Zero(example.rows(), example.cols());
      affineRepDec = sdn::Matrix::Zero(example.rows(), example.cols());
    }

    // Forward Propagate
    encode(example);
    decode(hiddenRepresentation);

    // Present Example
    outputGradient = costFunc->jacobian(reconstruction, example).transpose();
    sdn::Matrix thisError = costFunc->output(reconstruction, example);
    errorTracker.addValues(thisError);
       
    // Back Prop
    decodeBackProp( hiddenRepresentation, outputGradient, hiddenGradient);
    encodeBackProp( example, hiddenGradient, inputGradient);

    // Obtain Pre Bias Values   
    if (ones2.cols() != affineRepEnc.cols())
      ones2 = Matrix::Ones(1, affineRepEnc.cols());
    preBias = affineRepEnc - affineEncoder.bias * ones2;

    // Gradient Updates (weight decay is performed in the affineEncoder itself)
    gradientUpdate();

    // ============== ADDITIONAL OPERATIONS ==============

    // Sparsity Updates
    activationTracker.addValues( sigmEncoder.normalizeActivationValues( hiddenRepresentation ) );
    if (learningRateSparsity > 0) {
      sdn::Matrix sparsityError = activationTracker.getMean().cwise() - targetActivation;
      affineEncoder.bias -= affineEncoder.learningRate*learningRateSparsity*sparsityError;
    }

    // ============== BIAS SHIFT COMPUTATIONS ==============
    if (errorTracker.getMean().sum() < acceptableReconstructionError &&
        thisError.sum() < acceptableReconstructionError &&
        biasShift > 0) {
      affineEncoder.bias.cwise() -= biasShift;
    }

    // ============== KURTOSIS COMPUTATIONS ==============
    
    
    // Set up the ones matrix ( input dimension )
    sdn::Matrix& m = affineRepEnc;  // easier to reference
    
    /*
    if (ones.rows() != m.rows())
      ones = Matrix::Ones(m.rows(), 1);
    
    // Mean, multiply by ones to get the right size
    mean =  ones * (1.0/double(m.rows()) * m.colwise().sum());

    // 2nd Central Moment
    vcm2 = (m - mean).cwise().square();
    cm2 = (1.0/double(m.rows())) * vcm2.colwise().sum();
    
    // 4th Central Moment
    vcm4 = vcm2.cwise().square();
    cm4 = (1.0/double(m.rows())) * vcm4.colwise().sum();

    // UnNormalized Kurtosis (Hyvarinen)
    kurt = cm4 - 3.0 * (cm2.cwise() * cm2);
    unnormalizedKurtTracker.addValues( kurt );
    
    // Normalized Kurtosis
    kurt = (cm4.cwise() / (cm2.cwise() * cm2)).cwise() - 3.0;
    normalizedKurtTracker.addValues( kurt );
    */
    
    // ============== KURTOSIS GRADIENT UPDATE ==============
    
    
    if (learningRateKurtosis > 0) {
     
      if (optimizeUnitKurtosis) {
        
        // Optimize the Standard Kurtosis Measure
        // (Assume unity variance, zero mean, optimize 4th order for each feature/hidden unit)
        
        // Weghts Derivative
        weightsDerv = 4 * ((preBias.cwise().square().cwise() * preBias) * example.transpose());
        
        // ------------ START Derivative Checking ---------
        // Note that this check can fail if the weight matrix is changed before it reaches here
        /*        
        Matrix deltaWeights = affineEncoder.weights;
        RandomizeWeightMatrix().randomize( deltaWeights, 100 );
        deltaWeights *= 1e-10/sqrt(deltaWeights.cwise().square().sum());
        Matrix newPreBias = (affineEncoder.weights+deltaWeights)*example;
        Matrix newDiff = (newPreBias.cwise().square().cwise().square()) - 
                         (preBias.cwise().square().cwise().square()) ;
        std::cout << "analytic change: " << (weightsDerv.cwise() * deltaWeights).sum() << "\n";
        std::cout << "finite diff change: " << newDiff.sum() << "\n";
        */
        // ------------ END Derivative Checking ---------
         
        
        // Apply the gradient
        // + because we are maximizing
        affineEncoder.weights += affineEncoder.learningRate * learningRateKurtosis * weightsDerv;
        
      } else {
    
        // We do the gradient for the normalized kurtosis
        // 1st Central Moment
        vcm1 = (m - mean);
        cm1 = (1.0/double(m.rows())) * vcm1.colwise().sum();
        
        // 3rd Central Moment
        vcm3 = vcm2.cwise() * vcm1;
        cm3 = (1.0/double(m.rows())) * vcm3.colwise().sum();
        
        // First part of the gradient
        dervK1 = ( 4.0/(double(m.rows())) * ( 
                  -1.0/(double(m.rows()))*(ones*cm3) + 
                  vcm3 ) ).cwise() / (ones * (cm2.cwise().square()));
        
        // Second part of the gradient
        dervK2 = (( 2.0/(double(m.rows())) * ( 
                   -1.0/(double(m.rows()))*(ones*cm1) + 
                  vcm1 ) ).cwise() / (ones * (cm2.cwise().square().cwise() * cm2))).cwise() * (2.0 * ones * cm4);
        
        // Total derv
        dervK = dervK1 - dervK2;
        
        
        // ------------ START Derivative Checking ---------
        /*
        Matrix deltaAffineRep = m;
        RandomizeWeightMatrix().randomize( deltaAffineRep, 100 );
        deltaAffineRep *= 1e-10/sqrt(deltaAffineRep.cwise().square().sum());
        std::cout << "analytic k change: " << (dervK.cwise() * deltaAffineRep).sum() << "\n";
        std::cout << "finite diff k change: " << computeColwiseKurtosis(m+deltaAffineRep) - kurt  << "\n";
        */
        // ------------ END Derivative Checking ---------
        

        // Back prop this
        dervK = -dervK; // reverse since we want to maximize this
        affineEncoder.backProp( example, dervK, inputGradient);

        // Gradient Updates (weight decay is performed in the affineEncoder itself)
        sdn::Scalar oldLR = affineEncoder.learningRate;
        affineEncoder.learningRate = affineEncoder.learningRate * learningRateKurtosis;
        // bool oldLockBias = affineEncoder.lockBias;
        // affineEncoder.lockBias = true;
        affineEncoder.gradientUpdate();
        // affineEncoder.lockBias = oldLockBias;
        affineEncoder.learningRate = oldLR;

      }    
    }


  }

  const sdn::Matrix& JngiamAutoEncoderNetwork::encode(const sdn::Matrix& input) {
    
    affineEncoder.forwardProp(input, affineRepEnc);
    sigmEncoder.forwardProp(affineRepEnc, hiddenRepresentation);
    
    return hiddenRepresentation;
  }

  const sdn::Matrix& JngiamAutoEncoderNetwork::decode(const sdn::Matrix& hidden) {
    
    SDN_ASSERT(affineEncoder.inputSize > 0);
    
    affineDecoder.forwardProp(hidden, affineRepDec);
    if (noDecTanh) {
      reconstruction = affineRepDec;
    } else {
      sigmDecoder.forwardProp(affineRepDec, reconstruction);
    }
    
    if (normalizeOutputToL2Norm > 0) {
      // normalize reconstruction
      // to have l2norm = normalizeOutputToL2Norm
      // this rescales the output
      reconstruction *=  normalizeOutputToL2Norm  / sqrt(reconstruction.cwise().square().sum());
    }
    
    if (learnOutputScaling) {
      reconstruction *= outputScale;
    }
    
    return reconstruction;
  }

  void JngiamAutoEncoderNetwork::encodeBackProp(const sdn::Matrix& input,
            const sdn::Matrix& hiddenErrorDerivatives,
            sdn::Matrix& inputErrorDerivatives) {
    sdn::Matrix tempGradient;
    sigmEncoder.backProp(affineRepEnc, hiddenErrorDerivatives, tempGradient);
    affineEncoder.backProp(input, tempGradient, inputErrorDerivatives);
  }

  void JngiamAutoEncoderNetwork::decodeBackProp(const sdn::Matrix& hidden,
            const sdn::Matrix& outputErrorDerivatives,
            sdn::Matrix& hiddenErrorDerivatives) {
              
    sdn::Matrix tempGradient, newOutputDerv;
    if (learnOutputScaling) {
      // outputScaleGradient
      outputScaleGradient = ((outputErrorDerivatives.cwise() * reconstruction) / outputScale).sum() / outputErrorDerivatives.cols();
      newOutputDerv = outputScale*outputErrorDerivatives;
    }
        
    if (noDecTanh) {
      tempGradient = (learnOutputScaling ? newOutputDerv : outputErrorDerivatives);
    } else {
      sigmDecoder.backProp(affineRepDec, (learnOutputScaling ? newOutputDerv : outputErrorDerivatives), tempGradient);
    }
    affineDecoder.backProp(hidden, tempGradient, hiddenErrorDerivatives);
  }

  void JngiamAutoEncoderNetwork::encodeGradientUpdate() {
    affineEncoder.gradientUpdate();
  }

  void JngiamAutoEncoderNetwork::decodeGradientUpdate() {
    outputScale -= affineDecoder.learningRate * outputScaleGradient * outputScaleGradientLearningRate;
    affineDecoder.gradientUpdate();
  }

  const sdn::Matrix& JngiamAutoEncoderNetwork::getHiddenRepresentation() {
    return hiddenRepresentation;
  }
  
  const sdn::Matrix& JngiamAutoEncoderNetwork::getReconstruction() {
    return reconstruction;
  }
  
  void JngiamAutoEncoderNetwork::setupGoodInitialParameters() {
    
    if (affineEncoder.inputSize > 0) {
      RandomizeWeightMatrix().randomize(affineEncoder.weights, affineEncoder.inputSize);
    }
      
    if (hiddenLayerSize > 0) {
      RandomizeWeightMatrix().randomize(affineDecoder.weights, hiddenLayerSize);
      if (initialBias == 0 && targetActivation != 0)
        affineEncoder.bias = sigmEncoder.getBiasInitialParameters(targetActivation, hiddenLayerSize);
      else 
        affineEncoder.bias = Matrix::Constant(hiddenLayerSize, 1, initialBias);
    }
    
  }
  
  // --------------------------------------------------------------------------

  void JngiamAutoEncoderNetwork::setConfigurableParameters(sdn::Scalar defTargetActivation,
                                                             sdn::Scalar defLearningRateSparsity,
                                                             unsigned int defHiddenLayerSize,
                                                             sdn::Scalar defAveragingConstant) {

    // Shallow Loading

    targetActivation = defTargetActivation;
    hiddenLayerSize = defHiddenLayerSize;
    learningRateSparsity = defLearningRateSparsity;
    activationAveragingConstant = defAveragingConstant;
    
    initPersistentParameters();
    initTransientParameters();
  }

  void JngiamAutoEncoderNetwork::loadXML(XMLNode& node) {

    AttributeParser p;
    
    p.addDouble("targetActivation", 0);
    p.addDouble("initialBias", 0);
    
    p.addDouble("activationAveragingConstant");
    p.addDouble("kurtAveragingConstant", 0);
    p.addDouble("learningRateSparsity", 0);
    p.addDouble("learningRateKurtosis", 0);
    p.addInt("hiddenLayerSize");
    p.addBool("optimizeUnitKurtosis", 0);

    p.addDouble("biasShift", 0);
    p.addDouble("acceptableReconstructionError", 0);
    
    p.addBool("noDecTanh", false);
    p.addDouble("normalizeOutputToL2Norm", 0);
    
    p.addBool("learnOutputScaling", false);
    p.addDouble("outputScaleGradientLearningRate", 0.01);
    
    p.parse(node);

    targetActivation = p.getDouble("targetActivation");
    initialBias = p.getDouble("initialBias");
    activationAveragingConstant = p.getDouble("activationAveragingConstant");
    kurtAveragingConstant = p.getDouble("kurtAveragingConstant");
    learningRateSparsity = p.getDouble("learningRateSparsity");
    learningRateKurtosis = p.getDouble("learningRateKurtosis");
    hiddenLayerSize = p.getInt("hiddenLayerSize");
    optimizeUnitKurtosis = p.getBool("optimizeUnitKurtosis");

    biasShift = p.getDouble("biasShift");
    acceptableReconstructionError = p.getDouble("acceptableReconstructionError");

    noDecTanh = p.getBool("noDecTanh");
    normalizeOutputToL2Norm = p.getDouble("normalizeOutputToL2Norm");
    
    learnOutputScaling = p.getBool("learnOutputScaling");
    outputScaleGradientLearningRate = p.getDouble("outputScaleGradientLearningRate");

    // Deep Loading
    XMLNode affineEncoderNode = node.getChildNode("Encoder");
    if (!affineEncoderNode.isEmpty())
      affineEncoder.loadXML(affineEncoderNode);

    XMLNode affineDecoderNode = node.getChildNode("Decoder");
    if (!affineDecoderNode.isEmpty())
      affineDecoder.loadXML(affineDecoderNode);

    
    initPersistentParameters();
    initTransientParameters();

  }
  
  void JngiamAutoEncoderNetwork::initTransientParameters() {
    
    if (hiddenLayerSize > 0) {
      hiddenRepresentation = sdn::Matrix::Zero(hiddenLayerSize, 1);
      hiddenGradient = sdn::Matrix::Zero(hiddenLayerSize, 1);
      affineRepEnc =  sdn::Matrix::Zero(hiddenLayerSize, 1);
    }
    
    if (affineEncoder.inputSize > 0) {
      inputGradient = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
      outputGradient = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
      affineRepDec = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
    }
    
    activationTracker.setConfigurableParameters(activationAveragingConstant,
                                                targetActivation,
                                                SIZE_AUTOSET);
                                                
    normalizedKurtTracker.setConfigurableParameters(kurtAveragingConstant,
                                          TRACKER_AUTOSET,
                                          SIZE_AUTOSET);
    
    unnormalizedKurtTracker.setConfigurableParameters(kurtAveragingConstant,
                                          TRACKER_AUTOSET,
                                          SIZE_AUTOSET);
    
    errorTracker.setConfigurableParameters(0.001, TRACKER_AUTOSET, SIZE_AUTOSET);

  }
  
  void JngiamAutoEncoderNetwork::initPersistentParameters() {
    
    affineEncoder.initPersistentParameters();
    affineDecoder.initPersistentParameters();
    
    sigmEncoder.initPersistentParameters();
    sigmDecoder.initPersistentParameters();

    setupGoodInitialParameters();

    outputScale = 1;
  }

  const boost::shared_ptr<CostFunction> JngiamAutoEncoderNetwork::getCostFunction() {
    return costFunc;
  }

  void JngiamAutoEncoderNetwork::serialize(serial::Archive & ar)  {
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1030;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
        
      case 1030:
        ar.attribute("outputScale", outputScale);
        ar.attribute("learnOutputScaling", learnOutputScaling);
      
      case 1020:
        ar.attribute("normalizeOutputToL2Norm", normalizeOutputToL2Norm);
        
      case 1010:
        ar.attribute("noDecTanh", noDecTanh);
        ar.attribute("initialBias", initialBias);
        
      case 1000:
      
        // Persistent
        affineEncoder.serialize(ar);
        affineDecoder.serialize(ar);
        sigmEncoder.serialize(ar);
        sigmDecoder.serialize(ar);
        
        // Configurable
        ar.attribute("targetActivation", targetActivation);
        ar.attribute("activationAveragingConstant", activationAveragingConstant);
        ar.attribute("kurtAveragingConstant", kurtAveragingConstant);
        ar.attribute("learningRateSparsity", learningRateSparsity);
        ar.attribute("learningRateKurtosis", learningRateKurtosis);
        ar.attribute("biasShift", biasShift);
        ar.attribute("acceptableReconstructionError", acceptableReconstructionError);
        ar.attribute("optimizeUnitKurtosis", optimizeUnitKurtosis);
        ar.attribute("hiddenLayerSize", hiddenLayerSize);
    
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }

  JngiamAutoEncoderNetwork::JngiamAutoEncoderNetwork() {
    costFunc = boost::shared_ptr<CostFunction> (new SumOfSquaresCostFunction());
  }
  
  JngiamAutoEncoderNetwork::~JngiamAutoEncoderNetwork() {
  }

  
}

