//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: autoTuneScript.cc
//Description: 

#include "autoTuneScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn {

  class AutoTuneScriptConfigParser : public ConfigParser {

  protected:
    AutoTuneScript & parentScript;

  public:
  
    AutoTuneScriptConfigParser(AutoTuneScript & parent) : parentScript(parent) {}
    ~AutoTuneScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::AutoTuneScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::AutoTuneScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  AutoTuneScript::AutoTuneScript() {
    loadUsingConfigManager();
  }
  
  AutoTuneScript::~AutoTuneScript() {
  }
  
  void AutoTuneScript::loadXML(XMLNode& node) {
    
    AttributeParser p;

    p.addString("encodingNetworkFile", "");
    
    p.addString("trainDataset");
    p.addString("validateDataset");

    p.addString("useWhiteningMatrixFile", "");
    p.addString("useMeanMatrixFile", "");
    
    p.addInt("inputHeight");
    p.addInt("inputWidth");

    p.addInt("nExamples", 1);
    p.addInt("hyperTuneStep", 1);

    p.parse(node);
    
    // Note that values are set *directly* into the network
    std::string encodingNetworkFile = p.getString("encodingNetworkFile");
    useWhiteningMatrixFile = p.getString("useWhiteningMatrixFile");
    useMeanMatrixFile = p.getString("useMeanMatrixFile");

    trainDatasetName = p.getString("trainDataset");
    validateDatasetName = p.getString("validateDataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    nExamples = p.getInt("nExamples");
    
    hyperTuneStep = p.getInt("hyperTuneStep");
    
    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network.loadXML(networkNode); 
    
    // Load another network to use
    if (encodingNetworkFile != "") {
      
      // Load up encodingNetworkFile
      std::ifstream fin;
      fin.open(encodingNetworkFile.c_str(), std::ios::binary);
      SDN_ASSERT_MSG(fin.good(), "Unable to open " << encodingNetworkFile);
      sdn::serial::BinInArchive ar(fin);
      boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
      fin.close();
      
      SDN_ASSERT_MSG(net, "Unable to load network from " << encodingNetworkFile);

      // Cast to Just Any AutoEncoder
      boost::shared_ptr<sdn::AbstractAutoEncoderNetwork> anet = 
        boost::dynamic_pointer_cast<sdn::AbstractAutoEncoderNetwork> (net);
      
      SDN_ASSERT_MSG(anet, net->className() << " is not an autoencoder network!");
      
      autoEncNetwork = anet;
      
      SDN_MSG("Loaded " << autoEncNetwork->className() << " as the encoding network!");
    }

  }

  void AutoTuneScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new AutoTuneScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();

  }
  
  void AutoTuneScript::initTransientParameters() {

    trainDataset =  DatasetRegistry().make(trainDatasetName);
    if (!trainDataset) {
      SDN_ERROR("Could not load dataset with name " << trainDatasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    trainDataset->requestExampleSize(inputHeight, inputWidth);

    validateDataset =  DatasetRegistry().make(validateDatasetName);
    if (!validateDataset) {
      SDN_ERROR("Could not load dataset with name " << validateDatasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    validateDataset->requestExampleSize(inputHeight, inputWidth);

    if (useWhiteningMatrixFile != "") {
      SDN_ASSERT(useMeanMatrixFile != "");
      std::ifstream fin;

      fin.open(useWhiteningMatrixFile.c_str());
      serial::BinInArchive arinW(fin);
      serializeMatrix(arinW, whiteMat);
      fin.close();

      fin.open(useMeanMatrixFile.c_str());
      serial::BinInArchive arinM(fin);
      serializeMatrix(arinM, meanMat);
      fin.close();
      
      invWhite = whiteMat.transpose();
      for(int i=0;i<invWhite.cols();i++) {
        invWhite.col(i) /= invWhite.col(i).cwise().square().sum();
      }
  
      
    } else {
      whiteMat = sdn::Matrix::Zero(1,1);
      meanMat = sdn::Matrix::Zero(1,1);
    }


    // We are requesting column vector examples
    oneInputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
    
    if (whiteMat.size() > 1) {
      inputExample = sdn::Matrix::Zero(whiteMat.rows(), nExamples);
      targetOutput = sdn::Matrix::Zero(trainDataset->numLabels(), nExamples);
    } else {
      inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, nExamples);
      targetOutput = sdn::Matrix::Zero(trainDataset->numLabels(), nExamples);
    }
    

  }
  
  void AutoTuneScript::initPersistentParameters() {

    network.initTransientParameters();
    network.initPersistentParameters();

    clearFile("actstats.dat");
    clearFile("outputscales.dat");
    clearFile("suptrerror.dat");
    clearFile("supvlerror.dat");
    clearFile("outputscales.dat");
    clearFile("targetact.dat");
    
  }
  
  void AutoTuneScript::serialize(serial::Archive & ar) {
    SDN_FATAL("AutoTuneScript is not serializable yet");
  }

  
  void AutoTuneScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    // ----- TRAINING SET --------------
    targetOutput.fill(0.0);

    for(int i=0 ; i < nExamples; i++) {
      
      int label = trainDataset->getRandomExampleAsColumn(oneInputExample);
      
      if (whiteMat.size() > 1) {
        SDN_FATAL("Keep away from whitening here first");
        // Using Whitening
        whiteInputExample = whiteMat * (oneInputExample - meanMat);
        whiteInputExample /= 5; // Divide by data max after whitening, MNIST Specific
        //normInput.normalizePatch(whiteInputExample);
      } else {
        normInput.normalizePatch(oneInputExample);
      }
      
      sdn::Matrix& theExample = (whiteMat.size() > 1 ? whiteInputExample : oneInputExample);
      
      if (autoEncNetwork) {
        encodedExample = autoEncNetwork->encode(theExample);
        if (inputExample.rows() != encodedExample.rows())
          inputExample.resize(encodedExample.rows(), nExamples);
        inputExample.col(i) = encodedExample.col(0);
      } else {
        inputExample.col(i) = theExample.col(0);
      }
      targetOutput(label, i) = 1.0;
     
    }
    
    // do fancy stuff here
    
    // do not learn output scaling until later
    sdn::Scalar temp1, temp2, temp3;
    if (manager->currentIteration <= 50000) { // choice of 50k is rather arbitary
      temp1 = network.encodeScale.learningRate;
      temp2 = network.decodeScale.learningRate;
      network.encodeScale.learningRate = 0;
      network.decodeScale.learningRate = 0;
    }
    
    network.trainTrainingExample(inputExample, targetOutput);
    
    if (manager->currentIteration <= 50000) {
      network.encodeScale.learningRate = temp1;
      network.decodeScale.learningRate = temp2;
    }
    
    // ----- VALIDATION SET -----
    
    if (manager->currentIteration % hyperTuneStep != 0) {
      // Skip on validation set
      return;
    }
    
    targetOutput.fill(0.0);

    for(int i=0 ; i < nExamples; i++) {
      
      int label = validateDataset->getRandomExampleAsColumn(oneInputExample);

      if (whiteMat.size() > 1) {
        // Using Whitening
        whiteInputExample = whiteMat * (oneInputExample - meanMat);
        whiteInputExample /= 5; // Divide by data max after whitening, MNIST Specific
        // normInput.normalizePatch(whiteInputExample);
      } else {
        normInput.normalizePatch(oneInputExample);
      }
      
      sdn::Matrix& theExample = (whiteMat.size() > 1 ? whiteInputExample : oneInputExample);

      
      if (autoEncNetwork) {
        encodedExample = autoEncNetwork->encode(oneInputExample);
        if (inputExample.rows() != encodedExample.rows())
          inputExample.resize(encodedExample.rows(), nExamples);
        inputExample.col(i) = encodedExample.col(0);
      } else {
        inputExample.col(i) = theExample.col(0);
      }
      targetOutput(label, i) = 1.0;
      
    }
    
    // do fancy stuff here
    
    // do not learn output scaling until later
    if (manager->currentIteration <= 50000) { // choice of 50k is rather arbitary
      temp1 = network.encodeScale.learningRate;
      temp2 = network.decodeScale.learningRate;
      temp3 = network.hyperEncScaleLearningRate ;
      network.encodeScale.learningRate = 0;
      network.decodeScale.learningRate = 0;
      network.hyperEncScaleLearningRate = 0;
    }
    
    network.trainValidationExample(inputExample, targetOutput);
    
    if (manager->currentIteration <= 50000) {
      network.encodeScale.learningRate = temp1;
      network.decodeScale.learningRate = temp2;
      network.hyperEncScaleLearningRate = temp3;
    }
  }
  
  // ---- Report Progress Callback ----
  void AutoTuneScript::reportProgress(std::ostream& o) {
    
    o << "[" << itos(manager->currentIteration, 10) << "] Act Mean: " << network.activationTracker.getMean().sum() / network.activationTracker.getMean().size()  << "\n";
    o << "[" << itos(manager->currentIteration, 10) << "] Reconstruction Error: " << network.reconstructionErrorTracker.getMean().sum() / network.reconstructionErrorTracker.getMean().size()  << "\n";
    o << "[" << itos(manager->currentIteration, 10) << "] Sup Train Error: " << network.supervisedTrainingErrorTracker.getMean().sum() / network.supervisedTrainingErrorTracker.getMean().size()  << "\n";
    o << "[" << itos(manager->currentIteration, 10) << "] Sup Valid Error: " << network.supervisedValidationErrorTracker.getMean().sum() / network.supervisedValidationErrorTracker.getMean().size()  << "\n";
    
    // output stuff to files
    appendToFile("actstats.dat", "%u %f\n", manager->currentIteration, network.activationTracker.getMean().sum() / network.activationTracker.getMean().size() );
    appendToFile("reconerror.dat", "%u %f\n", manager->currentIteration, network.reconstructionErrorTracker.getMean().sum() / network.reconstructionErrorTracker.getMean().size() );
    appendToFile("suptrerror.dat", "%u %f\n", manager->currentIteration, network.supervisedTrainingErrorTracker.getMean().sum() / network.supervisedTrainingErrorTracker.getMean().size() );
    appendToFile("supvlerror.dat", "%u %f\n", manager->currentIteration, network.supervisedValidationErrorTracker.getMean().sum() / network.supervisedValidationErrorTracker.getMean().size() );
    appendToFile("outputscales.dat", "%u %f %f\n", manager->currentIteration, network.encodeScale.scale, network.decodeScale.scale  );
    appendToFile("targetact.dat", "%u %f\n", manager->currentIteration, network.targetActivation  );
    
  }
  
  // ---- Visualization Callback ----
  const sdn::Matrix& AutoTuneScript::getEncoderWeights() {
    if (whiteMat.size() > 1) {
      temp = network.encodeAffine.weights*invWhite.transpose();
      return temp;
    } else
      return network.encodeAffine.weights;
  }

  const int AutoTuneScript::getWeightPatchHeight()  {
    return inputHeight;
  }
  
  const sdn::Matrix& AutoTuneScript::getDecoderWeights() {
    if (whiteMat.size() > 1) {
      temp = network.decodeAffine.weights*invWhite.transpose();
      return temp;
    } else
      return network.decodeAffine.weights;
  }
  
  // ---- Save Model Callback ----
  const void AutoTuneScript::saveModelInDirectory(std::string dir)  {
    
    
    // write using the serializer
    network.quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");
    
    std::ofstream fout;

    fout.open( quickString("%s%010u-ew.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.encodeAffine.weights;
    fout.close();

    fout.open( quickString("%s%010u-eb.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.encodeAffine.bias;
    fout.close();

    fout.open( quickString("%s%010u-es.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.encodeScale.scale;
    fout.close();

    fout.open( quickString("%s%010u-dw.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.decodeAffine.weights;
    fout.close();

    fout.open( quickString("%s%010u-db.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.decodeAffine.bias;
    fout.close();

    fout.open( quickString("%s%010u-ds.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network.decodeScale.scale;
    fout.close();


  }

 
}
