//Author: Jiquan <jngiam@cs.stanford.edu>
//File: jngiamScript.cc
//Description: 

#include "jngiamScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn {

  class JngiamScriptConfigParser : public ConfigParser {

  protected:
    JngiamScript & parentScript;

  public:
  
    JngiamScriptConfigParser(JngiamScript & parent) : parentScript(parent) {}
    ~JngiamScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::JngiamScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::JngiamScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  JngiamScript::JngiamScript() {
    loadUsingConfigManager();
  }
  
  JngiamScript::~JngiamScript() {
  }
  
  void JngiamScript::loadXML(XMLNode& node) {
    
    AttributeParser p;

    p.addString("encodingNetworkFile", "");
    
    p.addString("network");
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    p.addInt("nExamples", 1);
    p.addBool("useL2Normalization");
    p.addBool("noScaleNormalization", false);
    
    p.parse(node);
    
    // Note that values are set *directly* into the network
    std::string encodingNetworkFile = p.getString("encodingNetworkFile");
    networkName = p.getString("network");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    nExamples = p.getInt("nExamples");
    useL2Normalization = p.getBool("useL2Normalization");
    noScaleNormalization = p.getBool("noScaleNormalization");
    
    // We need to create the network here because the XML parameters
    // for it need to be read directly to the network    
    network = boost::dynamic_pointer_cast<JngiamAutoEncoderNetwork> (AbstractNetworkRegistry().make(networkName));
    if (!network) {
      SDN_ERROR("Could not load network with name " << networkName);
      SDN_ERROR("Is " << networkName << " an JngiamAutoEncoderNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network->loadXML(networkNode); 
    
    // Load another network to use
    if (encodingNetworkFile != "") {
      
      // Load up encodingNetworkFile
      std::ifstream fin;
      fin.open(encodingNetworkFile.c_str(), std::ios::binary);
      SDN_ASSERT_MSG(fin.good(), "Unable to open " << encodingNetworkFile);
      sdn::serial::BinInArchive ar(fin);
      boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
      fin.close();
      
      SDN_ASSERT_MSG(net, "Unable to load network from " << encodingNetworkFile);

      // Cast to Just Any AutoEncoder
      boost::shared_ptr<sdn::AbstractAutoEncoderNetwork> anet = 
        boost::dynamic_pointer_cast<sdn::AbstractAutoEncoderNetwork> (net);
      
      SDN_ASSERT_MSG(anet, net->className() << " is not an autoencoder network!");
      
      autoEncNetwork = anet;
      
      SDN_MSG("Loaded " << autoEncNetwork->className() << " as the encoding network!");
    }

  }

  void JngiamScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new JngiamScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();

  }
  
  void JngiamScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, nExamples);
    oneInputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
    
  }
  
  void JngiamScript::initPersistentParameters() {
    network->initTransientParameters();
    network->initPersistentParameters();
    clearFile("normkurtstats.dat");
    clearFile("unnormkurtstats.dat");
    clearFile("kurtstats.dat");
    clearFile("actstats.dat");
    clearFile("outputscale.dat");
  }
  
  void JngiamScript::serialize(serial::Archive & ar) {
    SDN_FATAL("JngiamScript is not serializable yet");
  }

  
  void JngiamScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    for(int i=0 ; i < nExamples; i++) {
      
      dataset->getRandomExampleAsColumn(oneInputExample);
      
      if (!noScaleNormalization) {
        if (useL2Normalization) {
          double mean = inputExample.sum() / (inputExample.rows() * inputExample.cols());
          inputExample.cwise() -= mean;
          inputExample /= sqrt(inputExample.cwise().square().sum());
        } else {
          normInput.normalizePatch(oneInputExample);
        }
      } else {
        double mean = inputExample.sum() / (inputExample.rows() * inputExample.cols());
        inputExample.cwise() -= mean;
      }
      
      if (autoEncNetwork) {
        encodedExample = autoEncNetwork->encode(oneInputExample);
        if (inputExample.rows() != encodedExample.rows())
          inputExample.resize(encodedExample.rows(), nExamples);
        inputExample.col(i) = encodedExample.col(0);
      } else {
        inputExample.col(i) = oneInputExample.col(0);
      }
      
    }
    
    // do fancy stuff here
    
    // do not learn output scaling until later
    sdn::Scalar temp;
    if (manager->currentIteration <= 50000) { // choice of 50k is rather arbitary
      temp = network->outputScaleGradientLearningRate;
      network->outputScaleGradientLearningRate = 0;
    }
    
    network->unsupervisedTrain(inputExample);
    
    if (manager->currentIteration <= 50000) {
      network->outputScaleGradientLearningRate = temp;
    }
  }
  
  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> JngiamScript::getReconstructionCostFunction() {
    return network->getCostFunction();
  }

  const sdn::Matrix& JngiamScript::getReconstruction() {
    return network->getReconstruction();
  }
  
  const sdn::Matrix& JngiamScript::getOriginal() {
    return inputExample;
  }
 
  // ---- Report Progress Callback ----
  void JngiamScript::reportProgress(std::ostream& o) {
    
    o << "[" << itos(manager->currentIteration, 10) << "] Norm Kurt Mean: " << network->normalizedKurtTracker.getMean() << "\n";
    o << "[" << itos(manager->currentIteration, 10) << "] Unnorm Kurt Mean: " << network->unnormalizedKurtTracker.getMean() << "\n";
    o << "[" << itos(manager->currentIteration, 10) << "] Act Mean: " << network->activationTracker.getMean().sum() / network->activationTracker.getMean().size()  << "\n";
    
    // output stuff to files
    appendToFile("normkurtstats.dat", "%u %f\n", manager->currentIteration, network->normalizedKurtTracker.getMean()[0]);
    appendToFile("unnormkurtstats.dat", "%u %f\n", manager->currentIteration, network->unnormalizedKurtTracker.getMean()[0]);
    appendToFile("actstats.dat", "%u %f\n", manager->currentIteration, network->activationTracker.getMean().sum() / network->activationTracker.getMean().size() );
    appendToFile("outputscale.dat", "%u %f\n", manager->currentIteration, network->outputScale );
    
  }
  
  // ---- Visualization Callback ----
  const sdn::Matrix& JngiamScript::getEncoderWeights() {
    
    if (autoEncNetwork) {
      // return a weight matrix that sums the autoEncNetwork weight matrix
      boost::shared_ptr<JngiamAutoEncoderNetwork> net = boost::dynamic_pointer_cast<JngiamAutoEncoderNetwork> (autoEncNetwork);
      SDN_ASSERT_MSG (net, "2nd Layer Visualization Requires First Layer to be a JngiamAutoEncoderNetwork");
      // combine the weights!
      tempVizMatrix = network->affineEncoder.weights * net->affineEncoder.weights;
      return tempVizMatrix;
    } else
      return network->affineEncoder.weights;
  }
  
  const int JngiamScript::getWeightPatchHeight()  {
    return inputHeight;
  }
  
  const sdn::Matrix& JngiamScript::getDecoderWeights() {
    return network->affineDecoder.weights;
  }
  
  // ---- Save Model Callback ----
  const void JngiamScript::saveModelInDirectory(std::string dir)  {
    
    
    // write using the serializer
    network->quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");
    
    std::ofstream fout;

    fout.open( quickString("%s%010u-ew.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineEncoder.weights;
    fout.close();

    fout.open( quickString("%s%010u-eb.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineEncoder.bias;
    fout.close();

    fout.open( quickString("%s%010u-dw.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineDecoder.weights;
    fout.close();

    fout.open( quickString("%s%010u-db.amat", dir.c_str(), manager->currentIteration).c_str() );
    fout << network->affineDecoder.bias;
    fout.close();

  }

 
}
