//Author: NAME <name@stanford.edu>
//File: templateScript.cc
//Description: 

#include "templateScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn {

  class TemplateScriptConfigParser : public ConfigParser {

  protected:
    TemplateScript & parentScript;

  public:
  
    TemplateScriptConfigParser(TemplateScript & parent) : parentScript(parent) {}
    ~TemplateScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::TemplateScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::TemplateScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  TemplateScript::TemplateScript() {
    loadUsingConfigManager();
  }
  
  TemplateScript::~TemplateScript() {
  }
  
  void TemplateScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("network");
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    
    p.parse(node);
    
    // Note that values are set *directly* into the network
    networkName = p.getString("network");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    
    // We need to create the network here because the XML parameters
    // for it need to be read directly to the network    
    network = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> (AbstractNetworkRegistry().make(networkName));
    if (!network) {
      SDN_ERROR("Could not load network with name " << networkName);
      SDN_ERROR("Is " << networkName << " an AbstractAutoEncoderNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network->loadXML(networkNode); 

  }

  void TemplateScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new TemplateScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();

  }
  
  void TemplateScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
  }
  
  void TemplateScript::initPersistentParameters() {
    network->initTransientParameters();
    network->initPersistentParameters();
  }
  
  void TemplateScript::serialize(serial::Archive & ar) {
    SDN_FATAL("TemplateScript is not serializable yet");
  }

  
  void TemplateScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // do fancy stuff here
    normInput.normalizePatch(inputExample);
    network->unsupervisedTrain(inputExample);
    
  }
  
  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> TemplateScript::getReconstructionCostFunction() {
    return network->getCostFunction();
  }

  const sdn::Matrix& TemplateScript::getReconstruction() {
    return network->getReconstruction();
  }
  
  const sdn::Matrix& TemplateScript::getOriginal() {
    return inputExample;
  }
 
 
}
