//Author: Daniel Chia, Zhenghao Chen, Pang Wei Koh
//File: invarianceTestBlock.cc
//Description: Implements invariance metric

#include <ext/numeric>
#include "invarianceTestBlock.h"
#include "../../../stable/math/costFunction.h"
#include "../../../stable/helpers.h"
#include "../../../stable/scriptRuntimeManager.h"
#include "../../../registry/scriptRegistry.h"
#include "sdn/base.h"
#include <iostream>
#include <math.h>

using namespace std;

namespace sdn {


	void InvarianceTestBlock::loadXML(XMLNode & node) {

		AttributeParser p;

		p.addDouble("quantile");
		p.addDouble("neuronFraction");
		p.addBool("displayTrajectory");
		p.addString("trajectorySavePath");
		p.parse(node);

		quantile = p.getDouble("quantile");
//		cout << "quantile is " << quantile << endl;
		neuronFraction = p.getDouble("neuronFraction");
		displayTrajectory = p.getBool("displayTrajectory");
		trajectorySavePath = p.getString("trajectorySavePath");
		initPersistentParameters();
		initTransientParameters();

	}


	void InvarianceTestBlock::invarianceTest(boost::shared_ptr<sdn::AbstractAutoEncoderNetwork> networkInput, int patchWidth, int patchHeight){
		this->network = networkInput;
		this->patchWidth = patchWidth;
		this->patchHeight = patchHeight;
		itd = new NaturalTestDataset(patchWidth, patchHeight);
		firingRateTest();
		cout << "natural test, network score " << networkScoreOut << endl;
	}

	void InvarianceTestBlock::serialize(serial::Archive & ar) {
		SDN_FATAL(className() << " not serializable yet");
	}


	void InvarianceTestBlock::computeThresholdAndGlobalCount(std::vector<Scalar> activations,
			Scalar &threshold,
			Scalar &globalCount)
	{
		SDN_ASSERT(quantile <= 1 && quantile >= 0);
		std::vector<int> indices(activations.size());
		iota(indices.begin(), indices.end(), 0);
		indirect_less_than<Scalar> igt(&(activations[0]));
		std::sort(indices.begin(), indices.end(),igt);

		int index_quantile = (int) (indices.size()*quantile);
		threshold = activations[indices[index_quantile]];
		globalCount = 0;//indices.size() - index_quantile;
		for (unsigned int i=0;i<activations.size();i++){
			if (activations[i] >= threshold)
				globalCount = globalCount + 1;
		}
	}


	// compute average firing rate of top scoring neurons in a network with sign compensation
	void InvarianceTestBlock::computeThresholdScore(const Matrix localPositive, 
			const Matrix middlePositive,
			const Matrix globalPositive,
			const Matrix localNegative, 
			const Matrix middleNegative,
			const Matrix globalNegative,
			Scalar &networkScore)
	{
		SDN_ASSERT(localPositive.rows() == middlePositive.rows() && 
				localPositive.cols() == middlePositive.cols());
		SDN_ASSERT(localNegative.rows() == middleNegative.rows() && 
				localNegative.cols() == middleNegative.cols());
		SDN_ASSERT(localNegative.rows() == localPositive.rows() && 
				localNegative.cols() == localPositive.cols());
		std::vector<Scalar> max_scores;
		int poscount = 0;
		for (int i=0;i<localPositive.rows();i++){
			for (int j=0;j<localPositive.cols();j++){
				Scalar positive_score;
				if (middlePositive(i,j) != 0){
					positive_score = localPositive(i,j) / middlePositive(i,j) / globalPositive(i,j);
				}
				else{
					positive_score = 0;
				}
				Scalar negative_score;
				if (middleNegative(i,j) != 0){
					negative_score = localNegative(i,j) / middleNegative(i,j) / globalNegative(i,j);
				}
				else{
					negative_score = 0;
				}
				Scalar score = MAX(positive_score, negative_score);
				if (positive_score > negative_score) 
					poscount ++ ;
//					cout << "1 " ;
//				else 
//					cout << "0 " ;
//				cout << endl;				
				max_scores.push_back(score);
			}
		}
		std::vector<int> indices(max_scores.size());
		iota(indices.begin(), indices.end(), 0);
		indirect_greater_than<Scalar> igt(&(max_scores[0]));
		sort(indices.begin(), indices.end(),igt);
		int k = (int) (((float)neuronFraction * (float) localPositive.rows() * (float) localPositive.cols()) +0.5);
		Scalar mean_score = 0;
		for (int i=0; i<k;i++){
			mean_score = mean_score + max_scores[indices[i]];
		}
		networkScore = mean_score / k;
		cout << "networkScore for this test : " << networkScore << endl;
		cout << "pos count is " << poscount << "out of " << localPositive.rows()*localPositive.cols() << endl;
	}


	// this function returns the threshold that activates quantile percentage of cells
	// and use this for counting the number of cells that activate globally
	void InvarianceTestBlock::computeGlobalStatistics(
			int sign,
			Matrix &thresholds,
			Matrix &globalCounts)
	{
		SDN_ASSERT(sign==-1 || sign==1);
		std::vector<Matrix> exhaustivePatches;
		itd->getExhaustivePatches(exhaustivePatches);

		std::vector<Matrix> responses;
		std::vector<Scalar> activations;

		// run through all random patches
		for (unsigned int k=0;k<exhaustivePatches.size();k++){
			// collect the response for a random patch
			//	network->processExampleToHighestLayer(exhaustivePatches[k]);
			//Matrix highestLayer = network->getHighestLayerActivities();
			Matrix highestLayer = network->encode(exhaustivePatches[k]);

			responses.push_back(highestLayer);
		}		
		thresholds = Matrix::Zero(responses[0].rows(),responses[0].cols());
		globalCounts = Matrix::Zero(responses[0].rows(),responses[0].cols());
		for (int i=0;i<responses[0].rows();i++){
			for (int j=0;j<responses[0].cols();j++){
				std::vector<Scalar> activations;
				for (unsigned int k=0;k<responses.size();k++){
					activations.push_back(sign*responses[k](i,j));
				}
				Scalar threshold;
				Scalar globalCount;
				computeThresholdAndGlobalCount(activations, threshold, globalCount);
				thresholds(i,j) = threshold;
				globalCounts(i,j) = globalCount;
			}
		}

		globalCounts = globalCounts / responses.size();
/*
		cout << "the global count is: " << endl;
		for (int i=0; i<globalCounts.rows(); i++)
			for (int j=0; j<globalCounts.cols(); j++) 
				cout << globalCounts(i,j) << " ";
		cout<<endl;

		cout << endl << "the thresholds are: " << endl;
		for (int i=0; i<thresholds.rows(); i++)
			for (int j=0; j<thresholds.cols(); j++) 
				cout << thresholds(i,j) << " ";
		cout<<endl;

		cout << "the quantile is " << quantile << " and the neuronFraction is " << neuronFraction << endl; 
*/
	}

	// compute the firing rate for local transformed images
	void InvarianceTestBlock::computeLocalStatistics(
			int inv_type, 
			int sign,
			Matrix thresholds, 
			Matrix &localCounts,
			Matrix &middleCounts)
	{
		SDN_ASSERT(sign==NEGATIVE || sign==POSITIVE);
		std::vector<Matrix> inexhaustivePatches;
		std::vector<Scalar> inexhaustivePhases;
		std::vector<Scalar> inexhaustiveOrientations;
		std::vector<Scalar> inexhaustiveFrequencies;

		itd->getInexhaustivePatches(inv_type, 
				inexhaustivePatches);  
		//	network->processExampleToHighestLayer(inexhaustivePatches[0]);
		//	Matrix highestLayer = network->getHighestLayerActivities();
		Matrix highestLayer = network->encode(inexhaustivePatches[0]);

		nHiddenRows = highestLayer.rows();
		nHiddenCols = highestLayer.cols();
		patchSize = inexhaustivePatches[0].size();

		localCounts = Matrix::Zero(nHiddenRows, nHiddenCols);
		middleCounts = Matrix::Zero(nHiddenRows, nHiddenCols);

		char * imageFileName;

		//Storing local trajectories may take up too much space --> store index of patches to trade computational speed for space?
		//Could reorder the loops to add patches as they are found but would require many more computations of activations
		//std::vector<std::vector<std::vector<Matrix>>> localTrajectories (nHiddenRows*nHiddenCols, std::vector<std::vector<Matrix>>());
		trajectoryIndices.resize(trajectoryIndices.size() + nHiddenRows*nHiddenCols);
		int trajectoryCount = 0;
		std::vector<Matrix> trajectoryTest;
		//Finding number of patches in each local trajectory, should be 10 -- currently hardcoded as a constant in naturalImageDataset
		//TODO: Find better way of doing this
		itd->getLocalPatches(1,1,trajectoryTest);
		trajectoryFrameCount = trajectoryTest.size();
		//cout << "size of inexhaustivePatches is " << inexhaustivePatches.size() << endl;
		// run through each hidden unit

		int encodeCount = 0;

		std::vector<Matrix> locallyVaryingPatches;

//		cout << "*********** computing local statistics: inv_type = " << inv_type << " , sign = " << sign << "***********" << endl;
//	 	cout << "inexhaustivePatches.size() = " << inexhaustivePatches.size() << endl;
//		cout << "nHiddenRows = " << nHiddenRows << " , nHiddenCols = " << nHiddenCols << endl;

		for (unsigned int i=0;i<inexhaustivePatches.size();i++){
			//if (i % 50 == 0)
				//cout << "at patch " << i << endl;
			// check hidden unit (r,c) response for a particular patch index by i
			//	network->processExampleToHighestLayer(inexhaustivePatches[i]);
			//	highestLayer = network->getHighestLayerActivities();

			
			highestLayer = network->encode(inexhaustivePatches[i]);
			encodeCount++;

			int flag = 0;

			std::vector<Matrix> highestLayer_temp;

			for (int r=0;r<nHiddenRows;r++){
				for (int c=0;c<nHiddenCols;c++){			

					Scalar val = sign*highestLayer(r,c);
					// run through each patch in the middle frame
					// if the response is higher than a threshold, we can use this as a stimulus
					if (val >= thresholds(r,c)){

						middleCounts(r,c) = middleCounts(r,c) + 1;
						// get locally varying data
						if (!flag){
							itd->getLocalPatches(inv_type, i, locallyVaryingPatches);
							for (unsigned int k=0;k<locallyVaryingPatches.size();k++){
								highestLayer_temp.push_back(network->encode(locallyVaryingPatches[k]));
								encodeCount++;
							}

							flag = 1;
						}

						//store local trajectories for display later
//						if (displayTrajectory){
							trajectoryIndices[r*nHiddenCols+c].push_back(i);
							//localTrajectories[r*nHiddenCols+c].push_back(locallyVaryingPatches);							
							trajectoryCount++;
//						}

						// run through network trajectory and check the response
						for (unsigned int k=0;k<locallyVaryingPatches.size();k++){
							//	network->processExampleToHighestLayer(locallyVaryingPatches[k]);
							//		Matrix highestLayer_temp = network->getHighestLayerActivities();

//							Matrix highestLayer_temp = network->encode(locallyVaryingPatches[k]);


							Scalar trajVal = sign*highestLayer_temp[k](r,c);
							if (trajVal >= thresholds(r,c)){
								localCounts(r,c) = localCounts(r,c) + 1.0 / (locallyVaryingPatches.size()-1);
							}
						}
						localCounts(r,c) = localCounts(r,c) - 1.0 / (locallyVaryingPatches.size()-1);
					}
				}
			}
		}
/*	 	cout << "locallyVaryingPatches.size() = " << locallyVaryingPatches.size() << endl;
		cout << "encodeCount = " << encodeCount << endl;
		cout << "middleCounts: " << endl;	
		for (int i=0; i<middleCounts.rows(); i++)
			for (int j=0; j<middleCounts.cols(); j++) 
				cout << middleCounts(i,j) << " ";
		cout<<endl;

				
		cout << "*********** done computing local statistics ***********" << endl;
*/

		//Display all local trajectories
		if (displayTrajectory){
			if (trajectoryCount == 0){
				cout << "No local trajectories to display" << endl;
			} else {
				
				for (int r = 0; r < nHiddenRows;r++){
					for (int c = 0;c < nHiddenCols;c++){
					//	int r = 0;
						PatchDisplay viz("Firing Rate Test trajectories",patchWidth,patchHeight,1,trajectoryFrameCount,trajectoryIndices[r*nHiddenCols+c].size());
						for (unsigned int k = 0;k < trajectoryIndices[r*nHiddenCols+c].size();k++){
							std::vector<Matrix> localTrajectory;
							itd->getLocalPatches(inv_type,trajectoryIndices[r*nHiddenCols+c][k],localTrajectory);
							//cout << "size of localTrajectory is " << localTrajectory.size() << endl;
							for (int l = 0;l < trajectoryFrameCount;l++){
								Vector trajectoryFrame(patchWidth*patchHeight);
								Matrix trajectoryFrameMat (patchWidth, patchHeight);
								
								matrixToVector(localTrajectory[l],trajectoryFrame);
								
								vectorToMatrix(trajectoryFrame,trajectoryFrameMat);
																																																																																																														
								viz.addPatch(trajectoryFrameMat,0,0);
							}
						}
						imageFileName = (char *)trajectorySavePath.c_str();
						sprintf(imageFileName, "%s/trajectory_%d_%d_%d.png",imageFileName, r*nHiddenCols+c, inv_type, sign);
						viz.saveImage(imageFileName);
						viz.clearDisplay();
					}
				}
				//viz.enableWindow();
				//SDN_MSG("Press any key");
				//viz.wait(0);

				
				//viz.saveImage("
			}
		}
		//TODO: Implement saving local trajectories onto file
	}


	// perform firing rate test, local over global firing rates
	void InvarianceTestBlock::firingRateTest()
	{
		networkScoreOut = 0;
		templateScoreOut = 0;
		maxpoolingScoreOut = 0;

		Matrix networkThresholdsPositive;
		Matrix globalPositive;
		Matrix networkThresholdsNegative;
		Matrix globalNegative;

		int counter = 0;
		int numCategory = itd->getNumCategory();
		std::vector<int> counterPerCategory;
		for (int i=0;i<numCategory;i++){
			scorePerCategory.push_back(0);
			counterPerCategory.push_back(0);
		}
		cout << "computing global statistics" << endl;
		computeGlobalStatistics(POSITIVE, 
				networkThresholdsPositive, 
				globalPositive);
		cout << "computing global statistics" << endl;
		computeGlobalStatistics(NEGATIVE, 
				networkThresholdsNegative, 
				globalNegative);
		std::vector<Scalar> scores;

		for (int inv_type = 0;inv_type<itd->getNumCollection();inv_type++){
			cout << "invariance type = " << itd->getType(inv_type) << endl;
			// compute local firing rate with sign = positive
			Matrix localPositive;
			Matrix middlePositive;
			cout << "computing local statistics" << endl;
			computeLocalStatistics(
					inv_type, POSITIVE,   
					networkThresholdsPositive,  
					localPositive, 
					middlePositive);
			// compute local firing rate with sign = negative
			Matrix localNegative;
			Matrix middleNegative;
			cout << "computing local statistics" << endl;
			computeLocalStatistics(
					inv_type, NEGATIVE,  
					networkThresholdsNegative, 
					localNegative, 
					middleNegative);
			// use both local firing rates to compute a firing rate score
			Scalar score = 0;
			cout << "computing threshold" << endl;
			computeThresholdScore(localPositive, 
					middlePositive,
					globalPositive,
					localNegative, 
					middleNegative,
					globalNegative,
					score);
			int category = itd->getCategory(inv_type);
			scorePerCategory[category] = scorePerCategory[category] + score;
			counterPerCategory[category] = counterPerCategory[category] + 1;


			//computeThresholdScore(networkLocalCounts, 
			//middleCounts,
			//networkGlobalCounts,
			//score);
			networkScoreOut = networkScoreOut + score;
			scores.push_back(score);
			counter = counter + 1;
		}
		for (unsigned int i=0;i<scorePerCategory.size();i++){
			scorePerCategory[i] = scorePerCategory[i] / counterPerCategory[i];
		}
		cout << "*all scores ";
		for (unsigned int i=0;i<scores.size();i++){
			cout << scores[i] << " ";
		}
		cout << std::endl;

		networkScoreOut = networkScoreOut /counter;
		templateScoreOut = templateScoreOut /counter;
		maxpoolingScoreOut = maxpoolingScoreOut /counter;
	}


	// Initialize gradient computations by computing global and local statistics, and storing the trajectories

	void InvarianceTestBlock::initializeGradient(int patchWidth, int patchHeight, int inv_type)
	{

		Matrix networkThresholdsPositive;
		Matrix globalPositive;

		network = boost::dynamic_pointer_cast<sdn::AbstractAutoEncoderNetwork> (invNetwork);

		itd = new NaturalTestDataset(patchWidth, patchHeight);

		gradientInvType = inv_type;

		computeGlobalStatistics(POSITIVE, 
				networkThresholdsPositive, 
				globalPositive);

		std::vector<Scalar> scores;

		cout << "invariance type = " << itd->getType(gradientInvType) << endl;

		// compute local firing rate with sign = positive
		Matrix localPositive;

		computeLocalStatistics(
				gradientInvType, POSITIVE,   
				networkThresholdsPositive,  
				localPositive, 
				middlePositive);

	}




	void InvarianceTestBlock::computeGradient(
			Matrix &weightsGradient,
			Matrix &biasGradient)
	{

		weightsGradient = sdn::Matrix::Zero(nHiddenRows * nHiddenCols, patchSize);	
		biasGradient = sdn::Matrix::Zero(nHiddenRows * nHiddenCols, 1);

		// run through each hidden unit
		for (int r=0;r<nHiddenRows;r++){
			for (int c=0;c<nHiddenCols;c++){		

				int rnHc = r*nHiddenCols+c;

				if (middlePositive(r,c) == 0){
			  	  continue;
				}

				
				for (unsigned int k=0; k<trajectoryIndices[rnHc].size(); k++){
	
					// get locally varying data
					std::vector<Matrix> localTrajectory;
					itd->getLocalPatches(gradientInvType, trajectoryIndices[rnHc][k], localTrajectory);

					// calculate activations 

					for (unsigned int i=0; i<localTrajectory.size(); i++){
					
						Scalar temp = (invNetwork->affineEncoder).bias(rnHc, 0);

						Matrix test, test2;
					   
						test = (invNetwork->affineEncoder).weights.row(rnHc);
						
						test2 = test * localTrajectory[i];						

						cout << "testing.." << endl;
						temp = temp + tanh(test2(0,0));
						cout << "done!" << endl;

						weightsGradient.row(rnHc) -= (1.0 - (temp * temp)) * localTrajectory[i];
						biasGradient(rnHc, 0) -= (1.0 - (temp * temp));
					}
				}

				weightsGradient.row(rnHc) /= middlePositive(r, c);
				biasGradient(rnHc, 0) /= middlePositive(r, c);
			}
		}

	}



}


