//Author: Daniel Chia <danchia@stanford.edu>
//File: danchiaSupervisedScript.cc
//Description: 

#include "danchiaSupervisedScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"


namespace sdn {

  class DanchiaSupervisedScriptConfigParser : public ConfigParser {

  protected:
    DanchiaSupervisedScript & parentScript;

  public:
  
    DanchiaSupervisedScriptConfigParser(DanchiaSupervisedScript & parent) : parentScript(parent) {}
    ~DanchiaSupervisedScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::DanchiaSupervisedScript");
    }

    void readConfig(XMLNode & node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::DanchiaSupervisedScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  DanchiaSupervisedScript::DanchiaSupervisedScript() {
    loadUsingConfigManager();
  }
  
  DanchiaSupervisedScript::~DanchiaSupervisedScript() {
  }
  
  void DanchiaSupervisedScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("encodingNetworkFile", "");
    p.addString("supervisedNetworkName");
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    p.addBool("fineTuneUnsupervisedNetwork");
    
    p.parse(node);
    
    // Set Values
    encodingNetworkFile = p.getString("encodingNetworkFile");
    supervisedNetworkName = p.getString("supervisedNetworkName");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    fineTuneUnsupervisedNetwork = p.getBool("fineTuneUnsupervisedNetwork");
  
    // ---------------------------------------------------
    
    // Load AutoEncoder Part
    
		if (encodingNetworkFile != "") {
			std::ifstream fin;

			fin.open(encodingNetworkFile.c_str(), std::ios::binary);
			SDN_ASSERT_MSG(fin.good(), "Unable to open " << encodingNetworkFile);
			
			sdn::serial::BinInArchive ar(fin);
			boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
			fin.close();

			SDN_ASSERT_MSG(net, "Unable to load network from file: " << encodingNetworkFile);

			encodingNetwork = boost::dynamic_pointer_cast<sdn::AbstractAutoEncoderNetwork> (net);

			SDN_ASSERT_MSG(encodingNetwork, encodingNetwork->className() << " is not an autoencoder network!");

			SDN_MSG("Loaded " << encodingNetwork->className() << " as encoder");
		}
		
    // ---------------------------------------------------

    // Load Supervised Part
    
    supervisedNetwork = boost::dynamic_pointer_cast<AbstractSupervisedNetwork> (AbstractNetworkRegistry().make(supervisedNetworkName));
    if (!supervisedNetwork) {
      SDN_ERROR("Could not load network with name " << supervisedNetworkName);
      SDN_ERROR("Is " << supervisedNetworkName << " an AbstractSupervisedNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

		SDN_MSG("Loaded " << supervisedNetwork->className() << " as supervised layer");

    // Deep Loading
    XMLNode supNetworkNode = node.getChildNode("SupervisedNetworkParameters");
    SDN_ASSERT_MSG(!supNetworkNode.isEmpty(), "You need to specify <SupervisedNetworkParameters> !")
    supervisedNetwork->loadXML(supNetworkNode); 

    
  }

  void DanchiaSupervisedScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new DanchiaSupervisedScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();
  }
  
  void DanchiaSupervisedScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
    
    // set the output vector size
    targetOutput = sdn::Matrix::Zero(dataset->numLabels(), 1);
   
  }
  
  void DanchiaSupervisedScript::initPersistentParameters() {
  	SDN_MSG(encodingNetwork->className()); 
    // Load by XML should have setup the networks, here we link them up
    supervisedNetwork->useAutoEncoderNetwork(encodingNetwork);
    
  }
  
  void DanchiaSupervisedScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void DanchiaSupervisedScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // Setup target output
    targetOutput.fill(0.0);
    targetOutput(label, 0) = 1.0;
    
    // Setup input
    normInput.normalizePatch(inputExample);
    
		// Supervised Training
		supervisedNetwork->supervisedTrain(inputExample, targetOutput);
		if (fineTuneUnsupervisedNetwork) {
			supervisedNetwork->fineTuneAutoEncoder(inputExample);
		}
   
  }
  
  // Callbacks ---------------------------------------------------------
  
  void DanchiaSupervisedScript::reportProgress(std::ostream& o) {

  }
  
  const boost::shared_ptr<CostFunction> DanchiaSupervisedScript::getSupervisedCostFunction() {
    return supervisedNetwork->getCostFunction();
  }

  const sdn::Matrix& DanchiaSupervisedScript::getPrediction() {
    // prevent crashes. return something that gives 0 error
    if (supervisedNetwork->getPrediction().rows() != targetOutput.rows() ||
        supervisedNetwork->getPrediction().cols() != targetOutput.cols())
      return targetOutput;
    else
      return supervisedNetwork->getPrediction();
  }
  
  const sdn::Matrix& DanchiaSupervisedScript::getTarget() {
    return targetOutput;
  }
  
  const boost::shared_ptr<CostFunction> DanchiaSupervisedScript::getReconstructionCostFunction() {
    return encodingNetwork->getCostFunction();
  }

  const sdn::Matrix& DanchiaSupervisedScript::getReconstruction() {
    // prevent crashes. return something that gives 0 error
    if (encodingNetwork->getReconstruction().rows() != inputExample.rows() ||
        encodingNetwork->getReconstruction().cols() != inputExample.cols())
      return inputExample;
    else
      return encodingNetwork->getReconstruction();
  }
  
  const sdn::Matrix& DanchiaSupervisedScript::getOriginal() {
    return inputExample;
  }
  
  // ---- Save Model Callback ----
  const void DanchiaSupervisedScript::saveModelInDirectory(std::string dir) {
    encodingNetwork->quickSaveBinary(dir + "encNet_" + itos(manager->currentIteration, 10) + ".dat");
    supervisedNetwork->quickSaveBinary(dir + "supNet_" + itos(manager->currentIteration, 10) + ".dat");
  }
 
 
}
