//Author: Daniel Chia, Zhenghao Chen, Pang Wei Koh
//File: invarianceTweakScript.cc
//Description: Takes in a StandardAutoEncoderNetwork and tries to artificially tweak its invariance score. No callbacks supported.

#include "invarianceTweakScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"
#include "sdn/registry.h"

namespace sdn {

  class InvarianceTweakScriptConfigParser : public ConfigParser {

  protected:
    InvarianceTweakScript & parentScript;

  public:
  
    InvarianceTweakScriptConfigParser(InvarianceTweakScript & parent) : parentScript(parent) {}
    ~InvarianceTweakScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::InvarianceTweakScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::InvarianceTweakScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  InvarianceTweakScript::InvarianceTweakScript() {
    loadUsingConfigManager();
  }
  
  InvarianceTweakScript::~InvarianceTweakScript() {
  }
  
  void InvarianceTweakScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("networkFile");
	 p.addInt("patchHeight");
	 p.addInt("patchWidth");
	 p.addInt("invType");

    p.parse(node);
    
    // Note that values are set *directly* into the network
    networkFile = p.getString("networkFile");
	 patchHeight = p.getInt("patchHeight");
	 patchWidth = p.getInt("patchWidth");
	 invType = p.getInt("invType");
    
    // ---------------------------------------------------
    
    // Load AutoEncoder Part
    
		if (networkFile != "") {
			std::ifstream fin;

			fin.open(networkFile.c_str(), std::ios::binary);
			
			SDN_ASSERT_MSG(fin.good(), "Unable to open " << networkFile);
			
			sdn::serial::BinInArchive ar(fin);
			boost::shared_ptr<sdn::AbstractNetwork> net = sdn::AbstractNetworkRegistry().load(ar);
			fin.close();

			SDN_ASSERT_MSG(net, "Unable to load network from file: " << networkFile);

			origNetwork = boost::dynamic_pointer_cast<sdn::StandardAutoEncoderNetwork> (net);

			SDN_ASSERT_MSG(origNetwork, origNetwork->className() << " is not a standard autoencoder network!");

			SDN_MSG("Loaded " << origNetwork->className() << " as encoder");

			//Copy over original network to our new invariance tweaking network

    		network = boost::dynamic_pointer_cast<sdn::InvarianceTweakingStandardNetwork> (AbstractNetworkRegistry().make("sdn::InvarianceTweakingStandardNetwork"));
		
			network->affineEncoder.weights = origNetwork->affineEncoder.weights;
			network->affineEncoder.bias = origNetwork->affineEncoder.bias;
			network->affineEncoder.learningRate = origNetwork->affineEncoder.learningRate;
			network->affineEncoder.weightDecay = origNetwork->affineEncoder.weightDecay;
			network->affineEncoder.inputSize = origNetwork->affineEncoder.inputSize;
			network->affineEncoder.outputSize = origNetwork->affineEncoder.outputSize;
			
			network->targetActivation = origNetwork->targetActivation;
			network->activationAveragingConstant = origNetwork->activationAveragingConstant; 
			network->learningRateSparsity = origNetwork->learningRateSparsity; 
			network->hiddenLayerSize = origNetwork->hiddenLayerSize;

			network->patchWidth = patchWidth;
			network->patchHeight = patchHeight;
			network->invType = invType;

			network->invTestBlock.invNetwork = network;
			

			SDN_MSG("Copied original network over to a invarianceTweakingStandardNetwork");

			XMLNode invarianceNetworkNode = node.getChildNode("InvarianceNetwork");
    		SDN_ASSERT_MSG(!invarianceNetworkNode.isEmpty(), "You need to specify <InvarianceNetwork> !")
		
			network->loadXML (invarianceNetworkNode);

		}

		// ---------------------------------------------------

  }

  void InvarianceTweakScript::loadUsingConfigManager() {

	  // The following line will load the XML config
	  ConfigListener listener
		  (boost::shared_ptr<ConfigParser>
			(new InvarianceTweakScriptConfigParser(*this)));

	  // We have this here instead of loadXML because loadXML might get
	  // called multiple times by the configParser
	  initPersistentParameters();
	  initTransientParameters();

  }

  void InvarianceTweakScript::initTransientParameters() {
	  /*
		  dataset =  DatasetRegistry().make(datasetName);
		  if (!dataset) {
		  SDN_ERROR("Could not load dataset with name " << datasetName);
		  ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
		  SDN_FATAL("Quitting");
		  }
		  dataset->requestExampleSize(inputHeight, inputWidth);

	  // We are requesting column vector examples
	  inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);

		*/
  }

  void InvarianceTweakScript::initPersistentParameters() {
	  //   network->initTransientParameters();
	  //   network->initPersistentParameters();
  }

  void InvarianceTweakScript::serialize(serial::Archive & ar) {
	  SDN_FATAL("InvarianceTweakScript is not serializable yet");
  }


  void InvarianceTweakScript::executeOneIteration() {

	  SDN_ASSERT(manager);

	  //  int label = dataset->getRandomExampleAsColumn(inputExample);

	  // do fancy stuff here
	  //normInput.normalizePatch(inputExample);
	  network->unsupervisedTrain();

  }

	// ---- Invariance Callback ----

	const boost::shared_ptr<sdn::AbstractAutoEncoderNetwork> InvarianceTweakScript::getNetwork() {
		return network;
	}
 
	// ---- Visualization Callback ----
	const sdn::Matrix& InvarianceTweakScript::getEncoderWeights() {
		return network->affineEncoder.weights;
	}

	const int InvarianceTweakScript::getWeightPatchHeight() {
		return patchHeight;
	}

	const sdn::Matrix& InvarianceTweakScript::getDecoderWeights() {
		// Dummy function
		return network->affineEncoder.weights;
	}

	// ---- Save Model Callback ----
	const void InvarianceTweakScript::saveModelInDirectory(std::string dir) {
		network->quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");
	}


}
