//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: abstractNetwork.h
//Description: Interface for generic network.

#pragma once

#include "sdn/base.h"
#include "abstractBlock.h"
#include <boost/shared_ptr.hpp>

namespace sdn {

  // Forward Declarations
	class CostFunction;  
  namespace serial {
    class Archive;
  }

  // 

  class AbstractNetwork : public AbstractBlock {
  
  public:

    // The key point of this class is to give meaning to "networks"
    // Mainly that networks are classes which can be trained on

    // Intentionally disabled. 
    // Derived classes should implement this with default parameter values;
    // virtual void setConfigurableParameters() = 0;


    /* These are inherited from AbstractBlock 
     
       virtual void loadXML(XMLNode& node) = 0;

       virtual void initTransientParameters() = 0;
       virtual void initPersistentParameters() = 0;

       virtual void serialize(serial::Archive & ar) = 0;
    */
    
    
    virtual void unsupervisedTrain(const sdn::Matrix& example) { 
      SDN_FATAL("Unsupervised Training Not Implemented");
    }
    
    virtual void supervisedTrain(const sdn::Matrix& example, 
                                 const sdn::Matrix& label) { 
      SDN_FATAL("Supervised Training Not Implemented");
    }

		// Since networks are trainable, they must have an associated
		// CostFunction
    virtual const boost::shared_ptr<CostFunction> getCostFunction() = 0;
           
  };
  
  class AbstractAutoEncoderNetwork : public AbstractNetwork {
  
  public:

    virtual void unsupervisedTrain(const sdn::Matrix& example) = 0;
  
    virtual const sdn::Matrix& encode(const sdn::Matrix& input) = 0;
    virtual const sdn::Matrix& decode(const sdn::Matrix& hidden) = 0;

    virtual void encodeBackProp(const sdn::Matrix& input,
                                const sdn::Matrix& hiddenErrorDerivatives,
                                sdn::Matrix& inputErrorDerivatives) = 0;

    virtual void decodeBackProp(const sdn::Matrix& hidden,
                                const sdn::Matrix& outputErrorDerivatives,
                                sdn::Matrix& hiddenErrorDerivatives) = 0;
                                
    virtual void encodeGradientUpdate() = 0;
    virtual void decodeGradientUpdate() = 0;

    virtual const sdn::Matrix& getReconstruction() = 0;
    virtual const sdn::Matrix& getHiddenRepresentation() = 0;
    
    virtual void gradientUpdate() {
      encodeGradientUpdate();
      decodeGradientUpdate();
    }
    
  };
  

  class AbstractSupervisedNetwork : public AbstractNetwork {
  
  public:

    virtual void supervisedTrain(const sdn::Matrix& example, 
                                 const sdn::Matrix& label) = 0;

    // Equivalent to classify
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output) = 0;
    virtual const sdn::Matrix& getPrediction() = 0;

    // Backprops through the supervised layers only
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives) = 0;
    
    virtual void gradientUpdate() = 0;


    virtual void useAutoEncoderNetwork(boost::shared_ptr<AbstractAutoEncoderNetwork> encodingNetwork) = 0;
    // Backprops through the autoencoder + gradient update it
    virtual void fineTuneAutoEncoder(const sdn::Matrix& input) = 0;
    
  };
  
/*
 * 
 * For some reason, this file has started collecting notes (temporarily)
 * 
 * 
 * Here's a short note on parameters
 * 
 * - Configurable Parameters: Model parameters that we set/search for the best
 * - Persistent Parameters: Model parameters we learn (weights, bias, etc)
 * - Transient Parameters: Things that exist temporarily (gradients, error derivatives, activation values)
 * 
 * Program Flow (possible ways):
 * 
 * 1) Set Configurable Parameters -> init the persistent and transient
 *     a) in the code base, you can do this by calling
 *         setConfigurableParameters OR
 *         loadXML
 * 
 *        * It is the job of setConfigurableParameters|loadXML to call the initPersistent and initTransient
 * 
 * 2) Serialize Read In the Configurable and Persistent, then call initTransient
 * 
 */

/*
  =============================== REDESIGN CORE

  serialization
  
  running scripts: 2 methods
  1) load from config  (training scripts etc)
  2) load from command line (visualization etc)

  =============================== READY TO WORK

  documentation, high level concepts
  visualization block, other blocks which are relevant
  unit testing


  callbacks for sparsity, saving for matlab analysis ?? whatelse

  =============================== POST REdeSIGN 

  more scripts:
    - slrun
    - slview ... etc should all be scripts

  examples:
    - config files, out of the box what works
    - how to change code
    - what/where to code,  rules that you should follow
  
       
  organization of code, delete old code:  
     - separation of stable/unstable code
  
  */
}
