//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: script.h
//Description: Scripts can be executed by the SDN app. 
//             A script describes how the network should be setup and trained
//             Also, provides a xml-config interface, saving and loading features


/* Scripts Envisioned
 * 
 * Load Network and Run Invariance Testing
 * Load Network and Run N examples through it for statistics
 * ...
 * 
 * Train Network on X Dataset
 * 
 * 
 * Notes
 * 
 * Scripts can call each other
 * A ScriptRuntimeManager will help to manage scripts + datasets
 */

#pragma once

#include "sdn/base.h"
#include "abstractNetwork.h"

namespace sdn {

  class ScriptRuntimeManager;

  class AbstractScript : public AbstractBlock {
    
  public:
    ScriptRuntimeManager* manager;

    /* These are inherited from AbstractBlock 
     
       virtual void loadXML(XMLNode& node) = 0;

       virtual void initTransientParameters() = 0;
       virtual void initPersistentParameters() = 0;

       virtual void serialize(serial::Archive & ar) = 0;
    */
  
    virtual void loadUsingConfigManager() = 0;
    virtual void executeOneIteration() = 0;
    
    virtual void serialize(serial::Archive & ar) = 0;

    
  };
  
  class AbstractScriptCallback : public AbstractBlock {
    
  public:

    /* These are inherited from AbstractBlock 
     
       virtual void loadXML(XMLNode& node) = 0;

       virtual void initTransientParameters() = 0;
       virtual void initPersistentParameters() = 0;

       virtual void serialize(serial::Archive & ar) = 0;
    */
    
    virtual void execute(ScriptRuntimeManager& scriptManager,
                         AbstractScript& script,
                         unsigned int iteration) = 0;
   

    /**
     * How many steps after execute() until execute() should be called again.
     * Called after a call to execute(). Returns the number of steps which
     * should occur until execute is called again.
     * Typically configured by XML to be enforced by runtime manager.
     * Default is every iteration.
     *
     * @return number of steps to let pass. 0 is never again
     */
    virtual long int itersToNextCall() { return 1; }

    /**
     * Specifies timestep at which callback should first be run.
     * Typically configured by XML to be enforced by runtime manager.
     * Default is from first iteration.
     *
     * @return Timestep when callback is to be first run. 0 indicates beginning
     */
    virtual long int firstIter() { return 0; } 
    
  };
  
  typedef AbstractScript Script;
  typedef AbstractScriptCallback ScriptCallback;
  
}
