//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: sharedBlocks.cc
//Description: Implementation of generic blocks used everywhere

#include "sharedBlocks.h"
#include "sdn/base.h"
#include <cxxabi.h>
#include <Eigen/Array>

namespace sdn {


  void NormalizeInputBlock::normalizePatch(sdn::Matrix& patch) {
    double mean = patch.sum() / (patch.rows() * patch.cols());
    patch.cwise() -= mean;
    double maxAbs = (patch.maxCoeff() > fabs(patch.minCoeff()) ? patch.maxCoeff() : fabs(patch.minCoeff()));
    patch /= (maxAbs+ 1e-6);
    patch *= 0.8;
  }

  void RandomizeWeightMatrix::randomize(sdn::Matrix& weights,
                                        unsigned int inputSize) {
    RandomRealGenerator<double> rrg;
    rrg.setSeed(time(NULL));
    for (int i = 0; i < weights.rows(); i++)
      for (int j = 0; j < weights.cols(); j++)
        weights(i,j) = rrg.randUnif(-1.0,1.0)/sqrt(inputSize);
  }


  // ExponentiallyWeightedMeanBlock ------------------------------------------------

  void ExponentiallyWeightedMeanBlock::setConfigurableParameters(sdn::Scalar defAveragingConstant,
                                                                 sdn::Scalar defInitialValue ,
                                                                 int defTrackingSize)  {
    averagingConstant = defAveragingConstant;
    trackingSize = defTrackingSize;
    initialValue = defInitialValue;

    initPersistentParameters();
    initTransientParameters();
  }
  
  void ExponentiallyWeightedMeanBlock::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addDouble("averagingConstant");
    p.addDouble("initialValue", TRACKER_AUTOSET);
    p.addDouble("trackingSize", SIZE_AUTOSET);
    
    p.parse(node);
    
    averagingConstant = p.getDouble("averagingConstant");
    initialValue = p.getDouble("initialValue");
    trackingSize = p.getDouble("trackingSize");
    
    initPersistentParameters();
    initTransientParameters();

  }

    
  void ExponentiallyWeightedMeanBlock::initTransientParameters() {
    if (trackingSize < 0 || isnan(initialValue)) {
      activations = sdn::Matrix::Zero(1,1);
    } else {
      activations = sdn::Matrix::Constant(trackingSize,1,initialValue);
    }
  }

  
  void ExponentiallyWeightedMeanBlock::initPersistentParameters() {
  }
  
  void ExponentiallyWeightedMeanBlock::serialize(serial::Archive & ar) {
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;
    version = ar.beginBase(this,version,className());
    switch(version) {
      case 1000:
        ar.attribute("averagingConstant", averagingConstant);
        ar.attribute("initialValue", initialValue);
        ar.attribute("trackingSize", trackingSize);
        break;
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }

  void ExponentiallyWeightedMeanBlock::addValues(const sdn::Matrix& values) {
    
    if (trackingSize < 0) {
      trackingSize = values.size();
      // initTransientParameters();
      activations = sdn::Matrix::Constant(values.rows(),values.cols(),initialValue);
    }
    
    if (isnan(initialValue)) {
      initialValue = 0;
      activations = values;
    }
    
    SDN_ASSERT(values.size() == trackingSize);
    
    activations = values * averagingConstant + activations * (1.0 - averagingConstant);
  
  }

  const sdn::Matrix& ExponentiallyWeightedMeanBlock::getMean() {
    return activations;
  }  
  

}
