//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: standardBlocks.cc
//Description: Implementation of blocks used in AutoEncoders

#include "standardBlocks.h"
#include "sdn/base.h"
#include <Eigen/Array>

namespace sdn {

  // ScaleBlock ------------------------------------------------

  void ScaleBlock::setConfigurableParameters(sdn::Scalar defLearningRate) {
    learningRate = defLearningRate;
    initPersistentParameters();
    initTransientParameters();
  }
  
  void ScaleBlock::loadXML(XMLNode& node) {
    AttributeParser p;

    p.addDouble("learningRate");
    p.parse(node);
    learningRate =p.getDouble("learningRate");
        
    initPersistentParameters();
    initTransientParameters();

  }
  
  void ScaleBlock::initTransientParameters()  {
    scaleGradient = 0;
  }
  
  void ScaleBlock::initPersistentParameters() {
    scale = 1.0;
  }
  
  void ScaleBlock::serialize(serial::Archive & ar) {
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
      case 1000:
      
        // Persistent
        ar.attribute("scale", scale);
        
        // Configurable
        ar.attribute("learningRate", learningRate);
    
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }
  
  ScaleBlock::ScaleBlock() {}
  ScaleBlock::~ScaleBlock() {}

  void ScaleBlock::forwardProp(const sdn::Matrix& input,
                           sdn::Matrix& output) {
    output = scale*input;
  }
 
  void ScaleBlock::backProp(const sdn::Matrix& input,
                        const sdn::Matrix& outputErrorDerivatives,
                        sdn::Matrix& inputErrorDerivatives) {
    inputErrorDerivatives = scale * outputErrorDerivatives;
    scaleGradient = (outputErrorDerivatives.cwise() * input).sum() / outputErrorDerivatives.cols();
  }
  
  void ScaleBlock::gradientUpdate() {
    scale -= learningRate * scaleGradient;
  }
  
  // AffineBlock ------------------------------------------------
  
  AffineBlock::AffineBlock() {
  }
  
  AffineBlock::~AffineBlock() {
  }

  void AffineBlock::setConfigurableParameters(sdn::Scalar defLearningRate,
                                              sdn::Scalar defWeightDecay,
                                              int defInputSize,
                                              int defOutputSize) {
    learningRate = defLearningRate;
    weightDecay = defWeightDecay;
    outputSize = defOutputSize;
    inputSize = defInputSize;
    
    initPersistentParameters();
    initTransientParameters();
  }

  void AffineBlock::loadXML(XMLNode& node) {

    AttributeParser p;

    p.addDouble("learningRate");
    p.addDouble("weightDecay", NO_WEIGHT_DECAY);
    p.addInt("outputSize", SIZE_AUTOSET);
    p.addInt("inputSize", SIZE_AUTOSET);
        
    p.parse(node);

    weightDecay = p.getDouble("weightDecay");
    learningRate =p.getDouble("learningRate");
    outputSize = p.getInt("outputSize");
    inputSize = p.getInt("inputSize");
        
    initPersistentParameters();
    initTransientParameters();

  }
  
  void AffineBlock::initTransientParameters() {
    if (outputSize > 0 && inputSize > 0) {
      weightsGradient = sdn::Matrix::Zero(outputSize, inputSize);
      biasGradient = sdn::Matrix::Zero(outputSize, 1);
    }
  }
  
  void AffineBlock::initPersistentParameters() {
    if (outputSize > 0 && inputSize > 0) {
      weights = sdn::Matrix::Zero(outputSize, inputSize);
      bias = sdn::Matrix::Zero(outputSize, 1);

      RandomizeWeightMatrix().randomize(weights, inputSize);
    }
  }
  
  void AffineBlock::serialize(serial::Archive & ar) {
    
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
      case 1000:
      
        // Persistent
        serializeMatrix(ar, weights);
        serializeMatrix(ar, bias);
        
        // Configurable
        ar.attribute("learningRate", learningRate);
        ar.attribute("weightDecay", weightDecay);
        ar.attribute("outputSize", outputSize);
        ar.attribute("inputSize", inputSize);
    
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }
  
  void AffineBlock::forwardProp(const sdn::Matrix& input,
                                sdn::Matrix& output) {

    if (outputSize <= 0 || inputSize <= 0)  {
      if (inputSize <= 0)
        inputSize = input.rows();
      if (outputSize <= 0)
        outputSize = output.rows();
      initPersistentParameters();
      initTransientParameters();
    }

    if (weights.cols() != input.rows()) {
      SDN_FATAL(className() << " weight matrix #columns does not match input #rows");
    }

    // Multiple examples, hence, need to use ones * bias
    if (ones.cols() != input.cols())
      ones = Matrix::Ones(1, input.cols());
    
    output = weights * input + bias * ones;
    
  }
 
  void AffineBlock::backProp(const sdn::Matrix& input,
                             const sdn::Matrix& outputDerivatives,
                             sdn::Matrix& inputDerivatives) {
    
    if (outputSize <= 0 || inputSize <= 0)  {
      inputSize = input.size();
      outputSize = outputDerivatives.size();
      initTransientParameters();
      initPersistentParameters();
    } 
    
    inputDerivatives = weights.transpose() * outputDerivatives;
    weightsGradient = outputDerivatives * input.transpose();
    biasGradient    = outputDerivatives;
    
    if (outputDerivatives.cols() > 1) {
      // Handle Multiple Examples
      weightsGradient *= 1.0 / outputDerivatives.cols();
      biasGradient = biasGradient.rowwise().sum() / biasGradient.cols();
    }
    
  }
  
  void AffineBlock::gradientUpdate() {
    
    // If weightDecay enabled
    if (weightDecay > 0) 
      weights *= (1.0 - learningRate * weightDecay);  
        
    // Assume Minimizing the Objective
    weights -= learningRate * weightsGradient;    
    bias    -= learningRate * biasGradient;
    
  }


  
}
