//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: standardBlocks.h
//Description: Implementation of blocks used in AutoEncoders

#pragma once

#include "sdn/base.h"
#include "../math/activationFunction.h"
#include "../defaultParameters.h"
#include "../abstractBlock.h"
#include "sharedBlocks.h"
#include <Eigen/Array>

namespace sdn {


  /// ScaleBlock class
  /** Expects input to be a matrix of column vector only
   *  each column vector is one example
  */
  class ScaleBlock : public ComposableBlock {
  
  public:
  
    virtual void setConfigurableParameters(sdn::Scalar defLearningRate);
    virtual void loadXML(XMLNode& node);
    
    virtual void initTransientParameters() ;
    virtual void initPersistentParameters();
    
    virtual void serialize(serial::Archive & ar);
    
    ScaleBlock();
    ~ScaleBlock();

    // Persistent Parameters
    sdn::Scalar scale;
    
    // Configurable Parameters
    sdn::Scalar learningRate;
    
    // Transient Parameters
    sdn::Scalar scaleGradient;
    
    // Operations
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output);
   
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives);
    
    virtual void gradientUpdate();
    
  };
  


  /// AffineBlock class
  /** Expects input to be a matrix of column vector only
   *  each column vector is one example
  */
  class AffineBlock : public ComposableBlock {
  
  public:
  
    virtual void setConfigurableParameters(sdn::Scalar defLearningRate,
                                           sdn::Scalar defWeightDecay,
                                           int defInputSize,
                                           int defOutputSize);
    virtual void loadXML(XMLNode& node);
    
    virtual void initTransientParameters() ;
    virtual void initPersistentParameters();
    
    virtual void serialize(serial::Archive & ar);
    
    AffineBlock();
    ~AffineBlock();

    // Persistent Parameters
    sdn::Matrix weights;                // Each row is a feature detector
    sdn::Matrix bias;
    
    // Configurable Parameters
    sdn::Scalar learningRate;
    sdn::Scalar weightDecay;
    int outputSize;
    int inputSize;
    
    // Transient Parameters
    sdn::Matrix weightsGradient;
    sdn::Matrix biasGradient;
    sdn::Matrix ones; 
    
    // Operations
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output);
   
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives);
    
    virtual void gradientUpdate();
    
  };
  
 
  /// ActFunctionBlock class
  /** Takes any activation function and makes a block out of it
  */
  template <class ActivationFunction> 
  class ActFunctionBlock : public ComposableBlock {
  
  public:

    ActivationFunction actFunc;

    virtual void setConfigurableParameters() {}
    virtual void loadXML(XMLNode& node) {}
  
    virtual void initTransientParameters() {}
    virtual void initPersistentParameters() {}
    
    virtual void serialize(serial::Archive & ar) {
      SDN_DEBUG(className() << "::" << __FUNCTION__);
      int version = 1000;
      version = ar.beginBase(this,version,className());
      switch(version) {
        case 1000:
          actFunc.serialize(ar);
          break;
        default:
          SDN_FATAL("Unrecognized " << className() << " version");
        }
      ar.endBase();
      // Input Case: Initialize Transient
      if (ar.in()) {
        initTransientParameters();
      }
    }
    
    // Persistent Parameters
    // Transient Parameters
    sdn::Matrix backPropDerv;
        
    // Operations
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output) {
      actFunc.output(input, output);
    }
   
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives) {
      
      // Assumes each column is an example
      if (   inputErrorDerivatives.rows() != outputErrorDerivatives.rows() 
          || inputErrorDerivatives.cols() != outputErrorDerivatives.cols()) {
        inputErrorDerivatives = Matrix::Zero(outputErrorDerivatives.rows(), 
                                             outputErrorDerivatives.cols());
      }
      
      if (actFunc.diagonalOptimizationSupported()) {
        actFunc.outputDerivative(input,
                                 inputErrorDerivatives,
                                 true);
        inputErrorDerivatives.cwise() *= outputErrorDerivatives;
      } else {
        for (int i=0;i<input.cols();i++) {
          
          actFunc.outputDerivative(input.col(i), 
                                   backPropDerv,
                                   true);
                                   
          // For certain activation functions (e.g. SoftMax),
          // the gradients between each pair of values is non-zero
          // so we have to handle all of those 
          inputErrorDerivatives.col(i) = backPropDerv.transpose() *
                                         outputErrorDerivatives.col(i);
        }
      }

    }
    
    virtual void gradientUpdate() {}
    
    virtual sdn::Matrix getBiasInitialParameters(sdn::Scalar targetActivation,
                                          unsigned int layerSize) {
      if (actFunc.hasInverse())
        return Matrix::Constant(layerSize, 1, 
                                actFunc.inverse(
                                 (actFunc.max() - actFunc.min()) *
                                 targetActivation +
                                  actFunc.min()));
      else 
        return Matrix::Zero(layerSize, 1);
    }
    
    virtual sdn::Matrix normalizeActivationValues(sdn::Matrix& actValues) {
      return (actValues.cwise() - actFunc.min()) / (actFunc.max() - actFunc.min());
    }
  };
  
  /// Define Common Activation Function Blocks
  class TanhBlock : public ActFunctionBlock<TanhActivationFunction> {};
  class SoftmaxBlock : public ActFunctionBlock<SoftmaxActivationFunction>  {};
  class SquareBlock : public ActFunctionBlock <SquareActivationFunction> {};
   
 
  /// Composable "StandardLayer" AffineAndTransferBlock Block
  /** 
   */
  template <class ActivationFunction> 
  class AffineAndTransferBlock : public ComposableBlock {
    
  public:
  
    // Persistent Parameters
    AffineBlock affineB;
    ActFunctionBlock<ActivationFunction> activationB;
    
    // Transient
    sdn::Matrix affineRepresentation;
    sdn::Matrix affineErrorDerivatives;
  
    virtual void setConfigurableParameters(sdn::Scalar defLearningRate,
                                           sdn::Scalar defWeightDecay,
                                           int defInputSize,
                                           int defOutputSize) {
      affineB.setConfigurableParameters(defLearningRate, defWeightDecay, defInputSize, defOutputSize);
      initTransientParameters();
      initPersistentParameters();
    }
    
    virtual void loadXML(XMLNode& node) {
      affineB.loadXML(node);
      initTransientParameters();
      initPersistentParameters();
    }
    
    virtual void initTransientParameters() {}
    
    virtual void initPersistentParameters() {
      affineB.initPersistentParameters();
      activationB.initPersistentParameters();
    }
    
    virtual void serialize(serial::Archive & ar) {

      SDN_DEBUG(className() << "::" << __FUNCTION__);
      int version = 1000;
      version = ar.beginBase(this,version,className());
      switch(version) {
        case 1000:
          affineB.serialize(ar);
          activationB.serialize(ar);
          break;
        default:
          SDN_FATAL("Unrecognized " << className() << " version");
        }
      ar.endBase();
      // Input Case: Initialize Transient
      if (ar.in()) {
        initTransientParameters();
      }

    }
    
    AffineAndTransferBlock() {};
    ~AffineAndTransferBlock() {}
    
    // Operations
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output) {
      affineB.forwardProp(input, affineRepresentation);
      activationB.forwardProp(affineRepresentation, output);
    }
   
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives) {
      activationB.backProp(affineRepresentation, outputErrorDerivatives, affineErrorDerivatives);
      affineB.backProp(input, affineErrorDerivatives, inputErrorDerivatives);
    }
    
    virtual void gradientUpdate() {
      activationB.gradientUpdate();
      affineB.gradientUpdate();
    }
    
  };

  /// Define Common StandardLayer Blocks
  // We use a class here so that the auto naming code will work fine with it
  class AffineTanhBlock: public AffineAndTransferBlock<TanhActivationFunction> {};
  class AffineSoftmaxBlock: public AffineAndTransferBlock<SoftmaxActivationFunction> {};

  /// SparseBiasUpdate AffineAndTransferBlock
  /** This block implements two important ideas:
   *  1) the bias update hack rule over the AffineAndTransferBlock 
   *  2) good methods to initialize the weights/bias 
   */
  template <class ActivationFunction> 
  class SparseBiasUpdateAffineAndTransferBlock : public ComposableBlock {
    
  public:
  
    // Persistent Parameters
    AffineAndTransferBlock<ActivationFunction> transferBlock;
    
    // Configurable
    double activationAveragingConstant;
    double learningRateSparsity;
    double targetActivation;
    
    // Transient
    ExponentiallyWeightedMeanBlock activationTracker;
    sdn::Matrix sparsityError;
  
    virtual void setConfigurableParameters(sdn::Scalar defTargetActivation,
                                           sdn::Scalar defActivationAveragingConstant,
                                           sdn::Scalar defLearningRateSparsity,
                                           sdn::Scalar defLearningRate,
                                           sdn::Scalar defWeightDecay,
                                           int defInputSize,
                                           int defOutputSize) {
      learningRateSparsity = defLearningRateSparsity;
      activationAveragingConstant = defActivationAveragingConstant;
      transferBlock.setConfigurableParameters(defLearningRate, defWeightDecay, defInputSize, defOutputSize);
      initTransientParameters();
      initPersistentParameters();
    }
    
    virtual void loadXML(XMLNode& node) {
      
      AttributeParser p;
      
      // Ignore the Config for the transferBlock
      p.ignoreDouble("learningRate");
      p.ignoreDouble("weightDecay");
      p.ignoreInt("outputSize");
      p.ignoreInt("inputSize");

      p.addDouble("targetActivation");
      p.addDouble("learningRateSparsity");
      p.addDouble("activationAveragingConstant");
      
      p.parse(node);
      
      targetActivation = p.getDouble("targetActivation");
      learningRateSparsity = p.getDouble("learningRateSparsity");
      activationAveragingConstant = p.getDouble("activationAveragingConstant");

      // The parameters for the included block must be in a child node
      // Since we only include one block, we demand that there's
      // exactly 1 child node
      // 
      // In general, you can use getChildNode("nodename") in place of 
      // just selecting one node.
      //
      // We use getChildNode(0) because  this template allows for many
      // variations of the included block (by varying the activation
      // function)
      
      if (node.nChildNode() <= 0) {
        SDN_FATAL("Node " << (node.getName() ? node.getName() : "unknown") << " needs a child node to specify the Affine/Tanh Parameters");
      } else if (node.nChildNode() > 1){
        SDN_FATAL("Node " << (node.getName() ? node.getName() : "unknown") << " has too many children");
      } else {
        // Pass the config down to the transferBlock
        XMLNode transferBlockNode = node.getChildNode(0);
        transferBlock.loadXML(transferBlockNode);
      }
     
      initTransientParameters();
      initPersistentParameters();
      
    }
    
    virtual void initTransientParameters() {}
    virtual void initPersistentParameters() {
      transferBlock.initPersistentParameters();
      activationTracker.setConfigurableParameters(activationAveragingConstant,
                                                  targetActivation,
                                                  SIZE_AUTOSET);

      if (transferBlock.affineB.inputSize > 0 && transferBlock.affineB.outputSize > 0 ) {
        transferBlock.affineB.bias = transferBlock.activationB.getBiasInitialParameters(targetActivation, transferBlock.affineB.outputSize);
      }
    }
    
    virtual void serialize(serial::Archive & ar) {
      SDN_DEBUG(className() << "::" << __FUNCTION__);
      int version = 1000;
      version = ar.beginBase(this,version,className());
      switch(version) {
        case 1000:
          ar.attribute("activationAveragingConstant", activationAveragingConstant);
          ar.attribute("learningRateSparsity", learningRateSparsity);
          ar.attribute("targetActivation", targetActivation);
          transferBlock.serialize(ar);
          break;
        default:
          SDN_FATAL("Unrecognized " << className() << " version");
        }
      ar.endBase();
      // Input Case: Initialize Transient
      if (ar.in()) {
        initTransientParameters();
      }
    }
    
    SparseBiasUpdateAffineAndTransferBlock() {}
    ~SparseBiasUpdateAffineAndTransferBlock() {}
    
    // Operations
    virtual void forwardProp(const sdn::Matrix& input,
                             sdn::Matrix& output) {
                               
      transferBlock.forwardProp(input, output);
      
      // MULTIPLE EXAMPLE NOTE: We might need to average output before addint to the tracker
      activationTracker.addValues( transferBlock.activationB.normalizeActivationValues(output) );
      
    }
   
    virtual void backProp(const sdn::Matrix& input,
                          const sdn::Matrix& outputErrorDerivatives,
                          sdn::Matrix& inputErrorDerivatives) {
      transferBlock.backProp(input, outputErrorDerivatives, inputErrorDerivatives);
    }
    
    virtual void gradientUpdate() {
      
      transferBlock.gradientUpdate();
      
      sparsityError = activationTracker.getMean().cwise() - targetActivation;
      transferBlock.affineB.bias -= transferBlock.affineB.learningRate*learningRateSparsity*sparsityError;
      
    }
    
  };
  
  /// Define Common StandardLayer Blocks
  // We use a class here so that the auto naming code will work fine with it
  class SparseAffineTanhBlock : public SparseBiasUpdateAffineAndTransferBlock<TanhActivationFunction> {};
  
  
}
