// Copyright 2009, All Rights Reserved.
// Authors: Andrew Saxe (asaxe@stanford.edu)
//          Ian Goodfellow (ia3n@cs.stanford.edu)

/** \file common-activation-functions.h
    \brief  Defines a number of common activation functions.
*/

#include "activationFunction.h"
#include <math.h>
#include <assert.h>
#include <limits>
#include "sdn/base.h"
#include <Eigen/Array>

using namespace sdn;

//Implemented as a macro rather than an inline function to make sure
//Eigen gets to work its lazy evaluation magic as well as possible
#define SDN_DIAGONAL_JACOBIAN( expr ) if (allowDiagonalOptimization) \
    result = (expr); \
  else \
    result = ((expr).col(0)).asDiagonal(); 

Matrix ActivationFunction::output(const Matrix & x)
{
  SDN_DEBUG("Someone's using the slow version of ActivationFunction::output!");
  Matrix temp;
  output(x,temp);
  return temp;
}

Scalar ActivationFunction::inverse(Scalar x) {
  SDN_FATAL("Inverse not implemented");
  return -1;
}


void LinearActivationFunction::output(const Matrix& x, Matrix & result) {
  result = x;
}

void LinearActivationFunction::output(const Vector & x, Vector & result)
{
  result = x;
}

void LinearActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
  SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));
  SDN_DIAGONAL_JACOBIAN(Vector::Ones(x.rows()));
}

bool LinearActivationFunction::diagonalOptimizationSupported() {
  return true;
}


Scalar LinearActivationFunction::max() {
  return std::numeric_limits<Scalar>::max();
}

Scalar LinearActivationFunction::min() {
  return std::numeric_limits<Scalar>::min();
}

void LogisticActivationFunction::output(const Matrix& x, Matrix & result) {
  result = (1+(-x).cwise().exp().cwise()).cwise().inverse();
}

void LogisticActivationFunction::output(const Vector& x, Vector & result) {
  result = (1+(-x).cwise().exp().cwise()).cwise().inverse();
}

void LogisticActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
  SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));
  output(x, o_);

  SDN_DIAGONAL_JACOBIAN(o_.cwise()*((-o_).cwise()+1))
}

bool LogisticActivationFunction::diagonalOptimizationSupported() {
  return true;
}

Scalar LogisticActivationFunction::max() {
  return 1.0;
}

Scalar LogisticActivationFunction::min() {
  return 0.0;
}


void TanhActivationFunction::output(const Matrix& x,
Matrix & result) {
  result.resize(x.rows(),x.cols());

  for(int r = 0; r < x.rows(); r++)
    for(int c = 0; c < x.cols(); c++)
      result(r,c) = tanh(x(r,c));

  //SDN_MSG("Input: " << x);
  //SDN_MSG("Output: "<< result);

}

void TanhActivationFunction::output(const Vector & x,
				    Vector & result) 
{
  result.resize(x.size());
  
  for (int i = 0; i < x.size(); i++)
    result(i) = tanh(x(i));
}

void TanhActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
 SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));
 output(x, o_);
 SDN_DIAGONAL_JACOBIAN((-(o_.cwise().square())).cwise() + 1);
}

bool TanhActivationFunction::diagonalOptimizationSupported() {
  return true;
}

Scalar TanhActivationFunction::max() {
  return 1.0;
}

Scalar TanhActivationFunction::min() {
  return -1.0;
}

Scalar TanhActivationFunction::inverse(Scalar x) {
  return atanh(x);
}

void SoftSignActivationFunction::output(const Matrix& x, Matrix & result) {
  result =  x.cwise()/(1+x.cwise().abs().cwise());
}

void SoftSignActivationFunction::output(const Vector& x, Vector & result) {
  result = x.cwise()/(1+x.cwise().abs().cwise());
}

void SoftSignActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
  SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));

  SDN_DIAGONAL_JACOBIAN ((x.cwise().abs().cwise()  + 1).cwise().square().cwise().inverse());
}

bool SoftSignActivationFunction::diagonalOptimizationSupported() {
  return true;
}

Scalar SoftSignActivationFunction::max() {
  return 1.0;
}

Scalar SoftSignActivationFunction::min() {
  return -1.0;
}

void SoftmaxActivationFunction::output(const Matrix& x, Matrix & result) {

 result = x.cwise().exp();
 normalizationConstants_ = result.colwise().sum();
  
  for(int c = 0; c < result.cols(); c++)
    result.col(c) /= normalizationConstants_(c);
}

void SoftmaxActivationFunction::output(const Vector & x, Vector & result)
{
  result = x.cwise().exp();
  
  result /= result.sum();
}


void SoftmaxActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
  SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));
  output(x.col(0), o_); // Ok, x must be a vector in softmax
  result =  o_.asDiagonal()-o_*o_.transpose();
}

bool SoftmaxActivationFunction::diagonalOptimizationSupported() {
  return false;
}

ActivationFunction::~ActivationFunction()
{

}

LinearActivationFunction::~LinearActivationFunction()
{

}

LogisticActivationFunction::~LogisticActivationFunction()
{

}


TanhActivationFunction::~TanhActivationFunction()
{

}

SoftSignActivationFunction::~SoftSignActivationFunction()
{

}


SoftmaxActivationFunction::~SoftmaxActivationFunction()
{

}


Scalar SoftmaxActivationFunction::max() {
  return 1.0;
}

Scalar SoftmaxActivationFunction::min() {
  return 0;
}


Registry<ActivationFunction> & sdn::ActivationFunctionRegistry()
{
  static Registry<ActivationFunction> registry("sdn::ActivationFunction");
  return registry;
}

void ActivationFunction::serialize(serial::Archive & ar)
{
  int version = 3;

  version =   ar.beginBase(this, version, "sdn::ActivationFunction");

  switch (version)
    {
    case 3:
      ar.endBase();
      break;
    default:
      SDN_FATAL("Invalid ActivationFunction version requested");
    }
}

void TanhActivationFunction::serialize(serial::Archive & ar)
{
  int version = 5;

  version = ar.beginDerived(this, version, "sdn::TanhActivationFunction");

  switch(version)
    {
    case 5:
      ar.beginSuper();
      ActivationFunction::serialize(ar);
      ar.endSuper();
      ar.endDerived();
      break;
    default:
      SDN_FATAL("Invalid TanhActivationFunction version requested");
    }
}

SDN_AUTOREGISTER(ActivationFunction,TanhActivationFunction);

void SquareActivationFunction::output(const Matrix& x, Matrix & result) {
  result = x.cwise().square();
}

void SquareActivationFunction::output(const Vector & x, Vector & result)
{
  result = x.cwise().square();
}

Scalar SquareActivationFunction::inverse(const Scalar x){
  return sqrt(x);//x.cwise().sqrt();
}

void SquareActivationFunction::outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization) {
  SDN_ASSERT(x.cols() == 1 || (allowDiagonalOptimization && diagonalOptimizationSupported()));
  //SDN_DIAGONAL_JACOBIAN(Vector::Ones(x.rows()));
  result = x*2;
}

bool SquareActivationFunction::diagonalOptimizationSupported() {
  return false;
}


Scalar SquareActivationFunction::max() {
  return std::numeric_limits<Scalar>::max();
}

Scalar SquareActivationFunction::min() {
  return std::numeric_limits<Scalar>::min();
}

SquareActivationFunction::~SquareActivationFunction() {
}
