// Copyright 2009, All Rights Reserved.
// Authors: Andrew Saxe (asaxe@stanford.edu)
//          Ian Goodfellow (ia3n@cs.stanford.edu)                                                                                
/** \file commonActivationFunctions.h
    \brief  Declares a number of common activation functions.
*/

#pragma once

#include "sdn/base.h"

namespace sdn {

template <typename T>
class Registry;

namespace serial
{
	class Archive;
}

class ActivationFunction {
 public:

  /** 
   Each column represents one example. Due to the way
   outputDerivative is invoked, the same vector-to-vector
   function must be applied to each column. 
*/
  virtual Matrix output(const Matrix& x);
  
  /**

  \param result Need not be correctly sized in advance
  */
  virtual void output(const Matrix & x, Matrix & result) = 0;

  virtual void output(const Vector & x, Vector & result) = 0;



  ///Should return the Jacobian.
/** 
    \param x must be a column vector. 
      \param result The return value of the function will be placed in this matrix.
	Need not be correctly sized in advance (though it isn't advisable to call
	it with a wrong-sized matrix in a tight loop, as this will waste time
	continually allocating/deallocating memory).

    \param allowDiagonalOptimiziation If true, diagonal matrices will be returned as 1-dimensional, and it will be up to the client to make them into square
    matrices
  */
  virtual void outputDerivative(const Matrix& x, Matrix & result, bool allowDiagonalOptimization = false) = 0;

  virtual ~ActivationFunction();
  ///Returns the maximum possible value returned by Output.
  virtual Scalar max() = 0;
  ///Returns the minimum possible value returned by Output.
  virtual Scalar min() = 0;

  /// Returns the inverse of the activation function
  virtual Scalar inverse(Scalar x);

  virtual bool hasInverse() const { return false;}

  virtual bool diagonalOptimizationSupported() { return false; }

  virtual void serialize(serial::Archive & ar);


};


/// Implements a linear activation function.
/**
   The linear function is \f[f(x) = x\f].
   Suitable for Gaussian distributions.
 */

class LinearActivationFunction : public ActivationFunction {
 public: 
  void output(const Matrix& x, Matrix & result);
  void output(const Vector& x, Vector & result);
  void  outputDerivative(const Matrix& x, Matrix & result,
 bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~LinearActivationFunction();
  Scalar max();
  Scalar min();

};


/// Implements a logistic activation function.
/**
   The logistic function is \f[f(x)= \frac{1}{1+e^{-x}} \f].
   Suitable for Bernoulli distributions.
 */

class LogisticActivationFunction : public ActivationFunction {
 public:
  void output(const Matrix& x, Matrix & result);
  void output(const Vector& x, Vector & result);

  void outputDerivative(const Matrix& x, Matrix & result,
bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~LogisticActivationFunction();
  Scalar max();
  Scalar min();
 protected:
  Matrix o_;

};


/// Implements a tanh activation function.
/**
   The tanh function is \f[ f(x) = tanh(x)\f].
 */
class TanhActivationFunction : public ActivationFunction {
 public:
  void output(const Matrix& x, Matrix & result);
  void output(const Vector&x, Vector & result);

  /** 
 \todo this calls output, even though in the context of backprop
  the results of output should always be available already. in
  order to prevent choking on constant allocation / deallocation
  of memory, I had to made a cached temporary o_. On GPU this will
  waste too much space, so we should make an alternate method that
  assumes output is available in the storage layer. likewise, we
  also cache result in singleLayerNetwork, which could probably be
  avoided.
  */
  void outputDerivative(const Matrix& x, Matrix & result,
bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~TanhActivationFunction();
  Scalar max();
  Scalar min();

  Scalar inverse(Scalar x);
  bool hasInverse() const { return true;}
  bool hasOutput() const { return true; }

  virtual void serialize(serial::Archive & ar);

 protected:
  Matrix o_;

};

/// Implements a soft sign activation function.
/**
   The soft sign function is \f[ f(x) = \frac{x}{1+|x|}\f].
 */

class SoftSignActivationFunction : public ActivationFunction {
 public:
  void output(const Matrix& x, Matrix & result);
  void output(const Vector& x, Vector & result);
  void outputDerivative(const Matrix& x, Matrix & result,
bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~SoftSignActivationFunction();
  Scalar max();
  Scalar min();

};


/// Implements a softmax activation function.
 
/**
   The softmax function is \f[f_i(x) = \frac{e^{x_i}}{\sum_j e^{x_j}} \f].
   Suitable for multinomial distributions.
  
   This only supports a multinomial distribution over a single variable.
 */

 
class SoftmaxActivationFunction : public ActivationFunction {
 public:
  void output(const Matrix& x, Matrix & result);
  void output(const Vector& x, Vector & result);
  void outputDerivative(const Matrix& x, Matrix & result,
bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~SoftmaxActivationFunction();
  Scalar max();
  Scalar min();
 protected:
   Vector normalizationConstants_;
   Vector o_;


};
///Implements a squared activation function

class SquareActivationFunction : public ActivationFunction {
 public: 
  void output(const Matrix& x, Matrix & result);
  void output(const Vector& x, Vector & result);	
  void  outputDerivative(const Matrix& x, Matrix & result,
  bool allowDiagonalOptimization);
  virtual bool diagonalOptimizationSupported();
  virtual ~SquareActivationFunction();
  Scalar max();
  Scalar min();
  Scalar inverse (const Scalar x);

};




 Registry<ActivationFunction> & ActivationFunctionRegistry();

}

//SDN_AUTOREGISTER_H(ActivationFunction, TanhActivationFunction);

