// Copyright 2009, All Rights Reserved.
/// \author Andrew Saxe (asaxe@stanford.edu)
/// \author Ian Goodfellow                                                                                
/** 
    \file costFunction.h
    \brief  Declares a number of common error functions.
*/

#ifndef CORE__COMMON_COST_FUNCTIONS_H_
#define CORE__COMMON_COST_FUNCTIONS_H_

#include "sdn/base.h"
#include "objectiveTerm.h"
#include <boost/shared_ptr.hpp>

namespace sdn
{
  namespace serial
  {
	class Archive;
  }

  template <typename T> class Registry;


  /** \todo remove the actual/desired terminology and call it
      BinaryObjectiveTerm
      
      we may want to use this to implement sparsity as a function
      of a matrix estimating firing rate and a matrix for the biases
 */

  /// An ObjectiveTerm that is a function of two matrices
  class CostFunction : public ObjectiveTerm  {
 public:



    /// Returns the output of the function (must be a column vector)
    virtual Matrix output(const Matrix & actual, const Matrix & desired)
    {
      return lambda * output_(actual, desired);
    }

    /// Returns the Jacobian (only makes sense if input is vector-valued)
    virtual Matrix jacobian(const Matrix & actual, const Matrix & desired)
    {
      return lambda * jacobian_(actual, desired);
    }

    // Returns the derivatives with respect to the input matrix
    virtual Matrix matrixDerivative(const Matrix & actual, const Matrix & desired)
    {
      return lambda * matrixDerivative_(actual,desired);
    }
    
    /// \todo-- why does the serialization code not call SumOfSquaresCostFunction::serialize if there is a CostFunction::serialize? Is there something funny with the declaration of these classes that is screwing up the vtable? 
    virtual void serializeAbstract(serial::Archive & ar);
    virtual void serialize(serial::Archive & ar) = 0;

    virtual ~CostFunction();
    
    void registerActual(boost::shared_ptr<Matrix> actual)
    {
      actual_ = actual;
    }
    
    void registerDesired(boost::shared_ptr<Matrix> desired)
    {
      desired_ = desired;
    }
    
  protected:
    
    boost::shared_ptr<Matrix> actual_;
    boost::shared_ptr<Matrix> desired_;
    
    
    Matrix output_()
      {
	return lambda * output_(*actual_, *desired_);
      }
    
    Matrix jacobian_()
      {
	return lambda * jacobian_(*actual_, *desired_);
      }
    
    Matrix matrixDerivative_()
      {
	return lambda * matrixDerivative_(*actual_, *desired_);
      }
    
    
    /// Returns the output of the function (must be a column vector)
    virtual Matrix output_(const Matrix & actual, const Matrix & desired) = 0;
    
    /// Returns the Jacobian (only makes sense if input is vector-valued)
    virtual Matrix jacobian_(const Matrix & actual, const Matrix & desired)
    {
      SDN_FATAL("This objective term does not compute a Jacobian");
      return Matrix();
    }
    
    // Returns the derivatives with respect to the input matrix
    virtual Matrix matrixDerivative_(const Matrix & actual, const Matrix & desired)
    {
      SDN_FATAL("This objective term does not compute matrix derivatives");
      return Matrix();
    }
  };


  /// Implements a sum of squares cost function.
  /**
     The sum of squares cost function is \f[f(x,t) = \frac{1}{2}||x-t||^2\f].
     Suitable for Gaussian distributions.
    */
  class SumOfSquaresCostFunction : public CostFunction {
  public: 
    Matrix output_(const Matrix& actual, 
			    const Matrix& desired);
    Matrix jacobian_(const Matrix& actual, 
			      const Matrix& desired);
    
    virtual ~SumOfSquaresCostFunction();
    
    virtual void serialize(serial::Archive & ar);
    
  };
  
  /// Implements a CrossEntropyCostFunction function.
  /**
     Suitable for Classification.
    */
  class CrossEntropyCostFunction : public CostFunction {
  public: 
    Matrix output_(const Matrix& actual, 
			    const Matrix& desired);
    Matrix jacobian_(const Matrix& actual, 
			      const Matrix& desired);
    
    virtual ~CrossEntropyCostFunction();
    
    virtual void serialize(serial::Archive & ar);
    
  };
  
  Registry<CostFunction> & CostFunctionRegistry();
  
}

//SDN_AUTOREGISTER_H(CostFunction, SumOfSquaresCostFunction);

#endif //CORE__COMMON_COST_FUNCTIONS_H
