/// \author Ian Goodfellow (Ia3n@cs.stanford.edu)
/// \file objectiveTerm.h
/// \brief Abstract base class for a term of the objective function


#pragma once

#include "sdn/base.h"

namespace sdn
{
  
  /// \todo-- do we really need to support vector-value objectives? this would all be a lot cleaner if we just did scalar valued objectives and used matrix derivatives rather than the Jacobian everywhere
  class ObjectiveTerm
  {
    
  public:
    
    /// the term is multiplied by this value
    double lambda;
    
    
 public:
    
    ObjectiveTerm(double scalar = 1.0)
      {
	lambda = scalar;
      }
    
    virtual ~ObjectiveTerm()
      {
      }
    
    
    /// Returns the output of the function . May only have one column
    virtual Matrix output()
    {
      /// \todo-- speed this up by returning an Eigen expression, or does it work that way?
      return lambda * output_();
    }
    
    /// Returns the Jacobian. Only makes sense if function has vector-valued input
    virtual Matrix jacobian()
    {
      return lambda * jacobian_();
    }

    /// Returns the derivatives with respect to the input matrix (only makes sense if the output is scalar)
    virtual Matrix matrixDerivative()
    {
      return lambda * matrixDerivative_();
    }
    

    void serialize(serial::Archive& ar) {
     
      SDN_DEBUG("ObjectiveTerm::Serialize");

      int version = 211;

      version = ar.beginBase(this,version,"sdn::ObjectiveTerm");
  
      switch (version)
	{
	case 211:
	  ar.attribute("lambda",lambda);
	  ar.endBase();
	  break;
	default:
	  SDN_FATAL("Unrecognized ObjectiveTerm version");
	}
    }
  protected:
    
    ///Computes the output before multiplication by $\lambda$
    virtual Matrix output_() = 0;
    
    //Computes the Jacobian before multiplication by $\lambda$
    virtual Matrix jacobian_()
    {
      SDN_FATAL("This objective term does not compute a Jacobian");
      return Matrix();
    }

    //Computes the matrix derivative before multiplication by $\lambda$
    virtual Matrix matrixDerivative_()
    {
      SDN_FATAL("This objective term does not compute a matrix derivative.");
      return Matrix();
    }
    
    
    
  };
  
}
