/// \author Ian Goodfellow
/// \file unaryObjectiveTerm.h
/// \brief Abstract base class for objective terms that are a function of a single matrix

#pragma once
#include "objectiveTerm.h"
#include <string>
#include <boost/shared_ptr.hpp>

namespace sdn
{

  template <typename T> class Registry;


  /// An ObjectiveTerm that is a function of a single matrix

  class UnaryObjectiveTerm : public ObjectiveTerm
  {
  public:
    
    static std::string getClassName() { return "sdn::UnaryObjectiveTerm"; }

    /// Returns the output of the function (must be a column vector)
    virtual Matrix output(const Matrix & m) = 0;

    /// Returns the Jacobian (only makes sense if input is vector-valued)
    virtual Matrix jacobian(const Matrix & m)
    {
      SDN_FATAL("This objective term does not compute a Jacobian");
      return Matrix();
    }

    // Returns the derivatives with respect to the input matrix (only makes sense if output is scalar)
    virtual Matrix matrixDerivative(const Matrix & m)
    {
      SDN_FATAL("This objective term does not compute matrix derivatives");
      return Matrix();
    }


    virtual void serialize(serial::Archive & ar);

    virtual ~UnaryObjectiveTerm()
      {
       
      }

    void registerArgument(boost::shared_ptr<Matrix> arg)
    {
      arg_ = arg;
    }

  protected:
    
    boost::shared_ptr<Matrix> arg_;

    Matrix output_()
      {
	return output(*arg_);
      }

    Matrix jacobian_()
      {
	return jacobian(*arg_);
      }

    Matrix matrixDerivative_()
      {
	return matrixDerivative(*arg_);
      }

  };

  Registry<UnaryObjectiveTerm> & UnaryObjectiveTermRegistry();

}
