//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: composableAutoEncoderNetwork.cc
//Description: Uses ComposableBlocks to create a Network

#include "composableAutoEncoderNetwork.h"

// Check for at least one encoding block and at least one decoding block
#define COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(x) SDN_ASSERT(endEncodeBlock >= 0 && endEncodeBlock+1 < x.size())

namespace sdn {

  ComposableAutoEncoderNetwork::ComposableAutoEncoderNetwork() {
    endEncodeBlock = -1;
    costFunctionName = "sdn::SumOfSquaresCostFunction";
  }
  
  ComposableAutoEncoderNetwork::~ComposableAutoEncoderNetwork() {
  }
  
  void ComposableAutoEncoderNetwork::setConfigurableParameters() {
    SDN_WARN("ComposableAutoEncoderNetwork requires using loadXML for setup");
  }
  
  void ComposableAutoEncoderNetwork::loadXML(XMLNode& node)  {

    // Setting the End Encode Block makes it an AutoEncoder
    AttributeParser p;
    
    p.addInt("endEncodeBlock");
    p.addString("costFunction", "sdn::SumOfSquaresCostFunction");
    
    p.parse(node);
    
    endEncodeBlock = p.getInt("endEncodeBlock");
    costFunctionName = p.getString("costFunction");
    
    
    // Load up some blocks awesomeness
    XMLNode architectureNode = node.getChildNode("Architecture");
    SDN_ASSERT_MSG(!architectureNode.isEmpty(), "You need to specify <Architecture> !")
    LoadComposableBlocksBlock().loadXMLComposableBlocks(blocks, architectureNode);
    
    initPersistentParameters();
    initTransientParameters();
    
  }
  
  void ComposableAutoEncoderNetwork::initTransientParameters() {

    // Initialize the CostFunction
    reconstructionCostFunction = CostFunctionRegistry().make(costFunctionName);
    if (!reconstructionCostFunction) {
      SDN_ERROR("Could not load CostFunction with name " << costFunctionName);
      ERROR_OUT_LIST_REGISTRY_NAMES(CostFunction);
      SDN_FATAL("Quitting");
    }
    
    SDN_ASSERT_MSG(blocks.size() > 0, "No blocks in composable network!");
    
    // Setup the intermediate variables
    intermediateRepresentations.resize(blocks.size());
    intermediateDerivatives.resize(blocks.size());
    
  }
  
  void ComposableAutoEncoderNetwork::initPersistentParameters() {
    // Blocks are loaded by the XML loader or the serializer
  }
  
  void ComposableAutoEncoderNetwork::serialize(serial::Archive & ar)  {

    if (ar.in()) {
      // reading, empty our own structure
      blocks.clear();
    }
    int numBlocks = blocks.size();

    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;
    version = ar.beginBase(this,version,className());
    switch(version) {
      case 1000:
      
      
        ar.attribute("endEncodeBlock", endEncodeBlock);
        ar.attribute("costFunctionName", costFunctionName);
        ar.attribute("numBlocks", numBlocks);
        
        // serialize the blocks
        for(int i=0;i<numBlocks;i++) {
          
          // read/write the block name
          std::string blockName;
          if (!ar.in())  {
            blockName = blocks[i]->className();
          }
          ar.attribute("block" + itos(i), blockName);
          
          // read/write the block itself
          if (ar.in()) {
            
            // Try to load up this block by name
            boost::shared_ptr<ComposableBlock> newBlock;
            newBlock = ComposableBlockRegistry().make(blockName);
            if (!newBlock) {
              SDN_ERROR("Could not load ComposableBlock with name " << blockName);
              ERROR_OUT_LIST_REGISTRY_NAMES(ComposableBlock);
              SDN_FATAL("Quitting");
            }
            
            // Add Block to our List
            blocks.push_back(newBlock);
           
          }
          
          // actual read/write
          blocks[i]->serialize(ar);

        }
        
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
    }
    ar.endBase();
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
       
  }
  
  void ComposableAutoEncoderNetwork::unsupervisedTrain(const sdn::Matrix& example) {
    
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(blocks);
    
    // Forward Propagate
    encode(example);
    decode(getHiddenRepresentation());

    // Present Example
    sdn::Matrix& outputGradient = getOutputGradient();
    outputGradient = reconstructionCostFunction->jacobian(getReconstruction(), example).transpose();
    
    // Back Prop
    decodeBackProp( getHiddenRepresentation(), outputGradient, getHiddenGradient());
    encodeBackProp( example, getHiddenGradient(), inputDerivatives);

    // Gradient Updates
    gradientUpdate();

  }
  
  const boost::shared_ptr<CostFunction> ComposableAutoEncoderNetwork::getCostFunction() {
    return reconstructionCostFunction;
  }

  const sdn::Matrix& ComposableAutoEncoderNetwork::encode(const sdn::Matrix& input) {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateRepresentations);
    
    // Run encode for first block
    blocks[0]->forwardProp(input, intermediateRepresentations[0]);
    
    // Run forward prop for each intermediate block
    for (int i=1;i<=endEncodeBlock;i++)
      blocks[i]->forwardProp(intermediateRepresentations[i-1], intermediateRepresentations[i]);
          
    return intermediateRepresentations[endEncodeBlock];
  }
  
  const sdn::Matrix& ComposableAutoEncoderNetwork::decode(const sdn::Matrix& hidden) {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(blocks);
    SDN_ASSERT(intermediateRepresentations.size() == blocks.size());
    
    intermediateRepresentations[endEncodeBlock] = hidden;
    
    // Run encode for each future blocks
    for (int i=endEncodeBlock+1;i<blocks.size();i++)
      blocks[i]->forwardProp(intermediateRepresentations[i-1], intermediateRepresentations[i]);

    return intermediateRepresentations[blocks.size() - 1];
  }

  void ComposableAutoEncoderNetwork::encodeBackProp(const sdn::Matrix& input,
                                                    const sdn::Matrix& hiddenErrorDerivatives,
                                                    sdn::Matrix& inputErrorDerivatives) {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);

    intermediateDerivatives[endEncodeBlock] = hiddenErrorDerivatives;
    // run backprop for blocks
    for (int i=endEncodeBlock;i>0;i--) {
      blocks[i]->backProp(intermediateRepresentations[i-1], intermediateDerivatives[i], intermediateDerivatives[i-1]);
    }
    blocks[0]->backProp(input, intermediateDerivatives[0], inputErrorDerivatives);
    
  }

  void ComposableAutoEncoderNetwork::decodeBackProp(const sdn::Matrix& hidden,
                                                    const sdn::Matrix& outputErrorDerivatives,
                                                    sdn::Matrix& hiddenErrorDerivatives) {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);
    SDN_ASSERT(intermediateDerivatives.size() == blocks.size());
    
    intermediateDerivatives[blocks.size()-1] = outputErrorDerivatives;
    
    // run backprop for blocks
    for (int i=blocks.size()-1 ; i > endEncodeBlock + 1 ; i--) {
      blocks[i]->backProp(intermediateRepresentations[i-1], intermediateDerivatives[i], intermediateDerivatives[i-1]);
    }
    blocks[endEncodeBlock+1]->backProp(hidden, 
                                       intermediateDerivatives[endEncodeBlock+1], 
                                       intermediateDerivatives[endEncodeBlock]);
                                       
    hiddenErrorDerivatives = intermediateDerivatives[endEncodeBlock];
  }

  void ComposableAutoEncoderNetwork::encodeGradientUpdate() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);
    for (int i=0;i<=endEncodeBlock;i++)
      blocks[i]->gradientUpdate();
  }
  
  void ComposableAutoEncoderNetwork::decodeGradientUpdate() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);
    SDN_ASSERT(intermediateDerivatives.size() == blocks.size());
    for (int i=endEncodeBlock+1;i<blocks.size();i++)
      blocks[i]->gradientUpdate();
  }

  const sdn::Matrix& ComposableAutoEncoderNetwork::getHiddenRepresentation() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateRepresentations);
    return intermediateRepresentations[endEncodeBlock];
  }
  
  const sdn::Matrix& ComposableAutoEncoderNetwork::getReconstruction() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(blocks);
    SDN_ASSERT(intermediateRepresentations.size() == blocks.size());
    return intermediateRepresentations[blocks.size() - 1];
  }
  
  sdn::Matrix& ComposableAutoEncoderNetwork::getHiddenGradient() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);
    return intermediateDerivatives[endEncodeBlock];
  }
  
  sdn::Matrix& ComposableAutoEncoderNetwork::getOutputGradient() {
    COMPOSABLEAUTOENCODER_NETWORK_CHECK_AUTOENCODER_MODE(intermediateDerivatives);
    SDN_ASSERT(intermediateDerivatives.size() == blocks.size());
    return intermediateDerivatives[blocks.size() - 1];
  }
  

}
