//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: composableAutoEncoderNetwork.h
//Description: Uses ComposableBlocks to create an AutoEncoder

#pragma once

#include "sdn/base.h"
#include "../abstractNetwork.h"
#include "../../registry/composableBlockRegistry.h"
#include "../blocks/composableBlocks.h"
#include "../blocks/sharedBlocks.h"
#include "../blocks/standardBlocks.h"
#include "../defaultParameters.h"
#include "../math/costFunction.h"
#include "../math/normPenalty.h"

namespace sdn {
  
  /// ComposableAutoEncoderNetwork class
  /** 
  */
  class ComposableAutoEncoderNetwork : public AbstractAutoEncoderNetwork {
  public:
  
    virtual void setConfigurableParameters();
    virtual void loadXML(XMLNode& node) ;
    
    virtual void initTransientParameters();
    virtual void initPersistentParameters();
    
    virtual void serialize(serial::Archive & ar) ;
    
    virtual void unsupervisedTrain(const sdn::Matrix& example);
    virtual const boost::shared_ptr<CostFunction> getCostFunction();
    
    virtual const sdn::Matrix& encode(const sdn::Matrix& input);
    virtual const sdn::Matrix& decode(const sdn::Matrix& hidden);

    virtual void encodeBackProp(const sdn::Matrix& input,
                                const sdn::Matrix& hiddenErrorDerivatives,
                                sdn::Matrix& inputErrorDerivatives);

    virtual void decodeBackProp(const sdn::Matrix& hidden,
                                const sdn::Matrix& outputErrorDerivatives,
                                sdn::Matrix& hiddenErrorDerivatives);

    virtual void encodeGradientUpdate();
    virtual void decodeGradientUpdate();
    
    virtual const sdn::Matrix& getHiddenRepresentation();
    virtual const sdn::Matrix& getReconstruction();

    // These are internal functions for consistency, autoencoder specific helpers
    virtual sdn::Matrix& getHiddenGradient();
    virtual sdn::Matrix& getOutputGradient();
    
    ComposableAutoEncoderNetwork();
    ~ComposableAutoEncoderNetwork();

    // Persistent-Transient Parameters
    
    // We do not actually serialize the blocks, but instead the parameters
    // In effect, no pointers are serialized
    std::vector<boost::shared_ptr<ComposableBlock> > blocks;
    
    // Configurable Parameters
    int endEncodeBlock; // means *after* endEncodeBlock is completed, encoding ends there
    std::string costFunctionName;
    
    // Transient Parameters
    boost::shared_ptr<CostFunction> reconstructionCostFunction;
    std::vector<sdn::Matrix> intermediateRepresentations;// Representations for the values *after* the block (so [blocks.size()-1] = reconstruction)
    std::vector<sdn::Matrix> intermediateDerivatives;    // Gradients for the values *after* the block (so [blocks.size()-1] = outputGradient)
    sdn::Matrix inputDerivatives; // Gradient at the inputs (how to change the inputs to decrease error)
    
  };
  
}
