//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: composableSupervisedNetwork.cc
//Description: Uses ComposableBlocks to create an AutoEncoder

#include "composableSupervisedNetwork.h"
#include "../math/costFunction.h"
#include "../../registry/networkRegistry.h"

namespace sdn {

  void ComposableSupervisedNetwork::loadXML(XMLNode& node) {

    // Setting the End Encode Block makes it an AutoEncoder
    AttributeParser p;
    
    p.addString("costFunction", "sdn::SumOfSquaresCostFunction");
    p.parse(node);
    costFunctionName = p.getString("costFunction");
    
    // Load up some blocks awesomeness
    XMLNode architectureNode = node.getChildNode("Architecture");
    SDN_ASSERT_MSG(!architectureNode.isEmpty(), "You need to specify <Architecture> !")
    LoadComposableBlocksBlock().loadXMLComposableBlocks(blocks, architectureNode);
    
    initPersistentParameters();
    initTransientParameters();
    
  }
  
  void ComposableSupervisedNetwork::initTransientParameters() {
    
    // Initialize the CostFunction
    supervisedErrorCostFunction = CostFunctionRegistry().make(costFunctionName);
    if (!supervisedErrorCostFunction) {
      SDN_ERROR("Could not load CostFunction with name " << costFunctionName);
      ERROR_OUT_LIST_REGISTRY_NAMES(CostFunction);
      SDN_FATAL("Quitting");
    }
    
    SDN_ASSERT_MSG(blocks.size() > 0, "No blocks in composable network!");
    
    // Setup the intermediate variables
    intermediateRepresentations.resize(blocks.size());
    intermediateDerivatives.resize(blocks.size());

  }
  
  void ComposableSupervisedNetwork::initPersistentParameters() {
    // Blocks are loaded by the XML loader or the serializer
  }
  
  void ComposableSupervisedNetwork::serialize(serial::Archive & ar) {

    if (ar.in()) {
      // reading, empty our own structure
      blocks.clear();
    }
    int numBlocks = blocks.size();
    std::string encodingNetworkName = (encodingNetwork ? encodingNetwork->className() : "");
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;
    version = ar.beginBase(this,version,className());
    switch(version) {
      case 1000:
      
      
        ar.attribute("costFunctionName", costFunctionName);
        ar.attribute("encodingNetworkName", encodingNetworkName);
        
        if (ar.in()) {
          encodingNetwork = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> (AbstractNetworkRegistry().make(encodingNetworkName));
          if (!encodingNetwork) {
            SDN_ERROR("Could not load network with name " << encodingNetworkName);
            ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
            SDN_FATAL("Quitting");
          }
        }
        
        encodingNetwork->serialize(ar);
        
        ar.attribute("numBlocks", numBlocks);
        
        // serialize the blocks
        for(int i=0;i<numBlocks;i++) {
          
          // read/write the block name
          std::string blockName;
          if (!ar.in())  {
            blockName = blocks[i]->className();
          }
          ar.attribute("block" + itos(i), blockName);
          
          // read/write the block itself
          if (ar.in()) {
            
            // Try to load up this block by name
            boost::shared_ptr<ComposableBlock> newBlock;
            newBlock = ComposableBlockRegistry().make(blockName);
            if (!newBlock) {
              SDN_ERROR("Could not load ComposableBlock with name " << blockName);
              ERROR_OUT_LIST_REGISTRY_NAMES(ComposableBlock);
              SDN_FATAL("Quitting");
            }
            
            // Add Block to our List
            blocks.push_back(newBlock);
           
          }
          
          // actual read/write
          blocks[i]->serialize(ar);

        }
        
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
    }
    ar.endBase();
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
  }

  void ComposableSupervisedNetwork::supervisedTrain(const sdn::Matrix& example, 
                                                    const sdn::Matrix& label)  {
      
      SDN_ASSERT(supervisedErrorCostFunction);                                                
      
      forwardProp(example, prediction);
      
      sdn::Matrix& outputGradient = getOutputGradient();
      outputGradient = supervisedErrorCostFunction->jacobian(getPrediction(), label).transpose();
      
      backProp(example, outputGradient, inputDerivatives);
      
      gradientUpdate();
  }

  const boost::shared_ptr<CostFunction> ComposableSupervisedNetwork::getCostFunction() {
    return supervisedErrorCostFunction;
  }

  void ComposableSupervisedNetwork::forwardProp(const sdn::Matrix& input,
                                                sdn::Matrix& output)  {

    SDN_ASSERT(blocks.size() > 0);

    // Check if using AutoEncoder
    if (encodingNetwork) {
      // Transform the input before encoding into first block
      blocks[0]->forwardProp(encodingNetwork->encode(input),
                             intermediateRepresentations[0]);
    } else {
      // Run encode for first block
      blocks[0]->forwardProp(input, intermediateRepresentations[0]);
    }
  
    // Run forward prop for each intermediate block
    for (int i=1;i<blocks.size();i++)
      blocks[i]->forwardProp(intermediateRepresentations[i-1], intermediateRepresentations[i]);
      
    output = intermediateRepresentations[blocks.size() - 1];
    
  }

  void ComposableSupervisedNetwork::backProp(const sdn::Matrix& input,
                                             const sdn::Matrix& outputErrorDerivatives,
                                             sdn::Matrix& inputErrorDerivatives)  {

    SDN_ASSERT(intermediateDerivatives.size() == blocks.size());
    
    intermediateDerivatives[blocks.size()-1] = outputErrorDerivatives;

    // run backprop for blocks
    for (int i=blocks.size()-1;i>0;i--) {
      blocks[i]->backProp(intermediateRepresentations[i-1], intermediateDerivatives[i], intermediateDerivatives[i-1]);
    }
    
    if (encodingNetwork) {
      blocks[0]->backProp(encodingNetwork->getHiddenRepresentation(), intermediateDerivatives[0], inputErrorDerivatives);
    } else {
      blocks[0]->backProp(input, intermediateDerivatives[0], inputErrorDerivatives);
    }

  }
  
  void ComposableSupervisedNetwork::gradientUpdate()  {
    for (int i=0;i<blocks.size();i++) {
      blocks[i]->gradientUpdate();
    }
  }

  sdn::Matrix& ComposableSupervisedNetwork::getOutputGradient() {
    SDN_ASSERT(intermediateDerivatives.size() == blocks.size());
    return intermediateDerivatives[blocks.size() - 1];
  }

  const sdn::Matrix& ComposableSupervisedNetwork::getPrediction() {
    return prediction;
  }

  void ComposableSupervisedNetwork::useAutoEncoderNetwork(boost::shared_ptr<AbstractAutoEncoderNetwork> encodingNet)  {
    encodingNetwork = encodingNet;
  }

  void ComposableSupervisedNetwork::fineTuneAutoEncoder(const sdn::Matrix& input) {
    
    SDN_ASSERT_MSG(encodingNetwork, "You need to add an AutoEncoder before you can fine tune one!");
    
    // Backprop the AutoEncoder and Gradient Update
    encodingNetwork->encodeBackProp(input,
                                    inputDerivatives,
                                    fineTuneInputDerivatives);
    encodingNetwork->encodeGradientUpdate();
  }
  
}
