//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: StackedAutoEncoderNetwork.cc
//Description: Uses AutoEncoderNetworks to create a Deep Network

#include "stackedAutoEncoderNetwork.h"
#include "../math/costFunction.h"
#include "../../registry/networkRegistry.h"
#include "sdn/base.h"

namespace sdn {
	
	StackedAutoEncoderNetwork::StackedAutoEncoderNetwork() {
    costFunctionName = "sdn::SumOfSquaresCostFunction";
	}
  
	StackedAutoEncoderNetwork::~StackedAutoEncoderNetwork() {
	}
  
  void StackedAutoEncoderNetwork::setConfigurableParameters() {
    initPersistentParameters();
    initTransientParameters();
  }
  
  void StackedAutoEncoderNetwork::loadXML(XMLNode& node) {
    
    // Only loads the cost function name
    
    
    AttributeParser p;
    
    p.addString("costFunction", "sdn::SumOfSquaresCostFunction");
    p.parse(node);
    costFunctionName = p.getString("costFunction");

    
    initPersistentParameters();
    initTransientParameters();
  }

  void StackedAutoEncoderNetwork::initTransientParameters() {
    reconstructionCostFunction = CostFunctionRegistry().make(costFunctionName);
    if (!reconstructionCostFunction) {
      SDN_ERROR("Could not load CostFunction with name " << costFunctionName);
      ERROR_OUT_LIST_REGISTRY_NAMES(CostFunction);
      SDN_FATAL("Quitting");
    }
		
		encodeIntermediateRepresentations.resize(networkStack.size());
		decodeIntermediateRepresentations.resize(networkStack.size());

		encodeIntermediateDerivatives.resize(networkStack.size());
		decodeIntermediateDerivatives.resize(networkStack.size());
  }
	
  void StackedAutoEncoderNetwork::initPersistentParameters() {
    
  }
  
  void StackedAutoEncoderNetwork::serialize(serial::Archive & ar) {
    if (ar.in()) {
      // reading, empty our own structure
      networkStack.clear();
    }
    int numNetworks = networkStack.size();
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;
    version = ar.beginBase(this,version,className());
    switch(version) {
      case 1000:
      
      
        ar.attribute("costFunctionName", costFunctionName);
        
        ar.attribute("numNetworks", numNetworks);
        
        // serialize the blocks
        for(int i=0;i<numNetworks;i++) {
          
          // read/write the block name
          std::string networkName;
          if (!ar.in())  {
            networkName = networkStack[i]->className();
          }
          ar.attribute("network" + itos(i), networkName);
          
          // read/write the block itself
          if (ar.in()) {
            
            // Try to load up this block by name
            boost::shared_ptr<AbstractAutoEncoderNetwork> newNetwork;
            newNetwork = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> (AbstractNetworkRegistry().make(networkName));
            if (!newNetwork) {
              SDN_ERROR("Could not load network with name " << networkName);
              ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
              SDN_FATAL("Quitting");
            }
            
            // Add Block to our List
            networkStack.push_back(newNetwork);
           
          }
          
          // actual read/write
          networkStack[i]->serialize(ar);

        }
        
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
    }
    ar.endBase();
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
	}
	
  void StackedAutoEncoderNetwork::unsupervisedTrain(const sdn::Matrix& example) {

		// Empty Stack 
		if (networkStack.size() == 0) {
			SDN_WARN("Called " + std::string(__FUNCTION__) + " on an empty network");
			inputToEmptyNetwork = example;
			return ;
		}
    
		// Training a StackedAutoEncoder Network implies that you are finetuning the whole network
		encode(example);
		decode(getHiddenRepresentation());
		
    // Present Example
    sdn::Matrix& outputGradient = getOutputGradient();
    outputGradient = reconstructionCostFunction->jacobian(getReconstruction(), example).transpose();
    
    // Back Prop
    decodeBackProp( getHiddenRepresentation(), getOutputGradient(), getHiddenGradient());
    encodeBackProp( example, getHiddenGradient(), inputDerivatives);
    
    // Gradient Updates 
    gradientUpdate();
	}
  
  const boost::shared_ptr<CostFunction> StackedAutoEncoderNetwork::getCostFunction() {
    return reconstructionCostFunction;
  }
	
  const sdn::Matrix& StackedAutoEncoderNetwork::encode(const sdn::Matrix& input) {
		
		// Empty stack, treat this as the "identity" network
		if (networkStack.size() <= 0) {
			inputToEmptyNetwork = input;
			return inputToEmptyNetwork;
		}
    
	
		// We call encode on each network in the stack
		encodeIntermediateRepresentations[0] = networkStack[0]->encode(input);
		for (int i=1 ; i < networkStack.size() ; i++) {
			encodeIntermediateRepresentations[i] = networkStack[i]->encode(encodeIntermediateRepresentations[i-1]);
		}
		
    return getHiddenRepresentation();
  }
  
  const sdn::Matrix& StackedAutoEncoderNetwork::decode(const sdn::Matrix& hidden) {

		// Empty stack, treat this as the "identity" network
		if (networkStack.size() == 0) {
			inputToEmptyNetwork = hidden;
			return inputToEmptyNetwork;
		}
		
		// Call decode on each network -- in reverse order!
		decodeIntermediateRepresentations[networkStack.size() - 1] = networkStack[networkStack.size() - 1]->decode(hidden);
		for (int i=networkStack.size() - 2 ; i >= 0 ; i--) {
			decodeIntermediateRepresentations[i] = networkStack[i]->decode(decodeIntermediateRepresentations[i+1]);
		}
		
    return getReconstruction();
  }

  void StackedAutoEncoderNetwork::encodeBackProp(const sdn::Matrix& input,
																								 const sdn::Matrix& hiddenErrorDerivatives,
																								 sdn::Matrix& inputErrorDerivatives) {
		
		// Empty stack, treat this as the "identity" network
		if (networkStack.size() == 0) {
			// just copy over
			inputErrorDerivatives = hiddenErrorDerivatives;
			return ;
		}

		encodeIntermediateDerivatives[networkStack.size() - 1] = hiddenErrorDerivatives;
    // run backprop for network stack
    for (int i=networkStack.size() - 1 ; i > 0 ; i-- ) {
      networkStack[i]->encodeBackProp(encodeIntermediateRepresentations[i-1], 
																			encodeIntermediateDerivatives[i], 
																			encodeIntermediateDerivatives[i-1]);
    }
    networkStack[0]->encodeBackProp(input, encodeIntermediateDerivatives[0], inputErrorDerivatives);

  }

  void StackedAutoEncoderNetwork::decodeBackProp(const sdn::Matrix& hidden,
																								 const sdn::Matrix& outputErrorDerivatives,
																								 sdn::Matrix& hiddenErrorDerivatives) {

		// Empty stack, treat this as the "identity" network
		if (networkStack.size() == 0) {
			// just copy over
			hiddenErrorDerivatives = outputErrorDerivatives;
			return ;
		}
		
		decodeIntermediateDerivatives[0] = outputErrorDerivatives;
    
		// run backprop for blocks
    for (int i=0;i<networkStack.size()-1;i++) {
      networkStack[i]->decodeBackProp(decodeIntermediateRepresentations[i+1], 
																			decodeIntermediateDerivatives[i], 
																			decodeIntermediateDerivatives[i+1]);
    }
		
		networkStack[networkStack.size()-1]->decodeBackProp(hidden, 
																												decodeIntermediateDerivatives[networkStack.size() - 1], 
																												hiddenErrorDerivatives);
																						
  }

  void StackedAutoEncoderNetwork::encodeGradientUpdate() {
		for (int i=0 ; i < networkStack.size() ; i++) {
			networkStack[i]->encodeGradientUpdate();
		}
  }

  void StackedAutoEncoderNetwork::decodeGradientUpdate() {
		for (int i=0 ; i < networkStack.size() ; i++) {
			networkStack[i]->decodeGradientUpdate();
		}
  }
	
	sdn::Matrix& StackedAutoEncoderNetwork::getOutputGradient() {
		return decodeIntermediateDerivatives[0];
	}

	sdn::Matrix& StackedAutoEncoderNetwork::getHiddenGradient() {
		SDN_ASSERT(encodeIntermediateDerivatives.size() == networkStack.size());
    return encodeIntermediateDerivatives[networkStack.size()-1];
	}
	
  const sdn::Matrix& StackedAutoEncoderNetwork::getHiddenRepresentation() {
		SDN_ASSERT(encodeIntermediateRepresentations.size() == networkStack.size());
    if (networkStack.size() == 0)
      return inputToEmptyNetwork;
    else
      return encodeIntermediateRepresentations[networkStack.size()-1];
  }
  
  const sdn::Matrix& StackedAutoEncoderNetwork::getReconstruction() {
		SDN_ASSERT(decodeIntermediateRepresentations.size() == networkStack.size());
    if (networkStack.size() == 0)
      return inputToEmptyNetwork;
    else
      return decodeIntermediateRepresentations[0];
  }
	
	void StackedAutoEncoderNetwork::addAutoEncoder(boost::shared_ptr<AbstractAutoEncoderNetwork> anAutoEncoder) {

    // Add the auto encoder
    networkStack.push_back(anAutoEncoder);

		// Further Setup Transient Parameters
		
		encodeIntermediateRepresentations.push_back(sdn::Matrix());
		decodeIntermediateRepresentations.push_back(sdn::Matrix());

		encodeIntermediateDerivatives.push_back(sdn::Matrix());
		decodeIntermediateDerivatives.push_back(sdn::Matrix());
		
	}


}
