//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: StandardAutoEncoderNetwork.cc
//Description: Uses AutoEncoderBlocks to create a Network

#include "../blocks/sharedBlocks.h"
#include "standardAutoEncoderNetwork.h"
#include "sdn/base.h"
#include <Eigen/Array>

namespace sdn {
  
  void StandardAutoEncoderNetwork::unsupervisedTrain(const sdn::Matrix& example) {
   
    if (reconstruction.size() != example.size()) {
      reconstruction = sdn::Matrix::Zero(example.rows(), example.cols());
      affineRepDec = sdn::Matrix::Zero(example.rows(), example.cols());
    }

    // Forward Propagate
    encode(example);
    decode(hiddenRepresentation);

    // Present Example
    outputGradient = costFunc->jacobian(reconstruction, example).transpose();
    
    // Back Prop
    decodeBackProp( hiddenRepresentation, outputGradient, hiddenGradient);
    encodeBackProp( example, hiddenGradient, inputGradient);

    // Gradient Updates (weight decay is performed in the affineEncoder itself)
    gradientUpdate();

    // ============== ADDITIONAL OPERATIONS ==============

    // Sparsity Updates
    activationTracker.addValues( sigmEncoder.normalizeActivationValues( hiddenRepresentation ) );
    sdn::Matrix sparsityError = activationTracker.getMean().cwise() - targetActivation;
    affineEncoder.bias -= affineEncoder.learningRate*learningRateSparsity*sparsityError;
    
  }

  const sdn::Matrix& StandardAutoEncoderNetwork::encode(const sdn::Matrix& input) {
    
    affineEncoder.forwardProp(input, affineRepEnc);
    sigmEncoder.forwardProp(affineRepEnc, hiddenRepresentation);
    
    return hiddenRepresentation;
  }

  const sdn::Matrix& StandardAutoEncoderNetwork::decode(const sdn::Matrix& hidden) {
    
    SDN_ASSERT(affineEncoder.inputSize > 0);
    
    affineDecoder.forwardProp(hidden, affineRepDec);
    sigmDecoder.forwardProp(affineRepDec, reconstruction);
    
    return reconstruction;
  }

  void StandardAutoEncoderNetwork::encodeBackProp(const sdn::Matrix& input,
            const sdn::Matrix& hiddenErrorDerivatives,
            sdn::Matrix& inputErrorDerivatives) {
    sdn::Matrix tempGradient;
    sigmEncoder.backProp(affineRepEnc, hiddenErrorDerivatives, tempGradient);
    affineEncoder.backProp(input, tempGradient, inputErrorDerivatives);
  }

  void StandardAutoEncoderNetwork::decodeBackProp(const sdn::Matrix& hidden,
            const sdn::Matrix& outputErrorDerivatives,
            sdn::Matrix& hiddenErrorDerivatives) {
    sdn::Matrix tempGradient;
    sigmDecoder.backProp(affineRepDec, outputErrorDerivatives, tempGradient);
    affineDecoder.backProp(hidden, tempGradient, hiddenErrorDerivatives);
  }

  void StandardAutoEncoderNetwork::encodeGradientUpdate() {
    affineEncoder.gradientUpdate();
  }

  void StandardAutoEncoderNetwork::decodeGradientUpdate() {
    affineDecoder.gradientUpdate();
  }

  const sdn::Matrix& StandardAutoEncoderNetwork::getHiddenRepresentation() {
    return hiddenRepresentation;
  }
  
  const sdn::Matrix& StandardAutoEncoderNetwork::getReconstruction() {
    return reconstruction;
  }
  
  void StandardAutoEncoderNetwork::setupGoodInitialParameters() {
    
    if (affineEncoder.inputSize > 0) {
      RandomizeWeightMatrix().randomize(affineEncoder.weights, affineEncoder.inputSize);
    }
      
    if (hiddenLayerSize > 0) {
      RandomizeWeightMatrix().randomize(affineDecoder.weights, hiddenLayerSize);
      affineEncoder.bias = sigmEncoder.getBiasInitialParameters(targetActivation, hiddenLayerSize);
    }
    
  }
  
  // --------------------------------------------------------------------------

  void StandardAutoEncoderNetwork::setConfigurableParameters(sdn::Scalar defTargetActivation,
                                                             sdn::Scalar defLearningRateSparsity,
                                                             unsigned int defHiddenLayerSize,
                                                             sdn::Scalar defAveragingConstant) {

    // Shallow Loading

    targetActivation = defTargetActivation;
    hiddenLayerSize = defHiddenLayerSize;
    learningRateSparsity = defLearningRateSparsity;
    activationAveragingConstant = defAveragingConstant;
    
    
    initPersistentParameters();
    initTransientParameters();
  }

  void StandardAutoEncoderNetwork::loadXML(XMLNode& node) {

    AttributeParser p;
    
    p.addDouble("targetActivation");
    p.addDouble("activationAveragingConstant");
    p.addDouble("learningRateSparsity");
    p.addInt("hiddenLayerSize");
    
    p.parse(node);

    targetActivation = p.getDouble("targetActivation");
    activationAveragingConstant = p.getDouble("activationAveragingConstant");
    learningRateSparsity = p.getDouble("learningRateSparsity");
    hiddenLayerSize = p.getInt("hiddenLayerSize");

    // Deep Loading
    XMLNode affineEncoderNode = node.getChildNode("Encoder");
    if (!affineEncoderNode.isEmpty())
      affineEncoder.loadXML(affineEncoderNode);

    XMLNode affineDecoderNode = node.getChildNode("Decoder");
    if (!affineDecoderNode.isEmpty())
      affineDecoder.loadXML(affineDecoderNode);

    initPersistentParameters();
    initTransientParameters();

  }
  
  void StandardAutoEncoderNetwork::initTransientParameters() {
    if (hiddenLayerSize > 0) {
      hiddenRepresentation = sdn::Matrix::Zero(hiddenLayerSize, 1);
      hiddenGradient = sdn::Matrix::Zero(hiddenLayerSize, 1);
      affineRepEnc =  sdn::Matrix::Zero(hiddenLayerSize, 1);
    }
    if (affineEncoder.inputSize > 0) {
      inputGradient = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
      outputGradient = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
      affineRepDec = sdn::Matrix::Zero(affineEncoder.inputSize, 1);
    }
    activationTracker.setConfigurableParameters(activationAveragingConstant,
                                                targetActivation,
                                                SIZE_AUTOSET);
    
  }
  
  void StandardAutoEncoderNetwork::initPersistentParameters() {
    
    affineEncoder.initPersistentParameters();
    affineDecoder.initPersistentParameters();
    
    sigmEncoder.initPersistentParameters();
    sigmDecoder.initPersistentParameters();

    setupGoodInitialParameters();

  }

  const boost::shared_ptr<CostFunction> StandardAutoEncoderNetwork::getCostFunction() {
    return costFunc;
  }

  void StandardAutoEncoderNetwork::serialize(serial::Archive & ar)  {
    SDN_DEBUG(className() << "::" << __FUNCTION__);
    int version = 1000;

    version = ar.beginBase(this,version,className());

    switch(version)
      {
      case 1000:
      
        // Persistent
        affineEncoder.serialize(ar);
        affineDecoder.serialize(ar);
        sigmEncoder.serialize(ar);
        sigmDecoder.serialize(ar);
        
        // Configurable
        ar.attribute("targetActivation", targetActivation);
        ar.attribute("activationAveragingConstant", activationAveragingConstant);
        ar.attribute("hiddenLayerSize", hiddenLayerSize);
        ar.attribute("learningRateSparsity", learningRateSparsity);
    
        break;
        
      default:
        SDN_FATAL("Unrecognized " << className() << " version");
      }
    ar.endBase();
      
    // Input Case: Initialize Transient
    if (ar.in()) {
      initTransientParameters();
    }
        
  }

  StandardAutoEncoderNetwork::StandardAutoEncoderNetwork() {
    costFunc = boost::shared_ptr<CostFunction> (new SumOfSquaresCostFunction());
  }
  
  StandardAutoEncoderNetwork::~StandardAutoEncoderNetwork() {
  }

  
}
