//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: ScriptRuntimeManager.cc
//Description: Loads up a script / dataset and runs stuff !

#include "scriptRuntimeManager.h"
#include "../registry/scriptRegistry.h"
#include "blocks/sharedBlocks.h"
#include "scripts/callbacks/sharedCallbacks.h"
#include "sdn/base.h"

namespace sdn {
  
  class ScriptRuntimeManagerConfigParser : public ConfigParser {

  protected:
    ScriptRuntimeManager & parentManager;

  public:
  
    ScriptRuntimeManagerConfigParser(ScriptRuntimeManager & parent) : parentManager(parent) {}
        
    ~ScriptRuntimeManagerConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      topics.push_back("sdn::ScriptRuntimeManager");
      topics.push_back("UseScriptCallbacks");
    }

    void readConfig(XMLNode & node) {
      parentManager.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::ScriptRuntimeManager ...... script=\"<String>\" />\n";
    }
  };

  // ---------------------------------------------------------------

  void ScriptRuntimeManager::loadCallBack(XMLNode& node) {
    const char * cname;
    SDN_ASSERT(cname = node.getName());
    std::string name (cname);

    boost::shared_ptr<ScriptCallback> cb =  ScriptCallbackRegistry().make(name);
    
    if (!cb) {
      SDN_ERROR("Could not load ScriptCallback with name " << name);
      ERROR_OUT_LIST_REGISTRY_NAMES(ScriptCallback);
      SDN_FATAL("Quitting");
    }
    
    cb->loadXML(node);
    callbacks.push_back(cb);
    callbackNextIter.push_back(cb->firstIter());
  }

  void ScriptRuntimeManager::loadXML(XMLNode& node) {
    const char * name;
    SDN_ASSERT(name = node.getName());
    if (std::string(name) == "UseScriptCallbacks") {
      for (int i = 0; i < node.nChildNode(); i++) {
        XMLNode childNode = node.getChildNode(i);
        loadCallBack(childNode);
      }
    } else { // must be "sdn::ScriptRuntimeManager"
      AttributeParser p;
      p.addString("script");
      p.addInt("trainIterations");
      p.parse(node);
      scriptName = p.getString("script");
      trainIterations = p.getInt("trainIterations");
    }
  }

  void ScriptRuntimeManager::init() {

    // Check if name mangling is OK to auto obtain class names
    if (className() != "sdn::ScriptRuntimeManager") {
      SDN_FATAL("Check Class Name Demangling Failed: Got " << className() << " expected sdn::ScriptRuntimeManager");
    } else {
      SDN_DEBUG("Check Class Name Demangling Passed: " << className());
    }
    

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new ScriptRuntimeManagerConfigParser(*this)));
        
    // Attempt to load the dataset and script from the registrys
    script = ScriptRegistry().make(scriptName);
    if (!script) {
      SDN_ERROR("Could not load script with name " << scriptName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Script);
      SDN_FATAL("Quitting");
    }
    script->manager = this;

  }


  void ScriptRuntimeManager::run() {
    
    for(currentIteration=0;currentIteration<trainIterations;currentIteration++) {
      
      script->executeOneIteration();
      
      // Possible Optimzation for the Following Using a P-Queue
      // if we ever optimize this, note DisableOtherCallback
      for (int j = 0;j<callbacks.size();j++)
        if (callbackNextIter[j] == currentIteration) {
          callbacks[j]->execute(*this, *script, currentIteration);
          callbackNextIter[j] = currentIteration + callbacks[j]->itersToNextCall();
        }
    }
    
    
  }


  ScriptRuntimeManager::ScriptRuntimeManager() {
    init();
  }
  
  ScriptRuntimeManager::~ScriptRuntimeManager() {
    
    
  }
  
}
