//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: sharedCallbacks.cc
//Description: 

#include "sharedCallbacks.h"
#include "../../math/costFunction.h"
#include "../../helpers.h"
#include "../../scriptRuntimeManager.h"
#include "../../../registry/scriptRegistry.h"
#include "sdn/base.h"

namespace sdn {

  void ReconstructionErrorCallback::loadXML(XMLNode & node) {

    AttributeParser p;

    p.addDouble("averagingConstant");
    p.addInt("reportStep");
    p.addInt("firstIter", 0);
    p.addString("saveFile");

    p.parse(node);

    averagingConstant = p.getDouble("averagingConstant");
    reportStep = p.getInt("reportStep");
    runFirstIter = p.getInt("firstIter");
    saveFile = p.getString("saveFile");

    initPersistentParameters();
    initTransientParameters();

  }

  // We call this every iteration to actually compute the error
  long int ReconstructionErrorCallback::itersToNextCall() {
    return 1;
  }
  
  long int ReconstructionErrorCallback::firstIter() {
    return runFirstIter;
  }
  
  void ReconstructionErrorCallback::execute(ScriptRuntimeManager& scriptManager,
                                            AbstractScript& script,
                                            unsigned int iteration) {
    
    ReconstructionErrorCallbackInterface* rScript;
    SDN_ASSERT(rScript = dynamic_cast<ReconstructionErrorCallbackInterface*> (&script));
    
    sdn::Matrix error = rScript->getReconstructionCostFunction()->output(rScript->getReconstruction(), rScript->getOriginal());
    errorTracker.addValues(error);
    
    if (iteration % reportStep == 0) {
      std::cout << "[" << itos(iteration,10) << "] Reconstruction Error (" 
                << (getClassName<CostFunction> (*(rScript->getReconstructionCostFunction())))
                << ") : " << errorTracker.getMean() << "\n";
                
      if (saveFile != "" ) {
        appendToFile(saveFile, "%010u %f\n", iteration, errorTracker.getMean()[0]);
      }
    }
  }

  void ReconstructionErrorCallback::initTransientParameters() {
    errorTracker.setConfigurableParameters(averagingConstant, TRACKER_AUTOSET, SIZE_AUTOSET);
    // clear out old error file
    if (saveFile != "") {
      clearFile(saveFile);
    }
  }
  
  void ReconstructionErrorCallback::initPersistentParameters() {
  }
  
  void ReconstructionErrorCallback::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << " not serializable yet");
  }


  // ------------------------------------------------------------------   
  
  void SupervisedErrorCallback::loadXML(XMLNode & node) {

    AttributeParser p;

    p.addDouble("averagingConstant");
    p.addInt("reportStep");
    p.addInt("firstIter", 0);
		p.addString("saveFile");

    p.parse(node);

    averagingConstant = p.getDouble("averagingConstant");
    reportStep = p.getInt("reportStep");
    runFirstIter = p.getInt("firstIter");
		saveFile = p.getString("saveFile");

    initPersistentParameters();
    initTransientParameters();

  }

  // We call this every iteration to actually compute the error
  long int SupervisedErrorCallback::itersToNextCall() {
    return 1;
  }
  
  long int SupervisedErrorCallback::firstIter() {
    return runFirstIter;
  }
  
  void SupervisedErrorCallback::execute(ScriptRuntimeManager& scriptManager,
                                        AbstractScript& script,
                                        unsigned int iteration) {
    
    SupervisedErrorCallbackInterface* rScript;
    SDN_ASSERT(rScript = dynamic_cast<SupervisedErrorCallbackInterface*> (&script));
    
    sdn::Matrix error = rScript->getSupervisedCostFunction()->output(rScript->getTarget(), rScript->getPrediction());
    errorTracker.addValues(error);
    
    if (iteration % reportStep == 0) {
      std::cout << "[" << itos(iteration,10) << "] Supervised Error (" 
                << (getClassName<CostFunction> (*(rScript->getSupervisedCostFunction()))) 
                << ") : " << errorTracker.getMean() << "\n";

			if (saveFile != "") {
				appendToFile(saveFile, "%010u %f\n", iteration, errorTracker.getMean()[0]);
			}
		}
   
  }

  void SupervisedErrorCallback::initTransientParameters() {
    errorTracker.setConfigurableParameters(averagingConstant, TRACKER_AUTOSET, SIZE_AUTOSET);
  }
  
  void SupervisedErrorCallback::initPersistentParameters() {
		if (saveFile !="") {
			clearFile(saveFile);
		}
  }
  
  void SupervisedErrorCallback::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << " not serializable yet");
  }


  // ------------------------------------------------------------------ 
  
  void ProgressReportCallback::loadXML(XMLNode & node) {
    AttributeParser p;
    p.addInt("reportStep");
    p.parse(node);
    reportStep = p.getInt("reportStep");
    initPersistentParameters();
    initTransientParameters();
  }

  void ProgressReportCallback::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << " not serializable yet");
  }

  void ProgressReportCallback::execute(ScriptRuntimeManager& scriptManager,
                                       AbstractScript& script,
                                       unsigned int iteration) {
    ProgressReportCallbackInterface* prScript;
    SDN_ASSERT(prScript = dynamic_cast<ProgressReportCallbackInterface*> (&script));
    prScript->reportProgress(std::cout);
  }

  long int ProgressReportCallback::itersToNextCall() {
    return reportStep;
  }
  
  long int ProgressReportCallback::firstIter() {
    return 0;
  }
  
  // ------------------------------------------------------------------ 

  void DisableOtherCallback::loadXML(XMLNode & node) {

    AttributeParser p;
    p.addInt("disableIter");
    p.addString("targetCallback");
    p.parse(node);
    disableIter = p.getInt("disableIter");
    targetCallback = p.getString("targetCallback");
    
    // Test and See whether targetCallback exists
    boost::shared_ptr<ScriptCallback> cb =  ScriptCallbackRegistry().make(targetCallback);
    SDN_ASSERT_MSG(cb, " targetCallback in " << (node.getName() ? node.getName() : "unknown") << " does not exist!");
    
    initPersistentParameters();
    initTransientParameters();
    

  }


  long int DisableOtherCallback::itersToNextCall() {
    // Runs only once
    return -1;
  }
  
  long int DisableOtherCallback::firstIter() {
    return disableIter;
  }
  
  void DisableOtherCallback::execute(ScriptRuntimeManager& scriptManager,
                                     AbstractScript& script,
                                     unsigned int iteration) {
    
    SDN_ASSERT(disableIter == iteration);
    
    // go through the manager and find the target
    for (int j = 0;j<scriptManager.callbacks.size();j++)
      if (scriptManager.callbacks[j]->className() == targetCallback) {
        
        // Disable it!
        scriptManager.callbackNextIter[j] = -1;
        
      }
  }

  void DisableOtherCallback::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << " not serializable yet");
  }
  
  // ------------------------------------------------------------------

  void SaveModelCallback::loadXML(XMLNode & node) {

    AttributeParser p;

    p.addInt("step");
    p.addInt("firstIter", 0);
    p.addString("saveDir");

    p.parse(node);

    step = p.getInt("step");
    runFirstIter = p.getInt("firstIter");
    saveDir = p.getString("saveDir");

    initPersistentParameters();
    initTransientParameters();

    if (saveDir == "")
      saveDir = ".";
    if (saveDir[saveDir.size()-1] != '/')
      saveDir = saveDir + "/";

  }

  // We call this every iteration to actually compute the error
  long int SaveModelCallback::itersToNextCall() {
    return step;
  }
  
  long int SaveModelCallback::firstIter() {
    return runFirstIter;
  }
  
  void SaveModelCallback::execute(ScriptRuntimeManager& scriptManager,
                                  AbstractScript& script,
                                  unsigned int iteration) {
    
    SaveModelCallbackInterface* rScript;
    SDN_ASSERT(rScript = dynamic_cast<SaveModelCallbackInterface*> (&script));
    rScript->saveModelInDirectory(saveDir);
  }

  void SaveModelCallback::initTransientParameters() {}
  
  void SaveModelCallback::initPersistentParameters() {}
  
  void SaveModelCallback::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << " " << __FUNCTION__ << "not implemented yet");
  }


}
