//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: StandardAutoEncoderScript.cc
//Description: 

#include "standardAutoEncoderScript.h"
#include "../scriptRuntimeManager.h"
#include "sdn/base.h"
#include "sdn/dataset.h"

namespace sdn {

  class StandardAutoEncoderScriptConfigParser : public ConfigParser {

  protected:
    StandardAutoEncoderScript & parentScript;

  public:
  
    StandardAutoEncoderScriptConfigParser(StandardAutoEncoderScript & parent) : parentScript(parent) {}
    ~StandardAutoEncoderScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::StandardAutoEncoderScript");
    }

    void readConfig(XMLNode & node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::StandardAutoEncoderScript ...... learningRate=\"<Scalar>\" learningRateSparsity=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  StandardAutoEncoderScript::StandardAutoEncoderScript() {
    loadUsingConfigManager();
  }
  
  StandardAutoEncoderScript::~StandardAutoEncoderScript() {
  }
  
  void StandardAutoEncoderScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    
    p.parse(node);
    
    // Note that values are set *directly* into the network

    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth = p.getInt("inputWidth");

    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network.loadXML(networkNode);
    
  }

  void StandardAutoEncoderScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new StandardAutoEncoderScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();
  }
  
  void StandardAutoEncoderScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
  }
  
  void StandardAutoEncoderScript::initPersistentParameters() {
    network.initPersistentParameters();
    network.initTransientParameters();
  }
  
  void StandardAutoEncoderScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void StandardAutoEncoderScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // do fancy stuff here
    normInput.normalizePatch(inputExample);
    network.unsupervisedTrain(inputExample);
    
  }
  

  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> StandardAutoEncoderScript::getReconstructionCostFunction() {
    return network.costFunc;
  }
  
  const sdn::Matrix& StandardAutoEncoderScript::getReconstruction() {
    return network.reconstruction;
  }
  
  const sdn::Matrix& StandardAutoEncoderScript::getOriginal() {
    return inputExample;
  }

  const sdn::Matrix& StandardAutoEncoderScript::getEncoderWeights() {
    return network.affineEncoder.weights;
  }

  const int StandardAutoEncoderScript::getWeightPatchHeight() {
    return inputHeight;
  }
  const sdn::Matrix& StandardAutoEncoderScript::getDecoderWeights() {
     return network.affineDecoder.weights;
  }

  // ---- Save Model Callback ----
  const void StandardAutoEncoderScript::saveModelInDirectory(std::string dir) {
    network.quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");
  }
 
 
}
