//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: supervisedWithPreTrainingScript.cc
//Description: 

#include "supervisedWithPreTrainingScript.h"
#include "../scriptRuntimeManager.h"
#include "../../registry/networkRegistry.h"
#include "sdn/base.h"
#include "sdn/dataset.h"


namespace sdn {

  class SupervisedWithPreTrainingScriptConfigParser : public ConfigParser {

  protected:
    SupervisedWithPreTrainingScript & parentScript;

  public:
  
    SupervisedWithPreTrainingScriptConfigParser(SupervisedWithPreTrainingScript & parent) : parentScript(parent) {}
    ~SupervisedWithPreTrainingScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::SupervisedWithPreTrainingScript");
    }

    void readConfig(XMLNode & node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::SupervisedWithPreTrainingScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  SupervisedWithPreTrainingScript::SupervisedWithPreTrainingScript() {
    loadUsingConfigManager();
  }
  
  SupervisedWithPreTrainingScript::~SupervisedWithPreTrainingScript() {
  }
  
  void SupervisedWithPreTrainingScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("encodingNetworkName", "");
    p.addString("supervisedNetworkName");
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    p.addInt("unsupervisedTrainingIterations");
    p.addBool("fineTuneUnsupervisedNetwork");
    
    p.parse(node);
    
    // Set Values
    encodingNetworkName = p.getString("encodingNetworkName");
    supervisedNetworkName = p.getString("supervisedNetworkName");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    unsupervisedTrainingIterations = p.getInt("unsupervisedTrainingIterations");
    fineTuneUnsupervisedNetwork = p.getBool("fineTuneUnsupervisedNetwork");
  
    // ---------------------------------------------------
    
    // Load AutoEncoder Part
    
    encodingNetwork = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> (AbstractNetworkRegistry().make(encodingNetworkName));
    if (!encodingNetwork) {
      if (encodingNetworkName == "" ) {
        SDN_MSG("No pre-training network specified.");
        SDN_MSG("Setting unsupervisedTrainingIterations = -1");
        SDN_MSG("Setting fineTuneUnsupervisedNetwork    = false");
        unsupervisedTrainingIterations=-1;
        fineTuneUnsupervisedNetwork=false;
      } else {
        SDN_ERROR("Could not load network with name " << encodingNetworkName);
        SDN_ERROR("Is " << encodingNetworkName << " an AbstractAutoEncoderNetwork?");
        ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
        SDN_FATAL("Quitting");
      }
    } else {
      // Deep Loading
      XMLNode encNetworkNode = node.getChildNode("EncodingNetworkParameters");
      SDN_ASSERT_MSG(!encNetworkNode.isEmpty(), "You need to specify <EncodingNetworkParameters> !")
      encodingNetwork->loadXML(encNetworkNode); 
    }

    // ---------------------------------------------------

    // Load Supervised Part
    
    supervisedNetwork = boost::dynamic_pointer_cast<AbstractSupervisedNetwork> (AbstractNetworkRegistry().make(supervisedNetworkName));
    if (!supervisedNetwork) {
      SDN_ERROR("Could not load network with name " << supervisedNetworkName);
      SDN_ERROR("Is " << supervisedNetworkName << " an AbstractSupervisedNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

    // Deep Loading
    XMLNode supNetworkNode = node.getChildNode("SupervisedNetworkParameters");
    SDN_ASSERT_MSG(!supNetworkNode.isEmpty(), "You need to specify <SupervisedNetworkParameters> !")
    supervisedNetwork->loadXML(supNetworkNode); 

    
  }

  void SupervisedWithPreTrainingScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new SupervisedWithPreTrainingScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();
  }
  
  void SupervisedWithPreTrainingScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
    
    // set the output vector size
    targetOutput = sdn::Matrix::Zero(dataset->numLabels(), 1);
   
  }
  
  void SupervisedWithPreTrainingScript::initPersistentParameters() {
    
    // Load by XML should have setup the networks, here we link them up
    supervisedNetwork->useAutoEncoderNetwork(encodingNetwork);
    
  }
  
  void SupervisedWithPreTrainingScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void SupervisedWithPreTrainingScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // Setup target output
    targetOutput.fill(0.0);
    targetOutput(label, 0) = 1.0;
    
    // Setup input
    normInput.normalizePatch(inputExample);
    
    if (manager->currentIteration <= unsupervisedTrainingIterations) {
      // Still Doing Unsupervised Training
      encodingNetwork->unsupervisedTrain(inputExample);
    } else {
      // Supervised Training
      supervisedNetwork->supervisedTrain(inputExample, targetOutput);
      if (fineTuneUnsupervisedNetwork) {
        supervisedNetwork->fineTuneAutoEncoder(inputExample);
      }
    }
   
  }
  
  // Callbacks ---------------------------------------------------------
  
  void SupervisedWithPreTrainingScript::reportProgress(std::ostream& o) {

  }
  
  const boost::shared_ptr<CostFunction> SupervisedWithPreTrainingScript::getSupervisedCostFunction() {
    return supervisedNetwork->getCostFunction();
  }

  const sdn::Matrix& SupervisedWithPreTrainingScript::getPrediction() {
    // prevent crashes. return something that gives 0 error
    if (supervisedNetwork->getPrediction().rows() != targetOutput.rows() ||
        supervisedNetwork->getPrediction().cols() != targetOutput.cols())
      return targetOutput;
    else
      return supervisedNetwork->getPrediction();
  }
  
  const sdn::Matrix& SupervisedWithPreTrainingScript::getTarget() {
    return targetOutput;
  }
  
  const boost::shared_ptr<CostFunction> SupervisedWithPreTrainingScript::getReconstructionCostFunction() {
    return encodingNetwork->getCostFunction();
  }

  const sdn::Matrix& SupervisedWithPreTrainingScript::getReconstruction() {
    // prevent crashes. return something that gives 0 error
    if (encodingNetwork->getReconstruction().rows() != inputExample.rows() ||
        encodingNetwork->getReconstruction().cols() != inputExample.cols())
      return inputExample;
    else
      return encodingNetwork->getReconstruction();
  }
  
  const sdn::Matrix& SupervisedWithPreTrainingScript::getOriginal() {
    return inputExample;
  }
  
  // ---- Save Model Callback ----
  const void SupervisedWithPreTrainingScript::saveModelInDirectory(std::string dir) {
    encodingNetwork->quickSaveBinary(dir + "encNet_" + itos(manager->currentIteration, 10) + ".dat");
    supervisedNetwork->quickSaveBinary(dir + "supNet_" + itos(manager->currentIteration, 10) + ".dat");
  }
 
 
}
