//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: trainStackedAutoEncodersScript.cc
//Description: 

#include "../scriptRuntimeManager.h"
#include "../../registry/networkRegistry.h"
#include "trainStackedAutoEncodersScript.h"
#include "sdn/base.h"
#include "sdn/dataset.h"


namespace sdn {
  
  class TrainStackedAutoEncodersScriptConfigParser : public ConfigParser {

  protected:
    TrainStackedAutoEncodersScript & parentScript;

  public:
  
    TrainStackedAutoEncodersScriptConfigParser(TrainStackedAutoEncodersScript & parent) : parentScript(parent) {}
    ~TrainStackedAutoEncodersScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back(parentScript.className());
    }

    void readConfig(XMLNode & node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<" << parentScript.className() << " ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  TrainStackedAutoEncodersScript::TrainStackedAutoEncodersScript() {
    trainAll = false;
    stackedNetwork = boost::shared_ptr<StackedAutoEncoderNetwork> (new StackedAutoEncoderNetwork());
    loadUsingConfigManager();
  }
  
  TrainStackedAutoEncodersScript::~TrainStackedAutoEncodersScript() {
  }
  
  void TrainStackedAutoEncodersScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    
    p.parse(node);
    
    // Note that values are set *directly* into the network
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");

    // Deep Loading
    XMLNode stackedNetworkNode = node.getChildNode("StackedNetworkConfig");
    SDN_ASSERT_MSG(!stackedNetworkNode.isEmpty(), "You need to specify StackedNetworkConfig !")
    
    stackConfig.clear();
    
    // Load and Check Validity
    loadStackConfig(stackedNetworkNode);
    stackedNetwork->loadXML(stackedNetworkNode); // load costFunctionName

  }

  void TrainStackedAutoEncodersScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new TrainStackedAutoEncodersScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();
  }
  
  void TrainStackedAutoEncodersScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
  }
  
  void TrainStackedAutoEncodersScript::initPersistentParameters() {
  }
  
  void TrainStackedAutoEncodersScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void TrainStackedAutoEncodersScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    // Check if we need to add a network to the stack
    for(int i=0;i<stackConfig.size();i++) {
      if (stackConfig[i].addNetworkIteration == manager->currentIteration) {

        // Add any network on newNetwork to the stack
        if (newNetwork) {
          stackedNetwork->addAutoEncoder(newNetwork);
          SDN_MSG("Adding existing newNetwork onto the Network Stack.");
          SDN_MSG("Current Network Stack Size: " << stackedNetwork->networkStack.size());
        }

        // First we move whatever new network we were training 
        trainAll = stackConfig[i].trainAll;

        // Create Network
        newNetwork = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> 
                      (AbstractNetworkRegistry().make(stackConfig[i].networkName));
        SDN_ASSERT(newNetwork);

        SDN_MSG("Creating a new Network " << newNetwork->className());

        // Deep Loading
        newNetwork->loadXML(stackConfig[i].networkConfig); 
        
        // If trainAll, we add the network to the stack immediately
        if (trainAll) {
          stackedNetwork->addAutoEncoder(newNetwork);
          SDN_MSG("Adding newly created newNetwork onto the Network Stack (training all at once).");
          SDN_MSG("Current Network Stack Size: " << stackedNetwork->networkStack.size());
          // Empty the old pointer
          newNetwork.reset();
        }
          
      }
    }
    
    // Train!
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // Do fancy stuff here
    normInput.normalizePatch(inputExample);

    if (!trainAll) {
      // Encode with the stack
      // Then train the new network
      if (stackedNetwork->networkStack.size() > 0) {
        stackedNetwork->encode(inputExample);
        newNetwork->unsupervisedTrain(stackedNetwork->getHiddenRepresentation());
      } else {
        newNetwork->unsupervisedTrain(inputExample);
      }
    } else {
      // Encode it all the way and then backprop it 
      // The newNetwork should already be on the stack!
      stackedNetwork->unsupervisedTrain(inputExample);
    }
    
  }
  
  void TrainStackedAutoEncodersScript::loadStackConfig(XMLNode& node) {
    
    // We expect a series of 
    
    //  AddAutoEncoder 
    //    iteration="N" trainAll="0/1"
    //    With a child node NetworkParameters
    
    // This functions checks that the structure node follows the above structure
    // and that each network can be created successfully
    
    // We'll create a bunch of networks and throw them away
    
    // We need to create the network here because the XML parameters
    // for it need to be read directly to the network    
    
    for(int i=0;i<node.nChildNode();i++) {
      
      XMLNode newNetworkNode = node.getChildNode(i);
      
      // Check Node Name
      SDN_ASSERT(std::string(newNetworkNode.getName()) =="AddAutoEncoder");
      
      // Check Parameters
      AttributeParser p;
      p.addInt("iteration"); p.addBool("trainAll", false); p.addString("networkName");
      p.parse(newNetworkNode);
      
      long int addIter = p.getInt("iteration");
      bool enabletrainAll = p.getBool("trainAll");
      std::string tempNetworkName = p.getString("networkName");
      
      // Check Network Structure
      boost::shared_ptr<AbstractAutoEncoderNetwork> tempNetwork =
        boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> 
        (AbstractNetworkRegistry().make(tempNetworkName));
      if (!tempNetwork) {
        SDN_ERROR("Could not load network with name " << tempNetworkName);
        SDN_ERROR("Is " << tempNetworkName << " an AbstractAutoEncoderNetwork?");
        ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
        SDN_FATAL("Quitting");
      }

      // Deep Loading
      XMLNode tempNetworkNode = newNetworkNode.getChildNode("NetworkParameters");
      SDN_ASSERT_MSG(!tempNetworkNode.isEmpty(), "You need to specify NetworkParameters !")
      tempNetwork->loadXML(tempNetworkNode); 
      
      // All Operations Pass, add to stack
      stackConfig.push_back(newNetworkConfig(addIter, enabletrainAll, 
                                             tempNetworkName, tempNetworkNode));
    }
  }

  
  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> TrainStackedAutoEncodersScript::getReconstructionCostFunction() {
    if (trainAll) {
      return stackedNetwork->getCostFunction();
    } else {
      return newNetwork->getCostFunction();
    }
  }

  const sdn::Matrix& TrainStackedAutoEncodersScript::getReconstruction() {
    if (trainAll) {
      return stackedNetwork->getReconstruction();
    } else {
      return newNetwork->getReconstruction();
    }
  }
  
  const sdn::Matrix& TrainStackedAutoEncodersScript::getOriginal() {
    if (trainAll) {
      return inputExample;
    } else {
      if (stackedNetwork->networkStack.size() > 0) {
        return stackedNetwork->getHiddenRepresentation();
      } else {
        return inputExample;
      }
    }
  }
 
  // ---- Save Model Callback ----
  const void TrainStackedAutoEncodersScript::saveModelInDirectory(std::string dir) {
    stackedNetwork->quickSaveBinary(dir + "stackNet_" + itos(manager->currentIteration, 10) + ".dat");
    newNetwork->quickSaveBinary(dir + "currentNet_" + itos(manager->currentIteration, 10) + ".dat");
  }
 
 
}
