//Author: Jiquan Ngiam <jngiam@cs.stanford.edu>
//File: unsupervisedTrainingScript.cc
//Description: 

#include "unsupervisedTrainingScript.h"
#include "../scriptRuntimeManager.h"
#include "../../registry/networkRegistry.h"
#include "sdn/base.h"
#include "sdn/dataset.h"


namespace sdn {

  class UnsupervisedTrainingScriptConfigParser : public ConfigParser {

  protected:
    UnsupervisedTrainingScript & parentScript;

  public:
  
    UnsupervisedTrainingScriptConfigParser(UnsupervisedTrainingScript & parent) : parentScript(parent) {}
    ~UnsupervisedTrainingScriptConfigParser() {}
    
    void topics(std::vector<std::string> & topics) {
      // Only Have ONE topic
      // If not, loadXML in your Script will be called multiple times
      // This might be a concern
      topics.push_back("sdn::UnsupervisedTrainingScript");
    }

    void readConfig(XMLNode& node) {
      parentScript.loadXML(node);
    }
    
    void usage(std::ostream & os) const {
      os << "<sdn::UnsupervisedTrainingScript ...... script=\"<Scalar>\" network=\"<Scalar>\" \" />\n";
    }
  };
  
  // -------------------------------------------------------------------

  UnsupervisedTrainingScript::UnsupervisedTrainingScript() {
    loadUsingConfigManager();
  }
  
  UnsupervisedTrainingScript::~UnsupervisedTrainingScript() {
  }
  
  void UnsupervisedTrainingScript::loadXML(XMLNode& node) {
    
    AttributeParser p;
    
    p.addString("network");
    p.addString("dataset");
    p.addInt("inputHeight");
    p.addInt("inputWidth");
    
    p.parse(node);
    
    // Note that values are set *directly* into the network
    networkName = p.getString("network");
    datasetName = p.getString("dataset");
    inputHeight = p.getInt("inputHeight");
    inputWidth  = p.getInt("inputWidth");
    
    // We need to create the network here because the XML parameters
    // for it need to be read directly to the network    
    network = boost::dynamic_pointer_cast<AbstractAutoEncoderNetwork> (AbstractNetworkRegistry().make(networkName));
    if (!network) {
      SDN_ERROR("Could not load network with name " << networkName);
      SDN_ERROR("Is " << networkName << " an AbstractAutoEncoderNetwork?");
      ERROR_OUT_LIST_REGISTRY_NAMES(AbstractNetwork);
      SDN_FATAL("Quitting");
    }

    // Deep Loading
    XMLNode networkNode = node.getChildNode("NetworkParameters");
    SDN_ASSERT_MSG(!networkNode.isEmpty(), "You need to specify NetworkParameters !")
    network->loadXML(networkNode); 

  }

  void UnsupervisedTrainingScript::loadUsingConfigManager() {

    // The following line will load the XML config
    ConfigListener listener
      (boost::shared_ptr<ConfigParser>
        (new UnsupervisedTrainingScriptConfigParser(*this)));

    // We have this here instead of loadXML because loadXML might get
    // called multiple times by the configParser
    initPersistentParameters();
    initTransientParameters();

  }
  
  void UnsupervisedTrainingScript::initTransientParameters() {

    dataset =  DatasetRegistry().make(datasetName);
    if (!dataset) {
      SDN_ERROR("Could not load dataset with name " << datasetName);
      ERROR_OUT_LIST_REGISTRY_NAMES(Dataset);
      SDN_FATAL("Quitting");
    }
    dataset->requestExampleSize(inputHeight, inputWidth);

    // We are requesting column vector examples
    inputExample = sdn::Matrix::Zero(inputHeight*inputWidth, 1);
  }
  
  void UnsupervisedTrainingScript::initPersistentParameters() {
    network->initTransientParameters();
    network->initPersistentParameters();
  }
  
  void UnsupervisedTrainingScript::serialize(serial::Archive & ar) {
    SDN_FATAL(className() << "::" << __FUNCTION__ << " not implemented");
  }

  
  void UnsupervisedTrainingScript::executeOneIteration() {
    
    SDN_ASSERT(manager);
    
    int label = dataset->getRandomExampleAsColumn(inputExample);
    
    // do fancy stuff here
    normInput.normalizePatch(inputExample);
    network->unsupervisedTrain(inputExample);
    
  }
  
  // ---- Reconstruction Callback ----
  const boost::shared_ptr<CostFunction> UnsupervisedTrainingScript::getReconstructionCostFunction() {
    return network->getCostFunction();
  }

  const sdn::Matrix& UnsupervisedTrainingScript::getReconstruction() {
    return network->getReconstruction();
  }
  
  const sdn::Matrix& UnsupervisedTrainingScript::getOriginal() {
    return inputExample;
  }
 
  // ---- Save Model Callback ----
  const void UnsupervisedTrainingScript::saveModelInDirectory(std::string dir) {
    network->quickSaveBinary(dir + "net_" + itos(manager->currentIteration, 10) + ".dat");
  }
 
}
