/// \author Ian Goodfellow


#include "sdn/base.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include "gtest/gtest.h"
#include <string>


using namespace sdn;
using namespace std;

TEST(AttributeParser, CorrectlyFormatted)
{
	string xml = "<sampleNode color=\"red\" amount=\"12\" quantity=\"3.14\" flag=\"true\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);
	p.addDouble("quantity");
	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	p.parse(node);

	ASSERT_EQ(p.getInt("numStripes").newlyConfigured, false);
	ASSERT_EQ(p.getInt("numStripes").configured, false);
	ASSERT_EQ(p.getInt("numStripes").hasDefault, true);
	ASSERT_EQ(p.getInt("numStripes").defaultValue, 7);
	ASSERT_EQ(p.getInt("numStripes").value, 7);

	ASSERT_EQ(p.getDouble("quantity").value, 3.14);
	ASSERT_EQ(p.getDouble("quantity").hasDefault, false);

	ASSERT_EQ(p.getBool("flag").value,true);
	ASSERT_EQ(p.getBool("flag").configured,true);
	
	ASSERT_EQ(p.getInt("amount").value,12);
	ASSERT_EQ(p.getInt("amount").newlyConfigured,true);

	ASSERT_EQ(p.getString("color").value, "red");

	xml = "<nameDoesntMatter color=\"blue\" />";

	node = parseXML(xml);

	p.parse(node);

	ASSERT_EQ(p.getString("color").value, "blue");

	ASSERT_EQ(p.getString("color").newlyConfigured, true);
	ASSERT_EQ(p.getBool("flag").newlyConfigured, false);

	p.reset();

	ASSERT_EQ(p.getString("color").configured, false);
}

TEST(AttributeParser, BadAdd1)
{
  AttributeParser p;
  p.addInt("zebra");
  ASSERT_EXIT(p.addString("zebra","maybe"), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, BaddAdd2)
{
  AttributeParser p;
  p.addDouble("constant",0.5);
  ASSERT_EXIT(p.addBool("constant",true), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, BadGet)
{
	string xml = "<sampleNode color=\"red\" amount=\"12\" quantity=\"3.14\" flag=\"true\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);
	p.addDouble("quantity");
	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	p.parse(node);

	ASSERT_EXIT(p.getBool("numStripes"), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, BadInt)
{
	string xml = "<sampleNode color=\"red\" amount=\"12e\" quantity=\"3.14\" flag=\"true\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);
	p.addDouble("quantity");
	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	ASSERT_EXIT(p.parse(node), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, BadDouble)
{
	string xml = "<sampleNode color=\"red\" amount=\"12\" quantity=\"3.14 george\" flag=\"true\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);
	p.addDouble("quantity");
	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	ASSERT_EXIT(p.parse(node), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, BadBool)
{
	string xml = "<sampleNode color=\"red\" amount=\"12\" quantity=\"3.14\" flag=\"2\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);
	p.addDouble("quantity");
	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	ASSERT_EXIT(p.parse(node), ::testing::KilledBySignal(SIGABRT), "");
}

TEST(AttributeParser, UndeclaredAttribute)
{
	string xml = "<sampleNode color=\"red\" amount=\"12\" quantity=\"3.14\" flag=\"1\" />";

	XMLNode node = parseXML(xml);

	AttributeParser p;
	p.addInt("numStripes",7);

	p.addBool("flag");
	p.addInt("amount");
	p.addString("color");
	
	ASSERT_EXIT(p.parse(node), ::testing::KilledBySignal(SIGABRT), "");
}
