#include "sdn/cuda.h"
#include "gtest/gtest.h"
#include <fstream>
#include <float.h>
#include <stdio.h>
#include <cublas.h>

using namespace sdn;

TEST(CublasInit, Initialization) {
  CudaMatrix::Ones(10, 10) + CudaMatrix::Zero(10, 10);
}

TEST(CudaMatrix, AllocationBenchmark) {
  for(int i = 0; i < 1000; i++)
    CudaMatrix::Ones(1000, 1000);
}

TEST(CudaMatrix, GettingSetting) {
  CudaMatrix m(64, 64);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      m(i,j) = i*i-j*j;

  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(i*i-j*j, m(i,j));
}

TEST(CudaMatrix, StaticAllocators) {
  CudaMatrix m = CudaMatrix::Zero(1000, 1000);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(0, m(i,j));
  
  m = CudaMatrix::Ones(999, 1001);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(1, m(i,j));

  m = CudaMatrix::Constant(1001, 999, 3.14);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(3.14, m(i,j));

  m = CudaMatrix::Random(1000, 1000);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_NEAR(0, m(i,j), 1);
}

TEST(CudaMatrix, AdditionSubtraction) {
  CudaMatrix m1(641, 641);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = i+j+1;

  CudaMatrix m2(641, 641);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      m2(i,j) = i*j;

  CudaMatrix sum = m1 + m2;
  for(int i = 0; i < sum.rows(); i++)
    for(int j = 0; j < sum.cols(); j++)
      ASSERT_FLOAT_EQ(i+j+1+i*j, sum(i,j));

  CudaMatrix diff = m2 - m1;
  for(int i = 0; i < diff.rows(); i++)
    for(int j = 0; j < diff.cols(); j++)
      ASSERT_FLOAT_EQ(i*j-i-j-1, diff(i,j));
}

TEST(CudaMatrix, Multiplication) {
  CudaMatrix m1(1024, 256);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = j / 3.0;

  CudaMatrix m2(256, 256);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      m2(i,j) = j * 2.0;

  CudaMatrix m3(256, 1024);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      m3(i,j) = i;

  CudaMatrix prod = m1 * m2;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m1.cols()*(m1.cols()-1)*j/3.0, prod(i,j));

  prod = m2 * m3;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m2.cols()*(m2.cols()-1)*(2*m2.cols()-1)/3.0, prod(i,j));
}

TEST(CudaMatrix, Transpose) {
  CudaMatrix m1(256, 256);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = i;

  CudaMatrix m2(128, 256);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      m2(i,j) = j;

  CudaMatrix m3(128, 256);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      m3(i,j) = i;

  CudaMatrix prod = m2 * m1.transpose();
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m2.cols()*(m2.cols()-1)*j/2.0, prod(i,j));

  prod = m3.transpose() * m2;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m3.rows()*(m3.rows()-1)*j/2.0, prod(i,j));

  prod = prod.transpose();
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m3.rows()*(m3.rows()-1)*i/2.0, prod(i,j));

  CudaMatrix m4 = CudaMatrix::Ones(256, 128);
  CudaMatrix sum = m4.transpose() + m3;
  for(int i = 0; i < sum.rows(); i++)
    for(int j = 0; j < sum.cols(); j++)
      ASSERT_FLOAT_EQ(i+1, sum(i,j));
      
  CudaMatrix m5(3, 5);
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      m5(i, j) = i*j+3;
  m5.transposeThis();
  m5.forceTransposeCPU();
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_FLOAT_EQ(j*i+3, m5(i, j));
}

TEST(CudaMatrix, ScalarMultiplication) {
  CudaMatrix m(512, 256);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      m(i,j) = i+j;
  
  m = m * 1.5;
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(1.5*(i+j), m(i,j));

  m = 0.5 * m;
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      ASSERT_FLOAT_EQ(0.75*(i+j), m(i,j));
}

TEST(CudaMatrix, MixedBasic) {
  CudaMatrix A(5, 5);
  CudaMatrix B(5, 3);
  CudaMatrix x1(5, 1);
  CudaMatrix x2(5, 1);
  CudaMatrix x3(5, 1);
  CudaMatrix y(3, 1);

  A(0, 0) = 2; A(0, 1) = 3; A(0, 2) = 1; A(0, 3) = 6; A(0, 4) = 6;
  A(1, 0) = 8; A(1, 1) = 3; A(1, 2) = 3; A(1, 3) = 7; A(1, 4) = 6;
  A(2, 0) = 2; A(2, 1) = 2; A(2, 2) = 7; A(2, 3) = 6; A(2, 4) = 9;
  A(3, 0) = 2; A(3, 1) = 3; A(3, 2) = 1; A(3, 3) = 1; A(3, 4) = 6;
  A(4, 0) = 7; A(4, 1) = 10; A(4, 2) = 1; A(4, 3) = 5; A(4, 4) = 3;

  B(0, 0) = 4; B(0, 1) = 4; B(0, 2) = 5;
  B(1, 0) = 3; B(1, 1) = 5; B(1, 2) = 4;
  B(2, 0) = 5; B(2, 1) = 3; B(2, 2) = 4;
  B(3, 0) = 4; B(3, 1) = 3; B(3, 2) = 3;
  B(4, 0) = 4; B(4, 1) = 3; B(4, 2) = 3;

  y(0, 0) = 2;
  y(1, 0) = 1;
  y(2, 0) = -2;

  x1(0, 0) = -4; x2(0, 0) = 1; x3(0, 0) = -2;
  x1(1, 0) = 0; x2(1, 0) = 1; x3(1, 0) = 1;
  x1(2, 0) = 1; x2(2, 0) = 2; x3(2, 0) = -1;
  x1(3, 0) = 5; x2(3, 0) = 2; x3(3, 0) = -3;
  x1(4, 0) = -4; x2(4, 0) = 3; x3(4, 0) = -4;

  CudaMatrix x = x1 + x2 - 2 * x3;
  ASSERT_FLOAT_EQ(1, x(0, 0));
  ASSERT_FLOAT_EQ(-1, x(1, 0));
  ASSERT_FLOAT_EQ(5, x(2, 0));
  ASSERT_FLOAT_EQ(13, x(3, 0));
  ASSERT_FLOAT_EQ(7, x(4, 0));

  CudaMatrix result = x.transpose() * A * B * y;
  ASSERT_FLOAT_EQ(1885, result(0, 0));

  CudaMatrix result2 = 3.14 * result;
  ASSERT_FLOAT_EQ(5918.9, result2(0, 0));
}

TEST(CudaMatrix, CudaMatrixAsDiagonal) {
  CudaMatrix rowMatrix(1, 23);
  CudaMatrix colMatrix(23, 1);

  for(int i = 0; i < 23; i++) {
    rowMatrix(0,i) = i;
    colMatrix(i,0) = i;
  }

  CudaMatrix diagonalFromRow = rowMatrix.asDiagonal();
  CudaMatrix diagonalFromCol = colMatrix.asDiagonal();

  for(int i = 0; i < 23; i++) {
    for(int j = 0; j < 23; j++) {
      if(i != j) {
        ASSERT_FLOAT_EQ(0, diagonalFromRow(i,j));
        ASSERT_FLOAT_EQ(0, diagonalFromCol(i,j));
      }
      else {
        ASSERT_FLOAT_EQ(i, diagonalFromRow(i,j));
        ASSERT_FLOAT_EQ(i, diagonalFromCol(i,j));
      }
    }
  }

  CudaMatrix m = CudaMatrix::Random(300, 400);
  diagonalFromRow = m.row(10).asDiagonal();
  for(int i = 0; i < diagonalFromRow.rows(); i++)
    for(int j = 0; j < diagonalFromRow.cols(); j++)
      ASSERT_FLOAT_EQ((i == j ? m(10,i) : 0), diagonalFromRow(i,j));
  
  diagonalFromCol = m.col(250).asDiagonal();
  for(int i = 0; i < diagonalFromCol.rows(); i++)
    for(int j = 0; j < diagonalFromCol.cols(); j++)
      ASSERT_FLOAT_EQ((i == j ? m(i,250) : 0), diagonalFromCol(i,j));

  m = (m + CudaMatrix::Random(400, 300).transpose()).transpose();
  diagonalFromRow = m.row(300).asDiagonal();
  for(int i = 0; i < diagonalFromRow.rows(); i++)
    for(int j = 0; j < diagonalFromRow.cols(); j++)
      ASSERT_FLOAT_EQ((i == j ? m(300,i) : 0), diagonalFromRow(i,j));
  
  diagonalFromCol = m.col(120).asDiagonal();
  for(int i = 0; i < diagonalFromCol.rows(); i++)
    for(int j = 0; j < diagonalFromCol.cols(); j++)
      ASSERT_FLOAT_EQ((i == j ? m(i,120) : 0), diagonalFromCol(i,j));
}

TEST(CudaMatrix, BlockOperations) {
  CudaMatrix m1 = CudaMatrix::Ones(100, 200);
  CudaMatrix m2 = CudaMatrix::Zero(200, 300);

  m1.block(10, 10, 20, 30) = m2.block(100, 200, 20, 30);

  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      ASSERT_FLOAT_EQ((i >= 10 && i < 30 && j >= 10 && j < 40) ? 0 : 1,
		      m1(i,j));

  m2.block(20, 50, 10, 20).transpose() = m1.block(50, 60, 20, 10);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ((i >= 20 && i < 30 && j >= 50 && j < 70) ? 1 : 0,
  		      m2(i,j));

  CudaMatrix m3 = m2.block(10, 40, 20, 30);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ((i >= 10 && i < 20 && j >= 10 && j < 30) ? 1 : 0,
		      m3(i,j));
  
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = i+j;

  m3.row(5) = m1.block(10, 10, m3.cols(), 1).transpose();
  for(int i = 0; i < m3.cols(); i++)
    ASSERT_FLOAT_EQ(20+i, m3(5,i));

  m2.block(2, 4, m3.rows(), 1).transpose() = m3.col(2).transpose();
  for(int i = 0; i < m3.rows(); i++)
    ASSERT_FLOAT_EQ((i == 5 ? 22 : 0), m2(2+i,4));

  m1 = CudaMatrix::Random(100, 200);
  m2 = CudaMatrix::Random(200, 300);
  m3 = CudaMatrix::Ones(100, 100);
  
  m3.block(10, 20, 30, 40) += m2.block(20, 30, 30, 40);
  for(int i = 10; i < 40; i++)
    for(int j = 20; j < 60; j++)
      ASSERT_FLOAT_EQ(1+m2(10+i,10+j), m3(i,j));

  m3.block(10, 20, 30, 40) -= m1.block(30, 40, 30, 40);
  for(int i = 10; i < 40; i++)
    for(int j = 20; j < 60; j++)
      ASSERT_FLOAT_EQ(1+m2(10+i,10+j)-m1(20+i,20+j), m3(i,j));

  m3.block(10, 20, 30, 40) = CudaMatrix::Ones(30, 40);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(1, m3(i,j));

  m3.block(20, 30, 40, 50).transpose().cwise() *= m1.block(50, 50, 50, 40);
  for(int i = 20; i < 60; i++)
    for(int j = 30; j < 80; j++)
      ASSERT_FLOAT_EQ(m1(20+j,30+i), m3(i,j));

  m3.block(20, 30, 40, 50).cwise() /= m2.block(10, 10, 50, 40).transpose();
  for(int i = 20; i < 60; i++)
    for(int j = 30; j < 80; j++)
      ASSERT_FLOAT_EQ(m1(20+j,30+i) / m2(-20+j,-10+i), m3(i,j));

  m3.block(20, 30, 40, 50).transpose() = CudaMatrix::Zero(50, 40);
  m3.block(20, 30, 40, 50) += 1.0;
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(1, m3(i,j));

  m3.block(40, 10, 20, 20).transpose() -= 2.0;
  for(int i = 40; i < 60; i++)
    for(int j = 10; j < 20; j++)
      ASSERT_FLOAT_EQ(-1, m3(i,j));

  m3.block(40, 10, 20, 20) *= 3.14;
  for(int i = 40; i < 60; i++)
    for(int j = 10; j < 20; j++)
      ASSERT_FLOAT_EQ(-3.14, m3(i,j));

  m1 = CudaMatrix::Zero(100, 100);
  m1.col(20) = m3.row(10).transpose() * 3.0;
  for(int i = 0; i < m1.rows(); i++)
    ASSERT_FLOAT_EQ(m3(10,i) * 3.0, m1(i,20));

  m1.row(30).transpose() = (3.4 + m3.col(20).cwise()) / 2.0;
  for(int i = 0; i < m1.rows(); i++)
    ASSERT_FLOAT_EQ((3.4 + m3(i,20)) / 2.0, m1(30,i));

  m2 = CudaMatrix::Ones(100, 120);
  m2 = m2.block(10, 20, 30, 40);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(m2(i,j), 1.0);
}

TEST(CudaMatrix, CwiseOperations) {
  CudaMatrix m1(30, 40);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = i+j;

  CudaMatrix m2 = m1.cwise() * 123.45;
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(123.45*(i+j), m2(i,j));

  CudaMatrix m3 = 123.45 * m1.cwise();
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(m2(i,j), m3(i,j));

  m2 = m2.transpose().cwise() + 999;
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(123.45*(i+j)+999, m2(i,j));

  m3 = 999 + m3.transpose().cwise();
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(m2(i, j), m3(i, j));

  m2 = m2.transpose().cwise() - 111;
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(123.45*(i+j)+888, m2(i,j));

  m3 = 111 - m3.transpose().cwise();
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(-m2(i,j), m3(i,j));

  m2 = m2.cwise() / 10;
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(12.345*(i+j)+88.8, m2(i,j));

  m3 = m3.cwise() * m2 / 10;
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_FLOAT_EQ(-m2(i,j)*m2(i,j), m3(i,j));

  CudaMatrix m4 = m1.transpose().cwise().exp();
  for(int i = 0; i < m4.rows(); i++)
    for(int j = 0; j < m4.cols(); j++)
      ASSERT_NEAR(1, exp(1.0*i+j)/m4(i,j), 1e-5);

  CudaMatrix m5 = m2.transpose().cwise() * m4;
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_NEAR(1, exp(1.0*i+j)*(12.345*(i+j)+88.8)/m5(i,j), 1e-5);

  m5 = m3.cwise() + m4.transpose();
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_NEAR(1, (exp(1.0*i+j)-m2(i,j)*m2(i,j))/m5(i,j), 1e-5);

  m5 = m5.cwise().abs();
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_NEAR(1, fabs(exp(1.0*i+j)-m2(i,j)*m2(i,j))/m5(i,j), 1e-5);

  m5 = m2.cwise().square();
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_FLOAT_EQ(m2(i,j)*m2(i,j), m5(i,j));

  m5 = m1.cwise().pow(0.5);
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_FLOAT_EQ(sqrt(1.0*i+j), m5(i,j));

  m5 = (m1.cwise() + 1).cwise().inverse();
  for(int i = 0; i < m5.rows(); i++)
    for(int j = 0; j < m5.cols(); j++)
      ASSERT_FLOAT_EQ(1.0/(1.0*i+j+1.0), m5(i,j));
}

TEST(CudaMatrix, MixedBlockCwise) {
  CudaMatrix m1(100, 250);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = j;

  CudaMatrix m2(100, 100);
  m2.setConstant(0.3);

  CudaMatrix m3 = (m1.transpose().cwise() + 7) * (m2.cwise() + 1);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_NEAR(1, 130*(i+7) / m3(i,j), 1e-5);

  m2 = m1.block(10, 90, 10, 80).cwise() / m3.block(20, 30, 80, 10).transpose();
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ(m1(10+i,90+j)/m3(20+j,30+i), m2(i,j));

  m2 = ((0.7 * m1.cwise()).cwise() + 2.4).transpose().cwise() * (m3 / 3.3);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      ASSERT_FLOAT_EQ((0.7*m1(j,i)+2.4)*(m3(i,j)/3.3), m2(i,j));

  m3 = 0.5 * m3 * (m1.cwise() - 1.4) / 4.7;
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      ASSERT_NEAR(1, 6500*(i+7)*(j-1.4)/4.7/m3(i,j), 1e-5);
}

TEST(CudaMatrix, FullReductions) {
  CudaMatrix m1 = CudaMatrix::Ones(1000, 2500);
  ASSERT_FLOAT_EQ(1000*2500, m1.sum());
  ASSERT_FLOAT_EQ(sqrt(1000*2500), m1.norm());
  ASSERT_FLOAT_EQ(1, m1.minCoeff());
  ASSERT_FLOAT_EQ(1, m1.maxCoeff());

  // Kahan summation for reference sum
  m1 = CudaMatrix::Random(100, 300).transpose();
  float sum = 0, sum1 = 0, c = 0, c1 = 0, y, t, minv = FLT_MAX, maxv = FLT_MIN;
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++) {
      y = m1(i,j) - c;
      t = sum + y;
      c = (t - sum) - y;
      sum = t;
      y = m1(i,j)*m1(i,j) - c1;
      t = sum1 + y;
      c1 = (t - sum1) - y;
      sum1 = t;
      minv = fmin(minv, m1(i,j));
      maxv = fmax(maxv, m1(i,j));
    }
  ASSERT_NEAR(1, sum/m1.sum(), 1e-4);
  ASSERT_NEAR(1, sum/m1.transpose().sum(), 1e-4);
  ASSERT_NEAR(1, sqrt(sum1)/m1.norm(), 1e-4);
  ASSERT_NEAR(1, sqrt(sum1)/m1.transpose().norm(), 1e-4);
  ASSERT_FLOAT_EQ(minv, m1.minCoeff());
  ASSERT_FLOAT_EQ(minv, m1.transpose().minCoeff());
  ASSERT_FLOAT_EQ(maxv, m1.maxCoeff());
  ASSERT_FLOAT_EQ(maxv, m1.transpose().maxCoeff());

  sum = 0; sum1 = 0; c = 0; c1 = 0; minv = FLT_MAX; maxv = FLT_MIN;
  for(int i = 30; i < 190; i++)
    for(int j = 10; j < 80; j++) {
      y = m1(i,j) - c;
      t = sum + y;
      c = (t - sum) - y;
      sum = t;
      y = m1(i,j)*m1(i,j) - c1;
      t = sum1 + y;
      c1 = (t - sum1) - y;
      sum1 = t;
      minv = fmin(minv, m1(i,j));
      maxv = fmax(maxv, m1(i,j));
    }
  ASSERT_NEAR(1, sum/m1.block(30, 10, 160, 70).transpose().sum(), 1e-4);
  ASSERT_NEAR(1, sqrt(sum1)/m1.block(30, 10, 160, 70).transpose().norm(), 1e-4);
  ASSERT_FLOAT_EQ(minv, m1.block(30, 10, 160, 70).transpose().minCoeff());
  ASSERT_FLOAT_EQ(maxv, m1.block(30, 10, 160, 70).transpose().maxCoeff());
}

TEST(CudaMatrix, PartialReductions) {
  CudaMatrix m = CudaMatrix::Ones(1000, 800);
  CudaMatrix colwise = m.colwise().sum();
  for(int i = 0; i < colwise.cols(); i++)
    ASSERT_NEAR(1, m.rows()/colwise(0,i), 1e-6);
  
  CudaMatrix rowwise = m.rowwise().sum();
  for(int i = 0; i < rowwise.rows(); i++)
  ASSERT_NEAR(1, m.cols()/rowwise(i,0), 1e-6);

  // Kahan summation for reference sum
  m = CudaMatrix::Random(300, 500);
  colwise = m.block(100, 200, 50, 150).cwise().abs().colwise().sum();
  float ref[1024], y, t, c;
  for(int i = 0; i < colwise.cols(); i++) {
    ref[i] = 0; c = 0;
    for(int j = 100; j < 150; j++) {
      y = fabs(m(j,200+i)) - c;
      t = ref[i] + y;
      c = (t - ref[i]) - y;
      ref[i] = t;
    }
    ASSERT_NEAR(1, ref[i]/colwise(0,i), 1e-6);
  }

  rowwise = m.block(200, 50, 100, 300).cwise().exp().rowwise().sum();
  for(int i = 0; i < rowwise.rows(); i++) {
    ref[i] = 0; c = 0;
    for(int j = 50; j < 350; j++) {
      y = exp(m(200+i,j)) - c;
      t = ref[i] + y;
      c = (t - ref[i]) - y;
      ref[i] = t;
    }
    ASSERT_NEAR(1, ref[i]/rowwise(i,0), 1e-6);
  }

  colwise = m.transpose().colwise().sum();
  for(int i = 0; i < colwise.cols(); i++) {
    ref[i] = 0; c = 0;
    for(int j = 0; j < m.cols(); j++) {
      y = m(i,j) - c;
      t = ref[i] + y;
      c = (t - ref[i]) - y;
      ref[i] = t;
    }
    ASSERT_NEAR(1, ref[i]/colwise(0,i), 1e-6);
  }

  m = CudaMatrix::Zero(9, 1);
  m << 0.00894616, 0.0110368, 0.0821217, 0.113627, 0.0296039, 0.0342584,
    0.393644, 0.00613252, 0.164561;
  ASSERT_NEAR(1, m.colwise().sum()(0,0)/0.843932, 1e-6);
}

TEST(CudaMatrix, Comparison) {
  CudaMatrix m1 = CudaMatrix::Constant(200, 300, 314);
  CudaMatrix m2 = CudaMatrix::Constant(200, 100, 3.14);
  ASSERT_EQ(m1, m2 * CudaMatrix::Ones(100, 300));

  m2.cwise() += 1e-6;
  ASSERT_EQ(m1, m2 * CudaMatrix::Ones(100, 300));

  m2.cwise() -= 1e-4;
  ASSERT_TRUE(m1 != m2 * CudaMatrix::Ones(100, 300));

  m1 = CudaMatrix::Random(100, 100);
  ASSERT_TRUE(m1 != m2);

  m2 = CudaMatrix::Random(100, 100);
  ASSERT_FALSE(m1 == m2);

  m2 = m1;
  ASSERT_EQ(m1, m2);
  ASSERT_EQ(m1.block(10, 10, 20, 20), m2.block(10, 10, 20, 20));
  ASSERT_EQ(m1.block(10, 10, 20, 20).transpose(),
	    m2.block(10, 10, 20, 20).transpose());
}

TEST(CudaMatrix, CommaInitializer) {
  CudaMatrix m(6, 5);
  m <<
    1, 2, 3, 4, 5,
    6, 7, 8, 9, 10,
    11, 12, 13, 14, 15,
    16, 17, 18, 19, 20,
    21, 22, 23, 24, 25,
    26, 27, 28, 29, 30;

  for(int i = 0; i < 6; i++)
    for(int j = 0; j < 5; j++)
      ASSERT_FLOAT_EQ(5*i+j+1, m(i,j));
}

// benchmarking test
/*
TEST(CudaMatrix, TimingTest) {
  std::ofstream out("cudaMatrixTiming.csv");
  for(int size = 16; size <= 4096; size+=16) {
    clock_t start = clock();
    for(int i = 0; i < 25; i++) {
      CudaMatrix A = CudaMatrix::Ones(size, size);
      CudaMatrix B = CudaMatrix::Ones(size, size);
      CudaMatrix C = CudaMatrix::Ones(size, size);
      CudaMatrix D = A * B + C;
      ASSERT_FLOAT_EQ(1.0*size + 1, D(0, 0));
    }
    clock_t diff = clock() - start;
    double ms = (1000.0 * diff) / CLOCKS_PER_SEC;
    out << size << "," << ms << std::endl;
    if (size % 256 == 0) std::cout << size << std::endl;
  }
  out.close();
}
*/

TEST(CudaVector, StaticAllocators) {
  CudaVector v = CudaVector::Zero(100000);
  for(int i = 0; i < v.size(); i++)
    ASSERT_FLOAT_EQ(0, v(i));

  v = CudaVector::Ones(100001);
  for(int i = 0; i < v.size(); i++)
    ASSERT_FLOAT_EQ(1, v(i));
}

TEST(CudaVector, MatrixToVector) {
  CudaMatrix m = CudaMatrix::Random(100, 200);
  CudaVector v = m.col(10);

  for(int i = 0; i < v.size(); i++)
    ASSERT_FLOAT_EQ(m(i,10), v(i));

  v = m.row(40);
  for(int i = 0; i < v.size(); i++)
    ASSERT_FLOAT_EQ(m(40,i), v(i));

  m = v.transpose() * v;
  ASSERT_NEAR(1, v.norm()*v.norm()/m(0,0), 1e-5);
}

TEST(CudaVector, MixedBasic) {
  CudaMatrix m1 = CudaMatrix::Ones(256, 64);
  CudaVector v1(64);
  for(int i = 0; i < v1.size(); i++)
    v1(i) = i;

  CudaVector v2 = 2.0 * m1 * v1;
  for(int i = 0; i < v2.size(); i++)
    ASSERT_FLOAT_EQ(v1.size()*(v1.size()-1), v2(i));
  
  CudaMatrix m2 = CudaMatrix::Ones(64, 256);
  CudaVector v3 = CudaVector::Ones(64);
  CudaMatrix m3 = v1.transpose() * m2 * m1 * v3;
  ASSERT_FLOAT_EQ(m1.rows()*v3.size()*v1.size()*(v1.size()-1)/2.0, m3(0,0));

  v2 = (4.0 * (v1 + v3).transpose() * m2).transpose() - CudaVector::Zero(256);
  for(int i = 0; i < v2.size(); i++)
    ASSERT_FLOAT_EQ(2.0*v1.size()*(v1.size()+1), v2[i]);
}
