#include "sdn/cuda.h"
#include "gtest/gtest.h"
#include <fstream>
#include <float.h>
#include <stdio.h>
#include <cublas.h>
#include <time.h>
#include <stdlib.h>
#include <vector>

using namespace sdn;

TEST(CublasInit, Initialization) {
  CudaMatrix::Ones(10, 10) + CudaMatrix::Zero(10, 10);
}

TEST(SparseCudaMatrix, Constructor) {
  CudaMatrix m(64, 64);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      m(i,j) = i*i-j*j+1;
      
  SparseCudaMatrix ms1(m);
  for(int i = 0; i < ms1.rows(); i++)
    for(int j = 0; j < ms1.cols(); j++)
      ASSERT_FLOAT_EQ(i*i-j*j+1, ms1(i, j));
      
  SparseCudaMatrix ms2;
  ms2 = m;
  for(int i = 0; i < ms2.rows(); i++)
    for(int j = 0; j < ms2.cols(); j++)
      ASSERT_FLOAT_EQ(i*i-j*j+1, ms2(i, j));

  for(int i = 0; i < ms1.rows(); i++)
    for(int j = 0; j < ms1.cols(); j++)
      ASSERT_FLOAT_EQ(ms1(i, j), ms2(i, j));
}

TEST(SparseCudaMatrix, AdditionSubtraction) {
  int k = 641/2;
  CudaMatrix m1(641, 641);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = (i-k)+(j-k)+1;

  CudaMatrix m2(641, 641);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++) {
      if (m1(i, j) == 0) m2(i, j) = 0;
      else m2(i, j) = (i+1)*(j+1);
    }

  SparseCudaMatrix ms1(m1), ms2(m2);
  SparseCudaMatrix sum;
  sum = ms1 + ms2;

  for(int i = 0; i < sum.rows(); i++)
    for(int j = 0; j < sum.cols(); j++)
      ASSERT_FLOAT_EQ(m1(i, j)+m2(i, j), sum(i,j));

  SparseCudaMatrix diff = ms2 - ms1;
  for(int i = 0; i < diff.rows(); i++)
    for(int j = 0; j < diff.cols(); j++)
      ASSERT_FLOAT_EQ(m2(i, j)-m1(i, j), diff(i,j));
}

TEST(SparseCudaMatrix, CWiseOperations) {
  int k = 641/2;
  CudaMatrix m1(641, 641);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = (i-k)+(j-k)+1;

  CudaMatrix m2(641, 641);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++) {
      if (m1(i, j) == 0) m2(i, j) = 0;
      else m2(i, j) = (i+1)*(j+1);
    }

  SparseCudaMatrix ms1(m1), ms2(m2);
  SparseCudaMatrix prod;
  prod = ms1.cwise() * ms2;

  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m1(i, j)*m2(i, j), prod(i, j));
}

TEST(SparseCudaMatrix, TransposeThis) {
  CudaMatrix m(3, 2);
  m(0, 0) = 1; m(0, 1) = 2;
  m(1, 0) = -1; m(1, 1) = 5;
  m(2, 0) = -2; m(2, 1) = 314;

  SparseCudaMatrix sm(m);

  sm.transposeThis();
  for(int i = 0; i < sm.rows(); i++)
    for(int j = 0; j < sm.cols(); j++)
      ASSERT_FLOAT_EQ(sm(i, j), m(j, i));
      
  sm.transposeThis();
  for(int i = 0; i < sm.rows(); i++)
    for(int j = 0; j < sm.cols(); j++)
      ASSERT_FLOAT_EQ(sm(i, j), m(i, j));
}

TEST(SparseCudaMatrix, Multiplication) {
  srand(time(NULL));
  CudaMatrix m1(1024, 256);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = (rand()%5)-2;

  CudaMatrix m2(256, 256);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      m2(i,j) = (rand()%7)-3;

  CudaMatrix m3(256, 1024);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      m3(i,j) = i;

  SparseCudaMatrix sm1(m1), sm2(m2), sm3(m3);

  CudaMatrix prod = m1 * m2;
  SparseCudaMatrix sprod = m1 * m2;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(prod(i, j), sprod(i, j));

  sprod = sm1 * sm2;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(prod(i, j), sprod(i, j));

  prod = m2 * m3;
  sprod = m2 * m3;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(prod(i, j), sprod(i, j));
  sprod = sm2 * sm3;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(prod(i, j), sprod(i, j));
}

TEST(SparseCudaMatrix, Transpose) {
  CudaMatrix m1(256, 256);
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++)
      m1(i,j) = i;

  CudaMatrix m2(128, 256);
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++)
      m2(i,j) = j;

  CudaMatrix m3(128, 256);
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      m3(i,j) = i;

  SparseCudaMatrix sm1(m1), sm2(m2), sm3(m3);
  SparseCudaMatrix prod = sm2 * sm1.transpose();
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m2.cols()*(m2.cols()-1)*j/2.0, prod(i,j));

  prod = sm3.transpose() * sm2;
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m3.rows()*(m3.rows()-1)*j/2.0, prod(i,j));

  prod = prod.transpose();
  for(int i = 0; i < prod.rows(); i++)
    for(int j = 0; j < prod.cols(); j++)
      ASSERT_FLOAT_EQ(m3.rows()*(m3.rows()-1)*i/2.0, prod(i,j));
}

TEST(SparseCudaMatrix, ScalarMultiplication) {
  CudaMatrix m(512, 256);
  for(int i = 0; i < m.rows(); i++)
    for(int j = 0; j < m.cols(); j++)
      m(i,j) = i-j;
  
  SparseCudaMatrix sm(m);
  sm = sm * 1.5;
  for(int i = 0; i < sm.rows(); i++)
    for(int j = 0; j < sm.cols(); j++)
      ASSERT_FLOAT_EQ(1.5*(i-j), sm(i,j));

  sm = 0.5 * sm;
  for(int i = 0; i < sm.rows(); i++)
    for(int j = 0; j < sm.cols(); j++)
      ASSERT_FLOAT_EQ(0.75*(i-j), sm(i,j));
}

TEST(SparseCudaMatrix, SpecialOperator) {
  CudaMatrix m1(123, 84);
  CudaMatrix m2(m1.cols(), 863);
  CudaMatrix m3(m1.rows(), m2.cols());
  
  CudaMatrix m4(m1.rows(), m1.cols());
  CudaMatrix m5(m2.rows(), m2.cols());

  srand(time(NULL));
  for(int i = 0; i < m1.rows(); i++)
    for(int j = 0; j < m1.cols(); j++) {
      m1(i, j) = (rand()%153)-76;
      m4(i, j) = (rand()%153)-76;
    }
      
  for(int i = 0; i < m2.rows(); i++)
    for(int j = 0; j < m2.cols(); j++) {
      m2(i, j) = (rand()%283)-141;
      m5(i, j) = (rand()%283)-141;
    }
      
  for(int i = 0; i < m3.rows(); i++)
    for(int j = 0; j < m3.cols(); j++)
      m3(i, j) = (rand()%7)-3;

  SparseCudaMatrix sm3(m3);
  CudaMatrix res = m3.cwise()*(m1*m2);
  SparseCudaMatrix sres = denseTimesDenseCWiseTimesSparse(m1, m2, sm3);
  
  for(int i = 0; i < sres.rows(); i++)
    for(int j = 0; j < sres.cols(); j++)
      ASSERT_FLOAT_EQ(res(i, j), sres(i, j));
      
  SparseCudaMatrix sres2 = denseTimesDenseCWiseTimesSparse(m4, m5, sm3);
  sres+sres2;
}

TEST(SparseCudaMatrix, SpeedBenchmarking) {
  char mat3[]="/home/jaehyun/trunk/tests/lib/cuda/test_sparse_matrices/sparse_matrix_2.txt";
  FILE *fp = fopen(mat3, "r");
  std::vector<sparseEntry> entries;
  int i, j;
  float v;
  int m, n;
  fscanf(fp, "%d%d", &m, &n);
  while(fscanf(fp, "%d%d%f", &i, &j, &v)==3) {
    sparseEntry newEntry;
    newEntry.row = i;
    newEntry.col = j;
    newEntry.val = v;
    entries.push_back(newEntry);
  }
  fclose(fp);
  
  SparseCudaMatrix sm3(m, n, entries);
  int iter = 10;
  std::vector<double> times;
  for (int n_ex = 20; n_ex <= 1000; n_ex += 20) {
    double total = 0;
    for (i=0; i<iter; i++) {
      CudaMatrix m1=CudaMatrix::Random(sm3.rows(), n_ex /**/);
      CudaMatrix m2=CudaMatrix::Random(m1.cols(), sm3.cols());
      double start_time = clock();
      SparseCudaMatrix res = denseTimesDenseCWiseTimesSparse(m1, m2, sm3);
      double end_time = clock();
      total+=end_time-start_time;
    }
    printf("%.5lf secs per operation\n", (total/(double)(iter))/CLOCKS_PER_SEC);
    times.push_back((total/(double)(iter))/CLOCKS_PER_SEC);
  }
  std::ofstream times_file ("/home/jaehyun/trunk/tests/lib/cuda/test_sparse_matrices/times_2.txt");
  for(i = 0; i < (int)times.size(); i++)
    times_file << times[i] << std::endl;
}


TEST(SparseCudaMatrix, SpeedBenchmarkingVaryImgSize) {
  std::vector<double> times;
  for (int fc = 1; fc <= 17; fc++) {
    char sparseMatrixFile[256];
    sprintf(sparseMatrixFile, "/home/jaehyun/trunk/tests/lib/cuda/vary_img_size/sparse_matrix_%d.txt", fc);
    // char mat3[]="/home/jaehyun/trunk/tests/lib/cuda/test_sparse_matrices/sparse_matrix_2.txt";
    FILE *fp = fopen(sparseMatrixFile, "r");
    std::vector<sparseEntry> entries;
    int i, j;
    float v;
    int m, n;
    fscanf(fp, "%d%d", &m, &n);
    while(fscanf(fp, "%d%d%f", &i, &j, &v)==3) {
      sparseEntry newEntry;
      newEntry.row = i;
      newEntry.col = j;
      newEntry.val = v;
      entries.push_back(newEntry);
    }
    fclose(fp);
    
    SparseCudaMatrix sm3(m, n, entries);
    int iter = 1;
    int batchSize = 1000;
    double total=0;

    for (i=0; i<iter; i++) {
      CudaMatrix m1=CudaMatrix::Random(sm3.rows(), batchSize /**/);
      CudaMatrix m2=CudaMatrix::Random(m1.cols(), sm3.cols());
      double start_time = clock();
      SparseCudaMatrix res = denseTimesDenseCWiseTimesSparse(m1, m2, sm3);
      double end_time = clock();
      total+=end_time-start_time;
    }
    printf("%.5lf secs per operation\n", (total/(double)(iter))/CLOCKS_PER_SEC);
    times.push_back((total/(double)(iter))/CLOCKS_PER_SEC);
  }

  std::ofstream times_file ("/home/jaehyun/trunk/tests/lib/cuda/vary_img_size/batch_1000/cuda_times.txt");
  for(int i = 0; i < (int)times.size(); i++)
    times_file << times[i] << ' ';
}
