#include "sdn/base.h"
#include "sdn/old.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include "gtest/gtest.h"


using namespace sdn;
using namespace boost;

TEST(Convolver, Constructor) {
  const Convolver c(3,1,2);
}

TEST(Convolver, KnownInputOutput) {
  
  Convolver cl(3, 1, 2);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  
  Matrix ret = cl.convolve(input);

  Matrix deconv = cl.deconvolve(ret);

  ASSERT_EQ(deconv.rows(),input.rows());
  ASSERT_EQ(deconv.cols(),input.cols());
  for(int r = 0; r < deconv.rows(); ++r) {
    for(int c = 0; c < deconv.cols(); ++c) {
      ASSERT_EQ(deconv(r,c), input(r,c));
    }
  }
  
  ASSERT_EQ(ret.rows(), 9);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(9,4);
  ans << 
    0, 1, 4, 5,
    1, 2, 5, 6,
    2, 3, 6, 7,
    4, 5, 8, 9,
    5, 6, 9, 10,
    6, 7, 10, 11,
    8, 9, 12, 13,
    9, 10, 13, 14,
    10, 11, 14, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}

TEST(Convolver, KnownInputOutput2) {
  //Tests the case when patch_width == patch_step. Non-overlapping patches.
  
  Convolver cl(2, 2, 2);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  
  Matrix ret = cl.convolve(input);

  Matrix deconv = cl.deconvolve(ret);
  
  
  ASSERT_EQ(ret.rows(), 4);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(4,4);
  ans << 
    0, 2, 8, 10,
    1, 3, 9, 11,
    4, 6, 12, 14,
    5, 7, 13, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}

int RandInt(int lower, int upper) {
  return (int)((Scalar)rand()*(Scalar)(upper-lower+1)/(Scalar)RAND_MAX) 
    + lower;
}

TEST(Convolver, RandomStressTest) {
  //Tests a number of convolution/deconvolutions to make sure that they 
  //match.
  srand(0);
  
  int num_tests = 10;

  for(int i = 0; i < num_tests; i++) {

    int input_channels = RandInt(1,100);
    int output_width = RandInt(1,10);
    int patch_width = RandInt(1,30);
    int patch_step = RandInt(1,patch_width);
    int input_width = patch_step*(output_width-1) + patch_width;
    int output_channels = input_channels*patch_width*patch_width;
    
    StorageNode in(input_channels);
    StorageNode out(output_channels);
    
    Convolver cl(patch_width, 
		 patch_step, 
		 output_width);
    
    Matrix input = Matrix::Random(input_channels,
				      input_width*input_width);
    
    Matrix ret = cl.convolve(input);
    
    Matrix deconv = cl.deconvolve(ret);
    
    ASSERT_EQ(ret.rows(), output_channels);
    ASSERT_EQ(ret.cols(), output_width*output_width);
    
    for(int r = 0; r < input.rows(); ++r) {
      for(int c = 0; c < input.cols(); ++c) {
	ASSERT_NEAR(input(r,c), deconv(r,c),.00001);
      }
    }
  }
}

TEST(ConvolutionLayer, KnownInputOutput) {

  shared_ptr<StorageNode> in = shared_ptr<StorageNode>(new StorageNode(1));
  shared_ptr<StorageNode> out = shared_ptr<StorageNode>(new StorageNode(9));
  
  ConvolutionLayer cl(3, 1, 2, in, out);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  in->setActivities(input);

  cl.forwardProp();

  Matrix ret = out->getActivities();

  out->setDerivatives(ret);

  cl.backProp();
  Matrix deconv = in->getDerivatives();

  ASSERT_EQ(deconv.rows(),input.rows());
  ASSERT_EQ(deconv.cols(),input.cols());
  for(int r = 0; r < deconv.rows(); ++r) {
    for(int c = 0; c < deconv.cols(); ++c) {
      ASSERT_EQ(deconv(r,c), input(r,c));
    }
  }
  
  ASSERT_EQ(ret.rows(), 9);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(9,4);
  ans << 
    0, 1, 4, 5,
    1, 2, 5, 6,
    2, 3, 6, 7,
    4, 5, 8, 9,
    5, 6, 9, 10,
    6, 7, 10, 11,
    8, 9, 12, 13,
    9, 10, 13, 14,
    10, 11, 14, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}

TEST(ConvolutionLayer, KnownInputOutput2) {
  //Tests the case when patch_width == patch_step. Non-overlapping patches.
  shared_ptr<StorageNode> in = shared_ptr<StorageNode>(new StorageNode(1));
  shared_ptr<StorageNode> out = shared_ptr<StorageNode>(new StorageNode(4));
  
  ConvolutionLayer cl(2, 2, 2, in, out);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  in->setActivities(input);
  
  cl.forwardProp();

  Matrix ret = out->getActivities();

  out->setDerivatives(ret);

  cl.backProp();
  Matrix deconv = in->getDerivatives();
  
  
  ASSERT_EQ(ret.rows(), 4);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(4,4);
  ans << 
    0, 2, 8, 10,
    1, 3, 9, 11,
    4, 6, 12, 14,
    5, 7, 13, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}

TEST(ConvolutionLayer, RandomStressTest) {
  //Tests a number of convolution/deconvolutions to make sure that they 
  //match.
  srand(0);
  
  int num_tests = 10;

  for(int i = 0; i < num_tests; i++) {

    int input_channels = RandInt(1,100);
    int output_width = RandInt(1,10);
    int patch_width = RandInt(1,30);
    int patch_step = RandInt(1,patch_width);
    int input_width = patch_step*(output_width-1) + patch_width;
    int output_channels = input_channels*patch_width*patch_width;
    
    shared_ptr<StorageNode> in = shared_ptr<StorageNode>(new StorageNode(input_channels));
    shared_ptr<StorageNode> out = shared_ptr<StorageNode>(new StorageNode(output_channels));
    
    ConvolutionLayer cl(patch_width, 
			patch_step, 
			output_width, 
			in, 
			out);
    
    Matrix input = Matrix::Random(input_channels,
				      input_width*input_width);
    
    in->setActivities(input);
    
    cl.forwardProp();

    Matrix ret = out->getActivities();

    out->setDerivatives(ret);

    cl.backProp();
    Matrix deconv = in->getDerivatives();
    
    ASSERT_EQ(ret.rows(), output_channels);
    ASSERT_EQ(ret.cols(), output_width*output_width);
    
    for(int r = 0; r < input.rows(); ++r) {
      for(int c = 0; c < input.cols(); ++c) {
	ASSERT_NEAR(input(r,c), deconv(r,c),.00001);
      }
    }
  }
}



TEST(DeconvolutionLayer, KnownInputOutput) {

  shared_ptr<StorageNode> in = ptr(new StorageNode(9));
  shared_ptr<StorageNode> out = ptr( new StorageNode(1));
  
  DeconvolutionLayer cl(3, 1, 2, in, out);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  out->setDerivatives(input);

  cl.backProp();

  Matrix ret = in->getDerivatives();

  in->setActivities(ret);

  cl.forwardProp();
  Matrix deconv = out->getActivities();

  ASSERT_EQ(deconv.rows(),input.rows());
  ASSERT_EQ(deconv.cols(),input.cols());
  for(int r = 0; r < deconv.rows(); ++r) {
    for(int c = 0; c < deconv.cols(); ++c) {
      ASSERT_EQ(deconv(r,c), input(r,c));
    }
  }
  
  ASSERT_EQ(ret.rows(), 9);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(9,4);
  ans << 
    0, 1, 4, 5,
    1, 2, 5, 6,
    2, 3, 6, 7,
    4, 5, 8, 9,
    5, 6, 9, 10,
    6, 7, 10, 11,
    8, 9, 12, 13,
    9, 10, 13, 14,
    10, 11, 14, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}

TEST(DeconvolutionLayer, KnownInputOutput2) {
  //Tests the case when patch_width == patch_step. Non-overlapping patches.
  shared_ptr<StorageNode> in = ptr(new StorageNode(4));
  shared_ptr<StorageNode> out = ptr(new StorageNode(1));
  
  DeconvolutionLayer cl(2, 2, 2, in, out);
  
  Matrix input = Matrix::Zero(1,16);
  for( int i = 0; i < input.cols(); ++i) {
      input(0,i) = i;
  }
  
  out->setDerivatives(input);
  
  cl.backProp();

  Matrix ret = in->getDerivatives();

  in->setActivities(ret);

  cl.forwardProp();
  //Matrix deconv = out.getActivities();
  
  
  ASSERT_EQ(ret.rows(), 4);
  ASSERT_EQ(ret.cols(), 4);
  
  Matrix ans = Matrix::Zero(4,4);
  ans << 
    0, 2, 8, 10,
    1, 3, 9, 11,
    4, 6, 12, 14,
    5, 7, 13, 15;

  for(int r = 0; r < ret.rows(); ++r) {
    for(int c = 0; c < ret.cols(); ++c) {
      ASSERT_EQ(ret(r,c), ans(r,c));
    }
  }
}


TEST(DeconvolutionLayer, RandomStressTest) {
  //Tests a number of convolution/deconvolutions to make sure that they 
  //match.
  srand(0);
  
  int num_tests = 10;

  for(int i = 0; i < num_tests; i++) {

    int input_channels = RandInt(1,100);
    int output_width = RandInt(1,10);
    int patch_width = RandInt(1,30);
    int patch_step = RandInt(1,patch_width);
    int input_width = patch_step*(output_width-1) + patch_width;
    int output_channels = input_channels*patch_width*patch_width;
    
    StorageNode in(output_channels);
    StorageNode out(input_channels);
    
    DeconvolutionLayer cl(patch_width, 
			patch_step, 
			output_width, 
			  owned_ptr(&in), 
			  owned_ptr(&out));
    
    Matrix input = Matrix::Random(input_channels,
				      input_width*input_width);
    
    out.setDerivatives(input);
    
    cl.backProp();

    Matrix ret = in.getDerivatives();

    in.setActivities(ret);

    cl.forwardProp();
    Matrix deconv = out.getActivities();
    
    ASSERT_EQ(ret.rows(), output_channels);
    ASSERT_EQ(ret.cols(), output_width*output_width);
    
    for(int r = 0; r < input.rows(); ++r) {
      for(int c = 0; c < input.cols(); ++c) {
	ASSERT_NEAR(input(r,c), deconv(r,c), .00001);
      }
    }
  }
}

