// Copyright 2009, All Rights Reserved.
// Author: Andrew Saxe (asaxe@stanford.edu)
                                                                                
/** \file gradient-checker.cc
    \brief Implements the GradientChecker class.
*/
                                                                                
#include "derivativeChecker.h"
#include "sdn/base.h"
#include <assert.h>

#include <Eigen/Array>
using namespace sdn;

DerivativeChecker::DerivativeChecker() {}

void DerivativeChecker::PrintComparison(double epsilon, 
					const Matrix& x) {
  std::cout << "Numerical:\n" << NumericalDerivative(epsilon, x) << "\n";
  std::cout << "Analytical:\n" << AnalyticalDerivative(x) << "\n";
}

void DerivativeChecker::PrintComparisonMatrix(double epsilon,
					      const Matrix& x) {
  std::cout << "Numerical:\n" << NumericalDerivativeMatrix(epsilon,x) << "\n";
  std::cout << "Analytical:\n" << AnalyticalDerivative(x) << "\n";
}

double DerivativeChecker::CheckDerivative(double epsilon, 
				      const Matrix& x) {
  Matrix num = NumericalDerivative(epsilon, x);
  Matrix exact = AnalyticalDerivative(x);

  #if 0
  
  SDN_MSG("Input: \n"<<x<<"\numerical: \n "<<num<<"\nExact: \n"<<exact);
#endif

  return (num - exact).norm() / (num + exact).norm();
}

double DerivativeChecker::CheckDerivativeMatrix(double epsilon,
						const Matrix& x) {
  Matrix num = NumericalDerivativeMatrix(epsilon, x);
  Matrix exact = AnalyticalDerivative(x);

  return (num - exact).cwise().square().sum() / 
    (num + exact).cwise().square().sum();
}

Matrix DerivativeChecker::NumericalDerivative(double epsilon, 
				   const Matrix& x) {

  Matrix numerical_deriv = Matrix::Zero(x.rows(),x.cols());
  Matrix perturb = Matrix::Zero(x.rows(), x.cols());

  for(int r = 0; r < x.rows(); r++) {
    for(int c = 0; c < x.cols(); c++) {
      perturb(r,c) = epsilon;

      double pos_perturb = EvaluateFunction(x + perturb);//Inefficient but
      double neg_perturb = EvaluateFunction(x - perturb);//correct.
      numerical_deriv(r,c) = (pos_perturb - neg_perturb)/(2*epsilon);

      perturb(r,c) = 0;
    }
  }
  return numerical_deriv;
}
Matrix DerivativeChecker::NumericalDerivativeMatrix(double epsilon,
    const Matrix& x) {

  Matrix perturb = Matrix::Constant(x.rows(), x.cols(),epsilon);

  Matrix pos_perturb = EvaluateFunctionMatrix(x + perturb);
  Matrix neg_perturb = EvaluateFunctionMatrix(x - perturb);
  return (pos_perturb - neg_perturb)/(2*epsilon);

 
}

double DerivativeChecker::EvaluateFunction(const Matrix& x) {
  assert(0); //Unsupported if not overridden
  return 0;
}

Matrix DerivativeChecker::EvaluateFunctionMatrix(const Matrix& x) {
  assert(0); //Unsupported if not overridden.
  return Matrix::Zero(1,1);
}


void JacobianChecker::PrintComparison(double epsilon, 
				      const Vector& x) {
  std::cout << "Numerical:\n" << NumericalJacobian(epsilon, x) << "\n";
  std::cout << "Analytical:\n" << AnalyticalJacobian(x) << "\n";
}

 
double JacobianChecker::CheckJacobian(double epsilon, 
				      const Vector& x) {
  Matrix num = NumericalJacobian(epsilon, x);
  Matrix exact = AnalyticalJacobian(x);


  //SDN_MSG("Numerical: ");
  //SDN_MSG( num << "\nExact: \n"<<exact);


  return sqrt((num - exact).cwise().square().sum()) / 
    sqrt((num + exact).cwise().square().sum());
}
  

  /// Returns a Jacobian computed numerically using 
  /// perturbations of size epsilon, at the point x.
Matrix JacobianChecker::NumericalJacobian(double epsilon, 
    const Vector& x) {

  Vector perturb = Vector::Zero(x.rows());
  Matrix tmp = EvaluateFunction(x); //Just to get the output dimension.
  Matrix numerical_jacob(tmp.rows(),x.rows());

  for(int c = 0; c < numerical_jacob.cols(); c++) {
    perturb(c) = epsilon;
    Vector pos_perturb = EvaluateFunction(x + perturb);//Inefficient but
    Vector neg_perturb = EvaluateFunction(x - perturb);//correct.
    numerical_jacob.col(c) = (pos_perturb - neg_perturb)/(2*epsilon);
      
    perturb(c) = 0;
  
  }
  return numerical_jacob;
}

DerivativeChecker::~DerivativeChecker()
{

}

JacobianChecker::~JacobianChecker()
{

}
