// Copyright 2009, All Rights Reserved.
// Author: Andrew Saxe (asaxe@stanford.edu)

/** \file gradient-checker.h
    \brief Performs a numerical test of a gradient computation.
*/

#ifndef CORE__DERIVATIVE_CHECKER_H_
#define CORE__DERIVATIVE_CHECKER_H_

#include "sdn/base.h"

/// A base class which can be extended to numerically check a gradient.
class DerivativeChecker {
 public:

  /// Instantiates a DerivativeChecker.
  DerivativeChecker();

  /// Prints both the numerical and analytical gradients at a point.
  void PrintComparison(double epsilon, const sdn::Matrix& x);
  
  /// Prints both the numerical and analytical gradients at a matrix of points.
  void PrintComparisonMatrix(double epsilon, const sdn::Matrix& x);

  /// Checks the gradient at the point x with perturbations of size epsilon
  /**
     Calculates the numerical and analytical gradient at a random point and
     returns the degree to which they match.

     \param epsilon The size of the perturbations to use in computing the
     numerical gradient.
     
     \param x The input point to evaluate at.

     \return A measure of how well the gradients match. In particular, it 
     returns the norm of the difference divided by the norm of the sums: 
     \f[\frac{||\dot f_{num} - \dot f_{exact}|| }{||\dot f_{num} + 
     \dot f_{exact}||} \f]
   */
  double CheckDerivative(double epsilon, const sdn::Matrix& x);

  /// Checks the derivative of a function of a single variable at a 
  /// matrix of points. Should probably refactor to make the sdn::Matrix
  /// versions of the functions a separate class.
  double CheckDerivativeMatrix(double epsilon, const sdn::Matrix& x);

  
  /// Returns a vector of derivatives computed numerically using 
  /// perturbations of size epsilon, at a random point.
  //sdn::Matrix NumericalDerivative(double epsilon);

  /// Returns a vector of derivatives computed numerically using 
  /// perturbations of size epsilon, at the point x.
  sdn::Matrix NumericalDerivative(double epsilon, 
				    const sdn::Matrix& x);

 /// Returns a matrix of derivatives computed numerically using 
  /// perturbations of size epsilon, at the points given in x.
  sdn::Matrix NumericalDerivativeMatrix(double epsilon, 
				    const sdn::Matrix& x);

  //sdn::Matrix AnalyticalDerivative();

  /// Overload this to return the analytical gradient at point x.
  virtual sdn::Matrix AnalyticalDerivative(const sdn::Matrix& x) = 0;
  
  /// Overload this to return the function evaluated at point x.
  virtual double EvaluateFunction(const sdn::Matrix& x);
  
  ///Evaluates a function at each point in a matrix.
  virtual sdn::Matrix EvaluateFunctionMatrix(const sdn::Matrix& x);

  virtual ~DerivativeChecker();
};

/// A base class which can be extended to numerically check a Jacobian.
class JacobianChecker {
 public:
  /// Prints both the numerical and analytical Jacobians at a point.
  void PrintComparison(double epsilon, const sdn::Vector& x);

  /// Checks the Jacobian at the point x with perturbations of size epsilon
  /**

     \param epsilon The size of the perturbations to use in computing the
     numerical Jacobian.
     
     \param x The input point to evaluate at.

     \return A measure of how well the Jacobians match. In particular,
     it returns the Frobenius norm of the difference divided by the
     norm of the sums: \f[\frac{||\dot J_{num} - \dot J_{exact}||_{F}
     }{||\dot J_{num} + \dot J_{exact}||_{F}} \f]
   */
  double CheckJacobian(double epsilon, const sdn::Vector& x);
  

  /// Returns a Jacobian computed numerically using 
  /// perturbations of size epsilon, at the point x.
  sdn::Matrix NumericalJacobian(double epsilon, 
				    const sdn::Vector& x);

  /// Overload this to return the analytical Jacobian at point x.
  virtual sdn::Matrix AnalyticalJacobian(const sdn::Vector& x) = 0;
  
  /// Overload this to return the outputs evaluated at point x.
  virtual sdn::Vector EvaluateFunction(const sdn::Vector& x) = 0;
  

  virtual ~JacobianChecker();
};

#endif // CORE__DERIVATIVE_CHECKER_H_
