/// \author Ian Goodfellow

#include "sdn/old.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include "gtest/gtest.h"

using namespace sdn;

/// Checks that the gratings are drawn with the right dynamic range
TEST(Grating, Range)
{
  Scalar tolerance = 1e-3;
  unsigned rows=200;
  unsigned cols=200;
  Scalar f = 4.0;
  Scalar theta = 3.14 / 4.0;
  Scalar phase = 0;
  Scalar minValue = 12.0;
  Scalar maxValue = 13.7;

  Grating g(rows,cols,minValue,maxValue);
  g.setFrequency(f);
  g.setOrientation(theta);
  g.setPhaseShift(phase);

  Matrix m = g.getImage();

  ASSERT_LE(m.minCoeff(), minValue+tolerance);
  ASSERT_GE(m.maxCoeff(), maxValue-tolerance);
}

/// Checks that two gratings phase-shifted by 180 degrees cancel out
TEST(Grating, PhaseShiftCancel) {
  Scalar tolerance = 1e-4;
  unsigned rows=25;
  unsigned cols=25;
  Scalar f = 4.0;
  Scalar theta = 3.14 / 4.0;
  Scalar phase = 0;

  Grating g(rows,cols,-1.0,1.0);
  g.setFrequency(f);
  g.setOrientation(theta);
  g.setPhaseShift(phase);

  Matrix m = g.getImage();

  g.setPhaseShift(3.141592653589793238462);

  m += g.getImage();
  
    ASSERT_LE(m.cwise().abs().sum(), tolerance);
}

/// Checks that sum of two squared gratings in quadrature is 1
TEST(Grating, Quadrature) {
  Scalar tolerance = 1e-4;
  unsigned rows=25;
  unsigned cols=25;
  Scalar f = 4.0;
  Scalar theta = 3.14 / 7.0;
  Scalar phase = 0;

  Grating g(rows,cols,-1.0,1.0);
  g.setFrequency(f);
  g.setOrientation(theta);
  g.setPhaseShift(phase);
  g.setSamplingRate(1,1);//the condition for this test only holds if anti-aliasing is turned off

  Matrix m = g.getImage();

  g.setPhaseShift(3.141592653589793238462/2.0);

  m = m.cwise().square() + g.getImage().cwise().square();
  

  ASSERT_LE( abs(m.sum()-Scalar(rows*cols)), tolerance);
}

//todo-- test that empircal gradient of grating on a 2x2 window matches analytical gradient. this verifies that antialiasing is removing false edges
