/// \author Ian Goodfellow
#include "nnUtil.h"
#include <math.h>
#include <assert.h>
#include <iostream>
#include <fstream>
#include "sdn/base.h"
using namespace sdn;
using namespace std;





std::string parseLog(const std::string & filepath, std::map<std::string, std::vector<Scalar> > & output)
{
  ifstream ifs;
  ifs.open(filepath.c_str());

  if (ifs.fail())
    {
      cerr << "parseLog: could not open "<<filepath<<endl;
      return "";
    }
  
  string temp;

  ifs >> temp;

  if (temp != "<head>")
    {
      cerr << "parseLog: did not find <head> tag in "<<filepath<<endl;
      return "";
    }

  vector<string> columnNames;
  
  while (true)
    {
      ifs >> temp;
      if (ifs.fail())
	{
	  cerr << "parseLog: encounted EOF while reading header, or filestream died unexpectedly" << endl;
	  return "";
	}

      //Skip over record of what command was run
      if (temp == "<command>")
	{
	  while (true)
	    {
	      ifs >> temp;

	      if (ifs.fail())
		{
		  cerr << "parseLog: encountered EOF while reading command record, or filestream died unexpectedly" << endl;
		  return "";
		}

	      if (temp == "</head>")
		{
		  cerr << "parseLog: expected </command> before </head> " << endl;
		  return "";
		}

	      if (temp == "</command>")
		break;
	    }
	  continue;
	}


      if (temp == "</head>")
	break;

      columnNames.push_back(temp);
    }

  vector<vector<Scalar> > columnValues(columnNames.size());

  bool failed = false;
  while (!failed)
    {
      Scalar r;

      for (unsigned i = 0; i < columnNames.size(); i++)
	{
	  ifs >> r;

	  if (ifs.fail())
	    {
	      failed = true;
	      
	      if (i != 0)
		cerr << "parseLog: WARNING: encountered unexpected EOF while reading log file. Returning values read so far. (This probably just means you killed the process while it was writing to the log, but might mean that the log is formatted incorrectly)" << endl;
	      break;
	    }
	  else
	    columnValues[i].push_back(r);     
	}
    }

  for (unsigned i = 0; i < columnNames.size(); i++)
    output.insert( make_pair(columnNames[i], columnValues[i]));

  return strBaseName(filepath);
}

//Reciprocal of the number of overlaps occurring at point (i,j) in an image
Scalar averagingWeight(int i, int j, int patchStep, int inputWidth, int numStep)
{
  //input width REALLY IS the width of the code layer's receptive field
  //the size of the whole image is NOT an argument to this function-- that information
  //is contained in numStep

  //cout << "inputWidth: "<<inputWidth<<endl;
  //cout << "patchStep: "<<patchStep<<endl;

  assert(patchStep <= inputWidth);
  assert(patchStep > 0);

  
  int lastHorizStep = i/patchStep; //Depending on integer division to discard remainder

  if (lastHorizStep >= numStep)
    lastHorizStep = numStep-1;

  int firstHorizStep = (i - inputWidth + patchStep) / patchStep;

  if (firstHorizStep < 0)
    firstHorizStep = 0;

  //cout << "lh "<<lastHorizStep<<" fh "<<firstHorizStep << endl;

  int numHorizSteps = lastHorizStep - firstHorizStep + 1;

  int lastVertStep = j/patchStep;

  if (lastVertStep >= numStep)
    lastVertStep = numStep - 1;

  int firstVertStep = (j - inputWidth + patchStep) / patchStep;

  if (firstVertStep < 0)
    firstVertStep = 0;

  int numVertSteps = lastVertStep - firstVertStep + 1;

  //assert( Scalar (numHorizSteps * numVertSteps) > 0);

  return 1.0 / Scalar(numHorizSteps*numVertSteps);
}


Scalar sigmoidInverse(Scalar x)
{
  assert(0.0 < x & x < 1.0);
  return log(x/(1.0-x)); 
}
