/// \author Ian Goodfellow
/// Checks that new code matches behavior of old code

#include "sdn/old.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include "gtest/gtest.h"
#include "stackedEncoder.h"
#include "weightConversion.h"
#include <vector>

using namespace sdn;
using namespace std;
using namespace boost;


//Use this to share a static dataset between tests and reduce setup time
shared_ptr<Dataset> getDataset()
{
  /// \todo-- this depeneds on people having the ng_video dataset, maybe make a synthetic dataset generating class so that the unit tests can be distributed off campus
  SDN_MSG("Loading data");
  static shared_ptr<Dataset> d = DatasetRegistry().make("ng_video");
  SDN_MSG("Done loading data");
  return d;
}


TEST(OldCode, OneLayerNoSparsityNoWeightDecay)
{
  sdn::Logger::setLogLevel(LOGLEVEL_DEBUG);
  Scalar tolerance = 1e-6;
  

  int patchWidth = 1;
  int patchHeight = patchWidth;
  

  Scalar sparsityPenaltyMult = 0;
  Scalar learningRate = 0.01;
  Scalar biasLearningRate = learningRate;


  shared_ptr<Dataset> d = getDataset();
  d->requestExampleSize(patchHeight,patchWidth);

  int sqrtNhidden = 1;
  int nhidden = sqrtNhidden * sqrtNhidden;

  Scalar desiredSparsityMin = 0.5;
  Scalar desiredSparsityMax = 0.5;
  Scalar weightDecay = 0;


  stringstream options;
  options << "<sdnConfig> "  
	  << "<sdn::StackedNetworkConfig "
	  << "error_averaging_timeconstant=\".01\" "
	  << "input_width=\"" << patchWidth << "\" input_height=\"" << patchHeight << "\">"  
	  << "<Architecture> "
	  << "<sdn::Autoencoder num_hidden=\"" << nhidden <<"\""
	  << "learning_rate=\"" << learningRate << "\""
	  << "learning_rate_sparsity=\"0\""
	  << "target_sparsity=\"0\""
	  << "weight_decay=\"0\"/>"
	  << "</Architecture> "
	  << "</sdn::StackedNetworkConfig>  "
	  << "</sdnConfig>";



  XMLNode config = parseXML(options.str());

  ConfigManager::get().configure(config);


  StackedEncoder sa(patchWidth, sqrtNhidden, learningRate, biasLearningRate, 
		    weightDecay, desiredSparsityMin, desiredSparsityMax, sparsityPenaltyMult);

  StackedNetwork sn;

  //set them to have the same weights
  ASSERT_EQ(sn.numLayers(),2);
  shared_ptr<Layer> enl = sn.getLayer(0);
  shared_ptr<Layer> del = sn.getLayer(1);

  shared_ptr<StandardLayer> en = dynamic_pointer_cast<StandardLayer>(enl);
  shared_ptr<StandardLayer> de = dynamic_pointer_cast<StandardLayer>(del);


  ASSERT_EQ(sa.numEncoderLayers(),1);
  Matrix eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
  Matrix db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
  Matrix ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
  Matrix dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));


  en->setBiases(eb);
  en->setWeights(ewn);
  de->setBiases(db);
  de->setWeights(dwn);


  //Run it and see if they match
				   
  for (int iter = 0; iter < 5000; iter++)
    {
#if 0
      //test only forward prop by continuously copying StackedEncoder weights
      //to StackedNetwork

      SDN_DEBUG("Re-copying weights");
      eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
      db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
      ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
      dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
      
      
      en->setBiases(eb);
      en->setWeights(ewn);
      de->setBiases(db);
      de->setWeights(dwn);

      //SDN_DEBUG("Encoder biases " << endl << eb);
      //SDN_DEBUG("Decoder biases " << endl << db);

#endif


      Matrix patch = Matrix::Zero(patchHeight, patchWidth);
      d->getRandomExample(patch);

      Matrix vpatch = matrixToVector(patch);

      Scalar sa_errorEstimate;
      vector<vector<Scalar> > saSparsityStats;

      sa.forwardProp(patch, & sa_errorEstimate, &saSparsityStats);
      sn.forwardProp(vpatch);

      Vector sah = sa.getCodeVec(1);
      Vector snh = matrixToVector(en->getOutput());

      Vector diff = snh - sah;

      Scalar maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Hidden units diverged in iteration "<<iter);

	  
	  eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	  db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	
	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	
	}
      else
	{
	  //SDN_DEBUG("Hidden units matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);


      Vector sar = matrixToVector(sa.getReconstruction());
      Vector snr = matrixToVector(de->getOutput());

      diff = sar - snr;

      maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Reconstruction diverged in iteration "<<iter);

	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));

	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	}
      else
	{
	  //SDN_DEBUG("Reconstruction matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);


      sa.backProp();
      sa.updateWeights();

      sn.presentTarget(vpatch);
      sn.backProp();
      sn.gradientDescentUpdate();
      
      
    }

}





TEST(OldCode, OneLayerNoSparsityWeightDecay)
{
  sdn::Logger::setLogLevel(LOGLEVEL_DEBUG);
  Scalar tolerance = 1e-6;
  

  int patchWidth = 1;
  int patchHeight = patchWidth;
  

  Scalar sparsityPenaltyMult = 0;
  Scalar learningRate = 0.01;
  Scalar biasLearningRate = learningRate;

  SDN_MSG("Loading data");
  shared_ptr<Dataset> d = getDataset();
  d->requestExampleSize(patchHeight,patchWidth);
  SDN_MSG("Done loading data");

  int sqrtNhidden = 1;
  int nhidden = sqrtNhidden * sqrtNhidden;

  Scalar desiredSparsityMin = 0.5;
  Scalar desiredSparsityMax = 0.5;
  Scalar weightDecay = 0.1;


  stringstream options;
  options << "<sdnConfig> "  
	  << "<sdn::StackedNetworkConfig "
	  << "error_averaging_timeconstant=\".01\" "
	  << "input_width=\"" << patchWidth << "\" input_height=\"" << patchHeight << "\">"  
	  << "<Architecture> "
	  << "<sdn::Autoencoder num_hidden=\"" << nhidden <<"\""
	  << "learning_rate=\"" << learningRate << "\""
	  << "learning_rate_sparsity=\"0\""
	  << "target_sparsity=\"0\""
	  << "weight_decay=\"" << weightDecay << "\"/>"
	  << "</Architecture> "
	  << "</sdn::StackedNetworkConfig>  "
	  << "</sdnConfig>";



  XMLNode config = parseXML(options.str());

  ConfigManager::get().configure(config);


  StackedEncoder sa(patchWidth, sqrtNhidden, learningRate, biasLearningRate, 
		    weightDecay, desiredSparsityMin, desiredSparsityMax, sparsityPenaltyMult);

  StackedNetwork sn;

  //set them to have the same weights
  ASSERT_EQ(sn.numLayers(),2);
  shared_ptr<Layer> enl = sn.getLayer(0);
  shared_ptr<Layer> del = sn.getLayer(1);

  shared_ptr<StandardLayer> en = dynamic_pointer_cast<StandardLayer>(enl);
  shared_ptr<StandardLayer> de = dynamic_pointer_cast<StandardLayer>(del);


  ASSERT_EQ(sa.numEncoderLayers(),1);
  Matrix eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
  Matrix db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
  Matrix ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
  Matrix dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));


  en->setBiases(eb);
  en->setWeights(ewn);
  de->setBiases(db);
  de->setWeights(dwn);


  //Run it and see if they match
				   
  for (int iter = 0; iter < 5000; iter++)
    {
#if 0
      //test only forward prop by continuously copying StackedEncoder weights
      //to StackedNetwork

      SDN_DEBUG("Re-copying weights");
      eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
      db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
      ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
      dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
      
      
      en->setBiases(eb);
      en->setWeights(ewn);
      de->setBiases(db);
      de->setWeights(dwn);

      //SDN_DEBUG("Encoder biases " << endl << eb);
      //SDN_DEBUG("Decoder biases " << endl << db);

#endif


      Matrix patch = Matrix::Zero(patchHeight, patchWidth);
      d->getRandomExample(patch);

      Matrix vpatch = matrixToVector(patch);

      Scalar sa_errorEstimate;
      vector<vector<Scalar> > saSparsityStats;

      sa.forwardProp(patch, & sa_errorEstimate, &saSparsityStats);
      sn.forwardProp(vpatch);

      Vector sah = sa.getCodeVec(1);
      Vector snh = matrixToVector(en->getOutput());

      Vector diff = snh - sah;

      Scalar maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Hidden units diverged in iteration "<<iter);
	  
	  
	  eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	  db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	  ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	  dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	  
	  SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	  SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	  SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	  SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	  
	  
	}
      else
	{
	  //SDN_DEBUG("Hidden units matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);


      Vector sar = matrixToVector(sa.getReconstruction());
      Vector snr = matrixToVector(de->getOutput());

      diff = sar - snr;

      maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Reconstruction diverged in iteration "<<iter);

	  eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	  db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	  ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	  dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	  
	  SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	  SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	  SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	  SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	}
      else
	{
	  //SDN_DEBUG("Reconstruction matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);


      sa.backProp();
      sa.updateWeights();

      sn.presentTarget(vpatch);
      sn.backProp();
      sn.gradientDescentUpdate();
      
      
    }

}


TEST(OldCode, OneLayerWithSparsityNoWeightDecay)
{
  sdn::Logger::setLogLevel(LOGLEVEL_DEBUG);
  Scalar tolerance = 1e-5;
  

  int patchWidth = 1;
  int patchHeight = patchWidth;
  

  Scalar sparsityPenaltyMult = 10;
  Scalar learningRate = 0.01;
  Scalar biasLearningRate = learningRate;


  shared_ptr<Dataset> d = getDataset();
  d->requestExampleSize(patchHeight,patchWidth);

  int sqrtNhidden = 14;
  int nhidden = sqrtNhidden * sqrtNhidden;

  Scalar desiredSparsityMin = 0.02;
  Scalar desiredSparsityMax = 0.02;
  Scalar weightDecay = 0;


  stringstream options;
  options << "<sdnConfig> "  
	  << "<sdn::StackedNetworkConfig "
	  << "error_averaging_timeconstant=\".01\" "
	  << "input_width=\"" << patchWidth << "\" input_height=\"" << patchHeight << "\">"  
	  << "<Architecture> "
	  << "<sdn::Autoencoder num_hidden=\"" << nhidden <<"\""
	  << "learning_rate=\"" << learningRate << "\""
	  << "learning_rate_sparsity=\"" << sparsityPenaltyMult << "\""
	  << "target_sparsity=\"" << desiredSparsityMax << "\""
          << "sparsity_averaging_timeconstant=\".001\""
	  << "weight_decay=\"0\"/>"
	  << "</Architecture> "
	  << "</sdn::StackedNetworkConfig>  "
	  << "</sdnConfig>";



  XMLNode config = parseXML(options.str());

  ConfigManager::get().configure(config);


  StackedEncoder sa(patchWidth, sqrtNhidden, learningRate, biasLearningRate, 
		    weightDecay, desiredSparsityMin, desiredSparsityMax, sparsityPenaltyMult);

  StackedNetwork sn;

  //set them to have the same weights
  ASSERT_EQ(sn.numLayers(),2);
  shared_ptr<Layer> enl = sn.getLayer(0);
  shared_ptr<Layer> del = sn.getLayer(1);

  shared_ptr<StandardLayer> en = dynamic_pointer_cast<StandardLayer>(enl);
  shared_ptr<StandardLayer> de = dynamic_pointer_cast<StandardLayer>(del);


  ASSERT_EQ(sa.numEncoderLayers(),1);
  Matrix eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
  Matrix db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
  Matrix ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
  Matrix dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));


  en->setBiases(eb);
  en->setWeights(ewn);
  de->setBiases(db);
  de->setWeights(dwn);


  //Run it and see if they match
				   
  for (int iter = 0; iter < 5000; iter++)
    {
#if 0
      //test only forward prop by continuously copying StackedEncoder weights
      //to StackedNetwork

      //SDN_DEBUG("Re-copying weights");
      eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
      db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
      ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
      dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
      
      
      en->setBiases(eb);
      en->setWeights(ewn);
      de->setBiases(db);
      de->setWeights(dwn);

      //SDN_DEBUG("Encoder biases " << endl << eb);
      //SDN_DEBUG("Decoder biases " << endl << db);

#endif


      Matrix patch = Matrix::Zero(patchHeight, patchWidth);
      d->getRandomExample(patch);

      Matrix vpatch = matrixToVector(patch);

      Scalar sa_errorEstimate;
      vector<vector<Scalar> > saSparsityStats;

      sa.forwardProp(patch, & sa_errorEstimate, &saSparsityStats);
      sn.forwardProp(vpatch);

      Vector sah = sa.getCodeVec(1);
      Vector snh = matrixToVector(en->getOutput());

      //SDN_DEBUG("OUTPUT COMPARISON: Stacked Network (new):\n" << snh);
      //SDN_DEBUG("Stacked Autoencoder output (old):\n" << sah);

      Vector diff = snh - sah;

      Scalar maxDiff = diff.cwise().abs().maxCoeff();

    
      if (maxDiff > tolerance)
	{
	SDN_DEBUG("Hidden units diverged in iteration "<<iter);

	SDN_DEBUG("Reconstruction diverged in iteration "<<iter);
	
	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	
	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	
	
	
	}
      else
	{
	  //SDN_DEBUG("Hidden units matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);


      Vector sar = matrixToVector(sa.getReconstruction());
      Vector snr = matrixToVector(de->getOutput());

      diff = sar - snr;

      maxDiff = diff.cwise().abs().maxCoeff();

     
      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Reconstruction diverged in iteration "<<iter);

	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));

	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	}
      else
	{
	  //SDN_DEBUG("Reconstruction matched in iteration "<<iter);
	}

      ASSERT_LE(maxDiff, tolerance);

      
      sa.backProp();
      sa.updateWeights();

      sn.presentTarget(vpatch);
      sn.backProp();
      sn.gradientDescentUpdate();
      sn.gradientDescentUpdateSparsity();
      
      //      SDN_DEBUG("Sparsity statistics: " << saSparsityStats[0][0]);
      //SDN_DEBUG("----------------Stacked Network (new):");
      //sn.printDebugStuff();
      //SDN_DEBUG("----------------Stacked Encoder (old):");      
      //sa.printDebugStuff();
      

      
    }

}

TEST(OldCode, OneLayerWithSparsityNoWeightDecayLongRunning)
{
  sdn::Logger::setLogLevel(LOGLEVEL_DEBUG);
  Scalar tolerance = 1e-5;
  

  int patchWidth = 1;
  int patchHeight = patchWidth;
  

  Scalar sparsityPenaltyMult = 30;
  Scalar learningRate = 0.003;
  Scalar biasLearningRate = learningRate;


  shared_ptr<Dataset> d = getDataset();
  d->requestExampleSize(patchHeight,patchWidth);

  int sqrtNhidden = 14;
  int nhidden = sqrtNhidden * sqrtNhidden;

  Scalar desiredSparsityMin = 0.5;
  Scalar desiredSparsityMax = 0.5;
  Scalar weightDecay = 0;


  stringstream options;
  options << "<sdnConfig> "  
	  << "<sdn::StackedNetworkConfig "
	  << "error_averaging_timeconstant=\".01\" "
	  << "input_width=\"" << patchWidth << "\" input_height=\"" << patchHeight << "\">"  
	  << "<Architecture> "
	  << "<sdn::Autoencoder num_hidden=\"" << nhidden <<"\""
	  << "learning_rate=\"" << learningRate << "\""
	  << "learning_rate_sparsity=\"" << sparsityPenaltyMult << "\""
	  << "target_sparsity=\"" << desiredSparsityMax << "\""
          << "sparsity_averaging_timeconstant=\".001\""
	  << "weight_decay=\"0\"/>"
	  << "</Architecture> "
	  << "</sdn::StackedNetworkConfig>  "
	  << "</sdnConfig>";



  XMLNode config = parseXML(options.str());

  ConfigManager::get().configure(config);


  StackedEncoder sa(patchWidth, sqrtNhidden, learningRate, biasLearningRate, 
		    weightDecay, desiredSparsityMin, desiredSparsityMax, sparsityPenaltyMult);

  StackedNetwork sn;

  //set them to have the same weights
  ASSERT_EQ(sn.numLayers(),2);
  shared_ptr<Layer> enl = sn.getLayer(0);
  shared_ptr<Layer> del = sn.getLayer(1);

  shared_ptr<StandardLayer> en = dynamic_pointer_cast<StandardLayer>(enl);
  shared_ptr<StandardLayer> de = dynamic_pointer_cast<StandardLayer>(del);


  ASSERT_EQ(sa.numEncoderLayers(),1);
  Matrix eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
  Matrix db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
  Matrix ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
  Matrix dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));


  en->setBiases(eb);
  en->setWeights(ewn);
  de->setBiases(db);
  de->setWeights(dwn);


  //Run it and see if they match
				   
  for (int iter = 0; iter < 100000; iter++)
    {
      if(iter < 90000) {
	//test only forward prop by continuously copying StackedEncoder weights
	//to StackedNetwork
	
	//SDN_DEBUG("Re-copying weights");
	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	
	
	en->setBiases(eb);
	en->setWeights(ewn);
	de->setBiases(db);
	de->setWeights(dwn);
	
	//SDN_DEBUG("Encoder biases " << endl << eb);
	//SDN_DEBUG("Decoder biases " << endl << db);
      }

      Matrix patch = Matrix::Zero(patchHeight, patchWidth);
      d->getRandomExample(patch);

      Matrix vpatch = matrixToVector(patch);

      Scalar sa_errorEstimate;
      vector<vector<Scalar> > saSparsityStats;

      sa.forwardProp(patch, & sa_errorEstimate, &saSparsityStats);
      sn.forwardProp(vpatch);

      Vector sah = sa.getCodeVec(1);
      Vector snh = matrixToVector(en->getOutput());

      //SDN_DEBUG("OUTPUT COMPARISON: Stacked Network (new):\n" << snh);
      //SDN_DEBUG("Stacked Autoencoder output (old):\n" << sah);

      Vector diff = snh - sah;

      Scalar maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	SDN_DEBUG("Hidden units diverged in iteration "<<iter);

	SDN_DEBUG("Reconstruction diverged in iteration "<<iter);
	
	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));
	
	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	
	
	
	}
      else
	{
	  //SDN_DEBUG("Hidden units matched in iteration "<<iter);
	}
      if (iter % 5000 == 0) {
	SDN_DEBUG("Hidden diff : " << maxDiff);
      }
      ASSERT_LE(maxDiff, tolerance);


      Vector sar = matrixToVector(sa.getReconstruction());
      Vector snr = matrixToVector(de->getOutput());

      diff = sar - snr;

      maxDiff = diff.cwise().abs().maxCoeff();

      if (maxDiff > tolerance)
	{
	  SDN_DEBUG("Reconstruction diverged in iteration "<<iter);

	eb = stackedEncoderToStandardLayerBiases(sa.getEncoderBiases(0));
	db = stackedEncoderToStandardLayerBiases(sa.getDecoderBiases(0));
	ewn = stackedEncoderToStandardLayerWeights(sa.getEncoderWeights(0));
	dwn = stackedEncoderToStandardLayerWeights(sa.getDecoderWeights(0));

	SDN_DEBUG("encoder bias maxdiff: " << (eb-en->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("encoder weights maxdif: " << (ewn-en->getWeights()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder bias maxdiff: " << (db-de->getBiases()).cwise().abs().maxCoeff());
	SDN_DEBUG("decoder weights maxdiff: "<< (dwn-de->getWeights()).cwise().abs().maxCoeff());
	}
      else
	{
	  //SDN_DEBUG("Reconstruction matched in iteration "<<iter);
	}
      if (iter % 5000 == 0) {
	SDN_DEBUG("Recon diff : " << maxDiff);
      }
      ASSERT_LE(maxDiff, tolerance);

      
      sa.backProp();
      sa.updateWeights();

      sn.presentTarget(vpatch);
      sn.backProp();
      sn.gradientDescentUpdate();
      sn.gradientDescentUpdateSparsity();
      
      //      SDN_DEBUG("Sparsity statistics: " << saSparsityStats[0][0]);
      //SDN_DEBUG("----------------Stacked Network (new):");
      //sn.printDebugStuff();
      //SDN_DEBUG("----------------Stacked Encoder (old):");      
      //sa.printDebugStuff();
      

      
    }

}
