/// \author Ian Goodfellow

#include <highgui.h>
#include "nnUtil.h"
#include "stackedEncoder.h"
#include "weightConversion.h"
#include <fstream>
#include <iomanip>
using namespace sdn;

//Performance todo--
// 1. Could re-ordering any of the loops or array definitions
//    improve performance,
//    either by improving cache coherency or by turning more
//    loops into Eigen linear algebra calls?
// 2. Find out how to turn off bounds checking in Eigen, since
//    basically every operation in backprop is an array lookup
// 3. Test whether caching averagingWeight, testing averagingWeight
//    to see if tanh needs to be recomputed, etc. result in
//    speedup or not (this might be a layer-wise check, since
//    currently our constructors will never make averaging
//    necessary while reconstruncting layer nEncoderLayers-1)



using namespace std;
//using namespace util;

//------------------------------------------------------------------------------------------


Scalar randomReal(Scalar lower, Scalar upper)
{
  static RandomRealGenerator<Scalar> rrg;
  return rrg.randUnif(lower, upper);
}

StackedEncoder::StackedEncoder(int inputWidth, int outputWidth,		\
			       Scalar learningRate,  Scalar biasLearningRate, 	\
			       Scalar weightDecay, Scalar desiredSparsityMin,\
			       Scalar desiredSparsityMax, \
			       Scalar sparsityPenaltyMultiplier)
  : _nEncoderLayers(1), _nTotLayers(3),	_inputWidth(_nEncoderLayers+1),	\
    _patchStep(_nEncoderLayers+1), _numStep(_nEncoderLayers+1), \
    _outputWidth(_nEncoderLayers+1), \
    _learningRate(learningRate), _biasLearningRate(biasLearningRate),	\
    _weightDecay(2), _desiredSparsityMin(2), \
    _desiredSparsityMax(2), \
    _sparsityPenaltyMultiplier(2), \
    _learnAll(false), _errorEstimate(-1)//,
			      //    _L1weightsDisplay("Layer 1 Weights", inputWidth, inputWidth, 2, outputWidth*2, outputWidth)
{ 
  //Validate inputs
  assert(desiredSparsityMin <= desiredSparsityMax); 

  //Set size specifications

  //Visible layer
  _inputWidth[0] = inputWidth;
  _patchStep[0] = inputWidth;
  _numStep[0] = 1;
  _outputWidth[0] = inputWidth;

  //Encoder layer
  _inputWidth[1] = inputWidth;
  _patchStep[1] = inputWidth;
  _numStep[1] = 1;
  _outputWidth[1] = outputWidth;

  //Assemble all the weights (making no attempt to be efficient;
  //I assume we won't be constructing these in a tight loop)
  
  //Weights out of visible layer
  _weightsOut.push_back(vector<vector<Matrix> >(outputWidth,
						  vector<Matrix> (outputWidth,Matrix::Zero(inputWidth,inputWidth))));
  for (int i = 0; i < outputWidth; i++)
      assert(_weightsOut[0][i].size() == outputWidth);
  assert(_weightsOut[0].size() == outputWidth);

  _biasOut.push_back(Matrix::Zero(outputWidth, outputWidth));

  //Weights of hidden layer
  _weightsOut.push_back(vector<vector<Matrix> >(inputWidth, vector<Matrix> (inputWidth, Matrix::Zero(outputWidth, outputWidth))));

  _biasOut.push_back(Matrix::Zero(inputWidth, inputWidth));

  //Initialize weights to small random values
  for (unsigned i = 0; i < _weightsOut.size(); i++)
    for (unsigned j = 0; j < _weightsOut[i].size(); j++)
      for (unsigned k = 0; k < _weightsOut[i][j].size(); k++)
	for (unsigned l = 0; l < _weightsOut[i][j][k].rows(); l++)
	  for (unsigned m = 0; m < _weightsOut[i][j][k].cols(); m++)
	    _weightsOut[i][j][k](l,m) = randomReal(-0.05, 0.05);


    //Set up learning parameters-- this needs to happen before setting up biases
  _weightDecay[1] = weightDecay;
  _desiredSparsityMin[1] = desiredSparsityMin;
  _desiredSparsityMax[1] = desiredSparsityMax;			
  _sparsityPenaltyMultiplier[1] = sparsityPenaltyMultiplier;

  //Initialize encoding biases such that if weights were 0 we would be on
  //the upper bound of the desired firing rate
  for (unsigned i = 0; i < outputWidth; i++)
    for (unsigned j = 0; j < outputWidth; j++)
      _biasOut[0](i,j) = 0.5*sigmoidInverse(max(_desiredSparsityMax[1], (Scalar) 0.005));

  //Initialize decoding biases to 0
  _biasOut[1].setConstant(0);

  //Set up activations and deltas

  //visible layer
  //activationIns and deltas are unused
  _activationIn.push_back(Matrix::Zero(1,1));
  _activationInMulti.push_back(vector<vector<Matrix> >());
  _delta.push_back(Matrix::Zero(1,1));
  _deltaMulti.push_back(vector<vector<Matrix> >());
  //activationOut is used
  _activationOut.push_back(Matrix::Zero(inputWidth,inputWidth));

  //hidden layer
  //muti layers are not used, since it is not possible for there
  //to be averaging
  _activationInMulti.push_back(vector<vector<Matrix> >());
  _deltaMulti.push_back(vector<vector<Matrix> >());
  //activationIn, activationOut and the deltas are used
  _activationIn.push_back(Matrix::Zero(outputWidth,outputWidth));
  _activationOut.push_back(Matrix::Zero(outputWidth,outputWidth));
  _delta.push_back(Matrix::Zero(outputWidth,outputWidth));
 

  //reconstruction layer
  //_activationIn is not used, no need to allocate
  //_activationInMulti is used
  _activationInMulti.push_back(vector<vector<Matrix> > (1, vector<Matrix>(1, Matrix::Zero(inputWidth, inputWidth))));
  //delta is used
  _delta.push_back(Matrix::Zero(inputWidth,inputWidth));
  //activationOut is used
  _activationOut.push_back(Matrix::Zero(inputWidth,inputWidth));
  //deltaMulti is used
  _deltaMulti.push_back(vector<vector<Matrix> > (1, vector<Matrix>(1, Matrix::Zero(inputWidth, inputWidth))));
  assert(_deltaMulti[2][0][0].rows() == inputWidth);

  //Set up the sparsity estimates
  
  //Vis layer is unused
  _sparsityEstimate.push_back(Matrix::Zero(1,1));
  //All encoder layers are used (if _learnAll is on, otherwise only last is used)
  for (int i = 1; i <= _nEncoderLayers; i++)
    _sparsityEstimate.push_back(_desiredSparsityMax[i] * Matrix::Ones(_numStep[i]*_outputWidth[i], _numStep[i]*_outputWidth[i]));
  
  for (int i = 0; i <= _nEncoderLayers; i++)
    _sparsityPenaltyEstimate.push_back(-1.0);
  
  
  //_L1weightsDisplay.setRange(-1,1); 
}

StackedEncoder::StackedEncoder(int outputWidth, \
			       int patchStep, int numStep, \
			       StackedEncoder & smaller, bool learnAll,
			       Scalar weightDecay, Scalar desiredSparsityMin, Scalar
			       desiredSparsityMax, Scalar sparsityPenaltyMultiplier)
  : _nEncoderLayers(smaller._nEncoderLayers+1), \
    _nTotLayers(smaller._nTotLayers+2), _inputWidth(_nEncoderLayers+1), \
    _patchStep(_nEncoderLayers+1), _numStep(_nEncoderLayers+1), \
    _outputWidth(_nEncoderLayers+1), 
    _weightsOut(smaller._weightsOut), _biasOut(smaller._biasOut),	\
    _learningRate(smaller._learningRate), \
    _biasLearningRate(smaller._biasLearningRate),     \
    _weightDecay(smaller._weightDecay), \
    _desiredSparsityMin(smaller._desiredSparsityMin),	      \
    _desiredSparsityMax(smaller._desiredSparsityMax), \
    _sparsityPenaltyMultiplier(smaller._sparsityPenaltyMultiplier), \
    _learnAll(learnAll), _errorEstimate(0)//,
    //_L1weightsDisplay("Layer 1 Weights", smaller._inputWidth[1], smaller._inputWidth[1], 2, smaller._outputWidth[1]*2, smaller._outputWidth[1])
{ 
  //Validate inputs
  assert(desiredSparsityMin <= desiredSparsityMax); 

  //Set size specifications

  //Copy over old values
  for (int i = 0; i < _nEncoderLayers; i++)
    {
      _inputWidth[i] = smaller._inputWidth[i];
      _patchStep[i] = smaller._patchStep[i];
      _numStep[i] = smaller._numStep[i];
      _outputWidth[i] = smaller._outputWidth[i];
    }
  
  //Values for innermost code layer-- there is only one instantiation
  //of it, and it is connected to the entirety of the next layer down
  //(Note: this is the ONLY place in the code that assumes the innermost
  //layer is not convolutional. By writing a new constructor, we could
  //train the smallest network as convolutional from the start)
  _inputWidth[_nEncoderLayers] = _outputWidth[_nEncoderLayers-1] * numStep;
  _outputWidth[_nEncoderLayers] = outputWidth;
  _patchStep[_nEncoderLayers] = _inputWidth[_nEncoderLayers];
  _numStep[_nEncoderLayers] = 1;

  //_inputWidth and _outputWidth don't change for the intermediate layers,
  //but we do need to update their _numStep and _patchStep

  assert(_nEncoderLayers >= 2);
  
  //Second-deepest encoder layer
  _patchStep[_nEncoderLayers-1] = patchStep;
  _numStep[_nEncoderLayers-1] = numStep;

  //All other encoder layers
  for (int i = _nEncoderLayers-2; i >= 1; i--)
    {
      //we want ow[i]*ns[i] = (ns[i+1]-1)*ps[i+1]*iw[i+1]
      //ow[i] is fixed, so we update ns
      _numStep[i] = ((_numStep[i+1]-1)*_patchStep[i+1]+_inputWidth[i+1])/_outputWidth[i];

      if( _numStep[i] * _outputWidth[i] != (_numStep[i+1]-1)*_patchStep[i+1]+_inputWidth[i+1])
	{
	  cerr << "ERROR: The way you configured the network, layer "<<i<<"has output width "<<_outputWidth[i]<<", which cannot be replicated an integer number of times to make the required "<<(_numStep[i+1]-1)*_patchStep[i+1]+_inputWidth[i+1]<<" inputs to layer "<<i+1<<endl;
	  assert(false);
	}
    }

  //Visible layer
  _outputWidth[0] = (_numStep[1]-1)*_patchStep[1]+_inputWidth[1];
  _inputWidth[0] = _outputWidth[0];


  for (int i = 0; i < _nEncoderLayers; i++)
    assert( _numStep[i]*_outputWidth[i] == ( _numStep[i+1] -1)*_patchStep[i+1] + _inputWidth[i+1]);

  //Assemble all the weights (making no attempt to be efficient;
  //I assume we won't be constructing these in a tight loop)


  //Note: v.insert(v.begin()+i, elem) results in v[i] == elem

  //Add weights into new hidden layer
  int iw = numStep * _outputWidth[_nEncoderLayers-1];
  assert(iw > 0);

  Matrix temp = Matrix::Zero(iw, iw);
  _weightsOut.insert(_weightsOut.begin()+_nEncoderLayers-1, \
		     vector<vector<Matrix> >(outputWidth, vector<Matrix>
					       (outputWidth, temp)));
  _biasOut.insert(_biasOut.begin()+_nEncoderLayers-1, Matrix::Zero(outputWidth, outputWidth));

  
  //Add weights out of new hidden layer
  _weightsOut.insert(_weightsOut.begin()+_nEncoderLayers, \
		     vector<vector<Matrix> >(iw, vector<Matrix> (iw, Matrix::Zero(outputWidth, outputWidth))));

  _biasOut.insert(_biasOut.begin()+_nEncoderLayers, temp);

  //Initialize weights to small random values
  for (unsigned i = _nEncoderLayers-1; i <= _nEncoderLayers; i++)
    for (unsigned j = 0; j < _weightsOut[i].size(); j++)
      for (unsigned k = 0; k < _weightsOut[i][j].size(); k++)
	for (unsigned l = 0; l < _weightsOut[i][j][k].rows(); l++)
	  for (unsigned m = 0; m < _weightsOut[i][j][k].cols(); m++)
	    {
	      _weightsOut[i][j][k](l,m) = randomReal(-0.05, 0.05);
	    }

  //Set up learning parameters-- this needs to happen before setting up biases
  _weightDecay.push_back(weightDecay);
  _desiredSparsityMin.push_back(desiredSparsityMin);
  _desiredSparsityMax.push_back(desiredSparsityMax);			
  _sparsityPenaltyMultiplier.push_back(sparsityPenaltyMultiplier);

  //Initialize encoding biases such that if weights were 0 we would be on
  //the upper bound of the desired firing rate
  for (unsigned i = 0; i < outputWidth; i++)
    for (unsigned j = 0; j < outputWidth; j++)
      _biasOut[_nEncoderLayers-1](i,j) = 0.5*sigmoidInverse(max(_desiredSparsityMax[1], Scalar(0.005)));

  //Initialize decoding biases to 0
  _biasOut[_nEncoderLayers].setConstant(0);



  //Set up activations and deltas

  //visible layer
  //activationIns and deltas are unused
  _activationIn.push_back(Matrix::Zero(1,1));
  _activationInMulti.push_back(vector<vector<Matrix> >());
  _delta.push_back(Matrix::Zero(1,1));
  _deltaMulti.push_back(vector<vector<Matrix> >());
  //activationOut is used, but will be handled when we do the other layers

  //_activationsOut for all layers except visible
  for (int i = 0; i <= _nEncoderLayers; i++)
    {
      int total_out = _outputWidth[i]*_numStep[i];
      assert(total_out > 0);
      Matrix temp2 = Matrix::Zero(total_out, total_out);
      _activationOut.push_back(temp2);
    }

  for (int i = _nEncoderLayers-1; i >= 0; i--)
    {
      _activationOut.push_back(_activationOut[i]);
    }
  assert(_activationOut.size() == _nTotLayers);

  //activationsIn and deltas for encoder layers
  for (int i = 1; i <= _nEncoderLayers; i++)
    {
      int total_out = _outputWidth[i]*_numStep[i];
      _activationIn.push_back(Matrix::Zero(total_out, total_out));
      _delta.push_back(Matrix::Zero(total_out, total_out));
      _activationInMulti.push_back(vector<vector<Matrix> >()); //unused
      _deltaMulti.push_back(vector<vector<Matrix> >()); //unused
    }

  int encoderLayer = _nEncoderLayers;
  //activationsIn and deltas for decoder layers
  for (int i = _nEncoderLayers+1; i < _nTotLayers; i++)
    {
      _delta.push_back(_activationOut[i]);
      //_activationsIn is not used, no need to allocate

      _activationInMulti.push_back(vector<vector<Matrix> >(_numStep[encoderLayer], vector<Matrix>(_numStep[encoderLayer], Matrix::Zero(_inputWidth[encoderLayer],_inputWidth[encoderLayer]))));
      _deltaMulti.push_back(_activationInMulti.back());

      encoderLayer--;
    } 


  //Set up the sparsity estimates

  //Vis layer is unused
  _sparsityEstimate.push_back(Matrix::Zero(1,1));
  //All encoder layers are used (if _learnAll is on, otherwise only last is used)
  for (int i = 1; i <= _nEncoderLayers; i++)
    _sparsityEstimate.push_back(_desiredSparsityMax[i] * Matrix::Ones(_numStep[i]*_outputWidth[i], _numStep[i]*_outputWidth[i]));

  for (int i = 0; i <= _nEncoderLayers; i++)
    _sparsityPenaltyEstimate.push_back(-1.0);

  //_L1weightsDisplay.setRange(-1,1); 

  _errorEstimate = -1.0;
}


void StackedEncoder::forwardProp(const Matrix & patch, Scalar * errorEstimate, std::vector<vector< Scalar> > * sparsityStats,
				 std::vector<Scalar> * sparsityPenaltyEstimate, int startLayer, int endLayer)
{
  if (endLayer == -1)
    endLayer = _nTotLayers -1;

  assert(startLayer < _nTotLayers - 1);

  //Check size of input
  assert(patch.rows() == _activationOut[startLayer].rows());
  assert(patch.cols() == _activationOut[startLayer].cols());


  //Input provides activation output for first layer
  _activationOut[startLayer] = patch;

  //Do forward prop through encoding layers
  for (int i = startLayer + 1; i <= min(_nEncoderLayers,endLayer); i++)
    {

      //cout << "stacked weightsOut "<<i-1<<endl;
      //for (int j = 0; j < _weightsOut[i][0].size(); j++)
      //for (int k = 0; k < _weightsOut[i].size(); k++)
      //  cout << _weightsOut[i][k][j] << endl << endl;

      //Iterate over positions of the weights we are "convolving"
      for (int j = 0; j < _numStep[i]; j++)
	{
	  for (int k = 0; k < _numStep[i]; k++)
	    {
	      //Load the image patch operated on by the current position
	      //of the weights (TODO-- find a faster way of doing this than
	      //copying the whole thing, if possible. Would declaring
	      //"input" as a block rather than a matrix do that?)
	      Matrix input = _activationOut[i-1].block \
		(j*_patchStep[i], k*_patchStep[i], _inputWidth[i], _inputWidth[i]); 



	      //Add bias term
	      _activationIn[i].block \
		(j*_outputWidth[i], k*_outputWidth[i], _outputWidth[i], \
		 _outputWidth[i]) = _biasOut[i-1];

	      //For each hidden unit
	      for (int l = 0; l < _outputWidth[i]; l++)
		{
		  for (int m = 0; m < _outputWidth[i]; m++)
		    {
		      int rowInd = j * _outputWidth[i]+l;
		      int colInd = k * _outputWidth[i]+m;
		      

		      //Add weights multiplied by inputs
		      _activationIn[i](rowInd,colInd) += (_weightsOut[i-1][l][m].cwise() * input).sum();

		      
		      /*
		      cout << "("<<rowInd<<","<<colInd<<") -> activationIn is "<<setprecision(24) << _activationIn[i](rowInd,colInd) << endl;
		      cout << "weights: "<<endl;
		      cout << _weightsOut[i-1][l][m] << endl;
		      cout << "bias: "<<_activationIn[i](rowInd,colInd) - (_weightsOut[i-1][l][m].cwise()*input).sum() << endl;
		      cout << "input: "<<endl;
		      cout << input << endl;
		      */

		      //Apply tanh
		      _activationOut[i](rowInd,colInd) = tanh(_activationIn[i](rowInd,colInd));
		    }
		}							    
	    }
	}

      //cout << "stackedEncoder activationIn "<<i<<endl;
      //cout << _activationIn[i] << endl;

      //cout << "stackedEncoder activationOut "<<i<<endl;
      //cout << _activationOut[i] << endl;
    }

  ///int i = _nEncoderLayers;

  //cout << "se _weightsOut "<<_nEncoderLayers<<endl;
  //for (int j = 0; j < _weightsOut[i][0].size(); j++)
  //for (int k = 0; k < _weightsOut[i].size(); k++)
  //  cout << _weightsOut[i][k][j] << endl << endl;

  //cout << "se biasOut "<< _biasOut[i] << endl;

  int encoderLayer = _nEncoderLayers;

  if (startLayer > _nEncoderLayers)
    encoderLayer = _nEncoderLayers - (startLayer+1-(_nEncoderLayers+1));

  //Do forward prop through decoding layers and output layer
  for (int i = max(_nEncoderLayers+1, startLayer+1); i <= min(endLayer,_nTotLayers-1); i++)
    {

      int ns = _numStep[encoderLayer];
      int ow = _outputWidth[encoderLayer];
      int iw = _inputWidth[encoderLayer];
      int ps = _patchStep[encoderLayer];

      _activationOut[i].setConstant(0);

      //Iterate over all instantiations of the weights
      for (int j = 0; j < ns; j++)
	{
	  for (int k = 0; k < ns; k++)
	    {
	      //TODO- find something faster than copying
	      Matrix input = _activationOut[i-1].block(j*ow,k*ow,ow,ow);

	      _activationInMulti[i][j][k] = _biasOut[i-1];
	      for (int l = 0; l < iw; l++)
		{
		  for (int m = 0; m < iw; m++)
		    {
		      _activationInMulti[i][j][k](l,m) += (_weightsOut[i-1][l][m].cwise()*input).sum();
		      _activationOut[i](j*ps+l,k*ps+m) += tanh(_activationInMulti[i][j][k](l,m));
		      //cout << "Incrementing ("<<j*ps+l<<","<<k*ps+m<<") by "<<_activationInMulti[i][j][k](l,m) << endl;
		    }
		}
	    }
	}

      
      //  cout << "stackedEncoder input "<<i<<endl;
      //for (int j = 0; j < ns; j++)
      //for (int k = 0; k < ns; k++)
      //  cout << _activationInMulti[i][k][j] << endl <<endl;

      //cout << setprecision(7) << endl;
      //cout << "stackedEncoder activation out before averaging "<<i<<endl;
      
      //cout << _activationOut[i] << endl;
      
      //if (i < _weightsOut.size())
      //{
      //  cout << "stackedEncoder weightsOut "<<i<<endl;
      //  for (int j = 0; j < _weightsOut[i][0].size(); j++)
      //    {
	//      for (int k = 0; k < _weightsOut[i].size(); k++)
	//	{
	//	  cout << _weightsOut[i][k][j] << endl << endl;
      //}
      //    }
      //	}
 
	
      //      cout << _activationOut[i] << endl;


      //So far, we have only SUMMED the post-synaptic outputs. Now
      //we've got to do the division step of the averaging
      int w = (ns-1)*ps+iw;
      assert(w == _activationOut[i].cols());
      for (int j = 0; j < w; j++)
	{
	  for (int k = 0; k < w; k++)
	    {
	      //cout << "Averaging weight: "<<averagingWeight(j,k,ps,iw,ns) << endl;
	      _activationOut[i](j,k) *= averagingWeight(j,k,ps,iw,ns); //todo-- matrix of counts instead?
	    }
	}

      encoderLayer--;
    }


  //It doesn't make sense to update the sparsity estimates if not driven by data
  //There's no way to estimate reconstruction error if there is no input
  if (startLayer != 0)
    return;


  int first = _nEncoderLayers;
  int last = _nEncoderLayers;

  if (_learnAll)
    first = 1;

  if (sparsityPenaltyEstimate)
    sparsityPenaltyEstimate->resize(_nEncoderLayers+1);

  if (sparsityStats)
    sparsityStats->resize(_nEncoderLayers+1);

  for (int i = first; i <= last; i++)
    {
      _sparsityEstimate[i] = 0.999 * _sparsityEstimate[i] + 0.001 * \
	(_activationOut[i].cwise()+1.0)/2.0;
      
      // cout << "_sparsityEstimate["<<i<<"](0,0): "<<_sparsityEstimate[i](0,0) <<endl;

      //Note-- autoencoder class doesn't actually compute a sparsity
      //penalty estimate, so I don't do so here either
      if (sparsityPenaltyEstimate)
	(*sparsityPenaltyEstimate)[i] = _sparsityPenaltyEstimate[i];

      if (sparsityStats)
	{
	  //Note-- this will report sparsity estimates for individual neurons,
	  //the gradients from these estimates are pooled when updating
	  //weights that are shared between _numStep x _numStep neurons
	  int ns = _numStep[i];
	  int ow = _outputWidth[i];
	  int nHid = ns*ns*ow*ow;
	  //Scalar * temp = new Scalar[nHid];
	  
	  //This is safe because we don't care about the order of the units;
	  //we're about to sort them anyway
	  vector<Scalar> hU(nHid);
	  memcpy(&hU[0], _sparsityEstimate[i].data(), nHid*sizeof(Scalar));
	  
	  //qsort_Scalar(temp, nHid);
	  sort(hU.begin(),hU.end());

	  //cout << "nnHid\t\tnnSort" << endl;
	  //for (int j = 0; j < nHid; j++)
	  //cout << _sparsityEstimate[i].data()[j] << "\t\t" << temp[j] << endl;

	  
	  (*sparsityStats)[i].resize(6);

	  (*sparsityStats)[i][0] = matrixMean(_sparsityEstimate[i]);
	  (*sparsityStats)[i][1] = hU[(int)round(nHid*0.1)]; // 10% percentile
	  (*sparsityStats)[i][2] = hU[(int)round(nHid*0.25)]; // 25% percentile
	  (*sparsityStats)[i][3] = hU[(nHid==1)?0:(int)round(nHid*0.5)]; // median
	  (*sparsityStats)[i][4] = hU[nHid-1-(int)round(nHid*0.25)]; // 75% percentile
	  (*sparsityStats)[i][5] = hU[nHid-1-(int)round(nHid*0.1)]; // 90% percentile	  					    
	  //delete [] temp;
	}
    }

  //Estimate errors  //TODO-- move this to backprop?
  Scalar thisError = 0;
  for (int i = 0; i < _inputWidth[0]; i++)
    for (int j = 0; j < _inputWidth[0]; j++)
      {
	//assert(!isinf(_activationOut[0](i,j)));
	//assert(!isinf(_activationOut[_nTotLayers-1](i,j)));
	thisError += pow(_activationOut[0](i,j) - _activationOut[_nTotLayers-1](i,j),2);
      }

  //assert(!isnan(thisError));
  //assert(!isinf(thisError));

  if (_errorEstimate < 0)
    _errorEstimate = thisError;
  else
    _errorEstimate = 0.9999 * _errorEstimate + 0.0001 * thisError;

  if (errorEstimate)
    *errorEstimate = _errorEstimate;
}


//#define OUTPUT_DELTA 1
//#define AVE_LAYER 1
//#define REST 1
void StackedEncoder::backProp()
{
  _deepBackprop = true; //TODO-- put this under client control somewhere
  //TODO-- if onyl learning with _deepBackprop=false, then it isn't
  //necessary to do full forwardProp. Just be careful not to break the
  //visualization code.

  int objectiveLayer = _nTotLayers - 1;
  int firstLearnableLayer = 0;
  int ref = 0;
  if (!_deepBackprop)
    {
      objectiveLayer = _nEncoderLayers+1;
      firstLearnableLayer = _nEncoderLayers-1;
      ref = _nEncoderLayers -1;
    }


  //Compute deltas for imaginary "averaging units" of output layer
  int idx = objectiveLayer;
  for (int i = 0; i < _activationOut[idx].rows(); i++)
    {
      for (int j = 0; j < _activationOut[idx].cols(); j++)
	{
	  _delta[idx](i, j) =		     \
	    (_activationOut[idx](i,j) -		     \
	     _activationOut[ref](i, j));
	  //cout << "output delta: "<<_delta[idx](i,j) << endl;
	  //cout << "output: "<<_activationOut[idx](i,j) << endl;
	  //cout << "input: "<<_activationOut[0](i,j) << endl;
	}
    }

  //  cout << "se _delta "<<idx<<endl<<_delta[idx] << endl;

	
  int encodingLayer = 1;

  if (!_deepBackprop)
    encodingLayer = _nEncoderLayers;

  //Compute deltas for layers whose output is averaged
  //(Note-- by output I mean output of the decoding layer, and input
  //I mean input to the decoding layer from a higher code layer)
  //cout << "Enter i loop"<<endl;
  for (int i = objectiveLayer - 1 ; i >= _nEncoderLayers; i--)
    {
      int ns = _numStep[encodingLayer];
      int iw = _inputWidth[encodingLayer];
      int ps = _patchStep[encodingLayer];
      int ow = _outputWidth[encodingLayer];

      //Iterate over all instantiations of the weights
      for (int j = 0; j < ns; j++)
	{
	  for (int k = 0; k < ns; k++)
	    {
	      //Compute deltas for output units-- each of them feeds into exactly
	      //one averaging unit, with weight given by the averagingWeight function

	      //Iterate over all output nodes
	      for (int l = 0; l < iw; l++)
		{
		  for (int m = 0; m < iw; m++)
		    {
		      /*
			debugging code-- can be deleted any time
		      cout << "dm " << _deltaMulti[i+1][j][k](l,m) << endl;
		      cout << "am " << _activationInMulti[i+1][j][k](l,m) << endl;
		      cout << "d " << _delta[i+1](j*ps+l,k*ps+m) << endl;
		      */

		      //Compute the delta for the output node, using
		      //the delta for the averaging node it feeds into
		      _deltaMulti[i+1][j][k](l,m) = \
			averagingWeight(j*ps+l,k*ps+m,ps,iw,ns) *			\
			(1.0-pow(tanh( _activationInMulti[i+1][j][k](l,m)),2)) * \
			_delta[i+1](j*ps+l,k*ps+m);


		      //cout << "_deltaMulti: " << _deltaMulti[i+1][j][k](l,m) << endl;
		      
		      //todo-- here we can speed things up by precomputing the
		      //averagingWeight function
		      //todo-- also, if averagingWeight=1 (no overlaps) then
		      // we can actually just read the tanh out of _activationOut
		    }
		}
	   

	      //Compute deltas for input units. In most cases these are imaginary
	      //averaging units so h'(a) = 1 but since layer _nEncoderLayers is
	      //not averaged, we must multiply by h'(a) there

	      //Iterate over all input nodes
	      for (int l = 0; l < ow; l++)
		{
		  for (int m = 0; m < ow; m++)
		    {
		      _delta[i](j*ow+l, k*ow+m) = 0;

		      //Iterate over all output nodes, summing their weighted
		      //contribution
		      for (int n = 0; n < iw; n++)
			{
			  for (int o = 0; o < iw; o++)
			    {
			      _delta[i](j*ow+l,k*ow+m) += _deltaMulti[i+1][j][k](n,o)*_weightsOut[i][n][o](l,m);
			    }
			}

		      //if (j*ow+l == 0 && k*ow+m == 1)
		      //{
			  //cout << i << " " << 
			  //j << " " 
			  //   << k << " "<<l<<" " << m<< endl;
			  //cout << "nn sum: "<<setprecision(24) << _delta[i](j*ow+l,k*ow+m) << endl;
		      //}

		      if (i == _nEncoderLayers)
			{
			  _delta[i](j*ow+l,k*ow+m) *= (1.0 - pow(_activationOut[i](j*ow+l,k*ow+m),2));
			  //if (j* ow+l == 0 && k*ow+m == 1)
			    //{
			      //cout << "nn h': " << (1.0 - sqr(_activationOut[i](j*ow+l,k*ow+l))) << endl;

			      //}
			}
 
		    }
		}

	    }
	}


      //      cout << "se _deltaMulti "<<i+1<<endl;
      /*      for (int j = 0; j < ns; j++)
	{
	  for (int k = 0; k < ns; k++)
	    {
	      cout << _deltaMulti[i+1][k][j] << endl << endl;
	    }
	}

      cout << "se _delta "<<i<<endl;
      cout << _delta[i] << endl;
      */
 
      encodingLayer++;
    }
  //cout << "Exit i loop"<<endl;


  int lastLayer  = _nEncoderLayers-1;
  
  if (lastLayer < 1)
    lastLayer = 1;

  if (_learnAll)
    lastLayer = 1;

  if (firstLearnableLayer > lastLayer)
    lastLayer = firstLearnableLayer;

  //Iterate over all encoding layers whose weights we are learning
  for (int i = _nEncoderLayers-1; i >= lastLayer; i--)
    {
      //Clear deltas
      _delta[i].setConstant(0);

      //Have layer i+1 send its weighted deltas down (parsing the
      //network structure is easier in this direction)
      for (int j = 0; j < _numStep[i+1]; j++)
	{
	  for (int k = 0; k < _numStep[i+1]; k++)
	    {
	      for (int l = 0; l < _outputWidth[i+1]; l++)
		{
		  for (int m = 0; m < _outputWidth[i+1]; m++)
		    {
		      for (int n = 0; n < _inputWidth[i+1]; n++)
			{
			  for (int o = 0; o < _inputWidth[i+1]; o++)
			    {
			      _delta[i](j*_patchStep[i+1]+n, k*_patchStep[i+1]+o) += _delta[i+1](j*_outputWidth[i+1]+l, k*_outputWidth[i+1]+m) * _weightsOut[i][l][m](n,o);
			    }
			}
		    }
		}
	    }
	}

      //Multiply deltas by h'
      for (int j = 0; j < _numStep[i]*_outputWidth[i]; j++)
	{
	  for (int k = 0; k < _numStep[i]*_outputWidth[i]; k++)
	    {
	      _delta[i](j,k) *= (1.0-pow(_activationOut[i](j,k),2));
	    }
	}

      //      cout << "se _delta "<<i<<endl<<_delta[i]<<endl;

    }
}

void StackedEncoder::updateWeights()
{

  int first = _nEncoderLayers-1;
  int last = _nEncoderLayers;

  if (_learnAll && _deepBackprop)
    {
      first = 0;
      last = _nTotLayers-2;
    }

  for (int i = first; i <= last; i++)
    {
      int encoderLayer;

      if (i < _nEncoderLayers)
	encoderLayer = i+1;
      else
	encoderLayer = 2 * _nEncoderLayers-i;

      Scalar weightDecay = _weightDecay[encoderLayer];
      int ns = _numStep[encoderLayer];
      int ow = _outputWidth[encoderLayer];
      int ps = _patchStep[encoderLayer];
      int iw = _inputWidth[encoderLayer];



      //      cout << "se _biasLearningRate: "<<_biasLearningRate<<endl;

      if (i < _nEncoderLayers)
	{
	  //cout << "nn before: "<<endl;
	  //cout << _biasOut[i] << endl;

	  for (int l = 0; l < ns; l++)
	    for (int m = 0; m < ns; m++)
	      _biasOut[i] -= _biasLearningRate * _delta[i+1].block(l*ow, m*ow, ow, ow);
	  
	  //cout << "nn after: "<<endl;
	  //cout << _biasOut[i] << endl;
	}
      else
	{
	  for (int l = 0; l < ns; l++)
	    {
	      for (int m = 0; m < ns; m++)
		_biasOut[i] -= _biasLearningRate * _deltaMulti[i+1][l][m];
	    }
	}


      int nw = _weightsOut[i].size();
      for (int j = 0; j < nw; j++)
	{
	  assert(nw == _weightsOut[i][j].size());
	  for (int k = 0; k < nw; k++)
	    {
	      //Weight decay-- do this first since it depends on the values of
	      //the weights
	      _weightsOut[i][j][k] *= (1.0 - weightDecay * _learningRate);

	      //Follow the backprop gradient and penalize for sparsity
	      if (i < _nEncoderLayers)
		{
		  assert(nw  == ow);
		  //Since nw = ow in this case, j and k are stepping over
		  //units within an output block

		  for (int l = 0; l < ns; l++)
		    {
		      for (int m = 0; m < ns; m++)
			{
			  int r = l*ow+j;
			  int c = m*ow+k;

			  //Backprop for encoder layers
			  _weightsOut[i][j][k] -= _learningRate * _delta[i+1](r,c) * _activationOut[i].block(l*ps, m*ps, iw, iw);
			  
			  //Encoder layers also have the sparsity penalty
			  Scalar se = _sparsityEstimate[i+1](r,c);
			  Scalar smx = _desiredSparsityMax[i+1];
			  Scalar smn = _desiredSparsityMin[i+1];

			  //SDN_DEBUG("SPARSITY EQN OLD: " << _biasOut[i](j,k) << " -= " << _learningRate << " * " << _sparsityPenaltyMultiplier[i+1] << " * (" << se << " - " << smx << ") or " << se-smx); 

			  if (se > smx)
			    {
			      _biasOut[i](j,k) -= (se - smx) * _sparsityPenaltyMultiplier[i+1] * _learningRate;
			    }
			  else if (se < smn)
			    {
			      _biasOut[i](j,k) += (smn-se) * _sparsityPenaltyMultiplier[i+1] * _learningRate;
			    }

			  //SDN_DEBUG("Which equals " << _biasOut[i](j,k));

			}
		    }

		}
	      else
		{
		  assert(nw == iw);
		  

		  //if (i == 3)
		    //{
		      //cout << "se _deltaMulti "<<i+1<<endl;
		      //for (int m = 0; m < ns; m++)
		      //{
		      //  for (int l = 0; l < ns; l++)
		      //    {
		      //      cout << _deltaMulti[i+1][l][m] << endl << endl;
		      //    }
		      //}

		    //cout << "se layer 3 updates: "<<endl;
		    //}

		  for (int l = 0; l < ns; l++)
		    {
		      for (int m = 0; m < ns; m++)
			{

			  //if (i == 3)
			  //cout << "("<<m<<","<<l<<"): "<<- _learningRate * _deltaMulti[i+1][m][l](j,k)*_activationOut[i].block(m*ow, l*ow, ow,ow) << endl;

			  //Backprop for decoder layers
			  _weightsOut[i][j][k] -= _learningRate * _deltaMulti[i+1][m][l](j,k) * _activationOut[i].block(m*ow,l*ow,ow,ow);
			  //Decoder layers do not have sparsity penalty

			}
		    }
		  //cout << "--------" << endl;
		}
	    }
	}
    }
}

void StackedEncoder::printDebugStuff() {
  SDN_DEBUG("StackedEncoder Debug");
  
  SDN_DEBUG("Encoder layer");
  int l = 0;
  SDN_DEBUG("weights:\n" << stackedEncoderToStandardLayerWeights(getEncoderWeights(l)));
  //SDN_DEBUG("error_d_wrt_w:\n" << error_derivative_wrt_weights_);
  //SDN_DEBUG("derivative:\n" << derivative_);
  SDN_DEBUG("sparsity_avg_act:\n" << _sparsityEstimate[l+1]);
  SDN_DEBUG("biases:\n" << stackedEncoderToStandardLayerBiases(getEncoderBiases(l)));
  //SDN_DEBUG("error_d_wrt_biases:\n" << error_derivative_wrt_biases_);
  //SDN_DEBUG("error_d_wrt_sparsity:\n" << error_derivative_wrt_sparsity_);
  SDN_DEBUG("net_input:\n" << _activationIn[l+1]);
  SDN_DEBUG("delta_:\n" << _delta[l+1]);
  //SDN_DEBUG("sparsity_avg_: " << sparsity_avg_);
  //SDN_DEBUG("target_sparsity_: " << target_sparsity_);
  //SDN_DEBUG("learning_rate: " << learning_rate);
  //SDN_DEBUG("learning_rate_sparsity: " << learning_rate_sparsity);
  //SDN_DEBUG("sparsity_averaging_timeconstant: " << sparsity_averaging_timeconstant);
  
  //SDN_DEBUG("Decoder layer");
  //l = 0;
  //SDN_DEBUG("weights:\n" << stackedEncoderToStandardLayerWeights(getDecoderWeights(l)));
  //SDN_DEBUG("error_d_wrt_w:\n" << error_derivative_wrt_weights_);
  //SDN_DEBUG("derivative:\n" << derivative_);
  //SDN_DEBUG("sparsity_avg_act:\n" << _sparsityEstimate[l+1]);
  //SDN_DEBUG("biases:\n" << stackedEncoderToStandardLayerBiases(getDecoderBiases(l)));
  //SDN_DEBUG("error_d_wrt_biases:\n" << error_derivative_wrt_biases_);
  //SDN_DEBUG("error_d_wrt_sparsity:\n" << error_derivative_wrt_sparsity_);
  //SDN_DEBUG("net_input:\n" << _activationIn[l+1]);
  //SDN_DEBUG("delta_:\n" << _delta[l+1]);
  //SDN_DEBUG("sparsity_avg_: " << sparsity_avg_);
  //SDN_DEBUG("target_sparsity_: " << target_sparsity_);
  //SDN_DEBUG("learning_rate: " << learning_rate);
  //SDN_DEBUG("learning_rate_sparsity: " << learning_rate_sparsity);
  //SDN_DEBUG("sparsity_averaging_timeconstant: " << sparsity_averaging_timeconstant);

}
void StackedEncoder::showLayer1Weights(bool allowRescalingColorMap, bool allowScalingUp)
{
  //_L1weightsDisplay.enableWindow();
  updateLayer1Weights(allowRescalingColorMap, allowScalingUp);
}

void StackedEncoder::updateLayer1Weights(bool allowRescalingColorMap, bool allowScalingUp)
{
  //_L1weightsDisplay.clearDisplay(); 
  Matrix out(_inputWidth[1], _inputWidth[1]);
  for (int i=0; i < _outputWidth[1]; i++)
    for (int j = 0; j < _outputWidth[1]; j++)
      {
	//_L1weightsDisplay.addPatch(_weightsOut[0][i][j], allowRescalingColorMap, allowScalingUp); 
	
	for (int k = 0; k < _inputWidth[1]; k++)
	  for (int l = 0; l < _inputWidth[1]; l++)
	    out(k,l) = _weightsOut[_nTotLayers-2][k][l](i,j);

	//_L1weightsDisplay.addPatch(out,allowRescalingColorMap,allowScalingUp);
      }
}

void StackedEncoder::saveLayer1WeightsImage(const char *filename)
{
  updateLayer1Weights();
  //_L1weightsDisplay.saveImage(filename);
}

/*
void StackedEncoder::debugCopyWeights(const AutoEncoder & o)
{
  //assumes the stacked encoder has a single layer and everything is
  //the same size as in the autoencoder
  
  int ind = 0;
  for (int i = 0; i < _weightsOut[0].size(); i++)
    {
      for (int j = 0; j < _weightsOut[0][i].size(); j++)
	    {
	      assert(_weightsOut[0][i][j].rows() == o._weightsIH[ind].rows());
	      _weightsOut[0][i][j] = o._weightsIH[ind];
	      assert(_biasOut[0](i,j) == o._biasH(ind,0)); //these are not randomized, so should be equal
	      ind++;
	    }
	}
  
  for (int i = 0; i < _weightsOut[1].size(); i++)
    {
      for (int j = 0; j < _weightsOut[1][i].size(); j++)
	{
	  ind = 0;
	  for (int k = 0; k < _weightsOut[1][i][j].rows(); k++)
	    for (int l = 0; l < _weightsOut[1][i][j].cols(); l++)
	      {
		_weightsOut[1][i][j](k,l) = o._weightsHO[i*o._patchWidth+j](ind,0);
		ind++;
	      }
	  assert(_biasOut[1](i,j) == o._biasO(i,j));
	}
    }
  
}




void StackedEncoder::debugCheckWeights(const AutoEncoder & o)
    {
      int ind = 0;
      for (int i = 0; i < _weightsOut[0].size(); i++)
	{
	  for (int j = 0; j < _weightsOut[0][i].size(); j++)
	    {
	      assert(_weightsOut[0][i][j].rows() == o._weightsIH[ind].rows());
	      for (int k = 0; k < _weightsOut[0][i][j].rows(); k++)
		{
		  for (int l = 0; l < _weightsOut[0][i][j].cols(); l++)
		    {
		      Scalar diff = _weightsOut[0][i][j](k,l) - o._weightsIH[ind](k,l);
		      if (diff < 0)
			diff = -diff;
		      if (diff > 1e-10)
			{
			  std::cout << "_weightsOut[0]["<<i<<"]["<<j<<"]("<<k<<","<<l<<"): "<<_weightsOut[0][i][j](k,l) << std::endl;
			  std::cout << "o._weightsIH["<<ind<<"]("<<k<<","<<l<<"): "<<o._weightsIH[ind](0,0)<<std::endl;
			  assert(false);
			}
		    }
		}

	      Scalar diff = _biasOut[0](i,j) - o._biasH(ind,0);
	      if (diff < 0)
		diff = -diff;

	      if (diff > 1e-10)
		{
		  std::cout << "Bias problem" << std::endl;
		  std::cout << "_biasOut[0]("<<i<<","<<j<<"): "<<_biasOut[0](i,j)<<std::endl;
		  std::cout << "o._biasH("<<ind<<",0): "<<o._biasH(ind,0)<<std::endl;
		  assert(false);
		}

	      ind++;
	    }
	}
      
      //todo-- decoder weights
    }
*/


void StackedEncoder::debugSetAllWeights(Scalar c)
{
  for (unsigned i = 0; i < _weightsOut.size(); i++)
    for (unsigned j = 0; j < _weightsOut[i].size(); j++)
      for (unsigned k = 0; k < _weightsOut[i][j].size(); k++)
	_weightsOut[i][j][k].setConstant(c);
  
  for (unsigned i = 0; i < _biasOut.size(); i++)
    _biasOut[i].setConstant(c);
}




void StackedEncoder::saveNN(const char *filename)
{
  cout << "Saving" << endl;

  std::ofstream outputFile(filename, ios::out | ios::binary);
  if (!outputFile.is_open()) {
    std::cerr << "StackedEncoder::saveNN: Tried to open " << filename << "\n";
    SDN_FATAL("StackedEncoder::saveNN: Error opening file.");
  }

  //Write a header so we can verify that the file is a
  //Stacked AutoEnCoder
  static char header[] = {'s','a','e','c'};
  outputFile.write((char *)header, 4);

  //Write a version number so we can change the file format later
  //and still load old files
  int version = 0;
  outputFile.write((char *)&version, sizeof(int));

  //Save all the fields of the class
  outputFile.write((char*)&_nEncoderLayers, sizeof(int));
  outputFile.write((char*)&_nTotLayers, sizeof(int));
  
  int s;

  s = _inputWidth.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_inputWidth[i], sizeof(int));

  s = _patchStep.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_patchStep[i], sizeof(int));
  
  s = _numStep.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_numStep[i], sizeof(int));

  s = _outputWidth.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    outputFile.write((char *)&_outputWidth[i], sizeof(int));


  s = _weightsOut.size();
  outputFile.write((char *)&s, sizeof(int));
  //cout << "Writing s: "<<s<<endl;
  for (int i = 0; i < s; i++)
    {
      int s1 = _weightsOut[i].size();
      outputFile.write((char *)&s1, sizeof(int));
      //cout << "Writing s1: "<<s1<<endl;

      for (int j = 0; j < s1; j++)
	{
	  int s2 = _weightsOut[i][j].size();
	  outputFile.write((char *)&s2, sizeof(int));
	  //cout << "Writing s2: "<<s2<<endl;
	  for (int k = 0; k < s2; k++)
	    {
	      saveMatrix(outputFile, _weightsOut[i][j][k]);
	    }
	}
    }

  s = _biasOut.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    saveMatrix(outputFile, _biasOut[i]);


  //just save the sizes of the temp datastructures
  s = _activationIn.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      int r = _activationIn[i].rows();
      int c = _activationIn[i].cols();
      outputFile.write((char *)&r, sizeof(int));
      outputFile.write((char *)&c, sizeof(int));
    }

  s = _activationOut.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      int r = _activationOut[i].rows();
      int c = _activationOut[i].cols();
      outputFile.write((char *)&r, sizeof(int));
      outputFile.write((char *)&c, sizeof(int));
    }

  s= _activationInMulti.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      int s1 = _activationInMulti[i].size();
      outputFile.write((char *)&s1, sizeof(int));
      for (int j = 0; j < s1; j++)
	{
	  int s2 = _activationInMulti[i][j].size();
	  outputFile.write((char *)&s2, sizeof(int));
	  for (int k = 0; k < s2; k++)
	    {
	      int r = _activationInMulti[i][j][k].rows();
	      int c = _activationInMulti[i][j][k].cols();
	      outputFile.write((char *)&r, sizeof(int));
	      outputFile.write((char *)&c, sizeof(int));
	    }
	}
    }

  s = _delta.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      int r = _delta[i].rows();
      int c = _delta[i].cols();
      outputFile.write((char *)&r, sizeof(int));
      outputFile.write((char *)&c, sizeof(int));
    }

  s= _deltaMulti.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      int s1 = _deltaMulti[i].size();
      outputFile.write((char *)&s1, sizeof(int));
      for (int j = 0; j < s1; j++)
	{
	  int s2 = _deltaMulti[i][j].size();
	  outputFile.write((char *)&s2, sizeof(int));
	  for (int k = 0; k < s2; k++)
	    {
	      int r = _deltaMulti[i][j][k].rows();
	      int c = _deltaMulti[i][j][k].cols();
	      outputFile.write((char *)&r, sizeof(int));
	      outputFile.write((char *)&c, sizeof(int));
	    }
	}
    }

  //Now we're back to real stuff that needs to be saved
  outputFile.write((char*)&_learningRate, sizeof(Scalar));
  outputFile.write((char*)&_biasLearningRate, sizeof(Scalar));

  s = _weightDecay.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)& (_weightDecay[0]), sizeof(Scalar)*s);
  s = _desiredSparsityMin.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)&(_desiredSparsityMin[0]), sizeof(Scalar)*s);
  s = _desiredSparsityMax.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)&(_desiredSparsityMax[0]), sizeof(Scalar)*s);
  s = _sparsityPenaltyMultiplier.size();
  outputFile.write((char*)&s, sizeof(int));
  outputFile.write((char*)& (_sparsityPenaltyMultiplier[0]), sizeof(Scalar)*s);

  outputFile.write((char *)&_learnAll,sizeof(bool));

  outputFile.write((char *)&_errorEstimate, sizeof(Scalar));

  s = _sparsityEstimate.size();
  outputFile.write((char *)&s, sizeof(int));
  for (int i = 0; i < s; i++)
    {
      saveMatrix(outputFile, _sparsityEstimate[i]);
    }

  //Only need to save the size of sparsity penalty estimate
  s = _sparsityPenaltyEstimate.size();
  outputFile.write((char *)&s, sizeof(int));


  //No need to save _L1weightsDisplay, it can be regenerated

  if (!outputFile.good())
    {
      SDN_FATAL("Error saving to file.");
    }

  outputFile.close();

  cout << "Done saving " << filename  << endl;
}

void StackedEncoder::loadNN(const char *filename)
{
  std::ifstream inputFile(filename, ios::in | ios::binary);
  if (!inputFile.is_open()) {
    std::cerr << "StackedEncoder::loadNN Tried to open " << filename << "\n";
    SDN_FATAL("StackedEncoder::loadNN: Error opening file."); 
    return;
  }

  //Read header
  static char header[5];
  header[4] = '\0';
  inputFile.read((char*)header, 4);
  
  if (strcmp(header,"saec") != 0)
    {
      cerr << "StackedEncoder::loadNN: file \""<<filename<<"\" does not have the right header" << endl;
      inputFile.close();
      exit(-1);
    }

  int version;
  inputFile.read((char *)&version,sizeof(int));

  assert(version==0);//if/when there are multiple versions, we can put
  //a switch here

  inputFile.read((char *)&_nEncoderLayers,sizeof(int));
  inputFile.read((char *)&_nTotLayers,sizeof(int));
  
  int s;

  inputFile.read((char *)&s, sizeof(int));
  _inputWidth.resize(s);
  inputFile.read((char *)&(_inputWidth[0]), sizeof(int)*s);

  inputFile.read((char *)&s, sizeof(int));
  _patchStep.resize(s);
  inputFile.read((char *)&(_patchStep[0]), sizeof(int)*s);

  //int buf[100];
  //inputFile.read((char *) buf, sizeof(int)*100);
  //assert(false);

  inputFile.read((char *)&s, sizeof(int));
  _numStep.resize(s);
  inputFile.read((char *)&(_numStep[0]), sizeof(int)*s);
  
  inputFile.read((char *)&s, sizeof(int));
  _outputWidth.resize(s);
  inputFile.read((char *)&(_outputWidth[0]), sizeof(int)*s);

  inputFile.read((char *)&s, sizeof(int));
  _weightsOut.resize(s);
  for (int i = 0; i < s; i++)
    {
      int s1;
      //cout << "s1 before read: "<<s1<<endl;
      inputFile.read((char *)&s1,sizeof(int));
      //cout << "s1 after read: "<<s1<<endl;
      //assert(!inputFile.bad());

      _weightsOut[i].resize(s1);
      for (int j = 0; j < s1; j++)
	{
	  int s2=2;
	  //cout << "s2 before read: "<<s2<<endl;
	  inputFile.read((char *) & s2, sizeof(int));
	  //cout << "s2 after read: "<<s2<<endl;
	  _weightsOut[i][j].resize(s2, Matrix::Zero(1,1));
	  for (int k = 0; k < s2; k++)
	    loadMatrix(inputFile, _weightsOut[i][j][k], true);
	}
    }

  inputFile.read((char *) & s, sizeof(int));
  _biasOut.resize(s, Matrix::Zero(1,1));
  for (int i = 0; i < s; i++)
    loadMatrix(inputFile, _biasOut[i], true);


  //Temp datastructures just need to be allocated with the right size
  inputFile.read((char *)&s, sizeof(int));
  _activationIn.resize(s, Matrix::Zero(1,1));
  for (int i = 0; i < s; i++)
    {
      int r, c;
      inputFile.read((char *) &r, sizeof(int));
      inputFile.read((char *) &c, sizeof(int));

      _activationIn[i] = Matrix::Ones(r,c);
    }

  inputFile.read((char *)&s, sizeof(int));
  _activationOut.resize(s, Matrix::Zero(1,1));
  for (int i = 0; i < s; i++)
    {
      int r, c;
      inputFile.read((char *) &r, sizeof(int));
      inputFile.read((char *) &c, sizeof(int));

      _activationOut[i] = Matrix::Ones(r,c);
    }

  inputFile.read((char *)&s, sizeof(int));
  _activationInMulti.resize(s);
  for (int i = 0; i < s; i++)
    {
      int s1;
      inputFile.read((char *)&s1, sizeof(int));
      _activationInMulti[i].resize(s1);
      for (int j = 0; j < s1; j++)
	{
	  int s2;
	  inputFile.read((char *)&s2, sizeof(int));
	  _activationInMulti[i][j].resize(s2, Matrix::Zero(1,1));
	  for (int k = 0; k < s2; k++)
	    {
	      int r, c;
	      inputFile.read((char *) &r, sizeof(int));
	      inputFile.read((char *) &c, sizeof(int));
	  
	      _activationInMulti[i][j][k] = Matrix::Ones(r,c);
	    }
	}
    }

  inputFile.read((char *)&s, sizeof(int));
  _delta.resize(s, Matrix::Zero(1,1));
  for (int i = 0; i < s; i++)
    {
      int r, c;
      inputFile.read((char *) &r, sizeof(int));
      inputFile.read((char *) &c, sizeof(int));

      _delta[i] = Matrix::Ones(r,c);
    }

 inputFile.read((char *)&s, sizeof(int));
  _deltaMulti.resize(s);
  for (int i = 0; i < s; i++)
    {
      int s1;
      inputFile.read((char *)&s1, sizeof(int));
      _deltaMulti[i].resize(s1);
      for (int j = 0; j < s1; j++)
	{
	  int s2;
	  inputFile.read((char *)&s2, sizeof(int));
	  _deltaMulti[i][j].resize(s2, Matrix::Zero(1,1));
	  for (int k = 0; k < s2; k++)
	    {
	      int r, c;
	      inputFile.read((char *) &r, sizeof(int));
	      inputFile.read((char *) &c, sizeof(int));
	  
	      _deltaMulti[i][j][k] = Matrix::Ones(r,c);
	    }
	}
    }


  inputFile.read((char*)&_learningRate, sizeof(Scalar));
  inputFile.read((char*)&_biasLearningRate, sizeof(Scalar));

  inputFile.read((char *)&s, sizeof(int));
  _weightDecay.resize(s);
  inputFile.read((char*)&(_weightDecay[0]), s*sizeof(Scalar));

  inputFile.read((char *)&s, sizeof(int));
  _desiredSparsityMin.resize(s);
  inputFile.read((char*)&(_desiredSparsityMin[0]), s*sizeof(Scalar));

  inputFile.read((char *)&s, sizeof(int));
  _desiredSparsityMax.resize(s);
  inputFile.read((char*)&(_desiredSparsityMax[0]), s*sizeof(Scalar));

  inputFile.read((char *)&s, sizeof(int));
  _sparsityPenaltyMultiplier.resize(s);
  inputFile.read((char*)&(_sparsityPenaltyMultiplier[0]), s*sizeof(Scalar));

  inputFile.read((char*)&_learnAll, sizeof(bool));

  inputFile.read((char *)&_errorEstimate, sizeof(Scalar));

  cout << "Error estimate: "<<_errorEstimate<<endl;

  inputFile.read((char *)&s, sizeof(int));
  _sparsityEstimate.resize(s, Matrix::Zero(1,1));
  for (int i = 0; i < s; i++)
    {
      loadMatrix(inputFile, _sparsityEstimate[i], true);
    }

  inputFile.read((char *)&s, sizeof(int));
  _sparsityPenaltyEstimate.resize(s);
  

  cout << "Read stacked autoencoder." << endl;
  cout << "Hidden units in deepest layer: " << _numStep.back()*_numStep.back()*_outputWidth.back()*_outputWidth.back() << "\nInput width: " << inputWidth()  
       << "\nLearning rate: " << _learningRate << "\n";

  cout << "Numstep: ";
  for (int i = 0; i < _patchStep.size(); i++)
    cout << _numStep[i] << "\t";
  cout << endl;

  cout << "Outputwidth: ";
  for (int i = 0; i < _outputWidth.size(); i++)
    cout << _outputWidth[i] << "\t";
  cout << endl;
       
  if (!inputFile.good())
    {
      SDN_FATAL("StackedEncoder::StackedEncoder: Error reading from file.");
    }

  //_L1weightsDisplay.setNewDisplayParam(_inputWidth[1], _inputWidth[1], 2, _outputWidth[1]*2, _outputWidth[1]);
  //_L1weightsDisplay.setRange(-1,1); 

  inputFile.close();

  return;
}

/*
void StackedEncoder::showBiases(int layer)
{
  PatchDisplay d("Biases",_biasOut[layer].cols(),_biasOut[layer].rows(),4,2,2);
  d.setRange(-1,1);
  d.addPatch(_biasOut[layer]);

  Matrix activations = _biasOut[layer];
  for (int i = 0; i < activations.rows(); i++)
    for (int j = 0; j < activations.cols(); j++)
      activations(i,j) = tanh(activations(i,j));

  d.addPatch(activations);


  Matrix inputs = _biasOut[layer];

  for (int i = 0; i < inputs.rows(); i++)
    {
      for (int j = 0; j < inputs.cols(); j++)
	{
	  inputs(i,j) -= _weightsOut[layer][i][j].sum();
	}
    }

  d.addPatch(inputs);

  Matrix outputs = inputs;

  for (int i = 0; i < inputs.rows(); i++)
    {
      for (int j = 0; j < inputs.cols(); j++)
	{
	  outputs(i,j) = tanh(outputs(i,j));
	}
    }


  d.addPatch(outputs);

  d.enableWindow();
  cout << "Biases "<<layer<<endl;




  //4 panes
  //upper left: biases
  //upper right: tanh(biases)
  //lower left: activationIn when all inputs are -1 (this is more analagous to what bias usually represents in a sigmoidal network)
  //lower right: tanh of the same

  cvWaitKey(0);
}

*/
