/// \author Ian Goodfellow
/// Pre-SDN stacked network, included for unit testing purposes
/// (StackedNetwork was not learning gabors, so we decided to check
/// it against the old code)


#pragma once
#include <vector>

#include <Eigen/Core>
#include <Eigen/Array>

#include "sdn/base.h"

sdn::Scalar randomReal(sdn::Scalar lower, sdn::Scalar upper);


class StackedEncoder
{
  public:

  friend class FastEncoder;

  //Constructs an autoencoder from file
  StackedEncoder(const char * filepath) //: _L1weightsDisplay(std::string("Layer 1 Weights"),1,1,1,1,1)
    {
      loadNN(filepath);
    }

  //Constructs an autoencoder with a single hidden layer
  StackedEncoder(int inputWidth, int outputWidth,			\
		 sdn::Scalar learningRate, sdn::Scalar biasLearningRate,		\
		   sdn::Scalar weightDecay, sdn::Scalar desiredSparsityMin,	\
		   sdn::Scalar desiredSparsityMax,				\
		   sdn::Scalar sparsityPenaltyMultiplier);

  //Given an autoencoder with n encoder layers, constructs an autoencoder with
  //n+1 encoder layers. outputWidth is the number of hidden units per side
  //of the final hidden layer. patchStep and numStep are assigned to the
  //next layer down.

  StackedEncoder(int outputWidth, int patchStep, int numStep,
		 StackedEncoder & smaller, bool learnAll, sdn::Scalar weightDecay,
		 sdn::Scalar desiredSparsityMin, sdn::Scalar desiredSparsityMax,
		 sdn::Scalar sparsityPenaltyMultiplier);
  
  int inputWidth() { return _inputWidth[0]; }


  void forwardProp(const sdn::Matrix & patch, sdn::Scalar * errorEstimate = NULL,
		   std::vector<std::vector< sdn::Scalar> > * sparsityStats = NULL,
		   std::vector<sdn::Scalar> * sparsityPenaltyEstimate = NULL, int startLayer = 0, int endLayer = -1);//sparsityPenaltyEstimate is not actually calcualted right now
 

  void debugSetConstantWeights(sdn::Scalar d)
  {
    for (unsigned i = 0; i < _weightsOut.size(); i++)
      for (unsigned j = 0; j < _weightsOut[i].size(); j++)
	for (unsigned k = 0; k < _weightsOut[i][j].size(); k++)
	  _weightsOut[i][j][k].setConstant(d);
  }

    void backProp();
    void updateWeights();

    const sdn::Matrix & getCode(int i) const
      { assert(i > 0 && i <= _nEncoderLayers); return _activationOut[i]; }
    const sdn::Matrix & getReconstruction(int i = 0) const
      { assert(i >= 0 && i < _nEncoderLayers); return _activationOut[_nTotLayers-1-i]; }

    sdn::Vector  getCodeVec(int i) const
      { return sdn::matrixToVector(getCode(i)); }


    const sdn::Matrix & outputValues() const
      { return _activationOut[_nTotLayers-1]; }

    void printDebugStuff();

    void debugSetAllWeights(sdn::Scalar c);
    const sdn::Matrix & debugDelMe1() { return _deltaMulti[2][0][0]; }
    const sdn::Matrix & debugDelMe2() { return _delta[1]; }
    //void debugCheckWeights(const AutoEncoder & o);
    //void debugCopyWeights(const AutoEncoder & o);
    void showLayer1Weights(bool allowRescalingColorMap , bool allowScalingUp );
    void saveLayer1WeightsImage(const char * filepath);
    void showBiases(int layer);

    void saveNN(const char *filename); 
    void loadNN(const char *filename);

    int numEncoderLayers() const { return _nEncoderLayers; }

    const std::vector<std::vector<sdn::Matrix> > & getEncoderWeights(unsigned i) const
      {
 	SDN_ASSERT(i < numEncoderLayers());
	return _weightsOut[i];
      }

    const sdn::Matrix & getEncoderBiases(unsigned i) const
      {
	SDN_ASSERT(i < numEncoderLayers());
	return _biasOut[i];
      }

    const std::vector<std::vector<sdn::Matrix> > & getDecoderWeights(unsigned i) const
      {
	SDN_ASSERT(i < numEncoderLayers());
	return _weightsOut[ numEncoderLayers() * 2 - 1 -i ];
      }

    const sdn::Matrix & getDecoderBiases(unsigned i) const
      {
	SDN_ASSERT(i < numEncoderLayers());
	return _biasOut[ numEncoderLayers() * 2 - 1 - i];
      }

    const std::vector<sdn::Matrix> & getActivationOut() const 
      {
	return _activationOut;}


 private:
    void updateLayer1Weights(bool allowRescalingColorMap = true, bool allowScalingUp = true);


  private:

  int _nEncoderLayers;                      // If have I -> L1 -> L2 -> L1 -> O, then this is 2.  
  // If have I -> L1 -> L2 -> L3 -> ... 0, then this is 3
  int _nTotLayers;                          // In the two examples above, would be respectively 5 and 7.
  // _nTotlayers =  2*_nEncoderLayers+1. 

  
  // Parameters specifying the dimensions of each layer 
  std::vector<int> _inputWidth, _patchStep, _numStep, _outputWidth; // vectors of size _nEncoderLayers+1.
    // Meanings of these parameters: 
    // _inputWidth[i]: Layer Li's activations is a function of a _inputWidth[i]x_inputWidth[i] patch of the lower level. 
    // _patchStep[i]: The input to each layer Li group is obtained by a square array Li-1 groups.  The next group over in Li steps
    //                    by patchStep[i]
    // _numStep[i]: The groups of layer Li form a numStep[i]xnumStep[i] array.
    // _outputWidth[i]: Each group of layer Li has size outputWidth[i]xoutputWidth[i]. Each of these outputWidth[i] units
    //                 connect to the same set of Li-1 units.  So, the total dimension (number of real numbers) of Li's 
    //                 output is numStep[i]^2 * outputWidth[i]^2

    //This should satisfy
    // (total width of layer Li =) _outputWidth[i-1]*_numStep[i-1] = (_numStep[i]-1)*_patchStep[i]+_inputWidth[i] 
    //For element 0 of the four arrays above, we must always have 
    //    _inputWidth = _patchStep = _outputWidth =  image size,
    //    and _numStep = 1

    // Weights
  std::vector<std::vector<std::vector<sdn::Matrix> > > _weightsOut;  // Weights for I->L1, ..., Ln->Ln-1, ..., L1->I. Thus there are _nTotLayers-1 of them.
  // _weightsOut[i][j][k] is the set of weights going from layer i to hidden unit (j,k) within its block in layer i+1 if layer i+1 is an encoding layer,
  //or if i+1 is a decoding layer, it is the set of weights going from all
  //hidden units in a block of layer i to pixel (j,k) within that block's output
  std::vector<sdn::Matrix> _biasOut;
  // _biasOut[i][j][k] is the weight connecting an imaginary bias unit in layer i
  // to hidden unit (j,k) in layer i+1
  
  // temporary datastructures for the current input 
  std::vector<sdn::Matrix> _activationIn, _activationOut;   // activationIn is the input to a node; apply the (say, tanh) activation function to this, and you get activationOut. Note: _activationIn[0] is unused, and the decoding layers use _activationInMulti instead
  std::vector<std::vector<std::vector<sdn::Matrix> > > _activationInMulti; //like _activationIn, but accounting for the fact that multiple networks feed to each unit.
  //_activationInMulti[i][j][k] is the presynaptic input to units reconstructed by block (j,k) of layer i-1, thus the first index that is actually used is
  //_nEncoderLayers+1.  For example, suppose we have a network I(0) -> L1(1) -> L2(2) -> L3(3) -> L2(4) -> L1(5) -> O(6).  
  // Consider layer 5, which is a decoding layer.  Each unit in L1(5) has an associated unit in L1(1), which was an encoding unit.
  // That unit is part of a larger group, all of which have the same input from I(0).  All the corresponding units in layer L1(5)
  // we now consider to be in the same group as well; and, they are all going to try to reconstruct the same sets of units in O(6). 
  // So, _activatinInMulti[6][j][k] would refer to the presynaptic reconstruction of layer 6 due to group [j][k] in layer 5. 
  

  
//the mapping between _deltas and units is the same as _activationIn and units
  //(the reason _activationIn is so awkward is that we need to store more detail
  //about it to compute the deltas)
  std::vector<sdn::Matrix> _delta; //only _delta[nTotLayers-1] down to _delta[_nEncoderLayers-1] is used if we are just learning the central set of weights
  std::vector<std::vector<std::vector<sdn::Matrix> > > _deltaMulti;
  //(For layers > _nEncoderLayers, _delta is actually the deltas for imaginary
  //averaging units, while _deltaMulti is the deltas  for the sigmoidal
  //units)

  // learning parameters
  sdn::Scalar _learningRate, _biasLearningRate;
  //vectors of size _nEncoderLayers+1, with element 0 unused
  std::vector<sdn::Scalar> _weightDecay, _desiredSparsityMin, _desiredSparsityMax,
_sparsityPenaltyMultiplier; 
  
 bool _learnAll; //if true, adjusts all weights using backprop rather
  //than just the central layer.
 bool _deepBackprop; //true = default behavior where the objective is reconstruction of the input. false = objective is reconstruction of 2nd deepest hidden layer.



 sdn::Scalar _errorEstimate;


 std::vector<sdn::Matrix> _sparsityEstimate; //element 0 is unused. element i is for encoder layer i
 std::vector<sdn::Scalar> _sparsityPenaltyEstimate; //element 0 is unused, and if _learnAll is false, the only used element is _nEncoderLayers

 //PatchDisplay _L1weightsDisplay;
};
