#include "sdn/old.h"
#include "derivativeChecker.h"
#include <Eigen/Core>
#include <Eigen/Array>
#include "gtest/gtest.h"

using namespace boost;
using namespace sdn;

TEST(StandardLayer, Constructor) {
  const StandardLayer sln();
}


class ErrorDerivativeChecker : public DerivativeChecker {

public:
  ErrorDerivativeChecker(shared_ptr<StorageNode> in, 
			 shared_ptr<StorageNode> out,
			 const Matrix& target) {
    sln_ = new StandardLayer();
    sln_->link(in,out);
    target_ = target;
    
  }

  ~ErrorDerivativeChecker() {
    delete sln_;
  }

  Matrix AnalyticalDerivative(const Matrix& x) {
    sln_->setOutput(x);
    sln_->presentTarget(target_);
    return sln_->getErrorDerivativeWrtOutput();
  }

  double EvaluateFunction(const Matrix& x) {
    sln_->setOutput(x);
    return sln_->calculateError(target_);
  }
  
  StandardLayer* sln_;
  Matrix target_;
};

TEST(StandardLayer, CalculateErrorDerivative) {
  Scalar epsilon = 1e-3;
  Scalar tolerance = 1e-3;
  int inputs = 5;
  int outputs = 7;
  StorageNode in(inputs);
  StorageNode out(outputs);
  Matrix target = Matrix::Random(outputs,1);

  ErrorDerivativeChecker edc(owned_ptr(&in), owned_ptr(&out), target);

  Matrix output = Matrix::Random(outputs,1);
  
  //std::cout << edc.NumericalDerivative(epsilon, output) << "\n";
  //std::cout << edc.AnalyticalDerivative(output) << "\n";
  ASSERT_LE(edc.CheckDerivative(epsilon, output), tolerance);
}

class GradientChecker : public DerivativeChecker {

public:
  GradientChecker(shared_ptr<StorageNode> in, 
		  shared_ptr<StorageNode> out,
		  const Matrix& input,
		  const Matrix& target) {
    sln_.link(in, out);
    input_ = input;
    target_ = target;
      }

  virtual ~GradientChecker() {
    // do nothing
  }

  Matrix AnalyticalDerivative(const Matrix& x) {
    sln_.setWeights(x);
    sln_.setAllBiases(0); /// \todo-- also check the bias gradient
    sln_.forwardProp(input_);
    sln_.presentTarget(target_);
    sln_.backProp();
    return sln_.errorDerivativeWrtWeights();
  }

  double EvaluateFunction(const Matrix& x) {
    sln_.setWeights(x);
    sln_.setAllBiases(0); /// \todo-- also check the bias gradient
    sln_.forwardProp(input_);
 
    return sln_.calculateError(target_);
  }
  
  StandardLayer sln_;
  Matrix input_;
  Matrix target_;
};


TEST(StandardLayer, GradientSinglePattern) {
  Scalar epsilon = 1e-3;
  Scalar tolerance = 1e-3;

  int inputs = 5;
  int outputs = 7;
  int patterns = 1;

  Matrix input = Matrix::Random(inputs, patterns);
  Matrix target = Matrix::Random(outputs,patterns);
  Matrix weights = Matrix::Random(outputs,inputs);

  StorageNode in(inputs);
  StorageNode out(outputs);
  GradientChecker gc(owned_ptr(&in), owned_ptr(&out), input, target); 
  ASSERT_LE(gc.CheckDerivative(epsilon, weights), tolerance);

}


TEST(StandardLayer, GradientManyPatterns) {
  Scalar epsilon = 1e-3;
  Scalar tolerance = 1e-3;

  int inputs = 5;
  int outputs = 7;
  int patterns = 10;

  Matrix input = Matrix::Random(inputs, patterns);
  Matrix target = Matrix::Random(outputs,patterns);
  Matrix weights = Matrix::Random(outputs,inputs);

  StorageNode in(inputs);
  StorageNode out(outputs);
  
  GradientChecker gc(owned_ptr(&in), owned_ptr(&out), input, target); 
 
  ASSERT_LE(gc.CheckDerivative(epsilon, weights), tolerance);

}


TEST(StandardLayer, GradientDescent) {
  Scalar learning_rate = 1e-1;
  Scalar tolerance = 1e-2;
  int iterations = 1000;

  int inputs = 5;
  int outputs = 7;
  int patterns = 3;

  StorageNode in(inputs);
  StorageNode out(outputs);
  StandardLayer sln;
  sln.link(owned_ptr(&in), owned_ptr(&out));

  Matrix input = Matrix::Random(inputs, patterns);
  Matrix target = Matrix::Random(outputs,patterns);
  
  for(int i = 0; i < iterations; i++) {
    sln.forwardProp(input);
    sln.presentTarget(target);
    sln.backProp();
    sln.gradientDescentUpdate(learning_rate);
  }

  ASSERT_LE(sln.calculateError(target), tolerance);  
}

class ErrorDerivativeAtInputChecker : public DerivativeChecker {

public:
  ErrorDerivativeAtInputChecker(boost::shared_ptr<StorageNode> in, 
				boost::shared_ptr<StorageNode> out,
				const Matrix& target) {
    sln_ = new StandardLayer();
    sln_->link(in,out);
    target_ = target;
    
  }

  ~ErrorDerivativeAtInputChecker() {
    delete sln_;
  }

  Matrix AnalyticalDerivative(const Matrix& x) {
    sln_->forwardProp(x);
    sln_->presentTarget(target_);
    sln_->backProp();
    return sln_->getErrorDerivativeWrtInput();
  }

  double EvaluateFunction(const Matrix& x) {
    sln_->forwardProp(x);
    return sln_->calculateError(target_);
  }
  
  StandardLayer* sln_;
  Matrix target_;
};

TEST(StandardLayer, ErrorDerivativeAtInput) {
  Scalar epsilon = 1e-3;
  Scalar tolerance = 1e-3;

  int inputs = 5;
  int outputs = 7;
  int patterns = 3;

  Matrix input = Matrix::Random(inputs, patterns);
  Matrix target = Matrix::Random(outputs,patterns);
  
  StorageNode in(inputs);
  StorageNode out(outputs);
  ErrorDerivativeAtInputChecker edc(owned_ptr(&in), owned_ptr(&out), target);
  
  
  ASSERT_LE(edc.CheckDerivative(epsilon,input), tolerance);  
}

TEST(StandardLayer, MultiLayerTest) {
  Scalar epsilon = 1e-1;
  Scalar tolerance = 1e-1; 
  int iterations = 1000;

  int inputs = 5;
  int hidden = 10;
  int outputs = 7;
  int patterns = 5;

  Matrix input = Matrix::Random(inputs, patterns);
  Matrix target = Matrix::Random(outputs,patterns);

  StorageNode in(inputs);
  StorageNode hid(hidden);
  StorageNode out(outputs);

  StandardLayer sln1;
  sln1.link(owned_ptr(&in), owned_ptr(&hid));
  StandardLayer sln2;
  sln2.link(owned_ptr(&hid), owned_ptr(&out));

  for(int i = 0; i < iterations; i++) {
    sln1.forwardProp(input);
    sln2.forwardProp();
    sln2.presentTarget(target);
    sln2.backProp();
    sln1.backProp();
    
    sln1.gradientDescentUpdate(epsilon);
    sln2.gradientDescentUpdate(epsilon);
    //std::cout << sln2.CalculateError(target) << "\n";
  }

  ASSERT_LE(sln2.calculateError(target), tolerance);  
}

TEST(StandardLayer, TransferLearningTest) {
  Scalar epsilon = 1e-1;
  Scalar tolerance = 1e-1; 
  int iterations = 1000;

  int inputs = 5;
  int hidden = 10;
  int outputs1 = 4;
  int outputs2 = 2;
  int patterns1 = 4;
  int patterns2 = 2;

  Matrix input1 = Matrix::Random(inputs, patterns1);
  Matrix target1 = Matrix::Random(outputs1,patterns1);

  Matrix input2 = Matrix::Random(inputs, patterns2);
  Matrix target2 = Matrix::Random(outputs2,patterns2);

  StorageNode in(inputs);
  StorageNode hid(hidden);
  StorageNode out1(outputs1);
  StorageNode out2(outputs2);

  StandardLayer sln1;
  sln1.link(owned_ptr(&in), owned_ptr(&hid));
  StandardLayer sln2;
  sln2.link(owned_ptr(&hid), owned_ptr(&out1));
  StandardLayer sln3;
  sln3.link(owned_ptr(&hid), owned_ptr(&out2));

  for(int i = 0; i < iterations; i++) {
    
    sln1.forwardProp(input1);
    sln2.forwardProp();
    sln2.presentTarget(target1);
    sln2.backProp();
    sln1.backProp();
    
    sln1.gradientDescentUpdate(epsilon);
    sln2.gradientDescentUpdate(epsilon);
    //std::cout << sln2.CalculateError(target1) << "\n";

    sln1.forwardProp(input2);
    sln3.forwardProp();
    sln3.presentTarget(target2);
    sln3.backProp();
    sln1.backProp();
    
    sln1.gradientDescentUpdate(epsilon);
    sln3.gradientDescentUpdate(epsilon);
    //std::cout << sln3.CalculateError(target2) << "\n";
  }
			
  EXPECT_LE(sln3.calculateError(target2), tolerance);
  EXPECT_LE(sln2.calculateError(target1), tolerance);  
}
